/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portions Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: daemon.c
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"

#include <linux/smp_lock.h>

#include "glock.h"
#include "inode.h"
#include "recovery.h"




/**
 * gfs_glockd - Glock kernel thread
 * @data: Pointer to GFS superblock
 *
 * A kernel thread which harvests gfs_bhdata_t structures laying
 * around. It also demoted glocks which are no longer being used.
 *
 * Returns: Process exit status
 */

int gfs_glockd(void *data)
{
  gfs_sbd_t *sdp = (gfs_sbd_t *)data;
  struct fs_struct *fs;
  unsigned long flags;

  ENTER(GFN_GLOCKD);


  GFS_ASSERT(sdp, );


  /*  Set up the thread  */

  lock_kernel();

  exit_mm(current);

  current->session = 1;
  current->pgrp = 1;
  current->tty = NULL;
  current->tty_old_pgrp = 0;

  exit_files(current);
  current->files = init_task.files;
  atomic_inc(&current->files->count);
	                
  exit_fs(current);       /* current->fs->count--; */
  fs = init_task.fs;
  current->fs = fs;
  atomic_inc(&fs->count);
	
  spin_lock_irqsave(&current->sigmask_lock, flags);
  flush_signals(current);
  flush_signal_handlers(current);
  sigfillset(&current->blocked);
  spin_unlock_irqrestore(&current->sigmask_lock, flags);

  osi_sprintf(current->comm, "gfs_glockd");

  sdp->sd_glockd_process = current;

  osi_sema_unlock(&sdp->sd_glockd_sema);


  /*
   * sdp->sd_glockd_ctrl is the number of seconds between each
   * run. When set to zero it causes the thread to exit.
   */

  while (osi_atomic_read(&sdp->sd_glockd_ctrl))
  {
    osi_sleep(osi_atomic_read(&sdp->sd_glockd_ctrl));

    IFDB(GFN_GLOCKD,
	 db_printf("\n\n\n  gfs_glockd wakes up\n");
	 db_printf("Entering gfs_glockd() core\n"););

    /*
     * Run the arch independent part of gfs_glockd
     */

    gfs_glockd_scan(sdp, sdp->sd_gl_heldtime);

    IFDB(GFN_GLOCKD,
	 db_printf("Exiting gfs_glockd() core\n"););
  }


  /*
   * Signal to umount that we have finished
   */

  osi_sema_unlock(&sdp->sd_glockd_sema);

  unlock_kernel();


  EXIT(GFN_GLOCKD);

  return 0;
}


/**
 * gfs_inoded - Inode harvesting kernel thread
 * @data: Pointer to GFS superblock
 *
 * A kernel thread that wakes up every so often and gets rid of all 
 * gfs_bhdata_t structures laying around.
 *
 * Returns: Process exit status
 */

int gfs_inoded(void *data)
{
  gfs_sbd_t *sdp = (gfs_sbd_t *)data;
  struct fs_struct *fs;
  unsigned long flags;
  int error;

  ENTER(GFN_INODED);


  GFS_ASSERT(sdp, );


  /*  Set up the thread  */

  lock_kernel();

  exit_mm(current);

  current->session = 1;
  current->pgrp = 1;
  current->tty = NULL;
  current->tty_old_pgrp = 0;

  exit_files(current);
  current->files = init_task.files;
  atomic_inc(&current->files->count);
	                
  exit_fs(current);       /* current->fs->count--; */
  fs = init_task.fs;
  current->fs = fs;
  atomic_inc(&fs->count);
	
  spin_lock_irqsave(&current->sigmask_lock, flags);
  flush_signals(current);
  flush_signal_handlers(current);
  sigfillset(&current->blocked);
  spin_unlock_irqrestore(&current->sigmask_lock, flags);

  osi_sprintf(current->comm, "gfs_inoded");

  sdp->sd_inoded_process = current;

  osi_sema_unlock(&sdp->sd_inoded_sema);


  /*
   * sdp->sd_inoded_ctrl is the number of seconds between each
   * run. When set to zero it causes the thread to exit.
   */

  while (osi_atomic_read(&sdp->sd_inoded_ctrl))
  {
    osi_sleep(osi_atomic_read(&sdp->sd_inoded_ctrl));

    IFDB(GFN_INODED,
	 db_printf("\n\n\n  gfs_inoded wakes up\n");
	 db_printf("Entering gfs_inoded() core\n"););

    /*
     * Run the arch independent part of gfs_inoded
     */

    error = gfs_dealloc_inodes(sdp);
    if (error)
      con_printf("gfs_inoded:  error = %d\n", error);

    IFDB(GFN_INODED,
	 db_printf("Exiting gfs_inoded() core\n"););
  }


  /*
   * Signal to umount that we have finished
   */

  osi_sema_unlock(&sdp->sd_inoded_sema);

  unlock_kernel();


  EXIT(GFN_INODED);

  return 0;
}


/**
 * gfs_recoverd - Recovery kernel thread
 * @data: Pointer to GFS superblock
 *
 * A kernel thread that frees any inodes that have a zero usage count and
 * no VFS inode pointer.
 *
 * Returns: Process exit status
 */

int gfs_recoverd(void *data)
{
  gfs_sbd_t *sdp = (gfs_sbd_t *)data;
  struct fs_struct *fs;
  unsigned long flags;
  int error;

  ENTER(GFN_RECOVERD);


  GFS_ASSERT(sdp, );


  /*  Set up the thread  */

  lock_kernel();

  exit_mm(current);

  current->session = 1;
  current->pgrp = 1;
  current->tty = NULL;
  current->tty_old_pgrp = 0;

  exit_files(current);
  current->files = init_task.files;
  atomic_inc(&current->files->count);
	                
  exit_fs(current);       /* current->fs->count--; */
  fs = init_task.fs;
  current->fs = fs;
  atomic_inc(&fs->count);
	
  spin_lock_irqsave(&current->sigmask_lock, flags);
  flush_signals(current);
  flush_signal_handlers(current);
  sigfillset(&current->blocked);
  spin_unlock_irqrestore(&current->sigmask_lock, flags);

  osi_sprintf(current->comm, "gfs_recoverd");

  sdp->sd_recoverd_process = current;

  osi_sema_unlock(&sdp->sd_recoverd_sema);


  /*
   * sdp->sd_recoverd_ctrl is the number of seconds between each
   * run. When set to zero it causes the thread to exit.
   */

  while (osi_atomic_read(&sdp->sd_recoverd_ctrl))
  {
    osi_sleep(osi_atomic_read(&sdp->sd_recoverd_ctrl));

    IFDB(GFN_RECOVERD,
	 db_printf("\n\n\n  gfs_recoverd wakes up\n");
	 db_printf("Entering gfs_recoverd() core\n"););

    /*
     * Run the arch independent part of gfs_recoverd
     */

    error = gfs_check_journals(sdp);
    if (error)
      con_printf("gfs_recoverd:  error = %d\n", error);

    IFDB(GFN_RECOVERD,
	 db_printf("Exiting gfs_recoverd() core\n"););
  }


  osi_sema_unlock(&sdp->sd_recoverd_sema);

  unlock_kernel();


  EXIT(GFN_RECOVERD);

  return 0;
}


