/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portions Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth Preslan <kpreslan@sistina.com>
** 
**    Module name:  dio.c
**
**    Description:
**      This module implements a bunch of routines that do I/O through the
**        buffer cache.
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"

#include <linux/blkdev.h>

#include "dio.h"
#include "gfs_linux.h"
#include "glock.h"
#include "inode.h"
#include "log.h"
#include "trans.h"





/**
 * rm_bh_internal - Remove a buffer from the list maintained for invalidation
 * @bh: The buffer head to remove
 *
 */

STATIC void rm_bh_internal(osi_buf_t *bh)
{
  gfs_bufdata_t *bd = BH_BUFDATA(bh);
  gfs_glock_t *gl;

  GFS_ASSERT(bd, );
  gl = bd->bd_gl;

  osi_spin_lock(&gl->gl_arch.gl_buflock);

  bh->b_rm_fn = NULL;
  bh->b_pdata = NULL;
  osi_list_del(&bd->bd_arch.bd_list);

  if (osi_list_empty(&gl->gl_arch.gl_bufs))
    osi_set_bit(GLF_DEMOTEME, &gl->gl_flags);

  osi_spin_unlock(&gl->gl_arch.gl_buflock);

  gfs_put_glstruct(gl);
  gfs_free_bufdata(bd);
}


/**
 * gfs_remove_bh - Remove a buffer from the list maintainted for invalidation
 * @bh: The buffer head to be removed
 *
 * This function is the one the buffer cache sees.  It checks to make sure
 * the buffer cache isn't throwing away an active buffer and then calls
 * rm_bh_internal().
 *
 */

void gfs_remove_bh(osi_buf_t *bh)
{
  GFS_ASSERT(!bh->b_count, );
  rm_bh_internal(bh);
}


/**
 * gfs_inval_bh - Invalidate all buffers associated with a glock
 * @sdp: the filesystem
 * @gl: the glock
 *
 */

void gfs_inval_bh(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  osi_list_head_t *tmp, *head;
  gfs_bufdata_t *bd;
  osi_buf_t *bh;

  ENTER(GFN_INVAL_BH);


 restart:
  osi_spin_lock(&gl->gl_arch.gl_buflock);

  tmp = head = &gl->gl_arch.gl_bufs;
  for (;;)
  {
    tmp = tmp->next;
    if (tmp == head)
      break;
    bd = osi_list_entry(tmp, gfs_bufdata_t, bd_arch.bd_list);
    bh = bd->bd_bh;

    GFS_ASSERT(!buffer_dirty(bh), );

    if (buffer_locked(bh))
    {
      osi_spin_unlock(&gl->gl_arch.gl_buflock);
      wait_on_buffer(bh);
      goto restart;
    }

    mark_buffer_uptodate(bh, FALSE);
  }

  osi_spin_unlock(&gl->gl_arch.gl_buflock);


  EXIT(GFN_INVAL_BH);
}


/**
 * gfs_sync_bh - Sync all buffers associated with a glock
 * @sdp: the filesystem
 * @gl: The glock
 * @wait: Non-zero means wait for completion of I/O
 *
 */

void gfs_sync_bh(gfs_sbd_t *sdp, gfs_glock_t *gl, int wait)
{
  osi_list_head_t *tmp, *head;
  gfs_bufdata_t *bd;
  osi_buf_t *bh;
  int pass = 0;

  ENTER(GFN_SYNC_BH);


 restart:
  osi_spin_lock(&gl->gl_arch.gl_buflock);

  tmp = head = &gl->gl_arch.gl_bufs;
  for(;;)
  {
    tmp = tmp->next;
    if (tmp == head)
      break;
    bd = osi_list_entry(tmp, gfs_bufdata_t, bd_arch.bd_list);
    bh = bd->bd_bh;

    if (buffer_dirty(bh) && !buffer_locked(bh))
    {
      osi_spin_unlock(&gl->gl_arch.gl_buflock);
      ll_rw_block(WRITE, 1, &bh);
      goto restart;
    }

    if (pass && buffer_locked(bh))
    {
      osi_spin_unlock(&gl->gl_arch.gl_buflock);
      wait_on_buffer(bh);
      goto restart;
    }
  }

  osi_spin_unlock(&gl->gl_arch.gl_buflock);


  if (wait && !(pass++))
    goto restart;


  EXIT(GFN_SYNC_BH);
}


/**
 * gfs_free_bh - Free all buffers associated with a glock
 * @sdp: the filesystem
 * @gl: The glock
 *
 * Remove all gfs_linux_bhdata_t structures on the list.  Be careful to avoid
 * races with other processes removeing them naturally through
 * gfs_remove_bh().  We do this by doing a getblk() on the buffer to
 * increment the count on the buffer.  (bh->b_count++ isn't good enough
 * to stop the race)
 *
 * Only called on unmount.   
 *
 */

void gfs_free_bh(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  gfs_bufdata_t *bd;
  osi_buf_t *bh;

  ENTER(GFN_FREE_BH);
  

 restart:
  osi_spin_lock(&gl->gl_arch.gl_buflock);

  while (!osi_list_empty(&gl->gl_arch.gl_bufs))
  {
    bd = osi_list_entry(gl->gl_arch.gl_bufs.next, gfs_bufdata_t, bd_arch.bd_list);
    GFS_ASSERT(bd, );

    bh = bd->bd_bh;
    GFS_ASSERT(bh, );

    bh->b_rm_fn = NULL;
    bh->b_pdata = NULL;
    osi_list_del(&bd->bd_arch.bd_list);
    gfs_put_glstruct(gl);

    osi_spin_unlock(&gl->gl_arch.gl_buflock);
    gfs_free_bufdata(bd);

    goto restart;
  }

  osi_spin_unlock(&gl->gl_arch.gl_buflock);


  EXIT(GFN_FREE_BH);
}


/**
 * clear_from_ail - make sure a buffer isn't in the AIL
 * @sdp: the filesystem
 * @bh: the buffer
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int clear_from_ail(gfs_sbd_t *sdp, osi_buf_t *bh)
{
  gfs_bufdata_t *bd = BH_BUFDATA(bh);
  int error = 0;

  ENTER(GFN_CLEAR_FROM_AIL);

  if (!bd)
    goto out;

  if (gfs_is_pinned(bh))
    gfs_log_flush(sdp, bd->bd_gl);

  if (bd->bd_arch.bd_ail_tr)
    error = gfs_dwrite(sdp, bh, DW_START | DW_WAIT);

 out:
  EXIT(GFN_CLEAR_FROM_AIL);

  return error;
}


/**
 * gfs_dgetblk_arch - Get a block
 * @sdp: The GFS superblock
 * @blkno: The block number
 * @gl: The glock associated with this block
 *
 * Returns: The buffer on success, or NULL
 */

osi_buf_t *gfs_dgetblk_arch(gfs_sbd_t *sdp, uint64 blkno, gfs_glock_t *gl)
{
  osi_buf_t *bh;
  gfs_bufdata_t *bd;

  ENTER(GFN_DGETBLK_ARCH);


  bh = getblk(sdp->sd_vsb->s_dev, blkno, sdp->sd_sb.sb_bsize);
  if (!bh)
    goto fail;


  if (gl)
  {
    osi_spin_lock(&gl->gl_arch.gl_buflock);


    /* buffer is from a deleted file */

    if ((bh->b_rm_fn == gfs_remove_bh) && (gl != BH_BUFDATA(bh)->bd_gl))
    {
      osi_spin_unlock(&gl->gl_arch.gl_buflock);

      if (clear_from_ail(sdp, bh) != 0)
	goto fail_brelse;
      rm_bh_internal(bh);

      osi_spin_lock(&gl->gl_arch.gl_buflock);
    }


    /* insert new bh into this lock's bufs list */

    if (bh->b_rm_fn != gfs_remove_bh) 
    {
      GFS_ASSERT(!bh->b_rm_fn, );

      osi_spin_unlock(&gl->gl_arch.gl_buflock);


      bd = gfs_alloc_bufdata();
      if (!bd)     
	goto fail_brelse;

      gfs_memset(bd, 0, sizeof(gfs_bufdata_t));

      bd->bd_bh = bh;
      bd->bd_gl = gl;

      osi_list_head_init(&bd->bd_new_list);
      osi_list_head_init(&bd->bd_incore_list);

      osi_sema_init(&bd->bd_lock);


      osi_spin_lock(&gl->gl_arch.gl_buflock);

      gfs_hold_glstruct(gl);
      osi_list_add(&bd->bd_arch.bd_list, &gl->gl_arch.gl_bufs);
      osi_clear_bit(GLF_DEMOTEME, &gl->gl_flags);

      bh->b_pdata = bd;
      bh->b_rm_fn = gfs_remove_bh;

      if (!sdp->sd_args.ar_localcaching)
	mark_buffer_uptodate(bh, FALSE);
    }


    osi_spin_unlock(&gl->gl_arch.gl_buflock);
  }


  EXIT(GFN_DGETBLK_ARCH);

  return bh;



 fail_brelse:
  brelse(bh);

 fail:
  EXIT(GFN_DGETBLK_ARCH);
  return NULL;
}


/**
 * gfs_dread_arch - Read a block from disk
 * @sdp: The GFS superblock
 * @blkno: The block number
 * @gl: The glock covering the block
 * @flags: flags to gfs_dreread()
 *
 * Returns: The buffer on success, NULL on failure
 */

osi_buf_t *gfs_dread_arch(gfs_sbd_t *sdp, uint64 blkno, gfs_glock_t *gl, int flags)
{
  osi_buf_t *bh;

  ENTER(GFN_DREAD_ARCH);


  bh = gfs_dgetblk_arch(sdp, blkno, gl);
  if (!bh)
  {
    EXIT(GFN_DREAD_ARCH);
    return bh;
  }

  if (gfs_dreread(sdp, bh, flags) != 0)
  {
    gfs_drelse(bh);
    bh = NULL;
  }


  EXIT(GFN_DREAD_ARCH);

  return bh;
}


/**
 * gfs_prep_new_buffer - Mark a new buffer we just gfs_dgetblk()ed uptodate
 * @bh: the buffer
 *
 */

void gfs_prep_new_buffer(osi_buf_t *bh)
{
  ENTER(GFN_PREP_NEW_BUFFER);

  lock_buffer(bh);

  mark_buffer_clean(bh);
  mark_buffer_uptodate(bh, TRUE);

  unlock_buffer(bh);

  EXIT(GFN_PREP_NEW_BUFFER);
}


/**
 * try_readahead - Do readahead on the device
 * @sdp: the filesystem
 * @bh: the buffer we're about to sleep on
 *
 */

STATIC void try_readahead(gfs_sbd_t *sdp, osi_buf_t *bh)
{
  osi_buf_t *tmp_bh;
  unsigned int major, minor;
  uint64 size, blk;
  unsigned int x;

  ENTER(GFN_TRY_READAHEAD);


  /*  All buffers should be on the same device, right?
      The device shouldn't shrink much while this function is running, right?  */

  major = MAJOR(bh->b_dev);
  minor = MINOR(bh->b_dev);

  if (!blk_size[major])
  {
    EXIT(GFN_TRY_READAHEAD);
    return;
  }

  if (sdp->sd_sb.sb_bsize_shift >= 10)
    size = blk_size[major][minor] >> (sdp->sd_sb.sb_bsize_shift - 10);
  else
    size = blk_size[major][minor] << (10 - sdp->sd_sb.sb_bsize_shift);


  for (x = 0; buffer_locked(bh) && x < sdp->sd_args.ar_buf_readahead; x++)
  {
    blk = BH_BLKNO(bh) + x + 1;

    if (blk >= size)
      break;

    tmp_bh = getblk(sdp->sd_vsb->s_dev, blk, sdp->sd_sb.sb_bsize);
    if (!tmp_bh)
      break;

    if (!buffer_uptodate(tmp_bh) && !buffer_locked(tmp_bh))
      ll_rw_block(READA, 1, &tmp_bh);

    brelse(tmp_bh);
  }


  EXIT(GFN_TRY_READAHEAD);
}


/**
 * gfs_dreread - Reread a block from disk
 * @sdp: the filesystem
 * @bh: The block to read
 * @flags: Flags that control the read
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_dreread(gfs_sbd_t *sdp, osi_buf_t *bh, int flags)
{
  int error = 0;

  ENTER(GFN_DREREAD);


  if (flags & DR_FORCE)
  {
    error = clear_from_ail(sdp, bh);
    if (error)
      goto out;
    if (!sdp->sd_args.ar_localcaching)
      mark_buffer_uptodate(bh, FALSE);
  }

  if ((flags & DR_START) && !buffer_uptodate(bh) && !buffer_locked(bh))
    ll_rw_block(READ, 1, &bh);

  if (flags & DR_WAIT)
  {
    /*  Do readahead if it's convenient and allowed  */

    if (buffer_locked(bh) &&
	sdp->sd_args.ar_localcaching &&
	sdp->sd_args.ar_buf_readahead)
      try_readahead(sdp, bh);

    wait_on_buffer(bh);

    if (!buffer_uptodate(bh))
    {
      con_printf("GFS:  read error on block %"FMT_64"u\n", BH_BLKNO(bh));
      error = -EIO;

      GFS_ASSERT(0, con_printf("Panicking because of read error on block %"FMT_64"u\n", BH_BLKNO(bh)););
    }
  }


 out:
  EXIT(GFN_DREREAD);

  return error;
}


/**
 * gfs_dwrite - Write a buffer
 * @sdp: the filesystem
 * @bh: The buffer to write
 * @flags: The type of write operation to do
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_dwrite(gfs_sbd_t *sdp, osi_buf_t *bh, int flags)
{
  gfs_bufdata_t *bd;
  gfs_glock_t *gl;
  int error = 0;

  ENTER(GFN_DWRITE);

  GFS_ASSERT(buffer_uptodate(bh), );


  if (flags & DW_DIRTY)
  {
    bd = BH_BUFDATA(bh);
    if (bd)
    {
      gl = bd->bd_gl;
      GFS_ASSERT(osi_atomic_read(&gl->gl_locked) &&
		 (gl->gl_pid == osi_pid() || gl->gl_pid == -1) &&
		 gl->gl_state == LM_ST_EXCLUSIVE, );
      osi_set_bit(GLF_DIRTY, &gl->gl_flags);
    }

    wait_on_buffer(bh);
    mark_buffer_dirty(bh, 0);
  }

  if ((flags & DW_START) && buffer_dirty(bh) && !buffer_locked(bh))
    ll_rw_block(WRITE, 1, &bh);

  if (flags & DW_WAIT)
  {
    wait_on_buffer(bh);

    if (!buffer_uptodate(bh) || buffer_dirty(bh))
    {
      con_printf("GFS:  write error on block %"FMT_64"u\n", BH_BLKNO(bh));
      error = -EIO;

      GFS_ASSERT(0, con_printf("Panicking because of write error on block %"FMT_64"u\n", BH_BLKNO(bh)););
    }
  }


  EXIT(GFN_DWRITE);

  return error;
}


/**
 * gfs_drelse_arch - Release a block
 * @bh: The block to release
 *
 */

void gfs_drelse_arch(osi_buf_t *bh)
{
  gfs_bufdata_t *bd = BH_BUFDATA(bh);

  ENTER(GFN_DRELSE_ARCH);

  if (bd && bd->bd_arch.bd_pinned)
    bh->b_count--;
  else
    brelse(bh);

  EXIT(GFN_DRELSE_ARCH);
}


/**
 * gfs_dforget_arch - Forget a data buffer
 * @bh: The buffer to release
 *
 */

void gfs_dforget_arch(osi_buf_t *bh)
{
  ENTER(GFN_DFORGET_ARCH);
  gfs_drelse_arch(bh);
  EXIT(GFN_DFORGET_ARCH);
}


/**
 * gfs_end_io_cb - End I/O Callback
 * @bh: The buffer
 * @uptodate: Flag indicating buffer state
 *
 */

STATIC void gfs_end_io_cb(osi_buf_t *bh, int uptodate)
{
  gfs_bufdata_t *bd;
  gfs_trans_t *tr;


  GFS_ASSERT(buffer_uptodate(bh), );
  GFS_ASSERT(uptodate,
	     con_printf("Panicking because of write error on block %"FMT_64"u\n", BH_BLKNO(bh)););

  bd = BH_BUFDATA(bh);
  GFS_ASSERT(bd, );
  GFS_ASSERT(!bd->bd_arch.bd_pinned, );

  tr = bd->bd_arch.bd_ail_tr;
  GFS_ASSERT(tr, );
  GFS_ASSERT(osi_atomic_read(&tr->tr_arch.tr_bufs), );


  /*  Sometimes we get callbacks at random times (who knows why?!?).
      If that happens, call the regular end_io callback, but don't
      take this buffer out of the AIL.  */

  if (buffer_dirty(bh))
  {
    bd->bd_arch.bd_old_cb(bh, uptodate);
    return;
  }


  osi_atomic_dec(&tr->tr_arch.tr_bufs);
  bd->bd_arch.bd_ail_tr = NULL;


  bh->b_end_io = bd->bd_arch.bd_old_cb;
  bh->b_end_io(bh, uptodate);
}


/**
 * gfs_dpin - Pin a metadata buffer in memory
 * @bh: The buffer to be pinned
 *
 */

void gfs_dpin(osi_buf_t *bh)
{
  gfs_bufdata_t *bd;

  ENTER(GFN_DPIN);

  GFS_ASSERT(buffer_uptodate(bh), );


  bd = BH_BUFDATA(bh);
  GFS_ASSERT(bd, );


  gfs_lock_buffer(bh);

  if (!bd->bd_arch.bd_pinned)
  {
    wait_on_buffer(bh);
    mark_buffer_clean(bh);
    bh->b_count++;

    if (bh->b_end_io != gfs_end_io_cb)
    {
      bd->bd_arch.bd_old_cb = bh->b_end_io;
      bh->b_end_io = gfs_end_io_cb;
    }
  }

  bd->bd_arch.bd_pinned++;

  gfs_unlock_buffer(bh);


  EXIT(GFN_DPIN);
}


/**
 * gfs_dunpin - Unpin a buffer
 * @bh: The buffer to unpin
 * @tr: The transaction in the AIL that contains this buffer
 *
 */

void gfs_dunpin(osi_buf_t *bh, gfs_trans_t *tr)
{
  gfs_bufdata_t *bd;
  int dec = FALSE;

  ENTER(GFN_DUNPIN);


  bd = BH_BUFDATA(bh);
  GFS_ASSERT(bd, );


  /*  Add the buffer to the AIL
      and get rid of an old reference if there is one  */

  if (tr)
  {
    if (bd->bd_arch.bd_ail_tr)
    {
      GFS_ASSERT(osi_atomic_read(&bd->bd_arch.bd_ail_tr->tr_arch.tr_bufs), );
      osi_atomic_dec(&bd->bd_arch.bd_ail_tr->tr_arch.tr_bufs);
    }
    bd->bd_arch.bd_ail_tr = tr;
  }


  gfs_lock_buffer(bh);

  GFS_ASSERT(bd->bd_arch.bd_pinned, );
  bd->bd_arch.bd_pinned--;

  if (!bd->bd_arch.bd_pinned)
  {
    mark_buffer_dirty(bh, 0);
    dec = TRUE;
  }

  gfs_unlock_buffer(bh);


  if (dec)
    bh->b_count--;


  EXIT(GFN_DUNPIN);
}


/**
 * gfs_is_pinned - check to see if a buffer is pinned
 * @bh: The buffer
 *
 * Returns:  TRUE if the buffer is pinned, FALSE if it isn't.
 */

int gfs_is_pinned(osi_buf_t *bh)
{
  gfs_bufdata_t *bd;
  int ret = FALSE;

  ENTER(GFN_IS_PINNED);

  bd = BH_BUFDATA(bh);
  if (bd)
  {
    gfs_lock_buffer(bh);
    ret = bd->bd_arch.bd_pinned;
    gfs_unlock_buffer(bh);
  }

  EXIT(GFN_IS_PINNED);

  return ret;
}


/**
 * gfs_end_buffer_io_sync - End I/O Callback
 * @bh: The buffer
 * @uptodate: Flag indicating buffer state
 *
 */

STATIC void gfs_end_buffer_io_sync(osi_buf_t *bh, int uptodate)
{
  mark_buffer_uptodate(bh, uptodate);
  unlock_buffer(bh);
}


/**
 * gfs_init_temp_bh - Initialize a temporary buffer head
 * @sdp: The GFS superblock
 * @blkno: The block number
 * @buf: The data area
 * @write: TRUE if this buffer will be writen, FALSE if it will be read
 *
 * Returns:  the temp buffer head
 */

osi_buf_t *gfs_init_temp_bh(gfs_sbd_t *sdp, uint64 blkno, char *buf, int write)
{
  osi_buf_t *bh;

  ENTER(GFN_INIT_TEMP_BH);

  bh = gmalloc(sizeof(osi_buf_t));
  if (!bh)
    goto out;

  gfs_memset(bh, 0, sizeof(osi_buf_t));

  bh->b_size = sdp->sd_sb.sb_bsize;
  bh->b_data = buf;
  bh->b_count = 1;
  bh->b_list = BUF_CLEAN;
  bh->b_dev = sdp->sd_vsb->s_dev;
  bh->b_blocknr = blkno;
  bh->b_end_io = gfs_end_buffer_io_sync;
  if (write)
    bh->b_state = (1ul << BH_Dirty) | (1ul << BH_Uptodate);

 out:
  EXIT(GFN_INIT_TEMP_BH);

  return bh;
}


/**
 * gfs_dispose_temp_bh - Get rid of a temporary buffer_head
 * @bh: The buffer head we no longer want
 *
 */

void gfs_dispose_temp_bh(osi_buf_t *bh)
{
  ENTER(GFN_DISPOSE_TEMP_BH);
  gfree(bh, sizeof(osi_buf_t));
  EXIT(GFN_DISPOSE_TEMP_BH);
}


/**
 * gfs_get_meta_buffer - Get a metadata buffer
 * @ip: The GFS inode
 * @depth: The depth in the metadata tree
 * @num: The block number (device relative) of the buffer
 * @new: Non-zero if we may create a new buffer
 *
 * Returns: The metadata buffer or NULL
 */

osi_buf_t *gfs_get_meta_buffer(gfs_inode_t *ip, int height, uint64 num, int new)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  osi_buf_t *bh = ip->i_cache[height];
  int flags = ((new) ? DR_FORCE : 0) | DR_START | DR_WAIT;

  ENTER(GFN_GET_META_BUFFER);


  if (bh && (BH_BLKNO(bh) == num))
  {
    if (gfs_dreread(sdp, bh, flags) != 0)
    {
      bh = NULL;
      goto out;
    }
    goto check_it;
  }


  gfs_flush_meta_cache(ip, height);

  bh = gfs_dread(sdp, num, ip->i_gl, flags);
  if (!bh)
    goto out;
  ip->i_cache[height] = bh;


 check_it:
  if (new)
  {
    gfs_trans_add_bh(sdp, bh);
    gfs_metatype_set(bh, GFS_METATYPE_IN, GFS_FORMAT_IN);
    gfs_buffer_clear_tail(bh, sizeof(gfs_indirect_t));
  }
  else
    gfs_metatype_check(bh, (height) ? GFS_METATYPE_IN : GFS_METATYPE_DI);


  bh->b_count++;
  buf_alloc_main(bh, __FILE__, __LINE__);


 out:
  EXIT(GFN_GET_META_BUFFER);

  return bh;
}


/**
 * gfs_get_data_buffer - Get a data buffer
 * @ip: The GFS inode
 * @num: The block number (device relative) of the data block
 * @new: Non-zero if this is a new allocation
 *
 * Returns: The data buffer or NULL
 */

osi_buf_t *gfs_get_data_buffer(gfs_inode_t *ip, uint64 block, int new)
{
  osi_buf_t *bh;

  ENTER(GFN_GET_DATA_BUFFER);


  if (block == ip->i_num.no_addr)
  {
    GFS_ASSERT(!new, );

    bh = gfs_dread(ip->i_sbd, block, ip->i_gl, DR_START | DR_WAIT);
    if (bh)
      gfs_metatype_check(bh, GFS_METATYPE_DI);
  }
  else if (gfs_is_jdata(ip))
  {
    if (new)
    {
      bh = gfs_dread(ip->i_sbd, block, ip->i_gl, DR_FORCE | DR_START | DR_WAIT);
      if (bh)
      {
	gfs_trans_add_bh(ip->i_sbd, bh);
	gfs_metatype_set(bh, GFS_METATYPE_JD, GFS_FORMAT_JD);
	gfs_buffer_clear_tail(bh, sizeof(gfs_jdata_t));
      }
    }
    else
    {
      bh = gfs_dread(ip->i_sbd, block, ip->i_gl, DR_START | DR_WAIT);
      if (bh)
	gfs_metatype_check(bh, GFS_METATYPE_JD);
    }
  }
  else
  {
    if (new)
    {
      bh = gfs_dgetblk(ip->i_sbd, block, ip->i_gl);
      if (bh) 
	gfs_prep_new_buffer(bh);
    }
    else
      bh = gfs_dread(ip->i_sbd, block, ip->i_gl, DR_START | DR_WAIT);
  }


  EXIT(GFN_GET_DATA_BUFFER);

  return bh;
}


