/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portions Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: inode.c
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"

#include <linux/utsname.h>
#include <linux/version.h>

#include "blklist.h"
#include "bmap.h"
#include "dcache.h"
#include "dio.h"
#include "dir.h"
#include "gfs_linux.h"
#include "glock.h"
#include "glops.h"
#include "file_linux.h"
#include "fs.h"
#include "inode.h"
#include "inode_linux.h"
#include "trans.h"
#include "gfs_dfsa.h"




/**
 * gfs_imatch - The match function used in gfs_iget()
 * @inode: The inode
 * @ino: The inode number (32 bits)
 * @ptr: The gfs_inode_t pointer we're looking for
 *
 * Returns: TRUE on a match, FALSE on a failure
 */

STATIC int gfs_imatch(struct inode *inode, unsigned long ino, void *ptr)
{
  gfs_inode_t *ip = vi2ino(inode);
  int ret = FALSE;

  ENTER(GFN_IMATCH);

  if (ip == (gfs_inode_t *)ptr)
    ret = TRUE;

  EXIT(GFN_IMATCH);

  return ret;
}


/**
 * gfs_iget - Gets an inode, does a 64 bit compare
 * @sb: The superblock
 * @ip: the gfs_inode_t to get the osi_vnode_t for
 *
 * Returns: An inode
 */

struct inode *gfs_iget(struct super_block *sb, gfs_inode_t *ip)
{
  struct inode *inode;
  ENTER(GFN_IGET);

  inode = iget4(sb, (unsigned long)ip->i_num.no_formal_ino, gfs_imatch, ip);

  EXIT(GFN_IGET);
  return inode;
}


/**
 * gfs_inode_attr_in - Copy attributes from the dinode into the VFS inode
 * @ip: The GFS inode
 *
 */

void gfs_inode_attr_in(gfs_inode_t *ip)
{
  struct inode *ino = ip->i_vino;

  ENTER(GFN_INODE_ATTR_IN);

  if (!ino)
    goto out;


  ino->i_mode = ip->i_di.di_mode & S_IALLUGO;

  switch (ip->i_di.di_type)
  {
  case FILE_REG:
    ino->i_mode |= S_IFREG;
    ino->i_rdev = NODEV;
    break;
  case FILE_DIR:
    ino->i_mode |= S_IFDIR;
    ino->i_rdev = NODEV;
    break;
  case FILE_LNK:
    ino->i_mode |= S_IFLNK;
    ino->i_rdev = NODEV;
    break;
  case FILE_BLK:
    ino->i_mode |= S_IFBLK;
    ino->i_rdev = MKDEV(ip->i_di.di_major, ip->i_di.di_minor);
    break;
  case FILE_CHR:
    ino->i_mode |= S_IFCHR;
    ino->i_rdev = MKDEV(ip->i_di.di_major, ip->i_di.di_minor);
    break;
  case FILE_FIFO:
    ino->i_mode |= S_IFIFO;
    ino->i_rdev = NODEV;
    break;
  case FILE_SOCK:
    ino->i_mode |= S_IFSOCK;
    ino->i_rdev = NODEV;
    break;
  default:
    GFS_ASSERT(0, con_printf("fino/addr = %"FMT_64"u/%"FMT_64"u, type = %u\n",
			     ip->i_num.no_formal_ino, ip->i_num.no_addr, ip->i_di.di_type););
    break;
  };

  ino->i_nlink = ip->i_di.di_nlink;
  ino->i_uid = ip->i_di.di_uid;
  ino->i_gid = ip->i_di.di_gid;
  ino->i_size = ip->i_di.di_size;
  ino->i_atime = ip->i_di.di_atime;
  ino->i_mtime = ip->i_di.di_mtime;
  ino->i_ctime = ip->i_di.di_ctime;
  ino->i_blksize = PAGE_SIZE;
  ino->i_blocks = ip->i_di.di_blocks * ip->i_sbd->sd_sb.sb_bsize / 512;
  ino->i_generation = ip->i_di.di_incarn;

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,2,13))
  ino->i_version = ++global_event;
#else
  ino->i_version = ++event;
#endif


 out:
  EXIT(GFN_INODE_ATTR_IN);
}


/**
 * gfs_inode_attr_out - Copy attributes from VFS inode into the dinode
 * @ip: The GFS inode
 *
 * Only copy out the attributes that we want the VFS layer
 * to be able to modify.
 */

void gfs_inode_attr_out(gfs_inode_t *ip)
{
  struct inode *ino = ip->i_vino;

  ENTER(GFN_INODE_ATTR_OUT);

  if (!ino)
    goto out;


  ip->i_di.di_mode = ino->i_mode & S_IALLUGO;
  ip->i_di.di_uid = ino->i_uid;
  ip->i_di.di_gid = ino->i_gid;
  ip->i_di.di_atime = ino->i_atime;
  ip->i_di.di_mtime = ino->i_mtime;
  ip->i_di.di_ctime = ino->i_ctime;


 out:
  EXIT(GFN_INODE_ATTR_OUT);
}


/**
 * gfs_lookup - Look up a filename in a directory and return its inode
 * @dir: The directory inode
 * @dentry: The dentry of the new inode
 *
 * Called by the VFS layer. Lock dir and call gfs_lookupi()
 *
 * Returns: 0 on success, -EXXXX on failure
 */

struct dentry *gfs_lookup(struct inode *dir, struct dentry *dentry)
{
  gfs_inode_t *dip = vi2ino(dir), *ip;
  gfs_sbd_t *sdp = dip->i_sbd;
  struct inode *inode = NULL;
  gfs_dcached_t *dcd;
#ifdef CONFIG_MOSIX_DFSA
  /* variables for the consistency checking of dfsa, needed before the addition
   * of a new dentry to the dcache
   */
  gfs_dcached_t *dcd2;
  struct list_head *head;
  struct list_head *tmp;
#endif CONFIG_MOSIX_DFSA
  int error;

  ENTER(GFN_LOOKUP);

  IFDB(GFN_LOOKUP,
       db_printf("  %"FMT_64"u, %s\n", dip->i_num.no_formal_ino, dentry->d_name.name););


  dcd = (gfs_dcached_t *)gmalloc(sizeof(gfs_dcached_t));
  if (!dcd)
  {
    error = -ENOMEM;
    goto fail;
  }
  gfs_memset(dcd, 0, sizeof(gfs_dcached_t));


  error = gfs_lookupi(dip, &dentry->d_name, TRUE, NULL, &ip);
  if (error != 0 && error != -ENOENT)
    goto fail_gfree;


  if (ip)
  {
    inode = gfs_iget(sdp->sd_vsb, ip);
    gfs_put_istruct(ip);
  }


#ifdef CONFIG_MOSIX_DFSA
  
  /* lookup in the inode alias list just in case that we have
   * a yet-unknown dentry pointing to this inode. It is being done in here
   * because no change in VFS is wanted.
   */

  /* do nothing in case of error */
  if (!inode)
    goto after_dfsa_part;
  
  head = &inode->i_dentry;
  tmp = head;
  /* 
   * Scans the list of d_alias to find an alias - part of the dcach maintenance
   * which prevents duplication of a dentry 
   */
  while ((tmp = tmp->next) != head)
  {
    struct dentry *entry = list_entry(tmp, struct dentry, d_alias);
	  
    dcd2 = (gfs_dcached_t *) entry->d_fsdata;
	  
    /* 
     * check if the current alias is a match, if so the existing dentry 
     * will be returned, and no new dentry will be allocated.
     */
	  
    /* 
     * case of alias which is the same according flags. First compares
     * inodes for non-hardlinks, otherwise compare names
     */
    if (((entry->d_inode->i_ino == dentry->d_inode->i_ino) &&
         (entry->d_inode->i_nlink==1 || S_ISDIR(entry->d_inode->i_mode))) ||
        (dcd2 && (dcd2->dc_dfsa_flags & DFSA_YETUNKNOWN) &&
	 dcd2->dc_parent_ino == dir->i_ino &&
	 ((entry->d_name.len > FILENAME_START &&
           !osi_strcmp(entry->d_name.name, dentry->d_name.name)) ||
	  (entry->d_name.len <= FILENAME_START &&
	   dentry->d_name.len > MAX_PATH_LEN))))
    {
      char *str;
      struct qstr   *name = &dentry->d_name;
      struct dentry *parent = dentry->d_parent;

#ifdef CONFIG_MOSIX_DEBUG
      if(ds_debug & DSDEB_DFSA_GFS)
      {
        printk("--GFS lookup Found matching YETUNKNOWN\n");
      }
#endif CONFIG_MOSIX_DEBUG
		  
      /* 
       * we need to copy the name to the YETUNKNOWN dentry,
       * allocates memory if name too long
       */
       if (name->len > DNAME_INLINE_LEN-1)
       {
         str = kmalloc((name->len +16) & ~15, GFP_BUFFER);
         if (!str)
         {
           printk("GFS DFSA Error: No mem\n");
           break;
         }
       } 
       else
         str = entry->d_iname; 
		  
       if (dname_external(entry)) 
         kfree(entry->d_name.name);
		  
       gfs_memcpy(str, name->name, name->len);
       str[name->len] = 0;
       entry->d_name.name = str;
       entry->d_name.len = name->len;
       entry->d_name.hash = name->hash;
		  
       /* Marks the dentry as regular */
       dcd2->dc_dfsa_flags &= ~DFSA_YETUNKNOWN;
		
       /* entry was belonged to root so must decrease root count */
       entry->d_parent->d_count--;

       /* updates parent, sb and sons of dentry */
       if(parent) 
       {
         entry->d_parent = dget(parent);
         entry->d_sb = parent->d_sb;
         list_del(&entry->d_child);
         list_add(&entry->d_child, &parent->d_subdirs);
       } 
		  
       iput(inode);
       gfree(dcd, sizeof(gfs_dcached_t));
       /* return the dentry found, and increases its reference count */
       return dget(entry);
    }
  } 

 after_dfsa_part:

#endif CONFIG_MOSIX_DFSA


  /*  Set up the dcache data and operations  */

  if (!vdc2dcd(dentry))
  {
    dcd->dc_vn = dip->i_gl->gl_vn;

    vdc2dcd(dentry) = dcd;
    dentry->d_op = &gfs_dops;
  }
  else
    gfree(dcd, sizeof(gfs_dcached_t));


  d_add(dentry, inode);


  IFDB(GFN_LOOKUP,
       db_printf("  error = %d\n", error););

  EXIT(GFN_LOOKUP);
 
  return NULL;



 fail_gfree:
  gfree(dcd, sizeof(gfs_dcached_t));

 fail:
  EXIT(GFN_LOOKUP);
  return ERR_PTR(error);
}


/**
 * gfs_create - Create a file
 * @dir: The directory in which to create the file
 * @dentry: The dentry of the new file
 * @mode: The permissions of the new file
 *
 * Returns: 0 on success, -EXXXX on failure
 */

STATIC int gfs_create(struct inode *dir, struct dentry *dentry, int mode)
{
  gfs_inode_t *dip = vi2ino(dir), *ip;
  gfs_sbd_t *sdp = dip->i_sbd;
  struct inode *inode;
  osi_buf_t *dibh;
  int new;
  int error;

  ENTER(GFN_CREATE);


  error = gfs_createi(dip, &dentry->d_name, FILE_REG, FALSE, NULL, &new, &ip);
  if (error)
  {
    EXIT(GFN_CREATE);
    return error;
  }

  GFS_ASSERT(ip, );


  if (new)
  {
    ip->i_di.di_mode = mode;
    ip->i_di.di_uid = current->fsuid;
    ip->i_di.di_gid = current->fsgid;

    dibh = gfs_get_inode_buffer(ip);
    GFS_ASSERT(dibh, );

    gfs_dinode_out(&ip->i_di, BH_DATA(dibh));

    gfs_drelse(dibh);

    osi_set_bit(GIF_CREATED, &ip->i_flags);

    gfs_trans_end(sdp);
    gfs_inplace_relse(dip);
  }


  gfs_gunlock_i(ip, 0);
  gfs_gunlock_i(dip, 0);


  inode = gfs_iget(sdp->sd_vsb, ip);
  gfs_put_istruct(ip);


  d_instantiate(dentry, inode);


  EXIT(GFN_CREATE);
  
  return 0;
}


/**
 * gfs_link - Link to a file
 * @old_dentry: The inode to link
 * @dir: Add link to this directory
 * @dentry: The name of the link
 *
 * Link the inode in "old_dentry" into the directory "dir" with the
 * name in "dentry".
 *
 * Returns: 0 on success, -EXXXX on failure
 */

STATIC int gfs_link(struct dentry *old_dentry, struct inode *dir, struct dentry *dentry)
{
  gfs_inode_t *dip = vi2ino(dir);
  gfs_sbd_t *sdp = dip->i_sbd;
  gfs_inode_t *ip = vi2ino(old_dentry->d_inode);
  gfs_ipreserv_t *res = &dip->i_res;
  gfs_lockop_t lock[2];
  int error;

  ENTER(GFN_LINK);


  if (ip->i_di.di_type == FILE_DIR)
  {
    error = -EPERM;
    goto fail;
  }

 
  gfs_memset(lock, 0, 2 * sizeof(gfs_lockop_t));

  lock[0].lo_ip = dip;
  lock[1].lo_ip = ip;

  error = gfs_glock_m(sdp, 2, lock);
  if (error)
    goto fail;


  error = permission(dir, MAY_WRITE | MAY_EXEC);
  if (error)
    goto fail_gunlock;

  error = gfs_dir_search(dip, &dentry->d_name, NULL, NULL);
  switch (error)
  {
  case -ENOENT:
    break;
  case 0:
    error = -EEXIST;
    goto fail_gunlock;
  default:
    goto fail_gunlock;
  }


  res->ir_diblks = 0;
  res->ir_metablks = sdp->sd_max_dirres;
  res->ir_datablks = 0;
  res->ir_target = dip->i_di.di_goal_rgrp;
  res->ir_trans = TRANS_LINK;

  error = gfs_inplace_reserv(dip);
  if (error)
    goto fail_gunlock;


  /*  Trans may require:
      two dinode blocks, directory modifications to add an entry,
      and RG bitmap blocks to allocate from  */

  error = gfs_trans_begin(sdp, TRANS_LINK, 2 + sdp->sd_max_dirres + res->ir_rgd->rd_ri.ri_length, 0);
  if (error)
    goto fail_ipres;

  error = gfs_dir_add(dip, &dentry->d_name, &ip->i_num, ip->i_di.di_type);
  if (error)
    goto fail_end_trans; 

  error = gfs_change_nlink(ip, +1);
  if (error)
    goto fail_end_trans;

  gfs_trans_end(sdp);


  gfs_inplace_relse(dip);
  gfs_gunlock_m(sdp, 2, lock);


  ip->i_vino->i_count++;

  d_instantiate(dentry, ip->i_vino);


  EXIT(GFN_LINK);

  return 0;



 fail_end_trans:
  gfs_trans_mark(sdp, TRF_POISONED);
  gfs_trans_end(sdp);

 fail_ipres:
  gfs_inplace_relse(dip);

 fail_gunlock:
  gfs_gunlock_m(sdp, 2, lock);

 fail:
  EXIT(GFN_LINK);
  return error;
}


/**
 * gfs_unlink - Unlink a file
 * @dir: The inode of the directory containing the file to unlink
 * @dentry: The file itself
 *
 * Unlink a file.  Call gfs_unlinki()
 *
 * Returns: 0 on success, -EXXXX on failure
 */

STATIC int gfs_unlink(struct inode *dir, struct dentry *dentry)
{
  gfs_inode_t *dip = vi2ino(dir);
  gfs_sbd_t *sdp = dip->i_sbd;
  gfs_inode_t *ip = vi2ino(dentry->d_inode);
  gfs_lockop_t lock[2];
  int error;

  ENTER(GFN_UNLINK);


  error = gfs_limit_unlinked(sdp);
  if (error)
    goto fail;


  gfs_memset(lock, 0, 2 * sizeof(gfs_lockop_t));

  lock[0].lo_ip = dip;
  lock[1].lo_ip = ip;

  error = gfs_glock_m(sdp, 2, lock);
  if (error)
    goto fail;

  
  if (!gfs_drevalidate(dentry, 0))
  {
    error = -ENOENT;
    goto fail_gunlock;
  }

  error = permission(dir, MAY_WRITE | MAY_EXEC);
  if (error)
    goto fail_gunlock;


  /*  Trans may require:
      Two dinode blocks and one modified directory leaf block
      and one nopen tag.  */

  error = gfs_trans_begin(sdp, TRANS_UNLINK, 3, 1);
  if (error)
    goto fail_gunlock;

  error = gfs_unlinki(dip, &dentry->d_name, ip);
  if (error)
    goto fail_end_trans;

  gfs_trans_end(sdp);


  gfs_gunlock_m(sdp, 2, lock);


  if (!error)
    d_delete(dentry);


  EXIT(GFN_UNLINK);

  return 0;



 fail_end_trans:
  gfs_trans_mark(sdp, TRF_POISONED);
  gfs_trans_end(sdp);

 fail_gunlock:
  gfs_gunlock_m(sdp, 2, lock);  

 fail:
  EXIT(GFN_UNLINK);
  return error;
}


/**
 * gfs_symlink - Create a symlink
 * @dir: The directory to create the symlink in
 * @dentry: The dentry to put the symlink in
 * @symname: The thing which the link points to
 *
 * Returns: 0 on success, -EXXXX on failure
 */

STATIC int gfs_symlink(struct inode *dir, struct dentry *dentry, const char *symname)
{
  gfs_inode_t *dip = vi2ino(dir), *ip;
  gfs_sbd_t *sdp = dip->i_sbd;
  struct inode *inode;
  osi_buf_t *dibh;
  int size;
  int error;

  ENTER(GFN_SYMLINK);


  /*  Must be stuffed with a null terminator for gfs_follow_link()  */

  size = osi_strlen(symname);
  if (size > sdp->sd_sb.sb_bsize - sizeof(gfs_dinode_t) - 1)
  {
    error = -ENAMETOOLONG;
    goto fail;
  }


  error = gfs_createi(dip, &dentry->d_name, FILE_LNK, TRUE, NULL, NULL, &ip);
  if (error)
    goto fail;

  GFS_ASSERT(ip, );


  ip->i_di.di_mode = 0777;
  ip->i_di.di_uid = current->fsuid;
  ip->i_di.di_gid = current->fsgid;
  ip->i_di.di_size = size;
  ip->i_di.di_type = FILE_LNK;

  dibh = gfs_get_inode_buffer(ip);
  GFS_ASSERT(dibh, );

  gfs_dinode_out(&ip->i_di, BH_DATA(dibh));
  gfs_memcpy(BH_DATA(dibh) + sizeof(gfs_dinode_t), symname, size);

  gfs_drelse(dibh);


  gfs_trans_end(sdp);

  gfs_inplace_relse(dip);

  gfs_gunlock_i(ip, 0);
  gfs_gunlock_i(dip, 0);


  inode = gfs_iget(sdp->sd_vsb, ip);
  gfs_put_istruct(ip);


  d_instantiate(dentry, inode);


  EXIT(GFN_SYMLINK);

  return 0;



 fail:
  EXIT(GFN_SYMLINK);
  return error;
}


/**
 * gfs_mkdir - Make a directory
 * @dir: The parent directory of the new one
 * @dentry: The dentry of the new directory
 * @mode: The mode of the new directory
 *
 * Returns: 0 on success, -EXXXX on failure
 */

STATIC int gfs_mkdir(struct inode *dir, struct dentry *dentry, int mode)
{
  gfs_inode_t *dip = vi2ino(dir), *ip;
  gfs_sbd_t *sdp = dip->i_sbd;
  struct inode *inode;
  osi_buf_t *dibh;
  gfs_dinode_t *di;
  gfs_dirent_t *dent;
  int error;

  ENTER(GFN_MKDIR);


  error = gfs_createi(dip, &dentry->d_name, FILE_DIR, TRUE, NULL, NULL, &ip);
  if (error)
  {
    EXIT(GFN_MKDIR);
    return error;
  }

  GFS_ASSERT(ip, );


  dibh = gfs_get_inode_buffer(ip);
  GFS_ASSERT(dibh, );

  di = (gfs_dinode_t *)BH_DATA(dibh);

  di->di_mode = cpu_to_gfs32(mode);
  di->di_uid = cpu_to_gfs32(current->fsuid);
  di->di_gid = cpu_to_gfs32(current->fsgid);
  di->di_nlink = cpu_to_gfs32(2);
  di->di_size = cpu_to_gfs64(sdp->sd_sb.sb_bsize - sizeof(gfs_dinode_t));

  di->di_flags = cpu_to_gfs32(GFS_DIF_JDATA);
  di->di_payload_format = cpu_to_gfs32(GFS_FORMAT_DE);
  di->di_type = cpu_to_gfs16(FILE_DIR);

  error = gfs_dirent_alloc(ip, dibh, 1, &dent);
  GFS_ASSERT(!error, );  /*  This should never fail  */

  dent->de_inum = di->di_num;  /*  already GFS endian  */
  dent->de_hash = gfs_hash(".", 1);
  dent->de_hash = cpu_to_gfs32(dent->de_hash);
  dent->de_type = cpu_to_gfs16(FILE_DIR);
  gfs_memcpy((char *)(dent + 1), ".", 1);
  di->di_entries = cpu_to_gfs32(1);

  error = gfs_dirent_alloc(ip, dibh, 2, &dent);
  GFS_ASSERT(!error, );  /*  This should never fail  */

  gfs_inum_out(&dip->i_num, (char *)&dent->de_inum);
  dent->de_hash = gfs_hash("..", 2);
  dent->de_hash = cpu_to_gfs32(dent->de_hash);
  dent->de_type = cpu_to_gfs16(FILE_DIR);
  gfs_memcpy((char *)(dent + 1), "..", 2);
  di->di_entries = cpu_to_gfs32(2);

  gfs_dinode_in(&ip->i_di, (char *)di);

  gfs_drelse(dibh);
  

  error = gfs_change_nlink(dip, +1);
  GFS_ASSERT(!error, );


  gfs_trans_end(sdp);

  gfs_inplace_relse(dip);

  gfs_gunlock_i(ip, 0);
  gfs_gunlock_i(dip, 0);


  inode = gfs_iget(sdp->sd_vsb, ip);
  gfs_put_istruct(ip);


  d_instantiate(dentry, inode);


  EXIT(GFN_MKDIR);

  return 0;
}


/**
 * gfs_rmdir - Remove a directory
 * @dir: The parent directory of the directory to be removed
 * @dentry: The dentry of the directory to remove
 *
 * Remove a directory. Call gfs_rmdiri()
 *
 * Returns: 0 on success, -EXXXX on failure
 */

STATIC int gfs_rmdir(struct inode *dir, struct dentry *dentry)
{
  gfs_inode_t *dip = vi2ino(dir);
  gfs_sbd_t *sdp = dip->i_sbd;
  gfs_inode_t *ip = vi2ino(dentry->d_inode);
  gfs_lockop_t lock[2];
  int error;

  ENTER(GFN_RMDIR);


  error = gfs_limit_unlinked(sdp);
  if (error)
    goto fail;


  gfs_memset(lock, 0, 2 * sizeof(gfs_lockop_t));

  lock[0].lo_ip = dip;
  lock[1].lo_ip = ip;

  error = gfs_glock_m(sdp, 2, lock);
  if (error)
    goto fail;


  if (!gfs_drevalidate(dentry, 0))
  {
    error = -ENOENT;
    goto fail_gunlock;
  }

  error = permission(dir, MAY_WRITE | MAY_EXEC);
  if (error)
    goto fail_gunlock;


  /*  Trans may require:
      Two dinode blocks, one directory leaf block containing the
      entry to be rmdired, two leaf blocks containing . and .. of
      the directory being rmdired, and one nopen tag  */

  error = gfs_trans_begin(sdp, TRANS_RMDIR, 5, 1);
  if (error)
    goto fail_gunlock;

  error = gfs_rmdiri(dip, &dentry->d_name, ip);
  if (error)
    goto fail_end_trans;

  gfs_trans_end(sdp);

  
  gfs_gunlock_m(sdp, 2, lock);


  if (!error)
    d_delete(dentry);


  EXIT(GFN_RMDIR);

  return 0;



 fail_end_trans:
  gfs_trans_mark(sdp, TRF_POISONED);
  gfs_trans_end(sdp);

 fail_gunlock:
  gfs_gunlock_m(sdp, 2, lock);  

 fail:
  EXIT(GFN_RMDIR);
  return error;
}


/**
 * gfs_mknod - Make a special file
 * @dir: The directory in which the special file will reside
 * @dentry: The dentry of the special file
 * @mode: The mode of the special file
 * @rdev: The device specification of the special file
 *
 */

int gfs_mknod(struct inode *dir, struct dentry *dentry, int mode, int rdev)
{
  gfs_inode_t *dip = vi2ino(dir), *ip;
  gfs_sbd_t *sdp = dip->i_sbd;
  struct inode *inode;
  osi_buf_t *dibh;
  kdev_t dev;
  uint16 type = 0;
  uint32 major = 0, minor = 0;
  int error;

  ENTER(GFN_MKNOD);


  switch (mode & S_IFMT)
  {
  case S_IFBLK:
    type = FILE_BLK;
    dev = to_kdev_t(rdev);
    major = MAJOR(dev);
    minor = MINOR(dev);
    break;
  case S_IFCHR:
    type = FILE_CHR;
    dev = to_kdev_t(rdev);
    major = MAJOR(dev);
    minor = MINOR(dev);
    break;
  case S_IFIFO:
    type = FILE_FIFO;
    break;
  case S_IFSOCK:
    type = FILE_SOCK;
    break;
  default:
    GFS_ASSERT(0, con_printf("mode = %d\n", mode););
    break;
  };


  error = gfs_createi(dip, &dentry->d_name, type, TRUE, NULL, NULL, &ip);
  if (error)
  {
    EXIT(GFN_MKNOD);
    return error;
  }

  GFS_ASSERT(ip, );


  ip->i_di.di_mode = mode;
  ip->i_di.di_uid = current->fsuid;
  ip->i_di.di_gid = current->fsgid;
  ip->i_di.di_major = major;
  ip->i_di.di_minor = minor;

  ip->i_di.di_type = type;


  dibh = gfs_get_inode_buffer(ip);
  GFS_ASSERT(dibh, );

  gfs_dinode_out(&ip->i_di, BH_DATA(dibh));

  gfs_drelse(dibh);


  gfs_trans_end(sdp);

  gfs_inplace_relse(dip);

  gfs_gunlock_i(ip, 0);
  gfs_gunlock_i(dip, 0);


  inode = gfs_iget(sdp->sd_vsb, ip);
  gfs_put_istruct(ip);


  d_instantiate(dentry, inode);


  EXIT(GFN_MKNOD);

  return 0;
}


/**
 * gfs_is_subdir - Return TRUE if @new_dentry is a subdir of @old_dentry
 * @new_dentry:
 * @old_dentry:
 *
 * Do the same thing as Linux's is_subdir(), but do drevalidates to make
 * sure the dcache is consistent.
 *
 * Returns: TRUE if new_dentry is a subdir of old_dentry, FALSE otherwise
 */

STATIC int gfs_is_subdir(struct dentry *new_dentry, struct dentry *old_dentry)
{
  int ret = FALSE;
  int error;

  ENTER(GFN_IS_SUBDIR);
 

  while (TRUE)
  {
    if (new_dentry != old_dentry)
    {
      if (new_dentry->d_parent == new_dentry)
	break;

      if (!gfs_drevalidate(new_dentry, 0))
      {
	error = -ENOENT;
	EXIT(GFN_IS_SUBDIR);
	return error;
      }

      new_dentry = new_dentry->d_parent;
      continue;
    }

    ret = TRUE;
    break;
  }


  EXIT(GFN_IS_SUBDIR);
 
  return ret;
}


/**
 * gfs_rename - Rename a file
 * @odir: Parent directory of old file name
 * @odentry: The old dentry of the file
 * @ndir: Parent directory of new file name
 * @ndentry: The new dentry of the file
 *
 * Returns: 0 on success, -EXXXX on failure
 */

int gfs_rename(struct inode *odir, struct dentry *odentry,
	       struct inode *ndir, struct dentry *ndentry)
{
  gfs_inode_t *odip = vi2ino(odir), *ndip = vi2ino(ndir);
  gfs_sbd_t *sdp = odip->i_sbd;
  gfs_inode_t *ip = vi2ino(odentry->d_inode), *nip = NULL;
  osi_qstr_t name;
  gfs_ipreserv_t *res = &ndip->i_res;
  gfs_lockop_t lock[4];
  int num_lock;
  int dir_rename = FALSE;
  int error;

  ENTER(GFN_RENAME);


  error = gfs_limit_unlinked(sdp);
  if (error)
    goto fail;


  if (ndentry->d_inode)
  {
    nip = vi2ino(ndentry->d_inode);

    if (ip == nip)
    {
      error = 0;
      goto fail;
    }
  }


  /*  Make sure we aren't trying to move a dirctory into it's subdir  */

  if (ip->i_di.di_type == FILE_DIR && odip != ndip)
  {
    error = gfs_glock(sdp, sdp->sd_rename_gl, 0);
    if (error)
      goto fail;

    error = gfs_is_subdir(ndentry->d_parent, odentry);
    switch (error)
    {
    case TRUE:
      error = -EINVAL;
      goto fail_renlock;
      break;
    case FALSE:
      break;
    default:
      goto fail_renlock;
      break;
    }

    dir_rename = TRUE;
  }


  gfs_memset(lock, 0, 4 * sizeof(gfs_lockop_t));

  lock[0].lo_ip = odip;
  lock[1].lo_ip = ndip;

  num_lock = 2;

  if (nip)
  {
    lock[num_lock].lo_ip = nip;
    num_lock++;
  }

  if (dir_rename)
  {
    lock[num_lock].lo_ip = ip;
    num_lock++;
  }

  error = gfs_glock_m(sdp, num_lock, lock);
  if (error)
    goto fail_renlock;


  /*  Check out the old directory  */

  error = permission(odir, MAY_WRITE | MAY_EXEC);
  if (error)
    goto fail_gunlock;

  if (!gfs_drevalidate(odentry, 0))
  {
    error = -ENOENT;
    goto fail_gunlock;
  }


  /*  Check out the new directory  */

  error = permission(ndir, MAY_WRITE | MAY_EXEC);
  if (error)
    goto fail_gunlock;

  if (nip)
  {
    if (!gfs_drevalidate(ndentry, 0))
    {
      error = -ENOENT;
      goto fail_gunlock;
    }
  }
  else
  {
    error = gfs_dir_search(ndip, &ndentry->d_name, NULL, NULL);
    switch (error)
    {
    case -ENOENT:
      error = 0;
      break;
    case 0:
      error = -EEXIST;
    default:
      goto fail_gunlock;
    };
  }


  res->ir_diblks = 0;
  res->ir_metablks = sdp->sd_max_dirres;
  res->ir_datablks = 0;
  res->ir_target = ndip->i_di.di_goal_rgrp;
  res->ir_trans = TRANS_RENAME;

  error = gfs_inplace_reserv(ndip);
  if (error)
    goto fail_gunlock;  


  /*  Trans may require:
      Dinodes for the srcdir, srcino, dstdir, dstino.  Blocks for
      adding the entry to dstdir.  RG bitmaps for that allocation.
      One leaf block in the srcdir for removal of the entry.
      One leaf block for changing .. in srcino (if it's a directory).
      Two leaf blocks for removing . and .. from dstino (if it exists
      and it's a directory) and one nopen tag.  */

  error = gfs_trans_begin(sdp, TRANS_RENAME, 8 + sdp->sd_max_dirres + res->ir_rgd->rd_ri.ri_length, 1);
  if (error)
    goto fail_ipres;


  /*  Remove the target file, if it exists  */

  if (nip)
  {
    if (nip->i_di.di_type == FILE_DIR)
      error = gfs_rmdiri(ndip, &ndentry->d_name, nip);
    else
      error = gfs_unlinki(ndip, &ndentry->d_name, nip);

    if (error)
      goto fail_end_trans;
  }

  if (dir_rename)
  {
    gfs_change_nlink(ndip, +1);
    gfs_change_nlink(odip, -1);
    name.len = 2;
    name.name = "..";

    error = gfs_dir_mvino(ip, &name, &ndip->i_num, FILE_DIR);
    if (error)
      goto fail_end_trans;
  }

  error = gfs_dir_add(ndip, &ndentry->d_name, &ip->i_num, ip->i_di.di_type);
  if (error)
    goto fail_end_trans;

  error = gfs_dir_del(odip, &odentry->d_name);
  if (error)
    goto fail_end_trans;

  if (dir_rename)
    gfs_trans_add_gl(sdp, sdp->sd_rename_gl);


  gfs_trans_end(sdp);


  gfs_inplace_relse(ndip);

  gfs_gunlock_m(sdp, num_lock, lock);

  if (dir_rename)
    gfs_gunlock(sdp, sdp->sd_rename_gl, 0);


  EXIT(GFN_RENAME);
 
  return 0;



 fail_end_trans:
  gfs_trans_mark(sdp, TRF_POISONED);
  gfs_trans_end(sdp);

 fail_ipres:
  gfs_inplace_relse(ndip);

 fail_gunlock:
  gfs_gunlock_m(sdp, num_lock, lock);

 fail_renlock:
  if (dir_rename)
    gfs_gunlock(sdp, sdp->sd_rename_gl, 0);  

 fail:
  EXIT(GFN_RENAME);
  return error;
}


/**
 * gfs_irevalidate - Check inode hasn't changed
 * @dentry: The dentry
 *
 * Revalidate the inode. i.e. grab the Glock and sync the incore dinode
 * with the ondisk inode
 *
 * Returns: 0 on success, or error
 */

int gfs_irevalidate(struct dentry *dentry)
{
  gfs_inode_t *ip = vi2ino(dentry->d_inode);
  int error;

  ENTER(GFN_IREVALIDATE);


  error = gfs_glock_i(ip, GL_SHARED);
  if (!error)
    gfs_gunlock_i(ip, 0);


  EXIT(GFN_IREVALIDATE);

  return error;
}


/**
 * gfs_readlink - Read the value of a symlink
 * @dentry: the symlink
 * @buf: the buffer to read the symlink data into
 * @size: the size of the buffer
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int gfs_readlink(struct dentry *dentry, char *buf, int size)
{
  gfs_inode_t *ip = vi2ino(dentry->d_inode);
  osi_buf_t *dibh;
  int error;

  ENTER(GFN_READLINK);


  error = gfs_glock_i(ip, GL_SHARED);
  if (error)
    goto fail;

  error = GFS_UPDATE_ATIME(ip);
  if (error)
    goto fail_gunlock;


  dibh = gfs_get_inode_buffer(ip);
  if (!dibh)
  {
    error = -EIO;
    goto fail_gunlock;
  }

  if (size > ip->i_di.di_size)
    size = ip->i_di.di_size;

  if (osi_copy_to_user(buf, BH_DATA(dibh) + sizeof(gfs_dinode_t), size))
  {
    error = -EFAULT;
    goto fail_drelse;
  }

  gfs_drelse(dibh);


  gfs_gunlock_i(ip, 0);
  

  EXIT(GFN_READLINK);

  return size;



 fail_drelse:
  gfs_drelse(dibh);

 fail_gunlock:
  gfs_gunlock_i(ip, 0);

 fail:
  EXIT(GFN_READLINK);
  return error;
}


/**
 * sym_hostname - Substitute system host name
 * @str: String for result
 * @len: Length of result buffer
 *
 * Returns: Length of hostname
 */

STATIC unsigned int sym_hostname(char *str)
{
  unsigned int l = osi_strlen(system_utsname.nodename);

  if (str)
    gfs_memcpy(str, system_utsname.nodename, l);

  return l;
}


/**
 * sym_machine - Substitute machine type
 * @str: String for result
 * @len: Length of result buffer
 *
 * Returns: Length of machine type
 */

STATIC unsigned int sym_machine(char *str)
{
  unsigned int l = osi_strlen(system_utsname.machine);

  if (str)
    gfs_memcpy(str, system_utsname.machine, l);

  return l;
}


/**
 * sym_os - Substitute OS name
 * @str: String for result
 * @len: Length of result buffer
 *
 * Returns: Length of OS name
 */

STATIC unsigned int sym_os(char *str)
{
  unsigned int l = osi_strlen(system_utsname.sysname);

  if (str)
    gfs_memcpy(str, system_utsname.sysname, l);

  return l;
}


static struct gfs_symlink_ops symlink_ops[] =
{
  { "hostname}", 9, sym_hostname },
  { "mach}", 5, sym_machine },
  { "os}", 3, sym_os },
  { NULL, 0, NULL }
};


/**
 * gfs_follow_link - Follow a symbolic link
 * @dentry: The dentry of the link
 * @nd: Data that we pass to vfs_follow_link()
 *
 * We do CDSL expansion for the following strings:
 * {hostname}, {mach}, and {os}.  (The values substituted are the outputs
 * of "uname -n", "uname -m", and "uname -s" respectively.)
 *
 * This can handle symlinks of any size. It is optimised for symlinks
 * under GFS_FAST_SYMLINK_SIZE.
 *
 * Returns: 0 on success or error code
 */

struct dentry *gfs_follow_link(struct dentry *dentry, struct dentry *base, unsigned int follow)
{
  struct dentry *ret = NULL;
  gfs_inode_t *ip = vi2ino(dentry->d_inode);
  osi_buf_t *dibh;
  char array[GFS_FAST_SYMLINK_SIZE];
  char *buf = array;
  unsigned int len = 0;
  int error;

  ENTER(GFN_FOLLOW_LINK);


  error = gfs_glock_i(ip, GL_SHARED);
  if (error)
    goto fail;
 
  error = GFS_UPDATE_ATIME(ip);
  if (error)
    goto fail_gunlock;


  dibh = gfs_get_inode_buffer(ip);
  if (!dibh)
  {
    error = -EIO;
    goto fail_gunlock;
  }

  /*  Don't forget the trailing zero  */

  len = gfs_link_size(symlink_ops, BH_DATA(dibh) + sizeof(gfs_dinode_t)) + 1;

  if (len > GFS_FAST_SYMLINK_SIZE)
  {
    buf = gmalloc(len);
    if (!buf)
    {
      error = -ENOMEM;
      goto fail_drelse;
    }
  }

  gfs_link_expand(symlink_ops, buf, BH_DATA(dibh) + sizeof(gfs_dinode_t));

  IFDB(GFN_FOLLOW_LINK,
       db_printf("  symlink substituted %s -> %s\n",
		 BH_DATA(dibh) + sizeof(gfs_dinode_t),
		 buf););

  gfs_drelse(dibh);

 
  gfs_gunlock_i(ip, 0);  

  
  ret = lookup_dentry(buf, base, follow);

  if (buf != array)
    gfree(buf, len);


  EXIT(GFN_FOLLOW_LINK);

  return ret;



 fail_drelse:
  gfs_drelse(dibh);

 fail_gunlock:
  gfs_gunlock_i(ip, 0);  

 fail:
  EXIT(GFN_FOLLOW_LINK);
  return ERR_PTR(error);
}


/**
 * bmap_unstuff - Unstuff a stuffed dinode
 * @ip: the inode
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int bmap_unstuff(gfs_inode_t *ip)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  gfs_ipreserv_t *res = &ip->i_res;
  int error;

  ENTER(GFN_BMAP_UNSTUFF);


  error = gfs_glock_i(ip, 0);
  if (error)
    goto fail;

  if (!gfs_is_stuffed(ip))
  {
    error = 0;
    goto fail_gunlock;
  }


  res->ir_diblks = 0;
  res->ir_metablks = 0;
  res->ir_datablks = 1;
  res->ir_target = ip->i_di.di_goal_rgrp;
  res->ir_trans = TRANS_BMAP;

  error = gfs_inplace_reserv(ip);
  if (error)
    goto fail_gunlock;


  /*  Trans may require:
      one block for the dinode, one block for the data that was
      previously stuffed, and a RG to allocate from  */

  error = gfs_trans_begin(sdp, TRANS_BMAP, 2 + res->ir_rgd->rd_ri.ri_length, 0);
  if (error)
    goto fail_ipres;

  error = gfs_unstuff_dinode(ip);
  if (error)
    goto fail_end_trans;

  gfs_trans_end(sdp);


  gfs_inplace_relse(ip);
  gfs_gunlock_i(ip, 0);


  EXIT(GFN_BMAP_UNSTUFF);

  return 0;



 fail_end_trans:
  gfs_trans_mark(sdp, TRF_POISONED);
  gfs_trans_end(sdp);

 fail_ipres:
  gfs_inplace_relse(ip); 

 fail_gunlock:
  gfs_gunlock_i(ip, 0);

 fail:
  EXIT(GFN_BMAP_UNSTUFF);
  return error;
}


/**
 * gfs_bmap - map a block number in a file into a block on the disk
 * @inode: the VFS inode
 * @iblock: the block to map
 *
 * Returns: 0 if the block doesn't exist in the file, or the disk block if it does.
 *          return 0 on error (I guess)
 */

int gfs_bmap(struct inode *inode, int iblock)
{
  gfs_inode_t *ip = vi2ino(inode);
  uint64 block;
  int not_new = 0;
  int error;

  ENTER(GFN_BMAP);


  if (iblock < 0)
  {
    con_printf("GFS:  gfs_bmap() called with iblock < 0\n");
    goto fail;
  }


 restart:

  error = gfs_glock_i(ip, GL_SHARED);
  if (error)
    goto fail;

  if (gfs_is_stuffed(ip))
  {
    gfs_gunlock_i(ip, 0);

    if (!iblock)
    {
      error = bmap_unstuff(ip);
      if (error)
	goto fail;
      goto restart;
    }
    else
    {
      block = 0;
      goto out;
    }
  }

  error = gfs_inode_get_block(ip, iblock, &not_new, &block);

  gfs_gunlock_i(ip, 0);

  if (error)
    goto fail;


  if (block & 0xFFFFFFFF80000000ll)
  {
    con_printf("GFS:  gfs_bmap():  block maps out of \"int\" range\n");
    goto fail;
  }


 out:
  EXIT(GFN_BMAP);

  return block;



 fail:
  EXIT(GFN_BMAP);
  return 0;  /*  Fucked up!  */
}


/**
 * stuffed_readpage - Fill in a Linux page with stuffed file data
 * @ip: the file
 * @page: the page
 *
 * Lots of stuff lifted out of generic_readpage() and
 *   brw_page().
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int stuffed_readpage(gfs_inode_t *ip, struct page *page)
{
  osi_buf_t *dibh;

  ENTER(GFN_STUFFED_READPAGE);


  dibh = gfs_get_inode_buffer(ip);
  if (!dibh)
  {
    /*  Exit here leaves page->count incremented  */
    clear_bit(PG_locked, &page->flags);
    wake_up(&page->wait);
    EXIT(GFN_STUFFED_READPAGE);
    return -EIO;
  }

  gfs_memcpy((void *)page_address(page), BH_DATA(dibh) + sizeof(gfs_dinode_t), ip->i_di.di_size);
  gfs_memset((void *)page_address(page) + ip->i_di.di_size, 0, PAGE_SIZE - ip->i_di.di_size);

  clear_bit(PG_locked, &page->flags);
  set_bit(PG_uptodate, &page->flags);
  wake_up(&page->wait);


  /*  Equivalent of after_unlock_page()  */

  __free_page(page);


  ++current->maj_flt;


  gfs_drelse(dibh);

  EXIT(GFN_STUFFED_READPAGE);

  return 0;
}


/**
 * gfs_readpage - readpage with locking
 * @file: The file to read a page for
 * @page: The page to read
 *
 * Returns: 0 on success, error code otherwise
 */

int gfs_readpage(struct file *file, struct page *page)
{
  gfs_inode_t *ip = vi2ino(file->f_dentry->d_inode);
  int rv;

  ENTER(GFN_READPAGE);


  /*  Equivalent of generic_readpage()
      We must lock the page _before_ we try to acquire the
      glock, or someone else can try to read the page while we're
      blocking on the glock.  This causes panics.  */

  atomic_inc(&page->count);
  set_bit(PG_locked, &page->flags);

  /*  Equivalent of brw_page()  */

  clear_bit(PG_uptodate, &page->flags);
  clear_bit(PG_error, &page->flags);


  rv = gfs_glock_i(ip, GL_SHARED);
  if (rv)
  {
    /*  Exit here leaves page->count incremented  */
    clear_bit(PG_locked, &page->flags);
    wake_up(&page->wait);
    goto out;
  }

  if (gfs_is_stuffed(ip) && !page->index)
    rv = stuffed_readpage(ip, page);
  else
  {
    atomic_dec(&page->count);  /*  crap...  */
    rv = generic_readpage(file, page);
  }

  gfs_gunlock_i(ip, 0);


 out:
  EXIT(GFN_READPAGE);

  return rv;
}


/**
 * gfs_permission - check permissions on a file
 * @ip: the file
 * @mask: the type of access we want
 * @cred: the structure defining our credentials
 *
 * Returns: 0 for Permission Granted, -EXXX on failure
 */

int gfs_permission(gfs_inode_t *ip, int mask, osi_cred_t *cred)
{
  int error;

  ENTER(GFN_PERMISSION);

  GFS_ASSERT(ip->i_vino, );
  error = permission(ip->i_vino, mask);

  EXIT(GFN_PERMISSION);

  return error;
}


/**
 * gfs_drop_excess_inodes - drop unused inodes
 * @sdp: the filesystem
 *
 */

void gfs_drop_excess_inodes(gfs_sbd_t *sdp)
{
  ENTER(GFN_DROP_EXCESS_INODES);

  if (!sdp->sd_args.ar_localcaching)
    shrink_dcache_sb(sdp->sd_vsb);

  EXIT(GFN_DROP_EXCESS_INODES);
}





struct inode_operations gfs_file_iops =
{
  default_file_ops:     &gfs_file_fops,
  readpage:             gfs_readpage,
  bmap:                 gfs_bmap,
  revalidate:           gfs_irevalidate
};


struct inode_operations gfs_dir_iops =
{
  default_file_ops:     &gfs_dir_fops,
  create:               gfs_create,
  lookup:               gfs_lookup,
  link:                 gfs_link,
  unlink:               gfs_unlink,
  symlink:              gfs_symlink,
  mkdir:                gfs_mkdir,
  rmdir:                gfs_rmdir,
  mknod:                gfs_mknod,
  rename:               gfs_rename,
  revalidate:           gfs_irevalidate
};


struct inode_operations gfs_symlink_iops =
{
  readlink:             gfs_readlink,
  follow_link:          gfs_follow_link,
  revalidate:           gfs_irevalidate
};





