/******************************************************************************
*******************************************************************************
**
**    Copyright 2000 Lior Amar <lior@cs.huji.ac.il>
**    Portions Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**         Lior Amar <lior@cs.huji.ac.il>
**         Steven Whitehouse <steve@sistina.com>
** 
**    Module name: m_super.c
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"

#include <linux/config.h>

#ifdef CONFIG_MOSIX_DFSA

#include "dcache.h"
#include "gfs_linux.h"
#include "glock.h"
#include "glops.h"
#include "fs.h"
#include "inode.h"
#include "inode_linux.h"
#include "super.h"
#include "super_linux.h"
#include "gfs_dfsa.h"


 /**
  * gfs_dfsa_identify - Creates identification information that will later be 
  * used to construct a live dentry in another node. 
  * The full pathname of the file (up to a limited size) and the inode number,
  * parent's inode number are kept as identifiction elements.
  * @dentry : Dentry from which data will be taken to build the ident struct. 
  * @data   : The ident struct that will be built to identify the dentry.
  * returns - the size of the new constructed identification struct
  */
int
gfs_dfsa_identify (struct dentry *dentry, void **data)
{
  /* 
   * The identification struct that will be built and assigned to data
   */
  static struct gfs_ident info;
  /* Inode number of dentry */
  gfs_inode_t *ip = vi2ino(dentry->d_inode);
  /* Inode numbers of dentry's parent and temporary one */
  struct dentry *tmp, *parent;
  /* pointer to the pathname buffer */
  char *p;
  /* number of parents for the inode */
  int depth;
  /* available space in the pathname's buffer */
  int len_left;
  /* buffer for the pathname string */
  char buff[MAX_PATH_LEN];

#ifdef CONFIG_MOSIX_DEBUG
  if (ds_debug & DSDEB_DFSA_GFS)
    {
      printk("----------Inside gfs_identify:--------------------\n ");
      printk("Dentry name %s\n", dentry->d_name.name);
      printk("Inode number %ld\n", dentry->d_inode->i_ino);
      printk("------- Getting dentry path  --------\n");
    }
#endif /* CONFIG_MOSIX_DEBUG */

  /* Initializations before start of buffer filling */
  tmp = dentry;
  p = &buff[MAX_PATH_LEN - 1];
  *p = '\0';
  len_left = MAX_PATH_LEN - 1;
  /* Initially assumes that there is a place for full pathname */
  info.status = IDENT_FULL_NAME;
  info.first_index = 0;
  depth = 0;
  /* scans the path and fill the buffer from end to start */
  for (;;)
    {
      /* the dentry was removed since the call to the function */
      if (tmp->d_parent != tmp && list_empty (&tmp->d_hash))
	return -1;
      tmp = tmp->d_covers;
      parent = tmp->d_parent;
      if (tmp == parent)
	break;
      len_left -= tmp->d_name.len;
      /* case of too long name for the identification struct */
      if (len_left < 1)		/* we need space for the / */
	{
	  if (depth == 0)
	    info.status = IDENT_NO_NAME;
	  else
	    info.status = IDENT_FILE_NAME;
	  break;
	}
      /* updates the next name found in the buffer */
      p -= tmp->d_name.len;
      gfs_memcpy(p, tmp->d_name.name, tmp->d_name.len);
      /* updates the index of the first name in buffer - file name */
      if (depth == 0)
	info.first_index = p - buff;
      *(--p) = '/';
      len_left--;
      depth++;
      tmp = parent;
    }

  /* shift the buffer content to the right - in case space was left */
  info.first_index -= p - buff;
  gfs_memcpy(info.path, p, MAX_PATH_LEN - (p - buff));
  /* Getting inode and parent inode numbers  */
  if (dentry->d_inode != NULL)
    {
      ip = vi2ino(dentry->d_inode);
      info.ino = (ip == NULL) ? 0 : ip->i_num.no_formal_ino;
      if (ip == NULL)
	printk("NULL GFS Inode in dentry %s\n", dentry->d_name.name);
    }

  if ((parent = dentry->d_parent) && parent->d_inode != NULL)
    {
      ip = vi2ino(parent->d_inode);
      info.parent_ino = (ip == NULL) ? 0 : ip->i_num.no_formal_ino;
      if (ip == NULL)
	printk("NULL GFS Inode in dentry %s\n", parent->d_name.name);
    }

  *data = (void *)&info;
#ifdef CONFIG_MOSIX_DEBUG
  if (ds_debug & DSDEB_DFSA_GFS)
    {
      printk("BUFF : $$$$ %s $$$$ \n", info.path);
      printk("First Index = %d\n", info.first_index);
      printk("Inode         = %lld\n", info.ino);
      printk("Parent Inode  = %lld\n", info.parent_ino);
      printk("----------------------Out gfs_identify ----------\n");
      printk("size of gfs_ident is: %d\n", sizeof (info) - len_left);
    }
#endif /* CONFIG_MOSIX_DEBUG */
  /* sizeof(gfs_ident) is the actual size and not the total size of the
     struct - since the name can be shorter
   */
  printk("size of gfs_ident is: %d\n", sizeof (info) - len_left);
  return(sizeof (info) - len_left);
}

 /**
  * gfs_dfsa_compare()
  * 
  * @sb      : subber block of current file system
  * @dp      : dentry to be compare to an identification information.
  * @info    : identification struct to be compared with a dentry.
  * @infolen : length of info.
  * compares a dentry and identification information. Compares the inode
  * numbers to be sure that the info struct in describing the
  * given dentry.
  */

int
gfs_dfsa_compare (struct super_block *sb, struct dentry *dp,
	     void *info, int infolen)
{
  struct gfs_ident *i = (struct gfs_ident *)info;
  gfs_dcached_t *p = (gfs_dcached_t *)dp->d_fsdata;
  int res;
#ifdef CONFIG_MOSIX_DEBUG
  if (ds_debug & DSDEB_DFSA_GFS)
    {
      printk("--------Inside gfs_compare:---------------\n");
      printk("got name of dentry: dp->d_name.name\n");
    }
#endif /* CONFIG_MOSIX_DEBUG */

  if (infolen != sizeof(struct gfs_ident) || dp->d_inode->i_sb != sb)
    {
      printk("Wrong superblock or infolen mismatch\n");
      return 0;
    }

  /* decide if equal according a simple comparison of fields */
  res = p && (vi2ino (dp->d_inode)->i_num.no_formal_ino == i->ino);

#ifdef CONFIG_MOSIX_DEBUG
  if (ds_debug & DSDEB_DFSA_GFS)
    {
      printk("Identify Res = %d\n", res);
    }
#endif /* CONFIG_MOSIX_DEBUG */

  return res;
}

 /**
  * gfs_dfsa_compare_names - check if two names are equal or not.
  * Comapring names of info and dentry using the status field.
  * If status is IDENT_FILE_NAME then we can strcmp the names 
  * else if status is IDENT_NO_NAME then we only check that the lenght
  * of dentry_name is less then MAX_PATH_LEN. If yes that mean's that the
  * names dont match, otherwise we return that they match (taking the
  * chance that 2 files with the same inode (hardlink) will get the same dentry
  *
  * @status     : status of the name of a gfs_ident struct (full, partial, etc.)
  * @info_name  : name of a gfs_ident struct
  * @type       : type of dentry (regular or root-attached)
  * @dentry_name: qstr of dentry 
  *
  * Returns: False if the names are definitely not equal. Since not all of the
  * name may be known at time of comparing, a True return value may be a guess 
  * that the names are equal - this may result in a wrong hardlink reference but
  * will maintain cach consistency - so no duplicate dentry will be added. And
  * no real influence on the user.
  */

STATIC int gfs_dfsa_compare_names (int status, char *info_name, int type,
			struct qstr *dentry_name)
{
  int res = 0;
  /* regular dentry - so exists in dcache with full name */
  if (type == DFSA_REGULAR)
    {
      switch(status)
	{
	case IDENT_FILE_NAME:
	  res = !osi_strcmp(dentry_name->name, info_name) ? 1 : 0;
	  break;
	case IDENT_NO_NAME:
	  res = (dentry_name->len >= MAX_PATH_LEN) ? 1 : 0;
	  break;
	default:
	  res = 0;
	}
    }
  /* YETUNKNOWN dentry - so exists in dcache with partial name */
  else
    {
      switch (status)
	{
	case IDENT_FILE_NAME:
	  if (dentry_name->len > FILENAME_START &&
	      !osi_strcmp (&dentry_name->name[FILENAME_START], info_name))
	    {
	      printk ("--Name match\n");
	      res = 1;
	    }

	  break;
	case IDENT_NO_NAME:
	  res = dentry_name->len > FILENAME_START ? 0 : 1;
	  break;
	default:
	  res = 0;
	}
    }

#ifdef CONFIG_MOSIX_DEBUG
  if (ds_debug & DSDEB_DFSA_GFS)
    {
      printk("\n------------ gfs_dfsa_compare_names ------------\n");
      printk("Status %d\n", status);
      printk("Type   %d\n", type);
      printk("Name          ## %s ##\n", info_name);
      printk("Dentry Name   ## %s ##\n", dentry_name->name);
      printk("Res = %d \n", res);
      printk("\n------------------------------------------------\n");
    }
#endif /* CONFIG_MOSIX_DEBUG */

  return res;
}

/**
 * gfs_dfsa_reconstruct
 * Reconstructing a dentry from the identification information.
 * Currenly using the full name of the dentry to reconstruct a live dentry
 * and checking that the inode number is the same as in the info
 *
 * @sb : super block of the new dentry to be reconstructed.
 * @info : ident struct according to which the dentry will be reconstructed
 * @infolen : length of info.
 */

struct dentry *
gfs_dfsa_reconstruct (struct super_block *sb, void *info, int infolen)
{
  struct gfs_ident *i = (struct gfs_ident *) info;
  struct dentry *dp;
  struct inode *ip;
  int lookup_flag = 0;

#ifdef CONFIG_MOSIX_DEBUG
  if (ds_debug & DSDEB_DFSA_GFS)
    {
      printk("\n+--------Starting gfs_reconstruct:----------\n ");
      printk("Info inode = %Ld\n", i->ino);
      printk("Info  path = %s\n", i->path);
      printk("SB name %s\n", sb->s_root->d_name.name);
      printk("SB cover %s\n", sb->s_root->d_covers->d_name.name);
    }
#endif /* CONFIG_MOSIX_DEBUG */

  /* The simple 99.9% of the cases where the name is sufficieant. The 
   * dentry will be reconstructed as a regular dentry with full name 
   * so the consistency maintenance is the job of dcache.
   */
  if (i->status == IDENT_FULL_NAME)
    {
      lookup_flag |= LOOKUP_STAYFS;
      dp = lookup_dentry(i->path, NULL, lookup_flag);
      if (IS_ERR (dp))
	{
	  printk("lookup_dentry = NULL\n");
	  goto special_case;
	}

      /* 
       * There is a posibilty that the file was deleted and recreated
       * on the same inode while the info was in the way.
       */
      if (vi2ino(dp->d_inode)->i_num.no_formal_ino != i->ino)
	{
	  printk("wrong inode in reconstructed dentry\n");
	  goto special_case;
	}
    }
  /* 
   * Very rare case in which the pathname is too long. Dentry will be 
   * reconstructed as a special dentry attached to root. In this case a
   * special consitency test needed due to the usage of special dentries,
   * eliminates duplicate dentry.
   */
  else
    {
special_case:
      {
	gfs_inode_t *dip;
	gfs_dcached_t *dcd;
	struct qstr qname;
	int error;
	char buff[MAX_PATH_LEN + FILENAME_START + 10];
	gfs_sbd_t *sdp = vsb2sbd (sb);
	lm_lockname_t name;
	gfs_glock_t *gl;
	struct list_head *head;
	struct list_head *tmp;

	/* Trying to see if a local dentry for the inode already
	 * exist. Find the inode and look in the inods i_dentry list
         *
	 * First aqcuire gfs lock and gfs inode struct for inode 
         */
	gl = gfs_get_glstruct(sdp, i->ino, &gfs_inode_glops, NULL, CREATE);
	dip = gfs_get_istruct(sdp, gl, i->ino, CREATE);
	ip = iget(sb, i->ino);
	gfs_put_istruct(dip);
	gfs_put_glstruct(gl);
	if (!ip)
	  return NULL;
	dip = vi2ino(ip);

	/* 
         * looking in the i_dentry list for already existing dentrys, a
	 * match will be considered if both inode and parent's inode
	 * are matching, and name are matched 
	 */
	head = &ip->i_dentry;
	tmp = head;
	printk ("Starting while loop\n");
	while ((tmp = tmp->next) != head)
	  {
	    struct dentry *entry;
	    entry = list_entry (tmp, struct dentry, d_alias);
	    printk("Entry name %s \n", entry->d_name.name);
	    dcd = (gfs_dcached_t *) entry->d_fsdata;
	    if ((entry->d_inode->i_ino == i->ino) &&
		((entry->d_inode->i_nlink == 1) ||
		 S_ISDIR (entry->d_inode->i_ino)))
	      {
		goto found_dentry;
	      }

	    /* case of regular dentry */
	    if (dcd && !(dcd->dc_dfsa_flags & DFSA_YETUNKNOWN))
	      if ((entry->d_parent->d_inode->i_ino == i->parent_ino))
		{
		  printk("Found Regular dentry checking ..\n");
		  if (gfs_dfsa_compare_names
		      (i->status, &i->path[i->first_index], DFSA_REGULAR,
		       &entry->d_name)) goto found_dentry;
		  else
		    continue;
		}
	    /* case of special dentry */
	    if (dcd && (dcd->dc_dfsa_flags & DFSA_YETUNKNOWN) &&
		dcd->dc_parent_ino == i->parent_ino)
	      {
		printk("Found maching yetunknown dentry\n");
		if (gfs_dfsa_compare_names
		    (i->status, &i->path[i->first_index],
		     DFSA_YETUNKNOWN, &entry->d_name))
		  {
found_dentry:
		    iput(ip);
		    return dget(entry);
		  }
	      }

	  }

#ifdef CONFIG_MOSIX_DEBUG
	if (ds_debug & DSDEB_DFSA_GFS)
	  {
	    printk("Not Full Name attaching to root\n");
	  }
#endif /* CONFIG_MOSIX_DEBUG */

	/* 
         * In this stage: dentry must be added as special dentry, and
	 * consistency checks passed, so dentry is not in cache. Adds
	 * it as son of root, with special name 
         */
	if (i->status == IDENT_FILE_NAME)
	  sprintf(buff, "/%.16llx/%.16llx/%s",
		   i->ino, i->parent_ino, &i->path[i->first_index]);
	else
	  sprintf(buff, "/%.16llx/%.16llx/", i->ino, i->parent_ino);
	printk("New dentry name  #%s# \n", buff);
	qname.name = buff;
	qname.len = osi_strlen(buff);
	if (!(dp = d_alloc(sb->s_root, &qname)))
	  {
	    printk("Error in reconstruct d_alloc\n");
	    return NULL;
	  }
	dcd = (gfs_dcached_t *)gmalloc(sizeof(gfs_dcached_t));
	if (!dcd)
	  {
	    error = -ENOMEM;
	    return NULL;
	  }


	dcd->dc_vn = dip->i_vn;
	dcd->dc_parent_ino = i->parent_ino;
	dcd->dc_dfsa_flags = DFSA_YETUNKNOWN;
	vdc2dcd(dp) = dcd;
	dp->d_op = &gfs_dops;
	d_add(dp, ip);
      }
    }

  return dp;
}

#endif CONFIG_MOSIX_DFSA


