/******************************************************************************
*******************************************************************************
**
**    Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: mount.c
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"

#define __NO_VERSION__ 
#include <linux/module.h>

#include "mount_linux.h"





/**
 * atou - convert a ascii number to a unsigned int
 * @c: the number
 *
 * Returns: the number
 */

STATIC unsigned int atou(const char *c)
{
  unsigned int x = 0;
      
  while ('0' <= *c && *c <= '9')
  {
    x = x * 10 + (*c - '0');
    c++;
  }
      
  return x;
}


/**
 * gfs_make_args - Parse mount arguments
 * @data:
 * @args:
 *
 * Return: 1 if data was valid, 0 if data is not valid
 */

int gfs_make_args(char *data, gfs_args_t *args)
{
  char *this_token, *value;
  int ret = 1;
  unsigned int count = 0;

  ENTER(GFN_MAKE_ARGS);


  /*  If someone preloaded options, use those instead  */

  osi_sema_lock(&gfs_mount_args_lock);
  if (gfs_mount_args)
  {
    data = gfs_mount_args;
    count = gfs_mount_args_size;

    db_printf("Overriding mount options with:  %s\n", data);

    gfs_mount_args = NULL;
  }
  osi_sema_unlock(&gfs_mount_args_lock);


  gfs_memset(args, 0, sizeof(gfs_args_t));


  /*  Set some defaults  */

  args->ar_dir_alloc_method = GFS_DAMETH_DEFAULT;
  args->ar_buf_readahead = 8;
  args->ar_atime_quantum = GFS_ATIME_DEFAULT;


  /*  Split the options into tokens with the "," character and process
      them  */

  if (data)
  {
    for (this_token = osi_strtok(data, ","); this_token; this_token = osi_strtok(NULL, ","))
    {
      /* This splits the token in half if there is an "=" in it */
      value = osi_strchr(this_token, '=');
      if (value)
	*value++ = 0;

      /* This can be extended to only allow valid lock protocols */
      if (!osi_strcmp(this_token, "lockproto") && value)
	osi_strcpy(args->ar_lockproto, value);

      /* This can be extended to only allow valid lock table names */
      else if (!osi_strcmp(this_token, "locktable") && value)
	osi_strcpy(args->ar_locktable, value);

      /* This can be extended to only allow valid host data fields */
      else if (!osi_strcmp(this_token, "hostdata") && value)
	osi_strcpy(args->ar_hostdata, value);

      else if (!osi_strcmp(this_token, "diralloc") && value)
      {
	if (!osi_strcmp(value, "single"))
	  args->ar_dir_alloc_method = GFS_DAMETH_SINGLE;
	else if (!osi_strcmp(value, "roundrobin"))
	  args->ar_dir_alloc_method = GFS_DAMETH_RROBIN;
	else if (!osi_strcmp(value, "random"))
	  args->ar_dir_alloc_method = GFS_DAMETH_RANDOM;
	else
	  ret = 0;
      }

      else if (!osi_strcmp(this_token, "ignore_local_fs"))
	args->ar_ignore_local_fs = TRUE;

      else if (!osi_strcmp(this_token, "localflocks"))
	args->ar_localflocks = TRUE;

      else if (!osi_strcmp(this_token, "localcaching"))
	args->ar_localcaching = TRUE;

      else if (!osi_strcmp(this_token, "buf_readahead") && value)
	args->ar_buf_readahead = atou(value);

      else if (!osi_strcmp(this_token, "atime_quantum") && value)
	args->ar_atime_quantum = atou(value);

      else if (!osi_strcmp(this_token, "upgrade"))
	args->ar_upgrade = TRUE;

      else
	ret = 0;

      if (!ret)
	break;
    }
  }


  if (!ret)
    con_printf("GFS:  invalid mount option(s)\n");
  else
  {
    db_printf("ar_lockproto = \"%s\"\n", args->ar_lockproto);
    db_printf("ar_locktable = \"%s\"\n", args->ar_locktable);
    db_printf("ar_hostdata = \"%s\"\n", args->ar_hostdata);
    db_printf("ar_dir_alloc_method = %u\n", args->ar_dir_alloc_method);
    db_printf("ar_ignore_local_fs = %s\n", (args->ar_ignore_local_fs) ? "TRUE" : "FALSE");
    db_printf("ar_localflocks = %s\n", (args->ar_localflocks) ? "TRUE" : "FALSE");
    db_printf("ar_localcaching = %s\n", (args->ar_localcaching) ? "TRUE" : "FALSE");
    db_printf("ar_buf_readahead = %u\n", args->ar_buf_readahead);
    db_printf("ar_atime_quantum = %u\n", args->ar_atime_quantum);
    db_printf("ar_upgrade = %u\n", args->ar_upgrade);
  }


  /*  Free the option space  */

  if (count)
    gfree(data, count);


  EXIT(GFN_MAKE_ARGS);

  return ret;
}


/**
 * gfs_proc_write - Read in some mount options
 * @file: unused
 * @buffer: a buffer of mount options
 * @count: the length of the mount options
 * @data: unused
 *
 * Called when someone writes to /proc/fs/gfs.
 * It allows you to specify mount options when you can't do it
 * from mount.  i.e. from a inital ramdisk
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_proc_write(struct file *file, const char *buffer,
		   unsigned long count, void *data)
{
  int error;
  char *p;

  ENTER(GFN_PROC_WRITE);

  MOD_INC_USE_COUNT;
  osi_sema_lock(&gfs_mount_args_lock);


  if (gfs_mount_args)
  {
    gfree(gfs_mount_args, gfs_mount_args_size);
    gfs_mount_args = NULL;
  }

  if (!count)
  {
    error = 0;
    goto fail;
  }


  gfs_mount_args = (char *)gmalloc(count + 1);
  if (!gfs_mount_args)
  {
    error = -ENOMEM;
    goto fail;
  }
  gfs_mount_args_size = count + 1;

  error = osi_copy_from_user(gfs_mount_args, buffer, count);
  if (error)
  {
    error = -EFAULT;
    goto fail_gfree;
  }
  gfs_mount_args[count] = 0;


  /*  Get rid of extra newlines  */

  for (p = gfs_mount_args; *p; p++)
    if (*p == '\n')
      *p = 0;


  osi_sema_unlock(&gfs_mount_args_lock);
  MOD_DEC_USE_COUNT;

  return count;



 fail_gfree:
  gfree(gfs_mount_args, gfs_mount_args_size);
  gfs_mount_args = NULL;

 fail:
  osi_sema_unlock(&gfs_mount_args_lock);
  MOD_DEC_USE_COUNT;
  EXIT(GFN_PROC_WRITE);
  return error;
}

