/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portion Copyright 2000-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: dcache.c
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"

#include <linux/smp_lock.h>

#include "blklist.h"
#include "dcache.h"
#include "dir.h"
#include "gfs_linux.h"
#include "glock.h"





/**
 * gfs_drevalidate - Check directory lookup consistency
 * @dentry: the mapping to check
 * @flags: Who knows?
 *
 * Check to make sure the lookup necessary to arrive at this inode from its
 * parent is still good.
 *
 * Returns: 1 if the dentry is ok, 0 if it isn't
 */

int gfs_drevalidate(struct dentry *dentry, int flags)
{
  gfs_inode_t *dip = vi2ino(dentry->d_parent->d_inode);
  gfs_dcached_t *dcd;
  gfs_inode_t *ip;
  gfs_inum_t inum;
  int ret = 0;
  int error;

  ENTER(GFN_DREVALIDATE);

  lock_kernel();


  GFS_ASSERT(dip, );

  if (!dentry->d_inode || is_bad_inode(dentry->d_inode))
    goto out;

  ip = vi2ino(dentry->d_inode);
  GFS_ASSERT(ip, );


  error = gfs_glock_i(dip, GL_SHARED);
  if (error) 
    goto out_print;

  dcd = vdc2dcd(dentry);


  if (dcd && dcd->dc_vn == dip->i_gl->gl_vn)
  {
    ret = 1;
    goto out_gunlock;
  }

  error = gfs_dir_search(dip, &dentry->d_name, &inum, NULL);
  if (error)
    goto out_gunlock;

  if (inum.no_formal_ino == ip->i_num.no_formal_ino)
  {
    if (dcd)
      dcd->dc_vn = dip->i_gl->gl_vn;
    ret = 1;
  }


 out_gunlock:
  gfs_gunlock_i(dip, 0);

 out_print:
  IFDB(GFN_DREVALIDATE,
       db_printf("(%"FMT_64"u, %s --> %"FMT_64"u) is %s\n",
		 dip->i_num.no_formal_ino,
		 dentry->d_name.name,
		 ip->i_num.no_formal_ino,
		 (ret) ? "valid" : "invalid"););

 out:
  unlock_kernel();

  EXIT(GFN_DREVALIDATE);

  return ret;
}


/**
 * gfs_drelease - Free the GFS specific data from a dentry about to be freed
 * @dentry: the dentry being freed
 *
 */

void gfs_drelease(struct dentry *dentry)
{
  gfs_dcached_t *dcd = vdc2dcd(dentry);

  ENTER(GFN_DRELEASE);

  if (dcd)
    gfree(dcd, sizeof(gfs_dcached_t));

  vdc2dcd(dentry) = NULL;

  EXIT(GFN_DRELEASE);
}


struct dentry_operations gfs_dops = 
{
  d_revalidate:         gfs_drevalidate,
  d_release:            gfs_drelease
};
