/******************************************************************************
*******************************************************************************
**
**    Copyright 2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: glops.c
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "blklist.h"
#include "dio.h"
#include "glock.h"
#include "inode.h"
#include "log.h"
#include "page.h"
#include "super.h"





/**
 * sync_meta - Sync the dirty data for a regular metadata glock
 * @sdp: the filesystem
 * @gl: the glock
 * @flags: GFS_SYNC_WAIT or GFS_SYNC_INVISIBLE
 *
 */

STATIC void sync_meta(gfs_sbd_t *sdp, gfs_glock_t *gl, int flags)
{
  ENTER(GFN_SYNC_META);

  if (osi_test_bit(GLF_DIRTY, &gl->gl_flags))
  {
    gfs_log_flush(sdp, gl);
    gfs_sync_bh(sdp, gl, (flags & GFS_SYNC_WAIT));
  }
  if (!(flags & GFS_SYNC_INVISIBLE))
  {
    osi_clear_bit(GLF_DIRTY, &gl->gl_flags);
    osi_clear_bit(GLF_SYNC, &gl->gl_flags);
  }

  EXIT(GFN_SYNC_META);
}


/**
 * release_meta - prepare a regular metadata glock to be released
 * @sdp: the filesystem
 * @gl: the glock
 *
 */

STATIC void release_meta(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  ENTER(GFN_RELEASE_META);

  sync_meta(sdp, gl, GFS_SYNC_WAIT);

  if (!sdp->sd_args.ar_localcaching)
  {
    gfs_inval_bh(sdp, gl);
    gl->gl_vn++;
  }

  EXIT(GFN_RELEASE_META);
}


/**
 * sync_inode - Sync the dirty data for a inode glock
 * @sdp: the filesystem
 * @gl: the glock
 * @flags: GFS_SYNC_WAIT or GFS_SYNC_INVISIBLE
 *
 */

STATIC void sync_inode(gfs_sbd_t *sdp, gfs_glock_t *gl, int flags)
{
  ENTER(GFN_SYNC_INODE);

  if (osi_test_bit(GLF_DIRTY, &gl->gl_flags))
  {
    gfs_log_flush(sdp, gl);
    gfs_sync_pg(sdp, gl, (flags & GFS_SYNC_WAIT));
    gfs_sync_bh(sdp, gl, (flags & GFS_SYNC_WAIT));
  }
  if (!(flags & GFS_SYNC_INVISIBLE))
  {
    osi_clear_bit(GLF_DIRTY, &gl->gl_flags);
    osi_clear_bit(GLF_SYNC, &gl->gl_flags);
  }

  EXIT(GFN_SYNC_INODE);
}


/**
 * release_inode - prepare a inode glock to be released
 * @sdp: the filesystem
 * @gl: the glock
 *
 */

STATIC void release_inode(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  gfs_inode_t *ip = (gfs_inode_t *)gl->gl_object;

  ENTER(GFN_RELEASE_INODE);

  if (ip)
    gfs_flush_meta_cache(ip, 0);

  sync_inode(sdp, gl, GFS_SYNC_WAIT);

  if (!sdp->sd_args.ar_localcaching)
  {
    gfs_inval_pg(sdp, gl);
    gfs_inval_bh(sdp, gl);
    gl->gl_vn++;
  }

  EXIT(GFN_RELEASE_INODE);
}


/**
 * lock_inode - operation done after an inode lock is locked by a process
 * @sdp: the filesystem
 * @gl: the glock
 * @flags: the flags passed into gfs_glock()
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int lock_inode(gfs_sbd_t *sdp, gfs_glock_t *gl, int flags)
{
  gfs_inode_t *ip = (gfs_inode_t *)gl->gl_object;
  int lockone = !(flags & GL_SKIP) && (osi_atomic_read(&gl->gl_locked) == 1);
  int error = 0;

  ENTER(GFN_LOCK_INODE);

  if (lockone &&
      ip &&
      ip->i_vn != gl->gl_vn)
    error = gfs_copyin_dinode(ip);

  EXIT(GFN_LOCK_INODE);

  return error;
}


/**
 * unlock_inode - operation done before an inode lock is unlocked by a process
 * @sdp: the filesystem
 * @gl: the glock
 * @flags: the flags passed into gfs_gunlock()
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC void unlock_inode(gfs_sbd_t *sdp, gfs_glock_t *gl, int flags)
{
  gfs_inode_t *ip = (gfs_inode_t *)gl->gl_object;
  int lockone = !(flags & GL_SKIP) && (osi_atomic_read(&gl->gl_locked) == 1);

  ENTER(GFN_UNLOCK_INODE);

  if (osi_test_and_clear_bit(GLF_POISONED, &gl->gl_flags))
    gl->gl_vn++;
  else if (lockone &&
	   ip &&
	   osi_test_bit(GLF_DIRTY, &gl->gl_flags))
    gfs_inode_attr_in(ip);

  EXIT(GFN_UNLOCK_INODE);
}


/**
 * acquire_rgrp - operation done after a rgrp lock is acquired
 * @sdp: the filesystem
 * @gl: the glock
 *
 */

STATIC void acquire_rgrp(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  gfs_rgrpd_t *rgd = (gfs_rgrpd_t *)gl->gl_object;

  ENTER(GFN_ACQUIRE_RGRP);

  GFS_ASSERT(rgd && gl->gl_lvb, );

  gfs_rgrp_save_in(&rgd->rd_rs, gl->gl_lvb);

  EXIT(GFN_ACQUIRE_RGRP);
}


/**
 * release_rgrp - prepare a rgrp glock to be released
 * @sdp: the filesystem
 * @gl: the glock
 *
 */

STATIC void release_rgrp(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  gfs_rgrpd_t *rgd = (gfs_rgrpd_t *)gl->gl_object;

  ENTER(GFN_RELEASE_RGRP);

  GFS_ASSERT(rgd && gl->gl_lvb, );

  sync_meta(sdp, gl, GFS_SYNC_WAIT);

  if (!sdp->sd_args.ar_localcaching)
  {
    gfs_inval_bh(sdp, gl);
    gl->gl_vn++;
  }

  if (rgd->rd_rs_dirty)
  {
    rgd->rd_rs.rs_owner = 0;
    gfs_rgrp_save_out(&rgd->rd_rs, gl->gl_lvb);

    rgd->rd_rs_dirty = FALSE;
  }

  EXIT(GFN_RELEASE_RGRP);
}


/**
 * lock_rgrp - operation done after an rgrp lock is locked by a process
 * @sdp: the filesystem
 * @gl: the glock
 * @flags: the flags passed into gfs_glock()
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int lock_rgrp(gfs_sbd_t *sdp, gfs_glock_t *gl, int flags)
{
  gfs_rgrpd_t *rgd = (gfs_rgrpd_t *)gl->gl_object;
  int lockone = !(flags & GL_SKIP) && (osi_atomic_read(&gl->gl_locked) == 1);
  int error = 0;

  ENTER(GFN_LOCK_RGRP);

  GFS_ASSERT(rgd && gl->gl_lvb, );

  if (lockone)
    error = gfs_rgrp_read(rgd);

  EXIT(GFN_LOCK_RGRP);

  return error;
}


/**
 * unlock_rgrp - operation done before an rgrp lock is unlocked by a process
 * @sdp: the filesystem
 * @gl: the glock
 * @flags: the flags passed into gfs_gunlock()
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC void unlock_rgrp(gfs_sbd_t *sdp, gfs_glock_t *gl, int flags)
{
  gfs_rgrpd_t *rgd = (gfs_rgrpd_t *)gl->gl_object;
  int lockone = !(flags & GL_SKIP) && (osi_atomic_read(&gl->gl_locked) == 1);

  ENTER(GFN_UNLOCK_RGRP);

  GFS_ASSERT(rgd && gl->gl_lvb, );

  if (lockone)
    gfs_rgrp_relse(rgd);

  if (osi_test_and_clear_bit(GLF_POISONED, &gl->gl_flags))
    gl->gl_vn++;
  else if (lockone &&
	   osi_test_bit(GLF_DIRTY, &gl->gl_flags))
    gfs_rgrp_lvb_fill(rgd);

  EXIT(GFN_UNLOCK_RGRP);
}


/**
 * release_trans - prepare the transaction glock to be released
 * @sdp: the filesystem
 * @gl: the glock
 *
 * We want to sync the device even with localcaching.  Remember
 * that localcaching journal replay only marks buffers dirty.
 */

STATIC void release_trans(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  ENTER(GFN_RELEASE_TRANS);

  gfs_log_flush(sdp, NULL);
  if (osi_sync_dev(sdp->sd_vsb, TRUE))
    GFS_ASSERT(0, con_printf("Panicking because of write error\n"););

  EXIT(GFN_RELEASE_TRANS);
}





gfs_glock_operations_t gfs_meta_glops =
{
  go_sync:              sync_meta,
  go_release:           release_meta,
  go_free:              gfs_free_bh,
  go_type:              LM_TYPE_META,
  go_name:              "Misc On-Disk Metadata"
};

gfs_glock_operations_t gfs_inode_glops =
{
  go_sync:              sync_inode,
  go_release:           release_inode,
  go_lock:              lock_inode,
  go_unlock:            unlock_inode,
  go_free:              gfs_free_bh,
  go_type:              LM_TYPE_INODE,
  go_name:              "Inode"
};

gfs_glock_operations_t gfs_rgrp_glops =
{
  go_sync:              sync_meta,
  go_acquire:           acquire_rgrp,
  go_release:           release_rgrp,
  go_lock:              lock_rgrp,
  go_unlock:            unlock_rgrp,
  go_free:              gfs_free_bh,
  go_type:              LM_TYPE_RGRP,
  go_name:              "Inode"
};

gfs_glock_operations_t gfs_nopen_glops =
{
  go_type:              LM_TYPE_NOPEN,
  go_name:              "N-Open"
};

gfs_glock_operations_t gfs_flock_glops =
{
  go_type:              LM_TYPE_FLOCK,
  go_name:              "Flock"
};

gfs_glock_operations_t gfs_plock_head_glops =
{
  go_type:              LM_TYPE_PLOCK_HEAD,
  go_name:              "Plock head"
};

gfs_glock_operations_t gfs_plock_glops =
{
  go_type:              LM_TYPE_PLOCK,
  go_name:              "Plock"
};

gfs_glock_operations_t gfs_nondisk_glops =
{
  go_type:              LM_TYPE_NONDISK,
  go_name:              "Non-Disk"
};

gfs_glock_operations_t gfs_trans_glops =
{
  go_release:           release_trans,
  go_type:              LM_TYPE_NONDISK,
  go_name:              "Trans-Lock"
};
