/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portions Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: main.c
**
*******************************************************************************
******************************************************************************/

#define EXTERN
#define WANT_DEBUG_NAMES
#include "gfs.h"

#define __NO_VERSION__ 
#ifdef DEBUG_SYMBOLS
#define EXPORT_SYMTAB
#endif  /*  DEBUG_SYMBOLS  */
#include <linux/module.h>
#include <linux/version.h>
char kernel_version [] = UTS_RELEASE;
#include <linux/init.h>
#include <linux/kmod.h>
#include <linux/proc_fs.h>

#include "blklist.h"
#include "bmap.h"
#include "dcache.h"
#include "dio.h"
#include "fs.h"
#include "inode_linux.h"
#include "mount_linux.h"
#include "super_linux.h"





static DECLARE_FSTYPE_DEV(gfs_fs_type, "gfs", gfs_read_super);


struct proc_dir_entry *gfs_proc_entry = NULL;


#ifndef DEBUG_MEM
static kmem_cache_t *gfs_glock_cachep = NULL;
static kmem_cache_t *gfs_inode_cachep = NULL;
static kmem_cache_t *gfs_bufdata_cachep = NULL;
static kmem_cache_t *gfs_rgrpd_cachep = NULL;
#endif





gfs_glock_t *gfs_alloc_glock(void)
{
#ifdef DEBUG_MEM
  return gmalloc(sizeof(gfs_glock_t));
#else
  return kmem_cache_alloc(gfs_glock_cachep, GFP_BUFFER);
#endif
}

void gfs_free_glock(gfs_glock_t *gl)
{
#ifdef DEBUG_MEM
  gfree(gl, sizeof(gfs_glock_t));
#else
  kmem_cache_free(gfs_glock_cachep, gl);
#endif
}

gfs_inode_t *gfs_alloc_inode(void)
{
#ifdef DEBUG_MEM
  return gmalloc(sizeof(gfs_inode_t));
#else
  return kmem_cache_alloc(gfs_inode_cachep, GFP_BUFFER);
#endif
}

void gfs_free_inode(gfs_inode_t *ip)
{
#ifdef DEBUG_MEM
  gfree(ip, sizeof(gfs_inode_t));
#else
  kmem_cache_free(gfs_inode_cachep, ip);
#endif
}

gfs_bufdata_t *gfs_alloc_bufdata(void)
{
#ifdef DEBUG_MEM
  return gmalloc(sizeof(gfs_bufdata_t));
#else
  return kmem_cache_alloc(gfs_bufdata_cachep, GFP_BUFFER);
#endif
}

void gfs_free_bufdata(gfs_bufdata_t *bd)
{
#ifdef DEBUG_MEM
  gfree(bd, sizeof(gfs_bufdata_t));
#else
  kmem_cache_free(gfs_bufdata_cachep, bd);
#endif
}

gfs_rgrpd_t *gfs_alloc_rgrpd(void)
{
#ifdef DEBUG_MEM
  return gmalloc(sizeof(gfs_rgrpd_t));
#else
  return kmem_cache_alloc(gfs_rgrpd_cachep, GFP_KERNEL);
#endif
}

void gfs_free_rgrpd(gfs_rgrpd_t *rgd)
{
#ifdef DEBUG_MEM
  gfree(rgd, sizeof(gfs_rgrpd_t));
#else
  kmem_cache_free(gfs_rgrpd_cachep, rgd);
#endif
}


/**
 * init_gfs_fs - Register GFS as a filesystem
 *
 * Returns: 0 on success, error code on failure
 */

int __init init_gfs_fs(void)
{
  int error;


#ifndef DEBUG_MEM
  gfs_glock_cachep = kmem_find_general_cachep(sizeof(gfs_glock_t), GFP_BUFFER);
  gfs_inode_cachep = kmem_find_general_cachep(sizeof(gfs_inode_t), GFP_BUFFER);
  gfs_bufdata_cachep = kmem_find_general_cachep(sizeof(gfs_bufdata_t), GFP_BUFFER);
  gfs_rgrpd_cachep = kmem_find_general_cachep(sizeof(gfs_rgrpd_t), GFP_BUFFER);
  
  if (!gfs_glock_cachep || !gfs_inode_cachep ||
      !gfs_bufdata_cachep || !gfs_rgrpd_cachep)
  {
    osi_printf("GFS: cannot create GFS kmem caches\n");
    return -ENOMEM;
  }
#endif


  error = register_filesystem(&gfs_fs_type);
  if (error)
    return error;


  osi_sema_init(&gfs_mount_args_lock);

  gfs_proc_entry = create_proc_read_entry("fs/gfs", S_IFREG | 0200, NULL, NULL, NULL);
  gfs_proc_entry->write_proc = gfs_proc_write;


  /*  Initialize the random number generator  */

  gfs_random_number = (uint32)(osi_gettimeofday() & 0x00000000FFFFFFFFllu);


#ifdef DEBUG_MEM
  {
    int x;
    for (x = 0; x < GMALLOC_HASH_SIZE; x++)
      osi_list_head_init(&gmalloc_head[x]);
  }
  gfs_memset(&gm_stats, 0, sizeof(gmalloc_stats_t));
  osi_spinlock_init(&gmalloc_lock);
#endif

#ifdef DEBUG_BUF
  {
    int x;
    for (x = 0; x < BUF_HASH_SIZE; x++)
      osi_list_head_init(&buf_head[x]);
  }
  gfs_memset(&buf_stats, 0, sizeof(buf_stats_t));
  osi_spinlock_init(&buf_lock);
#endif

#ifdef DEBUG_TRACE
  gfs_memset(gfs_debug_flags, 0, DEBUG_FLAG_BYTES);
#define dbon(x) gfs_debug_flags[(x) / 8] |= (1 << ((x) % 8))
  dbon(GFN_FIND_JHEAD);
  dbon(GFN_FOREACH_DESCRIPTOR);
  dbon(GFN_DO_REPLAY_LOCAL);
  dbon(GFN_DO_REPLAY_MULTI);
  dbon(GFN_REPLAY_METADATA);
  dbon(GFN_RECOVER_JOURNAL);
  dbon(GFN_RECOVER_NOPEN);
#endif  /*  DEBUG_TRACE  */

#ifdef DEBUG_PRINT
  gfs_memset(gfs_icdb_buf, 0, GFS_ICDB_BUF_LEN);
  osi_sema_init(&gfs_icdb_lock);
#endif  /*  DEBUG_PRINT  */

#ifdef DEBUG_PROFILE
  {
    int x;
    gfs_memset(gfs_prof_stats, 0, DEBUG_FLAGS * sizeof(gfs_profile_t));
    for (x = 0; x < DEBUG_FLAGS; x++)
      gfs_prof_stats[x].min_micros = (uint64)-1;
  }
#endif  /*  DEBUG_PROFILE  */


  con_printf("GFS %s (built %s %s) installed\n",
	     GFS_RELEASE_NAME,
	     __DATE__, __TIME__);


  db_printf("sizeof(gfs_meta_header_t) = %d\n", (int)sizeof(gfs_meta_header_t));
  db_printf("sizeof(gfs_sb_t) = %d\n", (int)sizeof(gfs_sb_t));
  db_printf("sizeof(gfs_jindex_t) = %d\n", (int)sizeof(gfs_jindex_t));
  db_printf("sizeof(gfs_rindex_t) = %d\n", (int)sizeof(gfs_rindex_t));
  db_printf("sizeof(gfs_rgrp_t) = %d\n", (int)sizeof(gfs_rgrp_t));
  db_printf("sizeof(gfs_indirect_t) = %d\n", (int)sizeof(gfs_indirect_t));
  db_printf("sizeof(gfs_dinode_t) = %d\n", (int)sizeof(gfs_dinode_t));
  db_printf("sizeof(gfs_dirent_t) = %d\n", (int)sizeof(gfs_dirent_t));
  db_printf("sizeof(gfs_leaf_t) = %d\n", (int)sizeof(gfs_leaf_t));
  db_printf("sizeof(gfs_log_header_t) = %d\n", (int)sizeof(gfs_log_header_t));
  db_printf("sizeof(gfs_log_descriptor_t) = %d\n", (int)sizeof(gfs_log_descriptor_t));
  db_printf("sizeof(gfs_block_tag_t) = %d\n", (int)sizeof(gfs_block_tag_t));
  db_printf("\n");
  db_printf("sizeof(gfs_bitmap_t) = %d\n", (int)sizeof(gfs_bitmap_t));
  db_printf("sizeof(gfs_rgrpd_t) = %d\n", (int)sizeof(gfs_rgrpd_t));
  db_printf("sizeof(gfs_glock_t) = %d\n", (int)sizeof(gfs_glock_t));
  db_printf("sizeof(gfs_bufdata_t) = %d\n", (int)sizeof(gfs_bufdata_t));
  db_printf("sizeof(gfs_inode_t) = %d\n", (int)sizeof(gfs_inode_t));
  db_printf("sizeof(gfs_trans_t) = %d\n", (int)sizeof(gfs_trans_t));
  db_printf("sizeof(gfs_sbd_t) = %d\n", (int)sizeof(gfs_sbd_t));
  db_printf("sizeof(gfs_dcached_t) = %d\n", (int)sizeof(gfs_dcached_t));


  return 0;
}


/**
 * exit_gfs_fs - Unregister the file system
 *
 */

void __exit exit_gfs_fs(void)
{
  /*  Free any unused mount options  */

  osi_sema_lock(&gfs_mount_args_lock);
  if (gfs_mount_args)
  {
    gfree(gfs_mount_args, gfs_mount_args_size);
    gfs_mount_args = NULL;
  }
  osi_sema_unlock(&gfs_mount_args_lock);

  remove_proc_entry("fs/gfs", NULL);

  unregister_filesystem(&gfs_fs_type);

  buf_check();
  gmalloc_check();
}

MODULE_DESCRIPTION("Global File System");
MODULE_AUTHOR("The GFS Group");

module_init(init_gfs_fs);
module_exit(exit_gfs_fs);

/*  Symbols  */

#ifndef DEBUG_SYMBOLS
EXPORT_NO_SYMBOLS;
#endif  /*  DEBUG_SYMBOLS  */

