/******************************************************************************
*******************************************************************************
**
**    Copyright 2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: page.c
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"

#include <asm/pgalloc.h>
#include <linux/blkdev.h>

#include "dio.h"





/**
 * __zap_mapping_list - Zap a list of VM areas
 * @vma: The first entry in the list of VM areas
 *
 */

STATIC void __zap_mapping_list(struct vm_area_struct *vma)
{
  ENTER(GFN_ZAP_MAPPING_LIST);

  for ( ; vma; vma = vma->vm_next_share)
  {
    struct mm_struct *mm = vma->vm_mm;
    unsigned long start = vma->vm_start;
    unsigned long end = vma->vm_end;
    unsigned long len = end - start;

    flush_cache_range(mm, start, end);
    zap_page_range(mm, start, len);
    flush_tlb_range(mm, start, end);
  }

  EXIT(GFN_ZAP_MAPPING_LIST);
}


/**
 * zap_inode_mapping - Ensures a page fault on address space access
 * @mapping: The address space
 *
 */

STATIC void zap_inode_mapping(struct address_space *mapping)
{
  ENTER(GFN_ZAP_INODE_MAPPING);

  spin_lock(&mapping->i_shared_lock);
  __zap_mapping_list(mapping->i_mmap_shared);
  spin_unlock(&mapping->i_shared_lock);

  EXIT(GFN_ZAP_INODE_MAPPING);
}


/**
 * gfs_inval_pg - Invalidate all pages associated with a glock
 * @sdp: the filesystem
 * @gl: the glock
 *
 */

void gfs_inval_pg(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  gfs_inode_t *ip = (gfs_inode_t *)gl->gl_object;
  struct inode *inode;

  ENTER(GFN_INVAL_PG);

  if (ip)
  {
    inode = ip->i_vino;
    if (inode)
    {
      zap_inode_mapping(inode->i_mapping);
      if (inode->i_mapping)
        truncate_inode_pages(inode->i_mapping, 0);
    }
  }

  EXIT(GFN_INVAL_PG);
}


/**
 * gfs_sync_pg - sync the pages associated with a glock
 * @sdp: the filesystem
 * @gl: the glock
 * @wait:  if TRUE, wait for sync
 *
 * We probably also want to do a filemap_fdatasync() here
 * when we do shared-writeable mappings.
 *
 */

void gfs_sync_pg(gfs_sbd_t *sdp, gfs_glock_t *gl, int wait)
{
  gfs_inode_t *ip = (gfs_inode_t *)gl->gl_object;
  unsigned long last = -1;

  ENTER(GFN_SYNC_PG);

  if (ip)
  {
    struct inode *inode = ip->i_vino;

    /*  We should probably start worrying about the inode going away
	here.  */

    if (inode)
    {
      /*  This code doesn't necessarily handle pages dirtied by a
	  shared, writeable mapping.  We need something like
	  filemap_fdatasync() and filemap_fdatawait().

	  Even with that, there are still big problems on glock release
	  because there is a window after this sync stuff finished, but
	  before zap_inode_mapping() is completed.  */

      if (generic_buffer_fdatasync(inode, 0, last))
	GFS_ASSERT(0, con_printf("Panicking because of write error\n"););
    }
  }

  EXIT(GFN_SYNC_PG);
}


