/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portions Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: super.c
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"

#define __NO_VERSION__ 
#include <linux/module.h>
#include <linux/sched.h>
#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>
#include <linux/blkdev.h>

#include "blklist.h"
#include "bmap.h"
#include "daemon.h"
#include "dcache.h"
#include "dio.h"
#include "gfs_linux.h"
#include "glock.h"
#include "glops.h"
#include "file_linux.h"
#include "fs.h"
#include "inode.h"
#include "inode_linux.h"
#include "locking.h"
#include "log.h"
#include "mount_linux.h"
#include "recovery.h"
#include "super.h"
#include "super_linux.h"
#include "trans.h"





/**
 * gfs_read_super - Read in superblock
 * @sb: The VFS superblock
 * @data: Mount options
 * @silent: Don't complain if its not a GFS filesystem
 *
 * Returns: The VFS superblock, or NULL on error
 */

struct super_block *gfs_read_super(struct super_block *sb, void *data, int silent)
{
  gfs_sbd_t *sdp;
  gfs_glock_t *gl_sb, *gl_i;
  gfs_log_header_t head;
  gfs_bhlist_t *bl;
  gfs_tr_nop_t *nop;
  int super = TRUE, jindex = TRUE;
  int x;
  int error;

  ENTER(GFN_READ_SUPER);


  /*  Make sure there are no dirty buffers lying around  */

  if (osi_sync_dev(sb, TRUE))
  {
    con_printf("GFS: write error\n");
    error = -EIO;
    goto fail;
  }
  invalidate_buffers(sb->s_dev);


  sdp = (gfs_sbd_t *)gmalloc(sizeof(gfs_sbd_t));
  if (!sdp)
  {
    con_printf("GFS:  out of memory\n");
    goto fail;
  }
  gfs_memset(sdp, 0, sizeof(gfs_sbd_t));

  vsb2sbd(sb) = sdp;
  sdp->sd_vsb = sb;


  /*  Set up the buffer cache and fill in some fake values
      to allow us to read in the superblock.  */

  sdp->sd_sb.sb_bsize = get_hardsect_size(sb->s_dev);
  if (sdp->sd_sb.sb_bsize < GFS_BASIC_BLOCK)
    sdp->sd_sb.sb_bsize = GFS_BASIC_BLOCK;
  set_blocksize(sb->s_dev, sdp->sd_sb.sb_bsize);
  GFS_ASSERT(sizeof(gfs_sb_t) <= sdp->sd_sb.sb_bsize, );


  /* Init rgrp variables */

  osi_list_head_init(&sdp->sd_rglist);
  osi_rwlock_init(&sdp->sd_rglock);


  /*  Init Glock structures  */

  for (x = 0; x < GFS_GHASH_SIZE; x++)
  {
    osi_list_head_init(&sdp->sd_ghash[x].hb_notheld);
    osi_list_head_init(&sdp->sd_ghash[x].hb_held);
    osi_list_head_init(&sdp->sd_ghash[x].hb_perm);

    osi_rwlock_init(&sdp->sd_ghash[x].hb_lock);
  }

  sdp->sd_gl_heldtime = 300;  /*  Time in seconds  */


  if (!gfs_make_args((char *)data, &sdp->sd_args))
    goto fail_sbd_free;

  if (gfs_mount_lockproto(sdp, silent))
    goto fail_sbd_free;


  if (sdp->sd_lockstruct.ls_ops->local_fs && !sdp->sd_args.ar_ignore_local_fs)
  {
    db_printf("local filesystem\n");

    /*  Force local [p|f]locks  */
    db_printf("forcing 'localflocks' option\n");
    sdp->sd_args.ar_localflocks = TRUE;

    /*  Force local read ahead and caching  */
    db_printf("forcing 'localcaching' option\n");
    sdp->sd_args.ar_localcaching = TRUE;
  }


  /*  Start up the glockd thread  */

  osi_sema_init_lkd(&sdp->sd_glockd_sema);
  osi_atomic_set(&sdp->sd_glockd_ctrl, 5);

  error = kernel_thread(gfs_glockd, sdp, 0);
  if (error < 0)
  {
    con_printf("GFS:  can't start glockd thread (%d)\n", error);
    goto fail_gla_free;
  }
  osi_sema_lock(&sdp->sd_glockd_sema);


  
  error = gfs_glock_num(sdp, GFS_MOUNT_LOCK, &gfs_nondisk_glops, NULL,
			GL_PERM | LM_FLAG_NOEXP, NULL);
  if (error)
  {
    con_printf("GFS:  can't acquire mount glock (%d)\n", error);
    goto fail_glockd;
  }


  error = gfs_glock_num(sdp, GFS_LIVE_LOCK, &gfs_nondisk_glops, NULL,
			GL_SHARED | GL_PERM | GL_DISOWN | LM_FLAG_NOEXP, NULL);
  if (error)
  {
    con_printf("GFS:  can't acquire live glock (%d)\n", error);
    goto fail_gunlock_mount;
  }


  error = gfs_glock_num(sdp, GFS_SB_LOCK, &gfs_meta_glops, NULL,
			(sdp->sd_args.ar_upgrade) ? 0 : GL_SHARED, &gl_sb);
  if (error)
  {
    con_printf("GFS:  can't acquire superblock glock (%d)\n", error);
    goto fail_gunlock_live;
  }


  if (gfs_read_sb(sdp, gl_sb, silent) < 0)
    goto fail_gunlock_sb;


  /*  Set up the buffer cache and SB for real  */

  if (sdp->sd_sb.sb_bsize > PAGE_SIZE)
  {
    con_printf("GFS:  the blocksize (%u) is too big for the page size (%lu)\n",
	       sdp->sd_sb.sb_bsize, PAGE_SIZE);
    goto fail_gunlock_sb;
  }

  set_blocksize(sb->s_dev, sdp->sd_sb.sb_bsize);


  /*  Copy out mount flags  */

  sdp->sd_noatime = (sb->s_flags & (MS_NOATIME | MS_NODIRATIME));
  sdp->sd_sync_mount = FALSE;  /*  Let the VFS deal with it  */


  /*  Setup up Virtual Super Block  */

  sb->s_blocksize = sdp->sd_sb.sb_bsize;
  sb->s_blocksize_bits = sdp->sd_sb.sb_bsize_shift;
  sb->s_magic = GFS_MAGIC;
  sb->s_op = &gfs_sops;
  sb->s_flags |= MS_NOATIME | MS_NODIRATIME;
  sb->s_maxbytes = ~0ULL;


  gfs_compute_reserv(sdp);


  /*  Read in journal index inode  */

  error = gfs_get_jiinode(sdp);
  if (error)
  {
    con_printf("GFS:  can't get journal index inode (%d)\n", error);
    goto fail_gunlock_sb;
  }


  /*  Upgrade version numbers if we need to  */

  if (sdp->sd_args.ar_upgrade)
  {
    error = gfs_do_upgrade(sdp, gl_sb);
    if (error)
      goto fail_ji_free;
  }


  /*  Get a handle on the transaction glock  */

  sdp->sd_trans_gl = gfs_get_glstruct(sdp, GFS_TRANS_LOCK, &gfs_trans_glops, NULL, CREATE);
  if (!sdp->sd_trans_gl)
  {
    error = -EIO;
    goto fail_ji_free;
  }
  osi_set_bit(GLF_STICKY, &sdp->sd_trans_gl->gl_flags);


  /*  Load in the journal index  */

  error = gfs_jindex_hold(sdp, TRUE);
  if (error)
  {
    con_printf("GFS: can't read journal index (%d)\n", error);
    goto fail_trans_gl;
  }

  if (sdp->sd_lockstruct.ls_jid >= sdp->sd_journals)
  {
    con_printf("GFS:  there are %u journals, you're trying to mount #%u\n",
	       sdp->sd_journals, sdp->sd_lockstruct.ls_jid + 1);
    goto fail_gunlock_ji;
  }
  sdp->sd_jdesc = sdp->sd_jindex[sdp->sd_lockstruct.ls_jid];


  /*  Initialize some log stuff (some of which is needed for recovery).  */

  sdp->sd_log_seg_free = sdp->sd_jdesc.ji_nsegment - 1;

  osi_list_head_init(&sdp->sd_log_ail);
  osi_list_head_init(&sdp->sd_log_incore);
  osi_sema_init(&sdp->sd_log_lock);
  osi_list_head_init(&sdp->sd_log_bhlist_cache);
  osi_list_head_init(&sdp->sd_nopen_od_list);
  osi_list_head_init(&sdp->sd_nopen_ic_list);


  error = gfs_glock_num(sdp, sdp->sd_jdesc.ji_addr, &gfs_meta_glops, NULL,
			GL_PERM | LM_FLAG_NOEXP, &sdp->sd_journal_gl);
  if (error)
  {
    con_printf("GFS:  can't acquire the journal glock (%d)\n", error);
    goto fail_gunlock_ji;
  }


  if (sdp->sd_lockstruct.ls_first)
  {
    for (x = 0; x < sdp->sd_journals; x++)
    {
      error = gfs_recover_journal(sdp, x);
      if (error)
      {
	con_printf("GFS:  error recovering journal %u (%d)\n", x, error);
	goto fail_gunlock_journal;
      }
    }

    DO_PROF(LM_OTHERS_MAY_MOUNT,
	    sdp->sd_lockstruct.ls_ops->others_may_mount(sdp->sd_lockstruct.ls_lockspace););
    sdp->sd_lockstruct.ls_first = FALSE;
  }
  else
  {
    error = gfs_recover_journal(sdp, sdp->sd_lockstruct.ls_jid);
    if (error)
    {
      con_printf("GFS:  error recovering my journal (%d)\n", error);
      goto fail_gunlock_journal;
    }
  }


  gfs_jindex_release(sdp);
  jindex = FALSE;


  /*  Disown my Journal glock  */

  error = gfs_glock(sdp, sdp->sd_journal_gl, GL_DISOWN);
  GFS_ASSERT(!error, );
  gfs_gunlock(sdp, sdp->sd_journal_gl, 0);


  error = gfs_find_jhead(sdp, &sdp->sd_jdesc, &head);
  if (error)
  {
    con_printf("GFS:  can't find head of journal (%d)\n", error);
    goto fail_gunlock_journal;
  }
  GFS_ASSERT(head.lh_flags & GFS_LOG_HEAD_UNMOUNT, );


  /*  Initialize some head of the log stuff  */

  sdp->sd_sequence = head.lh_sequence;
  sdp->sd_log_head = head.lh_first + 1;


  /*  Start up the recover thread
      Make sure we clear all dirty bits or we might panic
      if the lock module reported us expired while we were
      replaying our journal.  */

  gfs_memset(sdp->sd_dirty_j, 0, GFS_CLIENTS / sizeof(int) / 8);

  osi_sema_init_lkd(&sdp->sd_recoverd_sema);
  osi_atomic_set(&sdp->sd_recoverd_ctrl, 60);

  error = kernel_thread(gfs_recoverd, sdp, 0);
  if (error < 0)
  {
    con_printf("GFS:  can't start recoverd thread (%d)\n", error);
    goto fail_gunlock_journal;
  }
  osi_sema_lock(&sdp->sd_recoverd_sema);


  /*  Recover nopen counts  */

  if (head.lh_last_nopd)
  {
    error = gfs_recover_nopen(sdp, &head);
    if (error)
    {
      con_printf("GFS:  can't recover nopen counts (%d)\n", error);
      goto fail_nopen_list;  /*  Clean up from a partial nopen read  */
    }
  }


  /*  Dump out the first list of open inodes.  */

  gfs_nopen_dump(sdp, TRUE);


  /*  Get a handle on the rename lock  */

  sdp->sd_rename_gl = gfs_get_glstruct(sdp, GFS_RENAME_LOCK, &gfs_nondisk_glops, NULL, CREATE);
  if (!sdp->sd_rename_gl)
  {
    error = -EIO;
    goto fail_nopen_list;
  }


  /*  Read in the resource index inode  */

  error = gfs_get_riinode(sdp);
  if (error)
  {
    con_printf("GFS:  can't get resource index inode (%d)\n", error);
    goto fail_rename_gl;
  }


  /*  Get the root inode  */

  error = gfs_glock_num(sdp, sdp->sd_sb.sb_root_di.no_formal_ino, &gfs_inode_glops, NULL, GL_SHARED, &gl_i);
  if (error)
  {
    con_printf("GFS:  can't acquire root glock (%d)\n", error);
    goto fail_ri_free;
  }

  sdp->sd_rooti = gfs_get_istruct(sdp, gl_i, &sdp->sd_sb.sb_root_di, CREATE);

  gfs_gunlock(sdp, gl_i, 0);
  gfs_put_glstruct(gl_i);

  if (!sdp->sd_rooti)
  {
    con_printf("GFS:  can't read in root inode\n");
    goto fail_ri_free;
  }


  /*  We're through with the superblock lock  */

  gfs_gunlock(sdp, gl_sb, 0);
  gfs_put_glstruct(gl_sb);
  super = FALSE;


  /*  Get the inode/dentry  */

  sb->s_root = d_alloc_root(gfs_iget(sb, sdp->sd_rooti));

  gfs_put_istruct(sdp->sd_rooti);

  if (!sb->s_root)
  {
    con_printf("GFS:  can't get root dentry\n");
    goto fail_ri_free;
  }


  /*  Start up the inoded thread  */

  osi_sema_init_lkd(&sdp->sd_inoded_sema);
  osi_atomic_set(&sdp->sd_inoded_ctrl, 5);

  error = kernel_thread(gfs_inoded, sdp, 0);
  if (error < 0)
  {
    con_printf("GFS:  can't start inoded thread (%d)\n", error);
    goto fail_dput;
  }
  osi_sema_lock(&sdp->sd_inoded_sema);


  /*  Set up the directory allocation method  */

  error = gfs_setup_dameth(sdp);
  if (error)
  {
    con_printf("GFS:  can't set directory allocation method (%d)\n", error);
    goto fail_inoded;
  }


  /*  Start listening to sync calls  */

  sb->s_dirt = 1;


  gfs_gunlock_num(sdp, GFS_MOUNT_LOCK, &gfs_nondisk_glops, NULL, GL_NOCACHE);


  EXIT(GFN_READ_SUPER);

  return sb;



 fail_inoded:
  osi_atomic_set(&sdp->sd_inoded_ctrl, 0);
  osi_wake_up_process(sdp->sd_inoded_process);
  osi_sema_lock(&sdp->sd_inoded_sema);

 fail_dput:
  dput(sb->s_root);
  while (invalidate_inodes(sdp->sd_vsb)) /*  Do Nothing  */;

 fail_ri_free:
  gfs_put_riinode(sdp);
  gfs_clear_rgrpd(sdp);

 fail_rename_gl:
  gfs_put_glstruct(sdp->sd_rename_gl);

 fail_nopen_list:
  while (!osi_list_empty(&sdp->sd_nopen_od_list))
  {
    nop = osi_list_entry(sdp->sd_nopen_od_list.next, gfs_tr_nop_t, nop_list);
    osi_list_del(&nop->nop_list);
    gfree(nop, sizeof(gfs_tr_nop_t));
  }
  while (!osi_list_empty(&sdp->sd_nopen_ic_list))
  {
    nop = osi_list_entry(sdp->sd_nopen_ic_list.next, gfs_tr_nop_t, nop_list);
    osi_list_del(&nop->nop_list);
    gfree(nop, sizeof(gfs_tr_nop_t));
  }
  while (!osi_list_empty(&sdp->sd_log_bhlist_cache))
  {
    bl = osi_list_entry(sdp->sd_log_bhlist_cache.next, gfs_bhlist_t, bl_list);
    osi_list_del(&bl->bl_list);
    gfree(bl, sizeof(gfs_bhlist_t));
  }

  osi_atomic_set(&sdp->sd_recoverd_ctrl, 0);
  osi_wake_up_process(sdp->sd_recoverd_process);
  osi_sema_lock(&sdp->sd_recoverd_sema);

 fail_gunlock_journal:
  gfs_gunlock(sdp, sdp->sd_journal_gl, 0);
  gfs_put_glstruct(sdp->sd_journal_gl);

 fail_gunlock_ji:
  if (jindex)
    gfs_jindex_release(sdp);

 fail_trans_gl:
  gfs_put_glstruct(sdp->sd_trans_gl);

 fail_ji_free:
  gfs_put_jiinode(sdp);
  gfs_clear_journals(sdp);

 fail_gunlock_sb:
  if (super)
  {
    gfs_gunlock(sdp, gl_sb, 0);
    gfs_put_glstruct(gl_sb);
  }

 fail_gunlock_live:
  gfs_gunlock_num(sdp, GFS_LIVE_LOCK, &gfs_nondisk_glops, NULL, 0);

 fail_gunlock_mount:
  gfs_gunlock_num(sdp, GFS_MOUNT_LOCK, &gfs_nondisk_glops, NULL, 0);

 fail_glockd:
  osi_atomic_set(&sdp->sd_glockd_ctrl, 0);
  osi_wake_up_process(sdp->sd_glockd_process);
  osi_sema_lock(&sdp->sd_glockd_sema);

 fail_gla_free:
  gfs_clear_gla(sdp);
  gfs_unmount_lockproto(sdp); 

 fail_sbd_free:
  gfree(sdp, sizeof(gfs_sbd_t));

 fail:
  vsb2sbd(sb) = NULL;
  sb->s_dev = NODEV;

  EXIT(GFN_READ_SUPER);
  return NULL;
}


/**
 * connect_inode_to_vnode - Connect a GFS inode to a Linux vnode
 * @ip: the GFS inode
 * @inode: the Linux vnode
 *
 */

STATIC void connect_inode_to_vnode(gfs_inode_t *ip, struct inode *inode)
{
  ENTER(GFN_CONNECT_INODE_TO_VNODE);

  gfs_hold_istruct(ip);

  /*  Horrible kludge due to race between inode creation and destruction.  */
  while (ip->i_vino)
    osi_schedule();

  ip->i_vino = inode;
  vi2ino(inode) = ip;

  gfs_inode_attr_in(ip);

  if (S_ISREG(inode->i_mode))
  {
    inode->i_op = &gfs_file_iops;
    inode->i_fop = &gfs_file_fops;
    inode->i_mapping->a_ops = &gfs_file_aops;
  }
  else if (S_ISDIR(inode->i_mode))
  {
    inode->i_op = &gfs_dir_iops;
    inode->i_fop = &gfs_dir_fops;
  }
  else if (S_ISLNK(inode->i_mode))
  {
    inode->i_op = &gfs_symlink_iops;
  }
  else
  {
    inode->i_op = &gfs_dev_iops;
    init_special_inode(inode, inode->i_mode, inode->i_rdev);
  }

  EXIT(GFN_CONNECT_INODE_TO_VNODE);
}


/**
 * gfs_iget_for_nfs - Get an inode based on it's number
 * @sdp: the filesystem
 * @inum: the inode number
 * @inode: the linux inode
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int gfs_iget_for_nfs(gfs_sbd_t *sdp, gfs_inum_t *inum, struct inode *inode)
{
  gfs_inode_t *ip;
  gfs_glock_t *gl;
  int error;

  ENTER(GFN_IGET_FOR_NFS);


  error = gfs_glock_num(sdp, inum->no_formal_ino, &gfs_inode_glops, NULL, GL_SHARED, &gl);
  if (error)
    goto fail;

  ip = gfs_get_istruct(sdp, gl, inum, CREATE);
  if (!ip)
    goto fail_gunlock;

  if (ip->i_di.di_flags & GFS_DIF_UNUSED)
  {
    error = -EINVAL;
    goto fail_idestroy;
  }

  connect_inode_to_vnode(ip, inode);

  gfs_put_istruct(ip);
  
  gfs_gunlock(sdp, gl, 0);
  gfs_put_glstruct(gl);


  EXIT(GFN_IGET_FOR_NFS);

  return 0;



 fail_idestroy:
  gfs_put_istruct(ip);
  gfs_destroy_istruct(ip);

 fail_gunlock:
  gfs_gunlock(sdp, gl, 0);
  gfs_put_glstruct(gl);

 fail:
  EXIT(GFN_IGET_FOR_NFS);
  return error;  
}


/**
 * gfs_panic_inode - Panic
 * @inode: A VFS inode
 *
 * This is just here to make KNFSD happy.  It's never called.
 */

STATIC void gfs_panic_inode(struct inode *inode)
{
  ENTER(GFN_PANIC_INODE);

  GFS_ASSERT(0, );

  EXIT(GFN_PANIC_INODE);
}


/**
 * gfs_read_inode - Read an inode
 * @inode: A VFS inode
 * @opaque: The inum of the dinode to read in
 *
 * Pull in an inode out of the GFS inode cache.
 *
 * This will change when we start making a distinction between
 * formal inode numbers and disk addresses.
 *
 */

STATIC void gfs_read_inode(struct inode *inode, void *opaque)
{
  ENTER(GFN_READ_INODE);


  if (opaque)
  {
    gfs_inode_t *ip;

    /*  This section runs when we're called from GFS via gfs_iget()/iget4().  */

    GFS_ASSERT(opaque, );
    ip = (gfs_inode_t *)opaque;

    connect_inode_to_vnode(ip, inode);
  }
  else
  {
    gfs_inum_t inum;
    int error;

    /*  This section runs when we're called from KNFSD.
	When we start using the 2.4 NFS patches, KNFSD will call some
	other function and this function won't contain this section.  */

    inum.no_formal_ino = inode->i_ino;
    inum.no_addr = inode->i_ino;  /*  This won't always be true  */

    error = gfs_iget_for_nfs(vsb2sbd(inode->i_sb), &inum, inode);
    if (error)
      make_bad_inode(inode);
  }


  EXIT(GFN_READ_INODE);
}


/**
 * gfs_statfs - Gather and return stats about the filesystem
 * @sb: The superblock
 * @statfsbuf: The buffer
 *
 * Returns: 0 on success or error code
 */

STATIC int gfs_statfs(struct super_block *sb, struct statfs *sfs)
{
  gfs_sbd_t *sdp = vsb2sbd(sb);
  gfs_usage_t usage;
  int error;

  ENTER(GFN_STATFS);


  error = gfs_stat_gfs(sdp, &usage, TRUE);
  if (error)
  {
    EXIT(GFN_STATFS);
    return error;
  }


  gfs_memset(sfs, 0, sizeof(struct statfs));

  sfs->f_type = GFS_MAGIC;
  sfs->f_bsize = usage.block_size;
  sfs->f_blocks = usage.total_blocks -
    (usage.used_dinode + usage.free_dinode) -
      (usage.used_meta + usage.free_meta);
  sfs->f_bfree = usage.free;
  sfs->f_bavail = usage.free;
  sfs->f_files = usage.used_dinode + usage.free_dinode;
  sfs->f_ffree = usage.free_dinode;
  sfs->f_namelen = GFS_FNAMESIZE;


  EXIT(GFN_STATFS);

  return 0;
}


/**
 * gfs_clear_inode - Deallocate an inode when VFS is done with it
 * @inode: The VFS inode
 *
 */

STATIC void gfs_clear_inode(struct inode *inode)
{
  gfs_inode_t *ip;
  gfs_sbd_t *sdp;

  ENTER(GFN_CLEAR_INODE);


  ip = vi2ino(inode);
  if (ip)
  {
    sdp = ip->i_sbd;

    ip->i_vino = NULL;
    vi2ino(inode) = NULL;

    gfs_put_istruct(ip);

    osi_atomic_inc(&sdp->sd_free_icount);
    if (osi_atomic_read(&sdp->sd_free_icount) > 50)
      osi_wake_up_process(sdp->sd_glockd_process);
  }


  EXIT(GFN_CLEAR_INODE);
}


/**
 * gfs_remount_fs - called when the FS is remounted
 * @sb:  the filesystem
 * @flags:  the remount flags
 * @data:  extra data passed in (not used right now)
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int gfs_remount_fs(struct super_block *sb, int *flags, char *data)
{
  gfs_sbd_t *sdp = vsb2sbd(sb);

  ENTER(GFN_REMOUNT_FS);

  sdp->sd_noatime = (*flags & (MS_NOATIME | MS_NODIRATIME));
  sdp->sd_sync_mount = FALSE;  /*  Let the VFS deal with it  */

  /*  Don't let the VFS update atimes.  */
  *flags |= MS_NOATIME | MS_NODIRATIME;

  EXIT(GFN_REMOUNT_FS);

  return 0;
}


/**
 * gfs_write_super - disk commit all incore transactions
 * @sb: the filesystem
 *
 * This function is called every time sync(2) is called.
 * After this exits, all dirty buffers and synced.
 */

STATIC void gfs_write_super(struct super_block *sb)
{
  gfs_sbd_t *sdp = vsb2sbd(sb);

  ENTER(GFN_WRITE_SUPER);

  gfs_log_flush(sdp, NULL);
  gfs_pull_tail(sdp);

  EXIT(GFN_WRITE_SUPER);
}


/**
 * gfs_put_super - Unmount the filesystem
 * @sb: The VFS superblock
 *
 */

STATIC void gfs_put_super(struct super_block *sb)
{
  gfs_sbd_t *sdp = vsb2sbd(sb);
  gfs_tr_nop_t *nop;
  gfs_bhlist_t *bl;

  ENTER(GFN_PUT_SUPER);


  /*  Get rid of any VFS inodes laying around  */
  while (invalidate_inodes(sdp->sd_vsb)) /*  Do Nothing  */;

  /*  Kill off the inode thread  */
  osi_atomic_set(&sdp->sd_inoded_ctrl, 0);
  osi_wake_up_process(sdp->sd_inoded_process);
  osi_sema_lock(&sdp->sd_inoded_sema);

  gfs_put_riinode(sdp);

  /*  Kill off the recoverd thread  */
  osi_atomic_set(&sdp->sd_recoverd_ctrl, 0);
  osi_wake_up_process(sdp->sd_recoverd_process);
  osi_sema_lock(&sdp->sd_recoverd_sema);

  gfs_put_jiinode(sdp);

  /*  Kill off the glock thread  */
  osi_atomic_set(&sdp->sd_glockd_ctrl, 0);
  osi_wake_up_process(sdp->sd_glockd_process);
  osi_sema_lock(&sdp->sd_glockd_sema);

  /*  Stop getting sync commands  */
  sb->s_dirt = 0;

  /*  Flush out all incore transactions  */
  gfs_log_flush(sdp, NULL);

  /*  Sync out all dirty inodes and data  */
  if (osi_sync_dev(sb, TRUE))
    GFS_ASSERT(0, con_printf("Panicking because of write error\n"););

  /*  Dump out what's left of the nopen list.  */
  gfs_nopen_dump(sdp, TRUE);

  /*  Clear out the AIL  */
  gfs_pull_tail(sdp);

  /*  Shutdown the log manager  */
  gfs_shutdown_log(sdp);

  /*  At this point, we're through modifying the disk  */


  /*  Release the journal lock and the mount lock  */

  gfs_put_glstruct(sdp->sd_trans_gl);
  gfs_put_glstruct(sdp->sd_rename_gl);

  gfs_gunlock(sdp, sdp->sd_journal_gl, 0);
  gfs_put_glstruct(sdp->sd_journal_gl);

  gfs_gunlock_num(sdp, GFS_LIVE_LOCK, &gfs_nondisk_glops, NULL, 0);


  /*  Get rid of rgrp bitmap structures  */
  gfs_clear_rgrpd(sdp);
  gfs_clear_journals(sdp);

  /*  Take apart glock structures and buffer lists  */
  gfs_clear_gla(sdp);

  /*  Unmount the locking protocol  */
  gfs_unmount_lockproto(sdp);

  /*  At this point, we're through participating in the lockspace  */


  /*  Get rid of any nopen count structures  */

  while (!osi_list_empty(&sdp->sd_nopen_od_list))
  {
    nop = osi_list_entry(sdp->sd_nopen_od_list.next, gfs_tr_nop_t, nop_list);
    osi_list_del(&nop->nop_list);
    gfree(nop, sizeof(gfs_tr_nop_t));
  }
  while (!osi_list_empty(&sdp->sd_nopen_ic_list))
  {
    nop = osi_list_entry(sdp->sd_nopen_ic_list.next, gfs_tr_nop_t, nop_list);
    osi_list_del(&nop->nop_list);
    gfree(nop, sizeof(gfs_tr_nop_t));
  }


  /*  Get rid of the gfs_bhlist_t cache  */

  while (!osi_list_empty(&sdp->sd_log_bhlist_cache))
  {
    bl = osi_list_entry(sdp->sd_log_bhlist_cache.next, gfs_bhlist_t, bl_list);
    osi_list_del(&bl->bl_list);
    gfree(bl, sizeof(gfs_bhlist_t));
  }


  gfree(sdp, sizeof(gfs_sbd_t));


  vsb2sbd(sb) = NULL;

  EXIT(GFN_PUT_SUPER);
}





struct super_operations gfs_sops =
{
  read_inode:           gfs_panic_inode,
  read_inode2:          gfs_read_inode,
  put_super:            gfs_put_super,
  statfs:               gfs_statfs,
  clear_inode:          gfs_clear_inode,
  remount_fs:           gfs_remount_fs,
  write_super:          gfs_write_super
};

