/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portions Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth Preslan <kpreslan@sistina.com>
** 
**    Module name:  dio.c
**
**    Description:
**      This module implements a bunch of routines that do I/O through the
**        buffer cache.
**
*******************************************************************************
******************************************************************************/

#define _LARGEFILE64_SOURCE
#include "gfs.h"

#include "dio.h"
#include "gfs_linux.h"
#include "glock.h"
#include "inode.h"
#include "log.h"
#include "trans.h"


extern int fsfd;  
extern uint64 last_fs_block;



STATIC void __inline__ mark_buffer_uptodate(osi_buf_t *bh, int uptodate)
{
  bh->b_state = uptodate;
}

STATIC unsigned int __inline__ buffer_uptodate(osi_buf_t *bh)
{
  return (bh->b_state);
}


/**
 * gfs_remove_bh - Remove a buffer from the list maintainted for invalidation
 * @bh: The buffer head to be removed
 *
 * This function is the one the buffer cache sees.  It checks to make sure
 * the buffer cache isn't throwing away an active buffer and then calls
 * rm_bh_internal().
 *
 */

void gfs_remove_bh(osi_buf_t *bh)
{
}


/**
 * gfs_inval_bh - Invalidate all buffers associated with a glock
 * @sdp: the filesystem
 * @gl: the glock
 *
 */

void gfs_inval_bh(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
}


/**
 * gfs_sync_bh - Sync all buffers associated with a glock
 * @sdp: the filesystem
 * @gl: The glock
 * @wait: Non-zero means wait for completion of I/O
 *
 */

void gfs_sync_bh(gfs_sbd_t *sdp, gfs_glock_t *gl, int wait)
{
}


/**
 * gfs_free_bh - Free all buffers associated with a glock
 * @sdp: the filesystem
 * @gl: The glock
 *
 * Remove all gfs_linux_bhdata_t structures on the list.  Be careful to avoid
 * races with other processes removeing them naturally through
 * gfs_remove_bh().  We do this by doing a getblk() on the buffer to
 * increment the count on the buffer.  (bh->b_count++ isn't good enough
 * to stop the race)
 *
 * Only called on unmount.   
 *
 */

void gfs_free_bh(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
}



/**
 * gfs_dgetblk_arch - Get a block
 * @sdp: The GFS superblock
 * @blkno: The block number
 * @gl: The glock associated with this block
 *
 * Returns: The buffer on success, or NULL
 */

osi_buf_t *gfs_dgetblk_arch(gfs_sbd_t *sdp, uint64 blkno, gfs_glock_t *gl)
{
  osi_buf_t *bh = NULL;

  ENTER(GFN_DGETBLK_ARCH);


  if (blkno > last_fs_block)
    goto fail;


  bh = (osi_buf_t *)malloc(sizeof(osi_buf_t));
  if (!bh)
    goto fail;

  
  gfs_memset(bh, 0, sizeof(osi_buf_t));

  bh->b_blocknr = blkno;
  bh->b_size = sdp->sd_sb.sb_bsize;
  bh->b_state = 0;
  
  bh->b_data = (char *)malloc(bh->b_size);

  if (!bh->b_data)
  {
    free(bh);
    bh = NULL;
    goto fail;
  }
  
  gfs_memset(bh->b_data, 0, bh->b_size);

 fail:
  EXIT(GFN_DGETBLK_ARCH);
  return bh;
}


/**
 * gfs_dread_arch - Read a block from disk
 * @sdp: The GFS superblock
 * @blkno: The block number
 * @gl: The glock covering the block
 * @flags: flags to gfs_dreread()
 *
 * Returns: The buffer on success, NULL on failure
 */

osi_buf_t *gfs_dread_arch(gfs_sbd_t *sdp, uint64 blkno, gfs_glock_t *gl, int flags)
{
  osi_buf_t *bh;

  ENTER(GFN_DREAD_ARCH);


  bh = gfs_dgetblk_arch(sdp, blkno, gl);
  if (!bh)
  {
    EXIT(GFN_DREAD_ARCH);
    return bh;
  }

  if (gfs_dreread(sdp, bh, flags) != 0)
  {
    gfs_drelse(bh);
    bh = NULL;
  }


  EXIT(GFN_DREAD_ARCH);

  return bh;
}


/**
 * gfs_prep_new_buffer - Mark a new buffer we just gfs_dgetblk()ed uptodate
 * @bh: the buffer
 *
 */

void gfs_prep_new_buffer(osi_buf_t *bh)
{
}


/**
 * gfs_dreread - Reread a block from disk
 * @sdp: the filesystem
 * @bh: The block to read
 * @flags: Flags that control the read
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_dreread(gfs_sbd_t *sdp, osi_buf_t *bh, int flags)
{
  int bytes;
  uint64 off;
  int error = 0;

  ENTER(GFN_DREREAD);


  if (flags & DR_FORCE)
  {
    mark_buffer_uptodate(bh, FALSE);
  }

  /* all reads through this function end up being synchronous on START, so 
     WAIT will just return immediately */

  if ((flags & DR_START) && !buffer_uptodate(bh))
  {
    off = lseek(fsfd, (uint64)bh->b_blocknr * bh->b_size, SEEK_SET);
    
    if (off != (uint64)bh->b_blocknr * bh->b_size)
    {
      db_printf("Bad seek in gfs_dreread\n");
      error = -EIO;
      goto out;
    }
    
    bytes = read(fsfd, bh->b_data, bh->b_size);
    
    if (bytes != bh->b_size)
    {
      db_printf("Bad read in gfs_dreread\n");
      error = -EIO;
      goto out;
    }
    
    mark_buffer_uptodate(bh, TRUE);
  }

  if (flags & DR_WAIT)
  {
    if (!buffer_uptodate(bh))
    {
      error = -EIO;
    }
  }


 out:
  EXIT(GFN_DREREAD);

  return error;
}


/**
 * gfs_dwrite - Write a buffer
 * @sdp: the filesystem
 * @bh: The buffer to write
 * @flags: The type of write operation to do
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_dwrite(gfs_sbd_t *sdp, osi_buf_t *bh, int flags)
{
  return -EIO;
}


/**
 * gfs_drelse_arch - Release a block
 * @bh: The block to release
 *
 */

void gfs_drelse_arch(osi_buf_t *bh)
{
  free(bh->b_data);
  free(bh);
}


/**
 * gfs_dforget - Forget a data buffer
 * @bh: The buffer to release
 *
 */

void gfs_dforget(osi_buf_t *bh)
{
  gfs_drelse_arch(bh);
}



/**
 * gfs_dpin - Pin a metadata buffer in memory
 * @bh: The buffer to be pinned
 *
 */

void gfs_dpin(osi_buf_t *bh)
{
}


/**
 * gfs_dunpin - Unpin a buffer
 * @bh: The buffer to unpin
 * @tr: The transaction in the AIL that contains this buffer
 *
 */

void gfs_dunpin(osi_buf_t *bh, gfs_trans_t *tr)
{
}


/**
 * gfs_is_pinned - check to see if a buffer is pinned
 * @bh: The buffer
 *
 * Returns:  TRUE if the buffer is pinned, FALSE if it isn't.
 */

int gfs_is_pinned(osi_buf_t *bh)
{
  return FALSE;
}


/**
 * gfs_init_temp_bh - Initialise a temporary buffer head
 * @sdp: The GFS superblock
 * @blkno: The block number
 * @buf: The data area
 * @write: TRUE if this buffer will be writen, FALSE if it will be read
 *
 * Returns:  the temp buffer head
 */

osi_buf_t *gfs_init_temp_bh(gfs_sbd_t *sdp, uint64 blkno, char *buf, int write)
{
  return NULL;
}


/**
 * gfs_dispose_temp_bh - Get rid of a temporary buffer_head
 * @bh: The buffer head we no longer want
 *
 */

void gfs_dispose_temp_bh(osi_buf_t *bh)
{
}


/**
 * gfs_get_meta_buffer - Get a metadata buffer
 * @ip: The GFS inode
 * @depth: The depth in the metadata tree
 * @num: The block number (device relative) of the buffer
 * @new: Non-zero if we may create a new buffer
 *
 * Returns: The metadata buffer or NULL
 */

osi_buf_t *gfs_get_meta_buffer(gfs_inode_t *ip, int height, uint64 num, int new)
{
  osi_buf_t *bh;

  bh = gfs_dgetblk_arch(ip->i_sbd, num, NULL);
  gfs_dreread(ip->i_sbd, bh, DR_START);
  return bh;
}


/**
 * gfs_get_data_buffer - Get a data buffer
 * @ip: The GFS inode
 * @num: The block number (device relative) of the data block
 * @new: Non-zero if this is a new allocation
 *
 * Returns: The data buffer or NULL
 */

osi_buf_t *gfs_get_data_buffer(gfs_inode_t *ip, uint64 block, int new)
{
  osi_buf_t *bh;

  bh = gfs_dgetblk_arch(ip->i_sbd, block, NULL);
  gfs_dreread(ip->i_sbd, bh, DR_START);
  return bh;
}


