/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portions Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: file.c
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"

#include "blklist.h"
#include "bmap.h"
#include "dcache.h"
#include "dio.h"
#include "dir.h"
#include "file.h"
#include "flock.h"
#include "gfs_linux.h"
#include "glock.h"
#include "glops.h"
#include "inode.h"
#include "inode_linux.h"
#include "ioctl.h"
#include "log.h"
#include "trans.h"



struct gfs_filldir_struct
{
};
typedef struct gfs_filldir_struct gfs_filldir_struct_t;





/**
 * gfs_copy2mem - Trivial copy function for gfs_readi()
 * @bh: The buffer to copy from, or NULL meaning zero the buffer
 * @buf: The buffer to copy/zero
 * @offset: The offset in the buffer to copy from
 * @size: The amount of data to copy/zero
 *
 * This function is used as an argument to gfs_readi() when we want
 * to read data as part of another GFS internal function.
 *
 * Returns: The amount of data copied/cleared
 */

int gfs_copy2mem(osi_buf_t *bh, char *buf, int offset, int size)
{
  ENTER(GFN_COPY2MEM);
  
  if (bh)
    gfs_memcpy(buf, BH_DATA(bh) + offset, size);
  else
    gfs_memset(buf, 0, size);

  EXIT(GFN_COPY2MEM);

  return size;
}


/**
 * gfs_copy2user - Used only for the ioctl() read function
 * @bh: The buffer
 * @buf: The destination of the data
 * @offset: The offset into the buffer
 * @size: The amount of data to copy
 *
 * Returns: Amount of data copied
 */

int gfs_copy2user(osi_buf_t *bh, char *buf, int offset, int size)
{
  int remainder;

  ENTER(GFN_COPY2USER);

  if (bh)
  {
    gfs_memcpy(buf, BH_DATA(bh) + offset, size);
    remainder = size;
  }
  else
  {
    gfs_memset(buf, 0, size);
    remainder = size;
  }

  EXIT(GFN_COPY2USER);

  return (remainder) ? -EFAULT : size;
}


/**
 * gfs_readi - Read a file
 * @ip: The GFS Inode
 * @buf: The buffer to place result into
 * @offset: File offset to begin reading from
 * @size: Amount of data to transfer
 * @copy_fn: Function to actually perform the copy
 *
 * The @copy_fn only copies a maximum of a single block at once so
 * we are safe calling it with int arguments. It is done so that
 * we don't needlessly put 64bit arguments on the stack and it
 * also makes the code in the @copy_fn nicer too.
 *
 * Returns: The amount of data actually copied or the error
 */

int64 gfs_readi(gfs_inode_t *ip, char *buf, uint64 offset, uint64 size,
		read_copy_fn_t copy_fn)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  osi_buf_t *bh;
  int64 copied = 0;
  uint64 block, nr;
  unsigned int amount;
  unsigned int bsize;
  int not_new = 0;
  int journaled = gfs_is_jdata(ip);
  int64 error = 0;

  ENTER(GFN_READI);


  if (offset >= ip->i_di.di_size)
    goto out;

  if ((offset + size) > ip->i_di.di_size)
    size = ip->i_di.di_size - offset;

  if (!size)
    goto out;


  bsize = (journaled) ? sdp->sd_jbsize : sdp->sd_sb.sb_bsize;

  block = offset / bsize;
  offset %= bsize;

  if (gfs_is_stuffed(ip))
    offset += sizeof(gfs_dinode_t);
  else if (journaled)
    offset += sizeof(gfs_jdata_t);


  while (copied < size)
  {
    amount = size - copied;
    if (amount > sdp->sd_sb.sb_bsize - offset)
      amount = sdp->sd_sb.sb_bsize - offset;

    error = gfs_inode_get_block(ip, block, &not_new, &nr);
    if (error)
      goto out;

    if (nr)
    {
      bh = gfs_get_data_buffer(ip, nr, not_new);
      if (!bh)
      {
	error = -EIO;
	goto out;
      }
    }
    else
      bh = NULL;

    error = copy_fn(bh, buf, (int)offset, amount);
    if (bh)
      gfs_drelse(bh);
    if (error < 0)
      goto out;

    copied += amount;
    buf += amount;
    block++;

    offset = (journaled) ? sizeof(gfs_jdata_t) : 0;
  }


 out:
  EXIT(GFN_READI);
  return (error < 0) ? error : copied;
}



/**
 * gfs_copy_from_mem - Trivial copy function for gfs_writei()
 * @ip: The file to write to
 * @bh: The buffer to copy to or clear
 * @buf: The buffer to copy from
 * @offset: The offset in the buffer to write to
 * @size: The amount of data to write
 * @new: Flag indicating that remaining space in the buffer should be zeroed
 *
 * Whilst this function does deal with stuffed dinodes, they are never new
 * buffers here, which is why a call to zero_buffer_ends is safe.
 *
 * Returns: The amount of data copied
 */

int gfs_copy_from_mem(gfs_inode_t *ip, osi_buf_t *bh, const char *buf, int offset, int size, int new)
{
  return 0;
}


/**
 * gfs_copy_from_user - Copy bytes from user space for gfs_writei()
 * @ip: The file to write to
 * @bh: The buffer to copy to or clear
 * @buf: The buffer to copy from
 * @offset: The offset in the buffer to write to
 * @size: The amount of data to write
 * @new: Flag indicating that remaining space in the buffer should be zeroed
 *
 * If the buffer is a new one, and the copy from user space fails,
 * then we must zero out the space that would have been occupied by
 * data.
 *
 * Returns: The amount of data copied
 */

int gfs_copy_from_user(gfs_inode_t *ip, osi_buf_t *bh, const char *buf, int offset, int size, int new)
{
  return 0;
}


/**
 * gfs_writei - Write bytes to a file
 * @ip: The GFS inode
 * @buf: The buffer containing information to be written
 * @offset: The file offset to start writing at
 * @size: The amount of data to write
 * @copy_fn: Function to do the actual copying
 *
 * Returns: The number of bytes correctly written or error code
 */

int64 gfs_writei(gfs_inode_t *ip, const char *buf, uint64 offset, uint64 size,
		 write_copy_fn_t copy_fn)
{
  return 0;
}






