/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portions Copyright 2000-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: bits.c
**
**    Bitmaps:
**
**         These routines are used by the block-list routines (blklist.c)
**    to keep track of block allocation.  Each block is represented by two
**    bits.  One bit indicates whether or not the block is used.  (1=used,
**    0=free)  The other bit indicates whether or not the block contains a
**    dinode or not.  (1=dinode, 0=data block)
**         So, each byte represents GFS_NBBY (i.e. 4) blocks.  
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "bits.h"



static const char valid_change[16] =
{
          /* current */

  /* n */  0, 1, 0, 0,
  /* e */  1, 0, 0, 0,
  /* w */  1, 0, 0, 1,
           0, 0, 1, 0
};





/**
 * gfs_setbit - Set a bit in the bitmaps
 * @buffer: the buffer that holds the bitmaps
 * @buflen: the length (in bytes) of the buffer
 * @block: the block to set
 * @new_state: the new state of the block
 *
 */

void gfs_setbit(unsigned char *buffer, unsigned int buflen,
		uint32 block, unsigned char new_state)
{
  unsigned char *byte, *end, cur_state;
  unsigned int bit;

  ENTER(GFN_SET_BITS);


  byte = buffer + (block / GFS_NBBY);
  bit = (block % GFS_NBBY) * GFS_BIT_SIZE;
  end = buffer + buflen;

  GFS_ASSERT(byte < end, );

  cur_state = (*byte >> bit) & GFS_BIT_MASK;
  GFS_ASSERT(valid_change[new_state * 4 + cur_state],
	     con_printf("cur_state = %u, new_state = %u\n", cur_state, new_state););

  *byte ^= cur_state << bit;
  *byte |= new_state << bit;


  EXIT(GFN_SET_BITS);
}


/**
 * gfs_bitfit - Find a free block in the bitmaps
 * @buffer: the buffer that holds the bitmaps
 * @buflen: the length (in bytes) of the buffer
 * @goal: the block to try to allocate
 * @old_state: the state of the block we're looking for
 *
 * Return: the block number that was allocated
 */

uint32 gfs_bitfit(unsigned char *buffer, unsigned int buflen,
		  uint32 goal, unsigned char old_state)
{
  unsigned char *byte, *end, alloc;
  uint32 blk = goal;
  unsigned int bit;

  ENTER(GFN_BITFIT);


  byte = buffer + (goal / GFS_NBBY);
  bit = (goal % GFS_NBBY) * GFS_BIT_SIZE;
  end = buffer + buflen;
  alloc = (old_state & 1) ? 0 : 0x55;


  while (byte < end)
  {
    if ((*byte & 0x55) == alloc)
    {
      blk += (8 - bit) >> 1;

      bit = 0;
      byte++;

      continue;
    }

    if (((*byte >> bit) & GFS_BIT_MASK) == old_state)
    {
      EXIT(GFN_BITFIT);
      return blk;
    }

    bit += GFS_BIT_SIZE;
    if (bit >= 8)
    {
      bit = 0;
      byte++;
    }

    blk++;
  }


  EXIT(GFN_BITFIT);
  
  return BFITNOENT;
}


/**
 * gfs_bitcount - count the number of bits in a certain state
 * @buffer: the buffer that holds the bitmaps
 * @buflen: the length (in bytes) of the buffer
 * @state: the state of the block we're looking for
 *
 * Returns: The number of bits
 */

uint32 gfs_bitcount(unsigned char *buffer, unsigned int buflen, unsigned char state)
{
  unsigned char *byte, *end;
  unsigned int bit;
  uint32 count = 0;

  ENTER(GFN_BITCOUNT);


  byte = buffer;
  bit = 0;
  end = buffer + buflen;


  while (byte < end)
  {
    if (((*byte >> bit) & GFS_BIT_MASK) == state)
      count++;

    bit += GFS_BIT_SIZE;
    if (bit >= 8)
    {
      bit = 0;
      byte++;
    }
  }


  EXIT(GFN_BITCOUNT);

  return count;
}


