/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portions Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: blklist.c
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "blklist.h"
#include "bits.h"
#include "dio.h"
#include "glock.h"
#include "glops.h"
#include "file.h"
#include "fs.h"
#include "log.h"
#include "trans.h"



#define AVL_MIN_RG_COUNT 10
#define rd_avl_empty ((struct gfs_rgrpd *)NULL)

#include "rg_avl.c"





/**
 * gfs_rgrpd_hold - increment the useage count on a resource group
 * @rgd: the resource group
 *
 * Returns: the rgrp that is passed in
 */

STATIC __inline__ gfs_rgrpd_t *gfs_rgrpd_hold(gfs_rgrpd_t *rgd)
{
  if (rgd)
    osi_atomic_inc(&rgd->rd_count);
  return rgd;
}


/**
 * __gfs_blk2rgrpd_list - Find the RGRP for a given block number (using list)
 * @sdp: the filesystem
 * @blk: the block number
 *
 * Find the Resource group for a given block number.
 *
 * Returns: The resource group, or NULL if not found
 */

gfs_rgrpd_t *__gfs_blk2rgrpd_list(gfs_sbd_t *sdp, uint64 blk)
{
  struct osi_list_head *tmp, *head;
  gfs_rgrpd_t *rgd;
  gfs_rindex_t *ri;

  ENTER(GFN_BLK2RGRPD_LIST);


  head = tmp = &sdp->sd_rglist;
  for (;;)
  {
    rgd = NULL;
    tmp = tmp->next;
    if (tmp == head)
      break;

    rgd = osi_list_entry(tmp, gfs_rgrpd_t, rd_list);
    ri = &rgd->rd_ri;

    if (blk >= (ri->ri_data1 + ri->ri_data))
      continue;
    if (blk >= ri->ri_data1)
      break;

    rgd = NULL;
    break;
  }

  EXIT(GFN_BLK2RGRPD_LIST);

  return rgd;
}


/**
 * __gfs_blk2rgrpd_tree - Find RGRP for a given data block (using tree)
 * @sdp: The GFS superblock
 * @n: The data block number
 *
 * Returns: Ths resource group, or NULL if not found
 */

STATIC gfs_rgrpd_t *__gfs_blk2rgrpd_tree(gfs_sbd_t *sdp, uint64 n)
{
  gfs_rgrpd_t *tree = sdp->sd_rgavl;
  gfs_rgrpd_t *ret;

  ENTER(GFN_BLK2RGRPD_TREE);

  for(;;)
  {
    ret = NULL;
    if (tree == rd_avl_empty)
      break;
    if (tree->rd_avl_key > n)
    {
      tree = tree->rd_avl_left;
    }
    else
    {
      ret = tree;
      if ((tree->rd_ri.ri_data1 <= n) &&
          (tree->rd_ri.ri_data1 + tree->rd_ri.ri_data) > n)
        break;
      tree = tree->rd_avl_right;
    }
  }

  EXIT(GFN_BLK2RGRPD_TREE);
  return ret;
}


/**
 * gfs_blk2rgrpd - Find resource group for a given data block number
 * @sdp: The GFS superblock
 * @n: The data block number
 *
 * Returns: Ths resource group, or NULL if not found
 */

gfs_rgrpd_t *gfs_blk2rgrpd(gfs_sbd_t *sdp, uint64 n)
{
  gfs_rgrpd_t *rgd;
  int set_cache = 0;

  ENTER(GFN_BLK2RGRPD);

  osi_read_lock(&sdp->sd_rglock);
  rgd = sdp->sd_rgcache;

  if (rgd && (n >= rgd->rd_ri.ri_data1) && (n < (rgd->rd_ri.ri_data1 + rgd->rd_ri.ri_data)))
    goto out;

  if (sdp->sd_rgavl)
    rgd = __gfs_blk2rgrpd_tree(sdp, n);
  else
    rgd = __gfs_blk2rgrpd_list(sdp, n);

  if (rgd && (sdp->sd_rgcache != rgd))
    set_cache = 1;

 out:
  gfs_rgrpd_hold(rgd);
  osi_read_unlock(&sdp->sd_rglock);
  if (set_cache)
  {
    osi_write_lock(&sdp->sd_rglock);
    sdp->sd_rgcache = rgd;
    osi_write_unlock(&sdp->sd_rglock);
  }

  EXIT(GFN_BLK2RGRPD);
  return rgd;
}


/**
 * gfs_rgrpd_get_by_num - get the @num-th RG
 * @sdp: The GFS superblock
 * @num: The RG number to get
 *
 * Returns: The rgrp, or NULL if no rgrp with this number
 */

gfs_rgrpd_t *gfs_rgrpd_get_by_num(gfs_sbd_t *sdp, unsigned int num)
{
  struct osi_list_head *tmp, *head;
  gfs_rgrpd_t *rgd = NULL;
  unsigned int x = 0;

  ENTER(GFN_RGRPD_GET_BY_NUM);

  osi_read_lock(&sdp->sd_rglock);

  head = &sdp->sd_rglist;
  for (tmp = head->next; tmp != head; tmp = tmp->next)
  {
    rgd = osi_list_entry(tmp, gfs_rgrpd_t, rd_list);

    if (x == num)
    {
      gfs_rgrpd_hold(rgd);
      break;
    }

    rgd = NULL;
    x++;
  }

  osi_read_unlock(&sdp->sd_rglock);

  EXIT(GFN_RGRPD_GET_BY_NUM);

  return rgd;
}


/**
 * __gfs_find_rgrpd_list - Find an rgrp by disk block (using the list)
 * @sdp: The GFS superblock
 * @n: The block number
 *
 */

STATIC gfs_rgrpd_t *__gfs_find_rgrpd_list(gfs_sbd_t *sdp, uint64 n)
{
  struct osi_list_head *tmp, *head;
  gfs_rgrpd_t *rgrp;

  ENTER(GFN_FIND_RGRPD_LIST);

  tmp = head = &sdp->sd_rglist;
  for (;;) 
  {
    rgrp = NULL;
    tmp = tmp->next;
    if (tmp == head)
      break;
    rgrp = osi_list_entry(tmp, gfs_rgrpd_t, rd_list);

    if (rgrp->rd_ri.ri_addr < n)
      continue;
    if (rgrp->rd_ri.ri_addr > n)
      break;

    EXIT(GFN_FIND_RGRPD_LIST);
    return rgrp;
  }

  EXIT(GFN_FIND_RGRPD_LIST);

  return NULL;
}


/**
 * __gfs_find_rgrpd_tree - Find an rgrp by disk block (using the AVL tree)
 * @sdp: The GFS superblock
 * @n: The diskblock number
 *
 */

STATIC gfs_rgrpd_t *__gfs_find_rgrpd_tree(gfs_sbd_t *sdp, uint64 n)
{
  gfs_rgrpd_t *tree = sdp->sd_rgavl;
  gfs_rgrpd_t *ret;

  ENTER(GFN_FIND_RGRPD_TREE);

  for (;;)
  {
    ret = NULL;
    if (tree == rd_avl_empty)
      break;
    if (tree->rd_avl_key >= n)
    {
      ret = tree;
      if (tree->rd_avl_key == n)
        break;
      tree = tree->rd_avl_left;
    }
    else
    {
      tree = tree->rd_avl_right;
    }
  }

  EXIT(GFN_FIND_RGRPD_TREE);
  return ret;
}


/**
 * __gfs_find_rgrpd - Find an rgrp by disk block
 * @sdp: The GFS superblock
 * @n: The disk block
 *
 * Lookup a resource group by disk block number. First we look in the
 * cache. We often look up the same resource group several times in
 * a row, so this saves time. If we have an AVL tree, then we used that
 * method to perform a lookup, otherwise we fall back to scanning the
 * ordered list.
 */

STATIC __inline__ gfs_rgrpd_t *__gfs_find_rgrpd(gfs_sbd_t *sdp, uint64 n)
{
  gfs_rgrpd_t *rgd = sdp->sd_rgcache;

  ENTER(GFN_FIND_RGRPD);

  if (rgd && (rgd->rd_avl_key == n))
    goto out;

  if (sdp->sd_rgavl)
    rgd = __gfs_find_rgrpd_tree(sdp, n);
  else
    rgd = __gfs_find_rgrpd_list(sdp, n);

 out:
  EXIT(GFN_FIND_RGRPD);
  return gfs_rgrpd_hold(rgd);
}


/**
 * build_rgrp_avl - Build an AVL tree of resource groups
 * @sdp: The GFS superblock
 *
 */

STATIC void build_rgrp_avl(gfs_sbd_t *sdp)
{
  struct gfs_rgrpd *rd;
  struct osi_list_head *tmp, *head;

  ENTER(GFN_BUILD_RGRP_AVL);

  sdp->sd_rgavl = NULL;
  tmp = head = &sdp->sd_rglist;
  for(;;)
  {
    tmp = tmp->next;
    if (tmp == head)
      break;
    rd = osi_list_entry(tmp, struct gfs_rgrpd, rd_list);
    avl_insert(rd, &sdp->sd_rgavl);
  }

  EXIT(GFN_BUILD_RGRP_AVL);
}


/**
 * __gfs_rgrpd_insert - Add a new rgrp into the list, ensure its unique
 * @rgd: The new rgrp
 *
 * Returns: NULL on success, the duplicate rgrpd if not unique
 */

STATIC void __gfs_rgrpd_insert(gfs_rgrpd_t *nrgd)
{
  gfs_sbd_t *sdp = nrgd->rd_sbd;
  gfs_rgrpd_t *rgd;
  osi_list_head_t *tmp, *head;

  ENTER(GFN_RGRPD_INSERT);

  tmp = head = &sdp->sd_rglist;
  for (;;)
  {
    tmp = tmp->next;
    if (tmp == head)
      break;
    rgd = osi_list_entry(tmp, gfs_rgrpd_t, rd_list);

    if (rgd->rd_ri.ri_addr > nrgd->rd_ri.ri_addr)
      break;
  }

  osi_list_add_prev(&nrgd->rd_list, tmp);

  if (sdp->sd_rgavl) /* Add to AVL tree as well */
    avl_insert(nrgd, &sdp->sd_rgavl);

  sdp->sd_rgcount++;
  if (sdp->sd_rgcount >= AVL_MIN_RG_COUNT && !sdp->sd_rgavl)
    build_rgrp_avl(sdp);

  EXIT(GFN_RGRPD_INSERT);
}


/**
 * gfs_rgrpd_get - Get an rgrp by disk block number
 * @sdp: The GFS superblock
 * @n: The block number
 * @create: May create a new rgrp structure
 *
 */

gfs_rgrpd_t *gfs_rgrpd_get(gfs_sbd_t *sdp, uint64 n, int create)
{
  gfs_rgrpd_t *rgrp, *nrgrp;

  ENTER(GFN_RGRPD_GET);

  osi_read_lock(&sdp->sd_rglock);
  rgrp = __gfs_find_rgrpd(sdp, n);
  osi_read_unlock(&sdp->sd_rglock);

  if (rgrp)
  {
    osi_write_lock(&sdp->sd_rglock);
    sdp->sd_rgcache = rgrp;
    osi_write_unlock(&sdp->sd_rglock);
    if (!create)
      goto out;
  }

  nrgrp = gfs_alloc_rgrpd();
  if (!nrgrp)
    goto out;

  gfs_memset(nrgrp, 0, sizeof(gfs_rgrpd_t));
  nrgrp->rd_ri.ri_addr = n;
  nrgrp->rd_sbd = sdp;

  gfs_rgrpd_hold(nrgrp);
  gfs_rgrpd_hold(nrgrp);

  /*
   * Just because we created a new resource group doesn't mean that we are
   * going to use it right away, so don't bother caching it.
   */
  osi_write_lock(&sdp->sd_rglock);
  if ((rgrp = __gfs_find_rgrpd(sdp, n)) == NULL)
    __gfs_rgrpd_insert(nrgrp);
  osi_write_unlock(&sdp->sd_rglock);

  if (rgrp)
    gfs_free_rgrpd(nrgrp);
  else
    rgrp = nrgrp;

 out:
  EXIT(GFN_RGRPD_GET);
  return rgrp;
}


/**
 * gfs_rgrpd_get_with_hint - Get rgrp with disk address >= hint
 * @sdp: The GFS superblock
 * @hint: The minimum disk address of the rgrp
 *
 * The implementation of this routine is fucked up.  It should
 * use the AVL trees.  The routine is called with arbitrary hints
 * from gfs_inplace_reserv() all the time.
 *
 * Returns: The rgrp, or NULL if no rgrp is below the hinted disk address
 */

gfs_rgrpd_t *gfs_rgrpd_get_with_hint(gfs_sbd_t *sdp, uint64 hint)
{
  struct osi_list_head *tmp, *head;
  gfs_rgrpd_t *rgd = NULL;

  ENTER(GFN_RGRPD_GET_WITH_HINT);

  osi_read_lock(&sdp->sd_rglock);
  tmp = head = &sdp->sd_rglist;
  for (;;)
  {
    tmp = tmp->next;
    rgd = NULL;
    if (tmp == head)
      break;
    rgd = osi_list_entry(tmp, gfs_rgrpd_t, rd_list);

    if (rgd->rd_ri.ri_addr < hint)
      continue;
    gfs_rgrpd_hold(rgd);

    break;
  }
  osi_read_unlock(&sdp->sd_rglock);

  EXIT(GFN_RGRPD_GET_WITH_HINT);
  return rgd;
}


/**
 * __gfs_rgrpd_put - Free an rgrpd structure
 * @sdp: The GFS superblock
 * @rgd: The rgrpd to put
 * 
 * Called only from gfs_rgrpd_put()
 */

void __gfs_rgrpd_put(gfs_rgrpd_t *rgd)
{
  ENTER(GFN_RGRPD_PUT);

  GFS_ASSERT(osi_list_empty(&rgd->rd_list), );
  GFS_ASSERT(rgd->rd_avl_left == NULL, );
  GFS_ASSERT(rgd->rd_avl_right == NULL, );

  if (rgd->rd_gl)
  {
    gfs_force_gl_drop(rgd->rd_sbd, rgd->rd_gl);
    if (rgd->rd_gl->gl_lvb)
      gfs_unhold_lvb(rgd->rd_sbd, rgd->rd_gl);
    rgd->rd_gl->gl_object = NULL;
    gfs_put_glstruct(rgd->rd_gl);
  }
  if (rgd->rd_bh)
    gfree(rgd->rd_bh, rgd->rd_ri.ri_length * sizeof(osi_buf_t *));
  if (rgd->rd_bits)
    gfree(rgd->rd_bits, rgd->rd_ri.ri_length * sizeof(gfs_bitmap_t));

  gfs_free_rgrpd(rgd);

  EXIT(GFN_RGRPD_PUT);
}


/**
 * gfs_clear_rgrpd - Clear up rgrps
 * @sdp: The GFS superblock
 *
 */

void gfs_clear_rgrpd(gfs_sbd_t *sdp)
{
  gfs_rgrpd_t *rgd;

  ENTER(GFN_CLEAR_RGRPD);

  for (;;)
  {
    rgd = NULL;
    osi_write_lock(&sdp->sd_rglock);
    if (!osi_list_empty(&sdp->sd_rglist))
    {
      rgd = osi_list_entry(sdp->sd_rglist.next, gfs_rgrpd_t, rd_list);
      osi_list_del(&rgd->rd_list);
      osi_list_head_init(&rgd->rd_list);
      avl_remove(rgd, &sdp->sd_rgavl);
      rgd->rd_avl_left = rgd->rd_avl_right = NULL;
      rgd->rd_avl_height = 0;
      sdp->sd_rgcount--;
      if (sdp->sd_rgcache == rgd)
        sdp->sd_rgcache = NULL;
    }
    osi_write_unlock(&sdp->sd_rglock);
    if (rgd == NULL)
      break;

    GFS_ASSERT(osi_atomic_read(&rgd->rd_count) == 1, );

    gfs_rgrpd_put(rgd);
  }

  EXIT(GFN_CLEAR_RGRPD);
}


/**
 * gfs_compute_bitstructs - Compute the bitmap sizes
 * @rgd: The resource group descriptor
 *
 * Returns: 0, or negative on error
 */

STATIC int compute_bitstructs(gfs_rgrpd_t *rgd)
{
  gfs_sbd_t *sdp = rgd->rd_sbd;
  gfs_bitmap_t *bits;
  uint32 length = rgd->rd_ri.ri_length;
  uint32 bytes_left, bytes;
  int x;
  int error = 0;

  ENTER(GFN_COMPUTE_BITSTRUCTS);


  rgd->rd_bh = (osi_buf_t **)gmalloc(length * sizeof(osi_buf_t *));
  if (!rgd->rd_bh)
  {
    error = -ENOMEM;
    goto out;
  }
  gfs_memset(rgd->rd_bh, 0, length * sizeof(osi_buf_t *));

  rgd->rd_bits = (gfs_bitmap_t *)gmalloc(length * sizeof(gfs_bitmap_t));
  if (!rgd->rd_bits)
  {
    gfree(rgd->rd_bh, length * sizeof(osi_buf_t *));
    rgd->rd_bh = NULL;
    error = -ENOMEM;
    goto out;
  }
  gfs_memset(rgd->rd_bits, 0, length * sizeof(gfs_bitmap_t));


  bytes_left = rgd->rd_ri.ri_bitbytes;

  for (x = 0; x < length; x++)
  {
    bits = &rgd->rd_bits[x];

    if (length == 1)
    {
      bytes = bytes_left;
      bits->bi_offset = sizeof(gfs_rgrp_t);
      bits->bi_start = 0;
      bits->bi_len = bytes;
    }
    else if (x == 0)
    {
      bytes = sdp->sd_sb.sb_bsize - sizeof(gfs_rgrp_t);
      bits->bi_offset = sizeof(gfs_rgrp_t);
      bits->bi_start = 0;
      bits->bi_len = bytes;
    }
    else if (x + 1 == length)
    {
      bytes = bytes_left;
      bits->bi_offset = sizeof(gfs_rgrpbits_t);
      bits->bi_start = rgd->rd_ri.ri_bitbytes - bytes_left;
      bits->bi_len = bytes;
    }
    else
    {
      bytes = sdp->sd_sb.sb_bsize - sizeof(gfs_rgrpbits_t);
      bits->bi_offset = sizeof(gfs_rgrpbits_t);
      bits->bi_start = rgd->rd_ri.ri_bitbytes - bytes_left;
      bits->bi_len = bytes;
    }

    bytes_left -= bytes;
  }


  GFS_ASSERT(!bytes_left, );
  GFS_ASSERT((rgd->rd_bits[length - 1].bi_start +
              rgd->rd_bits[length - 1].bi_len) * GFS_NBBY == rgd->rd_ri.ri_data,
             con_printf("start=%u len=%u offset=%u\n", 
                        rgd->rd_bits[length - 1].bi_start, 
                        rgd->rd_bits[length - 1].bi_len, 
                        rgd->rd_bits[length - 1].bi_offset);
	     gfs_rindex_print(&rgd->rd_ri, TRUE););


 out:
  EXIT(GFN_COMPUTE_BITSTRUCTS);

  return error;
}


/**
 * gfs_rgrpd_update - Update the rgrpd structures
 * @gl: The glock covering the rindex inode
 *
 * This is rather inefficient as it stands. I'm not too worried about
 * it for two reasons: firstly, the only time this is called is when
 * someone updates the resource index, which will be a very infrequent
 * operation, and secondly because when we move to per-block locking the
 * problem will go away anyway. For now though, this will do.
 *
 * Returns: 0 on successful update, error code otherwise
 */

STATIC int gfs_rgrp_update(gfs_inode_t *ip)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  gfs_rgrpd_t *rgd;
  gfs_rindex_t buf;
  unsigned int rg;
  int error;

  ENTER(GFN_RGRP_UPDATE);
  
  GFS_ASSERT((ip->i_di.di_size % sizeof(gfs_rindex_t)) == 0, );


  gfs_clear_rgrpd(sdp);


  for (rg = 0; ; rg++)
  {
    error = gfs_internal_read(ip, (char *)&buf, rg * sizeof(gfs_rindex_t), sizeof(gfs_rindex_t));
    if (!error)
      break;
    if (error != sizeof(gfs_rindex_t))
    {
      if (error > 0)
	error = -EIO;
      goto fail;
    }

    rgd = gfs_rgrpd_get(sdp, gfs64_to_cpu(buf.ri_addr), CREATE);
    if (!rgd)
      goto fail;

    gfs_rindex_in(&rgd->rd_ri, (char *)&buf);

    error = compute_bitstructs(rgd);
    if (error)
      goto fail;

    rgd->rd_gl = gfs_get_glstruct(sdp, rgd->rd_ri.ri_addr, &gfs_rgrp_glops, ip->i_gl, CREATE);
    if (!rgd->rd_gl)
    {
      error = -ENOMEM;
      goto fail;
    }
    rgd->rd_gl->gl_object = rgd;

    error = gfs_hold_lvb(sdp, rgd->rd_gl);
    if (error)
      goto fail;

    error = gfs_rgrp_lvb_init(rgd, FALSE);
    if (error)
      goto fail;

    gfs_rgrpd_put(rgd);
  }

  sdp->sd_riinode_vn = ip->i_gl->gl_vn;


  EXIT(GFN_RGRP_UPDATE);
  return 0;



 fail:
  gfs_clear_rgrpd(sdp);
  EXIT(GFN_RGRP_UPDATE);

  return error;
}


/**
 * gfs_rindex_hold - Grab a lock on the rindex
 * @sdp: The GFS superblock
 *
 * We grab a lock in the rindex inode to make sure that it doesn't
 * change whilst we are performing an operation. We keep this lock
 * for quite long periods of time compared to other locks. This
 * doesn't matter, since its shared and it is very, very rarely
 * accessed in the exclusive mode.
 *
 * Returns: 0 on success, error code otherwise
 */

int gfs_rindex_hold(gfs_sbd_t *sdp)
{
  gfs_inode_t *ip = sdp->sd_riinode;
  gfs_glock_t *gl = ip->i_gl;
  int error;

  ENTER(GFN_RINDEX_HOLD);

  error = gfs_glock_i(ip, GL_SHARED);
  if (error)
    goto out;

  if (sdp->sd_riinode_vn != gl->gl_vn)
  {
    error = gfs_rgrp_update(ip);
    if (error)
      gfs_gunlock_i(ip, 0);
  }

 out:
  EXIT(GFN_RINDEX_HOLD);
  return error;
}


/**
 * gfs_rindex_release - Release hold over rindex
 * @sdp: The GFS superblock
 *
 * Drop the lock on the resource index, allows filesystem grow and
 * shrink to take place once more.
 *
 */

void gfs_rindex_release(gfs_sbd_t *sdp)
{
  ENTER(GFN_RINDEX_RELEASE);
  gfs_gunlock_i(sdp->sd_riinode, 0);
  EXIT(GFN_RINDEX_RELEASE);
}


/**
 * gfs_rgrp_read - Read in a RG's bitmaps
 * @rgd: the gfs_rgrpd_t describing the RG to read in
 *
 * Read in RG bitmaps.  Must call gfs_rgrp_relse() it free the bitmaps.
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_rgrp_read(gfs_rgrpd_t *rgd)
{
  gfs_sbd_t *sdp = rgd->rd_sbd;
  gfs_glock_t *gl = rgd->rd_gl;
  int x, y, length = rgd->rd_ri.ri_length;
  int error;

  ENTER(GFN_RGRP_READ);


  for (x = 0; x < length; x++)
  {
    GFS_ASSERT(!rgd->rd_bh[x], );

    rgd->rd_bh[x] = gfs_dgetblk(sdp, rgd->rd_ri.ri_addr + x, gl);
    if (!rgd->rd_bh[x])
    {
      error = -EIO;
      goto fail;
    }
  }


  for (y = 0; y < length; y++)
    gfs_dreread(sdp, rgd->rd_bh[y], DR_START);

  for (y = length; y--; )
  {
    error = gfs_dreread(sdp, rgd->rd_bh[y], DR_WAIT);
    if (error)
      goto fail;
    gfs_metatype_check(rgd->rd_bh[y], (y) ? GFS_METATYPE_RB : GFS_METATYPE_RG);
  }

  gfs_rgrp_in(&rgd->rd_rg, BH_DATA(rgd->rd_bh[0]));


  EXIT(GFN_RGRP_READ);

  return 0;



 fail:
  while (x--)
  {
    gfs_drelse(rgd->rd_bh[x]);
    rgd->rd_bh[x] = NULL;
  }

  EXIT(GFN_RGRP_READ);
  return error;
}


/**
 * gfs_rgrp_relse - Release RG bitmaps read in with gfs_rgrp_read()
 * @rgd: the gfs_rgrpd_t describing the RG to read in
 *
 */

void gfs_rgrp_relse(gfs_rgrpd_t *rgd)
{
  int x, length = rgd->rd_ri.ri_length;
  
  ENTER(GFN_RGRP_RELSE);

  for (x = 0; x < length; x++)
  {
    gfs_drelse(rgd->rd_bh[x]);
    rgd->rd_bh[x] = NULL;
  }

  EXIT(GFN_RGRP_RELSE);
}


/**
 * gfs_rgrp_lvb_fill - copy RG usage data out of the gfs_rgrp_t into the gfs_rgrp_save_t
 * @rgd: the resource group data structure
 *
 */

void gfs_rgrp_lvb_fill(gfs_rgrpd_t *rgd)
{
  gfs_rgrp_t *rg = &rgd->rd_rg;
  gfs_rgrp_save_t *rs = &rgd->rd_rs;

  ENTER(GFN_RGRP_LVB_FILL);

  rs->rs_magic = GFS_MAGIC;
  rs->rs_free = rg->rg_free;
  rs->rs_useddi = rg->rg_useddi;
  rs->rs_freedi = rg->rg_freedi;
  rs->rs_usedmeta = rg->rg_usedmeta;
  rs->rs_freemeta = rg->rg_freemeta;  

  rgd->rd_rs_dirty = TRUE;

  EXIT(GFN_RGRP_LVB_FILL);
}


/**
 * gfs_rgrp_lvb_init - Init the data of a RG LVB
 * @rgd: the resource group data structure
 * @force: Fill in the data even if it seems that it's correct
 *
 * Returns:  0 on success, -EXXX on failure
 */

int gfs_rgrp_lvb_init(gfs_rgrpd_t *rgd, int force)
{
  int error = 0;

  ENTER(GFN_RGRP_LVB_INIT);


  if (!force)
  {
    error = gfs_glock_rg(rgd, GL_SHARED | GL_SKIP);
    if (error)
      goto out;

    if (rgd->rd_rs.rs_magic == GFS_MAGIC)
    {
      gfs_gunlock_rg(rgd, GL_SKIP);
      error = 0;
      goto out;
    }

    gfs_gunlock_rg(rgd, GL_SKIP);
  }


  error = gfs_glock_rg(rgd, 0);
  if (error)
    goto out;

  IFDB(GFN_RGRP_LVB_INIT,
       gfs_rindex_print(&rgd->rd_ri, FALSE);
       gfs_rgrp_print(&rgd->rd_rg, FALSE);
       gfs_rgrp_verify(rgd););

  gfs_rgrp_lvb_fill(rgd);
  rgd->rd_rs.rs_owner = 0;
  gfs_rgrp_save_out(&rgd->rd_rs, rgd->rd_gl->gl_lvb);
  gfs_sync_lvb(rgd->rd_sbd, rgd->rd_gl);
  rgd->rd_rs_dirty = FALSE;

  gfs_gunlock_rg(rgd, 0);


 out:
  EXIT(GFN_RGRP_LVB_INIT);

  return error;
}


/**
 * rgrp_lvb_own - Mark a RG LVB as being owned by a machine
 * @rgd: the resource group data structure
 *
 * If a machine ever sees a RG LVB that's owned by another
 * machine, it knows that the data in that LVB is stale because
 * the owning machine crashed.
 */

STATIC void rgrp_lvb_own(gfs_rgrpd_t *rgd)
{
  gfs_sbd_t *sdp = rgd->rd_sbd;

  ENTER(GFN_RGRP_LVB_OWN);

  if (rgd->rd_rs.rs_owner != sdp->sd_lockstruct.ls_jid + 1)
  {
    rgd->rd_rs.rs_owner = sdp->sd_lockstruct.ls_jid + 1;
    gfs_rgrp_save_out(&rgd->rd_rs, rgd->rd_gl->gl_lvb);
    gfs_sync_lvb(rgd->rd_sbd, rgd->rd_gl);
  }

  EXIT(GFN_RGRP_LVB_OWN);
}


/**
 * gfs_rgrp_verify - Verify that a resource group is consistent
 * @sdp: the filesystem
 * @rgd: the rgrp
 *
 * Somebody should have already called gfs_glock_rg() on this RG.
 */

void gfs_rgrp_verify(gfs_rgrpd_t *rgd)
{
  gfs_bitmap_t *bits = NULL;
  uint32 length = rgd->rd_ri.ri_length;
  uint32 count[4], tmp;
  int buf, x;

  ENTER(GFN_RGRP_VERIFY);


  for (x = 0; x < 4; x++)
  {
    count[x] = 0;

    for (buf = 0; buf < length; buf++)
    {
      bits = &rgd->rd_bits[buf];
      count[x] += gfs_bitcount(BH_DATA(rgd->rd_bh[buf]) + bits->bi_offset, bits->bi_len, x);
    }

    db_printf("count[%d] = %u\n", x, count[x]);
  }


  if (count[0] != rgd->rd_rg.rg_free)
    con_printf("free data mismatch:  %u != %u\n", count[0], rgd->rd_rg.rg_free);

  tmp = rgd->rd_ri.ri_data -
    (rgd->rd_rg.rg_usedmeta + rgd->rd_rg.rg_freemeta) -
      (rgd->rd_rg.rg_useddi + rgd->rd_rg.rg_freedi) -
	rgd->rd_rg.rg_free;
  if (count[1] != tmp)
    con_printf("used data mismatch:  %u != %u\n", count[1], tmp);

  if (count[2] != rgd->rd_rg.rg_freemeta)
    con_printf("free metadata mismatch:  %u != %u\n", count[2], rgd->rd_rg.rg_freemeta);

  tmp = rgd->rd_rg.rg_usedmeta + (rgd->rd_rg.rg_useddi + rgd->rd_rg.rg_freedi);
  if (count[3] != tmp)
    con_printf("used metadata mismatch:  %u != %u\n", count[3], tmp);


  EXIT(GFN_RGRP_VERIFY);
}


/**
 * try_rgrp_fit - See if a given reservation will fit in a given RG
 * @rgd: the RG data
 * @res: the gfs_ipreserv_t structure describing the reservation
 *
 * Sets the $ir_datares field in @res.
 * Sets the $ir_metares field in @res.
 *
 * Returns: 1 on success, 0 on failure
 */

STATIC int try_rgrp_fit(gfs_rgrpd_t *rgd, gfs_ipreserv_t *res)
{
  uint32 freeblks = rgd->rd_rg.rg_free;
  uint32 freemeta = rgd->rd_rg.rg_freemeta;
  uint32 metares = res->ir_metablks;
  uint32 datares = res->ir_datablks;

  ENTER(GFN_TRY_RGRP_FIT);


  /*  First take care of the data blocks required  */

  if (freeblks < res->ir_datablks)
    goto fail;

  freeblks -= res->ir_datablks;


  /*  Then take care of the dinodes  */

  if (rgd->rd_rg.rg_freedi < res->ir_diblks)
    metares += res->ir_diblks - rgd->rd_rg.rg_freedi;


  /*  Then take care of the metadata blocks  */

  while (freemeta < metares)
  {
    if (freeblks < GFS_META_CLUMP)
      goto fail;

    freeblks -= GFS_META_CLUMP;
    datares += GFS_META_CLUMP;
    freemeta += GFS_META_CLUMP;
  }


  res->ir_rgd = rgd;
  res->ir_metares = metares;
  res->ir_datares = datares;


  EXIT(GFN_TRY_RGRP_FIT);
  return 1;


 fail:
  EXIT(GFN_TRY_RGRP_FIT);
  return 0;
}


/**
 * gfs_inplace_reserv - Reserve space in the filesystem
 * @ip: the inode to reserve space for
 *
 * Acquire resource group locks to allow for the maximum allocation
 * described by "res".
 *
 * This should probably become more complex again, but for now, let's go
 * for simple (one resource group) reservations.
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_inplace_reserv(gfs_inode_t *ip)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  gfs_rgrpd_t *rgd;
  gfs_ipreserv_t *res = &ip->i_res;
  uint64 hint;
  int loops = 0;
  int error = 0;

  ENTER(GFN_INPLACE_RESERV);


  GFS_ASSERT(res->ir_diblks || res->ir_datablks || res->ir_metablks, ); 
  res->ir_rgd = NULL;
  res->ir_dialloced = 0;
  res->ir_metaalloced = 0;
  res->ir_dataalloced = 0;

  hint = res->ir_target;


  error = gfs_rindex_hold(sdp);
  if (error)
    return error;


  for (;;)
  {
    error = -ENOSPC;
    rgd = gfs_rgrpd_get_with_hint(sdp, hint);
    if (!rgd)
    {
      hint = 0;
      if (++loops >= 2)
      {
	gfs_rindex_release(sdp);	
	break;
      }
      continue;
    }


    error = gfs_glock_rg(rgd, 0);
    if (error)
    {
      gfs_rgrpd_put(rgd);
      gfs_rindex_release(sdp);	
      goto out;
    }

    /* Check if we may alloc the requested types of blocks here */

    if (try_rgrp_fit(rgd, res))
      break;

    gfs_gunlock_rg(rgd, 0);


    hint = rgd->rd_ri.ri_addr + 1;
    gfs_rgrpd_put(rgd);
  }


  IFDB(GFN_INPLACE_RESERV,
       db_printf("\nInplace reservation:\n");
       db_printf("    diblks = %u\n", res->ir_diblks);
       db_printf("    datablks = %u\n", res->ir_datablks);
       db_printf("    metablks = %u\n", res->ir_metablks);
       db_printf("    target = %"FMT_64"u\n", res->ir_target);
       db_printf("    ir_rgd = %"FMT_64"u\n", res->ir_rgd->rd_ri.ri_addr);
       db_printf("    ir_metares = %u\n", res->ir_metares);
       db_printf("    ir_datares = %u\n", res->ir_datares);
       db_printf("  error = %d\n", error););

 out:
  EXIT(GFN_INPLACE_RESERV);

  return error;
}


/**
 * gfs_inplace_relse - release an inplace reservation
 * @ip: the inode the reservation was taken out on
 *
 * Release a reservation made by gfs_inplace_reserv().
 */

void gfs_inplace_relse(gfs_inode_t *ip)
{
  gfs_ipreserv_t *res = &ip->i_res;
  gfs_rgrpd_t *rgd = res->ir_rgd;

  ENTER(GFN_INPLACE_RELSE);

  IFDB(GFN_INPLACE_RELSE,
       db_printf("  di alloced %d of %d\n", res->ir_dialloced, res->ir_diblks);
       db_printf("  meta alloced %d of %d\n", res->ir_metaalloced, res->ir_metares);
       db_printf("  data alloced %d of %d\n", res->ir_dataalloced, res->ir_datares););


  GFS_ASSERT(res->ir_dialloced <= res->ir_diblks,
	     con_printf("ir_dialloced = %u, ir_diblks = %u, trans = %d\n",
			res->ir_dialloced, res->ir_diblks, res->ir_trans););
  GFS_ASSERT(res->ir_metaalloced <= res->ir_metares,
	     con_printf("ir_metaalloced = %u, ir_metablks = %u, trans = %d\n",
			res->ir_metaalloced, res->ir_metablks, res->ir_trans););
  GFS_ASSERT(res->ir_dataalloced <= res->ir_datares,
	     con_printf("ir_dataalloced = %u, ir_datares = %u, trans = %d\n",
			res->ir_dataalloced, res->ir_datares, res->ir_trans););


  res->ir_rgd = NULL;

  gfs_gunlock_rg(rgd, 0);
  gfs_rgrpd_put(rgd);
  gfs_rindex_release(ip->i_sbd);

  EXIT(GFN_INPLACE_RELSE);
}


/**
 * blkalloc_internal - allocate a single block
 * @rgd: the resource group descriptor
 * @goal: the goal block in the RG
 * @old_state: the type of block to find
 * @new_state: the resulting block type
 * @do_it: if FALSE, we just find the block we would allocate
 *
 * This function never fails.
 *
 * Returns:  returns the block allocated
 */

STATIC uint32 blkalloc_internal(gfs_rgrpd_t *rgd,
				uint32 goal,
				unsigned char old_state, unsigned char new_state,
				int do_it)
{
  gfs_sbd_t *sdp = rgd->rd_sbd;
  gfs_bitmap_t *bits = NULL;
  uint32 length = rgd->rd_ri.ri_length;
  uint32 block = 0;
  int buf, x;

  ENTER(GFN_BLKALLOC_INTERNAL);


  for (buf = 0; buf < length; buf++)
  {
    bits = &rgd->rd_bits[buf];
    if (goal < (bits->bi_start + bits->bi_len) * GFS_NBBY)
      break;
  }

  GFS_ASSERT(buf < length, );
  goal -= bits->bi_start * GFS_NBBY;


  /*  "x <= length" because we're skipping over some of the first
      buffer when the goal is non-zero.  */

  for (x = 0; x <= length; x++)
  {
    block = gfs_bitfit(BH_DATA(rgd->rd_bh[buf]) + bits->bi_offset,
		       bits->bi_len, goal, old_state);
    if (block != BFITNOENT)
      break;

    buf = (buf + 1) % length;
    bits = &rgd->rd_bits[buf];
    goal = 0;
  }

  GFS_ASSERT(x <= length, );


  if (do_it)
  {
    gfs_trans_add_bh(sdp, rgd->rd_bh[buf]);
    gfs_setbit(BH_DATA(rgd->rd_bh[buf]) + bits->bi_offset,
	       bits->bi_len, block, new_state);
  }


  EXIT(GFN_BLKALLOC_INTERNAL);

  return bits->bi_start * GFS_NBBY + block;
}


/**
 * blkfree_internal - Free a block
 * @sdp: the filesystem
 * @bstart: the start of a run of blocks to free
 * @blen: the length of the block run
 * @new_state: the new state of the block
 *
 */

STATIC gfs_rgrpd_t *blkfree_internal(gfs_sbd_t *sdp, uint64 bstart, uint32 blen,
				     unsigned char new_state)
{
  gfs_rgrpd_t *rgd;
  gfs_bitmap_t *bits = NULL;
  uint32 length, rgrp_block, buf_block;
  int buf;

  ENTER(GFN_BLKFREE_INTERNAL);


  rgd = gfs_blk2rgrpd(sdp, bstart);
  GFS_ASSERT(rgd,
	     con_printf("block = %"FMT_64"u\n", bstart););

  length = rgd->rd_ri.ri_length;
  rgrp_block = bstart - rgd->rd_ri.ri_data1;


  while (blen--)
  {
    for (buf = 0; buf < length; buf++)
    {
      bits = &rgd->rd_bits[buf];
      if (rgrp_block < (bits->bi_start + bits->bi_len) * GFS_NBBY)
	break;
    }

    GFS_ASSERT(buf < length, );
    buf_block = rgrp_block - bits->bi_start * GFS_NBBY;
    rgrp_block++;

    gfs_trans_add_bh(sdp, rgd->rd_bh[buf]);
    gfs_setbit(BH_DATA(rgd->rd_bh[buf]) + bits->bi_offset,
	       bits->bi_len, buf_block, new_state);
  }


  EXIT(GFN_BLKFREE_INTERNAL);

  return rgd;
}


/**
 * clump_alloc - Allocate a clump of metadata
 * @sdp: the filesystem
 * @rgd: the resource group descriptor
 *
 * If this routine's performance sucks, try allocating
 * more than block at a time.
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int clump_alloc(gfs_rgrpd_t *rgd)
{
  gfs_sbd_t *sdp = rgd->rd_sbd;
  gfs_meta_header_t mh;
  osi_buf_t *bh[GFS_META_CLUMP];
  uint32 block, goal = 0;
  int x, y;
  int error;

  ENTER(GFN_CLUMP_ALLOC);


  if (osi_test_and_clear_bit(RDF_SYNC_B4_ALLOC, &rgd->rd_flags))
    gfs_log_flush(sdp, rgd->rd_gl);


  gfs_memset(&mh, 0, sizeof(gfs_meta_header_t));
  mh.mh_magic = GFS_MAGIC;
  mh.mh_type = GFS_METATYPE_NONE;


  for (x = 0; x < GFS_META_CLUMP; x++)
  {
    block = blkalloc_internal(rgd, goal, GFS_BLKST_FREE, GFS_BLKST_FREEMETA, TRUE);

    bh[x] = gfs_dgetblk(sdp, rgd->rd_ri.ri_data1 + block, rgd->rd_gl);
    if (!bh[x])
    {
      error = -EIO;
      goto fail;
    }
    gfs_prep_new_buffer(bh[x]);

    gfs_meta_header_out(&mh, BH_DATA(bh[x]));
    ((gfs_meta_header_t *)BH_DATA(bh[x]))->mh_generation = 0;  /*  Zero out the generation number.  */

    goal = block;
  }


  for (y = 0; y < GFS_META_CLUMP; y++)
    gfs_dwrite(sdp, bh[y], DW_DIRTY | DW_START);

  for (y = 0; y < GFS_META_CLUMP; y++)
  {
    error = gfs_dwrite(sdp, bh[y], DW_WAIT);
    if (error)
      goto fail;
  }

  for (y = 0; y < GFS_META_CLUMP; y++)
    gfs_drelse(bh[y]);


  GFS_ASSERT(rgd->rd_rg.rg_free >= GFS_META_CLUMP, );
  rgd->rd_rg.rg_free -= GFS_META_CLUMP;
  rgd->rd_rg.rg_freemeta += GFS_META_CLUMP;


  EXIT(GFN_CLUMP_ALLOC);

  return 0;



 fail:
  while (x--)
    gfs_drelse(bh[x]);

  EXIT(GFN_CLUMP_ALLOC);
  return error;
}


/**
 * gfs_blkalloc - Allocate a data block
 * @ip: the inode to allocate the data block for
 * @block: the block allocated
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_blkalloc(gfs_inode_t *ip, uint64 *block)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  gfs_ipreserv_t *res = &ip->i_res;
  gfs_rgrpd_t *rgd = res->ir_rgd;
  uint32 goal;
  int same;

  ENTER(GFN_BLKALLOC);

  GFS_ASSERT(rgd, );


  if (osi_test_and_clear_bit(RDF_SYNC_B4_ALLOC, &rgd->rd_flags))
    gfs_log_flush(sdp, rgd->rd_gl);


  /*  Any necessary quota checking goes here  */


  same = (rgd->rd_ri.ri_addr == ip->i_di.di_goal_rgrp);
  goal = (same) ? ip->i_di.di_goal_dblk : 0;

  *block = blkalloc_internal(rgd, goal, GFS_BLKST_FREE, GFS_BLKST_USED, TRUE);

  if (!same)
  {
    ip->i_di.di_goal_rgrp = rgd->rd_ri.ri_addr;
    ip->i_di.di_goal_mblk = 0;
  }
  ip->i_di.di_goal_dblk = *block;

  *block += rgd->rd_ri.ri_data1;


  GFS_ASSERT(rgd->rd_rg.rg_free, );
  rgd->rd_rg.rg_free--;

  gfs_trans_add_bh(sdp, rgd->rd_bh[0]);
  gfs_rgrp_out(&rgd->rd_rg, BH_DATA(rgd->rd_bh[0]));
  rgrp_lvb_own(rgd);

  res->ir_dataalloced++;


  EXIT(GFN_BLKALLOC);

  return 0;
}


/**
 * gfs_metaalloc - Allocate a metadata block to a file
 * @ip:  the file
 * @block: the block allocated
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_metaalloc(gfs_inode_t *ip, uint64 *block)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  gfs_ipreserv_t *res = &ip->i_res;
  gfs_rgrpd_t *rgd = res->ir_rgd;
  uint32 goal;
  int error;
  int same;

  ENTER(GFN_METAALLOC);

  GFS_ASSERT(rgd, );


  /*  Any necessary quota checking goes here  */


  if (!rgd->rd_rg.rg_freemeta)
  {
    error = clump_alloc(rgd);
    if (error)
    {
      EXIT(GFN_METAALLOC);
      return error;
    }
    res->ir_dataalloced += GFS_META_CLUMP;
  }


  same = (rgd->rd_ri.ri_addr == ip->i_di.di_goal_rgrp);
  goal = (same) ? ip->i_di.di_goal_mblk : 0;

  *block = blkalloc_internal(rgd, goal, GFS_BLKST_FREEMETA, GFS_BLKST_USEDMETA, TRUE);

  if (!same)
  {
    ip->i_di.di_goal_rgrp = rgd->rd_ri.ri_addr;
    ip->i_di.di_goal_dblk = 0;
  }
  ip->i_di.di_goal_mblk = *block;

  *block += rgd->rd_ri.ri_data1;


  GFS_ASSERT(rgd->rd_rg.rg_freemeta, );
  rgd->rd_rg.rg_freemeta--;
  rgd->rd_rg.rg_usedmeta++;

  gfs_trans_add_bh(sdp, rgd->rd_bh[0]);
  gfs_rgrp_out(&rgd->rd_rg, BH_DATA(rgd->rd_bh[0]));
  rgrp_lvb_own(rgd);

  res->ir_metaalloced++;


  EXIT(GFN_METAALLOC);

  return 0;
}


/**
 * gfs_dipick - Look for a free dinode
 * @sdp: the filesystem
 * @res: the reservation
 * @inum: the inode that the allocator will pick when you call gfs_dialloc()
 *
 */

void gfs_dipick(gfs_sbd_t *sdp, gfs_ipreserv_t *res, gfs_inum_t *inum)
{
  gfs_rgrpd_t *rgd = res->ir_rgd;
  unsigned char from, to;

  ENTER(GFN_DIPICK);

  GFS_ASSERT(rgd, );


  /*  If there are free dinodes use those first  */

  if (rgd->rd_rg.rg_freedi)
  {
    *inum = rgd->rd_rg.rg_freedi_list;
    GFS_ASSERT(inum->no_formal_ino && inum->no_addr, );
    goto out;
  }


  /*  Else, pick a metadata block  */

  if (rgd->rd_rg.rg_freemeta)
  {
    from = GFS_BLKST_FREEMETA;
    to = GFS_BLKST_USEDMETA;
  }
  else
  {
    from = GFS_BLKST_FREE;
    to = GFS_BLKST_FREEMETA;
  }

  inum->no_formal_ino = 0;
  inum->no_addr = blkalloc_internal(rgd, 0, from, to, FALSE);
  inum->no_addr += rgd->rd_ri.ri_data1;


 out:
  EXIT(GFN_DIPICK);
}


/**
 * dialloc_preexist - Allocate a preexisting unused dinode
 * @sdp: the filesystem
 * @res: the reservation
 * @expected_inum: the number of the dinode to be allocated
 * @expected_gl: the glock of the dinode to be allocated
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int dialloc_preexist(gfs_sbd_t *sdp, gfs_ipreserv_t *res,
			    gfs_inum_t *expected_inum, gfs_glock_t *expected_gl)

{
  gfs_rgrpd_t *rgd = res->ir_rgd;
  gfs_dinode_t *di;
  osi_buf_t *bh;
  gfs_inum_t next;
  uint32 flags;

  ENTER(GFN_DIALLOC_PREEXIST);

  GFS_ASSERT(GFS_INUM_EQUAL(&rgd->rd_rg.rg_freedi_list, expected_inum), );


  bh = gfs_dread(sdp, expected_inum->no_addr, expected_gl, DR_START | DR_WAIT);
  if (!bh)
  {
    EXIT(GFN_DIALLOC_PREEXIST);
    return -EIO;
  }

  di = (gfs_dinode_t *)BH_DATA(bh);

  flags = di->di_flags;
  flags = gfs32_to_cpu(flags);
  GFS_ASSERT(flags & GFS_DIF_UNUSED, );

  gfs_inum_in(&next, (char *)&di->di_next_unused);

  gfs_drelse(bh);


  rgd->rd_rg.rg_freedi--;
  rgd->rd_rg.rg_useddi++;
  rgd->rd_rg.rg_freedi_list = next;

  gfs_trans_add_bh(sdp, rgd->rd_bh[0]);
  gfs_rgrp_out(&rgd->rd_rg, BH_DATA(rgd->rd_bh[0]));
  rgrp_lvb_own(rgd);

  res->ir_dialloced++;


  EXIT(GFN_DIALLOC_PREEXIST);

  return 0;
}


/**
 * gfs_dialloc - Allocate a dinode
 * @sdp: the filesystem
 * @res: the reservation
 * @expected_inum: The expected inode number
 * @expected_gl: the glock of the dinode to be allocated
 *
 * Assumes expected_gl is held exclusively by the calling process.
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_dialloc(gfs_sbd_t *sdp, gfs_ipreserv_t *res,
		gfs_inum_t *expected_inum, gfs_glock_t *expected_gl)
{
  gfs_rgrpd_t *rgd = res->ir_rgd;
  uint64 block;
  uint32 goal;
  int error;

  ENTER(GFN_DIALLOC);

  GFS_ASSERT(rgd, );


  /*  Any necessary quota checking goes here  */


  if (rgd->rd_rg.rg_freedi)
  {
    error = dialloc_preexist(sdp, res, expected_inum, expected_gl);
    EXIT(GFN_DIALLOC);
    return error;
  }


  if (!rgd->rd_rg.rg_freemeta)
  {
    error = clump_alloc(rgd);
    if (error)
    {
      EXIT(GFN_DIALLOC);
      return error;
    }
    res->ir_dataalloced += GFS_META_CLUMP;
  }


  goal = expected_inum->no_addr - rgd->rd_ri.ri_data1;
  block = blkalloc_internal(rgd, goal, GFS_BLKST_FREEMETA, GFS_BLKST_USEDMETA, TRUE);
  block += rgd->rd_ri.ri_data1;

  GFS_ASSERT(block == expected_inum->no_addr, );


  GFS_ASSERT(rgd->rd_rg.rg_freemeta, );
  rgd->rd_rg.rg_freemeta--;
  rgd->rd_rg.rg_useddi++;

  gfs_trans_add_bh(sdp, rgd->rd_bh[0]);
  gfs_rgrp_out(&rgd->rd_rg, BH_DATA(rgd->rd_bh[0]));
  rgrp_lvb_own(rgd);

  res->ir_dialloced++;
  res->ir_metaalloced++;


  EXIT(GFN_DIALLOC);

  return 0;
}


/**
 * gfs_blkfree - free a piece of data
 * @ip: the inode these blocks are being free from
 * @bstart: the start of a run of blocks to free
 * @blen: the length of the block run
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_blkfree(gfs_inode_t *ip, uint64 bstart, uint32 blen)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  gfs_rgrpd_t *rgd;
  int error;

  ENTER(GFN_BLKFREE);

  rgd = blkfree_internal(sdp, bstart, blen, GFS_BLKST_FREE);

  rgd->rd_rg.rg_free += blen;

  osi_set_bit(RDF_SYNC_B4_ALLOC, &rgd->rd_flags);

  gfs_trans_add_bh(sdp, rgd->rd_bh[0]);
  gfs_rgrp_out(&rgd->rd_rg, BH_DATA(rgd->rd_bh[0]));
  rgrp_lvb_own(rgd);

  error = gfs_add_gl_dependency(sdp, ip->i_gl, rgd->rd_gl);
  gfs_rgrpd_put(rgd);

  EXIT(GFN_BLKFREE);

  return error;
}


/**
 * gfs_metafree - free a piece of metadata
 * @ip: the inode these blocks are being free from
 * @bstart: the start of a run of blocks to free
 * @blen: the length of the block run
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_metafree(gfs_inode_t *ip, uint64 bstart, uint32 blen)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  gfs_rgrpd_t *rgd;
  int error;

  ENTER(GFN_METAFREE);

  rgd = blkfree_internal(sdp, bstart, blen, GFS_BLKST_FREEMETA);

  GFS_ASSERT(rgd->rd_rg.rg_usedmeta >= blen, );
  rgd->rd_rg.rg_usedmeta -= blen;
  rgd->rd_rg.rg_freemeta += blen;

  gfs_trans_add_bh(sdp, rgd->rd_bh[0]);
  gfs_rgrp_out(&rgd->rd_rg, BH_DATA(rgd->rd_bh[0]));
  rgrp_lvb_own(rgd);

  error = gfs_add_gl_dependency(sdp, ip->i_gl, rgd->rd_gl);
  gfs_rgrpd_put(rgd);

  EXIT(GFN_METAFREE);

  return error;
}


/**
 * gfs_difree - free a piece of metadata
 * @ip: the inode representing the dinode to free
 * @rgd: the resource group that contains the dinode
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_difree(gfs_inode_t *ip, gfs_rgrpd_t *rgd)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  osi_buf_t *dibh;
  gfs_dinode_t *di;
  uint64 ctime;
  uint32 flags;
  int error;

  ENTER(GFN_DIFREE);


  dibh = gfs_dread(sdp, ip->i_num.no_addr, ip->i_gl, DR_START | DR_WAIT);
  if (!dibh)
  {
    error = -EIO;
    goto out;
  }
  di = (gfs_dinode_t *)BH_DATA(dibh);

  gfs_trans_add_bh(sdp, dibh);

  ctime = osi_current_time();
  di->di_ctime = cpu_to_gfs64(ctime);

  flags = gfs32_to_cpu(di->di_flags);
  flags |= GFS_DIF_UNUSED;
  di->di_flags = cpu_to_gfs32(flags);

  gfs_inum_out(&rgd->rd_rg.rg_freedi_list, (char *)&di->di_next_unused);

  gfs_drelse(dibh);


  GFS_ASSERT(rgd->rd_rg.rg_useddi, );
  rgd->rd_rg.rg_useddi--;
  rgd->rd_rg.rg_freedi++;
  rgd->rd_rg.rg_freedi_list = ip->i_num;

  gfs_trans_add_bh(sdp, rgd->rd_bh[0]);
  gfs_rgrp_out(&rgd->rd_rg, BH_DATA(rgd->rd_bh[0]));
  rgrp_lvb_own(rgd);


  error = gfs_add_gl_dependency(sdp, ip->i_gl, rgd->rd_gl);


 out:
  EXIT(GFN_DIFREE);

  return error;
}


/**
 * gfs_rlist_add - do stuff
 * @sdp: the filesystem
 * @rlist: the list of resource groups
 * @rgrp: the resource group to add
 *
 * Figure out what RG a block belongs to and add that RG to the list
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_rlist_add(gfs_sbd_t *sdp, gfs_rgrp_list_t *rlist, uint64 block)
{
  gfs_rgrpd_t *rgd, **tmp;
  unsigned int new_space;
  int x;
  int error = 0;
  int do_put = TRUE;

  ENTER(GFN_RLIST_ADD);


  GFS_ASSERT(rlist->rl_rgrps <= rlist->rl_space, );


  rgd = gfs_blk2rgrpd(sdp, block);
  GFS_ASSERT(rgd,
	     con_printf("block = %"FMT_64"u\n", block););


  for (x = 0; x < rlist->rl_rgrps; x++)
    if (rlist->rl_list[x] == rgd)
      goto out;


  if (rlist->rl_rgrps == rlist->rl_space)
  {
    new_space = rlist->rl_space + 10;

    tmp = (gfs_rgrpd_t **)gmalloc(new_space * sizeof(gfs_rgrpd_t *));
    if (!tmp)
    {
      error = -EIO;
      goto out;
    }

    if (rlist->rl_list)
    {
      gfs_memcpy(tmp, rlist->rl_list, rlist->rl_space * sizeof(gfs_rgrpd_t *));
      gfree(rlist->rl_list, rlist->rl_space * sizeof(gfs_rgrpd_t *));
    }

    rlist->rl_list = tmp;
    rlist->rl_space = new_space;
  }


  do_put = FALSE;
  rlist->rl_list[rlist->rl_rgrps++] = rgd;


 out:
  if (do_put)
    gfs_rgrpd_put(rgd);

  EXIT(GFN_RLIST_ADD);

  return error;
}


/**
 * gfs_rlist_sort - sort a list of resource groups
 * @list: the list of resource groups
 *
 * Dammit!!!  Get a better sorting algorithm.
 */

void gfs_rlist_sort(gfs_rgrp_list_t *rlist)
{
  gfs_rgrpd_t **list = rlist->rl_list, *tmp;
  uint32 rgrps = rlist->rl_rgrps;
  int x, changed;

  ENTER(GFN_RLIST_SORT);


  if (!rgrps)
  {
    EXIT(GFN_RLIST_SORT);
    return;
  }


  do
  {
    changed = FALSE;

    for (x = 1; x < rgrps; x++)
    {
      if (list[x - 1]->rd_ri.ri_addr > list[x]->rd_ri.ri_addr)
      {
	tmp = list[x - 1];
	list[x - 1] = list[x];
	list[x] = tmp;
	changed = TRUE;
      }
    }
  }
  while (changed);



  EXIT(GFN_RLIST_SORT);
}


/**
 * gfs_rlist_free - free a resource group list
 * @list: the list of resource groups
 *
 */

void gfs_rlist_free(gfs_rgrp_list_t *rlist)
{
  int x;

  ENTER(GFN_RLIST_FREE);

  if (rlist->rl_list)
  {
    for (x = 0; x < rlist->rl_rgrps; x++)
      gfs_rgrpd_put(rlist->rl_list[x]);

    gfree(rlist->rl_list, rlist->rl_space * sizeof(gfs_rgrpd_t *));
  }

  EXIT(GFN_RLIST_FREE);
}





