/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portions Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name:  bmap.c
**
**    Description:
**      These routines translate from extents in a file to block numbers.
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"

#include "blklist.h"
#include "bmap.h"
#include "dio.h"
#include "file.h"
#include "glock.h"
#include "inode.h"
#include "log.h"
#include "trans.h"



struct metapath
{
  uint64              mp_list[GFS_MAX_META_HEIGHT];
};
typedef struct metapath metapath_t;

typedef int (*block_call_t)(gfs_inode_t *ip, osi_buf_t *bh, uint64 *top, uint64 *bottom, int height, void *data);

struct strip_mine
{
  int                 sm_first;
  unsigned int        sm_height;
};
typedef struct strip_mine strip_mine_t;





/**
 * gfs_unstuff_dinode - Unstuff a dinode when the data has grown too big
 * @ip: The GFS inode to unstuff
 *
 * This routine unstuffs a dinode and returns it to a "normal" state such 
 * that the height can be grown in the traditional way.
 *
 * Returns: 0 on success, -EXXXX on failure
 */

int gfs_unstuff_dinode(gfs_inode_t *ip)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  osi_buf_t *bh, *dibh;
  uint64 block = 0;
  int error;
  int journaled = gfs_is_jdata(ip);

  ENTER(GFN_UNSTUFF_DINODE);

  GFS_ASSERT(gfs_is_stuffed(ip), );


  dibh = gfs_get_inode_buffer(ip);
  if (!dibh)
  {
    error = -EIO;
    goto fail;
  }


  if (ip->i_di.di_size)
  {
    /*  Get a free block, fill it with the stuffed data, and write it out to disk  */

    if (journaled)
      error = gfs_metaalloc(ip, &block);
    else
      error = gfs_blkalloc(ip, &block);

    if (error)
      goto fail_di_relse;

    bh = gfs_get_data_buffer(ip, block, TRUE);
    if (!bh)
    {
      error = -EIO;
      goto fail_di_relse;
    }

    gfs_buffer_copy_tail(bh, (journaled) ? sizeof(gfs_jdata_t) : 0,
			 dibh, sizeof(gfs_dinode_t));

    if (journaled)
      gfs_drelse(bh);
    else
    {
      error = gfs_dwrite(sdp, bh, DW_DIRTY | DW_START | DW_WAIT);
      if (error)
      {
	gfs_drelse(bh);
	goto fail_di_relse;
      }

      gfs_dforget(bh);
    }
  }


  /*  Set up the pointer to the new block  */

  gfs_trans_add_bh(sdp, dibh);

  gfs_buffer_clear_tail(dibh, sizeof(gfs_dinode_t));

  if (ip->i_di.di_size)
  {
    ((uint64 *)(BH_DATA(dibh) + sizeof(gfs_dinode_t)))[0] = cpu_to_gfs64(block);
    ip->i_di.di_blocks++; 
  }

  ip->i_di.di_height = 1;

  gfs_dinode_out(&ip->i_di, BH_DATA(dibh));
  gfs_drelse(dibh);


  EXIT(GFN_UNSTUFF_DINODE);

  return 0;


 fail_di_relse:
  gfs_drelse(dibh);

 fail:
  EXIT(GFN_UNSTUFF_DINODE);
  return error; 
}


/**
 * gfs_calc_tree_height - Calculate the height of a metadata tree
 * @ip: The GFS inode
 * @size: The proposed size of the file
 *
 * Work out how tall a metadata tree needs to be in order to accommodate a
 * file of a particular size. If size is less than the current size of
 * the inode, then the current size of the inode is used instead of the
 * supplied one.
 *
 * Returns: the height the tree should be
 */

unsigned int gfs_calc_tree_height(gfs_inode_t *ip, uint64 size)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  uint64 sz;
  unsigned int height;
  uint64 space, old_space;
  unsigned int bsize;

  ENTER(GFN_CALC_TREE_HEIGHT);


  sz = (size > ip->i_di.di_size) ? size : ip->i_di.di_size;

  if (sz <= (sdp->sd_sb.sb_bsize - sizeof(gfs_dinode_t)))
  {
    EXIT(GFN_CALC_TREE_HEIGHT);
    return 0;
  }


  bsize = (gfs_is_jdata(ip)) ? sdp->sd_jbsize : sdp->sd_sb.sb_bsize;

  height = 1;
  space = sdp->sd_diptrs * bsize;


  while (sz > space)
  {
    old_space = space;

    height++;
    space *= sdp->sd_inptrs;

    if (space / sdp->sd_inptrs != old_space ||
	space % sdp->sd_inptrs != 0)
      break;
  }


  EXIT(GFN_CALC_TREE_HEIGHT);

  return height;
}


/**
 * gfs_build_height - Build a metadata tree of the requested height
 * @ip: The GFS inode
 * @height: The height to build to
 *
 * This routine makes sure that the metadata tree is tall enough to hold
 * "size" bytes of data.
 *
 * Returns: 0 on success, -EXXXX on failure
 */

int build_height(gfs_inode_t *ip, int height)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  osi_buf_t *bh, *dibh;
  uint64 block;
  int error;

  ENTER(GFN_BUILD_HEIGHT);


  while (ip->i_di.di_height < height)
  {
    dibh = gfs_get_inode_buffer(ip);
    if (!dibh)
    {
      error = -EIO;
      goto fail;
    }


    /*  Get a new block, fill it with the old direct pointers and write it out  */

    error = gfs_metaalloc(ip, &block);
    if (error)
      goto fail_drelse;

    bh = gfs_dread(sdp, block, ip->i_gl, DR_FORCE | DR_START | DR_WAIT);
    if (!bh)
    {
      error = -EIO;
      goto fail_drelse;
    }

    gfs_trans_add_bh(sdp, bh);
    gfs_metatype_set(bh, GFS_METATYPE_IN, GFS_FORMAT_IN);
    gfs_buffer_copy_tail(bh, sizeof(gfs_indirect_t), dibh, sizeof(gfs_dinode_t));

    gfs_drelse(bh);


    /*  Set up the new direct pointer and write it out to disk  */

    gfs_trans_add_bh(sdp, dibh);

    gfs_buffer_clear_tail(dibh, sizeof(gfs_dinode_t));
    ((uint64 *)(BH_DATA(dibh) + sizeof(gfs_dinode_t)))[0] = cpu_to_gfs64(block);

    ip->i_di.di_height++;
    ip->i_di.di_blocks++;

    gfs_dinode_out(&ip->i_di, BH_DATA(dibh));
    gfs_drelse(dibh);
  }


  EXIT(GFN_BUILD_HEIGHT);

  return 0;



 fail_drelse:
  gfs_drelse(dibh);

 fail:
  EXIT(GFN_BUILD_HEIGHT);
  return error; 
}


/**
 * find_metapath - Find path through the metadata tree
 * @ip: The inode pointer
 * @mp: The metapath to return the result in
 * @block: The disk block to look up
 *
 *   This routine returns a metapath_t structure that defines a path through
 *   the metadata of inode "ip" to get to block "block".
 *
 *   Example:
 *   Given:  "ip" is a height 3 file, "offset" is 101342453, and this is a
 *   filesystem with a blocksize of 4096.
 *
 *   find_metapath() would return a metapath_t structure set to:
 *   mp_offset = 101342453, mp_height = 3, mp_list[0] = 0, mp_list[1] = 48,
 *   and mp_list[2] = 165.
 *
 *   That means that in order to get to the block containing the byte at
 *   offset 101342453, we would load the indirect block pointed to by pointer
 *   0 in the dinode.  We would then load the indirect block pointed to by
 *   pointer 48 in that indirect block.  We would then load the data block
 *   pointed to by pointer 165 in that indirect block.
 *
 *             ----------------------------------------
 *             | Dinode |                             |
 *             |        |                            4|
 *             |        |0 1 2 3 4 5                 9|
 *             |        |                            6|
 *             ----------------------------------------
 *                       |
 *                       |
 *                       V
 *             ----------------------------------------
 *             | Indirect Block                       |
 *             |                                     5|
 *             |            4 4 4 4 4 5 5            1|
 *             |0           5 6 7 8 9 0 1            2|
 *             ----------------------------------------
 *                                |
 *                                |
 *                                V
 *             ----------------------------------------
 *             | Indirect Block                       |
 *             |                         1 1 1 1 1   5|
 *             |                         6 6 6 6 6   1|
 *             |0                        3 4 5 6 7   2|
 *             ----------------------------------------
 *                                           |
 *                                           |
 *                                           V
 *             ----------------------------------------
 *             | Data block containing offset         |
 *             |            101342453                 |
 *             |                                      |
 *             |                                      |
 *             ----------------------------------------
 *
 */

STATIC void find_metapath(gfs_inode_t *ip, metapath_t *mp, uint64 block)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  unsigned int i;

  ENTER(GFN_FIND_METAPATH);

  for (i = ip->i_di.di_height; i--; )
  {
    mp->mp_list[i] = block % sdp->sd_inptrs;
    block /= sdp->sd_inptrs;
  }

  EXIT(GFN_FIND_METAPATH);
}


/**
 * metapointer - Return pointer to start of metadata in a buffer
 * @bh: The buffer
 * @level: The metadata level (0 = dinode)
 * @mp: The metapath 
 *
 * Return a pointer to the block number of the next level of the metadata
 * tree given a buffer containing the pointer to the current level of the
 * metadata tree.
 */ 

static __inline__ uint64 *metapointer(osi_buf_t *bh, int level, metapath_t *mp)
{
  int head_size = (level > 0) ? sizeof(gfs_indirect_t) : sizeof(gfs_dinode_t);
  return ((uint64 *)(BH_DATA(bh) + head_size)) + mp->mp_list[level];
}


/**
 * get_metablock - Get the next metadata block in metadata tree
 * @ip: The GFS inode
 * @bh: Buffer containing the pointers to metadata blocks
 * @level: The level of the tree (0 = dinode)
 * @mp: The metapath
 * @create: Non-zero if we may create a new meatdata block
 * @new: Used to indicate if we did create a new metadata block
 * @block: the returned disk block number
 *
 * Given a metatree, complete to a particular level, checks to see if the next
 * level of the tree exists. If not the next level of the tree is created.
 * The block number of the next level of the metadata tree is returned.
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int get_metablock(gfs_inode_t *ip,
			 osi_buf_t *bh, int level, metapath_t *mp,
			 int create,
			 int *new, uint64 *block)
{
  uint64 *ptr = metapointer(bh, level, mp);
  int error = 0;

  ENTER(GFN_GET_METABLOCK);

  *new = 0;
  *block = 0;


  if (*ptr)
  {
    *block = gfs64_to_cpu(*ptr);
    goto out;
  }

  if (!create)
    goto out;


  error = gfs_metaalloc(ip, block);
  if (error)
    goto out;

  gfs_trans_add_bh(ip->i_sbd, bh);

  *ptr = cpu_to_gfs64(*block);
  ip->i_di.di_blocks++;

  *new = 1;


 out:
  EXIT(GFN_GET_METABLOCK);
  return error;
}


/**
 * get_datablock - Get datablock number from metadata block
 * @ip: The GFS inode
 * @bh: The buffer containing pointers to datablocks
 * @mp: The metapath
 * @create: Non-zero if we may create a new data block
 * @new: Used to indicate if we created a new data block
 * @block: the returned disk block number
 *
 * Given a fully built metadata tree, checks to see if a particular data
 * block exists. It is created if it does not exist and the block number
 * on disk is returned.
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int get_datablock(gfs_inode_t *ip,
			 osi_buf_t *bh, metapath_t *mp,
			 int create,
			 int *new, uint64 *block)
{
  int level = ip->i_di.di_height - 1;
  uint64 *ptr = metapointer(bh, level, mp);
  int error = 0;

  ENTER(GFN_GET_DATABLOCK);

  *new = 0;
  *block = 0;


  if (*ptr)
  {
    *block = gfs64_to_cpu(*ptr);
    goto out;
  }

  if (!create)
    goto out;


  if (gfs_is_jdata(ip))
    error = gfs_metaalloc(ip, block);
  else
    error = gfs_blkalloc(ip, block);

  if (error)
    goto out;

  gfs_trans_add_bh(ip->i_sbd, bh);

  *ptr = cpu_to_gfs64(*block);
  ip->i_di.di_blocks++;

  *new = 1;


 out:
  EXIT(GFN_GET_DATABLOCK);
  return error;
}


/**
 * gfs_inode_get_block - Map a block from an inode to a disk block
 * @ip: The GFS inode
 * @blknum: The block number (inode relative)
 * @new: Value/Result argument (1 = may create/did create new blocks)
 *
 * Find the block number on the current device which corresponds to an
 * inode's block. If the block had to be created, "new" will be set.
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_inode_get_block(gfs_inode_t *ip,
			uint64 blknum,
			int *new, uint64 *block)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  osi_buf_t *bh;
  metapath_t mp;
  int create = *new;
  unsigned int bsize;
  unsigned int height;
  unsigned int i, end_of_metadata;
  int error = 0;

  ENTER(GFN_INODE_GET_BLOCK);

  *new = 0;
  *block = 0;


  if (gfs_is_stuffed(ip))
  {
    *block = ip->i_num.no_addr;
    goto out;
  }

  bh = gfs_get_inode_buffer(ip);
  if (!bh)
  {
    error = -EIO;
    goto fail;
  }

  if (create)
    gfs_trans_add_bh(sdp, bh);


  bsize = (gfs_is_jdata(ip)) ? sdp->sd_jbsize : sdp->sd_sb.sb_bsize;

  height = gfs_calc_tree_height(ip, (blknum + 1) * bsize);
  if (ip->i_di.di_height < height)
  {
    if (!create)
    {
      gfs_drelse(bh);
      goto out;
    }

    error = build_height(ip, height);
    if (error)
    {
      gfs_drelse(bh);
      goto fail;
    }
  }


  find_metapath(ip, &mp, blknum);
  end_of_metadata = ip->i_di.di_height - 1;

  for (i = 0; i < end_of_metadata; i++)
  {
    error = get_metablock(ip, bh, i, &mp, create, new, block);
    gfs_drelse(bh);
    if (error)
      goto fail;
    if (!*block)
      goto out;

    bh = gfs_get_meta_buffer(ip, i + 1, *block, *new);
    if (!bh)
    {
      error = -EIO;
      goto fail;
    }
  }

  error = get_datablock(ip, bh, &mp, create, new, block);
  gfs_drelse(bh);
  if (error)
    goto fail;


 out:
  if (*new)
  {
    bh = gfs_get_inode_buffer(ip);
    GFS_ASSERT(bh, );  /*  This buffer should be already pinned  */
    gfs_dinode_out(&ip->i_di, BH_DATA(bh));
    gfs_drelse(bh);
  }

  EXIT(GFN_INODE_GET_BLOCK);

  return 0;



 fail:
  EXIT(GFN_INODE_GET_BLOCK);
  return error;
}


/**
 * do_grow - Make a file look bigger than it is
 * @ip: the inode
 * @size: the size to set the file to
 *
 * Called with an exclusive lock on @ip.
 *
 * Returns: 0 on succes, -EXXX on failure
 */

STATIC int do_grow(gfs_inode_t *ip, uint64 size)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  gfs_ipreserv_t *res = &ip->i_res;
  osi_buf_t *dibh;
  int h;
  int error;

  ENTER(GFN_DO_GROW);


  res->ir_diblks = 0;
  res->ir_metablks = sdp->sd_max_height;
  res->ir_datablks = 1;
  res->ir_target = ip->i_di.di_goal_rgrp;
  res->ir_trans = TRANS_TRUNC_GROW;

  error = gfs_inplace_reserv(ip);
  if (error)
    goto fail;


  /*  Trans may require:
      Full extention of the metadata tree, block allocation,
      and a dinode modification  */

  error = gfs_trans_begin(sdp, TRANS_TRUNC_GROW, sdp->sd_max_height + res->ir_rgd->rd_ri.ri_length + 1, 0);
  if (error)
    goto fail_ipres;


  if (size > sdp->sd_sb.sb_bsize - sizeof(gfs_dinode_t))
  {
    if (gfs_is_stuffed(ip))
    {
      error = gfs_unstuff_dinode(ip);
      if (error)
	goto fail_end_trans;
    }
  
    h = gfs_calc_tree_height(ip, size);
    if (ip->i_di.di_height < h)
    {
      error = build_height(ip, h);
      if (error)
	goto fail_end_trans;
    }
  }


  ip->i_di.di_size = size;
  ip->i_di.di_mtime = ip->i_di.di_ctime = osi_current_time();
  
  dibh = gfs_get_inode_buffer(ip);
  if (!dibh)
  {
    error = -EIO;
    goto fail_end_trans;
  }

  gfs_trans_add_bh(sdp, dibh);
  gfs_dinode_out(&ip->i_di, BH_DATA(dibh));
  gfs_drelse(dibh);


  gfs_trans_end(sdp);

  gfs_inplace_relse(ip);


  EXIT(GFN_DO_GROW);

  return 0;



 fail_end_trans:
  gfs_trans_mark(sdp, TRF_POISONED);
  gfs_trans_end(sdp);

 fail_ipres:
  gfs_inplace_relse(ip);

 fail:
  EXIT(GFN_DO_GROW);
  return error; 
}


/**
 * recursive_scan - recursively scan through the end of a file
 * @ip: the inode
 * @mp: the path through the metadata to the point to start
 * @height: the height the recursion is at
 * @block: the indirect block to look at
 * @first: TRUE if this is the first block
 * @bc: the call to make for each piece of metadata
 * @data: data opaque to this function to pass to @bc
 *
 * When this is first called @height and @block should be zero and
 * @first should be TRUE.
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int recursive_scan(gfs_inode_t *ip, metapath_t *mp,
			  unsigned int height, uint64 block, int first,
			  block_call_t bc, void *data)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  osi_buf_t *bh = NULL;
  uint64 *top, *bottom;
  uint64 bn;
  int error;

  ENTER(GFN_RECURSIVE_SCAN);

  IFDB(GFN_RECURSIVE_SCAN,
       db_printf("  height = %d\n", height);
       db_printf("  block = %"FMT_64"u\n", block););


  if (!height)
  {
    bh = gfs_get_inode_buffer(ip);
    if (!bh)
    {
      error = -EIO;
      goto fail;
    }

    top = (uint64 *)(BH_DATA(bh) + sizeof(gfs_dinode_t)) + mp->mp_list[0];
    bottom = (uint64 *)(BH_DATA(bh) + sizeof(gfs_dinode_t)) + sdp->sd_diptrs;
  }
  else
  {
    bh = gfs_get_meta_buffer(ip, height, block, FALSE);
    if (!bh)
    {
      error = -EIO;
      goto fail;
    }

    top = (uint64 *)(BH_DATA(bh) + sizeof(gfs_indirect_t)) + ((first) ? mp->mp_list[height] : 0);
    bottom = (uint64 *)(BH_DATA(bh) + sizeof(gfs_indirect_t)) + sdp->sd_inptrs;
  }


  error = bc(ip, bh, top, bottom, height, data);
  if (error)
    goto fail;


  if (height < ip->i_di.di_height - 1)
    for ( ; top < bottom; top++, first = FALSE)
    {
      if (!*top)
	continue;

      bn = gfs64_to_cpu(*top);

      IFDB(GFN_RECURSIVE_SCAN,
	   db_printf("  Recursing into indirect block %"FMT_64"u\n", bn););

      error = recursive_scan(ip, mp,
			     height + 1, bn, first,
			     bc, data);
      if (error)
	goto fail;
    }
  

  gfs_drelse(bh);

  EXIT(GFN_RECURSIVE_SCAN);

  return 0;



 fail:
  if (bh)
    gfs_drelse(bh);

  EXIT(GFN_RECURSIVE_SCAN);
  return error;
}


/**
 * do_strip - Look for a layer a particular layer of the file and strip it off
 * @ip: the inode
 * @bh: A buffer of pointers
 * @top: The first pointer in the buffer
 * @bottom: One more than the last pointer
 * @height: the height this buffer is at
 * @data: a pointer to a strip_mine_t
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int do_strip(gfs_inode_t *ip, osi_buf_t *bh, uint64 *top, uint64 *bottom, int height, void *data)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  strip_mine_t *sm = (strip_mine_t *)data;
  gfs_rgrp_list_t rlist;
  gfs_rgrpd_t *rgd;
  osi_buf_t *dibh;
  uint64 bn, bstart;
  uint32 blen;
  uint64 *p;
  unsigned int blocks = 0;
  int x;
  int error;

  ENTER(GFN_DO_STRIP);


  if (!*top)
    sm->sm_first = FALSE;

  if (height != sm->sm_height)
  {
    error = 0;
    goto fail;
  }

  gfs_memset(&rlist, 0, sizeof(gfs_rgrp_list_t));

  if (sm->sm_first)
  {
    top++;
    sm->sm_first = FALSE;
  }


  error = gfs_rindex_hold(sdp);
  if (error)
    goto fail;

  bstart = 0;
  blen = 0;

  for (p = top; p < bottom; p++)
  {
    if (!*p)
      continue;

    bn = gfs64_to_cpu(*p);

    if (bstart + blen == bn)
      blen++;
    else
    {
      if (bstart)
      {
	error = gfs_rlist_add(sdp, &rlist, bstart);
	if (error)
	  goto fail_rlist_free;
      }

      bstart = bn;
      blen = 1;
    }
  }

  if (bstart)
  {
    error = gfs_rlist_add(sdp, &rlist, bstart);
    if (error)
      goto fail_rlist_free;
  }
  else
  {
    /*  Nothing to do  */
    error = 0;
    goto fail_rlist_free;
  }

  gfs_rlist_sort(&rlist);

  for (x = 0; x < rlist.rl_rgrps; x++)
  {
    rgd = rlist.rl_list[x];

    error = gfs_glock_rg(rgd, 0);
    if (error)
      goto fail_rg_gunlock;

    blocks += rgd->rd_ri.ri_length;
  }


  /*  Trans may require:
      All the bitmaps that were reserved. 
      One block for the dinode.
      One block for the indirect block being cleared.
      */

  error = gfs_trans_begin(sdp, TRANS_FILE_DEALLOC, blocks + 2, 0);
  if (error)
    goto fail_rg_gunlock;

  dibh = gfs_get_inode_buffer(ip);
  if (!dibh)
  {
    error = -EIO;
    goto fail_end_trans;
  }

  gfs_trans_add_bh(sdp, dibh);
  gfs_trans_add_bh(sdp, bh);

  bstart = 0;
  blen = 0;

  for (p = top; p < bottom; p++)
  {
    if (!*p)
      continue;

    bn = gfs64_to_cpu(*p);

    if (bstart + blen == bn)
      blen++;
    else
    {
      if (bstart)
      {
	if ((height == ip->i_di.di_height - 1) && !gfs_is_jdata(ip))
	  error = gfs_blkfree(ip, bstart, blen);
	else
	  error = gfs_metafree(ip, bstart, blen);

	if (error)
	  goto fail_drelse;
      }

      bstart = bn;
      blen = 1;
    }

    *p = 0;
    ip->i_di.di_blocks--;
  }

  if (bstart)
  {
    if ((height == ip->i_di.di_height - 1) && !gfs_is_jdata(ip))
      error = gfs_blkfree(ip, bstart, blen);
    else
      error = gfs_metafree(ip, bstart, blen);

    if (error)
      goto fail_drelse;
  }

  ip->i_di.di_mtime = ip->i_di.di_ctime = osi_current_time();

  gfs_dinode_out(&ip->i_di, BH_DATA(dibh));
  gfs_drelse(dibh);

  gfs_trans_end(sdp);


  for (x = 0; x < rlist.rl_rgrps; x++)
  {
    rgd = rlist.rl_list[x];
    gfs_gunlock_rg(rgd, 0);
  }
  gfs_rlist_free(&rlist);
  gfs_rindex_release(sdp);


  EXIT(GFN_DO_STRIP);

  return 0;



 fail_drelse:
  gfs_drelse(dibh);

 fail_end_trans:
  gfs_trans_mark(sdp, TRF_POISONED);
  gfs_trans_end(sdp);

 fail_rg_gunlock:
  while (x--)
  {
    rgd = rlist.rl_list[x];
    gfs_gunlock_rg(rgd, 0);
  }

 fail_rlist_free:
  gfs_rlist_free(&rlist);
  gfs_rindex_release(sdp);

 fail:
  EXIT(GFN_DO_STRIP);
  return error;
}



/**
 * gfs_shrink - make a file smaller
 * @ip: the inode
 * @size: the size to make the file
 *
 * Called with an exclusive lock on @ip.
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_shrink(gfs_inode_t *ip, uint64 size)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  gfs_glock_t *gl = ip->i_gl;
  metapath_t mp;
  strip_mine_t sm;
  osi_buf_t *bh;
  unsigned int height;
  uint64 bn;
  int not_new = 0;
  int error = 0;

  ENTER(GFN_SHRINK);


  GFS_ASSERT(!(gfs_is_jdata(ip) && size), );

  find_metapath(ip, &mp, (size) ? ((size - 1) >> sdp->sd_sb.sb_bsize_shift) : 0);


  /*  Get rid of all the data/metadata blocks  */

  for (height = ip->i_di.di_height; height--; )
  {
    sm.sm_first = (size) ? TRUE : FALSE;
    sm.sm_height = height;

    error = recursive_scan(ip, &mp, 0, 0, TRUE, do_strip, &sm);
    if (error)
      goto out;
  }


    /*  If we truncated in the middle of a block, zero out the leftovers.  */

  if (ip->i_di.di_height && (size % sdp->sd_sb.sb_bsize))
  {
    error = gfs_inode_get_block(ip, size >> sdp->sd_sb.sb_bsize_shift, &not_new, &bn);
    if (error)
      goto out;

    if (bn)
    {
      bh = gfs_dread(sdp, bn, gl, DR_START | DR_WAIT);
      if (!bh)
      {
	error = -EIO;
	goto out;
      }

      gfs_buffer_clear_tail(bh, size % sdp->sd_sb.sb_bsize);

      gfs_dwrite(sdp, bh, DW_DIRTY);
      gfs_drelse(bh);
    }
  }


  /*  Set the new size (and possibly the height)  */

  error = gfs_trans_begin(sdp, TRANS_SET_SIZE, 1, 0);
  if (error)
    goto out;

  bh = gfs_get_inode_buffer(ip);
  if (!bh)
  {
    gfs_trans_end(sdp);
    error = -EIO;
    goto out;
  }

  gfs_trans_add_bh(sdp, bh);

  ip->i_di.di_size = size;
  if (!size)
    ip->i_di.di_height = 0;
  ip->i_di.di_mtime = ip->i_di.di_ctime = osi_current_time();

  if (!ip->i_di.di_height && (size < sdp->sd_sb.sb_bsize - sizeof(gfs_dinode_t)))
    gfs_buffer_clear_tail(bh, sizeof(gfs_dinode_t) + size);

  gfs_dinode_out(&ip->i_di, BH_DATA(bh));
  gfs_drelse(bh);

  gfs_trans_end(sdp);


 out:
  EXIT(GFN_SHRINK);

  return error;
}


/**
 * gfs_truncate - make a file a give size
 * @ip: the inode
 * @size: the size to make the file
 * @cred: the caller's credentials
 *
 * The file size can grow, shrink, or stay the same size.
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_truncate(gfs_inode_t *ip, uint64 size, osi_cred_t *cred)
{
  int error;

  ENTER(GFN_TRUNCATE);


  GFS_ASSERT(ip->i_di.di_type == FILE_REG, );
  GFS_ASSERT(!gfs_is_jdata(ip), );


  error = gfs_glock_i(ip, 0);
  if (error)
    goto out;

  error = gfs_permission(ip, OSI_MAY_WRITE, cred);
  if (error)
  {
    gfs_gunlock_i(ip, 0);
    goto out;
  }


  if (size >= ip->i_di.di_size)  /*  Make bigger (or just modify the timestamps)  */
    error = do_grow(ip, size);
  else  /*  Make smaller  */
    error = gfs_shrink(ip, size);


  gfs_gunlock_i(ip, 0);


 out:
  EXIT(GFN_TRUNCATE);

  return error;
}


