/******************************************************************************
*******************************************************************************
**
**    Copyright 2000-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: 
**      dio.c
**
**    Description:
**      disk I/O (buffer cache) accounting wrappers
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "dio.h"
#include "fs.h"

#ifdef DEBUG_BUF



struct buf_call
{
  osi_list_head_t      list;

  osi_buf_t            *bh;
  unsigned int         count;

  char                 *file;
  unsigned int         line;
};
typedef struct buf_call buf_call_t;





/**
 * buf_hash - compute a hash on a buffer pointer
 * @bh: the buffer pointer
 *
 * Returns: the hash
 */

STATIC __inline__ uint32 buf_hash(osi_buf_t *bh)
{
  uint32 val;

  val = gfs_hash((char *)&bh, sizeof(osi_buf_t *));
  val &= (BUF_HASH_SIZE - 1);

  return val;
} 


/**
 * buf_alloc_brute - mark a buffer as allocated
 * @bh: the buffer
 * @file: the file containing the getblk
 * @line: the line it's on
 * @new: Set to TRUE if this is a new buffer
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int buf_alloc_brute(osi_buf_t *bh, char *file, unsigned int line, int *new)
{
  osi_list_head_t *tmp;
  buf_call_t *oc, *oc2;
  unsigned int index = buf_hash(bh);
  int found = FALSE;

  ENTER(GFN_BUF_ALLOC_BRUTE);

  *new = TRUE;


  oc = (buf_call_t *)gmalloc(sizeof(buf_call_t));
  if (!oc)
  { 
    EXIT(GFN_BUF_ALLOC_BRUTE);
    return -ENOMEM;
  }

  oc->bh = bh;
  oc->count = 1;

  oc->file = file;
  oc->line = line;


  osi_spin_lock(&buf_lock);

  for (tmp = buf_head[index].next; tmp != &buf_head[index]; tmp = tmp->next)
  {
    oc2 = osi_list_entry(tmp, buf_call_t, list);

    if (oc2->bh != bh)
      continue;

    *new = FALSE;

    oc2->count++;
    oc->count = oc2->count;

    if (oc2->line == line && !osi_strcmp(oc2->file, file) )
      found = TRUE;
  }

  if (!found)
    osi_list_add(&oc->list, &buf_head[index]);

  osi_spin_unlock(&buf_lock);


  if (found)
    gfree(oc, sizeof(buf_call_t));


  EXIT(GFN_BUF_ALLOC_BRUTE);

  return 0;
}


/**
 * buf_free_brute - mark a buffer as deallocated
 * @bh: the buffer
 * @file: the file containing the relse
 * @line: the line it's on
 * @last_one: Set to TRUE if this is the last allocation
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int buf_free_brute(osi_buf_t *bh, char *file, unsigned int line, int *last_one)
{
  unsigned int index = buf_hash(bh);
  osi_list_head_t *tmp, *head, *last;
  buf_call_t *oc;
  int found = FALSE;

  ENTER(GFN_BUF_FREE_BRUTE);

  *last_one = FALSE;


  osi_spin_lock(&buf_lock);

  last = head = &buf_head[index];
  for (tmp = head->next; tmp != head; tmp = last->next)
  {
    oc = osi_list_entry(tmp, buf_call_t, list);

    if (oc->bh != bh)
    {
      last = tmp;
      continue;
    }

    found = TRUE;

    GFS_ASSERT(oc->count,);
    oc->count--;

    if (!oc->count)
    {
      osi_list_del(&oc->list);
      gfree(oc, sizeof(buf_call_t));
      *last_one = TRUE;
    }
    else
      last = tmp;
  }

  osi_spin_unlock(&buf_lock);

   
  if (!found)
  {
    con_printf("GFS: freeing non allocated blk <%lX> (%s, %u)\n",
	       (unsigned long)bh, file, line);
    EXIT(GFN_BUF_FREE_BRUTE);
    return -EINVAL;
  }


  EXIT(GFN_BUF_FREE_BRUTE);

  return 0;
}


/**
 * buf_inc_stats - Increment the stats
 * @new:  TRUE if this is a new allocation
 *
 */

STATIC void buf_inc_stats(int new)
{
  ENTER(GFN_BUF_INC_STATS);


  osi_spin_lock(&buf_lock);

  if (new)
  {
    buf_stats.ct++;				
    buf_stats.ct_total++;				
    
    if (buf_stats.ct > buf_stats.ct_max)		
      buf_stats.ct_max = buf_stats.ct;		
  }

  buf_stats.refs++;				
  buf_stats.refs_total++;				
    
  if (buf_stats.refs > buf_stats.refs_max)		
    buf_stats.refs_max = buf_stats.refs;

  osi_spin_unlock(&buf_lock);


  EXIT(GFN_BUF_INC_STATS);						
}


/**
 * buf_dec_stats - Decrement the stats
 * @last_one:  TRUE if this was the last deallocation
 *
 */

STATIC void buf_dec_stats(int last_one)
{
  ENTER(GFN_BUF_DEC_STATS);

  osi_spin_lock(&buf_lock);

  if (last_one)
    buf_stats.ct--;

  buf_stats.refs--;

  osi_spin_unlock(&buf_lock);
  
  EXIT(GFN_BUF_DEC_STATS);
}


/**
 * buf_alloc_main - Handle a buffer allocation
 * @bh: the buffer
 * @file: the file the getblk occurs on
 * @line: the line
 *
 * Returns: 0 on success, -EXXX on failure
 */

int buf_alloc_main(osi_buf_t *bh, char *file, unsigned int line)
{
  int error;
  int new;

  ENTER(GFN_BUF_ALLOC_MAIN);

  error = buf_alloc_brute(bh, file, line, &new);
  if (error)
  {
    EXIT(GFN_BUF_ALLOC_MAIN);
    return error;
  }

  buf_inc_stats(new);

  EXIT(GFN_BUF_ALLOC_MAIN);

  return 0;
}


/**
 * gfs_dgetblk_wrapper - the wrapper around gfs_dgetblk()
 * @sdp: the filesystem
 * @blkno: the block number
 * @gl: the glock
 * @file: the file the getblk occurs on
 * @line: the line
 *
 * Returns: a new buffer
 */

osi_buf_t *gfs_dgetblk_wrapper(gfs_sbd_t *sdp, uint64 blkno, gfs_glock_t *gl, 
			       char *file, unsigned int line)
{
  osi_buf_t *bh = NULL;
  int error;

  ENTER(GFN_DGETBLK_WRAPPER);

  bh = gfs_dgetblk_arch(sdp, blkno, gl);
  if (!bh)
    goto out;

  error = buf_alloc_main(bh, file, line);
  if (error)
  {
    gfs_drelse(bh);
    bh = NULL;
  }

 out:
  EXIT(GFN_DGETBLK_WRAPPER);

  return bh;
}


/**
 * gfs_dread_wrapper - the wrapper around gfs_dread()
 * @sdp: the filesystem
 * @blkno: the block number
 * @gl: the glock
 * @flags: flags
 * @file: the file the dread occurs on
 * @line: the line
 *
 * Returns: a new buffer
 */

osi_buf_t *gfs_dread_wrapper(gfs_sbd_t *sdp, uint64 blkno, gfs_glock_t *gl, int flags, 
			     char *file, unsigned int line)
{
  osi_buf_t *bh = NULL;
  int error;

  ENTER(GFN_DREAD_WRAPPER);

  bh = gfs_dread_arch(sdp, blkno, gl, flags);
  if (!bh)
    goto out;

  error = buf_alloc_main(bh, file, line);
  if (error)
  {
    gfs_drelse(bh);
    bh = NULL;
  }

 out:
  EXIT(GFN_DREAD_WRAPPER);

  return bh;
}


/**
 * buf_free_main - Handle a buffer deallocation
 * @bh: the buffer
 * @file: the file the getblk occurs on
 * @line: the line
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int buf_free_main(osi_buf_t *bh, char *file, unsigned int line)
{
  int last_one;
  int error;

  ENTER(GFN_BUF_FREE_MAIN);

  if (!bh)
  {
    con_printf("GFS:  releasing NULL buffer (%s, %u)\n", file, line);
    EXIT(GFN_BUF_FREE_MAIN);
    return -EINVAL;
  }

  error = buf_free_brute(bh, file, line, &last_one);
  if (error)
  {
    EXIT(GFN_BUF_FREE_MAIN);
    return error;
  }

  buf_dec_stats(last_one);

  EXIT(GFN_BUF_FREE_MAIN);

  return 0;
}


/**
 * gfs_drelse_wrapper - the wrapper around gfs_drelse()
 * @bh: the buffer
 * @file: the file the gfs_drelse() occurs on
 * @line: the line
 *
 * Function description
 *
 * Returns: what is returned
 */

void gfs_drelse_wrapper(osi_buf_t *bh, char *file, unsigned int line)
{
  int error;

  ENTER(GFN_DRELSE_WRAPPER);

  error = buf_free_main(bh, file, line);
  if (error)
  {
    EXIT(GFN_DRELSE_WRAPPER);
    return;
  }

  gfs_drelse_arch(bh);

  EXIT(GFN_DRELSE_WRAPPER);
}


/**
 * gfs_dforget_wrapper - the wrapper around gfs_dforget()
 * @bh: the buffer
 * @file: the file the gfs_dforget() occurs on
 * @line: the line
 *
 * Function description
 *
 * Returns: what is returned
 */

void gfs_dforget_wrapper(osi_buf_t *bh, char *file, unsigned int line)
{
  int error;

  ENTER(GFN_DFORGET_WRAPPER);

  error = buf_free_main(bh, file, line);
  if (error)
  {
    EXIT(GFN_DFORGET_WRAPPER);
    return;
  }

  gfs_dforget_arch(bh);

  EXIT(GFN_DFORGET_WRAPPER);
}


/**
 * buf_check - Called at unload to print stats
 *
 */

void buf_check(void)
{
  buf_call_t *oc;
  unsigned int index;

  ENTER(GFN_BUF_CHECK);


  osi_spin_lock(&buf_lock);

  for (index = 0; index < BUF_HASH_SIZE; index++)
    while (!osi_list_empty(&buf_head[index]))
    {
      oc = osi_list_entry(buf_head[index].next, buf_call_t, list);
      osi_list_del(&oc->list);

      osi_spin_unlock(&buf_lock);


      con_printf("GFS:  unfreed buffer <%lX>, %"FMT_64"u, (%s, %u), %u\n", 
		 (unsigned long)oc->bh,
		 BH_BLKNO(oc->bh),
		 oc->file, oc->line, oc->count);

      gfree(oc, sizeof(buf_call_t));


      osi_spin_lock(&buf_lock);
    }

  osi_spin_unlock(&buf_lock);


  con_printf("GFS:  buf_ct:          %u\n", buf_stats.ct);
  con_printf("GFS:  buf_ct_max:      %u\n", buf_stats.ct_max);
  con_printf("GFS:  buf_ct_total:    %u\n", buf_stats.ct_total);
  con_printf("GFS:  refs:            %u\n", buf_stats.refs);
  con_printf("GFS:  refs_max:        %u\n", buf_stats.refs_max);
  con_printf("GFS:  refs_total:      %u\n", buf_stats.refs_total);


  EXIT(GFN_BUF_CHECK);
}





#endif  /*  DEBUG_BUF  */
