/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portions Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth Preslan <kpreslan@sistina.com>
** 
**    Module name:  dio.h
**
**    Description:
**      This module implements a bunch of routines that do I/O through the
**        buffer cache.
**
*******************************************************************************
******************************************************************************/

#ifndef __DIO_DOT_H__
#define __DIO_DOT_H__

#include "dio_arch.h"


/*  Arguments to gfs_dread()  */

#define DR_FORCE                (0x01000000)
#define DR_START                (0x02000000)
#define DR_WAIT                 (0x04000000)


/*  Arguments to gfs_dwrite()  */

#define DW_DIRTY                (0x00100000)
#define DW_START                (0x00200000)
#define DW_WAIT                 (0x00400000)


/*  Buffer accounting stuff

    the ct variables are changed each time we ask for
    a pointer to a buffer that we don't already have
    somewhere else. For

    bh = gfs_getblk(1024);
    bh1 = gfs_getblk(1024);

    ct = 1 and refs = 2
    */

#ifdef DEBUG_BUF

#define BUF_HASH_SIZE 4096
EXTERN osi_list_head_t buf_head[BUF_HASH_SIZE];

struct buf_stats
{
  unsigned int         ct; 
  unsigned int         ct_max;
  unsigned int         ct_total;
  unsigned int         refs;
  unsigned int         refs_max;
  unsigned int         refs_total;
};
typedef struct buf_stats buf_stats_t;

EXTERN buf_stats_t buf_stats;
EXTERN osi_spinlock_t buf_lock;

osi_buf_t *gfs_dgetblk_wrapper(gfs_sbd_t *sdp, uint64 blkno, 
			       gfs_glock_t *gl,
			       char *file, unsigned int line);
osi_buf_t *gfs_dread_wrapper(gfs_sbd_t *sdp, uint64 blkno, 
			     gfs_glock_t *gl, int flags,
			     char *file, unsigned int line);
void gfs_drelse_wrapper(osi_buf_t *bh, char *file, unsigned int line);
void gfs_dforget_wrapper(osi_buf_t *bh, char *file, unsigned int line);
int buf_alloc_main(osi_buf_t *bh, char *file, unsigned int line);
void buf_check(void);

#define gfs_dgetblk(sdp, blkno, gl) gfs_dgetblk_wrapper(sdp, blkno, gl, __FILE__, __LINE__)
#define gfs_dread(sdp, blkno, gl, flags) gfs_dread_wrapper(sdp, blkno, gl, flags, __FILE__, __LINE__)
#define gfs_drelse(bh) gfs_drelse_wrapper(bh, __FILE__, __LINE__)
#define gfs_dforget(bh) gfs_dforget_wrapper(bh, __FILE__, __LINE__)

#else /*  DEBUG_BUF  */

#define buf_alloc_main(bh, file, line)
#define buf_check()

#define gfs_dgetblk(sdp, blkno, gl) gfs_dgetblk_arch(sdp, blkno, gl)
#define gfs_dread(sdp, blkno, gl, flags) gfs_dread_arch(sdp, blkno, gl, flags)
#define gfs_drelse(bh) gfs_drelse_arch(bh)
#define gfs_dforget(bh) gfs_dforget_arch(bh)

#endif  /*  DEBUG_BUF  */



/*  Asynchronous I/O Routines  */

osi_buf_t *gfs_dgetblk_arch(gfs_sbd_t *sdp, uint64 blkno, gfs_glock_t *gl);
osi_buf_t *gfs_dread_arch(gfs_sbd_t *sdp, uint64 blkno, gfs_glock_t *gl, int flags);
void gfs_drelse_arch(osi_buf_t *bh);
void gfs_dforget_arch(osi_buf_t *bh);

void gfs_prep_new_buffer(osi_buf_t *bh);
int gfs_dreread(gfs_sbd_t *sdp, osi_buf_t *bh, int flags);
int gfs_dwrite(gfs_sbd_t *sdp, osi_buf_t *bh, int flags);

void gfs_dpin(osi_buf_t *bh);
void gfs_dunpin(osi_buf_t *bh, gfs_trans_t *tr);
int gfs_is_pinned(osi_buf_t *bh);

#define gfs_lock_buffer(bh) osi_sema_lock(&(BH_BUFDATA((bh)))->bd_lock)
#define gfs_unlock_buffer(bh) osi_sema_unlock(&(BH_BUFDATA((bh)))->bd_lock)

osi_buf_t *gfs_init_temp_bh(gfs_sbd_t *sdp, uint64 blocknr, char *data, int write);
void gfs_dispose_temp_bh(osi_buf_t *bh);



/*  Buffer Caching routines  */

osi_buf_t *gfs_get_meta_buffer(gfs_inode_t *ip, int height, uint64 num, int new);
osi_buf_t *gfs_get_data_buffer(gfs_inode_t *ip, uint64 block, int new);

static __inline__ void gfs_flush_meta_cache(gfs_inode_t *ip, int start)
{
  int i;

  for (i = start; i < GFS_MAX_META_HEIGHT; i++)
  {
    if (!ip->i_cache[i])
      break;
    gfs_drelse(ip->i_cache[i]);
    ip->i_cache[i] = NULL;
  }
}

static __inline__ osi_buf_t *gfs_get_inode_buffer(gfs_inode_t *ip)
{
  return gfs_get_meta_buffer(ip, 0, ip->i_num.no_addr, 0);
}



/*  Buffer Content Functions  */

/**
 * gfs_buffer_clear - Zeros out a buffer
 * @ip: The GFS inode
 * @bh: The buffer to zero
 *
 */

static __inline__ void gfs_buffer_clear(osi_buf_t *bh)
{
  gfs_memset(BH_DATA(bh), 0, BH_SIZE(bh));
}


/**
 * gfs_buffer_clear_tail - Clear buffer beyond the dinode
 * @bh: The buffer containing the on-disk inode
 * @head: the size of the head of the buffer
 *
 * Clears the remaining part of an on-disk inode that is not a dinode.
 * i.e. The data part of a stuffed inode, or the top level of metadata
 * of a non-stuffed inode.
 */

static __inline__ void gfs_buffer_clear_tail(osi_buf_t *bh, int head)
{
  gfs_memset(BH_DATA(bh) + head, 0, BH_SIZE(bh) - head);
}


/**
 * gfs_buffer_clear_ends - Zero out any bits of a buffer which are not being written
 * @bh: The buffer
 * @offset: Offset in buffer where write starts
 * @amount: Amount of data being written
 * @journaled: TRUE if this is a journaled buffer
 *
 */

static __inline__ void gfs_buffer_clear_ends(osi_buf_t *bh, int offset, int amount, int journaled)
{
  int z_off1 = (journaled) ? sizeof(gfs_jdata_t) : 0;
  int z_len1 = offset - z_off1;
  int z_off2 = offset + amount;
  int z_len2 = BH_SIZE(bh) - z_off2;

  if (z_len1)
    gfs_memset(BH_DATA(bh) + z_off1, 0, z_len1);

  if (z_len2)
    gfs_memset(BH_DATA(bh) + z_off2, 0, z_len2);
}


/**
 * gfs_buffer_copy_tail - copies the tail of one buffer to another
 * @to_bh: the buffer to copy to
 * @to_head: the size of the head of to_bh
 * @from_bh: the buffer to copy from
 * @from_head: the size of the head of from_bh
 *
 * from_head is guaranteed to bigger than to_head 
 */

static __inline__ void gfs_buffer_copy_tail(osi_buf_t *to_bh, int to_head,
					    osi_buf_t *from_bh, int from_head)
{
  gfs_memcpy(BH_DATA(to_bh) + to_head,
	     BH_DATA(from_bh) + from_head,
	     BH_SIZE(from_bh) - from_head);
  gfs_memset(BH_DATA(to_bh) + BH_SIZE(to_bh) + to_head - from_head,
	     0,
	     from_head - to_head);
}


/**
 * gfs_buffer_print - print a buffer to the debug console
 * @bh: the buffer
 * @string:  what to print before the contents of the buffer
 *
 */

static __inline__ void gfs_buffer_print(osi_buf_t *bh, char *string)
{
#ifdef DEBUG_PRINT
  unsigned int x, size = BH_SIZE(bh);
  unsigned char *c = BH_DATA(bh);

  db_printf("%s\n", string);

  for (x = 0; x < size; x++)
  {
    db_printf("%.2X ", c[x]);
    if (x % 16 == 15)
      db_printf("\n");
  }
  if (x % 16 != 0)
    db_printf("\n");
#endif  /*  DEBUG_PRINT  */
}



#endif  /*  __DIO_DOT_H__  */
