/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portions Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Seth Van Oort <vano0023@tc.umn.edu>
** 
**    Module name: dir.c
**
**    Description:
**	This module defines all GFS routines that have to do with directories.
**
**      Implements Extendible Hashing as described in:
**        "Extendible Hashing" by Fagin, et al in
**           __ACM Trans. on Database Systems__, Sept 1979.
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "blklist.h"
#include "bmap.h"
#include "dir.h"
#include "dio.h"
#include "file.h"
#include "fs.h"
#include "glock.h"
#include "inode.h"
#include "log.h"
#include "trans.h"



#define IS_LEAF     (1)
#define IS_DINODE   (2)

/*
Here's the layout of dirents which is essentially the same as that of ext2 
within a single block. The field de_name_len is the number of bytes
actually required for the name (no null terminator). The field de_rec_len
is the number of bytes allocated to the dirent. The offset of the next
dirent in the block is (dirent + dirent->de_rec_len). When a dirent is
deleted, the preceding dirent inherits its allocated space, ie
prev->de_rec_len += deleted->de_rec_len. Since the next dirent is obtained
by adding de_rec_len to the current dirent, this essentially causes the
deleted dirent to get jumped over when iterating through all the dirents.
When deleting the first dirent in a block, there is no previous dirent so
the field de_ino is set to zero to designate it as deleted. When allocating
a dirent, gfs_dirent_alloc iterates through the dirents in a block. If the
first dirent has (de_ino == 0) and de_rec_len is large enough, this first
dirent is allocated. Otherwise it must go through all the 'used' dirents
searching for one in which the amount of total space minus the amount of
used space will provide enough space for the new dirent.
There are two types of blocks in which dirents reside. In a stuffed dinode,
the dirents begin at offset sizeof(gfs_dinode_t) from the beginning of the block.
In leaves, they begin at offset sizeof (gfs_leaf_t) from the beginning of the
leaf block. The dirents reside in leaves when 

dip->i_di.di_regime == GFS_DIR_EXHASH.

The dirents are in the stuffed dinode when dip->i_di.di_regime == GFS_DIR_LINEAR.
When the dirents are in leaves, the actual contents of the directory file are
used as an array of 64-bit block pointers pointing to the leaf blocks. The
dirents are NOT in the directory file itself. There can be more than one block
pointer in the array that points to the same leaf. In fact, when a directory is
first converted from linear to exhash, all of the pointers point to the same
leaf. When a leaf is completely full, the size of the hash table can be doubled
unless it is already at the maximum size which is hard coded into 
GFS_DIR_MAX_DEPTH. After that, leaves are chained together in a linked list but
never before the maximum hash table size has been reached.
*/


/**
 * int gfs_filecmp - Compare two filenames
 * @file1: The first filename
 * @file2: The second filename
 * @len_of_file2: The length of the second file
 *
 * This routine compares two filenames and returns TRUE if they are equal.
 *
 * Returns: TRUE (!=0) if the files are the same, otherwise FALSE (0).
 */

int gfs_filecmp(osi_qstr_t *file1, char *file2, int len_of_file2)
{
  ENTER(GFN_FILECMP);


  if (file1->len != len_of_file2)
  {
    EXIT(GFN_FILECMP);
    return FALSE;
  }

  if (osi_memcmp(file1->name, file2, file1->len))
  {
    EXIT(GFN_FILECMP);
    return FALSE;
  }


  EXIT(GFN_FILECMP);

  return TRUE;
}


/**
 * dir_hash - Compute hash on a filename
 * @filename: The name to compute the hash upon
 *
 * Return: The hash value
 */

STATIC __inline__ uint32 dir_hash(osi_qstr_t *filename)
{
  return gfs_hash(filename->name, filename->len);
}


/**
 * gfs_dirent_first - Return the first dirent
 * @bh: The buffer
 * @dent: Pointer to list of dirents
 *
 * return first dirent whether bh points to leaf or stuffed dinode
 *
 * Returns: IS_LEAF or IS_DINODE
 */

STATIC int gfs_dirent_first(osi_buf_t *bh, gfs_dirent_t **dent)
{
  gfs_leaf_t *leaf;
  gfs_dinode_t *dinode;

  ENTER(GFN_DIRENT_FIRST);


  leaf = (gfs_leaf_t *)BH_DATA(bh);

  if (gfs32_to_cpu(leaf->lf_header.mh_type) == GFS_METATYPE_LF)
  {
    *dent = (gfs_dirent_t *)(BH_DATA(bh) + sizeof(gfs_leaf_t));

    EXIT(GFN_DIRENT_FIRST);
    return IS_LEAF;
  }
  else
  {
    dinode = (gfs_dinode_t *)BH_DATA(bh);
    GFS_ASSERT(gfs32_to_cpu(dinode->di_header.mh_type) == GFS_METATYPE_DI, );

    *dent = (gfs_dirent_t *)(BH_DATA(bh) + sizeof(gfs_dinode_t));

    EXIT(GFN_DIRENT_FIRST);
    return IS_DINODE;
  }
}


/**
 * gfs_dirent_next - Next dirent
 * @bh: The buffer
 * @dent: Pointer to list of dirents
 *
 * Returns: 0 on success, error code otherwise
 */

STATIC int gfs_dirent_next(osi_buf_t *bh, gfs_dirent_t **dent)
{
  gfs_dirent_t *tmp, *cur;
  char *bh_end;
  uint32 cur_rec_len;

  ENTER(GFN_DIRENT_NEXT);


  cur = *dent;
  bh_end = BH_DATA(bh) + BH_SIZE(bh);

  cur_rec_len = gfs16_to_cpu(cur->de_rec_len);

  if ((char *)cur + cur_rec_len >= bh_end)
  {
    GFS_ASSERT((char *)cur + cur_rec_len == bh_end, );

    EXIT(GFN_DIRENT_NEXT);
    return -ENOENT;
  }

  tmp = (gfs_dirent_t *)((char *)cur + cur_rec_len);

  GFS_ASSERT((char *)tmp + gfs16_to_cpu(tmp->de_rec_len) <= bh_end, );
  GFS_ASSERT(tmp->de_inum.no_formal_ino, );  /*  only the first dent could ever have de_ino == 0  */

  *dent = tmp;


  EXIT(GFN_DIRENT_NEXT);

  return 0;
}


/**
 * gfs_dirent_del - Delete a dirent
 * @dip: The GFS inode
 * @bh: The buffer
 * @prev: The previous dirent
 * @cur: The current dirent
 *
 * Returns: 0 on success, error code otherwise
 */

STATIC int gfs_dirent_del(gfs_inode_t *dip, osi_buf_t *bh, gfs_dirent_t *prev, gfs_dirent_t *cur)
{
  uint32 cur_rec_len, prev_rec_len;

  ENTER(GFN_DIRENT_DEL);

  GFS_ASSERT(cur->de_inum.no_formal_ino, );


  gfs_trans_add_bh(dip->i_sbd, bh);


  /*  If there is no prev entry, this is the first entry in the block.
      The de_rec_len is already as big as it needs to be.  Just zero
      out the inode number and return.  */

  if (!prev)
  {
    cur->de_inum.no_formal_ino = 0;  /*  No endianess worries  */
    EXIT(GFN_DIRENT_DEL);
    return 0;
  }


  /*  Combine this dentry with the previous one.  */

  prev_rec_len = gfs16_to_cpu(prev->de_rec_len);
  cur_rec_len = gfs16_to_cpu(cur->de_rec_len);

  GFS_ASSERT((char *)prev + prev_rec_len == (char *)cur, );
  GFS_ASSERT((char *)cur + cur_rec_len <= BH_DATA(bh) + BH_SIZE(bh), );

  prev_rec_len += cur_rec_len;
  prev->de_rec_len = cpu_to_gfs16(prev_rec_len);


  EXIT(GFN_DIRENT_DEL);

  return 0;
}


/**
 * gfs_dirent_alloc - Allocate a directory entry
 * @dip: The GFS inode
 * @bh: The buffer
 * @name_len: The length of the name
 * @dent_out: Pointer to list of dirents
 *
 * Returns: 0 on success, error code otherwise
 */

int gfs_dirent_alloc(gfs_inode_t *dip, osi_buf_t *bh, int name_len, gfs_dirent_t **dent_out)
{
  gfs_sbd_t *sdp = dip->i_sbd;
  gfs_dirent_t *dent, *new;
  gfs_leaf_t *leaf = NULL;
  gfs_dinode_t *dinode;
  int rec_len, type;
  int entries = 0, offset = 0, x = 0;

  ENTER(GFN_DIRENT_ALLOC);


  type = gfs_dirent_first(bh, &dent);

  if (type == IS_LEAF)
  {
    leaf = (gfs_leaf_t *)BH_DATA(bh);
    entries = gfs16_to_cpu(leaf->lf_entries);
    offset = sizeof(gfs_leaf_t);
  }
  else if (type == IS_DINODE)
  {
    dinode = (gfs_dinode_t *)BH_DATA(bh);
    entries = gfs32_to_cpu(dinode->di_entries);
    offset = sizeof(gfs_dinode_t);
  }
  else
  {
    GFS_ASSERT(0, );
  }


  rec_len = GFS_DIR_REC_LEN(name_len);


  if (!entries)
  {
    gfs_trans_add_bh(sdp, bh);

    dent->de_rec_len = BH_SIZE(bh) - offset;
    dent->de_rec_len = cpu_to_gfs16(dent->de_rec_len);
    dent->de_name_len = cpu_to_gfs16(name_len);

    *dent_out = dent;
    goto success;
  }


  do
  {
    uint32 cur_rec_len, cur_name_len;

    cur_rec_len = gfs16_to_cpu(dent->de_rec_len);
    cur_name_len = gfs16_to_cpu(dent->de_name_len);

    if ((!dent->de_inum.no_formal_ino && cur_rec_len >= rec_len) ||
	(cur_rec_len >= GFS_DIR_REC_LEN(cur_name_len) + rec_len))
    {
      gfs_trans_add_bh(sdp, bh);

      if (dent->de_inum.no_formal_ino)
      {
	new = (gfs_dirent_t *)((char *)dent + GFS_DIR_REC_LEN(cur_name_len));
	new->de_rec_len = cpu_to_gfs16(cur_rec_len - GFS_DIR_REC_LEN(cur_name_len));
	new->de_name_len = cpu_to_gfs16(name_len);

	dent->de_rec_len = cur_rec_len - gfs16_to_cpu(new->de_rec_len);
	dent->de_rec_len = cpu_to_gfs16(dent->de_rec_len);

	*dent_out = new;
	goto success;
      }

      dent->de_name_len = cpu_to_gfs16(name_len);

      *dent_out = dent;
      goto success;
    }

    GFS_ASSERT(x < entries, );

    if (dent->de_inum.no_formal_ino)
      x++;
  }
  while (gfs_dirent_next(bh, &dent) == 0);


  EXIT(GFN_DIRENT_ALLOC);

  return -ENOSPC;



 success:
  EXIT(GFN_DIRENT_ALLOC);
  return 0;
}


/**
 * leaf_search
 * @bh:
 * @filename:
 * @dent_out:
 * @dent_prev:
 *
 * Returns:
 */

STATIC int leaf_search(osi_buf_t *bh, osi_qstr_t *filename,
		       gfs_dirent_t **dent_out, gfs_dirent_t **dent_prev)
{
  uint32 hash;
  gfs_dirent_t *dent, *prev = NULL;
  int type, entries = 0, x = 0;

  ENTER(GFN_LEAF_SEARCH);


  type = gfs_dirent_first(bh, &dent);

  if (type == IS_LEAF)
  {
    gfs_leaf_t *leaf = (gfs_leaf_t *)BH_DATA(bh);
    entries = gfs16_to_cpu(leaf->lf_entries);
  }
  else if (type == IS_DINODE)
  {
    gfs_dinode_t *dinode = (gfs_dinode_t *)(BH_DATA(bh));
    entries = gfs32_to_cpu(dinode->di_entries);
  }
  else
  {
    GFS_ASSERT(0, );
  }


  hash = dir_hash(filename);

  do
  {
    if (!dent->de_inum.no_formal_ino)
    {
      prev = dent;
      continue;
    }

    if (gfs32_to_cpu(dent->de_hash) == hash &&
	gfs_filecmp(filename, (char *)(dent + 1), gfs16_to_cpu(dent->de_name_len)))
    {
      *dent_out = dent;
      if (dent_prev)
	*dent_prev = prev;

      EXIT(GFN_LEAF_SEARCH);
      return 0;
    }

    GFS_ASSERT(x < entries, );
    x++;
    prev = dent;
  }
  while (gfs_dirent_next(bh, &dent) == 0);


  EXIT(GFN_LEAF_SEARCH);

  return -ENOENT;
}


/**
 * get_leaf - Get leaf
 * @dip:
 * @leaf_no:
 * @bh_out:
 *
 * Returns: 0 on success, error code otherwise
 */

STATIC int get_leaf(gfs_inode_t *dip, uint64 leaf_no, osi_buf_t **bh_out)
{
  gfs_sbd_t *sdp = dip->i_sbd;
  osi_buf_t *bh;

  ENTER(GFN_GET_LEAF);


  bh = gfs_dread(sdp, leaf_no, dip->i_gl, DR_START | DR_WAIT);
  if (!bh)
  {
    EXIT(GFN_GET_LEAF);
    return -EIO;
  }
  gfs_metatype_check(bh, GFS_METATYPE_LF);

  *bh_out = bh;


  EXIT(GFN_GET_LEAF);

  return 0;
}


/**
 * get_leaf_nr - Get a leaf number associated with the index
 * @dip: The GFS inode
 * @index:
 * @leaf_out:
 *
 * Returns: 0 on success, error code otherwise
 */

STATIC int get_leaf_nr(gfs_inode_t *dip, uint32 index, uint64 *leaf_out)
{
  uint64 leaf_no;
  int64 error;

  ENTER(GFN_GET_LEAF_NR);


  error = gfs_internal_read(dip, (char *)&leaf_no, index * sizeof(uint64), sizeof(uint64));
  if (error != sizeof(uint64))
  {
    EXIT(GFN_GET_LEAF_NR);
    return (error < 0) ? error : -EIO;
  }

  *leaf_out = gfs64_to_cpu(leaf_no);


  EXIT(GFN_GET_LEAF_NR);

  return 0;
}


/**
 * get_first_leaf - Get first leaf
 * @dip: The GFS inode
 * @index:
 * @bh_out:
 *
 * Returns: 0 on success, error code otherwise
 */

STATIC int get_first_leaf(gfs_inode_t *dip, uint32 index, osi_buf_t **bh_out)
{
  uint64 leaf_no;
  int64 error;

  ENTER(GFN_GET_FIRST_LEAF);


  error = get_leaf_nr(dip, index, &leaf_no);
  if (!error)
    error = get_leaf(dip, leaf_no, bh_out);


  EXIT(GFN_GET_FIRST_LEAF);

  return error;
}


/**
 * get_next_leaf - Get next leaf
 * @dip: The GFS inode
 * @bh_in: The buffer
 * @bh_out:
 *
 * Returns: 0 on success, error code otherwise
 */

STATIC int get_next_leaf(gfs_inode_t *dip, osi_buf_t *bh_in, osi_buf_t **bh_out)
{
  gfs_leaf_t *leaf;
  int error;

  ENTER(GFN_GET_NEXT_LEAF);


  leaf = (gfs_leaf_t *)BH_DATA(bh_in);

  if (!leaf->lf_next)
    error = -ENOENT;
  else
    error = get_leaf(dip, gfs64_to_cpu(leaf->lf_next), bh_out);


  EXIT(GFN_GET_NEXT_LEAF);

  return error;
}


/**
 * linked_leaf_search - Linked leaf search
 * @dip: The GFS inode
 * @filename: The filename to search for
 * @dent_out:
 * @dent_prev:
 * @bh_out:
 *
 * Returns: 0 on sucess, error code otherwise
 */

STATIC int linked_leaf_search(gfs_inode_t *dip, osi_qstr_t *filename,
			      gfs_dirent_t **dent_out, gfs_dirent_t **dent_prev,
			      osi_buf_t **bh_out)
{
  osi_buf_t *bh = NULL, *bh_next;
  uint32 hsize, index;
  uint32 hash;
  int64 error;

  ENTER(GFN_LINKED_LEAF_SEARCH);


  hsize = 1 << dip->i_di.di_depth;
  GFS_ASSERT(hsize * sizeof(uint64) == dip->i_di.di_size, );

  /*  Figure out the address of the leaf node.  */

  hash = dir_hash(filename);
  index = hash >> (32 - dip->i_di.di_depth);
  GFS_ASSERT(index < hsize, );


  error = get_first_leaf(dip, index, &bh_next);
  if (error)
  {
    EXIT(GFN_LINKED_LEAF_SEARCH);
    return error;
  }


  /*  Find the entry  */

  do
  {
    if (bh)
      gfs_drelse(bh);

    bh = bh_next;

    error = leaf_search(bh, filename, dent_out, dent_prev);
    if (error)
    {
      if (error != -ENOENT)
      {
	gfs_drelse(bh);
	EXIT(GFN_LINKED_LEAF_SEARCH);
	return error;
      }
    }
    else
    {
      *bh_out = bh;
      EXIT(GFN_LINKED_LEAF_SEARCH);
      return 0;
    }

    error = get_next_leaf(dip, bh, &bh_next);
  }
  while (!error);

  gfs_drelse(bh);


  EXIT(GFN_LINKED_LEAF_SEARCH);

  return error;
}


/**
 * dir_make_exhash - Convet a stuffed directory into an ExHash directory
 * @dip: The GFS inode
 *
 * Returns: 0 on success, error code otherwise
 */

STATIC int dir_make_exhash(gfs_inode_t *dip)
{
  gfs_sbd_t *sdp = dip->i_sbd;
  gfs_dirent_t *dent;
  osi_buf_t *bh, *dibh;
  gfs_leaf_t *leaf;
  int y;
  uint32 x;
  uint64 *lp, bn;
  int error;

  ENTER(GFN_DIR_MAKE_EXHASH);


  /*  Sanity checks  */

  GFS_ASSERT(sizeof(gfs_leaf_t) <= sizeof(gfs_dinode_t), );
  GFS_ASSERT(!(dip->i_di.di_flags & GFS_DIF_EXHASH), );
  GFS_ASSERT(gfs_is_stuffed(dip), );


  dibh = gfs_get_inode_buffer(dip);
  if (!dibh)
  {
    error = -EIO;
    goto fail;
  }


  /*  Allocate a new block for the first leaf node  */

  error = gfs_metaalloc(dip, &bn);
  if (error)
    goto fail_drelse;


  /*  Turn over a new leaf  */

  bh = gfs_dread(sdp, bn, dip->i_gl, DR_FORCE | DR_START | DR_WAIT);
  if (!bh)
  {
    error = -EIO;
    goto fail_drelse;
  }

  gfs_trans_add_bh(sdp, bh);
  gfs_metatype_set(bh, GFS_METATYPE_LF, GFS_FORMAT_LF);
  gfs_buffer_clear_tail(bh, sizeof(gfs_meta_header_t));


  /*  Fill in the leaf structure  */

  leaf = (gfs_leaf_t *)BH_DATA(bh);

  GFS_ASSERT(dip->i_di.di_entries < (1 << 16), );

  leaf->lf_dirent_format = cpu_to_gfs32(GFS_FORMAT_DE);
  leaf->lf_entries = cpu_to_gfs16(dip->i_di.di_entries);


  /*  Copy dirents  */

  gfs_buffer_copy_tail(bh, sizeof(gfs_leaf_t), dibh, sizeof(gfs_dinode_t));


  /*  Find last entry  */

  x = 0;
  gfs_dirent_first(bh, &dent);

  do
  {
    if (!dent->de_inum.no_formal_ino)
      continue;

    if (++x == dip->i_di.di_entries)
      break;
  }
  while (gfs_dirent_next(bh, &dent) == 0);


  /*  Adjust the last dirent's record length
      (Remember that dent still points to the last entry.)  */

  dent->de_rec_len = gfs16_to_cpu(dent->de_rec_len) + sizeof(gfs_dinode_t) - sizeof(gfs_leaf_t);
  dent->de_rec_len = cpu_to_gfs16(dent->de_rec_len);

  gfs_drelse(bh);


  /*  We're done with the new leaf block, now setup the new
      hash table.  */

  gfs_trans_add_bh(sdp, dibh);
  gfs_buffer_clear_tail(dibh, sizeof(gfs_dinode_t));

  lp = (uint64 *)(BH_DATA(dibh) + sizeof(gfs_dinode_t));

  for (x = sdp->sd_hash_ptrs; x--; lp++)
    *lp = cpu_to_gfs64(bn);

  dip->i_di.di_size = sdp->sd_sb.sb_bsize / 2;
  dip->i_di.di_blocks++;
  dip->i_di.di_flags |= GFS_DIF_EXHASH;
  dip->i_di.di_payload_format = 0;

  for (x = sdp->sd_hash_ptrs, y = -1; x; x >>= 1, y++) ;
  dip->i_di.di_depth = y;

  gfs_dinode_out(&dip->i_di, BH_DATA(dibh));

  gfs_drelse(dibh);


  EXIT(GFN_DIR_MAKE_EXHASH);

  return 0;



 fail_drelse:
  gfs_drelse(dibh);

 fail:
  EXIT(GFN_DIR_MAKE_EXHASH);
  return error;  
}


/**
 * dir_split_leaf - Split a leaf block into two
 * @dip: The GFS inode
 * @index:
 * @leaf_no:
 *
 * Returns: 0 on success, error code on failure
 */

STATIC int dir_split_leaf(gfs_inode_t *dip, uint32 index, uint64 leaf_no)
{
  gfs_sbd_t *sdp = dip->i_sbd;
  osi_buf_t *nbh, *obh, *dibh;
  gfs_leaf_t *nleaf, *oleaf;
  gfs_dirent_t *dent, *prev = NULL, *next = NULL, *new;
  uint32 hsize, start, len, half_len, divider;
  uint64 bn, *lp;
  uint32 name_len;
  int x, moved = FALSE;
  int64 error;

  ENTER(GFN_DIR_SPLIT_LEAF);


  /*  Sanity Checks  */

  GFS_ASSERT(dip->i_di.di_flags & GFS_DIF_EXHASH, );

  hsize = 1 << dip->i_di.di_depth;
  GFS_ASSERT(hsize * sizeof(uint64) == dip->i_di.di_size, );
  GFS_ASSERT(index < hsize, );


  /*  Allocate the new leaf block  */

  error = gfs_metaalloc(dip, &bn);
  if (error)
    goto fail;


  /*  Get the new leaf block  */

  nbh = gfs_dread(sdp, bn, dip->i_gl, DR_FORCE | DR_START | DR_WAIT);
  if (!nbh)
  {
    error = -EIO;
    goto fail;
  }
  
  gfs_trans_add_bh(sdp, nbh);
  gfs_metatype_set(nbh, GFS_METATYPE_LF, GFS_FORMAT_LF);
  gfs_buffer_clear_tail(nbh, sizeof(gfs_meta_header_t));

  nleaf = (gfs_leaf_t *)BH_DATA(nbh);

  nleaf->lf_dirent_format = cpu_to_gfs32(GFS_FORMAT_DE);


  /*  Get the old leaf block  */

  error = get_leaf(dip, leaf_no, &obh);
  if (error)
    goto fail_nrelse;

  gfs_trans_add_bh(sdp, obh);

  oleaf = (gfs_leaf_t *)BH_DATA(obh);


  /*  Compute the start and len of leaf pointers in the hash table.  */

  len = 1 << (dip->i_di.di_depth - gfs16_to_cpu(oleaf->lf_depth));
  GFS_ASSERT(len != 1, );
  half_len = len >> 1;

  start = (index & ~(len - 1));


  /*  Change the pointers.
      Don't bother distinguishing stuffed from non-stuffed.
      This code is complicated enough already.  */

  lp = (uint64 *)gmalloc(half_len * sizeof(uint64));
  if (!lp)
  {
    error = -ENOMEM;
    goto fail_orelse;
  }

  error = gfs_internal_read(dip, (char *)lp, start * sizeof(uint64), half_len * sizeof(uint64));
  if (error != half_len * sizeof(uint64))
  {
    if (error >= 0)
      error = -EIO;
    goto fail_lpfree;
  }

  /*  Change the pointers  */

  for (x = 0; x < half_len; x++)
    lp[x] = cpu_to_gfs64(bn);

  error = gfs_internal_write(dip, (char *)lp, start * sizeof(uint64), half_len * sizeof(uint64));
  if (error != half_len * sizeof(uint64))
  {
    if (error >= 0)
      error = -EIO;
    goto fail_lpfree;
  }

  gfree(lp, half_len * sizeof(uint64));


  /*  Compute the divider  */

  divider = (start + half_len) << (32 - dip->i_di.di_depth);

  IFDB(GFN_DIR_SPLIT_LEAF,
       db_printf("  Divider = 0x%.8X\n", divider););


  /*  Copy the entries  */

  gfs_dirent_first(obh, &dent);

  do
  {
    next = dent;
    if (gfs_dirent_next(obh, &next))
      next = NULL;

    if (dent->de_inum.no_formal_ino &&
	(gfs32_to_cpu(dent->de_hash) < divider))
    {
      name_len = gfs16_to_cpu(dent->de_name_len);

      error = gfs_dirent_alloc(dip, nbh, name_len, &new);
      GFS_ASSERT(!error, );

      new->de_inum = dent->de_inum;  /*  No endianness worries  */
      new->de_hash = dent->de_hash;  /*  No endianness worries  */
      new->de_type = dent->de_type;  /*  No endianness worries  */
      gfs_memcpy((char *)(new + 1), (char *)(dent + 1), name_len);

      nleaf->lf_entries = gfs16_to_cpu(nleaf->lf_entries) + 1;
      nleaf->lf_entries = cpu_to_gfs16(nleaf->lf_entries);

      gfs_dirent_del(dip, obh, prev, dent);

      GFS_ASSERT(gfs16_to_cpu(oleaf->lf_entries), );
      oleaf->lf_entries = gfs16_to_cpu(oleaf->lf_entries) - 1;
      oleaf->lf_entries = cpu_to_gfs16(oleaf->lf_entries);

      if (!prev)
	prev = dent;

      moved = TRUE;
    }
    else
      prev = dent;

    dent = next;
  }
  while (dent);


  /*  If none of the entries got moved into the new leaf,
      artificially fill in the first entry.  */

  if (!moved)
  {
    error = gfs_dirent_alloc(dip, nbh, 0, &new);
    GFS_ASSERT(!error, );

    new->de_inum.no_formal_ino = 0;
  }


  oleaf->lf_depth = gfs16_to_cpu(oleaf->lf_depth) + 1;
  oleaf->lf_depth = cpu_to_gfs16(oleaf->lf_depth);
  nleaf->lf_depth = oleaf->lf_depth;


  dibh = gfs_get_inode_buffer(dip);
  GFS_ASSERT(dibh, );  /*  Pinned in gfs_internal_write()  */

  dip->i_di.di_blocks++;

  gfs_trans_add_bh(sdp, dibh);
  gfs_dinode_out(&dip->i_di, BH_DATA(dibh));
  gfs_drelse(dibh);


  gfs_drelse(obh);
  gfs_drelse(nbh);


  EXIT(GFN_DIR_SPLIT_LEAF);

  return 0;



 fail_lpfree:
  gfree(lp, half_len * sizeof(uint64));

 fail_orelse:
  gfs_drelse(obh);

 fail_nrelse:
  gfs_drelse(nbh);

 fail:
  EXIT(GFN_DIR_SPLIT_LEAF);
  return error;
}


/**
 * dir_double_exhash - Double size of ExHash table
 * @dip: The GFS dinode
 *
 * Returns: 0 on success, error code on failure
 */

STATIC int dir_double_exhash(gfs_inode_t *dip)
{
  gfs_sbd_t *sdp = dip->i_sbd;
  osi_buf_t *dibh;

  uint32 hsize;
  uint64 *buf;
  uint64 *from, *to;
  uint64 block;

  int x;
  int64 error = 0;

  ENTER(GFN_DIR_DOUBLE_EXHASH);


  /*  Sanity Checks  */

  GFS_ASSERT(dip->i_di.di_flags & GFS_DIF_EXHASH, );
  hsize = 1 << dip->i_di.di_depth;
  GFS_ASSERT(hsize * sizeof(uint64) == dip->i_di.di_size, );


  /*  Allocate both the "from" and "to" buffers in one big chunk  */

  buf = (uint64 *)gmalloc(3 * sdp->sd_hash_bsize);
  if (!buf)
  {
    error = -ENOMEM;
    goto out;
  }

  for (block = dip->i_di.di_size / sdp->sd_hash_bsize; block--;)
  {
    error = gfs_internal_read(dip, (char *)buf, block * sdp->sd_hash_bsize, sdp->sd_hash_bsize);
    if (error != sdp->sd_hash_bsize)
    {
      if (error >= 0)
	error = -EIO;
      goto out_gfree;
    }

    from = buf;
    to = (uint64 *)((char *)buf + sdp->sd_hash_bsize);

    for (x = sdp->sd_hash_ptrs; x--; from++)
    {
      *to++ = *from;  /*  No endianess worries  */
      *to++ = *from;
    }

    error = gfs_internal_write(dip, (char *)buf + sdp->sd_hash_bsize, 
			       block * sdp->sd_sb.sb_bsize, sdp->sd_sb.sb_bsize);
    if (error != sdp->sd_sb.sb_bsize)
    {
      if (error >= 0)
	error = -EIO;
      goto out_gfree;
    }
  }

  gfree(buf, 3 * sdp->sd_hash_bsize);


  dibh = gfs_get_inode_buffer(dip);
  GFS_ASSERT(dibh, );  /*  Pinned in gfs_internal_write()  */

  dip->i_di.di_depth++;

  gfs_trans_add_bh(sdp, dibh);
  gfs_dinode_out(&dip->i_di, BH_DATA(dibh));
  gfs_drelse(dibh);


  EXIT(GFN_DIR_DOUBLE_EXHASH);

  return 0;


 out_gfree:
  gfree(buf, 3 * sdp->sd_hash_bsize);
  
 out:
  EXIT(GFN_DIR_DOUBLE_EXHASH);
  return error;
}


/**
 * compare_dents - compare directory entries by hash value
 * @dent1: first dent
 * @dent2: second dent
 *
 * When comparing the hash entries of dent1 to dent2:
 *   gt: returns 1
 *   lt: returns -1
 *   eq: returns 0
 */

static int compare_dents(const void *dent1, const void *dent2)
{
  gfs_dirent_t **foo1 = (gfs_dirent_t **) dent1;
  gfs_dirent_t **foo2 = (gfs_dirent_t **) dent2;

  int ret = 0;
  
  ENTER(GFN_COMPARE_DENTS);
  if(gfs32_to_cpu((*foo1)->de_hash) > gfs32_to_cpu((*foo2)->de_hash)) {
    ret = 1;
  }
  else if(gfs32_to_cpu((*foo1)->de_hash) < gfs32_to_cpu((*foo2)->de_hash)) {
    ret = -1;
  }
  
  EXIT(GFN_COMPARE_DENTS);
  return ret;
}


/**
 * find_dent_index - Finds a dent in a array of sorted dents
 * @sorted_dents: pre-sorted array of dents
 * @fpos: hash of the dent we're looking for
 * @dent_no: contains the total number of entries stored in the array when it 
 * comes into the function, variable to store the index of the dent in on exit
 * 
 * Finds a dent in the pre-sorted array based on the hash value passed in.  Stores
 * the index in the index pointer.  Currently is a simple linear search.  This
 * should maybe be converted to at least a binary search, but I don't know
 * how much of a difference it would really make.  I guess it might make a big
 * difference in *really* large directories that have linked leaves.
 *
 */

STATIC int find_dent_index(gfs_dirent_t **sorted_dents, int64 *f_pos, uint16 *dent_no)
{
  uint16 cur_index;

  ENTER(GFN_FIND_DENT_INDEX);

  for (cur_index = 0; cur_index < *dent_no; cur_index++)
  {
    if ((gfs32_to_cpu((sorted_dents[cur_index])->de_hash) >> 1) >= (uint32)*f_pos)
    { 
      *dent_no = cur_index;
      EXIT(GFN_FIND_DENT_INDEX);
      return 0;
    }
  }

  if (*f_pos == 0)
  {	  
    *dent_no = 0;
    EXIT(GFN_FIND_DENT_INDEX);
    return 0;
  }

  EXIT(GFN_FIND_DENT_INDEX);
  
  return -ENOENT;
}


/* CLEAN_UP_BH - helper macro for do_filldir
 * @type - Leaf or stuffed dinode?
 * @orig_leaf_no - The leaf number to start cleaning at
 * @dip - the dinode pointer we're working with
 *
 * Since we can't release the buffer heads til we're done with the sorted
 * sorted list of dentries, but we need to release them at every exit point
 * in do_filldir, I created this macro to do so.  It goes through all the
 * linked leaves and releases the buffer heads twice - once for the current
 * loop, and once for the loop where the dentries were stuffed into the
 * sorted list in the first place
 */

#define CLEAN_UP_BH(type, orig_leaf_no, dip)                                 \
  if (type == IS_LEAF)                                                       \
  {                                                                          \
    /* grab the leaf no that was passed into the fxn; */                     \
    /* this is the head leaf */                                              \
    uint64 lf_no = orig_leaf_no;                                             \
    do                                                                       \
    {                                                                        \
      int err;                                                               \
      gfs_leaf_t *lf = NULL;                                                 \
      osi_buf_t *lfbh = NULL;                                                \
      err = get_leaf(dip, lf_no, &lfbh);                                     \
      /* if we get an error, it means we found an invalid leaf */            \
      if (err)                                                               \
      {                                                                      \
	/* We can exit 'cause noone else got this far either */              \
	break;                                                               \
      }                                                                      \
      lf = (gfs_leaf_t *)BH_DATA(lfbh);                                      \
      lf_no = gfs64_to_cpu(lf->lf_next);                                     \
      /* release the buffer twice; once for this get_leaf, */                \
      /* and once for the one above */                                       \
      gfs_drelse(lfbh);                                                      \
      gfs_drelse(lfbh);                                                      \
    }                                                                        \
    while (lf_no);                                                           \
  }


/**
 * do_filldir - Fill directory
 * @dip: The GFS inode
 * @f_pos:
 * @dirent:
 * @filldir: the buffer to fill with entries
 * @bh:  The disk order buffer
 * @index: Used to verify current dirent belongs in current leaf
 * @leaf_no: 
 *
 * Does the filldir given a leaf or stuffed dinode
 * seven arguments stink. Ideas appreciated. (put in structure ?) 
 * 
 * New strategy:
 *  o get the leaf or stuffed dinode (or leaves if there are linked leaves)
 *  o sort the entries by their hash
 *  o find the last entry read if there is one, if not start at the first in sorted order
 *  o send them to filldir in order starting with the entry found in step 3
 *
 * Returns: 0 on success, error code on failure
 */

STATIC int do_filldir(gfs_inode_t *dip, int64 *f_pos,
		      void *dirent, gfs_filldir_t filldir,
		      osi_buf_t *bh, int index, uint64 leaf_no)
{
  gfs_dirent_t *dent;
  gfs_leaf_t *leaf = NULL;
  int type, len, error;
  gfs_dirent_t **sorted_dents; /* Entries from current dent are sorted and 
				  run throught the filldir fxn */
  uint16 dent_no, cur_dent_no;
  uint32 entries = 0;
  /* variables used to calculate things for the linked leaves */
  osi_buf_t *leafbh;      
  uint64 orig_leaf_no = leaf_no;
	  
  ENTER(GFN_DO_FILLDIR);


  type = gfs_dirent_first(bh, &dent);
 
  /* we have to find all entries in the leaf and any leaves attached to the first
   * leaf in a linked list */
  if (type == IS_LEAF)
  {
    do    
    {
      error = get_leaf(dip, leaf_no, &leafbh);
      if (error)
      {
        EXIT(GFN_DO_FILLDIR);
        return error;
      }

      leaf = (gfs_leaf_t *)BH_DATA(leafbh);
      entries += gfs16_to_cpu(leaf->lf_entries);
      leaf_no = gfs64_to_cpu(leaf->lf_next);
      gfs_drelse(leafbh);
    } 
    while (leaf_no);
  }
  else
  {
    /* stuffed dinodes are easy */
    entries = dip->i_di.di_entries;
  } 

  
  /* if there aren't any entries, don't bother doing anything with the dent. */ 
  if (!entries)
  {
    EXIT(GFN_DO_FILLDIR);
    return 0;
  }

  
  sorted_dents = (gfs_dirent_t **)gmalloc(sizeof(gfs_dirent_t *) * entries);
  if (!sorted_dents) 
  {
    EXIT(GFN_DO_FILLDIR);
    return -ENOMEM;
  }
  /* Make sure we're starting from the beginning */
  dent_no = 0;


  /* Get all dent hash entries and stuff them somewhere for later
   * sorting and processing */
  if (type == IS_LEAF)
  {
    /* grab the leaf no that was passed into the fxn, this is the head leaf */
    leaf_no = orig_leaf_no;
    
    do
    {
      error = get_leaf(dip, leaf_no, &leafbh);
      /* if we get an error, it means we didn't find the right leaf */
      if (error)
      {
	gfree(sorted_dents, sizeof(gfs_dirent_t *) * entries);
        CLEAN_UP_BH(type, orig_leaf_no, dip);
	EXIT(GFN_DO_FILLDIR);
	return error;
      }
      type = gfs_dirent_first(leafbh, &dent);
      leaf = (gfs_leaf_t *)BH_DATA(leafbh);
      
      /* if there aren't any dents, don't bother with this leaf */
      if (!gfs16_to_cpu(leaf->lf_entries))
      {
        leaf_no = gfs64_to_cpu(leaf->lf_next);
	/* don't need to release this now, 'cause I'll do it later */
	/*gfs_drelse(leafbh);*/
        continue;
      }
      
      /* put the dents into the list */ 
      do
      {
        if (!dent->de_inum.no_formal_ino)
          continue;

        /* sanity checking */
        len = 1 << (dip->i_di.di_depth - gfs16_to_cpu(leaf->lf_depth));
        GFS_ASSERT((gfs32_to_cpu(dent->de_hash) >> (32 - dip->i_di.di_depth)) < (index + len), );

        sorted_dents[dent_no] = dent;
        dent_no++;	 
      }  
      while (gfs_dirent_next(leafbh, &dent) == 0);

      leaf_no = gfs64_to_cpu(leaf->lf_next);
      /* can't release the buffer now, 'cause I'm still referencing the 
       * dentry inside it - See CLEAN_UP_BH macro for more info on this
       */
    }
    while (leaf_no);
  }
  else
  {
    /* put the stuffed dinode entries into an array */
    do
    {
      if (!dent->de_inum.no_formal_ino)
        continue;

      sorted_dents[dent_no] = dent;
      dent_no++;
    }
    while (gfs_dirent_next(bh, &dent) == 0);
  }
 
  /* Sort hash entries */
  gfs_sort(sorted_dents, (dent_no - 1), sizeof(gfs_dirent_t*), compare_dents); 
 
  cur_dent_no = dent_no;
  /* Grab the current dent_no based on fpos */
  error = find_dent_index(sorted_dents, f_pos, &cur_dent_no);
  if (error)
  {
    /* If there is an error returned by find_dent_index, no matching entry was found */
    gfree(sorted_dents, sizeof(gfs_dirent_t *) * entries);
    CLEAN_UP_BH(type, orig_leaf_no, dip);
    EXIT(GFN_DO_FILLDIR);
    return 0;
  }
  else
  {
    /* Run the dents through the filldir fxn */
    for (; cur_dent_no < dent_no; cur_dent_no++) 
    {
      /* make sure the latest dent hash value is stored */
      *f_pos = (uint64)(gfs32_to_cpu((sorted_dents[cur_dent_no])->de_hash) >> 1);
     
      /* get the current dent */
      dent = sorted_dents[cur_dent_no];

      error = filldir(dirent, (char *)(dent + 1), gfs16_to_cpu(dent->de_name_len),
		      *f_pos, gfs64_to_cpu(dent->de_inum.no_formal_ino), gfs16_to_cpu(dent->de_type));
      /* If there is an error, the filldir buffer is full, and we need to go back to
       * userspace to let the calling program handle the buffer */
      if (error < 0)
      {	
	error = -EINVAL;
	gfree(sorted_dents, sizeof(gfs_dirent_t *) * entries);
	CLEAN_UP_BH(type, orig_leaf_no, dip);
	EXIT(GFN_DO_FILLDIR);
	return error;
      }
    }
  } 

  
  /* Increment the *f_pos by one, so the next time we come into the do_filldir fxn, 
   * we get the next entry instead of the last one in the current leaf */
  (*f_pos)++; 
  gfree(sorted_dents, sizeof(gfs_dirent_t *) * entries);
  CLEAN_UP_BH(type, orig_leaf_no, dip);

  EXIT(GFN_DO_FILLDIR);

  return 0;
}


/**
 * dir_e_search -
 * @dip: The GFS inode
 * @filename:
 * @inode:
 *
 * Returns:
 */

STATIC int dir_e_search(gfs_inode_t *dip, osi_qstr_t *filename,
			gfs_inum_t *inum, unsigned int *type)
{
  osi_buf_t *bh;
  gfs_dirent_t *dent;
  int64 error;

  ENTER(GFN_DIR_E_SEARCH);


  /*  Sanity Checks  */

  GFS_ASSERT(dip->i_di.di_flags & GFS_DIF_EXHASH, );


  error = linked_leaf_search(dip, filename, &dent, NULL, &bh);
  if (error)
  {
    EXIT(GFN_DIR_E_SEARCH);
    return error;
  }

  if (inum)
    gfs_inum_in(inum, (char *)&dent->de_inum);
  if (type)
    *type = gfs16_to_cpu(dent->de_type);

  gfs_drelse(bh);


  EXIT(GFN_DIR_E_SEARCH);

  return 0;
}


/**
 * dir_e_add -
 * @dip: The GFS inode
 * @filename:
 * @inode:
 * @type:
 *
 */

STATIC int dir_e_add(gfs_inode_t *dip, osi_qstr_t *filename,
		     gfs_inum_t *inum, unsigned int type)
{
  gfs_sbd_t *sdp = dip->i_sbd;
  osi_buf_t *bh, *nbh, *dibh;
  gfs_leaf_t *leaf, *nleaf;
  gfs_dirent_t *dent;
  uint32 hsize, index;
  uint32 hash;
  uint64 leaf_no, bn;
  int64 error;

  ENTER(GFN_DIR_E_ADD);


 restart:

  /*  Sanity Checks  */

  GFS_ASSERT(dip->i_di.di_flags & GFS_DIF_EXHASH, );
  hsize = 1 << dip->i_di.di_depth;
  GFS_ASSERT(hsize * sizeof(uint64) == dip->i_di.di_size, );


  /*  Figure out the address of the leaf node.  */

  hash = dir_hash(filename);
  index = hash >> (32 - dip->i_di.di_depth);
  GFS_ASSERT(index < hsize, );


  error = get_leaf_nr(dip, index, &leaf_no);
  if (error)
  {
    EXIT(GFN_DIR_E_ADD);
    return error;
  }


  /*  Add entry to the leaf  */

  while (TRUE)
  {
    error = get_leaf(dip, leaf_no, &bh);
    if (error)
    {
      EXIT(GFN_DIR_E_ADD);
      return error;
    }

    leaf = (gfs_leaf_t *)BH_DATA(bh);


    if (gfs_dirent_alloc(dip, bh, filename->len, &dent))
    {
      if (gfs16_to_cpu(leaf->lf_depth) < dip->i_di.di_depth)
      {							   /*  Can we split the leaf?  */
	gfs_drelse(bh);

	error = dir_split_leaf(dip, index, leaf_no);
	if (error)
	{
	  EXIT(GFN_DIR_E_ADD);
	  return error;
	}

	goto restart;
      }
      else if (dip->i_di.di_depth < GFS_DIR_MAX_DEPTH)
      {							   /*  Can we double the hash table?  */
	gfs_drelse(bh);

	error = dir_double_exhash(dip);
	if (error)
	{
	  EXIT(GFN_DIR_E_ADD);
	  return error;
	}

	goto restart;
      }
      else if (leaf->lf_next)
      {							   /*  Can we try the next leaf in the list?  */
	leaf_no = gfs64_to_cpu(leaf->lf_next);
	gfs_drelse(bh);
	continue;
      }
      else
      {							   /*  Create a new leaf and add it to the list.  */
	error = gfs_metaalloc(dip, &bn);
	if (error)
	{
	  gfs_drelse(bh);
	  EXIT(GFN_DIR_E_ADD);
	  return error;
	}

	nbh = gfs_dread(sdp, bn, dip->i_gl, DR_FORCE | DR_START | DR_WAIT);
	if (!nbh)
	{
	  gfs_drelse(bh);
	  EXIT(GFN_DIR_E_ADD);
	  return -EIO;
	}

	gfs_trans_add_bh(sdp, nbh);
	gfs_metatype_set(nbh, GFS_METATYPE_LF, GFS_FORMAT_LF);
	gfs_buffer_clear_tail(nbh, sizeof(gfs_meta_header_t));

	gfs_trans_add_bh(sdp, bh);
	leaf->lf_next = cpu_to_gfs64(bn);

	nleaf = (gfs_leaf_t *)BH_DATA(nbh);
	nleaf->lf_depth = leaf->lf_depth;
	nleaf->lf_dirent_format = cpu_to_gfs32(GFS_FORMAT_DE);

	if (gfs_dirent_alloc(dip, nbh, filename->len, &dent))
	  GFS_ASSERT(0, );

	dip->i_di.di_blocks++;

	gfs_drelse(bh);

	bh = nbh;
	leaf = nleaf;
      }
    }

    
    /*  If the gfs_dirent_alloc() succeeded, it pinned the "bh".  */

    gfs_inum_out(inum, (char *)&dent->de_inum);
    dent->de_hash = cpu_to_gfs32(hash);
    dent->de_type = cpu_to_gfs16(type);
    gfs_memcpy((char *)(dent + 1), filename->name, filename->len);

    leaf->lf_entries = gfs16_to_cpu(leaf->lf_entries) + 1;
    leaf->lf_entries = cpu_to_gfs16(leaf->lf_entries);

    gfs_drelse(bh);


    dibh = gfs_get_inode_buffer(dip);
    if (!dibh)
    {
      EXIT(GFN_DIR_E_ADD);
      return -EIO;
    }

    dip->i_di.di_entries++;
    dip->i_di.di_mtime = dip->i_di.di_ctime = osi_current_time();

    gfs_trans_add_bh(sdp, dibh);
    gfs_dinode_out(&dip->i_di, BH_DATA(dibh));
    gfs_drelse(dibh);


    EXIT(GFN_DIR_E_ADD);

    return 0;
  }


  EXIT(GFN_DIR_E_ADD);

  return -ENOENT;
}


/**
 * dir_e_del - 
 * @dip: The GFS inode
 * @filename:
 *
 * Returns:
 */

STATIC int dir_e_del(gfs_inode_t *dip, osi_qstr_t *filename)
{
  osi_buf_t *bh, *dibh;
  gfs_dirent_t *dent, *prev;
  gfs_leaf_t *leaf;
  unsigned int entries;
  int64 error;

  ENTER(GFN_DIR_E_DEL);


  /*  Sanity Checks  */

  GFS_ASSERT(dip->i_di.di_flags & GFS_DIF_EXHASH, );


  error = linked_leaf_search(dip, filename, &dent, &prev, &bh);
  if (error)
  {
    EXIT(GFN_DIR_E_DEL);
    return error;
  }

  gfs_dirent_del(dip, bh, prev, dent);  /*  Pins bh  */

  leaf = (gfs_leaf_t *)BH_DATA(bh);
  entries = gfs16_to_cpu(leaf->lf_entries);
  GFS_ASSERT(entries, );
  entries--;
  leaf->lf_entries = cpu_to_gfs16(entries);

  gfs_drelse(bh);


  GFS_ASSERT(dip->i_di.di_entries, );
  dip->i_di.di_entries--;

  dip->i_di.di_mtime = dip->i_di.di_ctime = osi_current_time();


  dibh = gfs_get_inode_buffer(dip);
  if (!dibh)
  {
    EXIT(GFN_DIR_E_DEL);
    return -EIO;
  }

  gfs_trans_add_bh(dip->i_sbd, dibh);
  gfs_dinode_out(&dip->i_di, BH_DATA(dibh));
  gfs_drelse(dibh);


  EXIT(GFN_DIR_E_DEL);

  return 0;
}


/**
 * dir_e_read - Reads the entries from a directory into a filldir buffer 
 * @dip: dinode pointer
 * @f_pos: the hash of the last entry read shifted to the right once
 * @dirent: buffer for the filldir function to fill 
 * @filldir: points to the filldir function to use
 *
 */

STATIC int dir_e_read(gfs_inode_t *dip, int64 *f_pos, void *dirent, gfs_filldir_t filldir)
{
  gfs_sbd_t *sdp = dip->i_sbd; /* */ 
  osi_buf_t *bh;      /* */
  gfs_leaf_t leaf;    /* */
  uint16 reread_lp;   /* 1 if the leaf pointer needs to be reread, 0 if not */
  uint32 hsize,        
         len,         /* */
	 ht_offset,   /* a multiple of sdp->sd_hash_ptrs.  used to find the offset in the
			 inode of the current leaf block. */
	 lp_offset,   /* offset in the current leaf */
	 index,       /* The current index based on the hash value stored in f_pos */
	 hash;        /* */
  uint64 *lp;         /* the leaf pointer */
  uint64 leaf_no;     /* */
  int64 error;        /* */

  ENTER(GFN_DIR_E_READ);


  /* we must re-read the lp the first time we come into this
   * it used to be that we could use the value of lp_offset
   * to determine this, but since we calculate lp_offset now
   * we can't be sure that it will be 0 the first time.  */
  reread_lp = 1;

  /*  Sanity Checks  */
  GFS_ASSERT(dip->i_di.di_flags & GFS_DIF_EXHASH, );
  hsize = 1 << dip->i_di.di_depth;
  GFS_ASSERT(hsize * sizeof(uint64) == dip->i_di.di_size, );

  lp = (uint64 *)gmalloc(sdp->sd_hash_bsize);
  if (!lp)
  {
    EXIT(GFN_DIR_E_READ);
    return -ENOMEM;
  }
  
  /* we need to shift the hash back to the left one so that the 
   * correct leaf is selected */
  hash = (uint32) ((*f_pos) << 1); 
 
  /* calculate the index based on the hash */
  index = hash >> (32 - dip->i_di.di_depth);

  GFS_ASSERT(index < hsize, );
  
  /* ht_offset is index rounded down to nearest half block size */
  ht_offset = index - (index % (sdp->sd_hash_ptrs));
  GFS_ASSERT(ht_offset < hsize, );
  /* lp_offset is difference between the index and the ht_offset */
  lp_offset = index - ht_offset;
  if (lp_offset >= sdp->sd_hash_ptrs)
    lp_offset = 0;
 
  while (index < hsize)
  {
    if (reread_lp)
    {	    	    
      error = gfs_internal_read(dip, (char *)lp, ht_offset * sizeof(uint64), sdp->sd_hash_bsize);
      if (error != sdp->sd_hash_bsize)
      {
	gfree(lp, sdp->sd_hash_bsize);
	EXIT(GFN_DIR_E_READ);
	return (error < 0) ? error : -EIO;
      }
      reread_lp = 0;
    }
    
    leaf_no = gfs64_to_cpu(lp[lp_offset]);
    
    error = get_leaf(dip, leaf_no, &bh);
    if (error)
    {
      gfree(lp, sdp->sd_hash_bsize);
      EXIT(GFN_DIR_E_READ);
      return error;
    }
    
    gfs_leaf_in(&leaf, BH_DATA(bh));
      
    len = 1 << (dip->i_di.di_depth - leaf.lf_depth);

    /* resets leaf pointer offset to the beginning of the leaf.  Without this, we
     * tend to end up missing several entries every time the filldir buffer fills 
     * make sure we aren't trying to mod by 0 */
    if (dip->i_di.di_depth - leaf.lf_depth)
      lp_offset = lp_offset - (lp_offset % (2 << (dip->i_di.di_depth - leaf.lf_depth - 1)));      

    error = do_filldir(dip, f_pos, dirent, filldir, bh, index, leaf_no);
    if (error)
    {
      if (error == -EINVAL)
        error = 0;
      gfree(lp, sdp->sd_hash_bsize);
	
      gfs_drelse(bh);

      EXIT(GFN_DIR_E_READ);
      return error;
    }
    gfs_drelse(bh);

    lp_offset += len;
    index += len;
    
    if (lp_offset >= sdp->sd_hash_ptrs)
    {
      /* the ht_offset should always be a multiple of the sd_hash_ptrs */
      ht_offset += sdp->sd_hash_ptrs;
      lp_offset = 0;
      reread_lp = 1;
    }
  }

  gfree(lp, sdp->sd_hash_bsize);


  EXIT(GFN_DIR_E_READ);

  return 0;
}


/**
 * dir_e_mvino -
 * @dip: The GFS inode
 * @filename:
 * @new_inode:
 *
 * Returns:
 */

STATIC int dir_e_mvino(gfs_inode_t *dip, osi_qstr_t *filename,
		       gfs_inum_t *inum, unsigned int new_type)
{
  gfs_sbd_t *sdp = dip->i_sbd;
  osi_buf_t *bh, *dibh;
  gfs_dirent_t *dent;
  int64 error;

  ENTER(GFN_DIR_E_MVINO);


  /*  Sanity Checks  */

  GFS_ASSERT(dip->i_di.di_flags & GFS_DIF_EXHASH, );


  error = linked_leaf_search(dip, filename, &dent, NULL, &bh);
  if (error)
  {
    EXIT(GFN_DIR_E_MVINO);
    return error;
  }

  gfs_trans_add_bh(sdp, bh);

  gfs_inum_out(inum, (char *)&dent->de_inum);
  dent->de_type = cpu_to_gfs16(new_type);

  gfs_drelse(bh);


  dibh = gfs_get_inode_buffer(dip);
  if (!dibh)
  {
    EXIT(GFN_DIR_E_MVINO);
    return -EIO;
  }

  dip->i_di.di_mtime = dip->i_di.di_ctime = osi_current_time();

  gfs_trans_add_bh(sdp, dibh);
  gfs_dinode_out(&dip->i_di, BH_DATA(dibh));
  gfs_drelse(dibh);
  

  EXIT(GFN_DIR_E_MVINO);

  return 0;
}


/**
 * dir_l_search -
 * @dip: The GFS inode
 * @filename:
 * @inode:
 *
 * Returns:
 */

STATIC int dir_l_search(gfs_inode_t *dip, osi_qstr_t *filename,
			gfs_inum_t *inum, unsigned int *type)
{
  osi_buf_t *dibh;
  gfs_dirent_t *dent;
  int64 error;

  ENTER(GFN_DIR_L_SEARCH);


  /*  Sanity checks  */

  GFS_ASSERT(!(dip->i_di.di_flags & GFS_DIF_EXHASH), );
  GFS_ASSERT(gfs_is_stuffed(dip), );


  dibh = gfs_get_inode_buffer(dip);
  if (!dibh)
  {
    EXIT(GFN_DIR_L_SEARCH);
    return -EIO;
  }


  error = leaf_search(dibh, filename, &dent, NULL);
  if (error)
  {
    gfs_drelse(dibh);
    EXIT(GFN_DIR_L_SEARCH);
    return error;
  }

  if (inum)
    gfs_inum_in(inum, (char *)&dent->de_inum);
  if (type)
    *type = gfs16_to_cpu(dent->de_type);


  gfs_drelse(dibh);


  EXIT(GFN_DIR_L_SEARCH);

  return 0;
}


/**
 * dir_l_add -
 * @dip: The GFS inode
 * @filename:
 * @inode:
 * @type:
 *
 * Returns:
 */

STATIC int dir_l_add(gfs_inode_t *dip, osi_qstr_t *filename,
		     gfs_inum_t *inum, unsigned int type)
{
  osi_buf_t *dibh;
  gfs_dirent_t *dent;
  int error = 0;

  ENTER(GFN_DIR_L_ADD);


  /*  Sanity checks  */

  GFS_ASSERT(!(dip->i_di.di_flags & GFS_DIF_EXHASH), );
  GFS_ASSERT(gfs_is_stuffed(dip), );


  dibh = gfs_get_inode_buffer(dip);
  if (!dibh)
  {
    EXIT(GFN_DIR_L_ADD);
    return -EIO;
  }


  if (gfs_dirent_alloc(dip, dibh, filename->len, &dent))
  {
    gfs_drelse(dibh);

    error = dir_make_exhash(dip);

    if (!error)
      error = dir_e_add(dip, filename, inum, type);

    EXIT(GFN_DIR_L_ADD);
    return error;
  }


  /*  gfs_dirent_alloc() pins  */

  gfs_inum_out(inum, (char *)&dent->de_inum);
  dent->de_hash = dir_hash(filename);
  dent->de_hash = cpu_to_gfs32(dent->de_hash);
  dent->de_type = cpu_to_gfs16(type);
  gfs_memcpy((char *)(dent + 1), filename->name, filename->len);


  dip->i_di.di_entries++;
  dip->i_di.di_mtime = dip->i_di.di_ctime = osi_current_time();

  gfs_dinode_out(&dip->i_di, BH_DATA(dibh));
  gfs_drelse(dibh);


  EXIT(GFN_DIR_L_ADD);

  return 0;
}


/**
 * dir_l_del - 
 * @dip: The GFS inode
 * @filename:
 *
 * Returns:
 */

STATIC int dir_l_del(gfs_inode_t *dip, osi_qstr_t *filename)
{
  osi_buf_t *dibh;
  gfs_dirent_t *dent, *prev;
  int64 error;

  ENTER(GFN_DIR_L_DEL);


  /*  Sanity checks  */

  GFS_ASSERT(!(dip->i_di.di_flags & GFS_DIF_EXHASH), );
  GFS_ASSERT(gfs_is_stuffed(dip), );


  dibh = gfs_get_inode_buffer(dip);
  if (!dibh)
  {
    EXIT(GFN_DIR_L_DEL);
    return -EIO;    
  }


  error = leaf_search(dibh, filename, &dent, &prev);
  if (error)
  {
    gfs_drelse(dibh);

    EXIT(GFN_DIR_L_DEL);
    return error;
  }

  gfs_dirent_del(dip, dibh, prev, dent);


  /*  gfs_dirent_del() pins  */

  GFS_ASSERT(dip->i_di.di_entries, );
  dip->i_di.di_entries--;

  dip->i_di.di_mtime = dip->i_di.di_ctime = osi_current_time();

  
  gfs_dinode_out(&dip->i_di, BH_DATA(dibh));
  gfs_drelse(dibh);


  EXIT(GFN_DIR_L_DEL);

  return 0;
}


/**
 * dir_l_read -
 * @dip:
 * @f_pos:
 * @dirent:
 * @filldir:
 *
 * Returns:
 */

STATIC int dir_l_read(gfs_inode_t *dip, int64 *f_pos, void *dirent, gfs_filldir_t filldir)
{
  osi_buf_t *dibh;
  int error;

  ENTER(GFN_DIR_L_READ);


  /*  Sanity checks  */

  GFS_ASSERT(!(dip->i_di.di_flags & GFS_DIF_EXHASH), );
  GFS_ASSERT(gfs_is_stuffed(dip), );


  dibh = gfs_get_inode_buffer(dip);
  if (!dibh)
  {
    EXIT(GFN_DIR_L_READ);
    return -EIO;
  }

  error = do_filldir(dip, f_pos, dirent, filldir, dibh, 0, 0);
  if (error == -EINVAL)
    error = 0;


  gfs_drelse(dibh);

  EXIT(GFN_DIR_L_READ);

  return error;
}


/**
 * dir_l_mvino -
 * @dip:
 * @filename:
 * @new_inode:
 *
 * Returns:
 */

STATIC int dir_l_mvino(gfs_inode_t *dip, osi_qstr_t *filename, 
		       gfs_inum_t *inum, unsigned int new_type)
{
  osi_buf_t *dibh;
  gfs_dirent_t *dent;
  int64 error;

  ENTER(GFN_DIR_L_MVINO);


  /*  Sanity checks  */

  GFS_ASSERT(!(dip->i_di.di_flags & GFS_DIF_EXHASH), );
  GFS_ASSERT(gfs_is_stuffed(dip), );


  dibh = gfs_get_inode_buffer(dip);
  if (!dibh)
  {
    EXIT(GFN_DIR_L_MVINO);
    return -EIO;
  }


  error = leaf_search(dibh, filename, &dent, NULL);
  if (error)
  {
    gfs_drelse(dibh);

    EXIT(GFN_DIR_L_MVINO);
    return error;
  }


  gfs_trans_add_bh(dip->i_sbd, dibh);

  gfs_inum_out(inum, (char *)&dent->de_inum);
  dent->de_type = cpu_to_gfs16(new_type);

  dip->i_di.di_mtime = dip->i_di.di_ctime = osi_current_time();

  gfs_dinode_out(&dip->i_di, BH_DATA(dibh));
  gfs_drelse(dibh);


  EXIT(GFN_DIR_L_MVINO);

  return 0;
}


/**
 * gfs_dir_search - Search a directory
 * @dip: The GFS inode
 * @filename:
 * @inode:
 *
 * This routine searches a directory for a file or another directory.
 * Assumes a glock is held on dip.
 *
 * Returns: Inode number if found, -EXXXX on failure.
 */

int gfs_dir_search(gfs_inode_t *dip, osi_qstr_t *filename,
		   gfs_inum_t *inum, unsigned int *type)
{
  int error;

  ENTER(GFN_DIR_SEARCH);

  if (dip->i_di.di_flags & GFS_DIF_EXHASH)
    error = dir_e_search(dip, filename, inum, type);
  else
    error = dir_l_search(dip, filename, inum, type);

  EXIT(GFN_DIR_SEARCH);

  return error;
}


/**
 * gfs_dir_add - Add new filename into directory
 * @dip: The GFS inode
 * @filename: The new name
 * @inode: The inode number of the entry
 * @type: The type of the entry
 *
 * Returns: 0 on success, error code on failure
 */

int gfs_dir_add(gfs_inode_t *dip, osi_qstr_t *filename,
		gfs_inum_t *inum, unsigned int type)
{
  int error;

  ENTER(GFN_DIR_ADD);

  if (dip->i_di.di_flags & GFS_DIF_EXHASH)
    error = dir_e_add(dip, filename, inum, type);
  else
    error = dir_l_add(dip, filename, inum, type);

  EXIT(GFN_DIR_ADD);

  return error;
}


/**
 * gfs_dir_del - Delete a directory entry
 * @dip: The GFS inode
 * @filename: The filename
 *
 * Returns: 0 on success, error code on failure
 */

int gfs_dir_del(gfs_inode_t *dip, osi_qstr_t *filename)
{
  int error;

  ENTER(GFN_DIR_DEL);

  if (dip->i_di.di_flags & GFS_DIF_EXHASH)
    error = dir_e_del(dip, filename);
  else
    error = dir_l_del(dip, filename);

  EXIT(GFN_DIR_DEL);

  return error;
}


/**
 * gfs_dir_read - Translate a GFS filename
 * @dip: The GFS inode
 * @f_pos:
 * @dirent:
 * @filldir:
 *
 * Returns: 0 on success, error code otherwise
 */

int gfs_dir_read(gfs_inode_t *dip, int64 *f_pos, void *dirent, gfs_filldir_t filldir)
{
  int error;

  ENTER(GFN_DIR_READ);

  if (dip->i_di.di_flags & GFS_DIF_EXHASH)
    error = dir_e_read(dip, f_pos, dirent, filldir);
  else
    error = dir_l_read(dip, f_pos, dirent, filldir);

  EXIT(GFN_DIR_READ);

  return error;
}


/**
 * gfs_dir_mvino - Change inode number of directory entry
 * @dip: The GFS inode
 * @filename:
 * @new_inode:
 *
 * This routine changes the inode number of a directory entry.  It's used
 * by rename to change ".." when a directory is moved.
 * Assumes a glock is held on dvp.
 *
 * Returns: 0 on success, -EXXXX on failure
 */

int gfs_dir_mvino(gfs_inode_t *dip, osi_qstr_t *filename,
		  gfs_inum_t *inum, unsigned int new_type)
{
  int error;

  ENTER(GFN_DIR_MVINO);

  if (dip->i_di.di_flags & GFS_DIF_EXHASH)
    error = dir_e_mvino(dip, filename, inum, new_type);
  else
    error = dir_l_mvino(dip, filename, inum, new_type);

  EXIT(GFN_DIR_MVINO);

  return error;
}


/**
 * leaf_free - Deallocate a directory leaf
 * @dip: the directory
 * @ht_offset: the hash table offset in the directory
 * @len: the number of pointers to this leaf
 * @leaf_no: the leaf number
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int leaf_free(gfs_inode_t *dip, uint32 ht_offset, uint32 len, uint64 leaf_no)
{
  gfs_sbd_t *sdp = dip->i_sbd;
  gfs_leaf_t tmp_leaf;
  gfs_rgrp_list_t rlist;
  gfs_rgrpd_t *rgd;
  osi_buf_t *bh;
  uint64 blk;
  unsigned int blocks = 0;
  char *ht;
  unsigned int x, size = len * sizeof(uint64);
  int error;

  ENTER(GFN_LEAF_FREE);


  gfs_memset(&rlist, 0, sizeof(gfs_rgrp_list_t));

  ht = (char *)gmalloc(size);
  if (!ht)
  {
    error = -ENOMEM;
    goto fail;
  }
  gfs_memset(ht, 0, size);


  error = gfs_rindex_hold(sdp);
  if (error)
    goto fail_gfree;

  /*  Count the number of leaves  */

  for (blk = leaf_no; blk; blk = tmp_leaf.lf_next)
  {
    error = get_leaf(dip, blk, &bh);
    if (error)
      goto fail_free_rlist;
    gfs_leaf_in(&tmp_leaf, BH_DATA(bh));
    gfs_drelse(bh);

    error = gfs_rlist_add(sdp, &rlist, blk);
    if (error)
      goto fail_free_rlist;
  }

  gfs_rlist_sort(&rlist);

  for (x = 0; x < rlist.rl_rgrps; x++)
  {
    rgd = rlist.rl_list[x];

    error = gfs_glock_rg(rgd, 0);
    if (error)
      goto fail_rg_gunlock;

    blocks += rgd->rd_ri.ri_length;    
  }
  

  /*  Trans may require:
      All the bitmaps that were reserved.
      One block for the dinode.
      All the hash blocks that will be changed.
      */

  error = gfs_trans_begin(sdp, TRANS_LEAF_DEALLOC, blocks + 1 + (DIV_RU(size, sdp->sd_jbsize) + 1), 0);
  if (error)
    goto fail_rg_gunlock;

  for (blk = leaf_no; blk; blk = tmp_leaf.lf_next)
  {
    error = get_leaf(dip, blk, &bh);
    if (error)
      goto fail_end_trans;
    gfs_leaf_in(&tmp_leaf, BH_DATA(bh));
    gfs_drelse(bh);

    error = gfs_metafree(dip, blk, 1);
    if (error)
      goto fail_end_trans;
  }

  error = gfs_internal_write(dip, ht, ht_offset * sizeof(uint64), size);
  if (error != size)
  {
    if (error >= 0)
      error = -EIO;
    goto fail_end_trans;
  }

  gfs_trans_end(sdp);


  for (x = 0; x < rlist.rl_rgrps; x++)
  {
    rgd = rlist.rl_list[x];
    gfs_gunlock_rg(rgd, 0);
  }
  gfs_rlist_free(&rlist);
  gfs_rindex_release(sdp);
  gfree(ht, size);


  EXIT(GFN_LEAF_FREE);

  return 0;



 fail_end_trans:
  gfs_trans_mark(sdp, TRF_POISONED);
  gfs_trans_end(sdp);

 fail_rg_gunlock:
  while (x--)
  {
    rgd = rlist.rl_list[x];
    gfs_gunlock_rg(rgd, 0);
  }

 fail_free_rlist:
  gfs_rlist_free(&rlist);
  gfs_rindex_release(sdp);

 fail_gfree:
  gfree(ht, size);

 fail:
  EXIT(GFN_LEAF_FREE);
  return error;
}


/**
 * gfs_dir_exhash_free - free all the leaf block in a directory
 * @dip: the directory
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_dir_exhash_free(gfs_inode_t *dip)
{
  gfs_sbd_t *sdp = dip->i_sbd;
  osi_buf_t *bh;
  gfs_leaf_t leaf;
  uint32 hsize, len, ht_offset, lp_offset;
  uint64 *lp;
  uint64 leaf_no;
  int64 error;

  ENTER(GFN_DIR_EXHASH_FREE);


  /*  Sanity Checks  */

  GFS_ASSERT(dip->i_di.di_flags & GFS_DIF_EXHASH, );
  hsize = 1 << dip->i_di.di_depth;
  GFS_ASSERT(hsize * sizeof(uint64) == dip->i_di.di_size, );


  lp = (uint64 *)gmalloc(sdp->sd_hash_bsize);
  if (!lp)
  {
    error = -ENOMEM;
    goto fail;
  }


  ht_offset = 0;
  lp_offset = 0;
  

  while (ht_offset < hsize)
  {
    if (!lp_offset)
    {
      error = gfs_internal_read(dip, (char *)lp, ht_offset * sizeof(uint64), sdp->sd_hash_bsize);
      if (error != sdp->sd_hash_bsize)
      {
	if (error >= 0)
	  error = -EIO;
	goto fail_gfree;
      }
    }

    leaf_no = gfs64_to_cpu(lp[lp_offset]);

    if (leaf_no)
    {
      error = get_leaf(dip, leaf_no, &bh);
      if (error)
	goto fail_gfree;
      gfs_leaf_in(&leaf, BH_DATA(bh));
      gfs_drelse(bh);

      len = 1 << (dip->i_di.di_depth - leaf.lf_depth);

      error = leaf_free(dip, ht_offset, len, leaf_no);
      if (error)
	goto fail_gfree;

      ht_offset += len;
      lp_offset += len;
    }
    else
    {
      ht_offset++;
      lp_offset++;
    }

    if (lp_offset >= sdp->sd_hash_ptrs)
      lp_offset = 0;
  }


  /*  Make this a regular file in case we crash.
      (We don't want to free these blocks a second time.)  */

  error = gfs_trans_begin(sdp, TRANS_DIR2FILE, 1, 0);
  if (error)
    goto fail_gfree;

  bh = gfs_get_inode_buffer(dip);
  if (!bh)
  {
    error = -EIO;
    goto fail_end_trans;
  }

  gfs_trans_add_bh(sdp, bh);
  ((gfs_dinode_t *)(BH_DATA(bh)))->di_type = cpu_to_gfs16(FILE_REG);

  gfs_drelse(bh);

  gfs_trans_end(sdp);


  gfree(lp, sdp->sd_hash_bsize);

  EXIT(GFN_DIR_EXHASH_FREE);

  return 0;



 fail_end_trans:
  gfs_trans_mark(sdp, TRF_POISONED);
  gfs_trans_end(sdp);

 fail_gfree:
  gfree(lp, sdp->sd_hash_bsize);

 fail:
  EXIT(GFN_DIR_EXHASH_FREE);
  return error;
}


