/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portions Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name:  gfs_ondisk.h
**
**    Description:  The GFS on-disk structures and the prototypes for the
**      routines to read and write them.
**
**    NOTE:
**      
**      If you add 8 byte fields to these structures, they must be 8 byte
**      aligned.  4 byte field must be 4 byte aligned, etc...
**
**      All structures must be a multiple of 8 bytes long.
**
**    GRIPES:
**      We should have forgetten about supporting 512B FS block sizes 
**      and made the di_reserved field in the gfs_dinode_t structure
**      much bigger.
**
**      de_rec_len in gfs_dirent_t should really have been a 32-bit value
**      as it now limits us to a 64k FS block size (with the current code
**      in dir.c).
**
*******************************************************************************
******************************************************************************/

#ifndef __GFS_ONDISK_DOT_H__
#define __GFS_ONDISK_DOT_H__

#include "global.h"



#define GFS_MAGIC               (0x01161970)
#define GFS_BASIC_BLOCK         (512)



/*  Lock numbers of the LM_TYPE_MISC type  */

#define GFS_MOUNT_LOCK          (0)
#define GFS_LIVE_LOCK           (1)
#define GFS_TRANS_LOCK          (2)
#define GFS_RENAME_LOCK         (3)



/*  Format numbers for various metadata types  */

#define GFS_FORMAT_SB       	(100)
#define GFS_FORMAT_RG       	(200)
#define GFS_FORMAT_RB           (300)
#define GFS_FORMAT_DI       	(400)
#define GFS_FORMAT_IN           (500)
#define GFS_FORMAT_LF           (600)
#define GFS_FORMAT_JD           (700)
#define GFS_FORMAT_LH           (800)
#define GFS_FORMAT_LD           (900)
/*  These don't have actual gfs_meta_header_t structures to go with them  */
#define GFS_FORMAT_JI           (1000)
#define GFS_FORMAT_RI           (1100)
#define GFS_FORMAT_DE           (1200)
/*  These are part of the superblock  */
#define GFS_FORMAT_FS           (1305)
#define GFS_FORMAT_MULTI        (1401)





/*
 *  An on-disk inode number
 */

#define GFS_INUM_EQUAL(ino1, ino2) \
(((ino1)->no_formal_ino == (ino2)->no_formal_ino) && \
 ((ino1)->no_addr == (ino2)->no_addr))

struct gfs_inum
{
  uint64              no_formal_ino;
  uint64              no_addr;
};
typedef struct gfs_inum gfs_inum_t;





/*
 *  Generic metadata head structure
 *
 *  Every inplace buffer logged in the journal must start with this.
 */

#define GFS_METATYPE_NONE       (0)
#define GFS_METATYPE_SB       	(1)
#define GFS_METATYPE_RG       	(2)
#define GFS_METATYPE_RB         (3)
#define GFS_METATYPE_DI       	(4)
#define GFS_METATYPE_IN         (5)
#define GFS_METATYPE_LF         (6)
#define GFS_METATYPE_JD         (7)
#define GFS_METATYPE_LH         (8)
#define GFS_METATYPE_LD         (9)

#define GFS_META_CLUMP          (64)

struct gfs_meta_header
{
  uint32              mh_magic;         /* Magic number */
  uint32              mh_type;          /* GFS_METATYPE_XX */
  uint64              mh_generation;    /* Generation number */
  uint32              mh_format;        /* GFS_FORMAT_XX */
  uint32              mh_pad;
};
typedef struct gfs_meta_header gfs_meta_header_t;

typedef struct gfs_meta_header gfs_rgrpbits_t;
typedef struct gfs_meta_header gfs_jdata_t;





/*
 *  super-block structure
 *
 *  It's probably good if SIZEOF_SB <= GFS_BASIC_BLOCK
 */

/*  Address of SuperBlock in GFS basic blocks  */
#define GFS_SB_ADDR     	(128)
/*  The lock number for the superblock (must be zero)  */
#define GFS_SB_LOCK             (0)

#define GFS_SBF_SUBPOOL_ALIGNED (0x00000001)

/*  Requirement:  GFS_LOCKNAME_LEN % 8 == 0
    Includes: the fencing zero at the end  */
#define GFS_LOCKNAME_LEN        (64)

struct gfs_sb
{
  /*  Order is important  */
  gfs_meta_header_t   sb_header;

  uint32              sb_fs_format;
  uint32              sb_multihost_format;
  uint32              sb_flags;

  /*  Important information  */
  uint32              sb_bsize;         /* fundamental fs block size in bytes */
  uint32              sb_bsize_shift;   /* log2(sb_bsize) */
  uint32              sb_seg_size;      /* Journal segment size in FS blocks */

  gfs_inum_t          sb_jindex_di;     /* journal index inode number (GFS_SB_LOCK) */
  gfs_inum_t          sb_rindex_di;     /* resource index inode number (GFS_SB_LOCK) */
  gfs_inum_t          sb_root_di;       /* root directory inode number (GFS_ROOT_LOCK) */

  char                sb_lockproto[GFS_LOCKNAME_LEN];  /* Type of locking this FS uses */
  char                sb_locktable[GFS_LOCKNAME_LEN];  /* Name of lock table for this FS */

  char                sb_reserved[128];
};
typedef struct gfs_sb gfs_sb_t;





/*
 *  journal index structure 
 */

struct gfs_jindex
{
  uint64              ji_addr;          /* starting block of the journal */
  uint32              ji_nsegment;      /* number of segments in journal */
  uint32              ji_pad;

  char                ji_reserved[64];
};
typedef struct gfs_jindex gfs_jindex_t;





/*
 *  resource index structure 
 */

struct gfs_rindex
{
  uint64              ri_addr;          /* rgrp block disk address */
  uint32              ri_length;        /* length of rgrp header in fs blocks */
  uint32              ri_pad;

  uint64              ri_data1;         /* first data location */
  uint32              ri_data;          /* num of data blocks in rgrp */

  uint32              ri_bitbytes;      /* number of bytes in data bitmaps */
  
  char                ri_reserved[64];
};
typedef struct gfs_rindex gfs_rindex_t;





/*
 *  resource group header structure
 *
 */

/* Number of blocks per byte in rgrp */
#define GFS_NBBY		(4)
#define GFS_BIT_SIZE            (2)
#define GFS_BIT_MASK            (0x00000003)

#define GFS_BLKST_FREE          (0)
#define GFS_BLKST_USED          (1)
#define GFS_BLKST_FREEMETA      (2)
#define GFS_BLKST_USEDMETA      (3)

struct gfs_rgrp
{
  gfs_meta_header_t   rg_header;

  uint32              rg_flags;         /* flags */

  uint32              rg_free;          /* number of free data blocks */

  uint32              rg_useddi;        /* number of dinodes */
  uint32              rg_freedi;        /* number of unused dinodes */
  gfs_inum_t          rg_freedi_list;   /* list of free dinodes */

  uint32              rg_usedmeta;      /* number of used metadata blocks (not including dinodes) */
  uint32              rg_freemeta;      /* number of unused metadata blocks */

  char                rg_reserved[64];
};
typedef struct gfs_rgrp gfs_rgrp_t;





/*
 *  dinode structure
 */

#define GFS_MAX_META_HEIGHT     (10)
#define GFS_DIR_MAX_DEPTH       (17)

/*  Dinode types  */
#define FILE_NON                (0)
#define FILE_REG                (1)
#define FILE_DIR                (2)
#define FILE_LNK                (5)
#define FILE_BLK                (7)
#define FILE_CHR                (8)
#define FILE_FIFO               (101)
#define FILE_SOCK               (102)

/*  Dinode flags  */
#define GFS_DIF_JDATA           (0x00000001)
#define GFS_DIF_EXHASH          (0x00000002)
#define GFS_DIF_UNUSED          (0x00000004)

struct gfs_dinode
{
  gfs_meta_header_t   di_header;

  gfs_inum_t          di_num;

  uint32              di_mode;          /* mode of file */
  uint32              di_uid;	        /* owner's user id */
  uint32              di_gid;           /* owner's group id */
  uint32              di_nlink;         /* number of links to this file */
  uint64              di_size;          /* number of bytes in file */
  uint64              di_blocks;        /* number of blocks in file */
  int64               di_atime;         /* time last accessed */
  int64               di_mtime;         /* time last modified */
  int64               di_ctime;         /* time last changed */
  uint32              di_major;         /* device major number */
  uint32              di_minor;         /* device minor number */

  uint64              di_rgrp;          /* dinode rgrp block number */
  uint64              di_goal_rgrp;     /* rgrp to alloc from next */
  uint32              di_goal_dblk;     /* data block goal */
  uint32              di_goal_mblk;     /* metadata block goal */
  uint32              di_flags;         /* flags */
  uint32              di_payload_format;  /* gfs_rindex_t, gfs_jindex_t, or gfs_dirent_t */
  uint16              di_type;          /* type of file */
  uint16              di_height;        /* height of metadata */
  uint32              di_incarn;        /* incarnation number */
  uint16              di_pad;

  /*  These only apply to directories  */
  uint16              di_depth;         /* Number of bits in the table */
  uint32              di_entries;       /* The number of entries in the directory */

  /*  This only applies to unused inodes  */
  gfs_inum_t          di_next_unused;

  char                di_reserved[64];
};
typedef struct gfs_dinode gfs_dinode_t;





/*
 *  indirect block header
 */

struct gfs_indirect
{
  gfs_meta_header_t   in_header;

  char                in_reserved[64];
};
typedef struct gfs_indirect gfs_indirect_t;





/*
 *  directory structure - many of these per directory file
 */

#define GFS_FNAMESIZE       	(256)

#define GFS_DIR_PAD (8)
#define GFS_DIR_ROUND (GFS_DIR_PAD - 1)
#define GFS_DIR_REC_LEN(name_len) ((sizeof(gfs_dirent_t) + (name_len) + GFS_DIR_ROUND) & \
				   (~GFS_DIR_ROUND))

struct gfs_dirent
{
  gfs_inum_t          de_inum;          /* Inode number */
  uint32              de_hash;          /* hash of the filename */
  uint16              de_rec_len;       /* the length of the dirent */
  uint16              de_name_len;      /* the length of the name */
  uint16              de_type;          /* type of dinode this points to */

  char                de_reserved[14];
};
typedef struct gfs_dirent gfs_dirent_t;





/*
 *  Header of leaf directory nodes
 */

struct gfs_leaf
{
  gfs_meta_header_t   lf_header;

  uint16              lf_depth;         /* Depth of leaf */
  uint16              lf_entries;       /* Number of dirents in leaf */
  uint32              lf_dirent_format; /* Format of the dirents */
  uint64              lf_next;          /* Next leaf, if overflow */

  char                lf_reserved[64];
};
typedef struct gfs_leaf gfs_leaf_t;





/*
 *  Log header structure
 */

#define GFS_LOG_HEAD_UNMOUNT    (0x00000001)

struct gfs_log_header
{
  gfs_meta_header_t   lh_header;

  uint32              lh_flags;         /* Flags */
  uint32              lh_pad;

  uint64              lh_first;         /* Block number of first header in this trans */
  uint64              lh_sequence;      /* Sequence number of this transaction */

  uint64              lh_tail;          /* Block number of log tail */
  uint64              lh_last_nopd;     /* block number of last nopen dump */

  char                lh_reserved[64];
};
typedef struct gfs_log_header gfs_log_header_t;





/*
 *  Log type descriptor
 */

#define GFS_LOG_DESC_METADATA	(300)
/*  ld_data1 is the number of metadata blocks in the descriptor,
    ld_data2 is the transaction type.
    */

#define GFS_NOP_DUMPS_PER_LOG   (4)
#define GFS_LOG_DESC_NOPOPEN    (400)
/*  ld_data1 is TRUE if this is a dump.
    ld_data2 is unused.
    */

#define GFS_LOG_DESC_NOPCLOSE   (401)
/*  ld_data1 is unused.
    ld_data2 is unused.
    */

#define GFS_LOG_DESC_LAST	(500)
/*  ld_data1 is unused.
    ld_data2 is unused.
    */

struct gfs_log_descriptor
{
  gfs_meta_header_t   ld_header;

  uint32              ld_type;          /* Type of data in this log chunk */
  uint32              ld_length;        /* Number of buffers in this chunk */
  uint32              ld_data1;         /* descriptor specific field */
  uint32              ld_data2;         /* descriptor specific field */

  char                ld_reserved[64];
};
typedef struct gfs_log_descriptor gfs_log_descriptor_t;





/*
 *  Metadata block tags
 */

struct gfs_block_tag
{
  uint64              bt_blkno;         /* inplace block number */
  uint32              bt_flags;         /* flags */
  uint32              bt_pad;
};
typedef struct gfs_block_tag gfs_block_tag_t;





/*  Endian functions  */

#define GFS_ENDIAN_BIG

#ifdef GFS_ENDIAN_BIG

#define gfs16_to_cpu osi_be16_to_cpu
#define gfs32_to_cpu osi_be32_to_cpu
#define gfs64_to_cpu osi_be64_to_cpu

#define cpu_to_gfs16 osi_cpu_to_be16
#define cpu_to_gfs32 osi_cpu_to_be32
#define cpu_to_gfs64 osi_cpu_to_be64

#else  /*  GFS_ENDIAN_BIG  */

#define gfs16_to_cpu osi_le16_to_cpu
#define gfs32_to_cpu osi_le32_to_cpu
#define gfs64_to_cpu osi_le64_to_cpu

#define cpu_to_gfs16 osi_cpu_to_le16
#define cpu_to_gfs32 osi_cpu_to_le32
#define cpu_to_gfs64 osi_cpu_to_le64

#endif  /*  GFS_ENDIAN_BIG  */


/*  Translation functions  */

void gfs_inum_in(gfs_inum_t *no, char *buf);
void gfs_inum_out(gfs_inum_t *no, char *buf);
void gfs_meta_header_in(gfs_meta_header_t *mh, char *buf);
void gfs_meta_header_out(gfs_meta_header_t *mh, char *buf);
void gfs_sb_in(gfs_sb_t *sb, char *buf);
void gfs_sb_out(gfs_sb_t *sb, char *buf);
void gfs_jindex_in(gfs_jindex_t *jindex, char *buf);
void gfs_jindex_out(gfs_jindex_t *jindex, char *buf);
void gfs_rindex_in(gfs_rindex_t *rindex, char *buf);
void gfs_rindex_out(gfs_rindex_t *rindex, char *buf);
void gfs_rgrp_in(gfs_rgrp_t *rgrp, char *buf);
void gfs_rgrp_out(gfs_rgrp_t *rgrp, char *buf);
void gfs_dinode_in(gfs_dinode_t *dinode, char *buf);
void gfs_dinode_out(gfs_dinode_t *dinode, char *buf);
void gfs_indirect_in(gfs_indirect_t *indirect, char *buf);
void gfs_indirect_out(gfs_indirect_t *indirect, char *buf);
void gfs_dirent_in(gfs_dirent_t *dirent, char *buf);
void gfs_dirent_out(gfs_dirent_t *dirent, char *buf);
void gfs_leaf_in(gfs_leaf_t *leaf, char *buf);
void gfs_leaf_out(gfs_leaf_t *leaf, char *buf);
void gfs_log_header_in(gfs_log_header_t *head, char *buf);
void gfs_log_header_out(gfs_log_header_t *head, char *buf);
void gfs_desc_in(gfs_log_descriptor_t *desc, char *buf);
void gfs_desc_out(gfs_log_descriptor_t *desc, char *buf);
void gfs_tag_in(gfs_block_tag_t *tag, char *buf);
void gfs_tag_out(gfs_block_tag_t *tag, char *buf);


/*  Printing functions  */

void gfs_inum_print(gfs_inum_t *no, int console);
void gfs_meta_header_print(gfs_meta_header_t *mh, int console);
void gfs_sb_print(gfs_sb_t *sb, int console);
void gfs_jindex_print(gfs_jindex_t *jindex, int console);
void gfs_rindex_print(gfs_rindex_t *rindex, int console);
void gfs_rgrp_print(gfs_rgrp_t *rgrp, int console);
void gfs_dinode_print(gfs_dinode_t *dinode, int console);
void gfs_indirect_print(gfs_indirect_t *indirect, int console);
void gfs_dirent_print(gfs_dirent_t *dirent, char *name, int console);
void gfs_leaf_print(gfs_leaf_t *leaf, int console);
void gfs_log_header_print(gfs_log_header_t *head, int console);
void gfs_desc_print(gfs_log_descriptor_t *desc, int console);
void gfs_tag_print(gfs_block_tag_t *tag, int console);


#endif  /*  __GFS_ONDISK_DOT_H__  */


