/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portions Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: glock.c
**
**    Description:
**	This module implements the GFS interface to the global locks.
**
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "glock.h"
#include "glops.h"
#include "inode.h"
#include "super.h"



DBTAG(GFS_LOCKING)

/*  Perform the test outlined in the table in glock.h.
    If the lock is already cached use the RELAXED_LOCKMODE_OK() test
    to avoid doing unecessary lock operations.  */

#define RELAXED_LOCKMODE_OK(requested, actual) ((requested) == (actual) || \
						((requested) == LM_ST_SHARED && (actual) == LM_ST_EXCLUSIVE))



STATIC void drop_glock(gfs_sbd_t *sdp, gfs_glock_t *gl);





/**
 * __lock_on_glock - Wait for exclusive access to a glock
 * @gl: The glock to wait for
 *
 */

STATIC void __lock_on_glock(gfs_glock_t *gl)
{
  osi_wait_vars(wait, "__lock_on_glock");

  do
  {
    osi_asleep(wait, gl->gl_wait_lock);
    if (osi_test_bit(GLF_LOCK, &gl->gl_flags))
      osi_await(wait);
    osi_unsleep(wait, gl->gl_wait_lock);
  }
  while (osi_test_and_set_bit(GLF_LOCK, &gl->gl_flags));
}


/**
 * lock_on_glock - Get exclusive access to a glock's fields
 * @gl: The glock to lock
 *
 * If we don't get the lock first time, call a non-inlined function to
 * wait until we do.
 */

STATIC inline void lock_on_glock(gfs_glock_t *gl)
{
  if (osi_test_and_set_bit(GLF_LOCK, &gl->gl_flags))
    __lock_on_glock(gl);
}


/**
 * trylock_on_glock - Try and get exclusive access to a glock, no waiting
 * @gl: The glock to attempt to lock
 *
 * Returns: Non-zero if successful
 */

STATIC inline int trylock_on_glock(gfs_glock_t *gl)
{
  return (osi_test_and_set_bit(GLF_LOCK, &gl->gl_flags) == 0);
}


/**
 * unlock_on_glock - End exclusive access to a glock's fields
 * @gl: The glock to unlock
 *
 */

STATIC inline void unlock_on_glock(gfs_glock_t *gl)
{
  GFS_ASSERT(osi_test_bit(GLF_LOCK, &gl->gl_flags), );
  osi_clear_bit(GLF_LOCK, &gl->gl_flags);
  osi_wakeup(gl->gl_wait_lock);
}


/**
 * glock_hash() - Turn glock number into hash bucket number
 * @lock: The glock number
 *
 * Returns: The number of the corresponding hash bucket
 */

STATIC inline unsigned int glock_hash(lm_lockname_t *name)
{
  unsigned int hash = (unsigned int)(name->ln_number);
  hash ^= (unsigned int)(name->ln_number >> 32);
  hash ^= (hash >> 16);
  hash ^= (hash >> 24);
  return hash & GFS_GHASH_MASK;
}


/**
 * gfs_blk2lockname() - convert a block number and a type into a lm_lockname_t
 * @sdp: the filesystem
 * @block: the block number
 * @type: the lock type
 * @name: the lm_lockname_t that gets filled in
 *
 */

void gfs_blk2lockname(gfs_sbd_t *sdp, uint64 block, unsigned int type, lm_lockname_t *name)
{
  ENTER(GFN_BLK2LOCKNAME);

  name->ln_number = block;
  name->ln_type = type;

  if (type != LM_TYPE_NONDISK)
    name->ln_number <<= sdp->sd_fsb2bb_shift;

  EXIT(GFN_BLK2LOCKNAME);
}


/**
 * gfs_lockname2blk() - convert a lockname to a block number
 * @sdp: the filesystem
 * @name: the lm_lockname_t
 *
 */

uint64 gfs_lockname2blk(gfs_sbd_t *sdp, lm_lockname_t *name)
{
  uint64 ret;

  ENTER(GFN_LOCKNAME2BLK);

  ret = name->ln_number;

  if (name->ln_type != LM_TYPE_NONDISK)
    ret >>= sdp->sd_fsb2bb_shift;

  EXIT(GFN_LOCKNAME2BLK);

  return ret;
}


/**
 * put_glstruct() - Decrement reference count on glstruct
 * @gl: The glstruct to put
 *
 */

STATIC __inline__ void put_glstruct(gfs_glock_t *gl)
{
  GFS_ASSERT(gl, );
  GFS_ASSERT(osi_atomic_read(&gl->gl_count), );

  osi_atomic_dec(&gl->gl_count);
}


/**
 * gfs_put_glstruct() - As put_glstruct(), but suitable for exporting
 * @gl: The glstruct to put
 *
 */

void gfs_put_glstruct(gfs_glock_t *gl)
{
  ENTER(GFN_PUT_GLSTRUCT);

  put_glstruct(gl);

  EXIT(GFN_PUT_GLSTRUCT);
}


/**
 * perm_glock() - Set GLF_PERM bit and move glstruct to the perm glock hash chain
 * @sdp: The GFS superblock
 * @gl: The glstruct to move
 *
 */

STATIC __inline__ void perm_glock(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  unsigned int bucket = gl->gl_bucket;

  osi_write_lock(&sdp->sd_ghash[bucket].hb_lock);
  if (!osi_test_and_set_bit(GLF_PERM, &gl->gl_flags))
  {
    osi_atomic_dec(&sdp->sd_ghash[bucket].hb_held_count);
    osi_atomic_inc(&sdp->sd_ghash[bucket].hb_perm_count);

    osi_list_del(&gl->gl_list);
    osi_list_add(&gl->gl_list, &sdp->sd_ghash[bucket].hb_perm);
  }
  osi_write_unlock(&sdp->sd_ghash[bucket].hb_lock);
}


/**
 * unperm_glock() - Clear GLF_PERM bit and move glstruct to the held glock hash chain
 * @sdp: The GFS superblock
 * @gl: The glstruct to move
 *
 */

STATIC __inline__ void unperm_glock(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  unsigned int bucket = gl->gl_bucket;

  osi_write_lock(&sdp->sd_ghash[bucket].hb_lock);
  if (osi_test_and_clear_bit(GLF_PERM, &gl->gl_flags))
  {
    osi_atomic_dec(&sdp->sd_ghash[bucket].hb_perm_count);
    osi_atomic_inc(&sdp->sd_ghash[bucket].hb_held_count);

    osi_list_del(&gl->gl_list);
    osi_list_add(&gl->gl_list, &sdp->sd_ghash[bucket].hb_held);
  }
  osi_write_unlock(&sdp->sd_ghash[bucket].hb_lock);
}


/**
 * hold_glock() - Set GLF_HELD bit and move glstruct to held glock hash chain
 * @sdp: The GFS superblock
 * @gl: The glstruct to move
 *
 */

STATIC __inline__ void hold_glock(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  unsigned int bucket = gl->gl_bucket;

  osi_write_lock(&sdp->sd_ghash[bucket].hb_lock);
  if (!osi_test_and_set_bit(GLF_HELD, &gl->gl_flags))
  {
    osi_atomic_dec(&sdp->sd_ghash[bucket].hb_notheld_count);
    osi_atomic_inc(&sdp->sd_ghash[bucket].hb_held_count);

    osi_list_del(&gl->gl_list);
    osi_list_add(&gl->gl_list, &sdp->sd_ghash[bucket].hb_held);

    osi_clear_bit(GLF_RELEXCL, &gl->gl_flags);
    osi_clear_bit(GLF_RELSHRD, &gl->gl_flags);
    osi_clear_bit(GLF_RELDFRD, &gl->gl_flags);
    osi_clear_bit(GLF_DEMOTEME, &gl->gl_flags);
    osi_clear_bit(GLF_DEPCHECK, &gl->gl_flags);
  }
  osi_write_unlock(&sdp->sd_ghash[bucket].hb_lock);
}


/**
 * unhold_glock() - Reset GLF_HELD bit and more glock to notheld hash chain
 * @sdp: The GFS superblock
 * @gl: The glstruct to move
 *
 */

STATIC __inline__ void unhold_glock(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  unsigned int bucket = gl->gl_bucket;

  osi_write_lock(&sdp->sd_ghash[bucket].hb_lock);
  if (osi_test_and_clear_bit(GLF_HELD, &gl->gl_flags))
  {
    osi_atomic_dec(&sdp->sd_ghash[bucket].hb_held_count);
    osi_atomic_inc(&sdp->sd_ghash[bucket].hb_notheld_count);

    osi_list_del(&gl->gl_list);
    osi_list_add(&gl->gl_list, &sdp->sd_ghash[bucket].hb_notheld);

    osi_time_stamp(&gl->gl_stamp);
  }
  osi_write_unlock(&sdp->sd_ghash[bucket].hb_lock);

  gl->gl_state = LM_ST_UNLOCKED;
}


/**
 * new_glock() - Get a new glstruct for a particular hash bucket
 * @sdp: The GFS superblock
 * @bucket: The hash bucket we wish our new glock to be in
 *
 * Get a new glock structure in the following way:
 *   - Look for an old glock at the end of the notheld hash chain
 *     Once the FS has been mounted for a while, I hope that most
 *     requests will be answered this way.
 *   - If we've got less held locks than the hard limit, grab a
 *     new glock struct from the slab cache. Otherwise we wait
 *     whilst we try and unlock some locks. If we manage to
 *     unlock some locks, try again.
 *   - Assuming that we are under the hard limit, schedule gfs_glockd
 *     if we are over the soft limit.
 *
 * Returns: A new gfs_glock_t
 */

STATIC gfs_glock_t *new_glock(gfs_sbd_t *sdp, unsigned int bucket)
{
  gfs_glock_t *gl = NULL;
  osi_list_head_t *head, *tmp;

  ENTER(GFN_NEW_GLOCK);


  /*
   * With this check we are not sure that there will actually be
   * any free glocks on our hash chain, but if the check fails,
   * there will certainly not be any free!
   */

  if (osi_atomic_read(&sdp->sd_ghash[bucket].hb_notheld_count))
  {
    osi_write_lock(&sdp->sd_ghash[bucket].hb_lock);

    head = tmp = &sdp->sd_ghash[bucket].hb_notheld;
    for (;;)
    {
      gl = NULL;
      tmp = tmp->prev;
      if (tmp == head)
	break;
      gl = osi_list_entry(tmp, gfs_glock_t, gl_list);

      if (osi_atomic_read(&gl->gl_count))
	continue;

      osi_list_del(&gl->gl_list);
      osi_atomic_dec(&sdp->sd_ghash[bucket].hb_notheld_count);

      break;
    }

    osi_write_unlock(&sdp->sd_ghash[bucket].hb_lock);
  }

  if (gl)
  {
    if (gl->gl_parent)
    {
      lock_on_glock(gl->gl_parent);
      osi_list_del(&gl->gl_parlist);
      unlock_on_glock(gl->gl_parent);
      put_glstruct(gl->gl_parent);
    }
    
    DO_PROF2(LM_PUT_LOCK,
	     sdp->sd_lockstruct.ls_ops->put_lock(gl->gl_lock););

    goto out;
  }


  /*
   * Now we know there is nothing free on our chain we try to
   * get a new glock from the slab cache.
   */

  gl = gfs_alloc_glock();


 out:
  EXIT(GFN_NEW_GLOCK);

  return gl;
}


/**
 * find_glstruct() - Find glstruct by lock number
 * @sdp: The GFS superblock
 * @name: The lock name
 * @parent: The parent lock of this lock
 * @bucket: The hash bucket in which we look for the lock
 *
 * Look down both hash chains in a hash bucket to try and
 * find a gfs_glock_t by number.
 *
 * Returns: NULL, or the gfs_glock_t with the requested number
 */

STATIC gfs_glock_t *find_glstruct(gfs_sbd_t *sdp,
				  lm_lockname_t *name, gfs_glock_t *parent,
				  uint32 bucket)
{
  osi_list_head_t *tmp, *head;
  gfs_glock_t *gl;

  ENTER(GFN_FIND_GLSTRUCT);

  
  tmp = head = &sdp->sd_ghash[bucket].hb_held;
  for (;;)
  {
    tmp = tmp->next;
    gl = NULL;
    if (tmp == head)
      break;
    gl = osi_list_entry(tmp, gfs_glock_t, gl_list);

    if (!lm_name_equal(&gl->gl_name, name) || gl->gl_parent != parent)
      continue;
    gfs_hold_glstruct(gl);
    goto out;
  }

  tmp = head = &sdp->sd_ghash[bucket].hb_notheld;
  for (;;)
  {
    tmp = tmp->next;
    gl = NULL;
    if (tmp == head)
      break;
    gl = osi_list_entry(tmp, gfs_glock_t, gl_list);

    if (!lm_name_equal(&gl->gl_name, name) || gl->gl_parent != parent)
      continue;
    gfs_hold_glstruct(gl);
    goto out;
  }

  tmp = head = &sdp->sd_ghash[bucket].hb_perm;
  for (;;)
  {
    tmp = tmp->next;
    gl = NULL;
    if (tmp == head)
      break;
    gl = osi_list_entry(tmp, gfs_glock_t, gl_list);

    if (!lm_name_equal(&gl->gl_name, name) || gl->gl_parent != parent)
      continue;
    gfs_hold_glstruct(gl);
    break;
  }


 out:
  EXIT(GFN_FIND_GLSTRUCT);
  return gl;
}


/**
 * gfs_get_glstruct() - Get a glstruct, or create one if one doesn't exist
 * @sdp: The GFS superblock
 * @number: the lock number
 * @glops: The glock_operations to use
 * @parent: The parent lock of this lock
 * @create: If FALSE, don't create the glock if it doesn't exist
 *
 * Returns: NULL, or a gfs_glock_t with the requested number
 */

gfs_glock_t *gfs_get_glstruct(gfs_sbd_t *sdp,
			      uint64 number, gfs_glock_operations_t *glops, gfs_glock_t *parent,
			      int create)
{
  lm_lockname_t name;
  uint32 bucket;
  gfs_glock_t *gl, *tmp;

  ENTER(GFN_GET_GLSTRUCT);


  gfs_blk2lockname(sdp, number, glops->go_type, &name);
  bucket = glock_hash(&name);

  osi_read_lock(&sdp->sd_ghash[bucket].hb_lock);
  gl = find_glstruct(sdp, &name, parent, bucket);
  osi_read_unlock(&sdp->sd_ghash[bucket].hb_lock);

  if (gl || !create)
    goto out;


  gl = new_glock(sdp, bucket);
  if (!gl)
    goto out;

  gfs_memset(gl, 0, sizeof(gfs_glock_t));

  DO_PROF(LM_GET_LOCK,
	  gl->gl_lock = sdp->sd_lockstruct.ls_ops->get_lock(sdp->sd_lockstruct.ls_lockspace, &name););
  if (!gl->gl_lock)
  {
    gfs_free_glock(gl);
    goto out;
  }

  osi_list_head_init(&gl->gl_list);
  gl->gl_name = name;
  osi_atomic_set(&gl->gl_count, 1);
  gl->gl_bucket = bucket;

  osi_spinlock_init(&gl->gl_head_lock);
  osi_init_wait_chan(gl->gl_wait_lock);
  osi_init_wait_chan(gl->gl_wait_unlock);

  gl->gl_ops = glops;

  gl->gl_state = LM_ST_UNLOCKED;

  gl->gl_parent = parent;
  osi_list_head_init(&gl->gl_parlist);
  osi_list_head_init(&gl->gl_children);

  osi_time_stamp(&gl->gl_stamp);

  osi_list_head_init(&gl->gl_new_list);
  osi_list_head_init(&gl->gl_new_bufs);

  osi_list_head_init(&gl->gl_incore_list);
  osi_list_head_init(&gl->gl_incore_bufs);

  osi_list_head_init(&gl->gl_depend);

  gfs_glock_arch_init(sdp, gl);

  if (parent)
  {
    gfs_hold_glstruct(parent);
    lock_on_glock(parent);
    osi_list_add(&gl->gl_parlist, &parent->gl_children);
    unlock_on_glock(parent);
  }


  osi_write_lock(&sdp->sd_ghash[bucket].hb_lock);
  if ((tmp = find_glstruct(sdp, &name, parent, bucket)) == NULL)
  {
    osi_list_add(&gl->gl_list, &sdp->sd_ghash[bucket].hb_notheld);
    osi_atomic_inc(&sdp->sd_ghash[bucket].hb_notheld_count);
  }
  osi_write_unlock(&sdp->sd_ghash[bucket].hb_lock);

  if (tmp)
  {
    lock_on_glock(parent);
    osi_list_del(&gl->gl_parlist);
    unlock_on_glock(parent);
    put_glstruct(parent);

    osi_list_head_init(&gl->gl_parlist);

    DO_PROF2(LM_PUT_LOCK,
	     sdp->sd_lockstruct.ls_ops->put_lock(gl->gl_lock););
    gfs_free_glock(gl);
    gl = tmp;
  }


 out:
  EXIT(GFN_GET_GLSTRUCT);

  return gl;
}


/**
 * release_glstruct() - Perform a few checks and then release gfs_glock_t
 * @sdp: The GFS superblock
 * @gl: The glstruct to release
 *
 */

STATIC void release_glstruct(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  ENTER(GFN_RELEASE_GLSTRUCT);


  GFS_ASSERT(osi_list_empty(&gl->gl_list), );
  GFS_ASSERT(!osi_atomic_read(&gl->gl_count), );
  GFS_ASSERT(!osi_atomic_read(&gl->gl_locked), );
  GFS_ASSERT(gl->gl_state == LM_ST_UNLOCKED, );
  GFS_ASSERT(osi_list_empty(&gl->gl_parlist), );
  GFS_ASSERT(osi_list_empty(&gl->gl_children), );
  GFS_ASSERT(!gl->gl_object, );
  GFS_ASSERT(!gl->gl_lvb, );

  DO_PROF(LM_PUT_LOCK,
	  sdp->sd_lockstruct.ls_ops->put_lock(gl->gl_lock););

  gfs_free_glock(gl);


  EXIT(GFN_RELEASE_GLSTRUCT);
}


/**
 * gfs_hold_lvb - attach a LVB from a glock
 * @sdp: The GFS superblock
 * @gl: The glock in question
 *
 */

int gfs_hold_lvb(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  int error = 0;

  ENTER(GFN_HOLD_LVB);

  lock_on_glock(gl);

  gl->gl_lvb_count++;
  if (gl->gl_lvb_count == 1)
  {
    gfs_hold_glstruct(gl);
    GFS_ASSERT(!gl->gl_lvb, );
    DO_PROF(LM_HOLD_LVB,
	    gl->gl_lvb = sdp->sd_lockstruct.ls_ops->hold_lvb(gl->gl_lock););
    if (!gl->gl_lvb)
    {
      put_glstruct(gl);
      gl->gl_lvb_count = 0;
      error = -ENOMEM;
    }
  }

  unlock_on_glock(gl);

  EXIT(GFN_HOLD_LVB);

  return error;
}


/**
 * gfs_unhold_lvb - detach a LVB from a glock
 * @sdp: The GFS superblock
 * @gl: The glock in question
 * 
 */

void gfs_unhold_lvb(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  ENTER(GFN_UNHOLD_LVB);

  lock_on_glock(gl);

  GFS_ASSERT(gl->gl_lvb_count, );
  gl->gl_lvb_count--;
  if (!gl->gl_lvb_count)
  {
    GFS_ASSERT(gl->gl_lvb, );
    DO_PROF(LM_UNHOLD_LVB,
	    sdp->sd_lockstruct.ls_ops->unhold_lvb(gl->gl_lock, gl->gl_lvb););
    gl->gl_lvb = NULL;
    put_glstruct(gl);
  }

  unlock_on_glock(gl);

  EXIT(GFN_UNHOLD_LVB);
}


/**
 * gfs_sync_lvb - sync a LVB
 * @sdp: The GFS superblock
 * @gl: The glock in question
 * 
 */

void gfs_sync_lvb(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  ENTER(GFN_SYNC_LVB);

  lock_on_glock(gl);

  GFS_ASSERT(gl->gl_lvb_count, );
  GFS_ASSERT(osi_atomic_read(&gl->gl_locked) &&
	     gl->gl_state == LM_ST_EXCLUSIVE, );
  DO_PROF(LM_SYNC_LVB,
	  sdp->sd_lockstruct.ls_ops->sync_lvb(gl->gl_lock, gl->gl_lvb););

  unlock_on_glock(gl);

  EXIT(GFN_SYNC_LVB);
}


/**
 * gfs_add_gl_dependency - Make an release-ordering dependency between to glocks
 * @sdp: the filesystem
 * @gl1: sync this glock
 * @gl2: before releasing this glock
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_add_gl_dependency(gfs_sbd_t *sdp, gfs_glock_t *gl1, gfs_glock_t *gl2)
{
  osi_list_head_t *tmp;
  gfs_gl_depend_t *dep;
  int error = 0;

  ENTER(GFN_ADD_GL_DEPENDENCY);


  /*  Look to see if this dependency already exists  */

  for (tmp = gl2->gl_depend.next; tmp != &gl2->gl_depend; tmp = tmp->next)
  {
    dep = osi_list_entry(tmp, gfs_gl_depend_t, gd_list);

    if (lm_name_equal(&dep->gd_name, &gl1->gl_name) && (dep->gd_parent == gl1->gl_parent))
    {
      osi_time_stamp(&dep->gd_time);
      goto out;
    }
  }


  /*  We need to add it new  */

  dep = gmalloc(sizeof(gfs_gl_depend_t));
  if (!dep)
  {
    error = -ENOMEM;
    goto out;
  }
  gfs_memset(dep, 0, sizeof(gfs_gl_depend_t));

  dep->gd_name = gl1->gl_name;
  dep->gd_parent = gl1->gl_parent;
  osi_time_stamp(&dep->gd_time);

  osi_list_add(&dep->gd_list, &gl2->gl_depend);
  osi_atomic_inc(&gl2->gl_num_dep);


 out:
  EXIT(GFN_ADD_GL_DEPENDENCY);

  return error;
}


/**
 * sync_dependencies - sync out any locks dependent on this one
 * @sdp: the filesystem
 * @gl: the glock
 *
 */

void sync_dependencies(gfs_sbd_t *sdp, gfs_glock_t *gl, unsigned int timeout)
{
  osi_list_head_t *tmp, *next, head;
  gfs_gl_depend_t *dep;
  gfs_glock_t *tmp_gl;

  ENTER(GFN_SYNC_DEPENDENCIES);

  osi_list_head_init(&head);


  for (next = gl->gl_depend.next; next != &gl->gl_depend; )
  {
    tmp = next;
    next = next->next;

    dep = osi_list_entry(tmp, gfs_gl_depend_t, gd_list);

    if (!osi_check_timeout(&dep->gd_time, timeout))
      continue;

    osi_list_del(&dep->gd_list);
    osi_atomic_dec(&gl->gl_num_dep);

    tmp_gl = find_glstruct(sdp, &dep->gd_name, dep->gd_parent, glock_hash(&dep->gd_name));
    if (!tmp_gl)
    {
      gfree(dep, sizeof(gfs_gl_depend_t));
      continue;
    } 

    osi_list_add(&dep->gd_list, &head);

    if (tmp_gl->gl_ops->go_sync)
      tmp_gl->gl_ops->go_sync(sdp, tmp_gl, GFS_SYNC_INVISIBLE);

    gfs_put_glstruct(tmp_gl);
  }


  while (!osi_list_empty(&head))
  {
    dep = osi_list_entry(head.next, gfs_gl_depend_t, gd_list);

    osi_list_del(&dep->gd_list);
    
    tmp_gl = find_glstruct(sdp, &dep->gd_name, dep->gd_parent, glock_hash(&dep->gd_name));
    if (tmp_gl)
    {
      if (tmp_gl->gl_ops->go_sync)
	tmp_gl->gl_ops->go_sync(sdp, tmp_gl, GFS_SYNC_WAIT | GFS_SYNC_INVISIBLE);
      gfs_put_glstruct(tmp_gl);
    }

    gfree(dep, sizeof(gfs_gl_depend_t));
  }


  EXIT(GFN_SYNC_DEPENDENCIES);
}


/**
 * handle_dropped - Handle a dropped glock
 * @sdp: The GFS superblock
 * @gl: The glock which has been dropped
 * @error: The error returned from the locking module
 *
 * Handle a dropped glock - Keep trying until we succeed
 */

STATIC void handle_dropped(gfs_sbd_t *sdp, gfs_glock_t *gl, int error)
{
  ENTER(GFN_HANDLE_DROPPED);

  GFS_ASSERT(error < 0, );

  do
  {
    con_printf("GFS:  error %d from locking module on lock (%"FMT_64"u, %u), retrying...\n",
	       error, gl->gl_name.ln_number, gl->gl_name.ln_type);
    osi_sleep_unint(1);
    DO_PROF(LM_RESET,
	    error = sdp->sd_lockstruct.ls_ops->reset(gl->gl_lock););
  }
  while (error);

  EXIT(GFN_HANDLE_DROPPED);
}


/**
 * cleanup_glock - Prepare a glock to be released
 * @sdp: the filesystem
 * @gl: the glock
 *
 */

STATIC void cleanup_glock(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  gfs_glock_operations_t *glo = gl->gl_ops;
  osi_list_head_t *tmp;
  gfs_glock_t *child;

  ENTER(GFN_CLEANUP_GLOCK);

  if (!osi_list_empty(&gl->gl_depend))
    sync_dependencies(sdp, gl, 0);

  for (tmp = gl->gl_children.next; tmp != &gl->gl_children; tmp = tmp->next)
  {
    child = osi_list_entry(tmp, gfs_glock_t, gl_parlist);

    lock_on_glock(child);

    GFS_ASSERT(!osi_atomic_read(&child->gl_locked), );
    if (child->gl_state != LM_ST_UNLOCKED)
      drop_glock(sdp, child);
    
    unlock_on_glock(child);
  }

  if (glo->go_release)
    glo->go_release(sdp, gl);

  EXIT(GFN_CLEANUP_GLOCK);
}


/**
 * drop_glock - Unlock a glock
 * @sdp: The GFS superblock
 * @gl: The glock to demote
 *
 */

STATIC void drop_glock(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  int ret;

  ENTER(GFN_DROP_GLOCK);


  GFS_ASSERT(gl->gl_state != LM_ST_UNLOCKED, );

  cleanup_glock(sdp, gl);

  IFDB(GFS_LOCKING,
       uint64 blk = gfs_lockname2blk(sdp, &gl->gl_name);
       db_printf("gl->gl_name = (%"FMT_64"u, %u) drop\n",
		 blk, gl->gl_name.ln_type););

  DO_PROF(LM_UNLOCK,
	  ret = sdp->sd_lockstruct.ls_ops->unlock(gl->gl_lock, gl->gl_state););

  IFDB(GFS_LOCKING,
       uint64 blk = gfs_lockname2blk(sdp, &gl->gl_name);
       db_printf("gl->gl_name = (%"FMT_64"u, %u) ret = %d\n",
		 blk, gl->gl_name.ln_type, ret););

  unhold_glock(sdp, gl);

  if (ret)
    handle_dropped(sdp, gl, ret);


  EXIT(GFN_DROP_GLOCK);
}


/**
 * remember_lockstate - Update state of lock
 * @sdp: The GFS superblock
 * @gl: The glock in question
 * @lck_ret: The return from the locking module
 *
 */

STATIC void remember_lockstate(gfs_sbd_t *sdp, gfs_glock_t *gl, int lck_ret)
{
  ENTER(GFN_REMEMBER_LOCKSTATE);

  gl->gl_state = lck_ret & LM_OUT_ST_MASK;

  if (!osi_test_bit(GLF_HELD, &gl->gl_flags) && gl->gl_state != LM_ST_UNLOCKED)
    hold_glock(sdp, gl);
  else if (osi_test_bit(GLF_HELD, &gl->gl_flags) && gl->gl_state == LM_ST_UNLOCKED)
    unhold_glock(sdp, gl);

  if (lck_ret & LM_OUT_NEED_E)
    osi_set_bit(GLF_RELEXCL, &gl->gl_flags);
  if ((lck_ret & LM_OUT_NEED_D))
    osi_set_bit(GLF_RELDFRD, &gl->gl_flags);
  if ((lck_ret & LM_OUT_NEED_S))
    osi_set_bit(GLF_RELSHRD, &gl->gl_flags);

  EXIT(GFN_REMEMBER_LOCKSTATE);
}


/**
 * xmote_glock - Promote a glock
 * @sdp: The GFS superblock
 * @gl: The glock in question
 * @req_state: the requested state
 * @flags: modifier flags to the lock call
 *
 * Returns: Zero on success, or error code
 */

STATIC int xmote_glock(gfs_sbd_t *sdp, gfs_glock_t *gl, unsigned int req_state, int flags)
{
  int lck_flags = 0, lck_ret;
  int error = 0;

  ENTER(GFN_XMOTE_GLOCK);


  GFS_ASSERT(req_state != LM_ST_UNLOCKED, );

  if (gl->gl_state == req_state)
    goto out;


  lck_flags = flags & (LM_FLAG_TRY | LM_FLAG_NOEXP); 


  if (gl->gl_state != LM_ST_UNLOCKED)
    cleanup_glock(sdp, gl);

  IFDB(GFS_LOCKING,
       uint64 blk = gfs_lockname2blk(sdp, &gl->gl_name);
       db_printf("gl->gl_name = (%"FMT_64"u, %u), gl->gl_state = %u, req_state = %u, lck_flags = %d\n",
		 blk, gl->gl_name.ln_type, gl->gl_state, req_state, lck_flags););

  for (;;)
  {
    DO_PROF(LM_LOCK,
	    lck_ret = sdp->sd_lockstruct.ls_ops->lock(gl->gl_lock, gl->gl_state, req_state, lck_flags););
    if (lck_ret >= 0)
      break;

    unhold_glock(sdp, gl);
    handle_dropped(sdp, gl, lck_ret);
  }

  IFDB(GFS_LOCKING,
       uint64 blk = gfs_lockname2blk(sdp, &gl->gl_name);
       db_printf("gl->gl_name = (%"FMT_64"u, %u), gl->gl_state = %u, req_state = %u, lck_flags = %d, lck_ret = %d\n",
		 blk, gl->gl_name.ln_type, gl->gl_state, req_state, lck_flags, lck_ret););

  remember_lockstate(sdp, gl, lck_ret);


  /*  Deal with each possible exit condition  */

  if (lck_ret & LM_OUT_CANCELED)
  {
    error = GLR_CANCELED;
  }
  else if ((lck_ret & LM_OUT_ST_MASK) == req_state)
  {
    /*  Success  */

    if (gl->gl_ops->go_acquire)
      gl->gl_ops->go_acquire(sdp, gl);

    error = 0;
  }
  else if (flags & LM_FLAG_TRY)
  {
    error = GLR_TRYFAILED;
  }
  else
  {
    GFS_ASSERT(0, );
  }


 out:
  EXIT(GFN_XMOTE_GLOCK);

  return error;
}


/**
 * request_glock_wait_or_abort - Test flags, possibly wait for lock
 * @gl: The glock in question
 * @flags: The flags relating to the lock request
 *
 * Returns: 0 if successful, GLR_TRYFAILED otherwise
 */

STATIC int request_glock_wait_or_abort(gfs_sbd_t *sdp, gfs_glock_t *gl, int flags)
{
  osi_wait_vars(wait, "gfs_glock");

  ENTER(GFN_REQUEST_GLOCK_WAIT_OR_ABORT);

 check_flags_and_wait:
  osi_spin_unlock(&gl->gl_head_lock);


  if (flags & LM_FLAG_TRY)
  {
    EXIT(GFN_REQUEST_GLOCK_WAIT_OR_ABORT);
    return GLR_TRYFAILED;
  }
  if (flags & LM_FLAG_NOEXP)
    DO_PROF(LM_CANCEL,
	    sdp->sd_lockstruct.ls_ops->cancel(gl->gl_lock, 1););


  osi_asleep(wait, gl->gl_wait_unlock);
  if (osi_atomic_read(&gl->gl_locked) ||
      (osi_test_bit(GLF_RECOVERY, &gl->gl_flags) && !(flags & LM_FLAG_NOEXP)))
    osi_await(wait);
  osi_unsleep(wait, gl->gl_wait_unlock);


  osi_spin_lock(&gl->gl_head_lock);

  if (osi_test_bit(GLF_RECOVERY, &gl->gl_flags) && !(flags & LM_FLAG_NOEXP))
    goto check_flags_and_wait;
  if (osi_atomic_read(&gl->gl_locked) && (gl->gl_pid != osi_pid()))
    goto check_flags_and_wait;

  osi_atomic_inc(&gl->gl_locked);
  gl->gl_pid = (flags & GL_DISOWN) ? -1 : osi_pid();

  osi_spin_unlock(&gl->gl_head_lock);


  EXIT(GFN_REQUEST_GLOCK_WAIT_OR_ABORT);

  return 0;
}


/**
 * request_glock_ownership - Serialise glock promotion operations
 * @gl: The glock in question
 * @flags: Flags for locking operation
 *
 * This is inline and thus fast. If any special conditions occur, then
 * request_glock_wait_or_abort() is called.
 *
 * Returns: 0 on success, GLR_ error code otherwise
 */

STATIC inline int request_glock_ownership(gfs_sbd_t *sdp, gfs_glock_t *gl, int flags)
{
  int error;

  ENTER(GFN_REQUEST_GLOCK_OWNERSHIP);

  osi_spin_lock(&gl->gl_head_lock);

  if (osi_test_bit(GLF_RECOVERY, &gl->gl_flags) && !(flags & LM_FLAG_NOEXP))
    goto wait;
  if (osi_atomic_read(&gl->gl_locked) && (gl->gl_pid != osi_pid()))
    goto wait;

  osi_atomic_inc(&gl->gl_locked);
  gl->gl_pid = (flags & GL_DISOWN) ? -1 : osi_pid();

  osi_spin_unlock(&gl->gl_head_lock);

  EXIT(GFN_REQUEST_GLOCK_OWNERSHIP);

  return 0;



 wait:
  error = request_glock_wait_or_abort(sdp, gl, flags);

  EXIT(GFN_REQUEST_GLOCK_OWNERSHIP);
  return error;
}


/**
 * gfs_glock - Locks a glock
 * @sdp: The GFS superblock
 * @gl: The glock to lock
 * @flags: flags
 *
 * This routine locks a glock.  Allows Glocks to be recursively held by the
 * same process.  If a different process tries to obtain a lock that is already
 * held, it sleeps.
 *
 * Returns: 0 on success, -EXXXX on failure
 */

int gfs_glock(gfs_sbd_t *sdp, gfs_glock_t *gl, int flags)
{
  gfs_glock_operations_t *glops = gl->gl_ops;
  unsigned int req_state;
  int error;

  ENTER(GFN_GLOCK);


  GFS_ASSERT(gl && osi_atomic_read(&gl->gl_count), );

  if (gl->gl_parent)
  {
    osi_spin_lock(&gl->gl_parent->gl_head_lock);
    GFS_ASSERT(osi_atomic_read(&gl->gl_parent->gl_locked) &&
	       gl->gl_parent->gl_pid == osi_pid(), );
    osi_spin_unlock(&gl->gl_parent->gl_head_lock);
  }

  switch (flags & (GL_SHARED | GL_DEFERRED))
  {
  case 0:
    req_state = LM_ST_EXCLUSIVE;
    break;
  case GL_SHARED:
    req_state = LM_ST_SHARED;
    break;
  case GL_DEFERRED:
    req_state = LM_ST_DEFERRED;
    break;
  case (GL_SHARED | GL_DEFERRED):
  default:
    GFS_ASSERT(0, );
  }

  if (flags & LM_FLAG_NOEXP)
    osi_set_bit(GLF_RECOVERY, &gl->gl_flags);


 try_again:
  error = request_glock_ownership(sdp, gl, flags);
  if (error)
    goto out;


  lock_on_glock(gl);

  if (flags & LM_FLAG_NOEXP)
    DO_PROF2(LM_CANCEL,
	     sdp->sd_lockstruct.ls_ops->cancel(gl->gl_lock, 0););

  if (!RELAXED_LOCKMODE_OK(req_state, gl->gl_state))
  {
    GFS_ASSERT(osi_atomic_read(&gl->gl_locked) == 1, );

    error = xmote_glock(sdp, gl, req_state, flags);
    if (error)
      goto do_error;
  }

  if (flags & GL_PERM)
    perm_glock(sdp, gl);

  if (glops->go_lock)
  {
    error = glops->go_lock(sdp, gl, flags);
    if (error)
      goto do_error;
  }

  unlock_on_glock(gl);


 out:
  if (flags & LM_FLAG_NOEXP)
    osi_clear_bit(GLF_RECOVERY, &gl->gl_flags);

  EXIT(GFN_GLOCK);

  return error;



 do_error:
  osi_atomic_dec(&gl->gl_locked);
  unlock_on_glock(gl);
  osi_wakeup(gl->gl_wait_unlock);

  if (error == GLR_CANCELED)
    goto try_again;

  goto out;
}


/**
 * gfs_gunlock - Unlocks a glock
 * @sdp: The GFS superblock
 * @gl: The glock in question
 * @flags: %GL_NOCACHE to prevent caching of the lock, %GL_SYNC to sync the lock
 *
 */

void gfs_gunlock(gfs_sbd_t *sdp, gfs_glock_t *gl, int flags)
{
  gfs_glock_operations_t *glops = gl->gl_ops;
  int need_wake_up = FALSE;

  ENTER(GFN_GUNLOCK);


  GFS_ASSERT(gl && osi_atomic_read(&gl->gl_count), );

  if (gl->gl_parent)
  {
    osi_spin_lock(&gl->gl_parent->gl_head_lock);
    GFS_ASSERT(osi_atomic_read(&gl->gl_parent->gl_locked) &&
	       gl->gl_parent->gl_pid == osi_pid(), );
    osi_spin_unlock(&gl->gl_parent->gl_head_lock);
  }

  if (glops->go_unlock)
    glops->go_unlock(sdp, gl, flags);


  lock_on_glock(gl);


  if (flags & GL_SYNC)
    osi_set_bit(GLF_SYNC, &gl->gl_flags);

  osi_spin_lock(&gl->gl_head_lock);
  GFS_ASSERT(osi_atomic_read(&gl->gl_locked), );
  GFS_ASSERT(gl->gl_pid == osi_pid() || gl->gl_pid == -1, );
  osi_spin_unlock(&gl->gl_head_lock);

  if (osi_atomic_dec_and_test(&gl->gl_locked))
  {
    need_wake_up = TRUE;

    if (osi_test_bit(GLF_PERM, &gl->gl_flags))
      unperm_glock(sdp, gl);

    if ((flags & GL_NOCACHE) || osi_test_and_clear_bit(GLF_RELEXCL, &gl->gl_flags))
      drop_glock(sdp, gl);
    else if (osi_test_and_clear_bit(GLF_RELSHRD, &gl->gl_flags))
      xmote_glock(sdp, gl, LM_ST_SHARED, LM_FLAG_TRY);
    else if (osi_test_and_clear_bit(GLF_RELDFRD, &gl->gl_flags))
      xmote_glock(sdp, gl, LM_ST_DEFERRED, LM_FLAG_TRY);
    else if (osi_test_bit(GLF_DIRTY, &gl->gl_flags))
    {
      GFS_ASSERT(gl->gl_state == LM_ST_EXCLUSIVE, );

      if (osi_test_bit(GLF_SYNC, &gl->gl_flags) || sdp->sd_sync_mount)
	if (glops->go_sync)
	  glops->go_sync(sdp, gl, GFS_SYNC_WAIT);
    }

    osi_time_stamp(&gl->gl_stamp);
  }

  unlock_on_glock(gl);


  /*
   * Don't wake up waiting processes until after we've cleared the
   * GLF_LOCK bit to avoid a process being woken up
   * after waiting on the gl_wait_unlock queue only to sleep again
   * on the gl_wait_lock wait queue.
   */

  if (need_wake_up)
    osi_wakeup(gl->gl_wait_unlock);

  EXIT(GFN_GUNLOCK);
}


/**
 * gfs_force_gl_drop - make sure that this machine isn't holding a lock
 * @sdp: the filesystem
 * @gl: the glock
 *
 * You would never call this while holding @gl.
 */

void gfs_force_gl_drop(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  int error;

  ENTER(GFN_FORCE_GL_DROP);

  error = request_glock_ownership(sdp, gl, 0);
  GFS_ASSERT(!error, );

  GFS_ASSERT(osi_atomic_read(&gl->gl_locked) == 1, );

  lock_on_glock(gl);

  if (gl->gl_state != LM_ST_UNLOCKED)
    drop_glock(sdp, gl);

  osi_atomic_dec(&gl->gl_locked);
  unlock_on_glock(gl);
  osi_wakeup(gl->gl_wait_unlock);

  EXIT(GFN_FORCE_GL_DROP);
}


/**
 * gfs_glock_num - Acquire a glock given its number
 * @sdp: The GFS superblock
 * @number: The lock number
 * @gops: The operations for the object covered by the lock
 * @parent: The parent lock of this lock
 * @flags: flags
 * @gl: Used to return a pointer to lock if successful
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_glock_num(gfs_sbd_t *sdp,
		  uint64 number, gfs_glock_operations_t *gops, gfs_glock_t *parent,
		  int flags, gfs_glock_t **gl)
{
  gfs_glock_t *tmp_gl;
  int error;

  ENTER(GFN_GLOCK_NUM);


  tmp_gl = gfs_get_glstruct(sdp, number, gops, parent, CREATE);
  if (!tmp_gl)
  {
    EXIT(GFN_GLOCK_NUM);
    return -ENOMEM;
  }

  error = gfs_glock(sdp, tmp_gl, flags);
  if (error)
  {
    gfs_put_glstruct(tmp_gl);
    EXIT(GFN_GLOCK_NUM);
    return error;
  }

  if (gl)
    *gl = tmp_gl;
  else
    gfs_put_glstruct(tmp_gl);


  EXIT(GFN_GLOCK_NUM);

  return 0;
}


/**
 * gfs_gunlock_num - Unlock a glock given its number
 * @sdp: The GFS superblock
 * @number: The lock number
 * @gops: The operations for the object covered by the lock
 * @parent: The parent lock of this lock
 * @flags: %GL_INCR to increment the version number, %GL_SYNC to sync lock
 *
 */

void gfs_gunlock_num(gfs_sbd_t *sdp,
		     uint64 number, gfs_glock_operations_t *gops, gfs_glock_t *parent,
		     int flags)
{
  gfs_glock_t *gl;

  ENTER(GFN_GUNLOCK_NUM);


  gl = gfs_get_glstruct(sdp, number, gops, parent, NO_CREATE);
  GFS_ASSERT(gl, );

  gfs_gunlock(sdp, gl, flags);

  gfs_put_glstruct(gl);


  EXIT(GFN_GUNLOCK_NUM);
}


/**
 * gfs_glock_m - Lock a bunch of glocks
 * @sdp: The GFS superblock
 * @num_lock: The number of locks
 * @ops: A vector containing pointers to the locks to be locked
 *
 * Acquire a bunch of locks in ascending numerical order
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_glock_m(gfs_sbd_t *sdp, int num_lock, gfs_lockop_t *ops)
{
  gfs_lockop_t *point[num_lock], *tmp;
  lm_lockname_t *name1, *name2;
  int x, y;
  int changed;
  int error = 0;

  ENTER(GFN_GLOCK_M);


  for (x = 0; x < num_lock; x++)
  {
    point[x] = &ops[x];
    if (point[x]->lo_ip)
      point[x]->lo_lock = point[x]->lo_ip->i_gl;
  }


  /*  Sort the locks in ascending numerical order.

      If there are locks with the same lock number, sort them so that any
      exclusive locks come before shared locks  */

  do
  {
    changed = FALSE;

    for (x = 0; x < num_lock - 1; x++)
    {
      name1 = &point[x]->lo_lock->gl_name;
      name2 = &point[x + 1]->lo_lock->gl_name;

      if (name1->ln_number < name2->ln_number)
	continue;

      if (name1->ln_number == name2->ln_number)
      {
	if (name1->ln_type < name2->ln_type)
	  continue;

	if (name1->ln_type == name2->ln_type)
	{
	  GFS_ASSERT(!((point[x]->lo_flags & GL_DEFERRED) || (point[x + 1]->lo_flags & GL_DEFERRED)), );
	  if (!((point[x]->lo_flags & GL_SHARED) && !(point[x + 1]->lo_flags & GL_SHARED)))
	    continue;
	}
      }

      tmp = point[x];
      point[x] = point[x + 1];
      point[x + 1] = tmp;
      changed = TRUE;
    }
  }
  while (changed);



  for (x = 0; x < num_lock; x++)
  {
    if (point[x]->lo_ip)
      error = gfs_glock_i(point[x]->lo_ip, point[x]->lo_flags);
    else
      error = gfs_glock(sdp, point[x]->lo_lock, point[x]->lo_flags);

    if (error)
      break;
  } 


  if (x < num_lock)
    for (y = 0; y < x; y++)
    {
      if (point[y]->lo_ip)
	gfs_gunlock_i(point[y]->lo_ip, point[y]->lo_flags);
      else
	gfs_gunlock(sdp, point[y]->lo_lock, point[y]->lo_flags);
    }


  EXIT(GFN_GLOCK_M);

  return error;
}


/**
 * gfs_gunlock_m - Unlock a bunch of glocks
 * @sdp: The GFS superblock
 * @num_lock: The number of locks to unlock
 * @ops: A vector containing pointers to the locks to be locked
 *
 */

void gfs_gunlock_m(gfs_sbd_t *sdp, int num_lock, gfs_lockop_t *ops)
{
  int x;

  ENTER(GFN_GUNLOCK_M);


  for (x = 0; x < num_lock; x++)
  {
    if (ops[x].lo_ip)
      gfs_gunlock_i(ops[x].lo_ip, ops[x].lo_flags);
    else
      gfs_gunlock(sdp, ops[x].lo_lock, ops[x].lo_flags);
  }


  EXIT(GFN_GUNLOCK_M);
}


/**
 * find_held_glstruct - find the held lock with a given lockname
 * @sdp: the filesystem
 * @name: the lock name
 *
 * This differs from find_glstruct() in that it:
 * 1) Identifies and acquires the hash bucket spinlock
 * 2) Doesn't look at glocks that aren't held
 * 3) Doesn't care about what a glock's parent is
 *
 * Returns: The glock or NULL
 */

STATIC gfs_glock_t *find_held_glstruct(gfs_sbd_t *sdp, lm_lockname_t *name)
{
  osi_list_head_t *tmp, *head;
  gfs_glock_t *gl;
  uint32 bucket = glock_hash(name);

  ENTER(GFN_FIND_HELD_GLSTRUCT);

  osi_read_lock(&sdp->sd_ghash[bucket].hb_lock);


  tmp = head = &sdp->sd_ghash[bucket].hb_held;
  for (;;)
  {
    tmp = tmp->next;
    gl = NULL;
    if (tmp == head)
      break;
    gl = osi_list_entry(tmp, gfs_glock_t, gl_list);

    if (!lm_name_equal(&gl->gl_name, name))
      continue;
    gfs_hold_glstruct(gl);
    goto out;
  }

  tmp = head = &sdp->sd_ghash[bucket].hb_perm;
  for (;;)
  {
    tmp = tmp->next;
    gl = NULL;
    if (tmp == head)
      break;
    gl = osi_list_entry(tmp, gfs_glock_t, gl_list);

    if (!lm_name_equal(&gl->gl_name, name))
      continue;
    gfs_hold_glstruct(gl);
    break;
  }


 out:
  osi_read_unlock(&sdp->sd_ghash[bucket].hb_lock);

  EXIT(GFN_FIND_HELD_GLSTRUCT);

  return gl;
}


/**
 * gfs_glock_cb - Callback used by locking module
 * @fsdata: Pointer to the superblock
 * @type: Type of callback
 * @data: Type dependent data pointer
 *
 * Called by the locking module when it wants to tell us something.
 * Either we need to drop a lock or another client expired.
 */

void gfs_glock_cb(lm_fsdata_t *fsdata, unsigned int type, void *data)
{
  gfs_sbd_t *sdp = (gfs_sbd_t *)fsdata;
  unsigned int journal;
  lm_lockname_t name;
  gfs_glock_t *gl;

  ENTER(GFN_GLOCK_CB);


  switch (type)
  {
  case LM_CB_EXPIRED:
    journal = *(unsigned int *)data;

    IFDB(GFS_LOCKING,
	 db_printf("  LM_CB_EXPIRED %u\n", journal););

    if (journal >= sdp->sd_journals)
      break;

    osi_set_bit(journal, sdp->sd_dirty_j);
    if (sdp->sd_recoverd_process)
      osi_wake_up_process(sdp->sd_recoverd_process);

    break;


  case LM_CB_NEED_E:
    name = *(lm_lockname_t *)data;

    IFDB(GFS_LOCKING,
	 uint64 blk = gfs_lockname2blk(sdp, &name);
	 db_printf("  LM_CB_NEED_E (%"FMT_64"u, %u)\n", blk, name.ln_type););

    gl = find_held_glstruct(sdp, &name);
    if (!gl)
      break;

    osi_set_bit(GLF_RELEXCL, &gl->gl_flags);

    if (trylock_on_glock(gl))
    {
      if (!osi_atomic_read(&gl->gl_locked) && gl->gl_state != LM_ST_UNLOCKED)
	drop_glock(sdp, gl);
      unlock_on_glock(gl);
    }

    put_glstruct(gl);

    break;


  case LM_CB_NEED_D:
    name = *(lm_lockname_t *)data;

    IFDB(GFS_LOCKING,
	 uint64 blk = gfs_lockname2blk(sdp, &name);
	 db_printf("  LM_CB_NEED_D (%"FMT_64"u, %u)\n", blk, name.ln_type););

    gl = find_held_glstruct(sdp, &name);
    if (!gl)
      break;

    osi_set_bit(GLF_RELDFRD, &gl->gl_flags);

    if (trylock_on_glock(gl))
    {
      if (!osi_atomic_read(&gl->gl_locked) && gl->gl_state != LM_ST_UNLOCKED)
	xmote_glock(sdp, gl, LM_ST_DEFERRED, LM_FLAG_TRY);
      unlock_on_glock(gl);
    }

    put_glstruct(gl);

    break;


  case LM_CB_NEED_S:
    name = *(lm_lockname_t *)data;

    IFDB(GFS_LOCKING,
	 uint64 blk = gfs_lockname2blk(sdp, &name);
	 db_printf("  LM_CB_NEED_S (%"FMT_64"u, %u)\n", blk, name.ln_type););

    gl = find_held_glstruct(sdp, &name);
    if (!gl)
      break;

    osi_set_bit(GLF_RELSHRD, &gl->gl_flags);

    if (trylock_on_glock(gl))
    {
      if (!osi_atomic_read(&gl->gl_locked) && gl->gl_state != LM_ST_UNLOCKED)
	xmote_glock(sdp, gl, LM_ST_SHARED, LM_FLAG_TRY);
      unlock_on_glock(gl);
    }

    put_glstruct(gl);

    break;


  case LM_CB_DROPLOCKS:
    IFDB(GFS_LOCKING,
	 db_printf("  LM_CB_DROPLOCKS\n"););

    gfs_drop_excess_inodes(sdp);
    gfs_pitch_inodes(sdp);
    gfs_glockd_scan(sdp, 0);

    break;


  default:
    GFS_ASSERT(0, con_printf("type = %u\n", type););
    break;
  }


  EXIT(GFN_GLOCK_CB);
}


/**
 * scan_cached_inodes() - Scans list of inodes to see if any can be freed
 * @sdp: The GFS superblock
 * @list: The list of inodes
 * @i: The hash bucket containing the glock in question
 *
 * Scans inodes on each glock to see if any can be freed.
 * If the glock is currently being locked by something,
 * continue on our way.
 *
 * Returns: the number of existing inodes +
 *          the number of glocks that couldn't be examined
 */

STATIC int scan_cached_inodes(gfs_sbd_t *sdp, osi_list_head_t *list, int i)
{
  gfs_glock_t *gl;
  gfs_inode_t *ip;
  osi_list_head_t *tmp;
  int missed;

  ENTER(GFN_SCAN_CACHED_INODES);


 restart:
  missed = 0;

  osi_read_lock(&sdp->sd_ghash[i].hb_lock);

  for (tmp = list->next; tmp != list; tmp = tmp->next)
  {
    gl = osi_list_entry(tmp, gfs_glock_t, gl_list);

    if (gl->gl_ops != &gfs_inode_glops)
      continue;

    if (!trylock_on_glock(gl))
    {
      missed++;
      continue;
    }

    if (osi_atomic_read(&gl->gl_locked))
    {
      unlock_on_glock(gl);
      missed++;
      continue;
    }

    ip = (gfs_inode_t *)gl->gl_object;
    if (!ip)
    {
      unlock_on_glock(gl);
      continue;
    }

    if (osi_atomic_read(&ip->i_count))
    {
      unlock_on_glock(gl);
      missed++;
      continue;
    }

    gfs_hold_glstruct(gl);

    osi_read_unlock(&sdp->sd_ghash[i].hb_lock);


    /*  Get rid of the inode  */

    gfs_destroy_istruct(ip);


    unlock_on_glock(gl);
    put_glstruct(gl);

    goto restart;
  }

  osi_read_unlock(&sdp->sd_ghash[i].hb_lock);


  EXIT(GFN_SCAN_CACHED_INODES);

  return missed;
}


/**
 * gfs_pitch_inodes - Throw out any inodes that are flagged to be thrown out
 * @sdp: the filesystem
 *
 */

void gfs_pitch_inodes(gfs_sbd_t *sdp)
{
  int i;

  ENTER(GFN_PITCH_INODES);

  if (osi_check_timeout(&sdp->sd_shrink_time, 60))
  {
    gfs_drop_excess_inodes(sdp);
    osi_time_stamp(&sdp->sd_shrink_time);
  }

  osi_atomic_set(&sdp->sd_free_icount, 0);

  for (i = 0; i < GFS_GHASH_SIZE; i++)
  {
    scan_cached_inodes(sdp, &sdp->sd_ghash[i].hb_held, i);
    scan_cached_inodes(sdp, &sdp->sd_ghash[i].hb_notheld, i);
  }

  EXIT(GFN_PITCH_INODES);
}


/**
 * scan_held_glocks() - Scans held glocks looking for demotion candidates
 * @sdp: The GFS superblock
 * @bucket: The hash bucket to scan
 * @timeout: The minimum age of a lock before we may demote it
 *
 * Scans held glocks and tries to demote any which are currently unused.
 * Only works over a single hash bucket.
 */

STATIC void scan_held_glocks(gfs_sbd_t *sdp, unsigned int bucket, int timeout)
{
  osi_list_head_t *head, *tmp;
  gfs_glock_t *gl;
  int action = 0;

  ENTER(GFN_SCAN_HELD_GLOCKS);


  /*  Set flags to check these locks for old dependencies  */

  osi_read_lock(&sdp->sd_ghash[bucket].hb_lock);

  head = &sdp->sd_ghash[bucket].hb_held;
  for (tmp = head->next; tmp != head; tmp = tmp->next)
  {
    gl = osi_list_entry(tmp, gfs_glock_t, gl_list);
    if (osi_atomic_read(&gl->gl_num_dep))
      osi_set_bit(GLF_DEPCHECK, &gl->gl_flags);
  }

  osi_read_unlock(&sdp->sd_ghash[bucket].hb_lock);


  do
  {
    osi_read_lock(&sdp->sd_ghash[bucket].hb_lock);

    tmp = head = &sdp->sd_ghash[bucket].hb_held;
    for (;;)
    {
      gl = NULL;
      tmp = tmp->prev;
      if (tmp == head)
	break;
      gl = osi_list_entry(tmp, gfs_glock_t, gl_list);

      if (!trylock_on_glock(gl))
	continue;
      
      if (!osi_atomic_read(&gl->gl_locked))
      {
	if ((osi_check_timeout(&gl->gl_stamp, timeout) ||
	     osi_test_bit(GLF_DEMOTEME, &gl->gl_flags)) &&
	    !osi_test_bit(GLF_STICKY, &gl->gl_flags))
	{
	  action = 1;
	  gfs_hold_glstruct(gl);
	  break;
	}
	if (osi_test_bit(GLF_DEPCHECK, &gl->gl_flags))
	{
	  action = 2;
	  gfs_hold_glstruct(gl);
	  break;
	}
      }

      unlock_on_glock(gl);
    }

    osi_read_unlock(&sdp->sd_ghash[bucket].hb_lock);


    if (gl)
    {
      switch (action)
      {
      case 1:
	drop_glock(sdp, gl);
	break;

      case 2:
	sync_dependencies(sdp, gl, (osi_atomic_read(&gl->gl_num_dep) > 100) ? 1 : 30);
	break;

      default:
	GFS_ASSERT(0, );
      }

      osi_clear_bit(GLF_DEPCHECK, &gl->gl_flags);

      unlock_on_glock(gl);
      put_glstruct(gl);
    }
  }
  while (gl);


  EXIT(GFN_SCAN_HELD_GLOCKS);
}


/**
 * gfs_glockd_scan - The architecture independant part of gfs_glockd
 * @sdp: The GFS superblock
 *
 * The architecture independant part of gfs_glockd. It does two
 * scans through the glocks, one for each hash queue. 
 *
 * Pass I: Look to see if any gfs_inode_t can be closed or
 *         deallocated.
 * Pass II: Look to see if there are any locked glocks which
 *          are no longer needed and are candidates for
 *          demotion.
 * Pass III: Look to see if there are any unlocked glocks which
 *           are spare and can thus be freed to the system.
 *
 */

void gfs_glockd_scan(gfs_sbd_t *sdp, unsigned int timeout)
{
  gfs_glock_t *gl;
  OSI_LIST_HEAD(dead);
  osi_list_head_t *head, *prev;
  int i;

  ENTER(GFN_GLOCKD_SCAN);


  IFDB(GFN_GLOCKD_SCAN,
       for (i = 0; i < GFS_GHASH_SIZE; i++)
	 db_printf("bucket %d: perm = %d, held = %d, notheld = %d\n",
		   i,
		   osi_atomic_read(&sdp->sd_ghash[i].hb_perm_count),
		   osi_atomic_read(&sdp->sd_ghash[i].hb_held_count),
		   osi_atomic_read(&sdp->sd_ghash[i].hb_notheld_count)););


  gfs_pitch_inodes(sdp);


  for (i = 0; i < GFS_GHASH_SIZE; i++)
  {
    /* Part I: Read lock */

    scan_held_glocks(sdp, i, timeout);


    /* Part II: Write lock */

    osi_write_lock(&sdp->sd_ghash[i].hb_lock);

    head = &sdp->sd_ghash[i].hb_notheld;
    prev = head->prev;

    for (;;)
    {
      osi_list_head_t *tmp = prev;

      prev = prev->prev;
      if (tmp == head)
	break;
      gl = osi_list_entry(tmp, gfs_glock_t, gl_list);

      if (osi_atomic_read(&gl->gl_count))
	continue;

      osi_list_del(&gl->gl_list);
      osi_list_add(&gl->gl_list, &dead);
      osi_atomic_dec(&sdp->sd_ghash[i].hb_notheld_count);
    }

    osi_write_unlock(&sdp->sd_ghash[i].hb_lock);


    while (!osi_list_empty(&dead))
    {
      gl = osi_list_entry(dead.next, gfs_glock_t, gl_list);

      osi_list_del(&gl->gl_list);
      osi_list_head_init(&gl->gl_list);

      if (gl->gl_parent)
      {
	lock_on_glock(gl->gl_parent);
	osi_list_del(&gl->gl_parlist);
	unlock_on_glock(gl->gl_parent);
	put_glstruct(gl->gl_parent);

	osi_list_head_init(&gl->gl_parlist);
      }

      release_glstruct(sdp, gl);
    }
  }


  EXIT(GFN_GLOCKD_SCAN);
}


/**
 * gfs_clear_gla - Called on umount to clear all glocks
 * @sdp: The GFS superblock
 *
 */

void gfs_clear_gla(gfs_sbd_t *sdp)
{
  gfs_glock_t *gl;
  osi_list_head_t *head, *tmp;
  int i;
  int cont;
  gfs_glock_operations_t *glo;

  ENTER(GFN_CLEAR_GLA);


  for (i = 0; i < GFS_GHASH_SIZE; i++)
  {
    while (scan_cached_inodes(sdp, &sdp->sd_ghash[i].hb_held, i))
    {
      osi_schedule();
    }
    while (scan_cached_inodes(sdp, &sdp->sd_ghash[i].hb_notheld, i))
    {
      osi_schedule();
    }
  }


  for (i = 0; i < GFS_GHASH_SIZE; i++)
  {
    osi_read_lock(&sdp->sd_ghash[i].hb_lock);
    GFS_ASSERT(osi_list_empty(&sdp->sd_ghash[i].hb_perm), );
    osi_read_unlock(&sdp->sd_ghash[i].hb_lock);
  }


  for (i = 0; i < GFS_GHASH_SIZE; i++)
  {
  restart_held:
    osi_read_lock(&sdp->sd_ghash[i].hb_lock);

    tmp = head = &sdp->sd_ghash[i].hb_held;
    for (;;)
    {
      gl = NULL;
      tmp = tmp->next;
      if (tmp == head)
	break;
      gl = osi_list_entry(tmp, gfs_glock_t, gl_list);

      if (!trylock_on_glock(gl))
      {
	osi_set_bit(GLF_RELEXCL, &gl->gl_flags);
	continue;
      }

      gfs_hold_glstruct(gl);

      break;
    }

    osi_read_unlock(&sdp->sd_ghash[i].hb_lock);

    if (gl)
    {
      GFS_ASSERT(!osi_atomic_read(&gl->gl_locked), );

      drop_glock(sdp, gl);
      unlock_on_glock(gl);
      put_glstruct(gl);

      goto restart_held;
    }

    /*  Make sure all locks are dropped  */

    if (osi_atomic_read(&sdp->sd_ghash[i].hb_held_count))
    {
      osi_schedule();
      goto restart_held;
    }
  }


  for (;;)
  {
    cont = FALSE;

    for (i = 0; i < GFS_GHASH_SIZE; i++)
    {
    restart_notheld:
      osi_write_lock(&sdp->sd_ghash[i].hb_lock);

      tmp = head = &sdp->sd_ghash[i].hb_notheld;
      for (;;)
      {
	gl = NULL;
	tmp = tmp->next;
	if (tmp == head)
	  break;
	gl = osi_list_entry(tmp, gfs_glock_t, gl_list);

	if (!trylock_on_glock(gl))
	  continue;

	GFS_ASSERT(!osi_atomic_read(&gl->gl_locked), );
	GFS_ASSERT(!gl->gl_object, );

	/*  We probably shouldn't call go_free() in while holding hb_lock.  */
        glo = gl->gl_ops;
	if (glo->go_free)
          glo->go_free(sdp, gl);

	if (osi_atomic_read(&gl->gl_count))
	{
	  unlock_on_glock(gl);
	  continue;
	}

	osi_list_del(&gl->gl_list);
	osi_atomic_dec(&sdp->sd_ghash[i].hb_notheld_count);

	break;
      }

      osi_write_unlock(&sdp->sd_ghash[i].hb_lock);

      if (gl)
      {
	unlock_on_glock(gl);
	osi_list_head_init(&gl->gl_list);

	if (gl->gl_parent)
	{
          lock_on_glock(gl->gl_parent);
	  osi_list_del(&gl->gl_parlist);
          unlock_on_glock(gl->gl_parent);
	  put_glstruct(gl->gl_parent);

	  osi_list_head_init(&gl->gl_parlist);
	}

	release_glstruct(sdp, gl);

	goto restart_notheld;
      }

      if (osi_atomic_read(&sdp->sd_ghash[i].hb_notheld_count))
	cont = TRUE;
    }

    if (!cont)
      break;

    osi_schedule();
  }


  EXIT(GFN_CLEAR_GLA);
}





/*
 *  Diagnostic routines to help debug distributed deadlock
 */


/**
 * dump_bucket - print out the lockstate for a given hash bucket
 * @sdp: the filesystem
 * @list: the bucket
 * @i: the bucket number
 *
 */
STATIC void dump_bucket(gfs_sbd_t *sdp, osi_list_head_t *list, int i)
{
  gfs_glock_t *gl;
  osi_list_head_t *tmp;
  uint64 blk;
  int locked;
  unsigned int try = 0;

  ENTER(GFN_DUMP_BUCKET);


  osi_read_lock(&sdp->sd_ghash[i].hb_lock);
  for (tmp = list->next; tmp != list; tmp = tmp->next)
  {
    gl = osi_list_entry(tmp, gfs_glock_t, gl_list);

    gfs_hold_glstruct(gl);
    osi_read_unlock(&sdp->sd_ghash[i].hb_lock);

    blk = gfs_lockname2blk(sdp, &gl->gl_name);
    locked = trylock_on_glock(gl);

    db_printf("(%u, %"FMT_64"u):  count=%d, flags=%d, sema=%s, locked=%d, pid=%ld, state=%u\n",
             gl->gl_name.ln_type, blk,
             osi_atomic_read(&gl->gl_count) - 1,
             gl->gl_flags,
             (locked) ? "U" : "L",
             osi_atomic_read(&gl->gl_locked),
             gl->gl_pid,
             gl->gl_state);
             
    if (locked)
      unlock_on_glock(gl);

    osi_read_lock(&sdp->sd_ghash[i].hb_lock);
    put_glstruct(gl);

    if (try++ > 10000)
    {
      db_printf("dammit!\n");
      break;
    }
  }

  osi_read_unlock(&sdp->sd_ghash[i].hb_lock);


  EXIT(GFN_DUMP_BUCKET);
}



/**
 * gfs_dump_lockstate - print out the current lockstate
 * @sdp: the filesystem
 *
 */
void gfs_dump_lockstate(gfs_sbd_t *sdp)
{
  int i;

  ENTER(GFN_DUMP_LOCKSTATE);

  db_printf("-----------  Start of lockstate dump  -----------\n");
  for (i = 0; i < GFS_GHASH_SIZE; i++)
  {
    dump_bucket(sdp, &sdp->sd_ghash[i].hb_perm, i);
    osi_schedule();
    dump_bucket(sdp, &sdp->sd_ghash[i].hb_held, i);
    osi_schedule();
    dump_bucket(sdp, &sdp->sd_ghash[i].hb_notheld, i);
    osi_schedule();
  }
  db_printf("-----------  End of lockstate dump  -----------\n");

  EXIT(GFN_DUMP_LOCKSTATE);
}




