/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portions Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: glock.h
**
**    Description:
**	This module implements the GFS interface to the global locks.
**
*******************************************************************************
******************************************************************************/

#ifndef __GFS_GLOCK_DOT_H__
#define __GFS_GLOCK_DOT_H__


/*  Flags passed in to gfs_glock().  These are passed down to the
    lock module in one way or another.

    GL_SHARED means that GFS is giving permission for the lock layers to acquire
    the lock in a shared state.

                                  GL_SHARED
      Lock actually
      acquired as     |      Set      |      Not set
    ------------------+---------------+--------------------
         Shared       |      GOOD     |       BAD!!!!
    ------------------+---------------+--------------------
         Exclusive    |       OK      |       GOOD 

    GL_DEFERRED asks for the lock in the deferred state.  It's mutual exclusive
    with GL_SHARED.  If a lock is asked for in the deferred state, it must
    be acquired in the deferred state.

    GL_PERM lets the lock layer know that this lock will be held for a long time.

    GL_DISOWN tells the lock layer to disallow recursive locking and allows a
    different process to unlock the lock.

    GL_SKIP tells the go_lock() and go_unlock() calls to do nothing

    LM_FLAG_NOEXP means that if this lock is expired, ignore that fact -- Treat the 
    lock exactly as it would if it wasn't expired.  (And the lock
    module shouldn't send GFS callbacks about finding this lock expired.)
    (This flag is probably ignored for a DLM-style lock module.)

    LM_FLAG_TRY means that the locking layers should only try to acquire the lock
    once.  If the lock can't be acquired immediately give up and return
    GLR_TRYFAILED.

    */

#define GL_SHARED         (0x00001000)
#define GL_DEFERRED       (0x00002000)
#define GL_PERM           (0x00004000)
#define GL_DISOWN         (0x00008000)
#define GL_SKIP           (0x00010000)
/*  Also LM_FLAG_TRY and LM_FLAG_NOEXP.  */


/*  Potential errors returned by gfs_glock() when used with some particular flags.
    GLR_CANCELED is never seen outside of glock.c  */

#define GLR_TRYFAILED (13)
#define GLR_CANCELED  (14)


/*  Flags passed in to gfs_gunlock()

    GL_SYNC causes all the data and metadata for this lock to be synced to
    disk before its released.  It doesn't do anything with transactions
    that have been commited incore.  In fact it's really dangerous to use
    this flag.  If you don't know why don't use it.

    GL_NOCACHE causes the lock layers to drop the lock immediately after
    it is unlocked.  It isn't cache (held) by the lock layers.

    */

#define GL_SYNC       (0x00100000)
#define GL_NOCACHE    (0x00200000)



struct gfs_lockop
{
  gfs_glock_t         *lo_lock;
  gfs_inode_t         *lo_ip;
  int                 lo_flags;
};
typedef struct gfs_lockop gfs_lockop_t;



/*  Lockname Functions  */

void gfs_blk2lockname(gfs_sbd_t *sdp, uint64 block, unsigned int type, lm_lockname_t *name);
uint64 gfs_lockname2blk(gfs_sbd_t *sdp, lm_lockname_t *name);


/*  Glock structure functions  */

gfs_glock_t *gfs_get_glstruct(gfs_sbd_t *sdp,
			      uint64 number, gfs_glock_operations_t *glops, gfs_glock_t *parent,
			      int create);
static __inline__ void gfs_hold_glstruct(gfs_glock_t *gl)
{
  osi_atomic_inc(&gl->gl_count);
}
void gfs_put_glstruct(gfs_glock_t *gl);


/*  Lock acquisition and release  */

int gfs_glock(gfs_sbd_t *sdp, gfs_glock_t *gl, int flags);
void gfs_gunlock(gfs_sbd_t *sdp, gfs_glock_t *gl, int flags);
void gfs_force_gl_drop(gfs_sbd_t *sdp, gfs_glock_t *gl);

#define gfs_glock_i(ip, flags) gfs_glock((ip)->i_sbd, (ip)->i_gl, (flags))
#define gfs_gunlock_i(ip, flags) gfs_gunlock((ip)->i_sbd, (ip)->i_gl, (flags))

#define gfs_glock_rg(rgd, flags) gfs_glock((rgd)->rd_sbd, (rgd)->rd_gl, (flags))
#define gfs_gunlock_rg(rgd, flags) gfs_gunlock((rgd)->rd_sbd, (rgd)->rd_gl, (flags))

int gfs_glock_num(gfs_sbd_t *sdp,
		  uint64 number, gfs_glock_operations_t *glops, gfs_glock_t *parent,
		  int flags,
		  gfs_glock_t **gl);
void gfs_gunlock_num(gfs_sbd_t *sdp,
		     uint64 number, gfs_glock_operations_t *glops, gfs_glock_t *parent,
		     int flags);

int gfs_glock_m(gfs_sbd_t *sdp, int num_lock, gfs_lockop_t *ops);
void gfs_gunlock_m(gfs_sbd_t *sdp, int num_lock, gfs_lockop_t *ops);


/*  Lock Value Block functions  */

int gfs_hold_lvb(gfs_sbd_t *sdp, gfs_glock_t *gl);
void gfs_unhold_lvb(gfs_sbd_t *sdp, gfs_glock_t *gl);
void gfs_sync_lvb(gfs_sbd_t *sdp, gfs_glock_t *gl);


/*  Inter-glock dependencies  */

int gfs_add_gl_dependency(gfs_sbd_t *sdp, gfs_glock_t *sync_this_gl, gfs_glock_t *before_releasing_this_gl);


/*  Only used in gfs_mount_lockproto() (and the locking modules)  */

void gfs_glock_cb(lm_fsdata_t *fsdata, unsigned int type, void *data);


/*  Crap  */

void gfs_pitch_inodes(gfs_sbd_t *sdp);
void gfs_glockd_scan(gfs_sbd_t *sdp, unsigned int timeout);
void gfs_clear_gla(gfs_sbd_t *sdp);

void gfs_dump_lockstate(gfs_sbd_t *sdp);


/*  These are architecture specific functions
    Either implement them or stub them out in your arch_* directory.  */

void gfs_glock_arch_init(gfs_sbd_t *sdp, gfs_glock_t *gl);


#endif /*  __GFS_GLOCK_DOT_H__  */
