/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portions Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name:  incore.h
**
**    Description:  The GFS in-core structures.
**
*******************************************************************************
******************************************************************************/

#ifndef __INCORE_DOT_H__
#define __INCORE_DOT_H__



/*  Generic Constants  */

#define GFS_CLIENTS             (1024)


/*  Structure prototypes  */

struct gfs_bitmap;
struct gfs_rgrpd;
struct gfs_bufdata;
struct gfs_gl_depend;
struct gfs_glock_operations;
struct gfs_glock;
struct gfs_ipreserv;
struct gfs_inode;
struct gfs_tune_pr;
struct gfs_tr_nop;
struct gfs_trans;
struct gfs_nopen_block;
struct gfs_bhlist;
struct gfs_ghash_bucket;
struct gfs_sbd;

typedef struct gfs_bitmap gfs_bitmap_t;
typedef struct gfs_rgrpd gfs_rgrpd_t;
typedef struct gfs_bufdata gfs_bufdata_t;
typedef struct gfs_gl_depend gfs_gl_depend_t;
typedef struct gfs_glock_operations gfs_glock_operations_t;
typedef struct gfs_glock gfs_glock_t;
typedef struct gfs_ipreserv gfs_ipreserv_t;
typedef struct gfs_inode gfs_inode_t;
typedef struct gfs_tune_pr gfs_tune_pr_t;
typedef struct gfs_tr_nop gfs_tr_nop_t;
typedef struct gfs_trans gfs_trans_t;
typedef struct gfs_nopen_block gfs_nopen_block_t;
typedef struct gfs_bhlist gfs_bhlist_t;
typedef struct gfs_ghash_bucket gfs_ghash_bucket_t;
typedef struct gfs_sbd gfs_sbd_t;



#include "incore_arch.h"



/*
 *  Structure containing information about the allocation bitmaps.
 *  There are one of these for each fs block that the bitmap for
 *  the resource group header covers.
 */

struct gfs_bitmap
{
  uint32              bi_offset;        /* The offset in the buffer of the first byte */
  uint32              bi_start;         /* The position of the first byte in this block */
  uint32              bi_len;           /* The number of bytes in this block */
}; 



/*
 *  Structure containing information Resource Groups
 */

#define RDF_SYNC_B4_ALLOC       (0)

struct gfs_rgrpd
{
  osi_list_head_t     rd_list;		/* Link with superblock */

  struct gfs_rgrpd    *rd_avl_left;
  struct gfs_rgrpd    *rd_avl_right;
  unsigned            rd_avl_height;

  osi_atomic_t        rd_count;         /* Reference count */
  gfs_glock_t         *rd_gl;           /* Glock for rgrp */

  int                 rd_flags;

  gfs_rindex_t        rd_ri;            /* Resource Index structure */
  gfs_rgrp_t          rd_rg;            /* Resource Group structure */

  int                 rd_rs_dirty;      /* TRUE if the rd_rs needs to be synced out */
  gfs_rgrp_save_t     rd_rs;            /* Resource Group LVB structure */

  gfs_bitmap_t        *rd_bits;
  osi_buf_t           **rd_bh;

  gfs_sbd_t           *rd_sbd;
};



/*
 *  Per-buffer data
 */

struct gfs_bufdata
{
  osi_buf_t           *bd_bh;           /* osi_buf_t which this struct belongs to */
  gfs_glock_t         *bd_gl;           /* Pointer to Glock struct for this bh */

  osi_list_head_t     bd_new_list;      /* List of buffers in this new transaction */
  osi_list_head_t     bd_incore_list;   /* List of buffers in this incore transaction */

  char                *bd_frozen;
  osi_sema_t          bd_lock;

  gfs_bufdata_arch_t  bd_arch;
};



/*
 *  Glock dependencies
 */

struct gfs_gl_depend
{
  osi_list_head_t     gd_list;

  lm_lockname_t       gd_name;
  gfs_glock_t         *gd_parent;

  osi_clock_ticks_t   gd_time;
};



/*
 *  Glock operations
 */

#define GFS_SYNC_WAIT           (1)
#define GFS_SYNC_INVISIBLE      (2)

struct gfs_glock_operations
{
  void                (*go_sync)(gfs_sbd_t *sdp, gfs_glock_t *gl, int flags);

  void                (*go_acquire)(gfs_sbd_t *sdp, gfs_glock_t *gl);
  void                (*go_release)(gfs_sbd_t *sdp, gfs_glock_t *gl);

  int                 (*go_lock)(gfs_sbd_t *sdp, gfs_glock_t *gl, int flags);
  void                (*go_unlock)(gfs_sbd_t *sdp, gfs_glock_t *gl, int flags);

  void                (*go_free)(gfs_sbd_t *sdp, gfs_glock_t *gl);

  int                 go_type;
  char                *go_name;
};



/*
 *  Glock Structure  (One per active Glock)
 */

/*
 * Glock flags. Can be changed atomically with bit setting functions, but
 * _only_ one at a time.
 */

#define GLF_HELD                (0)
#define GLF_PERM                (1)

#define GLF_LOCK                (2)

#define GLF_SYNC                (3)
#define GLF_DIRTY               (4)
#define GLF_POISONED            (5)

#define GLF_RELEXCL             (6)
#define GLF_RELDFRD		(7)
#define GLF_RELSHRD		(8)

#define GLF_STICKY              (9)
#define GLF_DEMOTEME            (10)

#define GLF_DEPCHECK            (11)

#define GLF_RECOVERY            (12)

/*
 * Explanations of the gfs_glock fields:
 *
 * gl_list   - Hash table hook
 * gl_name   - Unique name for lock
 * gl_count  - Reference count of gfs_glock
 * gl_flags  - Flags (bit definitions as above)
 * gl_bucket - Hash bucket that we inhabit
 * gl_head_lock - Spinlock which covers fields marked: (*)
 * gl_wait_lock - Wait for exclusive access to glock (see GLF_LOCK)
 * gl_wait_unlock - Wait for glock state to be LM_ST_UNLOCKED
 * gl_pid - PID of locking process (*)
 * gl_locked - Count of processes locking this glock (recursively) (*)
 * gl_ops - Operations which get called at various times
 * gl_state - The state of the lock
 *            LM_ST_UNLOCKED - Unlocked
 *            LM_ST_EXCLUSIVE - Enclusive lock
 *            LM_ST_SHARED - Shared lock
 *            LM_ST_DEFERRED - Another shared lock mode which doesn't share with
 *                             the LM_ST_SHARED mode.
 *
 * gl_lock - Private data for the lock module
 * gl_parent - The locks parent lock (if one exists)
 * gl_children - List of children of this lock
 * gl_lvb_count - Number of LVB references held on this glock
 * gl_lvb - LVB data attached by lock module
 * gl_vn - Version number (incremented when cache is not valid any more)
 * gl_stamp - Time of create or last unlock
 * gl_object - The object the glock is protecting
 *
 * gl_bufs - Buffers covered by this lock
 * gl_buflock - Spinlock covering the buffer lock
 *
 * gl_new_list - List of glocks in current trans
 * gl_new_bufs - List of buffers in current trans
 * gl_trans_t - The current transaction
 * 
 * gl_incore_list - List of glocks in incore transaction
 * gl_incore_bufs - List of buffers in incore transaction
 * gl_incode_tr - The incore transaction
 *
 * gl_num_dep - The number of glocks that need to be synced before this one can be released
 * gl_depend - The list of glocks that need to be synced before this one can be released
 */

struct gfs_glock
{
  osi_list_head_t     gl_list;          /* Pointers to other gfs_glock_t */
  lm_lockname_t       gl_name;          /* Name of the lock */
  osi_atomic_t        gl_count;         /* Usage Counter */
  int                 gl_flags;		/* Flags */
  unsigned int        gl_bucket;	/* Hash bucket we are in */

  osi_spinlock_t      gl_head_lock;     /* Spinlock covering various fields */
  osi_wait_chan_t     gl_wait_lock;     /* Wait to lock glock */
  osi_wait_chan_t     gl_wait_unlock;   /* Wait for glock to be unlocked */

  long                gl_pid;           /* Process ID of the locking process */
  osi_atomic_t        gl_locked;        /* Glock recursive lock count */
  gfs_glock_operations_t *gl_ops;       /* Glock operations */

  unsigned int        gl_state;         /* The state in which the lock is held */
  lm_lock_t           *gl_lock;         /* Private data for lock module */

  struct gfs_glock    *gl_parent;       /* This lock's parent (or NULL) */
  osi_list_head_t     gl_parlist;       /* Parent's list of children */
  osi_list_head_t     gl_children;      /* This lock's children */

  unsigned int        gl_lvb_count;     /* LVB reference count */
  lm_lvb_t            *gl_lvb;		/* LVB data */

  uint64              gl_vn;            /* Version number */
  osi_clock_ticks_t   gl_stamp;         /* Time of create or last unlock */
  void                *gl_object;       /* The object the glock is protecting */

  /*  Modified under the glock  (i.e. gl_locked > 0)  */

  osi_list_head_t     gl_new_list;      /* List of glocks in trans currently being built */
  osi_list_head_t     gl_new_bufs;      /* List of buffers for this lock in new trans */
  gfs_trans_t         *gl_new_tr;       /* The new transaction */

  /*  Modified under the log lock  */

  osi_list_head_t     gl_incore_list;   /* List of glocks in incore trans */
  osi_list_head_t     gl_incore_bufs;   /* List of buffers for this lock in incore trans */
  gfs_trans_t         *gl_incore_tr;    /* The incore transaction */

  osi_atomic_t        gl_num_dep;       /* Number of dependent Glocks */
  osi_list_head_t     gl_depend;        /* List of dependent Glocks */

  gfs_glock_arch_t    gl_arch;
};



/*
 *  In-Place Reservation structure
 */

struct gfs_ipreserv
{
  /*  Filled in by the caller to gfs_inplace_reserv()  */

  uint32              ir_diblks;     /* Number of dinodes to reserve */
  uint32              ir_metablks;   /* Number of metadata blocks to reserve */
  uint32              ir_datablks;   /* Number of datadata blocks to reserve */
  uint64              ir_target;     /* this is the target RG  */
  int                 ir_trans;      /* transaction type (for debugging) */

  /*  Filled in by gfs_inplace_reserv()  */

  gfs_rgrpd_t         *ir_rgd;
  uint32              ir_metares;
  uint32              ir_datares;

  /*  Filled in by gfs_blkalloc()  */

  uint32              ir_dialloced;
  uint32              ir_metaalloced;
  uint32              ir_dataalloced;
};



/*
 *  Incore inode structure
 */

#define GIF_CREATED             (0)

struct gfs_inode
{
  gfs_inum_t          i_num;

  osi_atomic_t        i_count;          /* Usage count */
  int                 i_flags;

  uint64              i_vn;             /* Clock associated with the glock */

  gfs_dinode_t        i_di;             /* Dinode Structure */
  gfs_glock_t         *i_gl;		/* Glock pointer */

  gfs_glock_t         *i_plock_gl;      /* plock head lock */
  osi_sema_t          i_plock_mux;      /* one pid at a time */
  osi_atomic_t        i_plock_cnt;      /* plock usage */

  osi_sema_t          i_flock_mux;      /* one pid at a time in list */
  osi_sema_t          i_flock_lmx;      /* one pid at a time in lock */
  int                 i_flock_st;       /* current flock state */
  osi_list_head_t     i_flock_pids;     /* list of pids that have this flock */

  gfs_sbd_t           *i_sbd;           /* GFS superblock pointer */
  osi_vnode_t         *i_vino;          /* FS's device independent ino */

  gfs_ipreserv_t      i_res;
  osi_buf_t           *i_cache[GFS_MAX_META_HEIGHT];
};



/*
 *  Transaction structures
 */

struct gfs_tr_nop
{
  osi_list_head_t     nop_list;

  gfs_inum_t          nop_inum;          /* The inode number */
  unsigned int        nop_type;          /* GFS_LOG_DESC_NOPOPEN, or GFS_LOG_DESC_NOPCLOSE */
};


#define TRF_POISONED            (0x00000001)
#define TRF_NOPEN_DUMP          (0x00000002)

struct gfs_trans
{
  osi_list_head_t     tr_list;          /* List of transactions */

  /* Initial creation stuff */

  unsigned int        tr_type;          /* The type of the transaction TRANS_CREATE, ...  */

  unsigned int        tr_blks_asked;    /* Number of log blocks asked to be reserved */
  unsigned int        tr_seg_reserved;  /* Number of segments reserved */

  /* Stuff filled in during creation */

  unsigned int        tr_flags;
  unsigned int        tr_num_gl;        /* Number of Glocks in the trans */
  unsigned int        tr_num_nop_o;     /* Number of Nopen opens in the trans */
  unsigned int        tr_num_nop_c;     /* Number of Nopen closes in the trans */
  unsigned int        tr_num_buf;       /* Number of buffers in the trans */

  osi_list_head_t     tr_gl;            /* List of glocks in this trans */
  osi_list_head_t     tr_nop_change;    /* List of nopen changes in this trans */
  osi_list_head_t     tr_nop_change_ic;

  /* Stuff modified during the commit */

  uint64              tr_log_head;      /* The current log head */
  uint64              tr_first_head;    /* First header block */ 

  osi_list_head_t     tr_bhlist;        /* List of buffers going to the log */

  gfs_trans_arch_t    tr_arch;
};



#define BLF_FAKE                (0x00000001)

struct gfs_bhlist
{
  osi_list_head_t     bl_list;

  osi_buf_t           *bl_bh;
  osi_buf_t           *bl_unlock;

  int                 bl_flags;
};



struct gfs_ghash_bucket
{
  osi_atomic_t        hb_notheld_count;
  osi_atomic_t        hb_held_count;
  osi_atomic_t        hb_perm_count;

  osi_list_head_t     hb_notheld;          /* !GLF_HELD */
  osi_list_head_t     hb_held;             /* GLF_HELD */
  osi_list_head_t     hb_perm;             /* (gl_locked > 0) for a long time */

  osi_rwlock_t        hb_lock;
}
#if defined(__KERNEL__)
__attribute__((__aligned__(SMP_CACHE_BYTES)))
#endif
;



/*
 *  Super Block Data Structure  (One per filesystem)
 */

#define GFS_GHASH_SHIFT         (9)
#define GFS_GHASH_SIZE          (1 << GFS_GHASH_SHIFT)
#define GFS_GHASH_MASK          (GFS_GHASH_SIZE - 1)

struct gfs_sbd
{
  gfs_sb_t            sd_sb;            /* Super Block */
  osi_vfs_t           *sd_vsb;          /* FS's device independent sb */


  gfs_args_t          sd_args;


  /* Resource group stuff */

  gfs_inode_t         *sd_riinode;      /* rindex inode */
  uint64              sd_riinode_vn;    /* Version number of the resource index inode */

  osi_list_head_t     sd_rglist;        /* List of resource groups */
  gfs_rgrpd_t         *sd_rgavl;        /* AVL tree of resource groups */
  gfs_rgrpd_t         *sd_rgcache;      /* Cache of previous lookup */
  unsigned int        sd_rgcount;       /* Count of resource groups */
  osi_rwlock_t        sd_rglock;        /* Lock upon list & AVL tree, etc */


  /*  Constants computed on mount  */

  uint32              sd_fsb2bb_shift;  /* Shift FS Block numbers to the left by
					   this to get buffer cache blocks  */
  uint32              sd_diptrs;        /* Number of pointers in a dinode */
  uint32              sd_inptrs;        /* Number of pointers in a indirect block */
  uint32              sd_jbsize;        /* Size of a journaled data block */
  uint32              sd_hash_bsize;    /* sizeof(exhash block) */
  uint32              sd_hash_ptrs;     /* Number of points in a hash block */
  uint32              sd_max_dirres;    /* Maximum space needed to add a directory entry */
  uint32              sd_max_height;    /* Maximum height of a file's metadata tree */


  /*  Mount flags  */

  int                 sd_noatime;
  int                 sd_sync_mount;


  /*  Lock Stuff  */

  gfs_ghash_bucket_t  sd_ghash[GFS_GHASH_SIZE];

  unsigned int        sd_gl_heldtime;   /* Time to cache locks for */

  lm_lockstruct_t     sd_lockstruct;


  /*  Inode Stuff  */

  gfs_inode_t         *sd_rooti;        /* FS's root inode */

  osi_atomic_t        sd_free_icount;   /* Approx free inodes hanging around */
  uint64              sd_dir_alloc_target; /* The address of the RG to allocate dirs from */
  osi_clock_ticks_t   sd_shrink_time;

  gfs_glock_t         *sd_rename_gl;     /* rename glock */


  /*  Daemon stuff  */

  osi_task_t         *sd_glockd_process;
  osi_atomic_t        sd_glockd_ctrl;
  osi_sema_t          sd_glockd_sema;

  osi_task_t         *sd_inoded_process;
  osi_atomic_t        sd_inoded_ctrl;
  osi_sema_t          sd_inoded_sema;

  osi_task_t         *sd_recoverd_process;
  osi_atomic_t        sd_recoverd_ctrl;
  osi_sema_t          sd_recoverd_sema;


  /*  Log stuff  */

  gfs_glock_t         *sd_trans_gl;     /* transaction glock */

  gfs_inode_t         *sd_jiinode;      /* jindex inode */
  uint64              sd_jiinode_vn;    /* Version number of the journal index inode */

  unsigned int        sd_journals;      /* Number of journals in the FS */
  gfs_jindex_t        *sd_jindex;       /* Array of Jindex structures describing this FS's journals */

  gfs_jindex_t        sd_jdesc;         /* Jindex structure describing this machine's journal */
  gfs_glock_t         *sd_journal_gl;   /* the glock for this machine's journal */

  uint64              sd_sequence;      /* Assigned to xactions in order they commit */
  uint64              sd_log_head;      /* Block number of next journal write */

  unsigned int        sd_log_seg_free;  /* Free segments in the log */
  osi_list_head_t     sd_log_ail;       /* gfs_trans_t structures that form the Active Items List 
					   "next" is the head, "prev" is the tail  */

  osi_list_head_t     sd_log_incore;    /* transactions that have been commited incore (but not ondisk)
					   "next" is the newest, "prev" is the oldest  */
  unsigned int        sd_log_buffers;   /* Number of buffers in the incore log */

  osi_sema_t          sd_log_lock;      /* Lock for access to log values */

  osi_list_head_t     sd_log_bhlist_cache;


  /*  Nopen crap  */

  osi_list_head_t     sd_nopen_od_list; /* List of inodes waiting to be deallocated */
  osi_atomic_t        sd_nopen_od_count;  /* Count of inodes waiting to be deallocated */

  osi_list_head_t     sd_nopen_ic_list;
  osi_atomic_t        sd_nopen_ic_count;  /* Like sd_nopen_count, but not
					     including deallocations already commited incore */

  uint64              sd_last_nopd;
  int                 sd_in_nopen_dump;


  /* Recovery stuff */

  int                 sd_dirty_j[GFS_CLIENTS / sizeof(int) / 8];
};



#define gfs_meta_check(bh) \
do \
{ \
  uint32 meta_check_magic = ((gfs_meta_header_t *)BH_DATA((bh)))->mh_magic; \
  meta_check_magic = gfs32_to_cpu(meta_check_magic); \
  GFS_ASSERT(meta_check_magic == GFS_MAGIC, \
	     gfs_meta_header_t meta_check_mh; \
	     con_printf("Bad metadata at %"FMT_64"u\n", BH_BLKNO((bh))); \
	     gfs_meta_header_in(&meta_check_mh, BH_DATA((bh))); \
	     gfs_meta_header_print(&meta_check_mh, TRUE);); \
} \
while (0)

#define gfs_metatype_check(bh, type) \
do \
{ \
  uint32 metatype_check_magic = ((gfs_meta_header_t *)BH_DATA((bh)))->mh_magic; \
  uint32 metatype_check_type = ((gfs_meta_header_t *)BH_DATA((bh)))->mh_type; \
  metatype_check_magic = gfs32_to_cpu(metatype_check_magic); \
  metatype_check_type = gfs32_to_cpu(metatype_check_type); \
  GFS_ASSERT(metatype_check_magic == GFS_MAGIC && \
	     metatype_check_type == (type), \
	     gfs_meta_header_t metatype_check_mh; \
	     con_printf("Bad metadata at %"FMT_64"u, should be %u\n", BH_BLKNO((bh)), (type)); \
	     gfs_meta_header_in(&metatype_check_mh, BH_DATA((bh))); \
	     gfs_meta_header_print(&metatype_check_mh, TRUE);); \
} \
while (0)

#define gfs_metatype_set(bh, type, format) \
do \
{ \
  gfs_meta_check((bh)); \
  ((gfs_meta_header_t *)BH_DATA((bh)))->mh_type = cpu_to_gfs32((type)); \
  ((gfs_meta_header_t *)BH_DATA((bh)))->mh_format = cpu_to_gfs32((format)); \
} \
while (0)



#endif  /*  __INCORE_DOT_H__  */




