/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portions Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: ioctl.c
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "blklist.h"
#include "bmap.h"
#include "dio.h"
#include "file.h"
#include "glock.h"
#include "glops.h"
#include "inode.h"
#include "super.h"
#include "trans.h"





#ifdef DEBUG_TRACE
/**
 * gfs_debug_change - Change debug level
 * @cmd: The command
 * @arg: Pointer to argument
 *
 * Returns: 0 on success or error code
 */

int gfs_debug_change(unsigned int cmd, void *arg)
{
  unsigned int flag;
  unsigned char *val;

  ENTER(GFN_DEBUG_CHANGE);


  if (osi_copy_from_user(&flag, arg, sizeof(unsigned int)))
  {
    EXIT(GFN_DEBUG_CHANGE);
    return -EFAULT;
  }


  if (flag >= DEBUG_FLAGS)
  {
    EXIT(GFN_DEBUG_CHANGE);
    return -EINVAL;
  }


  val = &gfs_debug_flags[flag / 8];


  if (cmd == GFS_DEBUG_ON)
    *val |= 1 << (flag % 8);
  else
    *val &= ~(1 << (flag % 8));


  EXIT(GFN_DEBUG_CHANGE);

  return 0;
}
#endif  /*  DEBUG_TRACE  */


#ifdef DEBUG_PRINT
/**
 * gfs_debug_dump - Dump debug information
 * @arg: Pointer to buffer
 *
 * Returns: Length of buffer used, or error code on failure
 */

int gfs_debug_dump(void *arg)
{
  char *p;
  int error = 0;

  ENTER(GFN_DEBUG_DUMP);


  /*  If they didn't give us a pointer,
      tell them how big a buffer to pass in.  */

  if (!arg)
  {
    EXIT(GFN_DEBUG_DUMP);
    return GFS_ICDB_BUF_LEN;
  }


  p = (char *)arg;


  osi_sema_lock(&gfs_icdb_lock);

  if (osi_copy_to_user(p, gfs_icdb_buf + gfs_icdb_point, GFS_ICDB_BUF_LEN - gfs_icdb_point))
    error = -EFAULT;
  if (!error && gfs_icdb_point)
    if (osi_copy_to_user(p + (GFS_ICDB_BUF_LEN - gfs_icdb_point), gfs_icdb_buf, gfs_icdb_point))
      error = -EFAULT;

  osi_sema_unlock(&gfs_icdb_lock);


  EXIT(GFN_DEBUG_DUMP);

  return error;
}


/**
 * gfs_print_frag - Print information about block locations
 * @ip: The inode to print information about
 *
 */

void gfs_print_frag(gfs_inode_t *ip)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  uint64 iblock, block, blocks;
  uint64 start = 0, daddr = 0, run = 0;
  int not_new = 0;
  int error;

  ENTER(GFN_PRINT_FRAG);


  db_printf("Extent list for inode %"FMT_64"u/%"FMT_64"u\n",
	    ip->i_num.no_formal_ino, ip->i_num.no_addr);

  error = gfs_glock_i(ip, GL_SHARED);
  if (error)
  {
    EXIT(GFN_PRINT_FRAG);
    return;
  }


  if (gfs_is_stuffed(ip))
  {
    db_printf("  stuffed\n");
    gfs_gunlock_i(ip, 0);
    EXIT(GFN_PRINT_FRAG);
    return;    
  }


  db_printf("%15s%15s%15s\n", "logical block", "extent length", "disk block");


  blocks = (ip->i_di.di_size + 1) >> sdp->sd_sb.sb_bsize_shift;
  for (iblock = 0; iblock < blocks; iblock++)
  {
    error = gfs_inode_get_block(ip, iblock, &not_new, &block);
    if (error)
    {
      gfs_gunlock_i(ip, 0);
      EXIT(GFN_PRINT_FRAG);
      return;    
    }

    if (daddr && (daddr + run == block))
    {
      run++;
    }
    else
    {
      if (daddr)
	db_printf("%15"FMT_64"u%15"FMT_64"u%15"FMT_64"u\n",
		  start, run, daddr);

      start = iblock;
      daddr = block;
      run = 1;
    }
  }


  if (daddr)
    db_printf("%15"FMT_64"u%15"FMT_64"u%15"FMT_64"u\n",
	      start, run, daddr);
 

  gfs_gunlock_i(ip, 0);


  EXIT(GFN_PRINT_FRAG);
}
#endif  /*  DEBUG_PRINT  */


#if DEBUG_PROFILE
/**
 * gfs_prof_dump - Profile dump
 * @arg: The buffer
 *
 * Returns: 0 on success, error code otherwise
 */

int gfs_prof_dump(void *arg)
{
  int x;
  int error = 0;

  ENTER(GFN_PROF_DUMP);


  /*  If they didn't give us a pointer,
      tell them how big a buffer to pass in.  */

  if (!arg)
  {
    EXIT(GFN_PROF_DUMP);
    return DEBUG_FLAGS;
  }


  con_printf("gfs_prof_neg_count = %"FMT_64"u\n", gfs_prof_neg_count);

  if (osi_copy_to_user(arg, gfs_prof_stats, DEBUG_FLAGS * sizeof(gfs_profile_t)))
    error = -EFAULT;
  else
  {
    gfs_memset(gfs_prof_stats, 0, DEBUG_FLAGS * sizeof(gfs_profile_t));
    for (x = 0; x < DEBUG_FLAGS; x++)
      gfs_prof_stats[x].min_micros = (uint64)-1;
    gfs_prof_neg_count = 0;
  }


  EXIT(GFN_PROF_DUMP);

  return error;
}
#endif  /*  DEBUG_PROFILE  */


/**
 * gfs_get_super - Dump the superblock into a buffer
 * @sb: The superblock
 * @ptr: The buffer pointer
 *
 * Returns: 0 or error code
 */

int gfs_get_super(gfs_sbd_t *sdp, void *arg)
{
  gfs_glock_t *gl;
  osi_buf_t *bh;
  int error;

  ENTER(GFN_GET_SUPER);


  error = gfs_glock_num(sdp, GFS_SB_LOCK, &gfs_meta_glops, NULL, GL_SHARED, &gl);
  if (error)
    goto out;

  bh = gfs_dread(sdp, GFS_SB_ADDR >> sdp->sd_fsb2bb_shift, gl, DR_START | DR_WAIT);
  if (!bh)
  {
    error = -EIO;
    goto out_gunlock;
  }

  error = osi_copy_to_user(arg, BH_DATA(bh), BH_SIZE(bh));

  gfs_drelse(bh);


 out_gunlock:
  gfs_gunlock(sdp, gl, 0);
  gfs_put_glstruct(gl);

 out:
  EXIT(GFN_GET_SUPER);
  return error;
}


/**
 * gfs_jread_ioctl - Read from a journaled data file via ioctl
 * @ip: The GFS inode
 * @arg: The argument from ioctl
 *
 * Returns: Amount of data copied or error
 */

int gfs_jread_ioctl(gfs_inode_t *ip, void *arg)
{
  gfs_jio_t jt;
  int ret;

  ENTER(GFN_JREAD_IOCTL);

  if (osi_copy_from_user(&jt, arg, sizeof(gfs_jio_t)))
  {
    ret = -EFAULT;
    goto out;
  }


  ret = gfs_glock_i(ip, GL_SHARED);
  if (ret)
    goto out;

  ret = gfs_readi(ip, jt.jio_data, jt.jio_offset, jt.jio_size, gfs_copy2user);

  gfs_gunlock_i(ip, 0);


 out:
  EXIT(GFN_JREAD_IOCTL);
  return ret;
}


/**
 * gfs_jwrite_ioctl - Write to a journaled file via ioctl
 * @ip: The GFS inode
 * @arg: The argument from ioctl
 *
 * Returns: Amount of data copied or error
 */

int gfs_jwrite_ioctl(gfs_inode_t *ip, void *arg)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  gfs_ipreserv_t *res = &ip->i_res;
  gfs_jio_t jt;
  unsigned int ind = 0, tmp, data_blocks, trans_reserv;
  int ret;

  ENTER(GFS_JWRITE_IOCTL);

  if (osi_copy_from_user(&jt, arg, sizeof(gfs_jio_t)))
  {
    ret = -EFAULT;
    goto out;
  }


  /*  Compute maximum number of data blocks we need.
      (We might need a block to unstuff the dinode)  */

  data_blocks = (jt.jio_size >> sdp->sd_sb.sb_bsize_shift) + 3;

  /*  Compute maximum number of indirect blocks we need  */

  for (tmp = data_blocks; tmp > sdp->sd_diptrs; )
  {
    tmp = DIV_RU(tmp, sdp->sd_inptrs) + 1;
    ind += tmp;
  }

  ind += 2 * (sdp->sd_max_height - 1);


  ret = gfs_glock_i(ip, 0);
  if (ret)
    goto out;


  res->ir_diblks = 0;
  if (gfs_is_jdata(ip))
  {
    res->ir_metablks = ind + data_blocks;
    res->ir_datablks = 0;
  }
  else
  {
    res->ir_metablks = ind;
    res->ir_datablks = data_blocks;
  }
  res->ir_target = ip->i_di.di_goal_rgrp;
  res->ir_trans = TRANS_JWRITE;

  ret = gfs_inplace_reserv(ip);
  if (ret)
    goto out_gunlock;


  /*  Trans may require:
      All blocks for a RG bitmap, all the "data" blocks, whatever indirect blocks we
      need, and a modified dinode  */

  if (gfs_is_jdata(ip))
    trans_reserv = 1 + data_blocks + res->ir_rgd->rd_ri.ri_length + ind;
  else
    trans_reserv = 1 + res->ir_rgd->rd_ri.ri_length + ind;

  ret = gfs_trans_begin(sdp, TRANS_JWRITE, trans_reserv, 0);
  if (ret)
    goto out_relse;

  ret = gfs_writei(ip, jt.jio_data, jt.jio_offset, jt.jio_size, gfs_copy_from_user);

  gfs_trans_end(sdp);


 out_relse:
  gfs_inplace_relse(ip);

 out_gunlock:
  ip->i_gl->gl_vn++;
  gfs_gunlock_i(ip, GL_SYNC);

 out:
  EXIT(GFS_JWRITE_IOCTL);

  return ret;
}


/**
 * gfs_stat_gfs_ioctl - Do a GFS specific statfs
 * @sdp: the filesystem
 * @arg: the gfs_usage_t structure
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_stat_gfs_ioctl(gfs_sbd_t *sdp, void *arg)
{
  gfs_usage_t usage;
  int error = 0;

  ENTER(GFN_STAT_GFS_IOCTL);


  error = gfs_stat_gfs(sdp, &usage, TRUE);
  if (error)
    goto out;

  if (osi_copy_to_user(arg, &usage, sizeof(gfs_usage_t)))
    error = -EFAULT;


 out:
  EXIT(GFN_STAT_GFS_IOCTL);

  return error;
}



