/******************************************************************************
*******************************************************************************
**
**    Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: main.c
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "dio.h"
#include "glock.h"
#include "locking.h"
#include "super.h"





/**
 * gfs_mount_lockproto - mount a locking protocol
 * @sdp: the filesystem
 * @args: mount arguements
 * @silent: if TRUE, don't complain if the FS isn't a GFS fs
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_mount_lockproto(gfs_sbd_t *sdp, int silent)
{
  gfs_sb_t sb;
  osi_buf_t *bh;
  char proto[256], table[256], host[256];
  uint64 block;
  int error;

  ENTER(GFN_MOUNT_LOCKPROTO);


  /*  Copy in values from "args"  */

  osi_strcpy(proto, sdp->sd_args.ar_lockproto);
  osi_strcpy(table, sdp->sd_args.ar_locktable);
  osi_strcpy(host, sdp->sd_args.ar_hostdata);


  /*  Try to autodetect  */

  if (!proto[0] || !table[0])
  {
    block = GFS_SB_ADDR * GFS_BASIC_BLOCK / sdp->sd_sb.sb_bsize;

    bh = gfs_dread(sdp, block, NULL, DR_FORCE | DR_START | DR_WAIT);
    if (!bh)
    {
      EXIT(GFN_MOUNT_LOCKPROTO);
      return -ENOBUFS;
    }
    gfs_sb_in(&sb, BH_DATA(bh));
    gfs_drelse(bh);

    error = gfs_check_sb(sdp, &sb, silent);
    if (error)
    {
      EXIT(GFN_MOUNT_LOCKPROTO);
      return error;
    }
      
    if (!proto[0])
      osi_strcpy(proto, sb.sb_lockproto);
      
    if (!table[0])
      osi_strcpy(table, sb.sb_locktable);
  }


  DO_PROF(LM_MOUNT,
	  error = lm_mount(proto, table, host,
			   gfs_glock_cb, sdp,
			   &sdp->sd_lockstruct););
  if (error)
  {
    con_printf("GFS:  can't mount proto = %s, table = %s, hostdata = %s\n",
	       proto, table, host);
    EXIT(GFN_MOUNT_LOCKPROTO);
    return -1;
  }

  GFS_ASSERT(sdp->sd_lockstruct.ls_lockspace, );
  GFS_ASSERT(sdp->sd_lockstruct.ls_ops, );  


  EXIT(GFN_MOUNT_LOCKPROTO);

  return 0;
}


/**
 * gfs_unmount_lockproto - Unmount lock protocol
 * @sdp: The GFS superblock
 *
 */

void gfs_unmount_lockproto(gfs_sbd_t *sdp)
{
  ENTER(GFN_UNMOUNT_LOCKPROTO);

  DO_PROF(LM_UNMOUNT,
	  sdp->sd_lockstruct.ls_ops->unmount(sdp->sd_lockstruct.ls_lockspace););

  EXIT(GFN_UNMOUNT_LOCKPROTO);
}


