/******************************************************************************
*******************************************************************************
**
**    Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: 
**      log.c
**
**    Description:
**      Log functions
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"

#include "dio.h"
#include "glock.h"
#include "log.h"
#include "recovery.h"
#include "trans.h"





/**
 * gfs_nopen_merge - Add/Remove from nopen the lists
 * @sdp: The GFS superblock
 * @nop: The nopen change
 * @list: the list to add to / remove from
 * @counter: the count of elements on the list
 *
 * Called with the log lock held
 */

void gfs_nopen_merge(gfs_sbd_t *sdp, gfs_tr_nop_t *nop_new,
		     osi_list_head_t *list, osi_atomic_t *counter)
{
  osi_list_head_t *tmp;
  gfs_tr_nop_t *nop = NULL;

  ENTER(GFN_NOPEN_MERGE);


  for (tmp = list->next; tmp != list; tmp = tmp->next)
  {
    nop = osi_list_entry(tmp, gfs_tr_nop_t, nop_list);

    if (GFS_INUM_EQUAL(&nop->nop_inum, &nop_new->nop_inum))
      break;
    else
      nop = NULL;
  }


  switch (nop_new->nop_type)
  {
  case GFS_LOG_DESC_NOPOPEN:
    GFS_ASSERT(!nop, );

    osi_list_add(&nop_new->nop_list, list);
    osi_atomic_inc(counter);

    break;


  case GFS_LOG_DESC_NOPCLOSE:
    GFS_ASSERT(nop, );

    osi_list_del(&nop->nop_list);
    gfree(nop_new, sizeof(gfs_tr_nop_t));
    gfree(nop, sizeof(gfs_tr_nop_t));
    osi_atomic_dec(counter);


    break;


  default:
    GFS_ASSERT(0, );
  }


  EXIT(GFN_NOPEN_MERGE);
}


/**
 * gfs_nopen_find - Find inode number in nopen list
 * @sdp: The GFS superblock
 * @inum: the inode number returned
 *
 * We need to acquire the log lock
 */

void gfs_nopen_find(gfs_sbd_t *sdp, gfs_inum_t *inum)
{
  gfs_tr_nop_t *nop;

  ENTER(GFN_NOPEN_FIND);

  osi_sema_lock(&sdp->sd_log_lock);

  if (osi_list_empty(&sdp->sd_nopen_ic_list))
  {
    gfs_memset(inum, 0, sizeof(gfs_inum_t));
    goto out;
  }

  nop = osi_list_entry(sdp->sd_nopen_ic_list.prev, gfs_tr_nop_t, nop_list);

  osi_list_del(&nop->nop_list);
  osi_list_add(&nop->nop_list, &sdp->sd_nopen_ic_list);

  *inum = nop->nop_inum;

 out:
  osi_sema_unlock(&sdp->sd_log_lock);

  EXIT(GFN_NOPEN_FIND);
}


/**
 * journal_segment_boundary - Discover if block is on journal segment boundary
 * @sdp: The GFS superblock
 * @block: The block number
 *
 *  This assumes that a subpool's starting block number is a multiple
 *  of the journal's segment size.
 *
 * Returns: TRUE if the block is on a journal segment boundary, FALSE otherwise
 */

STATIC __inline__ int journal_segment_boundary(gfs_sbd_t *sdp, uint64 block)
{
  return !(block % sdp->sd_sb.sb_seg_size);
}


/**
 * journal_distance - Compute distance between two journal blocks
 * @sdp: The GFS superblock
 * @newer: The most recent journal block of the pair
 * @older: The older journal block of the pair
 *
 *   Compute the distance (in the journal direction) between two
 *   blocks in the journal
 *
 * Returns: the distance in blocks
 */

STATIC __inline__ int64 journal_distance(gfs_sbd_t *sdp, uint64 newer, uint64 older)
{
  int64 dist;

  dist = newer - older;
  if (dist < 0)
    dist += sdp->sd_jdesc.ji_nsegment * sdp->sd_sb.sb_seg_size;

  return dist;
}


/**
 * current_tail - Find block number of current log tail
 * @sdp: The GFS superblock
 *
 * Find the block number of the current tail of the log.
 * Assumes that the log lock is held.
 *
 * Returns: The tail's block number
 */

STATIC uint64 current_tail(gfs_sbd_t *sdp)
{
  gfs_trans_t *tr;
  uint64 tail;

  ENTER(GFN_CURRENT_TAIL);

  if (osi_list_empty(&sdp->sd_log_ail))
  {
    tail = sdp->sd_log_head;

    if (!journal_segment_boundary(sdp, tail))
    {
      tail--;
      GFS_ASSERT(journal_segment_boundary(sdp, tail), );
    }
  }
  else
  {
    tr = osi_list_entry(sdp->sd_log_ail.prev, gfs_trans_t, tr_list);
    tail = tr->tr_first_head;
  }

  EXIT(GFN_CURRENT_TAIL);

  return tail;
}


/**
 * pull_tail_locked - move the tail of the log forward (if possible)
 * @sdp: the filesystem
 *
 * See if there are any unreferenced transactions at the end of the AIL.
 */

STATIC void pull_tail_locked(gfs_sbd_t *sdp)
{
  gfs_trans_t *tr;
  uint64 oldtail, newtail;
  int64 dist;
  unsigned int segments;

  ENTER(GFN_PULL_TAIL_LOCKED);


  oldtail = current_tail(sdp);


  while (!osi_list_empty(&sdp->sd_log_ail))
  {
    tr = osi_list_entry(sdp->sd_log_ail.prev, gfs_trans_t, tr_list);

    if (!gfs_trans_check_empty(sdp, tr))
      break;

    osi_list_del(&tr->tr_list);
    gfree(tr, sizeof(gfs_trans_t));
  }


  newtail = current_tail(sdp);

  if (oldtail != newtail)
  {
    dist = journal_distance(sdp, newtail, oldtail);

    segments = dist / sdp->sd_sb.sb_seg_size;
    GFS_ASSERT(segments * sdp->sd_sb.sb_seg_size == dist, );

    sdp->sd_log_seg_free += segments;
    GFS_ASSERT(sdp->sd_log_seg_free < sdp->sd_jdesc.ji_nsegment, );
  }


  EXIT(GFN_PULL_TAIL_LOCKED);
}


/**
 * gfs_pull_tail - move the tail of the log forward (if possible)
 * @sdp: the filesystem
 *
 * See if there are any unreferenced transactions at the end of the AIL.
 */

void gfs_pull_tail(gfs_sbd_t *sdp)
{
  ENTER(GFN_PULL_TAIL);

  osi_sema_lock(&sdp->sd_log_lock);
  pull_tail_locked(sdp);
  osi_sema_unlock(&sdp->sd_log_lock);

  EXIT(GFN_PULL_TAIL);
}


/**
 * gfs_log_reserv - Make a log reservation
 * @sdp: The GFS superblock
 * @segments: The number of segments to reserve
 *
 * Returns:  0 on success, -EXXX on failure
 */

int gfs_log_reserv(gfs_sbd_t *sdp, unsigned int segments)
{
  osi_clock_ticks_t start, flush;
  int first = TRUE;

  ENTER(GFN_LOG_RESERV);


  if (!segments || segments >= sdp->sd_jdesc.ji_nsegment)
  {
    con_printf("GFS:  error reserving log space (%u, %u)\n",
	       segments, sdp->sd_jdesc.ji_nsegment);
    EXIT(GFN_LOG_RESERV);
    return -EINVAL;
  }


  osi_time_stamp(&start);
  osi_time_stamp(&flush);  


  for (;;)
  {
    osi_sema_lock(&sdp->sd_log_lock);

    pull_tail_locked(sdp);

    if (sdp->sd_log_seg_free >= segments)
    {
      sdp->sd_log_seg_free -= segments;
      osi_sema_unlock(&sdp->sd_log_lock);
      break;
    }

    osi_sema_unlock(&sdp->sd_log_lock);


    if (first || osi_check_timeout(&flush, 1))
    {
      gfs_log_flush(sdp, NULL);
      if (osi_sync_dev(sdp->sd_vsb, FALSE))
	GFS_ASSERT(0, con_printf("Panicking because of write error\n"););

      osi_time_stamp(&flush);

      if (osi_check_timeout(&start, 60))
	con_printf("GFS:  can't make log reservation (asking for %u segments)\n",
		   segments);
    }

    osi_schedule();

    first = FALSE;    
  }


  EXIT(GFN_LOG_RESERV);

  return 0;
}


/**
 * gfs_log_relse - Release a given number of log segments
 * @sdp: The GFS superblock
 * @segments: The number of segments
 *
 */

void gfs_log_relse(gfs_sbd_t *sdp, unsigned int segments)
{
  ENTER(GFN_LOG_RELSE);

  osi_sema_lock(&sdp->sd_log_lock);
  sdp->sd_log_seg_free += segments;
  GFS_ASSERT(sdp->sd_log_seg_free < sdp->sd_jdesc.ji_nsegment, );
  osi_sema_unlock(&sdp->sd_log_lock);

  EXIT(GFN_LOG_RELSE);
}


/**
 * log_refund - Refund log segments to the free pool
 * @sdp: The GFS superblock
 * @tr: The tranaction to examine
 *
 * Look at the number of segments reserved for this transaction and the
 * number of segments actually needed for it.  If they aren't the
 * same, refund the difference to the free segment pool.
 *
 * Called with the log lock held
 */

STATIC void log_refund(gfs_sbd_t *sdp, gfs_trans_t *tr)
{
  unsigned int blocks, segments;

  ENTER(GFN_LOG_REFUND);


  blocks = gfs_struct2blk(sdp, tr->tr_num_buf, sizeof(gfs_block_tag_t)) + tr->tr_num_buf + 1;
  if (tr->tr_num_nop_o)
    blocks += gfs_struct2blk(sdp, tr->tr_num_nop_o, sizeof(gfs_inum_t));
  if (tr->tr_num_nop_c)
    blocks += gfs_struct2blk(sdp, tr->tr_num_nop_c, sizeof(gfs_inum_t));

  segments = gfs_blk2seg(sdp, blocks);

  if (tr->tr_seg_reserved > segments)
  {
    sdp->sd_log_seg_free += tr->tr_seg_reserved - segments;
    GFS_ASSERT(sdp->sd_log_seg_free < sdp->sd_jdesc.ji_nsegment, );

    tr->tr_seg_reserved = segments;
  }
  else
    GFS_ASSERT(tr->tr_seg_reserved == segments, );


  EXIT(GFN_LOG_REFUND);
}


/**
 * journal_incr_head - Increment journal head
 * @sdp: The GFS superblock
 * @tr: The transaction
 *
 * Increment journal head by one and "charge" this block to the current
 * transaction.  At the end of the journal, wrap head back to the start.
 *
 */

STATIC __inline__ void journal_incr_head(gfs_sbd_t *sdp, gfs_trans_t *tr)
{
  gfs_jindex_t *jdesc = &sdp->sd_jdesc;

  if (++tr->tr_log_head == jdesc->ji_addr + jdesc->ji_nsegment * sdp->sd_sb.sb_seg_size)
    tr->tr_log_head = jdesc->ji_addr;
}


/**
 * journal_next_segment - Set log head to next segment boundary
 * @sdp: The GFs superblock
 * @tr: The transaction
 *
 */

STATIC __inline__ void journal_next_segment(gfs_sbd_t *sdp, gfs_trans_t *tr)
{
  while (!journal_segment_boundary(sdp, tr->tr_log_head))
    journal_incr_head(sdp, tr);
}


/**
 * get_bhlist_entry - Get a gfs_bhlist_t structure
 * @sdp: the filesystem
 *
 * Get a gfs_bhlist_t structure -- Either out of the cache in the gfs_sbd_t
 * structure or gmalloc() a new one.
 */

STATIC gfs_bhlist_t *get_bhlist_entry(gfs_sbd_t *sdp)
{
  gfs_bhlist_t *bl;

  ENTER(GFN_GET_BHLIST_ENTRY);


  if (!osi_list_empty(&sdp->sd_log_bhlist_cache))
  {
    bl = osi_list_entry(sdp->sd_log_bhlist_cache.next, gfs_bhlist_t, bl_list);
    osi_list_del(&bl->bl_list);
  }
  else
  {
    bl = (gfs_bhlist_t *)gmalloc(sizeof(gfs_bhlist_t));
    if (!bl)
      goto out;
  }

  gfs_memset(bl, 0, sizeof(gfs_bhlist_t));

 out:
  EXIT(GFN_GET_BHLIST_ENTRY);

  return bl;
}


/**
 * journal_add_bh - Add a real buffer to the transaction list
 * @sdp: the filesystem
 * @tr: the transaction to add this buffer to
 * @bh: the buffer
 *
 */

STATIC int journal_add_bh(gfs_sbd_t *sdp, gfs_trans_t *tr, osi_buf_t *bh)
{
  gfs_bhlist_t *bl;

  ENTER(GFN_JOURNAL_ADD_BH);


  bl = get_bhlist_entry(sdp);
  if (!bl)
  { 
    EXIT(GFN_JOURNAL_ADD_BH);
    return -ENOMEM;
  }


  bl->bl_bh = bh;

  osi_list_add(&bl->bl_list, &tr->tr_bhlist);


  EXIT(GFN_JOURNAL_ADD_BH);

  return 0;
}


/**
 * journal_add_fakebh - Build a fake buffer head
 * @sdp: the filesystem
 * @tr: the transaction this is part of
 * @data: the data the buffer should point to
 * @unlock: a buffer that is unlocked as this gfs_bhlist_t is torn down
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int journal_add_fakebh(gfs_sbd_t *sdp, gfs_trans_t *tr, char *data, osi_buf_t *unlock)
{
  gfs_bhlist_t *bl;

  ENTER(GFN_JOURNAL_ADD_FAKEBH);


  bl = get_bhlist_entry(sdp);
  if (!bl)
  { 
    EXIT(GFN_JOURNAL_ADD_FAKEBH);
    return -ENOMEM;
  }


  bl->bl_bh = gfs_init_temp_bh(sdp, tr->tr_log_head, data, TRUE);
  if (!bl->bl_bh)
  {
    osi_list_add(&bl->bl_list, &sdp->sd_log_bhlist_cache);
    EXIT(GFN_JOURNAL_ADD_FAKEBH);
    return -ENOBUFS;
  }
  bl->bl_unlock = unlock;
  bl->bl_flags |= BLF_FAKE;

  osi_list_add(&bl->bl_list, &tr->tr_bhlist);  


  EXIT(GFN_JOURNAL_ADD_FAKEBH);

  return 0;
}


/**
 * journal_free_bhlist - free a transaction's bhlist
 * @sdp: the filesystem
 * @tr: the transaction
 *
 * Add the entries back into the cache in the superblock.
 */

STATIC void journal_free_bhlist(gfs_sbd_t *sdp, gfs_trans_t *tr)
{
  gfs_bhlist_t *bl;

  ENTER(GFN_JOURNAL_FREE_BHLIST)


  while (!osi_list_empty(&tr->tr_bhlist))
  {
    bl = osi_list_entry(tr->tr_bhlist.next, gfs_bhlist_t, bl_list);

    osi_list_del(&bl->bl_list);

    if (bl->bl_bh)
    {
      if (bl->bl_flags & BLF_FAKE)
	gfs_dispose_temp_bh(bl->bl_bh);
      else
	gfs_drelse(bl->bl_bh);
    }
    if (bl->bl_unlock)
      gfs_unlock_buffer(bl->bl_unlock);

    osi_list_add(&bl->bl_list, &sdp->sd_log_bhlist_cache);
  }


  EXIT(GFN_JOURNAL_FREE_BHLIST)
}


/**
 * journal_get_header - Get the journal header buffer
 * @sdp: The GFS superblock
 * @tr: The transaction
 * @next: TRUE is this is not a continuation of an existing transaction
 * @bhpp: Used to return a pointer to the buffer
 *
 */

STATIC int journal_get_header(gfs_sbd_t *sdp, gfs_trans_t *tr, int next, osi_buf_t **bhpp)
{
  gfs_log_header_t header;
  osi_buf_t *bh;
  int error;

  ENTER(GFN_JOURNAL_GET_HEADER);


  *bhpp = NULL;

  GFS_ASSERT(journal_segment_boundary(sdp, tr->tr_log_head), );


  bh = gfs_dgetblk(sdp, tr->tr_log_head, sdp->sd_journal_gl);
  if (!bh)
  {
    EXIT(GFN_JOURNAL_GET_HEADER);
    return -EIO;
  }
  gfs_prep_new_buffer(bh);

  gfs_buffer_clear(bh);
  gfs_memset(&header, 0, sizeof(header));

  if (next)
  {
    header.lh_header.mh_magic = GFS_MAGIC;
    header.lh_header.mh_type = GFS_METATYPE_LH;
    header.lh_header.mh_format = GFS_FORMAT_LH;
    header.lh_first = tr->tr_log_head;
    header.lh_sequence = sdp->sd_sequence + 1;
    header.lh_tail = current_tail(sdp);
    header.lh_last_nopd = sdp->sd_last_nopd;
  }
  else
  {
    header.lh_header.mh_magic = GFS_MAGIC;
    header.lh_header.mh_type = GFS_METATYPE_LH;
    header.lh_header.mh_format = GFS_FORMAT_LH;
    header.lh_first = tr->tr_first_head;
    header.lh_sequence = sdp->sd_sequence;
    header.lh_tail = current_tail(sdp);
    header.lh_last_nopd = sdp->sd_last_nopd;

    error = journal_add_bh(sdp, tr, bh);
    if (error)
    {
      gfs_drelse(bh);
      EXIT(GFN_JOURNAL_GET_HEADER);
      return error;
    }
  }

  gfs_log_header_out(&header, BH_DATA(bh));
  gfs_log_header_out(&header, BH_DATA(bh) + GFS_BASIC_BLOCK - sizeof(header));  

  journal_incr_head(sdp, tr);


  *bhpp = bh;


  EXIT(GFN_JOURNAL_GET_HEADER);

  return 0;
}


/**
 * journal_get_buffer - Get a buffer to use for control data
 * @sdp: The GFS superblock
 * @tr: The GFS transaction
 * @bhpp: Used to return the buffer
 *
 * Generate a regular buffer for use in the journal as control data.
 *
 */

STATIC int journal_get_buffer(gfs_sbd_t *sdp, gfs_trans_t *tr, osi_buf_t **bhpp)
{
  osi_buf_t *bh;
  int error;

  ENTER(GFN_JOURNAL_GET_BUFFER);


  *bhpp = NULL;


  if (journal_segment_boundary(sdp, tr->tr_log_head))
  {
    error = journal_get_header(sdp, tr, FALSE, &bh);
    if (error)
    {
      EXIT(GFN_JOURNAL_GET_BUFFER);
      return error;
    }
  }


  bh = gfs_dgetblk(sdp, tr->tr_log_head, sdp->sd_journal_gl);
  if (!bh)
  {
    EXIT(GFN_JOURNAL_GET_BUFFER);
    return -EIO;
  }
  gfs_prep_new_buffer(bh);

  gfs_buffer_clear(bh);

  
  error = journal_add_bh(sdp, tr, bh);
  if (error)
  {
    gfs_drelse(bh);
    EXIT(GFN_JOURNAL_GET_BUFFER);
    return error;
  }

  journal_incr_head(sdp, tr);


  *bhpp = bh;


  EXIT(GFN_JOURNAL_GET_BUFFER);

  return 0;
}


/**
 * gfs_struct2blk - compute stuff
 * @sdp: the filesystem
 * @nstruct: the number of structures
 * @ssize: the size of the structures
 *
 * Compute the number of log descriptor blocks needed to hold a certain number
 * of structures of a certain size.
 *
 * Returns: the number of blocks needed
 */

unsigned int gfs_struct2blk(gfs_sbd_t *sdp, unsigned int nstruct, unsigned int ssize)
{
  unsigned int blks;
  unsigned int first, second;

  ENTER(GFN_STRUCT2BLK);

  blks = 1;
  first = (sdp->sd_sb.sb_bsize - sizeof(gfs_log_descriptor_t)) / ssize;

  if (nstruct > first)
  {
    second = sdp->sd_sb.sb_bsize / ssize;
    blks += DIV_RU(nstruct - first, second);
  }

  EXIT(GFN_STRUCT2BLK);

  return blks;
}


/**
 * gfs_blk2seg - Convert number of blocks into number of segments
 * @sdp: The GFS superblock
 * @blocks: The number of blocks
 *
 * Returns: The number of journal segments
 */

unsigned int gfs_blk2seg(gfs_sbd_t *sdp, unsigned int blocks)
{
  unsigned int seg;

  ENTER(GFN_BLK2SEG);

  seg = DIV_RU(blocks, sdp->sd_sb.sb_seg_size - 1);

  EXIT(GFN_BLK2SEG);

  return seg;
}


/**
 * build_nop_desc - Build the log descriptor for an nopen change
 * @sdp: the filesystem
 * @tr: the transaction
 * @number: the number of nopen changes
 * @type: the type of changes to look for
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int build_nop_desc(gfs_sbd_t *sdp, gfs_trans_t *tr, unsigned int number, unsigned int type)
{
  osi_list_head_t *list, *tmp;
  gfs_tr_nop_t *nop;
  osi_buf_t *bh;
  gfs_log_descriptor_t desc;
  unsigned int blocks, offset, index = 0, blk = 1;
  int nopen_dump = tr->tr_flags & TRF_NOPEN_DUMP;
  int error;

  ENTER(GFN_BUILD_NOP_DESC);


  if (nopen_dump)
    list = &sdp->sd_nopen_od_list;
  else
    list = &tr->tr_nop_change;

  blocks = gfs_struct2blk(sdp, number, sizeof(gfs_inum_t));


  error = journal_get_buffer(sdp, tr, &bh);
  if (error)
    goto fail;

  gfs_memset(&desc, 0, sizeof(gfs_log_descriptor_t));
  desc.ld_header.mh_magic = GFS_MAGIC;
  desc.ld_header.mh_type = GFS_METATYPE_LD;
  desc.ld_header.mh_format = GFS_FORMAT_LD;
  desc.ld_type = type;
  desc.ld_length = blocks;
  desc.ld_data1 = (nopen_dump) ? TRUE : FALSE;

  gfs_desc_out(&desc, BH_DATA(bh));

  offset = sizeof(gfs_log_descriptor_t);


  for (tmp = list->next; tmp != list; tmp = tmp->next)
  {
    nop = osi_list_entry(tmp, gfs_tr_nop_t, nop_list);

    if (nop->nop_type != type)
      continue;

    if (offset + sizeof(gfs_inum_t) > sdp->sd_sb.sb_bsize)
    {
      offset = 0;
      bh = NULL;
    }

    if (!bh)
    {
      error = journal_get_buffer(sdp, tr, &bh);
      if (error)
	goto fail;
      blk++;
    }

    gfs_inum_out(&nop->nop_inum, BH_DATA(bh) + offset);

    offset += sizeof(gfs_inum_t);
    index++;
  }


  GFS_ASSERT(index == number, );
  GFS_ASSERT(blk == blocks, );


 fail:
  EXIT(GFN_BUILD_NOP_DESC);

  return 0;
}


/**
 * build_bhlist - Create buffers for a transaction
 * @sdp: The GFS superblock
 * @tr: The transaction
 *
 * Create the buffers required for a given transaction to be logged within
 * a unique descriptor chunk.  Add new buffers to the ordered bhlist
 * for the transaction.
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int build_bhlist(gfs_sbd_t *sdp, gfs_trans_t *tr)
{
  gfs_glock_t *gl;
  gfs_bufdata_t *bd;
  gfs_log_descriptor_t desc;
  gfs_block_tag_t tag;

  osi_list_head_t *gtmp, *btmp;
  osi_buf_t **cbh = NULL, *bh;
  unsigned int count_ctl = 0;
  int offset, index;
  int x, error = 0;
  char *data;

  ENTER(GFN_BUILD_BHLIST);


  /* figure out the number of regular (blk) buffers to be written and the number
     of control (ctl) buffers to be written */

  if (tr->tr_num_buf)
  {
    /* set up control buffers for descriptor and tags */

    count_ctl = gfs_struct2blk(sdp, tr->tr_num_buf, sizeof(gfs_block_tag_t));

    cbh = (osi_buf_t **)gmalloc(count_ctl * sizeof(osi_buf_t *));
    if (!cbh)
    {
      error = -ENOMEM;
      goto fail;
    }

    for (x = 0; x < count_ctl; x++)
    {
      error = journal_get_buffer(sdp, tr, &cbh[x]);
      if (error)
	goto fail;
    }


    /*  Set up metadata block list descriptor  */

    gfs_memset(&desc, 0, sizeof(gfs_log_descriptor_t));
    desc.ld_header.mh_magic = GFS_MAGIC;
    desc.ld_header.mh_type = GFS_METATYPE_LD;
    desc.ld_header.mh_format = GFS_FORMAT_LD;
    desc.ld_type = GFS_LOG_DESC_METADATA;
    desc.ld_length = count_ctl + tr->tr_num_buf;
    desc.ld_data1 = tr->tr_num_buf;
    desc.ld_data2 = tr->tr_type;
    gfs_desc_out(&desc, BH_DATA(cbh[0]));


    offset = sizeof(gfs_log_descriptor_t);
    index = 0;


    /*  Non-control buffers  */

    for (gtmp = tr->tr_gl.next; gtmp != &tr->tr_gl; gtmp = gtmp->next)
    {
      gl = osi_list_entry(gtmp, gfs_glock_t, gl_incore_list);

      for (btmp = gl->gl_incore_bufs.next; btmp != &gl->gl_incore_bufs; btmp = btmp->next)
      {
	bd = osi_list_entry(btmp, gfs_bufdata_t, bd_incore_list);


	/*  Write header if on boundary  */

	if (journal_segment_boundary(sdp, tr->tr_log_head))
	{	
	  error = journal_get_header(sdp, tr, FALSE, &bh);
	  if (error)
	    goto fail;
	}
      

	/*  Set up metadata buffer  */

	gfs_lock_buffer(bd->bd_bh);

	data = (bd->bd_frozen) ? bd->bd_frozen : BH_DATA(bd->bd_bh);

	error = journal_add_fakebh(sdp, tr, data, bd->bd_bh);
	if (error)
	{
	  gfs_unlock_buffer(bd->bd_bh);
	  goto fail;
	}


	/* fill in tag for this buffer */

	gfs_memset(&tag, 0, sizeof(tag));

	tag.bt_blkno = BH_BLKNO(bd->bd_bh);

      
	if (offset + sizeof(gfs_block_tag_t) > sdp->sd_sb.sb_bsize)
	{
	  index++;
	  offset = 0;
	}
	GFS_ASSERT(index < count_ctl, );

	gfs_tag_out(&tag, BH_DATA(cbh[index]) + offset);
      
	offset += sizeof(gfs_block_tag_t);     


	journal_incr_head(sdp, tr);
      }
    }

    gfree(cbh, count_ctl * sizeof(osi_buf_t *));
    cbh = NULL;
  }


  if (tr->tr_flags & TRF_NOPEN_DUMP)
  {
    error = build_nop_desc(sdp, tr, osi_atomic_read(&sdp->sd_nopen_od_count), GFS_LOG_DESC_NOPOPEN);
    if (error)
      goto fail;
  }
  else
  {
    if (tr->tr_num_nop_o)
    {
      error = build_nop_desc(sdp, tr, tr->tr_num_nop_o, GFS_LOG_DESC_NOPOPEN);
      if (error)
	goto fail;
    }
    if (tr->tr_num_nop_c)
    {
      error = build_nop_desc(sdp, tr, tr->tr_num_nop_c, GFS_LOG_DESC_NOPCLOSE);
      if (error)
	goto fail;
    }
  }


  GFS_ASSERT(!osi_list_empty(&tr->tr_bhlist), );


  EXIT(GFN_BUILD_BHLIST);

  return 0;



 fail:
  if (cbh)
    gfree(cbh, count_ctl * sizeof(osi_buf_t *));

  EXIT(GFN_BUILD_BHLIST);
  return error;
}


/**
 * check_seg_usage - Check that we didn't use too many segments
 * @sdp: The GFS superblock
 * @tr: The transaction
 *
 */

STATIC void check_seg_usage(gfs_sbd_t *sdp, gfs_trans_t *tr)
{
  int64 dist;
  unsigned int segments;

  ENTER(GFN_CHECK_SEG_USAGE);


  dist = journal_distance(sdp, tr->tr_log_head, tr->tr_first_head);

  segments = dist / sdp->sd_sb.sb_seg_size;
  GFS_ASSERT(segments * sdp->sd_sb.sb_seg_size == dist, );

  GFS_ASSERT(segments == tr->tr_seg_reserved, );
 

  EXIT(GFN_CHECK_SEG_USAGE);
}


/**
 * sync_trans - Add "last" descriptor to transaction and sync to disk
 * @sdp: The GFS superblock
 * @tr: The transaction
 *
 * Add the "last" descriptor on to the end of the current transaction
 * and sync it out to disk.  Don't commit it yet, though.
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int sync_trans(gfs_sbd_t *sdp, gfs_trans_t *tr)
{
  osi_list_head_t *tmp;
  gfs_log_descriptor_t desc;
  osi_buf_t *bh;
  gfs_bhlist_t *bl;
  uint64 blk;
  int error = 0;

  ENTER(GFN_SYNC_TRANS);


  /*  Build LAST descriptor  */

  error = journal_get_buffer(sdp, tr, &bh);
  if (error)
    goto fail;

  gfs_memset(&desc, 0, sizeof(gfs_log_descriptor_t));
  desc.ld_header.mh_magic = GFS_MAGIC;
  desc.ld_header.mh_type = GFS_METATYPE_LD;
  desc.ld_header.mh_format = GFS_FORMAT_LD;
  desc.ld_type = GFS_LOG_DESC_LAST;
  desc.ld_length = 1;
  for (blk = tr->tr_log_head; !journal_segment_boundary(sdp, blk); blk++)
    desc.ld_length++;
  gfs_desc_out(&desc, BH_DATA(bh));


  journal_next_segment(sdp, tr);

  check_seg_usage(sdp, tr);


  /*  Start I/O
      Go in "prev" direction to start the I/O in order.  (Does this really help?)  */

  for (tmp = tr->tr_bhlist.prev; tmp != &tr->tr_bhlist; tmp = tmp->prev)
  {
    bl = osi_list_entry(tmp, gfs_bhlist_t, bl_list);
    gfs_dwrite(sdp, bl->bl_bh, DW_DIRTY | DW_START);

    IFDB(GFN_SYNC_TRANS,
	 db_printf("  %"FMT_64"u:  0x%.8X\n",
		   BH_BLKNO(bl->bl_bh), gfs32_to_cpu(*(uint32 *)BH_DATA(bl->bl_bh))););
  }


  /*  Wait on I/O
      Go in "next" direction to minimize sleeps/wakeups.  */

  for (tmp = tr->tr_bhlist.next; tmp != &tr->tr_bhlist; tmp = tmp->next)
  {
    bl = osi_list_entry(tmp, gfs_bhlist_t, bl_list);

    error = gfs_dwrite(sdp, bl->bl_bh, DW_WAIT);
    if (error)
      goto fail;

    if (bl->bl_flags & BLF_FAKE)
      gfs_dispose_temp_bh(bl->bl_bh);
    else
      gfs_drelse(bl->bl_bh);
    bl->bl_bh = NULL;

    if (bl->bl_unlock)
    {
      gfs_unlock_buffer(bl->bl_unlock);
      bl->bl_unlock = NULL;
    }
  }


  EXIT(GFN_SYNC_TRANS);

  return 0;



 fail:
  EXIT(GFN_SYNC_TRANS);
  return error;
}


/**
 * commit_trans - Commit the current transaction
 * @sdp: The GFS superblock
 * @tr: The transaction
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int commit_trans(gfs_sbd_t *sdp, gfs_trans_t *tr)
{
  osi_buf_t *bh;
  int error = 0;

  ENTER(GFN_COMMIT_TRANS);


  /*  Write next header to commit  */

  error = journal_get_header(sdp, tr, TRUE, &bh);
  if (error)
  {
    EXIT(GFN_COMMIT_TRANS);
    return error;
  }

  error = gfs_dwrite(sdp, bh, DW_DIRTY | DW_START | DW_WAIT);
  gfs_drelse(bh);


  EXIT(GFN_COMMIT_TRANS);

  return error;

}


/**
 * disk_commit - Write a transaction to the on-disk journal
 * @sdp: The GFS superblock
 * @tr: The transaction
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int disk_commit(gfs_sbd_t *sdp, gfs_trans_t *tr)
{
  gfs_tr_nop_t *nop;
  uint64 last_nopd;
  int error = 0;

  ENTER(GFN_DISK_COMMIT);


  tr->tr_log_head = sdp->sd_log_head;
  tr->tr_first_head = tr->tr_log_head - 1;
  GFS_ASSERT(journal_segment_boundary(sdp, tr->tr_first_head), );
   

  error = build_bhlist(sdp, tr);
  if (error)
    goto out;

  error = sync_trans(sdp, tr);
  if (error)
    goto out;


  if (tr->tr_flags & TRF_NOPEN_DUMP)
  {
    /*  This commit header should point to the nopen dump we're
	commiting as the current one.  But save the copy of the
	old one in case we have problems commiting the dump.  */

    last_nopd = sdp->sd_last_nopd;
    sdp->sd_last_nopd = tr->tr_first_head;

    error = commit_trans(sdp, tr);
    if (error)
    {
      sdp->sd_last_nopd = last_nopd;
      goto out;
    }
  }
  else
  {
    error = commit_trans(sdp, tr);
    if (error)
      goto out;

    while (!osi_list_empty(&tr->tr_nop_change))
    {
      nop = osi_list_entry(tr->tr_nop_change.next, gfs_tr_nop_t, nop_list);
      osi_list_del(&nop->nop_list);
      gfs_nopen_merge(sdp, nop, &sdp->sd_nopen_od_list, &sdp->sd_nopen_od_count);
    }
  }


  sdp->sd_sequence++;
  sdp->sd_log_head = tr->tr_log_head;


 out:
  journal_free_bhlist(sdp, tr);

  EXIT(GFN_DISK_COMMIT);

  return error;
}


/**
 * increment_generation - Increment generation number on primary metadata
 * @tr: The transaction
 *
 * Increment the generation number on all primary metadata in the 
 * transaction.
 *
 *   Notice that gfs_dinode_out() and gfs_rgrp_out() don't copy out
 *   the generation number.  This is why we can get away with just
 *   incrementing it in the buffer and not in any other copies we
 *   have.  I know I'm going to forget about this and it'll bite me
 *   in the butt.
 *
 *   We also have to be careful to catch the frozen copy if there is one.
 *
 */

STATIC void increment_generation(gfs_trans_t *tr)
{
  osi_list_head_t *gtmp, *btmp;
  gfs_glock_t *gl;
  gfs_bufdata_t *bd;

  gfs_meta_header_t *mh, *mh2;
  uint64 tmp64;

  ENTER(GFN_INCREMENT_GENERATION);


  for (gtmp = tr->tr_gl.next; gtmp != &tr->tr_gl; gtmp = gtmp->next)
  {
    gl = osi_list_entry(gtmp, gfs_glock_t, gl_incore_list);

    for (btmp = gl->gl_incore_bufs.next; btmp != &gl->gl_incore_bufs; btmp = btmp->next)
    {
      bd = osi_list_entry(btmp, gfs_bufdata_t, bd_incore_list);

      gfs_meta_check(bd->bd_bh);


      gfs_lock_buffer(bd->bd_bh);

      mh = (gfs_meta_header_t *)BH_DATA(bd->bd_bh);

      tmp64 = gfs64_to_cpu(mh->mh_generation) + 1;
      tmp64 = cpu_to_gfs64(tmp64);

      if (bd->bd_frozen)
      {
	mh2 = (gfs_meta_header_t *)bd->bd_frozen;
	GFS_ASSERT(mh->mh_generation == mh2->mh_generation, );
	mh2->mh_generation = tmp64;
      }
      mh->mh_generation = tmp64;

      gfs_unlock_buffer(bd->bd_bh);
    }
  }


  EXIT(GFN_INCREMENT_GENERATION);
}


/**
 * add_trans_to_ail - Add a ondisk commited transaction to the AIL
 * @sdp: the filesystem
 * @new_tr: the transaction 
 *
 */

STATIC void add_trans_to_ail(gfs_sbd_t *sdp, gfs_trans_t *new_tr)
{
  gfs_glock_t *gl;
  gfs_bufdata_t *bd;

  ENTER(GFN_ADD_TRANS_TO_AIL);


  GFS_ASSERT(sdp->sd_log_buffers >= new_tr->tr_num_buf, );
  sdp->sd_log_buffers -= new_tr->tr_num_buf;


  gfs_trans_prep_for_ail(sdp, new_tr);


  while (!osi_list_empty(&new_tr->tr_gl))
  {
    gl = osi_list_entry(new_tr->tr_gl.next, gfs_glock_t, gl_incore_list);

    osi_list_del(&gl->gl_incore_list);
    osi_list_head_init(&gl->gl_incore_list);
    gl->gl_incore_tr = NULL;

    while (!osi_list_empty(&gl->gl_incore_bufs))
    {
      bd = osi_list_entry(gl->gl_incore_bufs.next, gfs_bufdata_t, bd_incore_list);

      osi_list_del(&bd->bd_incore_list);
      osi_list_head_init(&bd->bd_incore_list);

      gfs_dunpin(bd->bd_bh, new_tr);
    }
  }


  osi_list_add(&new_tr->tr_list, &sdp->sd_log_ail);


  EXIT(GFN_ADD_TRANS_TO_AIL);
}


/**
 * trans_combine - combine two transactions
 * @tr: the surviving transaction
 * @new_tr: the transaction that gets freed
 *
 * Assumes that the two transactions are independent.
 */

STATIC void trans_combine(gfs_trans_t *tr, gfs_trans_t *new_tr)
{
  gfs_glock_t *gl;
  gfs_tr_nop_t *nop;

  ENTER(GFN_TRANS_COMBINE);


  tr->tr_type = TRANS_COMPOUND;
  tr->tr_seg_reserved += new_tr->tr_seg_reserved;
  tr->tr_flags |= new_tr->tr_flags;
  tr->tr_num_gl += new_tr->tr_num_gl;
  tr->tr_num_buf += new_tr->tr_num_buf;
  tr->tr_num_nop_o += new_tr->tr_num_nop_o;
  tr->tr_num_nop_c += new_tr->tr_num_nop_c;


  /*  Combine the glocks in the two transactions  */

  while (!osi_list_empty(&new_tr->tr_gl))
  {
    gl = osi_list_entry(new_tr->tr_gl.next, gfs_glock_t, gl_incore_list);

    osi_list_del(&gl->gl_incore_list);
    osi_list_add(&gl->gl_incore_list, &tr->tr_gl);

    GFS_ASSERT(gl->gl_incore_tr == new_tr, );
    gl->gl_incore_tr = tr;
  }


  /*  Combine the nopen changes in the two transactions  */

  while (!osi_list_empty(&new_tr->tr_nop_change))
  {
    nop = osi_list_entry(new_tr->tr_nop_change.next, gfs_tr_nop_t, nop_list);
    osi_list_del(&nop->nop_list);
    osi_list_add(&nop->nop_list, &tr->tr_nop_change);
  }


  gfree(new_tr, sizeof(gfs_trans_t));


  EXIT(GFN_TRANS_COMBINE);
}


/**
 * gfs_log_flush - Flush all incore commited transactions for a glock
 * @sdp:  the filesystem
 * @gl:  the glock structure to flush
 *
 * We have to be careful about rename transactions.  If the transaction
 * We're looking for contains a rename, we must flush all other transactions
 * that contain renames prior to it.
 *
 * if @gl is NULL, flush all transactions
 */

void gfs_log_flush(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  gfs_trans_t *trans = NULL, *tr;
  int error;

  ENTER(GFN_LOG_FLUSH);


  osi_sema_lock(&sdp->sd_log_lock);


  if (osi_list_empty(&sdp->sd_log_incore))
    goto out;


  if (gl)
  {
    if (!gl->gl_incore_tr)
      goto out;

    trans = gl->gl_incore_tr;

    osi_list_del(&trans->tr_list);
  }
  else
  {
    while (!osi_list_empty(&sdp->sd_log_incore))
    {
      tr = osi_list_entry(sdp->sd_log_incore.next, gfs_trans_t, tr_list);

      osi_list_del(&tr->tr_list);

      if (trans)
	trans_combine(trans, tr);
      else
	trans = tr;
    }
  }


  GFS_ASSERT(trans, );

  log_refund(sdp, trans);

  IFDB(GFN_LOG_FLUSH,
       gfs_trans_print(sdp, trans, TRANS_IS_INCORE););


  /*  Actually do the stuff to commit the transaction  */

  increment_generation(trans);

  for (;;)
  {
    error = disk_commit(sdp, trans);
    if (!error)
      break;

    con_printf("GFS:  error (%d) commiting meta transaction (retrying)\n", error);
    osi_sleep_unint(1);
  }

  add_trans_to_ail(sdp, trans);


 out:
  osi_sema_unlock(&sdp->sd_log_lock);


  /*  Dump out an nopen entry if we need to.  */

  if (!sdp->sd_in_nopen_dump)
    gfs_nopen_dump(sdp, FALSE);


  EXIT(GFN_LOG_FLUSH);
}


/**
 * incore_nopen_merge - call gfs_nopen_merge() for this incore commit
 * @sdp: the filesystem
 * @new_tr: the new transaction
 *
 */

STATIC void incore_nopen_merge(gfs_sbd_t *sdp, gfs_trans_t *new_tr)
{
  gfs_tr_nop_t *nop;

  ENTER(GFN_INCORE_NOPEN_MERGE);

  while (!osi_list_empty(&new_tr->tr_nop_change_ic))
  {
    nop = osi_list_entry(new_tr->tr_nop_change_ic.next, gfs_tr_nop_t, nop_list);
    osi_list_del(&nop->nop_list);
    gfs_nopen_merge(sdp, nop, &sdp->sd_nopen_ic_list, &sdp->sd_nopen_ic_count);
  }

  EXIT(GFN_INCORE_NOPEN_MERGE);
}


/**
 * free_frozen - free a transaction's frozen data
 * @tr: the transaction
 *
 */

STATIC void free_frozen(gfs_trans_t *tr)
{
  gfs_glock_t *gl;
  gfs_bufdata_t *bd;
  osi_list_head_t *gtmp, *btmp;

  ENTER(GFN_FREE_FROZEN);


  for (gtmp = tr->tr_gl.next; gtmp != &tr->tr_gl; gtmp = gtmp->next)
  {
    gl = osi_list_entry(gtmp, gfs_glock_t, gl_new_list);

    for (btmp = gl->gl_new_bufs.next; btmp != &gl->gl_new_bufs; btmp = btmp->next)
    {
      bd = osi_list_entry(btmp, gfs_bufdata_t, bd_new_list);

      GFS_ASSERT(bd->bd_frozen, );
      gfree(bd->bd_frozen, BH_SIZE(bd->bd_bh));
      bd->bd_frozen = NULL;
    }
  }  


  EXIT(GFN_FREE_FROZEN);
}


/**
 * buf_move - the buffers on a glock of new and incore transactions
 * @sdp: the filesystem
 * @tr: the transaction this glock belongs to
 * @gl: the glock that holds the buffers
 *
 */

STATIC void buf_move(gfs_sbd_t *sdp, gfs_trans_t *tr, gfs_glock_t *gl)
{
  gfs_bufdata_t *bd;

  ENTER(GFN_BUF_MOVE);


  while (!osi_list_empty(&gl->gl_new_bufs))
  {
    bd = osi_list_entry(gl->gl_new_bufs.next, gfs_bufdata_t, bd_new_list);

    osi_list_del(&bd->bd_new_list);
    osi_list_head_init(&bd->bd_new_list);

    if (!osi_list_empty(&bd->bd_incore_list))
    {
      gfs_dunpin(bd->bd_bh, NULL);

      GFS_ASSERT(tr->tr_num_buf, );
      tr->tr_num_buf--;

      GFS_ASSERT(sdp->sd_log_buffers, );
      sdp->sd_log_buffers--;
    }
    else
      osi_list_add(&bd->bd_incore_list, &gl->gl_incore_bufs);
  }


  EXIT(GFN_BUF_MOVE);
}


/**
 * nop_move - merge the nopen changes in new_tr into trans
 * @sdp: the filesystem
 * @trans: the surviving transaction
 * @new_tr: the vanishing transaction
 *
 */

STATIC void nop_move(gfs_sbd_t *sdp, gfs_trans_t *trans, gfs_trans_t *new_tr)
{
  osi_list_head_t *ntmp;
  gfs_tr_nop_t *nop, *new_nop;

  ENTER(GFN_NOP_MOVE);


  while (!osi_list_empty(&new_tr->tr_nop_change))
  {
    new_nop = osi_list_entry(new_tr->tr_nop_change.next, gfs_tr_nop_t, nop_list);
    osi_list_del(&new_nop->nop_list);

    nop = NULL;
    for (ntmp = trans->tr_nop_change.next; ntmp != &trans->tr_nop_change; ntmp = ntmp->next)
    {
      nop = osi_list_entry(ntmp, gfs_tr_nop_t, nop_list);

      if (GFS_INUM_EQUAL(&nop->nop_inum, &new_nop->nop_inum))
	break;
      else
	nop = NULL;
    }

    if (nop)
    {
      GFS_ASSERT(nop->nop_type != new_nop->nop_type,
		 con_printf("nop->nop_type = %u, new_nop->nop_type = %u\n",
			    nop->nop_type, new_nop->nop_type);
		 con_printf("nop->nop_inum = %"FMT_64"u/%"FMT_64"u\n",
			    nop->nop_inum.no_formal_ino, nop->nop_inum.no_addr);
		 con_printf("new_tr->tr_type = %u\n", new_tr->tr_type););

      gfree(new_nop, sizeof(gfs_tr_nop_t));

      osi_list_del(&nop->nop_list);
      gfree(nop, sizeof(gfs_tr_nop_t));

      GFS_ASSERT(trans->tr_num_nop_o, );
      trans->tr_num_nop_o--;
      GFS_ASSERT(trans->tr_num_nop_c, );
      trans->tr_num_nop_c--;
    }
    else
      osi_list_add(&new_nop->nop_list, &trans->tr_nop_change);
  }


  EXIT(GFN_NOP_MOVE);
}


/**
 * incore_commit - commit a transaction in-core
 * @sdp: the filesystem
 * @new_tr: the transaction to commit
 *
 * Add the transaction @new_tr to the end of the incore commit list.
 * Pull up and merge an previously commited transactions that share
 * locks.  Also pull up any rename transactions that need it.
 */

STATIC void incore_commit(gfs_sbd_t *sdp, gfs_trans_t *new_tr)
{
  osi_list_head_t *gtmp;
  osi_list_head_t moved;
  gfs_trans_t *trans = NULL, *exist_tr;
  gfs_glock_t *gl;

  ENTER(GFN_INCORE_COMMIT);

  osi_sema_lock(&sdp->sd_log_lock);


  incore_nopen_merge(sdp, new_tr);

  free_frozen(new_tr);

  sdp->sd_log_buffers += new_tr->tr_num_buf;


  for (gtmp = new_tr->tr_gl.next; gtmp != &new_tr->tr_gl; gtmp = gtmp->next)
  {
    gl = osi_list_entry(gtmp, gfs_glock_t, gl_new_list);

    exist_tr = gl->gl_incore_tr;
    if (!exist_tr)
    {
      GFS_ASSERT(osi_list_empty(&gl->gl_incore_list), );
      continue;
    }
    else
      GFS_ASSERT(!osi_list_empty(&gl->gl_incore_list), );

    if (exist_tr != trans)
    {
      osi_list_del(&exist_tr->tr_list);
      if (trans)
	trans_combine(trans, exist_tr);
      else
	trans = exist_tr;
    }
  }


  if (trans)
  {
    trans->tr_type = TRANS_COMPOUND;
    trans->tr_seg_reserved += new_tr->tr_seg_reserved;
    trans->tr_flags |= new_tr->tr_flags;
    trans->tr_num_gl += new_tr->tr_num_gl;
    trans->tr_num_buf += new_tr->tr_num_buf;
    trans->tr_num_nop_o += new_tr->tr_num_nop_o;
    trans->tr_num_nop_c += new_tr->tr_num_nop_c;

    while (!osi_list_empty(&new_tr->tr_gl))
    {
      gl = osi_list_entry(new_tr->tr_gl.next, gfs_glock_t, gl_new_list);

      osi_list_del(&gl->gl_new_list);
      osi_list_head_init(&gl->gl_new_list);
      gl->gl_new_tr = NULL;

      if (gl->gl_incore_tr)
      {
	GFS_ASSERT(gl->gl_incore_tr == trans, );
	GFS_ASSERT(!osi_list_empty(&gl->gl_incore_list), );
	/*  gl_incore_bufs doesn't always contain buffers.  */

	GFS_ASSERT(trans->tr_num_gl, );
	trans->tr_num_gl--;
      }
      else
      {
	GFS_ASSERT(osi_list_empty(&gl->gl_incore_list), );
	GFS_ASSERT(osi_list_empty(&gl->gl_incore_bufs), );

	osi_list_add(&gl->gl_incore_list, &trans->tr_gl);
	gl->gl_incore_tr = trans;
      }

      buf_move(sdp, trans, gl);
    }

    nop_move(sdp, trans, new_tr);

    gfree(new_tr, sizeof(gfs_trans_t));
  }
  else
  {
    osi_list_head_init(&moved);

    gl = NULL;

    while (!osi_list_empty(&new_tr->tr_gl))
    {
      gl = osi_list_entry(new_tr->tr_gl.next, gfs_glock_t, gl_new_list);

      osi_list_del(&gl->gl_new_list);
      osi_list_head_init(&gl->gl_new_list);
      gl->gl_new_tr = NULL;

      GFS_ASSERT(!gl->gl_incore_tr, );
      GFS_ASSERT(osi_list_empty(&gl->gl_incore_list), );
      GFS_ASSERT(osi_list_empty(&gl->gl_incore_bufs), );

      osi_list_add(&gl->gl_incore_list, &moved);
      gl->gl_incore_tr = new_tr;

      buf_move(sdp, new_tr, gl);
    }

    osi_list_del(&moved);
    osi_list_add(&new_tr->tr_gl, &gl->gl_incore_list);

    trans = new_tr;
  }


  log_refund(sdp, trans);
  osi_list_add(&trans->tr_list, &sdp->sd_log_incore);


  osi_sema_unlock(&sdp->sd_log_lock);

  EXIT(GFN_INCORE_COMMIT);
}


/**
 * gfs_log_commit - Commit a transaction to the log
 * @sdp: the filesystem
 * @tr: the transaction
 *
 * Returns: 0 on success, -EXXX on failure
 */

void gfs_log_commit(gfs_sbd_t *sdp, gfs_trans_t *tr)
{
  ENTER(GFN_LOG_COMMIT);

  incore_commit(sdp, tr);

  osi_sema_lock(&sdp->sd_log_lock);
  if (sdp->sd_log_buffers > 1024)
  {
    osi_sema_unlock(&sdp->sd_log_lock);
    gfs_log_flush(sdp, NULL);
    gfs_pull_tail(sdp);
  }
  else
  {
    pull_tail_locked(sdp);
    osi_sema_unlock(&sdp->sd_log_lock);
  }

  EXIT(GFN_LOG_COMMIT);
}


/**
 * gfs_nopen_dump - make an Nopen Dump entry in the log
 * @sdp: the filesystem
 * @force: if TRUE, always make the dump even if one has been made recently
 *
 */

void gfs_nopen_dump(gfs_sbd_t *sdp, int force)
{
  gfs_trans_t tr;
  unsigned int blocks, segments, reserved_segments = 0;
  int have_res = FALSE, need_dump;
  int64 dist;
  int error;

  ENTER(GFN_NOPEN_DUMP);


  sdp->sd_in_nopen_dump = TRUE;


  for (;;)
  {
    osi_sema_lock(&sdp->sd_log_lock);

    dist = journal_distance(sdp, sdp->sd_log_head, sdp->sd_last_nopd);
    need_dump = (dist * GFS_NOP_DUMPS_PER_LOG >= sdp->sd_jdesc.ji_nsegment * sdp->sd_sb.sb_seg_size);

    blocks = gfs_struct2blk(sdp, osi_atomic_read(&sdp->sd_nopen_od_count), sizeof(gfs_inum_t));
    blocks++;  /*  last descriptor  */

    segments = gfs_blk2seg(sdp, blocks);

    if ((force || need_dump) && have_res && (reserved_segments >= segments))
    {
      /*  Refund any uneeded segments  */

      if (reserved_segments > segments)
      {
	sdp->sd_log_seg_free += reserved_segments - segments;
	GFS_ASSERT(sdp->sd_log_seg_free < sdp->sd_jdesc.ji_nsegment, );
      }

      gfs_memset(&tr, 0, sizeof(gfs_trans_t));
      osi_list_head_init(&tr.tr_bhlist);
      tr.tr_seg_reserved = segments;
      tr.tr_flags = TRF_NOPEN_DUMP;

      for (;;)
      {
	error = disk_commit(sdp, &tr);
	if (!error)
	{
	  /*  If there isn't anything the AIL, we won't get back the log
	      space we reserved unless we do it ourselves.  */

	  if (osi_list_empty(&sdp->sd_log_ail))
	  {
	    sdp->sd_log_seg_free += tr.tr_seg_reserved;
	    GFS_ASSERT(sdp->sd_log_seg_free < sdp->sd_jdesc.ji_nsegment, );
	  }

	  osi_sema_unlock(&sdp->sd_log_lock);
	  goto out;
	}

	con_printf("GFS:  error (%d) commiting nopen transaction (retrying)\n", error);
	osi_sleep_unint(1);
      }
    }

    osi_sema_unlock(&sdp->sd_log_lock);


    if (have_res)
    {
      gfs_log_relse(sdp, reserved_segments);
      have_res = FALSE;
    }


    if (force || need_dump)
    {
      error = gfs_log_reserv(sdp, segments);
      GFS_ASSERT(!error, );  /*  What next?  */

      have_res = TRUE;
      reserved_segments = segments;
    }
    else
      goto out;
  }


 out:
  sdp->sd_in_nopen_dump = FALSE;

  EXIT(GFN_NOPEN_DUMP);
}


/**
 * gfs_shutdown_log - write a shutdown header into a journal
 * @sdp: the filesystem
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_shutdown_log(gfs_sbd_t *sdp)
{
  gfs_jindex_t *jdesc = &sdp->sd_jdesc;
  gfs_log_header_t head;
  gfs_log_descriptor_t desc;
  osi_buf_t *bh;
  int error = 0;

  ENTER(GFN_SHUTDOWN_LOG);

  osi_sema_lock(&sdp->sd_log_lock);


  GFS_ASSERT(osi_list_empty(&sdp->sd_log_ail), );
  GFS_ASSERT(sdp->sd_log_seg_free == sdp->sd_jdesc.ji_nsegment - 1, );
  GFS_ASSERT(!sdp->sd_log_buffers, );
  GFS_ASSERT(journal_segment_boundary(sdp, sdp->sd_log_head - 1), );


  /*  Build a "last" log descriptor  */

  gfs_memset(&desc, 0, sizeof(gfs_log_descriptor_t));
  desc.ld_header.mh_magic = GFS_MAGIC;
  desc.ld_header.mh_type = GFS_METATYPE_LD;
  desc.ld_header.mh_format = GFS_FORMAT_LD;
  desc.ld_type = GFS_LOG_DESC_LAST;
  desc.ld_length = sdp->sd_sb.sb_seg_size - 1;


  /*  Write the descriptor  */

  bh = gfs_dgetblk(sdp, sdp->sd_log_head, sdp->sd_journal_gl);
  if (!bh)
  {
    error = -EIO;
    goto out;
  }
  gfs_prep_new_buffer(bh);

  gfs_buffer_clear(bh);
  gfs_desc_out(&desc, BH_DATA(bh));
  error = gfs_dwrite(sdp, bh, DW_DIRTY | DW_START | DW_WAIT);
  gfs_drelse(bh);
  if (error)
    goto out;


  /*  Move to the next header  */

  while (!journal_segment_boundary(sdp, sdp->sd_log_head))
  {
    if (++sdp->sd_log_head == jdesc->ji_addr + jdesc->ji_nsegment * sdp->sd_sb.sb_seg_size)
      sdp->sd_log_head = jdesc->ji_addr;
  }


  /*  Build the shutdown header  */

  gfs_memset(&head, 0, sizeof(gfs_log_header_t));
  head.lh_header.mh_magic = GFS_MAGIC;
  head.lh_header.mh_type = GFS_METATYPE_LH;
  head.lh_header.mh_format = GFS_FORMAT_LH;
  head.lh_flags = GFS_LOG_HEAD_UNMOUNT;
  head.lh_first = sdp->sd_log_head;
  head.lh_sequence = sdp->sd_sequence + 1;
  /*  Don't care about tail  */
  head.lh_last_nopd = (osi_atomic_read(&sdp->sd_nopen_od_count)) ? sdp->sd_last_nopd : 0;


  /*  Write out the shutdown header  */

  bh = gfs_dgetblk(sdp, head.lh_first, sdp->sd_journal_gl);
  if (!bh)
  {
    error = -EIO;
    goto out;
  }
  gfs_prep_new_buffer(bh);

  gfs_buffer_clear(bh);
  gfs_log_header_out(&head, BH_DATA(bh));
  gfs_log_header_out(&head, BH_DATA(bh) + GFS_BASIC_BLOCK - sizeof(gfs_log_header_t));
  error = gfs_dwrite(sdp, bh, DW_DIRTY | DW_START | DW_WAIT);
  gfs_drelse(bh);


 out:
  osi_sema_unlock(&sdp->sd_log_lock);

  EXIT(GFN_SHUTDOWN_LOG);

  return error;
}



