/******************************************************************************
*******************************************************************************
**
**    Adapted from linux/mm/mmap_avl.c in the Linux kernel source written
**    by Bruno Haible <haible@ma2s2.mathematik.uni-karlsruhe.de>. Original
**    comments preserved below.
**
**    Portions Copyright 2000-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Steven Whitehouse <steve@sistina.com>
** 
**    Module name: rg_avl.c
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"

/*
 * Searching a RG in the linear list sb->sd_rglist is horribly slow.
 * Use an AVL (Adelson-Velskii and Landis) tree to speed up this search
 * from O(n) to O(log n), where n is the number of RGs in the fs
 *
 * Written by Bruno Haible <haible@ma2s2.mathematik.uni-karlsruhe.de>.
 * Adapted for use in GFS by Steve Whitehouse <steve@sistina.com>
 */

/* We keep the list and tree sorted by address. */
#define rd_avl_key	rd_ri.ri_addr
#define rd_avl_key_t	uint64	/* typeof(rg->rd_key) */

/*
 * sb->sb_rgavl is the AVL tree corresponding to sb->sb_rglist
 * or, more exactly, its root.
 * A  struct gfs_rgrpd has the following fields:
 *   rd_avl_left     left son of a tree node
 *   rd_avl_right    right son of a tree node
 *   rd_avl_height   1+max(heightof(left),heightof(right))
 * The empty tree is represented as NULL.
 */

/* Since the trees are balanced, their height will never be large. */
#define avl_maxheight	64	/* Probably this can be made smaller */
#define heightof(tree)	((tree) == rd_avl_empty ? 0 : (tree)->rd_avl_height)

/*
 * Consistency and balancing rules:
 * 1. tree->rd_avl_height == 1+max(heightof(tree->rd_avl_left),heightof(tree->rd_avl_right))
 * 2. abs( heightof(tree->rd_avl_left) - heightof(tree->rd_avl_right) ) <= 1
 * 3. foreach node in tree->rd_avl_left: node->rd_avl_key <= tree->rd_avl_key,
 *    foreach node in tree->rd_avl_right: node->rd_avl_key >= tree->rd_avl_key.
 */

#ifdef DEBUG_AVL

/* Look up the nodes at the left and at the right of a given node. */
static void avl_neighbours (struct gfs_rgrpd * node, struct gfs_rgrpd * tree, struct gfs_rgrpd ** to_the_left, struct gfs_rgrpd ** to_the_right)
{
	rd_avl_key_t key = node->rd_avl_key;

	*to_the_left = *to_the_right = NULL;
	for (;;) {
		if (tree == rd_avl_empty) {
			printk("avl_neighbours: node not found in the tree\n");
			return;
		}
		if (key == tree->rd_avl_key)
			break;
		if (key < tree->rd_avl_key) {
			*to_the_right = tree;
			tree = tree->rd_avl_left;
		} else {
			*to_the_left = tree;
			tree = tree->rd_avl_right;
		}
	}
	if (tree != node) {
		printk("avl_neighbours: node not exactly found in the tree\n");
		return;
	}
	if (tree->rd_avl_left != rd_avl_empty) {
		struct gfs_rgrpd * node;
		for (node = tree->rd_avl_left; node->rd_avl_right != rd_avl_empty; node = node->rd_avl_right)
			continue;
		*to_the_left = node;
	}
	if (tree->rd_avl_right != rd_avl_empty) {
		struct gfs_rgrpd * node;
		for (node = tree->rd_avl_right; node->rd_avl_left != rd_avl_empty; node = node->rd_avl_left)
			continue;
		*to_the_right = node;
	}
	if ((*to_the_left && ((*to_the_left)->rd_list.next != node)) || (node->rd_list.next != *to_the_right))
		printk("avl_neighbours: tree inconsistent with list\n");
}

#endif

/*
 * Rebalance a tree.
 * After inserting or deleting a node of a tree we have a sequence of subtrees
 * nodes[0]..nodes[k-1] such that
 * nodes[0] is the root and nodes[i+1] = nodes[i]->{rd_avl_left|rd_avl_right}.
 */
static void avl_rebalance (struct gfs_rgrpd *** nodeplaces_ptr, int count)
{
	for ( ; count > 0 ; count--) {
		struct gfs_rgrpd ** nodeplace = *--nodeplaces_ptr;
		struct gfs_rgrpd * node = *nodeplace;
		struct gfs_rgrpd * nodeleft = node->rd_avl_left;
		struct gfs_rgrpd * noderight = node->rd_avl_right;
		int heightleft = heightof(nodeleft);
		int heightright = heightof(noderight);
		if (heightright + 1 < heightleft) {
			/*                                                      */
			/*                            *                         */
			/*                          /   \                       */
			/*                       n+2      n                     */
			/*                                                      */
			struct gfs_rgrpd * nodeleftleft = nodeleft->rd_avl_left;
			struct gfs_rgrpd * nodeleftright = nodeleft->rd_avl_right;
			int heightleftright = heightof(nodeleftright);
			if (heightof(nodeleftleft) >= heightleftright) {
				/*                                                        */
				/*                *                    n+2|n+3            */
				/*              /   \                  /    \             */
				/*           n+2      n      -->      /   n+1|n+2         */
				/*           / \                      |    /    \         */
				/*         n+1 n|n+1                 n+1  n|n+1  n        */
				/*                                                        */
				node->rd_avl_left = nodeleftright; nodeleft->rd_avl_right = node;
				nodeleft->rd_avl_height = 1 + (node->rd_avl_height = 1 + heightleftright);
				*nodeplace = nodeleft;
			} else {
				/*                                                        */
				/*                *                     n+2               */
				/*              /   \                 /     \             */
				/*           n+2      n      -->    n+1     n+1           */
				/*           / \                    / \     / \           */
				/*          n  n+1                 n   L   R   n          */
				/*             / \                                        */
				/*            L   R                                       */
				/*                                                        */
				nodeleft->rd_avl_right = nodeleftright->rd_avl_left;
				node->rd_avl_left = nodeleftright->rd_avl_right;
				nodeleftright->rd_avl_left = nodeleft;
				nodeleftright->rd_avl_right = node;
				nodeleft->rd_avl_height = node->rd_avl_height = heightleftright;
				nodeleftright->rd_avl_height = heightleft;
				*nodeplace = nodeleftright;
			}
		}
		else if (heightleft + 1 < heightright) {
			/* similar to the above, just interchange 'left' <--> 'right' */
			struct gfs_rgrpd * noderightright = noderight->rd_avl_right;
			struct gfs_rgrpd * noderightleft = noderight->rd_avl_left;
			int heightrightleft = heightof(noderightleft);
			if (heightof(noderightright) >= heightrightleft) {
				node->rd_avl_right = noderightleft; noderight->rd_avl_left = node;
				noderight->rd_avl_height = 1 + (node->rd_avl_height = 1 + heightrightleft);
				*nodeplace = noderight;
			} else {
				noderight->rd_avl_left = noderightleft->rd_avl_right;
				node->rd_avl_right = noderightleft->rd_avl_left;
				noderightleft->rd_avl_right = noderight;
				noderightleft->rd_avl_left = node;
				noderight->rd_avl_height = node->rd_avl_height = heightrightleft;
				noderightleft->rd_avl_height = heightright;
				*nodeplace = noderightleft;
			}
		}
		else {
			int height = (heightleft<heightright ? heightright : heightleft) + 1;
			if (height == node->rd_avl_height)
				break;
			node->rd_avl_height = height;
		}
	}
}

/* Insert a node into a tree. */
static inline void avl_insert (struct gfs_rgrpd * new_node, struct gfs_rgrpd ** ptree)
{
	rd_avl_key_t key = new_node->rd_avl_key;
	struct gfs_rgrpd ** nodeplace = ptree;
	struct gfs_rgrpd ** stack[avl_maxheight];
	int stack_count = 0;
	struct gfs_rgrpd *** stack_ptr = &stack[0]; /* = &stack[stackcount] */
	for (;;) {
		struct gfs_rgrpd * node = *nodeplace;
		if (node == rd_avl_empty)
			break;
		*stack_ptr++ = nodeplace; stack_count++;
		if (key < node->rd_avl_key)
			nodeplace = &node->rd_avl_left;
		else
			nodeplace = &node->rd_avl_right;
	}
	new_node->rd_avl_left = rd_avl_empty;
	new_node->rd_avl_right = rd_avl_empty;
	new_node->rd_avl_height = 1;
	*nodeplace = new_node;
	avl_rebalance(stack_ptr,stack_count);
}

/* Insert a node into a tree, and
 * return the node to the left of it and the node to the right of it.
 */
static inline void avl_insert_neighbours (struct gfs_rgrpd * new_node, struct gfs_rgrpd ** ptree,
	struct gfs_rgrpd ** to_the_left, struct gfs_rgrpd ** to_the_right)
{
	rd_avl_key_t key = new_node->rd_avl_key;
	struct gfs_rgrpd ** nodeplace = ptree;
	struct gfs_rgrpd ** stack[avl_maxheight];
	int stack_count = 0;
	struct gfs_rgrpd *** stack_ptr = &stack[0]; /* = &stack[stackcount] */
	*to_the_left = *to_the_right = NULL;
	for (;;) {
		struct gfs_rgrpd * node = *nodeplace;
		if (node == rd_avl_empty)
			break;
		*stack_ptr++ = nodeplace; stack_count++;
		if (key < node->rd_avl_key) {
			*to_the_right = node;
			nodeplace = &node->rd_avl_left;
		} else {
			*to_the_left = node;
			nodeplace = &node->rd_avl_right;
		}
	}
	new_node->rd_avl_left = rd_avl_empty;
	new_node->rd_avl_right = rd_avl_empty;
	new_node->rd_avl_height = 1;
	*nodeplace = new_node;
	avl_rebalance(stack_ptr,stack_count);
}

/* Removes a node out of a tree. */
static void avl_remove (struct gfs_rgrpd * node_to_delete, struct gfs_rgrpd ** ptree)
{
	rd_avl_key_t key = node_to_delete->rd_avl_key;
	struct gfs_rgrpd ** nodeplace = ptree;
	struct gfs_rgrpd ** stack[avl_maxheight];
	int stack_count = 0;
	struct gfs_rgrpd *** stack_ptr = &stack[0]; /* = &stack[stackcount] */
	struct gfs_rgrpd ** nodeplace_to_delete;
	for (;;) {
		struct gfs_rgrpd * node = *nodeplace;
#ifdef DEBUG_AVL
		if (node == rd_avl_empty) {
			/* what? node_to_delete not found in tree? */
			printk("avl_remove: node to delete not found in tree\n");
			return;
		}
#endif
		*stack_ptr++ = nodeplace; stack_count++;
		if (key == node->rd_avl_key)
			break;
		if (key < node->rd_avl_key)
			nodeplace = &node->rd_avl_left;
		else
			nodeplace = &node->rd_avl_right;
	}
	nodeplace_to_delete = nodeplace;
	/* Have to remove node_to_delete = *nodeplace_to_delete. */
	if (node_to_delete->rd_avl_left == rd_avl_empty) {
		*nodeplace_to_delete = node_to_delete->rd_avl_right;
		stack_ptr--; stack_count--;
	} else {
		struct gfs_rgrpd *** stack_ptr_to_delete = stack_ptr;
		struct gfs_rgrpd ** nodeplace = &node_to_delete->rd_avl_left;
		struct gfs_rgrpd * node;
		for (;;) {
			node = *nodeplace;
			if (node->rd_avl_right == rd_avl_empty)
				break;
			*stack_ptr++ = nodeplace; stack_count++;
			nodeplace = &node->rd_avl_right;
		}
		*nodeplace = node->rd_avl_left;
		/* node replaces node_to_delete */
		node->rd_avl_left = node_to_delete->rd_avl_left;
		node->rd_avl_right = node_to_delete->rd_avl_right;
		node->rd_avl_height = node_to_delete->rd_avl_height;
		*nodeplace_to_delete = node; /* replace node_to_delete */
		*stack_ptr_to_delete = &node->rd_avl_left; /* replace &node_to_delete->rd_avl_left */
	}
	avl_rebalance(stack_ptr,stack_count);
}

#ifdef DEBUG_AVL

/* print a list */
static void printk_list (struct osi_list_head *ptr)
{
	struct osi_list_head *tmp, *head;
	struct gfs_rgrpd *rd;

	printk("[");
	tmp = head = ptr;
	for(;;) {
		tmp = tmp->next;
		if (tmp == head)
			break;
		rd = osi_list_entry(tmp, struct gfs_rgrpd, rd_list);
		printk("%Lu(%lu)", rd->rd_avl_key, rd->rd_ri.ri_length);
		printk(" ");
	}
	printk("]");
}

/* print a tree */
static void printk_avl (struct gfs_rgrpd * tree)
{
	if (tree != rd_avl_empty) {
		printk("(");
		if (tree->rd_avl_left != rd_avl_empty) {
			printk_avl(tree->rd_avl_left);
			printk("<");
		}
		printk("%Lu(%lu)", tree->rd_avl_key, tree->rd_ri.ri_length);
		if (tree->rd_avl_right != rd_avl_empty) {
			printk(">");
			printk_avl(tree->rd_avl_right);
		}
		printk(")");
	}
}

static char *avl_check_point = "somewhere";

/* check a tree's consistency and balancing */
static void avl_checkheights (struct gfs_rgrpd * tree)
{
	int h, hl, hr;

	if (tree == rd_avl_empty)
		return;
	avl_checkheights(tree->rd_avl_left);
	avl_checkheights(tree->rd_avl_right);
	h = tree->rd_avl_height;
	hl = heightof(tree->rd_avl_left);
	hr = heightof(tree->rd_avl_right);
	if ((h == hl+1) && (hr <= hl) && (hl <= hr+1))
		return;
	if ((h == hr+1) && (hl <= hr) && (hr <= hl+1))
		return;
	printk("%s: avl_checkheights: heights inconsistent\n",avl_check_point);
}

/* check that all values stored in a tree are < key */
static void avl_checkleft (struct gfs_rgrpd * tree, rd_avl_key_t key)
{
	if (tree == rd_avl_empty)
		return;
	avl_checkleft(tree->rd_avl_left,key);
	avl_checkleft(tree->rd_avl_right,key);
	if (tree->rd_avl_key < key)
		return;
	printk("%s: avl_checkleft: left key %lu >= top key %lu\n",avl_check_point,tree->rd_avl_key,key);
}

/* check that all values stored in a tree are > key */
static void avl_checkright (struct gfs_rgrpd * tree, rd_avl_key_t key)
{
	if (tree == rd_avl_empty)
		return;
	avl_checkright(tree->rd_avl_left,key);
	avl_checkright(tree->rd_avl_right,key);
	if (tree->rd_avl_key > key)
		return;
	printk("%s: avl_checkright: right key %lu <= top key %lu\n",avl_check_point,tree->rd_avl_key,key);
}

/* check that all values are properly increasing */
static void avl_checkorder (struct gfs_rgrpd * tree)
{
	if (tree == rd_avl_empty)
		return;
	avl_checkorder(tree->rd_avl_left);
	avl_checkorder(tree->rd_avl_right);
	avl_checkleft(tree->rd_avl_left,tree->rd_avl_key);
	avl_checkright(tree->rd_avl_right,tree->rd_avl_key);
}

/* all checks */
static void avl_check (struct gfs_sbd_t *sdp, char *caller)
{
	avl_check_point = caller;
	avl_checkheights(sdp->sd_rgavl);
	avl_checkorder(sdp->sd_rgavl);
}

#endif
