/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portions Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: super.c
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "blklist.h"
#include "bmap.h"
#include "dio.h"
#include "file.h"
#include "format.h"
#include "fs.h"
#include "glock.h"
#include "glops.h"
#include "inode.h"
#include "recovery.h"
#include "super.h"





/**
 * gfs_check_sb - Check superblock
 * @sdp: the filesystem
 * @sb: The superblock
 * @silent: Don't print a message if the check fails
 *
 * Checks the version code of the FS is one that we understand how to
 * read and that the sizes of the various on-disk structures have not
 * changed.
 */

int gfs_check_sb(gfs_sbd_t *sdp, gfs_sb_t *sb, int silent)
{
  unsigned int x;
  int error = 0;

  ENTER(GFN_CHECK_SB);


  if (sb->sb_header.mh_magic != GFS_MAGIC ||
      sb->sb_header.mh_type != GFS_METATYPE_SB)
  {
    if (!silent)
      con_printf("GFS:  not a GFS filesystem\n");
    error = -EINVAL;
    goto out;
  }


  /*  If format numbers match exactly, we're done.  */ 

  if (sb->sb_fs_format == GFS_FORMAT_FS &&
      sb->sb_multihost_format == GFS_FORMAT_MULTI)
    goto out;


  if (sb->sb_fs_format != GFS_FORMAT_FS)
  {
    for (x = 0; gfs_old_fs_formats[x]; x++)
      if (gfs_old_fs_formats[x] == sb->sb_fs_format)
	break;

    if (!gfs_old_fs_formats[x])
    {
      con_printf("GFS:  code version (%u, %u) is incompatible with ondisk format (%u, %u)\n",
		 GFS_FORMAT_FS, GFS_FORMAT_MULTI,
		 sb->sb_fs_format, sb->sb_multihost_format);
      con_printf("GFS:  I don't know how to upgrade this FS\n");
      error = -EINVAL;
      goto out;
    }
  }


  if (sb->sb_multihost_format != GFS_FORMAT_MULTI)
  {
    for (x = 0; gfs_old_multihost_formats[x]; x++)
      if (gfs_old_multihost_formats[x] == sb->sb_multihost_format)
	break;

    if (!gfs_old_multihost_formats[x])
    {
      con_printf("GFS:  code version (%u, %u) is incompatible with ondisk format (%u, %u)\n",
		 GFS_FORMAT_FS, GFS_FORMAT_MULTI,
		 sb->sb_fs_format, sb->sb_multihost_format);
      con_printf("GFS:  I don't know how to upgrade this FS\n");
      error = -EINVAL;
      goto out;
    }
  }


  if (!sdp->sd_args.ar_upgrade)
  {
    con_printf("GFS:  code version (%u, %u) is incompatible with ondisk format (%u, %u)\n",
	       GFS_FORMAT_FS, GFS_FORMAT_MULTI,
	       sb->sb_fs_format, sb->sb_multihost_format);
    con_printf("GFS:  Use the \"upgrade\" mount option to upgrade the FS\n");
    con_printf("GFS:  See the HOWTO for more details\n");
    error = -EINVAL;
    goto out;
  }


 out:
  EXIT(GFN_CHECK_SB);

  return error;
}


/**
 * gfs_read_sb - Read super block
 * @sdp: The GFS superblock
 * @gl: the glock for the superblock (assumed to be held)
 * @silent: Don't print message if mount fails
 *
 */

int gfs_read_sb(gfs_sbd_t *sdp, gfs_glock_t *gl, int silent)
{
  osi_buf_t *bh;
  uint64 block;

  ENTER(GFN_READ_SB);


  block = GFS_SB_ADDR * GFS_BASIC_BLOCK / sdp->sd_sb.sb_bsize;

  bh = gfs_dread(sdp, block, gl, DR_START | DR_WAIT);
  if (!bh)
  {
    if (!silent)
      con_printf("GFS:  can't read superblock\n");
    EXIT(GFN_READ_SB);
    return -EIO;
  }

  GFS_ASSERT(sizeof(gfs_sb_t) <= BH_SIZE(bh), );

  gfs_sb_in(&sdp->sd_sb, BH_DATA(bh));

  gfs_drelse(bh);

  if (gfs_check_sb(sdp, &sdp->sd_sb, silent))
  {
    EXIT(GFN_READ_SB);
    return -EINVAL;
  }


  sdp->sd_fsb2bb_shift = ffs(sdp->sd_sb.sb_bsize / GFS_BASIC_BLOCK) - 1;
  sdp->sd_diptrs = (sdp->sd_sb.sb_bsize - sizeof(gfs_dinode_t)) / sizeof(uint64);
  sdp->sd_inptrs = (sdp->sd_sb.sb_bsize - sizeof(gfs_indirect_t)) / sizeof(uint64);
  sdp->sd_jbsize = sdp->sd_sb.sb_bsize - sizeof(gfs_jdata_t);
  sdp->sd_hash_bsize = sdp->sd_sb.sb_bsize / 2;
  sdp->sd_hash_ptrs = sdp->sd_hash_bsize / sizeof(uint64);


  EXIT(GFN_READ_SB);

  return 0;
}


/**
 * gfs_do_upgrade - upgrade a filesystem
 * @sdp: The GFS superblock
 *
 */

int gfs_do_upgrade(gfs_sbd_t *sdp, gfs_glock_t *gl_sb)
{
  gfs_glock_t *gl_j;
  gfs_log_header_t lh;
  osi_buf_t *bh;
  unsigned int x;
  uint64 block;
  int error = 0;

  ENTER(GFN_DO_UPGRADE);


  /*  If format numbers match exactly, we're done.  */ 

  if (sdp->sd_sb.sb_fs_format == GFS_FORMAT_FS &&
      sdp->sd_sb.sb_multihost_format == GFS_FORMAT_MULTI)
  {
    con_printf("GFS:  no upgrade necessary\n");
    sdp->sd_args.ar_upgrade = FALSE;
    goto out;
  }


  error = gfs_jindex_hold(sdp, FALSE);
  switch (error)
  {
  case 0:
    break;

  case GLR_TRYFAILED:
    error = -EBUSY;

  default:
    goto fail;
  }


  for (x = 0; x < sdp->sd_journals; x++)
  {
    error = gfs_glock_num(sdp, sdp->sd_jindex[x].ji_addr, &gfs_meta_glops, NULL,
			  LM_FLAG_TRY, &gl_j);
    switch (error)
    {
    case 0:
      break;

    case GLR_TRYFAILED:
      error = -EBUSY;
      
    default:
      goto fail_jindex_relse;
    }

    error = gfs_find_jhead(sdp, &sdp->sd_jindex[x], &lh);

    gfs_gunlock(sdp, gl_j, 0);
    gfs_put_glstruct(gl_j);

    if (error)
      goto fail_jindex_relse;

    if (!(lh.lh_flags & GFS_LOG_HEAD_UNMOUNT))
    {
      error = -EBUSY;
      goto fail_jindex_relse;
    }
  }


  /*  We don't need to journal this change because we're changing
      only one sector of one block.  We definitely don't want to have
      the journaling code running at this point.  */

  block = GFS_SB_ADDR * GFS_BASIC_BLOCK / sdp->sd_sb.sb_bsize;
  bh = gfs_dread(sdp, block, gl_sb, DR_START | DR_WAIT);
  if (!bh)
  {
    error = -EIO;
    goto fail_jindex_relse;
  }

  gfs_sb_in(&sdp->sd_sb, BH_DATA(bh));

  error = gfs_check_sb(sdp, &sdp->sd_sb, FALSE);
  GFS_ASSERT(!error, );

  sdp->sd_sb.sb_fs_format = GFS_FORMAT_FS;
  sdp->sd_sb.sb_multihost_format = GFS_FORMAT_MULTI;

  gfs_sb_out(&sdp->sd_sb, BH_DATA(bh));

  error = gfs_dwrite(sdp, bh, DW_DIRTY | DW_START | DW_WAIT);

  gfs_drelse(bh);


  gfs_jindex_release(sdp);


  if (!error)
  {
    con_printf("GFS:  upgrade successful\n");
    sdp->sd_args.ar_upgrade = FALSE;
  }


 out:
  EXIT(GFN_DO_UPGRADE);

  return error;



 fail_jindex_relse:
  gfs_jindex_release(sdp);

 fail:
  if (error == -EBUSY)
    con_printf("GFS:  can't upgrade:  the FS is still busy or contains dirty journals\n");
  else
    con_printf("GFS:  can't upgrade:  error = %d\n", error);
  EXIT(GFN_DO_UPGRADE);
  return error; 
}


/**
 * gfs_clear_journals - Clear all the journal index information
 * @sdp: The GFS superblock
 *
 */

void gfs_clear_journals(gfs_sbd_t *sdp)
{
  ENTER(GFN_CLEAR_JOURNALS);

  if (sdp->sd_jindex)
  {
    gfree(sdp->sd_jindex, sdp->sd_journals * sizeof(gfs_jindex_t));
    sdp->sd_jindex = NULL;
  }
  sdp->sd_journals = 0;

  EXIT(GFN_CLEAR_JOURNALS);
}


/**
 * gfs_journal_update - Update the journal index information
 * @ip: The journal index inode
 *
 * Returns: 0 on success, error code otherwise
 */

STATIC int gfs_ji_update(gfs_inode_t *ip)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  gfs_jindex_t buf;
  int rv;
  unsigned int j;

  ENTER(GFN_JI_UPDATE);

  GFS_ASSERT(ip->i_di.di_size % sizeof(gfs_jindex_t) == 0, );


  gfs_clear_journals(sdp);


  sdp->sd_jindex = (gfs_jindex_t *)gmalloc(ip->i_di.di_size);
  rv = -ENOMEM;
  if (!sdp->sd_jindex)
    goto out;
  gfs_memset(sdp->sd_jindex, 0, ip->i_di.di_size);


  for (j = 0; ; j++)
  {
    rv = gfs_internal_read(ip, (char *)&buf, j * sizeof(gfs_jindex_t), sizeof(gfs_jindex_t));
    if (!rv)
      break;
    if (rv != sizeof(gfs_jindex_t))
    {
      if (rv > 0)
	rv = -EIO;
      goto out;
    }

    gfs_jindex_in(sdp->sd_jindex + j, (char *)&buf);
  }


  GFS_ASSERT(j * sizeof(gfs_jindex_t) == ip->i_di.di_size, );
  GFS_ASSERT(j <= GFS_CLIENTS, );

  sdp->sd_journals = j;
  sdp->sd_jiinode_vn = ip->i_gl->gl_vn;


 out:
  EXIT(GFN_JI_UPDATE);
  return rv;
}


/**
 * gfs_jindex_hold - Grab a lock on the jindex
 * @sdp: The GFS superblock
 * @wait: block
 *
 * This is very similar to the gfs_rindex_hold() function, except that
 * in general we hold the jindex lock for longer periods of time and
 * we grab it far less frequently (in general) then the rgrp lock.
 *
 * Returns: 0 on success, error code otherwise
 */

int gfs_jindex_hold(gfs_sbd_t *sdp, int wait)
{
  gfs_inode_t *ip = sdp->sd_jiinode;
  gfs_glock_t *gl = ip->i_gl;
  int error;

  ENTER(GFN_JINDEX_HOLD);

  error = gfs_glock_i(ip, GL_SHARED | ((wait) ? 0 : LM_FLAG_TRY));
  if (error)
    goto out;

  if (sdp->sd_jiinode_vn != gl->gl_vn)
  {
    error = gfs_ji_update(ip);
    if (error)
      gfs_gunlock_i(ip, 0);
  }

 out:
  EXIT(GFN_JINDEX_HOLD);
  return error;
}


/**
 * gfs_ji_release - Release hold over journal index
 * @sdp: The GFs superblock
 *
 * Drop the lock on the journal index, allows the number and size of
 * journals to vary once more
 *
 */

void gfs_jindex_release(gfs_sbd_t *sdp)
{
  ENTER(GFN_JINDEX_RELEASE);
  gfs_gunlock_i(sdp->sd_jiinode, 0);
  EXIT(GFN_JINDEX_RELEASE);
}


/**
 * gfs_compute_reserv - Compute various maximum sizes
 * @sdp: The GFS superblock
 *
 */

void gfs_compute_reserv(gfs_sbd_t *sdp)
{
  gfs_inode_t ip;
  uint32 hash_blocks, ind_blocks, leaf_blocks;
  uint32 tmp_blocks;

  ENTER(GFN_COMPUTE_RESERV);


  /*  Compute maximum reservation required to add a entry to a directory  */

  hash_blocks = DIV_RU(sizeof(uint64) * (1 << GFS_DIR_MAX_DEPTH), sdp->sd_jbsize);

  ind_blocks = 0;
  for (tmp_blocks = hash_blocks; tmp_blocks > sdp->sd_diptrs; )
  {
    tmp_blocks = DIV_RU(tmp_blocks, sdp->sd_inptrs);
    ind_blocks += tmp_blocks;
  }
  
  leaf_blocks = (2 + GFS_DIR_MAX_DEPTH);
 
  sdp->sd_max_dirres =  hash_blocks + ind_blocks + leaf_blocks;


  /*  Compute the maximum height of a file's metadata tree  */

  ip.i_sbd = sdp;  /*  Damn kludge!!!  */
  ip.i_di.di_flags = GFS_DIF_JDATA;
  ip.i_di.di_size = (uint64)-1;

  sdp->sd_max_height = gfs_calc_tree_height(&ip, 0);

  GFS_ASSERT(sdp->sd_max_height <= GFS_MAX_META_HEIGHT, );


  EXIT(GFN_COMPUTE_RESERV);
}


/**
 * gfs_get_riinode - Initialise the rindex inode for the superblock
 * @sdp: The GFS superblock
 *
 * Returns: 0 on success, error code otherwise
 */

int gfs_get_riinode(gfs_sbd_t *sdp)
{
  gfs_glock_t *gl;
  int error;

  ENTER(GFN_GET_RIINODE);


  error = gfs_glock_num(sdp, sdp->sd_sb.sb_rindex_di.no_formal_ino, &gfs_inode_glops, NULL, GL_SHARED, &gl);
  if (error)
    goto out;

  sdp->sd_riinode = gfs_get_istruct(sdp, gl, &sdp->sd_sb.sb_rindex_di, CREATE);
  if (!sdp->sd_riinode)
  {
    gfs_gunlock(sdp, gl, 0);
    gfs_put_glstruct(gl);
    error = -ENOMEM;
    goto out;
  }

  sdp->sd_riinode_vn = gl->gl_vn - 1;
  osi_set_bit(GLF_STICKY, &gl->gl_flags);

  gfs_gunlock(sdp, gl, 0);
  gfs_put_glstruct(gl);


 out:
  EXIT(GFN_GET_RIINODE);

  return error;
}


/**
 * gfs_put_riinode - Put the rindex dinode
 * @sdp: The GFS superblock
 *
 */

void gfs_put_riinode(gfs_sbd_t *sdp)
{
  ENTER(GFN_PUT_RIINODE);
  gfs_put_istruct(sdp->sd_riinode);
  sdp->sd_riinode = NULL;
  EXIT(GFN_PUT_RIINODE);
}


/**
 * gfs_get_jiinode - Initialise the jindex inode for the superblock
 * @sdp: The GFS superblock
 *
 * Returns: 0 on success, error code otherwise
 */

int gfs_get_jiinode(gfs_sbd_t *sdp)
{
  gfs_glock_t *gl;
  int error;

  ENTER(GFN_GET_JIINODE);


  error = gfs_glock_num(sdp, sdp->sd_sb.sb_jindex_di.no_formal_ino, &gfs_inode_glops, NULL, GL_SHARED, &gl);
  if (error)
    goto out;

  sdp->sd_jiinode = gfs_get_istruct(sdp, gl, &sdp->sd_sb.sb_jindex_di, CREATE);
  if (!sdp->sd_jiinode)
  {
    gfs_gunlock(sdp, gl, 0);
    gfs_put_glstruct(gl);
    error = -ENOMEM;
    goto out;
  }

  sdp->sd_jiinode_vn = gl->gl_vn - 1;
  osi_set_bit(GLF_STICKY, &gl->gl_flags);

  gfs_gunlock(sdp, gl, 0);
  gfs_put_glstruct(gl);


 out:
  EXIT(GFN_GET_JIINODE);

  return error;
}


/**
 * gfs_put_jiinode - Put the jindex dinode
 * @sdp: The GFS superblock
 *
 */

void gfs_put_jiinode(gfs_sbd_t *sdp)
{
  ENTER(GFN_PUT_JIINODE);
  gfs_put_istruct(sdp->sd_jiinode);
  sdp->sd_jiinode = NULL;
  EXIT(GFN_PUT_JIINODE);
}


/**
 * gfs_setup_dameth - Setup the directory allocation method
 * @sdp: the filesystem
 *
 * Returns:  0 on success, -EXXX on failure
 */

int gfs_setup_dameth(gfs_sbd_t *sdp)
{
  gfs_rgrpd_t *rgd;
  unsigned int rg;
  int error = 0;

  ENTER(GFN_SETUP_DAMETH);


  error = gfs_rindex_hold(sdp);
  if (error)
    goto fail;


  switch (sdp->sd_args.ar_dir_alloc_method)
  {
  case GFS_DAMETH_SINGLE:
    /*  This code could be improved when we get LVBs.
	We could verify that two machines aren't
	picking the same RG.  */

    if (sdp->sd_rgcount >= sdp->sd_journals)
      rg = sdp->sd_rgcount * sdp->sd_lockstruct.ls_jid / sdp->sd_journals;
    else
      rg = sdp->sd_lockstruct.ls_jid % sdp->sd_rgcount;

    break;


  case GFS_DAMETH_RROBIN:
  case GFS_DAMETH_RANDOM:
    /*  Do the same thing for both of these:  just initialize
	sd_dir_alloc_target to a random value.  */

    rg = gfs_randrange(0, sdp->sd_rgcount);

    break;


  default:
    con_printf("GFS:  weird directory allocation method (%u)\n",
	       sdp->sd_args.ar_dir_alloc_method);
    error = -EINVAL;
    goto fail_ri_relse;
  }


  rgd = gfs_rgrpd_get_by_num(sdp, rg);
  if (!rgd)
  {
    con_printf("GFS:  can't determine initial resource group\n");
    error = -ENOENT;
    goto fail_ri_relse;
  }
  sdp->sd_dir_alloc_target = rgd->rd_ri.ri_addr;
  gfs_rgrpd_put(rgd);

  gfs_rindex_release(sdp);


  EXIT(GFN_SETUP_DAMETH);

  return 0;



 fail_ri_relse:
  gfs_rindex_release(sdp);

 fail:
  EXIT(GFN_SETUP_DAMETH);
  return error;
}


/**
 * gfs_update_datarget - Update the directory allocation target
 * @sdp: the filesystem
 * @res: the last inplace reservation
 *
 * Returns:  0 on success, -EXXX on failure
 */

int gfs_update_datarget(gfs_sbd_t *sdp, gfs_ipreserv_t *res)
{
  int error = 0;

  ENTER(GFN_UPDATE_DATARGET);

  
  switch (sdp->sd_args.ar_dir_alloc_method)
  {
  case GFS_DAMETH_SINGLE:
    sdp->sd_dir_alloc_target = res->ir_rgd->rd_ri.ri_addr;

    break;


  case GFS_DAMETH_RROBIN:
    sdp->sd_dir_alloc_target = res->ir_rgd->rd_ri.ri_addr + 1;

    break;


  case GFS_DAMETH_RANDOM:
    error = gfs_setup_dameth(sdp);

    break;


  default:
    con_printf("GFS:  weird directory allocation method (%u)\n",
	       sdp->sd_args.ar_dir_alloc_method);
    error = -EINVAL;
  }


  EXIT(GFN_UPDATE_DATARGET);

  return error;
}


/**
 * gfs_stat_gfs - Do a statfs
 * @sdp: the filesystem
 * @usage: the usage structure
 * @interruptible:  Stop if there is a signal pending
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_stat_gfs(gfs_sbd_t *sdp, gfs_usage_t *usage, int interruptible)
{
  gfs_rgrpd_t *rgd = NULL;
  uint64 hint = 0;
  int error;

  ENTER(GFN_STAT_GFS);


  gfs_memset(usage, 0, sizeof(gfs_usage_t));
  usage->block_size = sdp->sd_sb.sb_bsize;


  error = gfs_rindex_hold(sdp);
  if (error)
    goto fail;

  for (;;)
  {
    rgd = gfs_rgrpd_get_with_hint(sdp, hint);
    if (!rgd)
      break;


    for (;;)
    {
      error = gfs_glock_rg(rgd, GL_SHARED | GL_SKIP);
      if (error)
	goto fail_rgrpd_put;

      GFS_ASSERT(rgd->rd_rs.rs_magic == GFS_MAGIC, );


      if (!rgd->rd_rs.rs_owner || rgd->rd_rs.rs_owner == sdp->sd_lockstruct.ls_jid + 1)
      {
	IFDB(GFN_STAT_GFS,
	     gfs_rindex_print(&rgd->rd_ri, FALSE);
	     gfs_rgrp_save_print(&rgd->rd_rs, FALSE););

	usage->total_blocks += rgd->rd_ri.ri_data;
	usage->free += rgd->rd_rs.rs_free;
	usage->used_dinode += rgd->rd_rs.rs_useddi;
	usage->free_dinode += rgd->rd_rs.rs_freedi;
	usage->used_meta += rgd->rd_rs.rs_usedmeta;
	usage->free_meta += rgd->rd_rs.rs_freemeta;

	gfs_gunlock_rg(rgd, GL_SKIP);

	break;
      }
      else
      {
	gfs_gunlock_rg(rgd, GL_SKIP);

	error = gfs_rgrp_lvb_init(rgd, TRUE);
	if (error)
	  goto fail_rgrpd_put;

	continue;
      }
    }


    hint = rgd->rd_ri.ri_addr + 1;
    gfs_rgrpd_put(rgd);

    if (interruptible && osi_pending_signals())
    {
      error = -ERESTARTSYS;
      goto fail_rindex_relse;
    }
  }

  gfs_rindex_release(sdp);


  EXIT(GFN_STAT_GFS);

  return 0;



 fail_rgrpd_put:
  gfs_rgrpd_put(rgd);
  
 fail_rindex_relse:
  gfs_rindex_release(sdp);

 fail:
  EXIT(GFN_STAT_GFS);
  return error;
}




