/******************************************************************************
*******************************************************************************
**
**    Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: 
**      trans.c
**
**    Description:
**      Transaction functions
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"

#include "blklist.h"
#include "dio.h"
#include "glock.h"
#include "inode.h"
#include "log.h"
#include "trans.h"





/**
 * gfs_trans_print - Print a transaction to the console
 * @sdp: the filesystem
 * @tr: The GFS transaction
 * @where: Situation of transaction
 *
 */

void gfs_trans_print(gfs_sbd_t *sdp, gfs_trans_t *tr, int where)
{
  gfs_glock_t *gl;
  gfs_bufdata_t *bd;
  gfs_tr_nop_t *nop;
  osi_list_head_t *ntmp, *gtmp, *btmp, *head;
  char *type;
  uint32 meta_type;
  uint64 blk;

  ENTER(GFN_TRANS_PRINT);


  switch(tr->tr_type)
  {
  case TRANS_COMPOUND:
    type = "compound";
    break;
  case TRANS_CREATE:
    type = "create";
    break;
  case TRANS_LINK:
    type = "link";
    break;
  case TRANS_RENAME:
    type = "rename";
    break;
  case TRANS_WRITE:
    type = "write";
    break;
  case TRANS_BMAP:
    type = "bmap";
    break;
  case TRANS_TRUNC_GROW:
    type = "truncate grow";
    break;
  case TRANS_PREPWRITE:
    type = "prepare write";
    break;
  case TRANS_NOTE_CH:
    type = "notify change";
    break;
  case TRANS_SETATTR:
    type = "set attributes";
    break;
  case TRANS_DIR2FILE:
    type = "dir to file";
    break;
  case TRANS_UNLINK:
    type = "unlink";
    break;
  case TRANS_RMDIR:
    type = "rmdir";
    break;
  case TRANS_LEAF_DEALLOC:
    type = "leaf deallocate";
    break;
  case TRANS_FILE_DEALLOC:
    type = "file deallocate";
    break;
  case TRANS_I_DEALLOC:
    type = "inode deallocate";
    break;
  default:
    type = "unknown";
    break;
  }

  db_printf("Transaction:  %s  flags = 0x%.8X, res=%u segs (%u blocks orig)\n",
	    type, tr->tr_flags, tr->tr_seg_reserved, tr->tr_blks_asked);
  db_printf("  tr_num_buf = %u, tr_num_gl = %u, tr_num_nop_o = %u, tr_num_nop_c = %u\n",
	    tr->tr_num_buf, tr->tr_num_gl, tr->tr_num_nop_o, tr->tr_num_nop_c);

  for (ntmp = tr->tr_nop_change.next; ntmp != &tr->tr_nop_change; ntmp = ntmp->next)
  {
    nop = osi_list_entry(ntmp, gfs_tr_nop_t, nop_list);

    switch (nop->nop_type)
    {
    case GFS_LOG_DESC_NOPOPEN:
      db_printf("  inode %"FMT_64"u/%"FMT_64"u is now open\n",
		nop->nop_inum.no_formal_ino, nop->nop_inum.no_addr);
      break;
    case GFS_LOG_DESC_NOPCLOSE:
      db_printf("  inode %"FMT_64"u/%"FMT_64"u is now closed\n",
		nop->nop_inum.no_formal_ino, nop->nop_inum.no_addr);
      break;
    default:
      db_printf("  weird nopen value (%d)\n", nop->nop_type);
      break;
    }
  }


  for (gtmp = tr->tr_gl.next; gtmp != &tr->tr_gl; gtmp = gtmp->next)
  {
    if (where == TRANS_IS_NEW)
    {
      gl = osi_list_entry(gtmp, gfs_glock_t, gl_new_list);
      head = &gl->gl_new_bufs;
    }
    else if (where == TRANS_IS_INCORE)
    {
      gl = osi_list_entry(gtmp, gfs_glock_t, gl_incore_list);
      head = &gl->gl_incore_bufs;
    }
    else
      GFS_ASSERT(0, con_printf("where = %d\n", where););

    blk = gfs_lockname2blk(sdp, &gl->gl_name);
    db_printf("  Glock (%"FMT_64"u, %u)\n", blk, gl->gl_name.ln_type);

    for (btmp = head->next; btmp != head; btmp = btmp->next)
    {
      if (where == TRANS_IS_NEW)
	bd = osi_list_entry(btmp, gfs_bufdata_t, bd_new_list);
      else
	bd = osi_list_entry(btmp, gfs_bufdata_t, bd_incore_list);

      meta_type = ((gfs_meta_header_t *)BH_DATA(bd->bd_bh))->mh_type;
      meta_type = gfs32_to_cpu(meta_type);

      db_printf("    buffer %"FMT_64"u %u\n", BH_BLKNO(bd->bd_bh), meta_type);
    }
  }


  EXIT(GFN_TRANS_PRINT);
}


/**
 * new_trans - Make a new transaction structure
 * @sdp: The GFS superblock
 * 
 * Returns: the new transaction
 */

STATIC gfs_trans_t *new_trans(gfs_sbd_t *sdp)
{
  gfs_trans_t *tr = NULL;

  ENTER(GFN_GET_TRANS);


  tr = (gfs_trans_t *)gmalloc(sizeof(gfs_trans_t));
  if (tr)
  {
    gfs_memset(tr, 0, sizeof(gfs_trans_t));

    osi_list_head_init(&tr->tr_gl);
    osi_list_head_init(&tr->tr_nop_change);
    osi_list_head_init(&tr->tr_nop_change_ic);
    osi_list_head_init(&tr->tr_bhlist);

    GFS_ASSERT(!osi_get_transaction(), );
    osi_set_transaction(tr);
  }


  EXIT(GFN_GET_TRANS);

  return tr;
}


/**
 * gfs_trans_begin - Perpare to start a transaction
 * @sdp: The GFS superblock
 * @type: The type of the transaction
 * @meta_blocks: Reserve this many metadata blocks in the log
 * @extra_blocks: Number of non-metadata blocks to reserve
 *
 * Allocate the gfs_trans_t struct.  Do in-place and
 * log reservations.
 *
 * Returns: 0 on success, -EXXX on failure
 */

int gfs_trans_begin(gfs_sbd_t *sdp, int type, unsigned int meta_blocks, unsigned int extra_blocks)
{
  gfs_trans_t *tr;
  unsigned int blocks;
  int error;

  ENTER(GFN_TRANS_BEGIN);


  error = gfs_glock(sdp, sdp->sd_trans_gl, GL_SHARED);
  if (error)
    goto fail;


  /*  Actually get the transaction  */
  tr = new_trans(sdp);
  if (!tr)
  {
    error = -ENOMEM;
    goto fail_gunlock;
  }

  tr->tr_type = type;


  /*  Do log reservation  */

  tr->tr_blks_asked = meta_blocks;
  blocks = gfs_struct2blk(sdp, meta_blocks, sizeof(gfs_block_tag_t)) + meta_blocks + extra_blocks + 1;
  tr->tr_seg_reserved = gfs_blk2seg(sdp, blocks);


  error = gfs_log_reserv(sdp, tr->tr_seg_reserved);
  if (error)
    goto fail_free_trans;


  IFDB(GFN_TRANS_PRINT,
       gfs_trans_print(sdp, tr, TRANS_IS_NEW););


  EXIT(GFN_TRANS_BEGIN);

  return 0;



 fail_free_trans:
  osi_set_transaction(NULL);
  gfree(tr, sizeof(gfs_trans_t));

 fail_gunlock:
  gfs_gunlock(sdp, sdp->sd_trans_gl, 0);

 fail:
  EXIT(GFN_TRANS_BEGIN);
  return error;
}


/**
 * do_poison - poison a transaction
 * @sdp: the filesystem
 * @tr: the transaction
 *
 */

STATIC void do_poison(gfs_sbd_t *sdp, gfs_trans_t *tr)
{
  osi_list_head_t *gtmp, *btmp;
  gfs_glock_t *gl;
  gfs_bufdata_t *bd;
  gfs_tr_nop_t *nop;

  ENTER(GFN_DO_POISON);


  for (gtmp = tr->tr_gl.next; gtmp != &tr->tr_gl; gtmp = gtmp->next)
  {
    gl = osi_list_entry(gtmp, gfs_glock_t, gl_new_list);

    osi_set_bit(GLF_POISONED, &gl->gl_flags);

    for (btmp = gl->gl_new_bufs.next; btmp != &gl->gl_new_bufs; btmp = btmp->next)
    {
      bd = osi_list_entry(btmp, gfs_bufdata_t, bd_new_list);
      gfs_memcpy(BH_DATA(bd->bd_bh), bd->bd_frozen, sdp->sd_sb.sb_bsize);
    }
  }


  while (!osi_list_empty(&tr->tr_nop_change))
  {
    nop = osi_list_entry(tr->tr_nop_change.next, gfs_tr_nop_t, nop_list);
    osi_list_del(&nop->nop_list);
    gfree(nop, sizeof(gfs_tr_nop_t));
  }
  while (!osi_list_empty(&tr->tr_nop_change_ic))
  {
    nop = osi_list_entry(tr->tr_nop_change_ic.next, gfs_tr_nop_t, nop_list);
    osi_list_del(&nop->nop_list);
    gfree(nop, sizeof(gfs_tr_nop_t));
  }


  EXIT(GFN_DO_POISON);
}


/**
 * gfs_trans_end - End a transaction
 * @sdp: The GFS superblock
 *
 * If buffers were actually added to the transaction,
 * commit it.
 */

void gfs_trans_end(gfs_sbd_t *sdp)
{
  gfs_trans_t *tr;
  gfs_glock_t *gl;
  gfs_tr_nop_t *nop;
  osi_list_head_t *gtmp;
  int null_trans = TRUE;

  ENTER(GFN_TRANS_END);


  tr = osi_get_transaction();
  GFS_ASSERT(tr, );
  osi_set_transaction(NULL); 


  /*  Verify that all the glocks were passed to this transaction and look for null transactions  */

  for (gtmp = tr->tr_gl.next; gtmp != &tr->tr_gl; gtmp = gtmp->next)
  {
    gl = osi_list_entry(gtmp, gfs_glock_t, gl_new_list);
    GFS_ASSERT(osi_atomic_read(&gl->gl_locked) &&
	       gl->gl_pid == osi_pid() &&
	       gl->gl_state == LM_ST_EXCLUSIVE, );

    gfs_put_glstruct(gl);

    if (!osi_list_empty(&gl->gl_new_bufs))
      null_trans = FALSE;
  }


  if (null_trans)
  {
    while (!osi_list_empty(&tr->tr_gl))
    {
      gl = osi_list_entry(tr->tr_gl.next, gfs_glock_t, gl_new_list);

      osi_list_del(&gl->gl_new_list);
      osi_list_head_init(&gl->gl_new_list);
      gl->gl_new_tr = NULL;
    }

    while (!osi_list_empty(&tr->tr_nop_change))
    {
      nop = osi_list_entry(tr->tr_nop_change.next, gfs_tr_nop_t, nop_list);
      osi_list_del(&nop->nop_list);
      gfree(nop, sizeof(gfs_tr_nop_t));
    }
    while (!osi_list_empty(&tr->tr_nop_change_ic))
    {
      nop = osi_list_entry(tr->tr_nop_change_ic.next, gfs_tr_nop_t, nop_list);
      osi_list_del(&nop->nop_list);
      gfree(nop, sizeof(gfs_tr_nop_t));
    }

    gfs_log_relse(sdp, tr->tr_seg_reserved);
    gfree(tr, sizeof(gfs_trans_t));

    gfs_gunlock(sdp, sdp->sd_trans_gl, 0);

    EXIT(GFN_TRANS_END);
    return;
  }


  GFS_ASSERT(tr->tr_num_buf <= tr->tr_blks_asked,
	     gfs_trans_print(sdp, tr, TRANS_IS_NEW););


  /*  Handle poisoning  */

  if (tr->tr_flags & TRF_POISONED)
    do_poison(sdp, tr);


  IFDB(GFN_TRANS_PRINT,
       gfs_trans_print(sdp, tr, TRANS_IS_NEW););


  gfs_log_commit(sdp, tr);


  gfs_gunlock(sdp, sdp->sd_trans_gl, 0);


  EXIT(GFN_TRANS_END);
}


/**
 * gfs_trans_add_gl - Add a glock to a transaction
 * @sdp: the filesystem
 * @gl: the glock
 *
 * Add the given glock to this process's transaction
 */

void gfs_trans_add_gl(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  gfs_trans_t *tr;

  ENTER(GFN_TRANS_ADD_GL);


  if (!gl->gl_new_tr)
  {
    GFS_ASSERT(osi_atomic_read(&gl->gl_locked) &&
	       gl->gl_pid == osi_pid() &&
	       gl->gl_state == LM_ST_EXCLUSIVE, );
    GFS_ASSERT(osi_list_empty(&gl->gl_new_list), );
    GFS_ASSERT(osi_list_empty(&gl->gl_new_bufs), );

    tr = osi_get_transaction();
    GFS_ASSERT(tr, );

    osi_list_add(&gl->gl_new_list, &tr->tr_gl);
    gl->gl_new_tr = tr;

    gfs_hold_glstruct(gl);  /*  Released in gfs_trans_end()  */
    tr->tr_num_gl++;

    osi_set_bit(GLF_DIRTY, &gl->gl_flags);
  }
 

  EXIT(GFN_TRANS_ADD_GL);
}


/**
 * gfs_trans_add_bh - Add a buffer to the current transaction
 * @sdp: The GFS superblock
 * @bh: The buffer to add
 *
 * Add a buffer to the current transaction.  The glock for the buffer
 * should be held.  This pins the buffer as well.
 *
 * Call this as many times as you want during transaction formation.
 * It only does its work once.
 */

void gfs_trans_add_bh(gfs_sbd_t *sdp, osi_buf_t *bh)
{
  gfs_bufdata_t *bd;
  gfs_glock_t *gl;
  char *data;

  ENTER(GFN_TRANS_ADD_BH);


  gfs_meta_check(bh);


  bd = BH_BUFDATA(bh);

  if (!osi_list_empty(&bd->bd_new_list))
    goto out;


  gl = bd->bd_gl;

  if (!gl->gl_new_tr)
    gfs_trans_add_gl(sdp, gl);


  osi_list_add(&bd->bd_new_list, &gl->gl_new_bufs);
  gl->gl_new_tr->tr_num_buf++;


  gfs_dpin(bh);


  /*  FixMe!!!  */

  for (;;)
  {
    data = (char *)gmalloc(sdp->sd_sb.sb_bsize);
    if (data)
      break;

    con_printf("GFS:  Grrrr...\n");
    osi_sleep_unint(1);
  }


  gfs_lock_buffer(bh);

  GFS_ASSERT(!bd->bd_frozen, );
  bd->bd_frozen = data;

  gfs_memcpy(bd->bd_frozen, BH_DATA(bh), sdp->sd_sb.sb_bsize);

  gfs_unlock_buffer(bh);


 out:
  EXIT(GFN_TRANS_ADD_BH);
}


/**
 * gfs_trans_mark - Mark a flag in a transaction
 * @sdp: the filesystem
 * @flag: the flag to mark
 *
 */

void gfs_trans_mark(gfs_sbd_t *sdp, unsigned int flag)
{
  gfs_trans_t *tr;

  ENTER(GFN_TRANS_MARK);


  tr = osi_get_transaction();
  GFS_ASSERT(tr, );

  tr->tr_flags |= flag;


  EXIT(GFN_TRANS_MARK);
}


/**
 * gfs_trans_nopen_change - Add a nopen change to the current transaction
 * @sdp: the filesystem
 * @type: the type of nopen change (%GFS_LOG_DESC_NOP_OPEN or %GFS_LOG_DESC_NOP_CLOSE)
 * @inum: the inode number
 *
 * Returns:  0 on success, -EXXX on failure
 */

int gfs_trans_nopen_change(gfs_sbd_t *sdp, unsigned int type, gfs_inum_t *inum)
{
  gfs_trans_t *tr;
  gfs_tr_nop_t *nop;
  int x, error = 0;

  ENTER(GFN_TRANS_NOPEN_CHANGE);


  tr = osi_get_transaction();
  GFS_ASSERT(tr, );


  for (x = 0; x < 2; x++)
  {
    nop = gmalloc(sizeof(gfs_tr_nop_t));
    if (!nop)
    {
      error = -ENOMEM;
      goto out;
    }
    gfs_memset(nop, 0, sizeof(gfs_tr_nop_t));

    nop->nop_type = type;
    nop->nop_inum = *inum;

    if (x)
      osi_list_add(&nop->nop_list, &tr->tr_nop_change);
    else
      osi_list_add(&nop->nop_list, &tr->tr_nop_change_ic);
  }


  switch (type)
  {
  case GFS_LOG_DESC_NOPOPEN:
    tr->tr_num_nop_o++;
    break;

  case GFS_LOG_DESC_NOPCLOSE:
    tr->tr_num_nop_c++;
    break;

  default:
    GFS_ASSERT(0, con_printf("type = %u\n", type););
  }

  GFS_ASSERT(inum->no_formal_ino && inum->no_addr, );


 out:
  EXIT(GFN_TRANS_NOPEN_CHANGE);

  return error;
}


