/******************************************************************************
*
*    Copyright 1999 Regents of the University of Minnesota
*    Portions Copyright 1999-2001 Sistina Software, Inc.
*
*    This is free software released under the GNU General Public License.
*    There is no warranty for this software.  See the file COPYING for
*    details.
*
*    See the file CONTRIBUTORS for a list of contributors.
*
*    This file is maintained by:
*      Benjamin E Marzinski <marzinski@sistina.com>
*
* 
*/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <sys/mount.h>
#include <sys/mman.h>
#include <assert.h>

#include "header.h"

/* some private data structures. */
typedef struct {
   int fd;
   pthread_mutex_t recver;
   pthread_mutex_t sender;
   /* other info? */
} clients_t;

/* Table of clients that are logged in */
clients_t *Clients=NULL;
int client_cnt=0;
fd_set allset;

/**/
typedef struct Sema_s {
   int val;
   pthread_mutex_t mp;
   pthread_cond_t  cd;
} Sema;

#define CIRBUFSIZE 30
typedef struct cirbuf_s {
   clients_t *cl[CIRBUFSIZE];
   int       cnt;
   int       in;
   int       out;
   Sema      Space;
   Sema      Item;
} cirbuf_t;

cirbuf_t Buffy;

/* This global keeps the server running. */
static int Server_Runs = TRUE;

/* This file descriptor is the data we're serving. */
static int data_fd=-1;
static u64 datasize=0;
static char *datamap=NULL;
static int readonly = FALSE;

/******************************************************************************
 * dump_buffer()
 * prints out the contents of a buffer for inspection.  ignores formats.
 *
 * Need to get all zero lines absorbed.
 */
void dump_buffer(void *buf, int len)
{
	int i;
	u8 *c = buf;

	for(i=0; i<len; i++) {
		if( i % 16 == 0 )
			printf("\n%4u: ", i);
		printf("%2x ", c[i]);
	}
	printf("\n");
}
/******************************************************************************
 * heh, simple recv/send wrappers to get all of the beffer sent over the
 * net.  Had forgotten I needed this....
 */
int cpt_recv(int s, void *buf, size_t len, int flags)
{
	int res;
	do {
		res = recv(s, buf, len, flags);
		if(res<0) return res;
		len -= res;
		buf += res;
	}while(len>0);
	return res;
}

int cpt_send(int s, void *buf, size_t len, int flags)
{
	int res;
	do {
		res = send(s, buf, len, flags);
		if(res<0) return res;
		len -= res;
		buf += res;
	}while(len>0);
	return res;
}

/******************************************************************************
 * client_alloc()
 * expands the size of the clients table.
 */
#define NALLOC 10
void client_alloc(void)
{
   int i;
   if( Clients == NULL )
      Clients = malloc(NALLOC * sizeof(clients_t));
   else
      Clients = realloc(Clients, (client_cnt + NALLOC) * sizeof(clients_t));
   if(Clients == NULL) {
      perror("Can't alloc memory for client array");
      exit(1);
   }
   for(i=client_cnt; i < client_cnt+NALLOC; i++) {
      Clients[i].fd = -1;
#if 0 /* don't need to init these twice */
      pthread_mutex_init( &Clients[i].recver, NULL);
      pthread_mutex_init( &Clients[i].sender, NULL);
#endif
   }
   client_cnt += NALLOC;
   printf("realloced clients\n");
}
/******************************************************************************
 * client_add()
 */
int client_add(int fd)
{
   int i;
   
   if(Clients == NULL)
      client_alloc();
CAagain:
   for( i=0; i<client_cnt; i++) {
      if( Clients[i].fd == -1) {
         Clients[i].fd = fd;
         pthread_mutex_init( &Clients[i].recver, NULL);
         pthread_mutex_init( &Clients[i].sender, NULL);
         return i;
      }
   }

   /* client array must be full, expand it. */
   client_alloc();
   goto CAagain; /* now go add it. */
}
/******************************************************************************
 * client_del()
 * resetting the mutexes here may not be a good idea. do some checking.
 */
void client_del(int fd)
{
   int i;
   for(i=0; i<client_cnt; i++) {
      if( Clients[i].fd == fd) {
         Clients[i].fd = -1;
         pthread_mutex_destroy( &Clients[i].recver );
         pthread_mutex_destroy( &Clients[i].sender );
         return;
      }
   }
   /* pop an error? */
}

/******************************************************************************
 * home mafde semaphores....
 */
int Sema_init(Sema *S, int iv)
{
   int err;
   S->val = iv;
   err = pthread_mutex_init(&S->mp, NULL);
   if(err) return err;
   err = pthread_cond_init(&S->cd, NULL);
   if(err) return err;
   return 0;
}
int Sema_destroy(Sema *S)
{
   int err;
   err = pthread_mutex_destroy( &S->mp );
   if(err) return err;
   err = pthread_cond_destroy( &S->cd );
   if(err) return err;
   return 0;
}
void Sema_wait(Sema *S)
{
   pthread_mutex_lock( &S->mp );
   S->val--;
   if( S->val < 0)
      pthread_cond_wait( &S->cd, &S->mp );
   pthread_mutex_unlock( &S->mp );
}
void Sema_signal(Sema *S)
{
   pthread_mutex_lock( &S->mp );
   S->val++;
   if(S->val <= 0)
      pthread_cond_signal( &S->cd );
   pthread_mutex_unlock( &S->mp );
}

/******************************************************************************
 * Buffer things.
 */
int cirbuf_init( cirbuf_t *buf)
{
   buf->cnt = 0;
   buf->in  = 0;
   buf->out = 0;
   Sema_init( &buf->Space, CIRBUFSIZE );
   Sema_init( &buf->Item, 0);
   return 0;
}
int cirbuf_destroy( cirbuf_t *buf)
{
   Sema_destroy( &buf->Space );
   Sema_destroy( &buf->Item );
   return 0;
}

void cirbuf_insert(cirbuf_t *buf, clients_t *clt)
{
   Sema_wait( &buf->Space );

   assert( buf->cnt < CIRBUFSIZE );

   buf->cl[buf->in++] = clt;
   buf->in %= CIRBUFSIZE;
   buf->cnt++;

   Sema_signal( &buf->Item );
}

clients_t *cirbuf_remove( cirbuf_t *buf )
{
   clients_t *tmp;
   Sema_wait( &buf->Item );

   assert( buf->cnt > 0 );

   tmp = buf->cl[buf->out++];
   buf->out %= CIRBUFSIZE;
   buf->cnt--;

   Sema_signal( &buf->Space );

   return tmp;
}

/******************************************************************************
 * handleRequest()
 *
 * See if we can't get mmapping done instead of lseek&(read|write)
 */
int handleRequest(clients_t *Client, theader_t *ioHdr)
{
   unsigned long len;
   unsigned long long offset;
   int n;
   int clifd = Client->fd;
   pthread_mutex_t *sender = &Client->sender;
   char buf[20480];

   len = ntohl(ioHdr->len);
   offset = ntohll(ioHdr->offset);

#if 0
   printf("ioHdr: type=0x%x, key=0x%"FMT_64"x, len=0x%lx, offset=0x%llx\n",
              ioHdr->type, ioHdr->key, len, offset);
#endif

   if( len > 10240 ) {
      perror("Request too large");
      return -1;
   }
   if( len + offset > datasize) {
      perror("Out of bounds.");
      return -1;
   }

   switch(ioHdr->type) {
      case tioReadReq:
         printff("Doing a Read\n");
         /* Need to get len bytes from offset out of data_fd */
         if(datamap) {
            ioHdr->type = tioReadRpl;

            pthread_mutex_lock( sender );
            n = cpt_send(clifd, ioHdr, theader_len, 0);
            n = cpt_send(clifd, (datamap + offset), len, 0);
            pthread_mutex_unlock( sender );
         } else {
            if(lseek(data_fd, offset, SEEK_SET) <0)
              perror("Read: llseek");
            printff("Done seek\n");
            if(read(data_fd, buf, len) <0)
              perror("Read: read");
            printff("Done read\n");

            ioHdr->type = tioReadRpl;

            pthread_mutex_lock( sender );
            n = cpt_send(clifd, ioHdr, theader_len, 0);
            n = cpt_send(clifd, buf, len, 0);
            pthread_mutex_unlock( sender );
         }
         printff("Done send\n");
         break;
      case tioWriteReq:
         printff("Doing a Write\n");
         /* need a readonly check....*/
         /* we have len bytes that we want to write at offset in data_fd */
         if(datamap) {
            n = cpt_recv(clifd, (datamap + offset), len, 0); 
         } else {
            n = cpt_recv(clifd, buf, len, 0);
            if(n==0) return -1;
            if(lseek(data_fd, offset, SEEK_SET) <0)
              perror("Write: lseek");
            if(write(data_fd, buf, len) <0)
              perror("Write: write");
         }

         ioHdr->type = tioWriteRpl;
         ioHdr->len  = 0;
         ioHdr->err  = 0;
         ioHdr->offset = 0;
         /*expects an ack..*/
         pthread_mutex_lock( sender );
         n = cpt_send(clifd, ioHdr, theader_len, 0);
         pthread_mutex_unlock( sender );
         printff("Done send\n");
         break;
      case tioShutdown:
         /* Gotta close the connection, after replying.*/
         /* ioHdr should be exactly what we waht to send back, so we'll
         * just dump it right back.
         */
         pthread_mutex_lock( sender );
         n = cpt_send(clifd, &ioHdr, theader_len, 0);
         pthread_mutex_unlock( sender );
         printf("ioshutdown fd = %d\n", clifd);
         /* now close shop. */
         return -1;
         break;
      default:
         printf("Unrecognised header.\n");
         printf("ioHdr: type=0x%x, key=0x%"FMT_64"x, len=0x%lx, offset=0x%llx\n",
              ioHdr->type, ioHdr->key, len, offset);
         break;
   }
   return 0;
}

/******************************************************************************
 * drone()
 */
void *drone(void *args)
{
   clients_t *clt=NULL;
   int nread, clifd;
   theader_t ioHdr;
   pid_t mypid;

   mypid = getpid(); /* mostly for debugging */

   printf("drone %d: running\n",mypid);
   while(Server_Runs) {
      /* if I have this right, this will sleep here until an item is
       * queued.
       */
      printf("drone %d: waiting on cirbuf\n",mypid);
      clt = cirbuf_remove( &Buffy );
      assert( clt != NULL );
      printf("drone %d: found that fd %d has data\n", mypid, clt->fd);
      clifd = clt->fd;
      if( clifd == -1 ) continue;

      if(pthread_mutex_trylock( &clt->recver ) == 0) {
         if( (nread = cpt_recv(clifd, &ioHdr, sizeof(theader_t), 0)) <0)
            perror("drone: Read error");
         if(nread <= 0) {
            goto hungup;
         }else{
            if(handleRequest(clt, &ioHdr) <0) {
hungup:
               printf("drone %d: closed fd %d\n", mypid, clifd);
               client_del(clifd);
               FD_CLR(clifd, &allset);
               close(clifd);
            }
         }
         pthread_mutex_unlock( &clt->recver );
      }

   }
   return NULL;
}
/******************************************************************************
 * setkeepalive
 * quicky wrapper for setting the keepalive bit on a socket.
 *
 * This works, BUT if you take a look at the defaults for linux:
 *   cat /proc/sys/net/ipv4/tcp_keepalive_probes
 *   cat /proc/sys/net/ipv4/tcp_keepalive_time
 * you look at those. 9 probes, and 7200 somthings. I think seconds. That's
 * damn long. I reset the timeout on the server machine to 60, and it
 * actually works. if you wait for it.
 *   echo "60" > /proc/sys/net/ipv4/tcp_keepalive_time
 * So I'll leave this here, but we may need a faster method of detecting
 * client death.
 */
int setkeepalive(int sk)
{
   int iopt=1;
   return setsockopt(sk, SOL_SOCKET, SO_KEEPALIVE, &iopt, sizeof(int));
}

/******************************************************************************
 * loop()
 */
void loop(int listenfd)
{
   int i, maxi, clifd, maxfd, nread;
	struct sockaddr_in  theiraddr;
   fd_set rset;


   FD_ZERO(&allset);
   FD_SET(listenfd, &allset);
   maxfd = listenfd;
   maxi = -1;

   while(Server_Runs) {
      rset = allset;
      if( (nread = select(maxfd+1, &rset, NULL,NULL,NULL)) <0) {
         perror("loop: select error");
         exit(1);
      }

      if( FD_ISSET(listenfd, &rset)) {
         int len;
         len = sizeof(struct sockaddr_in);
         if((clifd = accept(listenfd,
                      (struct sockaddr *)&theiraddr, &len))<0){
            perror("loop: error in accept");
            exit(1);
         }
         if( setkeepalive(clifd) <0) {
            perror("loop: setkeepalive failed (ignoring)");
         }
         i = client_add(clifd);
         FD_SET(clifd, &allset);
         if(clifd > maxfd) maxfd = clifd;
         if(i > maxi) maxi=i;

         printf("Added new client[%d] fd=%d\n", i, clifd);
         {/*send login info*/
            /* can do this without locking io since there's no way someone
             * else can be doing io on this until we loop completely again.
             */
            tio_info_t serv_stat;
            serv_stat.devsize = htonll(datasize);
            serv_stat.readonly = readonly;
            nread = cpt_send(clifd, &serv_stat, sizeof(tio_info_t), 0);
         }
      }

      for( i=1; i<= maxi; i++) {
         if( (clifd = Clients[i].fd) <0 )
            continue;
         if(FD_ISSET(clifd, &rset)) {
            printf("loop: data on %d, queuing for drones\n", clifd);
            /* queue this clent for io */
            cirbuf_insert( &Buffy, &Clients[i] );

            /* <sigh> will have to do a read and shit.
             * oh well it was a good idea.
             * mayperhaps work, but may not be worth it.
             */
         }
      }
   }
}

/******************************************************************************
 * signal handler.
 * Send a SIGUSR1 to stop threads.
 */
void mysigact(int sig)
{
   if(sig==SIGUSR1)
      Server_Runs = FALSE;
}

/******************************************************************************
 * main()
 * Parse some options, checkout the world, listen to sockets, spawn
 * threads. You know, same old, same old.
 */
int main(int argc, char **argv)
{
   int listenfd;
   struct sockaddr_in  myaddr;
	pthread_t tmpTid;

	int c;
	extern char *optarg;
	extern int optind;
	int port = 14243;
   int threads = 1;

/* get options. */
	while((c = getopt(argc, argv, "s:t:p:rh?"))!= -1){
		switch(c){
         case 's':
            datasize=atoi(optarg);
			case 'p':
				port=atoi(optarg);
				break;
			case 'r':
				readonly=TRUE;
				break;
         case 't':
            threads=atoi(optarg);
            break;
			case '?':
			case 'h':
				printf("Usage: %s [-p port] [-s size] "
                   "[-t max thread cnt ] <device or file>\n",
                  argv[0]);
				return(1);
			break;
		}
	}

	/* setup signal handlers*/
   signal(SIGUSR1, mysigact);
	signal(SIGPIPE, SIG_IGN);

	/*open the device.*/
	data_fd = open( argv[optind], (readonly?O_RDONLY:O_RDWR)|O_SYNC );
	if(data_fd < 0) {
		perror("main: Failed to open data file");
		return 1;
	}
   if(datasize==0) {
		struct stat st;
		if(fstat(data_fd,&st)) {
			perror( "Could not fstat exported file");
			return 1;
		}
      if( S_ISBLK(st.st_mode) ) {
         if(ioctl(data_fd, BLKGETSIZE, (long*)&datasize) <0) {
            perror("Couldn't get sector count on block device");
            return 1;
         }
         datasize *= 512; /*sector size is hardcoded.*/
      }else{
         datasize=(u64)st.st_size;
      }
      if(datasize == 0) {
         fprintf(stderr,"Could not determin size of exported data\n");
         return 1;
      }
	}
   printf("Device size = %llx\n", datasize);

   /*attempt to mmap it. */
   datamap = mmap( NULL, datasize,
		        readonly?PROT_READ:(PROT_READ|PROT_WRITE),
		        MAP_SHARED,
		        data_fd, 0);
   if(datamap==(caddr_t)-1 || datamap==NULL) {
      printf("mmap() failed.  Will use lseek/read/write instead.\n");
      datamap=NULL;
   }
   
   /* setup socket */
	if( (listenfd = socket(AF_INET, SOCK_STREAM, 0)) <0 ) {
		perror("loop: Socket biffed");
		exit(1);
	}

	/* Set myself up to listen to the world. */
	myaddr.sin_family=AF_INET;
	myaddr.sin_addr.s_addr=INADDR_ANY;
	myaddr.sin_port=htons(port);
	if( bind(listenfd, (struct sockaddr *)&myaddr,
            sizeof(struct sockaddr_in)) <0) {
		perror("loop: Bind Biffed");
		exit(1);
	}

	if( listen(listenfd, 5) <0 )
		perror("loop: Listen Biffed");

   client_add(listenfd);

   /* initalize circular buffer (for incomming requests)*/
   cirbuf_init( &Buffy );

   /* Start drones loops */
   for(;threads > 0;threads--) {
      if( pthread_create( &tmpTid, NULL, drone, (void *)&listenfd) <0)
         perror("main: thread_create biffed");
      if( pthread_detach( tmpTid ) <0)
         perror("main: thread_detach biffed");
   }

   loop(listenfd);

   cirbuf_destroy( &Buffy );

   if(datamap)
      munmap(datamap, datasize);
	close(data_fd);
	return 0;
}

