/******************************************************************************
*
*    Copyright 1999 Regents of the University of Minnesota
*    Portions Copyright 1999-2001 Sistina Software, Inc.
*
*    This is free software released under the GNU General Public License.
*    There is no warranty for this software.  See the file COPYING for
*    details.
*
*    See the file CONTRIBUTORS for a list of contributors.
*
*    This file is maintained by:
*      Benjamin E Marzinski <marzinski@sistina.com>
*
* 
*/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <sys/mount.h>
#include <sys/mman.h>
#include <errno.h>

#include "header.h"

/* some private data structures. */
typedef struct {
  int fd;
  pthread_mutex_t mux;
  pthread_mutex_t sender;
  /* other info? */
} clients_t;

/* Table of clients that are logged in */
clients_t *Clients=NULL;
int client_cnt=0;

/* This is so there will only ever be one thread looking at accepts. */
pthread_mutex_t AcceptMux;

/* This global keeps the server running. */
static int Server_Runs = TRUE;

/* This file descriptor is the data we're serving. */
static int data_fd=-1;
static u64 datasize=0; /*bytes*/
static char *datamap=NULL;
static int readonly = FALSE;

/******************************************************************************
 * dump_buffer()
 * prints out the contents of a buffer for inspection.  ignores formats.
 *
 * Need to get all zero lines absorbed.
 */
void dump_buffer(void *buf, int len)
{
  int i;
  u8 *c = buf;

  for(i=0; i<len; i++) {
    if( i % 16 == 0 )
      printf("\n%4u: ", i);
    printf("%2x ", c[i]);
  }
  printf("\n");
}
/******************************************************************************
 * heh, simple recv/send wrappers to get all of the beffer sent over the
 * net.  Had forgotten I needed this....
 */
int cpt_recv(int s, void *buf, size_t len, int flags)
{
  int res;
  do {
    res = recv(s, buf, len, flags);
    if(res<0) return res;
    len -= res;
    buf += res;
  }while(len>0);
  return res;
}

int cpt_send(int s, void *buf, size_t len, int flags)
{
  int res;
  do {
    res = send(s, buf, len, flags);
    if(res<0) return res;
    len -= res;
    buf += res;
  }while(len>0);
  return res;
}

/******************************************************************************
 * client_alloc()
 * expands the size of the clients table.
 */
#define NALLOC 10
void client_alloc(void)
{
  int i;
  if( Clients == NULL )
    Clients = malloc(NALLOC * sizeof(clients_t));
  else
    Clients = realloc(Clients, (client_cnt + NALLOC) * sizeof(clients_t));
  if(Clients == NULL) {
    perror("Can't alloc memory for client array");
    exit(1);
  }
  for(i=client_cnt; i < client_cnt+NALLOC; i++) {
    Clients[i].fd = -1;
    pthread_mutex_init( &Clients[i].mux, NULL);
    pthread_mutex_init( &Clients[i].sender, NULL);
  }
  client_cnt += NALLOC;
  printf("realloced clients\n");
}
/******************************************************************************
 * client_add()
 */
int client_add(int fd)
{
  int i;
   
  if(Clients == NULL)
    client_alloc();
 CAagain:
  for( i=0; i<client_cnt; i++) {
    if( Clients[i].fd == -1) {
      Clients[i].fd = fd;
      pthread_mutex_init( &Clients[i].mux, NULL);
      pthread_mutex_init( &Clients[i].sender, NULL);
      return i;
    }
  }

  /* client array must be full, expand it. */
  client_alloc();
  goto CAagain; /* now go add it. */
}
/******************************************************************************
 * client_del()
 * resetting the mutexes here may not be a good idea. do some checking.
 */
void client_del(int fd)
{
  int i;
  for(i=0; i<client_cnt; i++) {
    if( Clients[i].fd == fd) {
      Clients[i].fd = -1;
      pthread_mutex_destroy( &Clients[i].mux );
      pthread_mutex_destroy( &Clients[i].sender );
      return;
    }
  }
  /* pop an error? */
}

/******************************************************************************
 * handleRequest()
 *
 * See if we can't get mmapping done instead of lseek&(read|write)
 */
int handleRequest(clients_t *Client, theader_t *ioHdr)
{
  unsigned long len;
  unsigned long long offset;
  int n;
  int clifd = Client->fd;
  pthread_mutex_t *sender = &Client->sender;
  char buf[20480];

  len = ntohl(ioHdr->len);
  offset = ntohll(ioHdr->offset);

#if 0
  printf("ioHdr: type=0x%x, key=0x%"FMT_64"x, len=0x%lx, offset=0x%"FMT_64"x\n",
	 ioHdr->type, ioHdr->key, len, offset);
#endif

  if( len > 10240 ) {
    perror("Request too large");
    return -1;
  }
  if( len + offset > datasize) {
    perror("Out of bounds.");
    return -1;
  }

  switch(ioHdr->type) {
  case tioReadReq:
    printff("Doing a Read\n");
    /* Need to get len bytes from offset out of data_fd */
    if(datamap) {
      ioHdr->type = tioReadRpl;

      pthread_mutex_lock( sender );
      n = cpt_send(clifd, ioHdr, theader_len, 0);
      n = cpt_send(clifd, (datamap + offset), len, 0);
      pthread_mutex_unlock( sender );
    } else {
      if(lseek(data_fd, offset, SEEK_SET) <0)
	perror("Read: llseek");
      printff("Done seek\n");
      if(read(data_fd, buf, len) <0)
	perror("Read: read");
      printff("Done read\n");

      ioHdr->type = tioReadRpl;

      pthread_mutex_lock( sender );
      n = cpt_send(clifd, ioHdr, theader_len, 0);
      n = cpt_send(clifd, buf, len, 0);
      pthread_mutex_unlock( sender );
    }
    printff("Done send\n");
    break;
  case tioWriteReq:
    printff("Doing a Write\n");
    /* need a readonly check....*/
    /* we have len bytes that we want to write at offset in data_fd */
    if(datamap) {
      n = cpt_recv(clifd, (datamap + offset), len, 0); 
    } else {
      n = cpt_recv(clifd, buf, len, 0);
      if(n==0) return -1;
      if(lseek(data_fd, offset, SEEK_SET) <0)
	perror("Write: lseek");
      if(write(data_fd, buf, len) <0)
	perror("Write: write");
    }

    ioHdr->type = tioWriteRpl;
    ioHdr->len  = 0;
    ioHdr->err  = 0;
    ioHdr->offset = 0;
    /*expects an ack..*/
    pthread_mutex_lock( sender );
    n = cpt_send(clifd, ioHdr, theader_len, 0);
    pthread_mutex_unlock( sender );
    printff("Done send\n");
    break;
  case tioShutdown:
    /* Gotta close the connection, after replying.*/
    /* ioHdr should be exactly what we waht to send back, so we'll
     * just dump it right back.
     */
    pthread_mutex_lock( sender );
    n = cpt_send(clifd, &ioHdr, theader_len, 0);
    pthread_mutex_unlock( sender );
    /* now close shop. */
    return -1;
    break;
  default:
    printf("Unrecognised header.\n");
    printf("ioHdr: type=0x%x, key=0x%"FMT_64"x, len=0x%lx, offset=0x%llx\n",
	   ioHdr->type, ioHdr->key, len, offset);
    break;
  }
  return 0;
}

/******************************************************************************
 * setkeepalive
 * quicky wrapper for setting the keepalive bit on a socket.
 *
 * This works, BUT if you take a look at the defaults for linux:
 *   cat /proc/sys/net/ipv4/tcp_keepalive_probes
 *   cat /proc/sys/net/ipv4/tcp_keepalive_time
 * you look at those. 9 probes, and 7200 somthings. I think seconds. That's
 * damn long. I reset the timeout on the server machine to 60, and it
 * actually works. if you wait for it.
 *   echo "60" > /proc/sys/net/ipv4/tcp_keepalive_time
 * So I'll leave this here, but we may need a faster method of detecting
 * client death.
 */
int setkeepalive(int sk)
{
  int iopt=1;
  return setsockopt(sk, SOL_SOCKET, SO_KEEPALIVE, &iopt, sizeof(int));
}

/******************************************************************************
 * loop()
 * the actual main loop of this thing
 *
 * Try using select this time.
 */
void *loop(void *args)
{
  int *t = (int *)args;
  int listenfd = *t;
  int i, maxi, clifd, maxfd, nread;
  int err = 0;
  theader_t ioHdr;
  struct sockaddr_in  theiraddr;
  fd_set rset, allset;
  pid_t mypid;

  mypid = getpid(); /* mostly for debugging */

  FD_ZERO(&allset);

  FD_SET(listenfd, &allset);
  maxfd = listenfd;
  maxi = -1;

  while(Server_Runs) {
    rset = allset;
    if( (nread = select(maxfd+1, &rset, NULL,NULL,NULL)) <0) {
      perror("loop: select error");
      exit(1);
    }

    /* If someone else is doing an accept, just skip over it. */
    if( pthread_mutex_trylock( &AcceptMux) == 0 ) {
      if( FD_ISSET(listenfd, &rset)) {
	int len;
	len = sizeof(struct sockaddr_in);
	if((clifd = accept(listenfd,
			   (struct sockaddr *)&theiraddr, &len))<0){
	  perror("loop: error in accept");
	  exit(1);
	}
	if( setkeepalive(clifd) <0) {
	  perror("loop: setkeepalive failed (ignoring)");
	}
	i = client_add(clifd);
	FD_SET(clifd, &allset);
	if(clifd > maxfd) maxfd = clifd;
	if(i > maxi) maxi=i;

	printf("Added new client[%d] fd=%d\n", i, clifd);
	{/*send login info*/
	  tio_info_t serv_stat;
	  serv_stat.devsize = htonll(datasize);
	  serv_stat.readonly = readonly;
	  nread = cpt_send(clifd, &serv_stat, sizeof(tio_info_t), 0);
	}
      }
      pthread_mutex_unlock(&AcceptMux);
    }

    for( i=1; i<= maxi; i++) {
      if( (clifd = Clients[i].fd) <0 )
	continue;
      if(FD_ISSET(clifd, &rset)) {
	if(pthread_mutex_trylock( &Clients[i].mux ) == 0) {
#if 0
	  printf("thread %d: got Clients[%d].mux\n",mypid,i);
#endif
	  if( (nread = cpt_recv(clifd, &ioHdr, sizeof(theader_t), 0)) <0)
	    perror("loop: Read error");
	  if(nread <= 0) {
	    goto hungup;
	  }else{
	    if(handleRequest(&Clients[i], &ioHdr) <0) {
	    hungup:
	      printf("loop: closed fd %d\n",clifd);
	      client_del(clifd);
	      FD_CLR(clifd, &allset);
	      close(clifd);
	    }
	  }
	  pthread_mutex_unlock( &Clients[i].mux );
#if 0
	  printf("thread %d: unlocked Clients[%d].mux\n",mypid,i);
	} else {
	  printf("thread %d: missed Clients[%d].mux\n",mypid,i);
#endif
	}
      }
    }
  }
  pthread_exit(&err);
}

/******************************************************************************
 * signal handler.
 * Send a SIGUSR1 to stop threads.
 */
void mysigact(int sig)
{
  if(sig==SIGUSR1)
    Server_Runs = FALSE;
}


int send_to_server(unsigned int val, uint16_t port){
  char addr[4] = {127,0,0,1};
  u32 *addr_32 = (u32 *)&addr[0];
  struct in_addr host;
  int sock_fd;
  struct sockaddr_in server;

  /*create the socket and fill the sockaddr with the inputs */
  host.s_addr = *addr_32;
  sock_fd = socket(AF_INET, SOCK_DGRAM, 0);
  if (sock_fd < 0){
    fprintf(stderr, "error creating socket: %s\n", strerror(errno));
    return -1;
  }
  server.sin_family = AF_INET;
  server.sin_addr = host;
  server.sin_port = htons(port);

  if (sendto(sock_fd, &val, sizeof(unsigned int), 0,
	     (struct sockaddr *)&server, sizeof(server)) < 0){
    perror("main: couldn't send the port number");
    close(sock_fd);
    return -1;
  }
  fprintf(stderr, "time to close\n");
  close(sock_fd);
  return 0;
}


/******************************************************************************
 * main()
 * Parse some options, checkout the world, listen to sockets, spawn
 * threads. You know, same old, same old.
 */
int main(int argc, char **argv)
{
  int listenfd;
  struct sockaddr_in  myaddr;
  pthread_t tmpTid;
  unsigned int port;
  int c;
  int length = sizeof(struct sockaddr_in);
  extern char *optarg;
  extern int optind;
  int threads = 1;

  /* get options. */
  if (argc < 2){
    printf("Usage: %s " /* [-p port] [-s size ]*/
	   "<device or file>\n", /*[-t max thread cnt ]*/
	   argv[0]);
    return(1);
  }
  while((c = getopt(argc, argv, "s:t:rh?"))!= -1){
    switch(c){
    case 's':
      datasize=atoi(optarg);
      /*
    case 'p':
      port=atoi(optarg);
      break;
      */
    case 'r':
      readonly=TRUE;
      break;
    case 't':
      threads=atoi(optarg);
      break;
    case '?':
    case 'h':
      printf("Usage: %s [-p port] " /*[-s size]*/
	     "<device or file>\n", /*[-t max thread cnt ]*/
	     argv[0]);
      return(1);
      break;
    }
  }

#if 0
  printf(" sizeof theader_t %u\n", sizeof(theader_t));
  printf(" sizeof tio_info_t %u\n", sizeof(tio_info_t));
#endif

  /* setup signal handlers*/
  signal(SIGUSR1, mysigact);
  signal(SIGPIPE, SIG_IGN);

  /*open the device.*/
  data_fd = open( argv[optind], (readonly?O_RDONLY:O_RDWR)|O_SYNC );
  if(data_fd < 0) {
    perror("main: Failed to open data file");
    return 1;
  }
  if(datasize==0) {
    struct stat st;
    if(fstat(data_fd,&st)) {
      perror( "Could not fstat exported file");
      return 1;
    }
    if( S_ISBLK(st.st_mode) ) {
      if(ioctl(data_fd, BLKGETSIZE, (long*)&datasize) <0) {
	perror("Couldn't get sector count on block device");
	return 1;
      }
      datasize *= 512; /*sector size is hardcoded.*/
    }else{
      datasize=(u64)st.st_size;
    }
    if(datasize == 0) {
      fprintf(stderr,"Could not determin size of exported data\n");
      return 1;
    }
  }
  printf("Device size = %"FMT_64"x\n", datasize);

  /*attempt to mmap it. */
  /*datamap = mmap( NULL, datasize,
		  readonly?PROT_READ:(PROT_READ|PROT_WRITE),
		  MAP_SHARED,
		  data_fd, 0);
  if(datamap==(caddr_t)-1 || datamap==NULL) {
    printf("mmap() failed.  Will use lseek/read/write instead.\n");
    datamap=NULL;
    }*/
   
  datamap=NULL;
  
  pthread_mutex_init( &AcceptMux, NULL);

  /* setup socket */
  if( (listenfd = socket(AF_INET, SOCK_STREAM, 0)) <0 ) {
    perror("loop: Socket biffed");
    exit(1);
  }

  /* Set myself up to listen to the world. */
  myaddr.sin_family=AF_INET;
  myaddr.sin_addr.s_addr=INADDR_ANY;
  myaddr.sin_port=0;
  if( bind(listenfd, (struct sockaddr *)&myaddr,
	   sizeof(struct sockaddr_in)) <0) {
    perror("loop: Bind Biffed");
    exit(1);
  }

  if( listen(listenfd, 5) <0 ){
    perror("loop: Listen Biffed");
    exit(1);
  }
  if (getsockname(listenfd, (struct sockaddr *)&myaddr, &length) < 0){
    perror("error geting the port number");
    exit(1);
  }							    

  client_add(listenfd);

  /* Start Accept loops */
  for(threads--;threads > 0;threads--) {
    if( pthread_create( &tmpTid, NULL, loop, (void *)&listenfd) <0)
      perror("main: thread_create biffed");
    if( pthread_detach( tmpTid ) <0)
      perror("main: thread_detach biffed");
  }

  /* Link the main thread to the last thread we create. This is kinda bad.
   * really should build a counting semaphore here, and have each worker
   * thread dec that, and when it hits the end, then the main thread can
   * exit. 
   * As it is now, when this attached thread dies, the app dies. Even if
   * there is activity on the other threads.
   *
   * But this will work for proof of concept just fine,
   */
  if( pthread_create( &tmpTid, NULL, loop, (void *)&listenfd) <0)
    perror("main: thread_create biffed");

  fprintf(stderr, "all threads created\n");
  
  port = (unsigned int)ntohs(myaddr.sin_port);
  if (send_to_server(port, 24314) < 0)
    exit(1);

  if( pthread_join( tmpTid, (void*)&c ) <0){
    perror("main: thread_join biffed");
    exit(1);
  }

  if(datamap)
    munmap(datamap, datasize);
  close(data_fd);
  return 0;
}










