/******************************************************************************
*******************************************************************************
**
**    Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: gfs_linux.h
**
*******************************************************************************
******************************************************************************/

#ifndef __OSI_LINUX22_DOT_H__
#define __OSI_LINUX22_DOT_H__

/*  Include Files
    These should only be the ones necessary to compile the common code.  */

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif  /*  MODVERSIONS  */
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <asm/uaccess.h>
#include <asm/spinlock.h>
#include <asm/atomic.h>
#include <linux/timer.h>
#include <linux/locks.h>



/*  Memory allocation abstraction  */

static __inline__ void *osi_malloc(unsigned int size)
{
  void *data;
  
  if (size <= 131072)
    data = kmalloc(size, GFP_BUFFER);
  else
    data = vmalloc(size);

  return data;
}

static __inline__ void osi_free(void *data, unsigned int size)
{
  if (size <= 131072)
    kfree(data);
  else
    vfree(data);
}



/*  Memory copy abstraction  */

#define osi_copy_to_user copy_to_user
#define osi_clear_user clear_user
#define osi_copy_from_user copy_from_user

#define osi_memset memset
#define osi_memcpy memcpy
#define osi_memcmp memcmp

#define osi_strlen strlen
#define osi_strstr strstr
#define osi_strcmp strcmp
#define osi_strtok strtok
#define osi_strchr strchr
#define osi_strcpy strcpy



/*  printf() abstraction  */

#define osi_printf printk
#define osi_sprintf sprintf



/*  panic abstractions  */

#define osi_panic hard_panic
#define osi_bug() BUG()



/*  GFS to VFS abstraction  */

typedef struct super_block osi_vfs_t;
typedef struct inode osi_vnode_t;
typedef struct file osi_vfile_t;



/*  I/O abstraction  */

typedef struct buffer_head osi_buf_t;

/*  Sync all dirty buffers  */

static __inline__ int osi_sync_dev(osi_vfs_t *vfs, int wait)
{
  return sync_buffers(vfs->s_dev, wait);
}



/*  Device type abstraction  */

typedef kdev_t osi_dev_t;

#define osi_make_dev MKDEV
#define osi_u2k_dev to_kdev_t
#define osi_k2u_dev kdev_t_to_nr



/*  Atomic Bit Manipulation interface  */

#define osi_test_bit test_bit
#define osi_set_bit set_bit
#define osi_clear_bit clear_bit
#define osi_test_and_set_bit test_and_set_bit
#define osi_test_and_clear_bit test_and_clear_bit

typedef atomic_t osi_atomic_t;

#define osi_atomic_inc atomic_inc
#define osi_atomic_dec atomic_dec
#define osi_atomic_read atomic_read
#define osi_atomic_dec_and_test atomic_dec_and_test
#define osi_atomic_set atomic_set
#define osi_atomic_add atomic_add
#define osi_atomic_sub atomic_sub



/*  Endianness conversion abstraction  */

#define osi_be16_to_cpu be16_to_cpu
#define osi_be32_to_cpu be32_to_cpu
#define osi_be64_to_cpu be64_to_cpu

#define osi_cpu_to_be16 cpu_to_be16
#define osi_cpu_to_be32 cpu_to_be32
#define osi_cpu_to_be64 cpu_to_be64

#define osi_le16_to_cpu le16_to_cpu
#define osi_le32_to_cpu le32_to_cpu
#define osi_le64_to_cpu le64_to_cpu

#define osi_cpu_to_le16 cpu_to_le16
#define osi_cpu_to_le32 cpu_to_le32
#define osi_cpu_to_le64 cpu_to_le64



/*  struct qstr definition 
 
    "quick string" -- eases parameter passing, but more importantly
    saves "metadata" about the string (ie length and the hash).
    */

typedef struct qstr osi_qstr_t;



/*  Semaphore abstraction

    All macros take a pointer to a osi_semaphore_t structure.
    osi_sema_down_try() returns TRUE if the down() succeeds 
    */

typedef struct semaphore osi_sema_t;

#define osi_sema_init(x) (*(x) = MUTEX)
#define osi_sema_init_lkd(x) {*(x) = MUTEX_LOCKED;}

#define osi_sema_lock(x) down((x))
#define osi_sema_lock_intr(x) down_interruptible((x))
#define osi_sema_trylock(x) (!down_trylock((x)))
#define osi_sema_unlock(x) up((x))



/*  Spinlock abstraction  */

typedef spinlock_t osi_spinlock_t;
typedef rwlock_t osi_rwlock_t;

#define osi_spinlock_init(x) (*(x) = SPIN_LOCK_UNLOCKED)
#define osi_rwlock_init(x) (*(x) = RW_LOCK_UNLOCKED)

#define osi_spin_lock(lock) spin_lock(lock)
#define osi_spin_unlock(lock) spin_unlock(lock)
#define osi_spin_trylock(lock) spin_trylock(lock)

#define osi_read_lock(lock) read_lock(lock)
#define osi_write_lock(lock) write_lock(lock)
#define osi_read_unlock(lock) read_unlock(lock)
#define osi_write_unlock(lock) write_unlock(lock)
#define osi_write_trylock(lock) write_trylock(lock)



/*  Process abstraction  */

#define osi_pid() ((long)(current->pid))
#define osi_get_transaction() ((gfs_trans_t *)current->fs_transaction)
#define osi_set_transaction(x) (current->fs_transaction = (void *)(x))

#define osi_wake_up_process(x) wake_up_process(x)
typedef struct task_struct osi_task_t;



/*  Time abstraction  */

#define osi_current_time() (CURRENT_TIME)

static __inline__ uint64_t osi_gettimeofday(void)
{
  struct timeval tv;
  do_gettimeofday(&tv);
  return((uint64_t)tv.tv_sec * 1000000 + tv.tv_usec);
}

typedef unsigned long osi_clock_ticks_t;
#define osi_time_stamp(x) (*(x) = jiffies)

static __inline__ int osi_check_timeout(osi_clock_ticks_t *t, int hz)
{
  return ((jiffies - *(t)) >= ((hz) * HZ));
}



/*  Timer abstraction  */

typedef struct timer_list osi_timer_t;
typedef void (osi_timer_func)(void *);
typedef void (linux_timer_func)(unsigned long);

/*  creates a new timer.  */
static __inline__ void osi_init_timer(osi_timer_t *t,
				      osi_timer_func *fp, void *data)
{
  init_timer( (struct timer_list *) t);
  ((struct timer_list *) t)->function = (linux_timer_func*)fp;
  ((struct timer_list *) t)->data = (unsigned long)data;
}

/*  deletes a timer. stopping it if it is running.  */
static __inline__ void osi_del_timer( osi_timer_t *t)
{
  del_timer((struct timer_list *) t);
}

/* sets a timer to go off in given seconds.
 * if executed on an existing running timer, resets it.
 */
static __inline__ void osi_set_timer( osi_timer_t * t, unsigned long sec)
{
   mod_timer((struct timer_list *) t, jiffies + (sec *HZ) );
}



/*  Schedule abstraction - Macro that makes a process temporarily
    give up control of the processor and lets other processes have
    a change to run. 
    Sleep abstraction - Sleep for x number of sections.  */

#define osi_schedule() do { schedule(); } while (0)
#define osi_sleep(x) do { \
			    current->state = TASK_INTERRUPTIBLE; \
			    schedule_timeout((x) * HZ); \
			} while (0)
#define osi_sleep_unint(x) do { \
				  current->state = TASK_UNINTERRUPTIBLE; \
				  schedule_timeout((x) * HZ); \
			      } while (0)
#define osi_hibernate() do { \
			       current->state = TASK_INTERRUPTIBLE; \
			       schedule(); \
			   } while (0)



/*  Wait queue abstraction  */

typedef struct wait_queue * osi_wait_chan_t;

#define osi_wait_vars(x, y)	struct wait_queue x = { current, NULL }
#define osi_init_wait_vars(x)   (x).task = current; (x).next = NULL;
#define osi_init_wait_chan(x)	init_waitqueue(&(x));
#define osi_asleep(x, y)	current->state = TASK_UNINTERRUPTIBLE; \
				add_wait_queue(&(y), &(x))

#define osi_await(x)		schedule()

#define osi_unsleep(x, y)	remove_wait_queue(&(y), &(x)); \
				current->state = TASK_RUNNING
#define osi_wakeup(x)		wake_up(&(x))



/*  Credentials structure  */

typedef void osi_cred_t;

#define OSI_MAY_READ MAY_READ
#define OSI_MAY_WRITE MAY_WRITE
#define OSI_MAY_EXEC MAY_EXEC



/* signals abstraction */

#define osi_pending_signals() signal_pending(current)



/*  Other Stuff  */



#endif  /*  __OSI_LINUX22_DOT_H__  */

