/******************************************************************************
*******************************************************************************
**
**    This code was shamelessly stolen from the Linux kernel.
**
**    It is Copyright Linus Torvalds and others.  It is licensed under the
**    GNU GPL.
**    
**    Module name: osi_list.h
**
**    Description:
**      This code implements standard list code across different OSs.
**
*******************************************************************************
******************************************************************************/

#ifndef __OSI_LIST_DOT_H__
#define __OSI_LIST_DOT_H__

/*
 * Simple doubly linked osi_list implementation.
 *
 * Some of the internal functions ("__xxx") are useful when
 * manipulating whole osi_lists rather than single entries, as
 * sometimes we already know the next/prev entries and we can
 * generate better code by using them directly rather than
 * using the generic single-entry routines.
 */

struct osi_list_head {
	struct osi_list_head *next, *prev;
};
typedef struct osi_list_head osi_list_head_t;

#define OSI_LIST_HEAD(name) \
	struct osi_list_head name = { &name, &name }

#define osi_list_head_init(ptr) do { \
	(ptr)->next = (ptr); (ptr)->prev = (ptr); \
} while (0)

/*
 * Insert a new entry between two known consecutive entries. 
 *
 * This is only for internal osi_list manipulation where we know
 * the prev/next entries already!
 */
static __inline__ void __osi_list_add(struct osi_list_head * new,
	struct osi_list_head * prev,
	struct osi_list_head * next)
{
	next->prev = new;
	new->next = next;
	new->prev = prev;
	prev->next = new;
}

/*
 * Insert a new entry after the specified head..
 */
static __inline__ void osi_list_add(struct osi_list_head *new, struct osi_list_head *head)
{
	__osi_list_add(new, head, head->next);
}

static __inline__ void osi_list_add_prev(struct osi_list_head *new, struct osi_list_head *head)
{
	__osi_list_add(new, head->prev, head);
}

/*
 * Delete a osi_list entry by making the prev/next entries
 * point to each other.
 *
 * This is only for internal osi_list manipulation where we know
 * the prev/next entries already!
 */
static __inline__ void __osi_list_del(struct osi_list_head * prev,
				  struct osi_list_head * next)
{
	next->prev = prev;
	prev->next = next;
}

static __inline__ void osi_list_del(struct osi_list_head *entry)
{
	__osi_list_del(entry->prev, entry->next);
}

static __inline__ int osi_list_empty(struct osi_list_head *head)
{
	return head->next == head;
}

/*
 * Splice in "osi_list" into "head"
 */
static __inline__ void osi_list_splice(struct osi_list_head *osi_list, struct osi_list_head *head)
{
	struct osi_list_head *first = osi_list->next;

	if (first != osi_list) {
		struct osi_list_head *last = osi_list->prev;
		struct osi_list_head *at = head->next;

		first->prev = head;
		head->next = first;

		last->next = at;
		at->prev = last;
	}
}

#define osi_list_entry(ptr, type, member) \
	((type *)((char *)(ptr)-(unsigned long)(&((type *)0)->member)))

#endif  /*  __OSI_LIST_DOT_H__  */
