/******************************************************************************
*******************************************************************************
**
**    Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland <teigland@sistina.com>
** 
**    Module name: osi_user.h
**
*******************************************************************************
******************************************************************************/

#ifndef __OSI_USER_DOT_H__
#define __OSI_USER_DOT_H__

/*  Include Files
    These should only be the ones necessary to compile the common code.  */


#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdint.h>
#include <time.h>

#include <asm/cache.h>  /* SMP_CACHE_BYTES */

#include "osi_endian.h"
#include "osi_list.h"


/*  Memory allocation abstraction  */

static __inline__ void *osi_malloc(unsigned int size)
{
  void *data;
  
  data = malloc(size);

  return data;
}

static __inline__ void osi_free(void *data, unsigned int size)
{
  free(data);
}



/*  Memory copy abstraction  */

#define osi_copy_to_user memcpy
#define osi_clear_user(x, y) memset((x), 0, (y))
#define osi_copy_from_user memcpy

#define osi_memset memset
#define osi_memcpy memcpy
#define osi_memcmp memcmp

#define osi_strlen strlen
#define osi_strstr strstr
#define osi_strcmp strcmp
#define osi_strtok strtok
#define osi_strchr strchr
#define osi_strcpy strcpy



/*  printf() abstraction  */

#define osi_printf printf
#define osi_sprintf sprintf



/*  panic abstractions  */

#define panic(fmt, args...) \
{ \
  fprintf(stderr, "%s: ", prog_name); \
  fprintf(stderr, fmt, ## args); \
  exit(-1); \
}

#define osi_panic panic

#define osi_bug() \
{ \
  fprintf(stderr, "%s: ", prog_name); \
  fprintf(stderr, "BUG\n"); \
  exit(-1); \
}



/*  GFS to VFS abstraction  */

typedef int osi_vfs_t;
typedef int osi_vnode_t;
typedef int osi_vfile_t;



/*  I/O abstraction  */

struct buffer_head
{
  unsigned long b_blocknr;
  unsigned long b_size;
  unsigned long b_state;
  char *b_data;
  char *b_pdata;
  osi_list_head_t b_list;
};
typedef struct buffer_head osi_buf_t;


/*  Sync all dirty buffers  */

static __inline__ int osi_sync_dev(osi_vfs_t *vfs, int wait)
{
  return 0;
}



/*  Device type abstraction  */

typedef dev_t osi_dev_t;

#define osi_make_dev MKDEV
#define osi_u2k_dev
#define osi_k2u_dev



/*  Atomic Bit Manipulation interface  */

#define osi_test_bit(nr, addr) (*(addr) &   (1 << nr))
#define osi_set_bit(nr, addr) (*(addr) |=  (1 << nr))
#define osi_clear_bit(nr, addr) (*(addr) &= ~(1 << nr))
#define osi_test_and_set_bit(nr, addr) (*(addr) |=  (1 << nr))
#define osi_test_and_clear_bit(nr, addr) (*(addr) &= ~(1 << nr))

typedef int osi_atomic_t;

#define osi_atomic_inc(x) (*(x) += 1)
#define osi_atomic_dec(x) (*(x) -= 1)
#define osi_atomic_read(x) (*(x))
#define osi_atomic_dec_and_test(x) (*(x) -= 1)
#define osi_atomic_set(x, y) (*(x) = (y))
#define osi_atomic_add(x, y) (*(y) += (x))
#define osi_atomic_sub(x, y) (*(y) -= (x))



/*  Endianness conversion abstraction  */

#define osi_be16_to_cpu be16_to_cpu
#define osi_be32_to_cpu be32_to_cpu
#define osi_be64_to_cpu be64_to_cpu

#define osi_cpu_to_be16 cpu_to_be16
#define osi_cpu_to_be32 cpu_to_be32
#define osi_cpu_to_be64 cpu_to_be64

#define osi_le16_to_cpu le16_to_cpu
#define osi_le32_to_cpu le32_to_cpu
#define osi_le64_to_cpu le64_to_cpu

#define osi_cpu_to_le16 cpu_to_le16
#define osi_cpu_to_le32 cpu_to_le32
#define osi_cpu_to_le64 cpu_to_le64



/*  struct qstr definition 
 
    "quick string" -- eases parameter passing, but more importantly
    saves "metadata" about the string (ie length and the hash).
    */

struct qstr {
  const unsigned char * name;
  unsigned int len;
  unsigned int hash;
};
typedef struct qstr osi_qstr_t;



/*  Semaphore abstraction

    All macros take a pointer to a osi_semaphore_t structure.
    osi_sema_down_try() returns TRUE if the down() succeeds 
    */

typedef int osi_sema_t;

#define osi_sema_init(x) (*(x) = 1)
#define osi_sema_init_lkd(x) (*(x) = 0)

#define osi_sema_lock(x) (*(x) -= 1)
#define osi_sema_lock_intr(x) (*(x) -= 1)
#define osi_sema_trylock(x) (*(x) -= 1)
#define osi_sema_unlock(x) (*(x) += 1)



/*  Spinlock abstraction  */

typedef int osi_spinlock_t;
typedef int osi_rwlock_t;

#define osi_spinlock_init(x) (*(x) = 0)
#define osi_rwlock_init(x) (*(x) = 0)

#define osi_spin_lock(lock)
#define osi_spin_unlock(lock)
#define osi_spin_trylock(lock)

#define osi_read_lock(lock)
#define osi_write_lock(lock)
#define osi_read_unlock(lock)
#define osi_write_unlock(lock)
#define osi_write_trylock(lock)



/*  Process abstraction  */

#define osi_pid() 1
#define osi_get_transaction() ((gfs_trans_t *)(NULL))
#define osi_set_transaction(x)

#define osi_wake_up_process(x)

typedef int osi_task_t;



/*  Time abstraction  */

#define osi_current_time() time(NULL)

static __inline__ uint64_t osi_gettimeofday(void)
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return((uint64_t)tv.tv_sec * 1000000 + tv.tv_usec);
}

typedef unsigned long osi_clock_ticks_t;
#define osi_time_stamp(x)

static __inline__ int osi_check_timeout(osi_clock_ticks_t *t, int hz)
{
  return 0;
}



/*  Timer abstraction  */

typedef int osi_timer_t;
typedef void (osi_timer_func)(void *);
typedef void (linux_timer_func)(unsigned long);

/*  creates a new timer.  */
static __inline__ void osi_init_timer(osi_timer_t *t,
				      osi_timer_func *fp, void *data)
{
}

/*  deletes a timer. stopping it if it is running.  */
static __inline__ void osi_del_timer( osi_timer_t *t)
{
}

/* sets a timer to go off in given seconds.
 * if executed on an existing running timer, resets it.
 */
static __inline__ void osi_set_timer( osi_timer_t * t, unsigned long sec)
{
}



/*  Schedule abstraction - Macro that makes a process temporarily
    give up control of the processor and lets other processes have
    a change to run. 
    Sleep abstraction - Sleep for x number of sections.  */

#define osi_schedule()
#define osi_sleep(x)
#define osi_sleep_unint(x)
#define osi_hibernate()


/*  Wait queue abstraction  */

typedef struct wait_queue * osi_wait_chan_t;

#define osi_wait_vars(x, y)
#define osi_init_wait_vars(x)
#define osi_init_wait_chan(x)
#define osi_asleep(x, y)

#define osi_await(x)

#define osi_unsleep(x, y)

#define osi_wakeup(x)



/*  Credentials structure  */

typedef void osi_cred_t;

#define OSI_MAY_READ 0
#define OSI_MAY_WRITE 0
#define OSI_MAY_EXEC 0

#define osi_pending_signals() 0


extern char *prog_name;
#define GFS_ENDIAN_BIG



#ifdef GFS_ENDIAN_BIG

#define gfs16_to_cpu osi_be16_to_cpu
#define gfs32_to_cpu osi_be32_to_cpu
#define gfs64_to_cpu osi_be64_to_cpu

#define cpu_to_gfs16 osi_cpu_to_be16
#define cpu_to_gfs32 osi_cpu_to_be32
#define cpu_to_gfs64 osi_cpu_to_be64

#else  /*  GFS_ENDIAN_BIG  */

#define gfs16_to_cpu osi_le16_to_cpu
#define gfs32_to_cpu osi_le32_to_cpu
#define gfs64_to_cpu osi_le64_to_cpu

#define cpu_to_gfs16 osi_cpu_to_le16
#define cpu_to_gfs32 osi_cpu_to_le32
#define cpu_to_gfs64 osi_cpu_to_le64

#endif  /*  GFS_ENDIAN_BIG  */


#define ERESTARTSYS EINTR


#endif  /*  __OSI_USER_DOT_H__  */



