/******************************************************************************
*******************************************************************************
**
**    Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland <teigland@sistina.com>
** 
**    Module name: moduleops.c
**
*******************************************************************************
******************************************************************************/

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif  /*  MODVERSIONS  */
#define __NO_VERSION__ 
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/file.h>
#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>

#include "memexp.h"
#include "ll_memexp.h"
#include "util.h"
#include "callbacks.h"
#include "cluster_info.h"
#include "client_state.h"
#include "mount_state.h"
#include "lock_state.h"
#include "daemon.h"
#include "dmep_tcp.h"
#include "dmep_scsi.h"
#include "dmep_debug.h"


memexp_t *list = NULL;



/**
 * memexp_mount
 *
 * @table_name: name of the cidev
 * @host_data: IP address
 * @cb: GFS callback function
 * @fsdata: opaque GFS handle
 * @jid: output jid
 * @first: output flag TRUE if this is the first mount
 *
 * Returns: handle to lock space
 */

lm_lockspace_t *memexp_mount(char *table_name, char *host_data,
			     lm_callback_t cb, lm_fsdata_t *fsdata,
			     unsigned int *jid, unsigned int *first)
{
  struct inode *cinode;
  struct file *cfile;
  struct dentry *cdentry;
  memexp_t *mep;
  char name[256], *port;
  uint32 ip;
  int error;


  MOD_INC_USE_COUNT;

  strcpy(name, table_name);

  inet_aton(host_data, &ip);



  /*  Check to make sure this lock table isn't already being used  */

  for (mep = list; mep; mep = mep->next)
    if (!strcmp(mep->table_name, name))
    {
      printk("%s:  \"%s\" is already in use\n", PROTO_NAME, name);
      goto fail;
    }



  /*  Set up our main structure  */

  mep = (memexp_t *)kmalloc(sizeof(memexp_t), GFP_KERNEL);
  if (!mep)
  {
    printk("%s:  out of memory\n", PROTO_NAME);
    goto fail;
  }
  memset(mep, 0, sizeof(memexp_t));

  strcpy(mep->table_name, name);

  mep->cb = cb;
  mep->fsdata = fsdata;



  /*  Get a pointer to the cidev and open it  */

  cdentry = lookup_dentry(name, NULL, 0);

  if (!cdentry || !cdentry->d_inode)
  {
    if (cdentry)
      dput(cdentry);

    printk("%s:  can't open cidev %s\n", PROTO_NAME, name);
    goto fail_free_mep;
  }

  cinode = cdentry->d_inode;
  cfile = get_empty_filp();
  cfile->f_op = blkdev_inode_operations.default_file_ops;

  if (!cfile->f_op->open || cfile->f_op->open(cinode, cfile))
  {
    put_filp(cfile);
    dput(cdentry);
    printk("%s:  can't open cidev %s\n", PROTO_NAME, name);
    goto fail_free_mep;
  }

  mep->cdentry = cdentry;
  mep->cfile = cfile;



  /*  Get information from cidev  */

  error = init_cluster_info(mep, ip);
  if (error)
  {
    printk("%s:  bad parameters from %s\n", PROTO_NAME, name);
    goto fail_dput;
  }
  show_cluster_info(mep);



  /*  Set up inter-machine callback socket  */
  
  if (open_cb_socket(mep) < 0)
    goto fail_rels_cluster;


   
  /*  Determine if lock source is device, server or debug */

  /* The debug lockdev_name string would look like, **
  ** 10.0.3.1:56789::/dev/pool/brassow              **
  ** So, we have 2 different lock devs              */
  if ((port = strstr(mep->lockdev_name, "::"))){
    mep->ll_memexp_ops = &ll_dmep_debug_ops;
  }else if ((port = strstr(mep->lockdev_name, ":"))){
    mep->ll_memexp_ops = &ll_dmep_tcp_ops;
  }else{
    mep->ll_memexp_ops = &ll_dmep_scsi_ops;
  }

  /*  Set up the dmep device or server  */

  error = mep->ll_memexp_ops->init(mep, mep->lockdev_name);
  if (error < 0)
  {
    printk("%s:  DMEP device or server init failed (%d)\n", PROTO_NAME, error);
    goto fail_closesock;
  }	



  /*  Set up LSM  */

  error = lsm_init(mep);
  if (error < 0)
  {
    printk("%s:  LSM init failed (%d)\n", PROTO_NAME, error);
    goto fail_rels_dmep;
  }



  /*  Wait for possible mount state recovery  */

  error = check_mount_state(mep, first);
  if (error < 0)
  {
    printk("%s:  mount lock check failed (%d)\n", PROTO_NAME, error);
    goto fail_lsm;
  }



  /*  Initialize client state  */

  error = client_state_init(mep);
  if (error < 0)
  {
    printk("%s:  client state init failed\n", PROTO_NAME);
    goto fail_lsm;
  }



  /*  Start lock callback thread  */

  osi_sema_init_lkd(&mep->cb_recvd_sem);

  error = kernel_thread(memexp_callback, mep, 0);
  if (error < 0)
  {
    printk("%s:  can't start callback thread (%d)\n", PROTO_NAME, error);
    goto fail_client_state;
  }

  down(&mep->cb_recvd_sem);
  osi_sema_init_lkd(&mep->cb_recvd_sem);



  /*  Start refresh thread  */

  osi_sema_init_lkd(&mep->lock_refresh_sem);

  error = kernel_thread(lock_refresh, mep, 0);
  if (error < 0)
  {
    printk("%s:  can't start refresh thread (%d)\n", PROTO_NAME, error);
    goto fail_callback;
  }



  /*  Start expiration monitor thread  */

  osi_sema_init_lkd(&mep->node_monitor_sem);

  error = kernel_thread(node_monitor, mep, 0);
  if (error < 0)
  {
    printk("%s:  can't start monitor thread (%d)\n", PROTO_NAME, error);
    goto fail_refresh;
  }



  /*  Success  */

  mep->next = list;
  list = mep;

  *jid = mep->jid;

  return((lm_lockspace_t *) mep);



 fail_refresh:
  mep->lock_refresh_stop = TRUE;
  wake_up_process(mep->lock_refresh_ts);
  down(&mep->lock_refresh_sem);

 fail_callback:
  mep->cb_recvd_stop = TRUE;   
  cid2ipaddr(mep, mep->cid, &ip);
  send_mess(mep, ip, 0, 0, 0, TRUE);
  down(&mep->cb_recvd_sem);

 fail_client_state:
  client_state_relse(mep);

 fail_lsm:
  lsm_shutdown(mep);

 fail_rels_dmep:
  mep->ll_memexp_ops->release(mep);

 fail_closesock:
  close_cb_socket(mep);

 fail_rels_cluster:
  release_cluster_info(mep);

 fail_dput:
  mep->cfile->f_op->release(mep->cdentry->d_inode, mep->cfile);
  dput(mep->cdentry);
  put_filp(mep->cfile);

 fail_free_mep:
  kfree(mep);

 fail:
  MOD_DEC_USE_COUNT;
  return(NULL);
}


/**
 * memexp_others_may_mount
 *
 * @lockspace: handle to specific lock space
 *
 * GFS calls this function if it was the first mounter after it's done
 * checking all the journals.
 *
 */

void memexp_others_may_mount(lm_lockspace_t *lockspace)
{
  memexp_t *mep = (memexp_t *)lockspace;

  others_may_mount(mep);
}



/**
 * memexp_umount
 *
 * @lockspace: handle to specific lock space
 *
 */

void memexp_unmount(lm_lockspace_t *lockspace)
{
  memexp_t *tmp, *last = NULL, *mep = (memexp_t *)lockspace;
  int done = FALSE;
  uint32 ip;

  if (list == mep)
  {
    list = list->next;
    done = TRUE;
  }
  else
    for (tmp = list; tmp; last = tmp, tmp = tmp->next)
      if (tmp == mep)
      {
	last->next = tmp->next;
	done = TRUE;
      }

  if (!done)
    printk("%s:  given a weird lockspace pointer\n", PROTO_NAME);
  else
  {
    mep->cb_recvd_stop = TRUE; 
    cid2ipaddr(mep, mep->cid, &ip);
    send_mess(mep, ip, 0, 0, 0, TRUE);
    down(&mep->cb_recvd_sem);
  
    mep->node_monitor_stop = TRUE;
    wake_up_process(mep->node_monitor_ts);
    down(&mep->node_monitor_sem);

    mep->lock_refresh_stop = TRUE;
    wake_up_process(mep->lock_refresh_ts);
    down(&mep->lock_refresh_sem);

    /* close and release stuff */

    client_state_relse(mep);

    lsm_shutdown(mep);

    mep->ll_memexp_ops->release(mep);

    release_cluster_info(mep);

    close_cb_socket(mep);

    mep->cfile->f_op->release(mep->cdentry->d_inode, mep->cfile);
    dput(mep->cdentry);
    put_filp(mep->cfile);

    kfree(mep);
  }

  MOD_DEC_USE_COUNT;
}







