/******************************************************************************
*******************************************************************************
**
**    Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland <teigland@sistina.com>
**
**    Module name: dmep_scsi.c
**
*******************************************************************************
******************************************************************************/

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif  /*  MODVERSIONS  */
#include <linux/kernel.h>
#include <asm/uaccess.h>
#include <linux/slab.h>

#include "osi.h"
#include "memexp.h"
#include "ll_memexp.h"
#include "pool_lock.h"
#include "pool_ondisk.h"
#include "pool_incore.h"


#define DMEP_SEGMENT_NUMBER 0


struct dmep_scsi_data
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0))
  struct nameidata nd;
#endif
  struct dentry *lockdev_dcache;
  struct inode *lockdev_inode;
};
typedef struct dmep_scsi_data dmep_scsi_data_t;


struct dump_info
{
  kdev_t dev;
  uint64 pbn;
};
typedef struct dump_info dump_info_t;


extern pool_t pools[];



/**
 * next_pool_dev - return the next lock device in the pool
 *
 * @pldev: the pool device
 * @in: the current lock device (NULL if first)
 * @out: the next lock device (after in)
 *
 * Returns: 0 if next device was found, -1 if no next device was found
 */

STATIC int next_pool_dev(kdev_t *pldev, kdev_t *in, kdev_t *out)
{
  pool_t *pp;
  subpool_t *sp;
  dmepdev_t *dev;
  int next = 0, i, j;

  pp = &pools[MINOR(*pldev)];
  
  for (i = 0; i < pp->p_num_spools; i++)
  {
    sp = &pp->p_spools[i];

    for (j = 0; j < sp->sp_dmepdevs; j++)
    {
      dev = &sp->sp_dmepdevlist[j];

      if (!in)
      {
	*out = dev->dm_dev;
	return(0);
      }

      if (next)
      {
	*out = dev->dm_dev;
	return(0);
      }

      if (dev->dm_dev == *in)
	next = TRUE;
    }
  }

  return(-1);
}



/**
 * fill_cdb - pack DMEP command parameters into a cdb structure
 *
 * @cdb: the cdb buffer
 * @lockname: contains number and type which form bid
 * @opcode: DMEP opcode for load or store
 * @action: which type of load or store
 * @segment: DMEP segment number
 * @len: the length
 *
 */

STATIC void fill_cdb(char *cdb, lm_lockname_t *lockname, uint8 opcode,
		     uint8 action, uint8 segment, uint32 len)
{
  uint64 number;
  number = lockname->ln_number;

  cdb[0] = opcode;
  cdb[1] = action & 0x1F;
  cdb[2] = segment;
  cdb[3] = lockname->ln_type & 0xFF;

  cdb[4] =  (number >> 56) & 0xFF;
  cdb[5] =  (number >> 48) & 0xFF;
  cdb[6] =  (number >> 40) & 0xFF;
  cdb[7] =  (number >> 32) & 0xFF;
  cdb[8] =  (number >> 24) & 0xFF;
  cdb[9] =  (number >> 16) & 0xFF;
  cdb[10] = (number >>  8) & 0xFF;
  cdb[11] = (number >>  0) & 0xFF;

  cdb[12] = (len >> 16) & 0xFF;
  cdb[13] = (len >>  8) & 0xFF;
  cdb[14] = (len >>  0) & 0xFF;

  cdb[15] = 0;
}



/**
 * fill_store - pack PDF for store command
 *
 * @cdb: the PDF buffer
 * @store: contains sequence number and pbn
 * @length:
 * @action:
 *
 */

STATIC void fill_store(char *out, buf_store_t *store, uint32 length, 
		       uint8 action)
{
  out[0] = (length >> 16) & 0xFF;
  out[1] = (length >>  8) & 0xFF;
  out[2] = (length >>  0) & 0xFF;

  out[3] = action & 0x1F;

  out[4] = (store->inuse ? 0x80 : 0x00);

  out[5] = 0;

  out[8]  = (store->seq >> 56) & 0xFF;
  out[9]  = (store->seq >> 48) & 0xFF;
  out[10] = (store->seq >> 40) & 0xFF;
  out[11] = (store->seq >> 32) & 0xFF;
  out[12] = (store->seq >> 24) & 0xFF;
  out[13] = (store->seq >> 16) & 0xFF;
  out[14] = (store->seq >>  8) & 0xFF;
  out[15] = (store->seq >>  0) & 0xFF;

  out[16] = (store->pbn >> 56) & 0xFF;
  out[17] = (store->pbn >> 48) & 0xFF;
  out[18] = (store->pbn >> 40) & 0xFF;
  out[19] = (store->pbn >> 32) & 0xFF;
  out[20] = (store->pbn >> 24) & 0xFF;
  out[21] = (store->pbn >> 16) & 0xFF;
  out[22] = (store->pbn >>  8) & 0xFF;
  out[23] = (store->pbn >>  0) & 0xFF;
}



/**
 * get_rdf_64 - extract 64 bit number from return data buffer
 *
 * @src: buffer
 * @val: number
 *
 */

STATIC void get_rdf_64(char *src, uint64 *val)
{
  uint64 number = 0;
  int i;

  for (i = 0; i < 8; i++)
    number += ((src[i] & 0xFF) << (8 * (7 - i)));

  *val = number;
}



/**
 * dmep_scsi_init
 *
 * @mep: the lock module "superblock"
 * @args: 
 *
 */

STATIC int dmep_scsi_init(memexp_t *mep, char *args)
{
  int			error = 0;
  dmep_scsi_data_t	*pdata;

  pdata = (dmep_scsi_data_t *) kmalloc(sizeof(dmep_scsi_data_t), GFP_BUFFER);

  if (!pdata)
    return(-ENOMEM);
    
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0))
  if (path_init(mep->lockdev_name, 0, &(pdata->nd)))
    error = path_walk(mep->lockdev_name, &(pdata->nd));

  if (!pdata->nd.dentry || !pdata->nd.dentry->d_inode ||
      IS_NODEV(pdata->nd.dentry->d_inode) || 
      !S_ISBLK(pdata->nd.dentry->d_inode->i_mode)){
    path_release(&pdata->nd);
    printk("%s:  can't open lock dev %s\n", PROTO_NAME, mep->lockdev_name);
    printk("pdata->nd.dentry                            :: %c\n",
	   (pdata->nd.dentry)? '+':'-');
    printk("pdata->nd.dentry->d_inode                   :: %c\n",
	   (pdata->nd.dentry)?
	   (pdata->nd.dentry->d_inode)? '+':'-' :
	   '-');
    printk("IS_NODEV(pdata->nd.dentry->d_inode)         :: %c\n",
	   (pdata->nd.dentry->d_inode)?
	   (IS_NODEV(pdata->nd.dentry->d_inode))? '+':'-' :
	   '-');
    printk("S_ISBLK(pdata->nd.dentry->d_inode->i_mode)  :: %c\n\n",
	   (pdata->nd.dentry->d_inode)?
	   (S_ISBLK(pdata->nd.dentry->d_inode->i_mode))? '+':'-' :
	   '-');
    kfree(pdata);
    return(-ENXIO);
  }

  pdata->lockdev_inode = pdata->nd.dentry->d_inode;
  error = blkdev_get(pdata->lockdev_inode->i_bdev, FMODE_READ,
		     O_RDONLY, BDEV_FILE);

  if (error)
  {
    path_release(&pdata->nd);
    printk("%s:  can't open lock dev %s\n",
	   PROTO_NAME, mep->lockdev_name);
    kfree(pdata);
    return(-EIO);
  }

  pdata->lockdev_dcache = pdata->nd.dentry;

#else
  pdata->lockdev_dcache = lookup_dentry(mep->lockdev_name, NULL, 0);


  if (!pdata->lockdev_dcache || !pdata->lockdev_dcache->d_inode)
  {
    if (pdata->lockdev_dcache)
      dput(pdata->lockdev_dcache);

    printk("%s:  can't open lock device %s\n", PROTO_NAME, mep->lockdev_name);
    kfree(pdata);
    return(-1);
  }
  
  pdata->lockdev_inode = pdata->lockdev_dcache->d_inode;
#endif

  mep->ll_memexp_data = pdata;

  return(error);
}



/**
 * dmep_scsi_release
 *
 * @mep: the lock module "superblock"
 *
 */

STATIC int dmep_scsi_release(memexp_t *mep)
{
  dmep_scsi_data_t *pdata = (dmep_scsi_data_t *) mep->ll_memexp_data;

  MEP_ASSERT(pdata, );

  dput(pdata->lockdev_dcache);
  
  kfree(pdata);

  mep->ll_memexp_data = NULL;

  return(0);
}



/**
 * dmep_scsi_load
 *
 * @mep: the lock module "superblock"
 * @lockname: holds lock number and type
 * @ret: the return values
 * @data: return data buffer copied here
 * @datalen: length of data buffer
 *
 */

STATIC int dmep_scsi_load(memexp_t *mep, lm_lockname_t *lockname, 
			  buf_load_t *ret, void *data, uint32 datalen)
{
  dmep_scsi_data_t *pdata = (dmep_scsi_data_t *) mep->ll_memexp_data;
  char cdb[16];
  char *retbuf;
  uint32 retlen;
  error_code_t ec;
  int error;

  MEP_ASSERT(pdata, );
  memset(ret, 0, sizeof(buf_load_t));
  memset(data, 0, datalen);


  /* intermediate buffer includes RDF and buffer */

  retlen = 24 + datalen;
  retbuf = (char *)kmalloc(retlen, GFP_BUFFER);
  memset(retbuf, 0, retlen);


  /* opcode: 0x85, action: 0, segment: n */

  fill_cdb(cdb, lockname, 0x85, 0, DMEP_SEGMENT_NUMBER, retlen);


  /* pool only uses the 8 byte lock number to map to a lock device */

  error = pool_memexp(pdata->lockdev_inode->i_rdev, LOGICAL_DEVICE, cdb,
		      lockname->ln_number, retbuf, retlen, &ec);
  if (error)
  {
    osi_printf("%s:  Error performing SCSI DMEP (load).\n", PROTO_NAME);
    osi_printf("\tSK   = %.2x\n"
	       "\tASC  = %.2x\n"
	       "\tASCQ = %.2x\n"
	       "\tSKS  = %.2x %.2x %.2x\n",
	       ec.sense_key, ec.asc, ec.ascq, ec.specific[0],
	       ec.specific[1], ec.specific[2]);
    if((ec.sense_key == 0xe) && (ec.asc == 0x26) && (ec.ascq == 0xe)){
      printk("Sequence number error (lck #%"FMT_64"u).\n",
	     lockname->ln_number);
      printk("This error should be imposible.\n"
	     "Contact your RAID vendor and report these lines.\n");
    }
    kfree(retbuf);
    return(error);
  }


  /* prepare info for returning */

  ret->inuse = retbuf[4] & 0x80;
  ret->full = retbuf[5];
  get_rdf_64(retbuf + 8, &ret->seq);
  get_rdf_64(retbuf + 16, &ret->pbn);
  memcpy(data, retbuf + 24, datalen);

  kfree(retbuf);

  return(0);
}



/**
 * dmep_scsi_store
 *
 * @mep: the lock module "superblock"
 * @lockname: holds lock number and type
 * @store: input store parameters, pbn and seq number
 * @data: buffer containing data to store
 * @datalen: length of data buffer
 *
 */

STATIC int dmep_scsi_store(memexp_t *mep, lm_lockname_t *lockname, 
			   buf_store_t *store, void *data, uint32 datalen)
{
  dmep_scsi_data_t *pdata = (dmep_scsi_data_t *) mep->ll_memexp_data;
  char cdb[16];
  char *outbuf;
  int outlen;
  error_code_t ec;
  int error = 0;

  MEP_ASSERT(pdata, );

  
  /* intermediate buffer holds PDF and data buffer */

  outlen = 24 + datalen;
  outbuf = (char *)kmalloc(outlen, GFP_BUFFER);  


  /* opcode: 0x89, action: 0, segment: n */

  fill_cdb(cdb, lockname, 0x89, 0, DMEP_SEGMENT_NUMBER, outlen);


  /* action: 0 */

  fill_store(outbuf, store, outlen, 0);


  /* opaque buffer is copied after the parameter data info */
  
  memcpy(outbuf + 24, data, datalen);


  /* pool only uses the 8 byte lock number to map to a lock device */

  error = pool_memexp(pdata->lockdev_inode->i_rdev, LOGICAL_DEVICE, cdb,
		      lockname->ln_number, outbuf, outlen, &ec);
  if (error)
  {
    osi_printf("%s:  Error performing SCSI DMEP (store).\n", PROTO_NAME);
    osi_printf("\tSK   = %.2x\n"
	       "\tASC  = %.2x\n"
	       "\tASCQ = %.2x\n"
	       "\tSKS  = %.2x %.2x %.2x\n",
	       ec.sense_key, ec.asc, ec.ascq, ec.specific[0],
	       ec.specific[1], ec.specific[2]);
    if((ec.sense_key == 0xe) && (ec.asc == 0x26) && (ec.ascq == 0xe)){
      printk("Sequence number error (lck #%"FMT_64"u).\n",
	     lockname->ln_number);
      printk("Sent 0x%.2x%.2x%.2x%.2x%.2x%.2x%.2x%.2x\n", 
	     outbuf[8],outbuf[9],outbuf[10],outbuf[11],
	     outbuf[12],outbuf[13],outbuf[14],outbuf[15]);
    }
  }
  
  kfree(outbuf);

  return(error);
}



/**
 * dmep_scsi_modepg - this is no longer really a mode page, but the name stuck
 *
 * @mep: the lock module "superblock"
 * @set: choose sense or select action
 * @numbufs: not used
 * @size: size of buffers
 *
 */

STATIC int dmep_scsi_modepg(memexp_t *mep, uint8 set, uint64 *numbufs, 
			    uint32 *size)
{

  dmep_scsi_data_t *pdata = (dmep_scsi_data_t *) mep->ll_memexp_data;
  char cdb[16];
  char *retbuf;
  uint32 retlen;
  error_code_t ec;
  int error, first = TRUE;
  lm_lockname_t lockname;
  kdev_t dev;
  uint32 bufsize;


  MEP_ASSERT(pdata, );

  *numbufs = 0;
  *size = 0;

  if (set){
    printk("dmep_scsi_modepg:: set != 0  FAILING\n");
    return(-1);
  }

  /* RDF will be returned in this buffer */

  retlen = 20;
  retbuf = (char *)kmalloc(retlen, GFP_BUFFER);


  /* this is ignored for sense config */

  lockname.ln_number = 0;
  lockname.ln_type = 0;


  /* opcode: 0x85, action: 2, segment: n */

  fill_cdb(cdb, &lockname, 0x85, 2, DMEP_SEGMENT_NUMBER, retlen);


  /* first real memexp device */

  error = next_pool_dev(&pdata->lockdev_inode->i_rdev, NULL, &dev);
  if (error < 0){
    printk("dmep_scsi_modepg:: Could not get first pool dev.  FAILING\n");
    goto fail;
  }

  /* sense on all real devices */

  while (1)
  {
    memset(retbuf, 0, retlen);

    error = pool_memexp(dev, REAL_DEVICE, cdb, 0, retbuf, retlen, &ec);
    if (error)
    {
      /* check values of ASC/Q */
      printk("dmep_scsi_modepg:: Could not complete DMEP call.  FAILING\n");
      goto fail;
    }

    bufsize = 0;
    bufsize = ((retbuf[16] & 0xFF) << 16) + 
              ((retbuf[17] & 0xFF) <<  8) +
              ((retbuf[18] & 0xFF) <<  0);

    if (!bufsize)
    {
      printk("%s:  Memory Export buffer size set to zero: %u,%u\n", PROTO_NAME,
	     MAJOR(dev), MINOR(dev));
      goto fail;
    }
    else
      printk("%s:  detected DMEP buffer size %u on device %u,%u\n",
	     PROTO_NAME, bufsize, MAJOR(dev), MINOR(dev));
 

    if (first)
    {
      *size = bufsize;
      first = FALSE;
    }
    else
      if (*size != bufsize)
      {
	printk("%s:  inconsistent Memory Export buffer sizes\n", PROTO_NAME);
	goto fail;
      }
    error = next_pool_dev(&pdata->lockdev_inode->i_rdev, &dev, &dev);
    if (error < 0)
    {
      kfree(retbuf);
      return(0);
    }
  }

 fail:

  kfree(retbuf);
  return(-1);
}



/**
 * dmep_scsi_dump
 *
 * @mep: the lock module "superblock"
 * @cookie: holds info this routine needs between calls
 * @heads: headers describing each buffer
 * @bufs: the buffers are packed in this space
 * @count: number of buffers which can be accepted
 * @size: the buffer size
 * @retcount: number of buffers actually returned
 *
 * Returns:  -EXXXX on error, 0 or 1 on success
 *            0: there are no more buffers to dump
 *            1: there are more buffers, call dump again
 */

STATIC int dmep_scsi_dump(memexp_t *mep, void **cookie, dump_status_t *heads,
		   uint8 *bufs, uint32 count, uint32 size, uint32 *retcount)
{
  dmep_scsi_data_t *pdata = (dmep_scsi_data_t *) mep->ll_memexp_data;
  char cdb[16];
  char *retbuf;
  uint32 retlen;
  dump_info_t *di;
  kdev_t dev, new;
  uint64 pbn = 0, pbn2 = 0;
  uint32 bytes, realcount;
  unsigned char *src;
  int error, i;
  lm_lockname_t lockname;
  unsigned int more;
  error_code_t ec;
  

  *retcount = 0;


  /* figure out which real scsi device and real pbn to start with */

  if (*cookie)
  {
    di = (dump_info_t *)(*cookie);
    dev = di->dev;
    pbn = di->pbn;
  }
  else
  {
    pbn = 0;
    error = next_pool_dev(&pdata->lockdev_inode->i_rdev, NULL, &dev);    
    if (error < 0)
      return(-1);

    di = (dump_info_t *)kmalloc(sizeof(dump_info_t), GFP_BUFFER);    
    if (!di)
      return(-ENOMEM);
    
    *cookie = di;
  }



  /* 8 byte header for dump command as a whole,
   * 28 byte header for each opaque buffer returned,
   * count opaque buffers returned,
   * size is the length of each opaque buffer */

  retlen = 8 + (28 * count) + (size * count);
  retbuf = (char *)kmalloc(retlen, GFP_BUFFER);
  memset(retbuf, 0, retlen);



  /* the type field is just "reserved" in this case */

  lockname.ln_number = pbn;
  lockname.ln_type = 0;



  /* action: 1, segment: n */

  fill_cdb(cdb, &lockname, 0x85, 1, DMEP_SEGMENT_NUMBER, retlen);


  
  /* the lock number field is unused with REAL_DEVICE */

  error = pool_memexp(dev, REAL_DEVICE, cdb, 0, retbuf, retlen, &ec);
  if (error)
  {
    /* check values of ASC/Q */
    osi_printf("%s:  Error performing SCSI DMEP.\n", PROTO_NAME);
    osi_printf("\tSK   = %.2x\n"
	       "\tASC  = %.2x\n"
	       "\tASCQ = %.2x\n"
	       "\tSKS  = %.2x %.2x %.2x\n",
	       ec.sense_key, ec.asc, ec.ascq, ec.specific[0],
	       ec.specific[1], ec.specific[2]);
    kfree(di);
    return(-1);
  }	



  /* pick off the 8 header bytes for the whole dump */

  bytes = 0;
  bytes = ((retbuf[0] & 0xFF) << 16) +
          ((retbuf[1] & 0xFF) <<  8) +
          ((retbuf[2] & 0xFF) <<  0);

  MEP_ASSERT((retbuf[3] & 0x1F) == 1, );

  more = (retbuf[4] & 0x80);
  


  /* actual number of returned buffers may differ from count */

  realcount = (bytes - 8) / (28 + size);

  if ((bytes - 8) % (28 + size))
    printk("%s:  warning - something is wrong with buffer byte counts\n", PROTO_NAME);
  

  
  /* split up the headers and the data	
     headers go into "heads" array and raw buffer contents go into "bufs" array */

  src = retbuf + 8;

  for (i = 0; i < realcount; i++)
  {
    src += 3;        /* reserved bytes */

    heads->lockname.ln_type = (unsigned int)(*src);
    get_rdf_64(src + 1, &heads->lockname.ln_number);
    
    src += 9;     /* bid (lock type + number) */

    get_rdf_64(src, &heads->seq);

    src += 8;     /* sequence number */

    get_rdf_64(src, &pbn2);

    src += 8;     /* pbn */

    memcpy(bufs, src, size);

    src += size;  /* opaque buffer */

    bufs += size;
    heads++;
  }



  /* return values */

  *retcount = realcount;

  
  if (!more)
  {
    error = next_pool_dev(&pdata->lockdev_inode->i_rdev, &dev, &new);

    if (error < 0)
    {
      kfree(di);
      return(0);
    }

    di->dev = new;
    di->pbn = 0;
  }
  else
    di->pbn = pbn2 + 1;


  return(1);
}



/**
 * dmep_scsi_stomith - not used
 *
 * @mep: the lock module "superblock"
 * @ipaddr:
 *
 */

STATIC int dmep_scsi_stomith(memexp_t *mep, uint32 ipaddr)
{
  return(0);
}



struct ll_memexp_ops ll_dmep_scsi_ops = 
{
 load: dmep_scsi_load,
 store: dmep_scsi_store,
 dump: dmep_scsi_dump,
 modepg: dmep_scsi_modepg,
 init: dmep_scsi_init,
 release: dmep_scsi_release,
 stomith: dmep_scsi_stomith
};
