/******************************************************************************
*******************************************************************************
**
**    Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland <teigland@sistina.com>
** 
**    Module name: ll_memexp.h  (Low Level defs)
**
*******************************************************************************
******************************************************************************/

#ifndef LL_MEMEXP_DOT_H
#define LL_MEMEXP_DOT_H


/*  Private lsm data  */

struct lsm_s
{
  uint8 segment; 
  uint32 bufsize;
  uint32 dump_length;
  uint32 dump_count;
};
typedef struct lsm_s lsm_t;


/*  Lock structure stored in DMEP buffer  */

struct lsm_lock_s
{
  uint8 state; 
  uint8 conv:4, lvb_perm:2, lvb_lock:2;
  uint16 expired_holders; 
  uint16 live_holders;
  uint16 lvb_holders;
  uint8 holders[MAX_CLIENTS / 8]; 
  uint8 lvb[MAX_CLIENTS / 8];
  uint8 expired[MAX_CLIENTS / 8]; 
  uint8 conversion[MAX_CLIENTS / 8]; 
};
typedef struct lsm_lock_s lsm_lock_t;


/*  Result of low level load  */

struct buf_load_s
{
  uint8 inuse;
  uint8 full;
  uint64 seq;
  uint64 pbn;
};
typedef struct buf_load_s buf_load_t;


/*  Input to low level store  */

struct buf_store_s
{
  uint8 inuse;
  uint64 seq;
  uint64 pbn;
};
typedef struct buf_store_s buf_store_t;


/*  Result of low level dump  */

struct dump_status_s
{
  lm_lockname_t lockname;
  uint64 seq;
  uint64 pbn;
};
typedef struct dump_status_s dump_status_t;


/*  Low level functions to DMEP device or server  */

struct ll_memexp_ops
{
  int (*load)(memexp_t *mep, lm_lockname_t *lockname, buf_load_t *ret, 
	      void *data, uint32 len);
  int (*store)(memexp_t *mep, lm_lockname_t *lockname, buf_store_t *store, 
	       void *data, uint32 len);
  int (*dump)(memexp_t *mep, void **cookie, dump_status_t *heads, uint8 *bufs,
	      uint32 cnt, uint32 sze, uint32 *ret);
  int (*modepg)(memexp_t *mep, uint8 set, uint64 *numbufs, uint32 *datasize);
  int (*init)(memexp_t *mep, char *args);
  int (*release)(memexp_t *mep);
  int (*stomith)(memexp_t *mep, uint32 ipaddr);
};
typedef struct ll_memexp_ops ll_memexp_ops_t;



/* A "store" operation requires the seq and pbn numbers from previous "load" */

#define prepare_store(bld, bst) \
{ \
  (bst)->inuse = 1; \
  (bst)->seq = (bld)->seq; \
  (bst)->pbn = (bld)->pbn; \
}



/* Routines to test, set and clear cid's in lock bitmaps */


static __inline__ int cid_in_list(unsigned char *list, uint16 cid)
{
  unsigned char byte = list[cid >> 3];
  unsigned char mask = (1 << (cid % 8));

  if (byte & mask)
    return(TRUE);

  return(FALSE);
}

#define test_holder(cid, lock) cid_in_list((lock)->holders, cid)
#define test_expired(cid, lock) cid_in_list((lock)->expired, cid)
#define test_conversion(cid, lock) cid_in_list((lock)->conversion, cid)
#define test_lvb(cid, lock) cid_in_list((lock)->lvb, cid)

static __inline__ int set_cid_list(unsigned char *list, uint16 cid)
{
  unsigned char *byte = &list[cid >> 3];
  unsigned char mask = (1 << (cid % 8));

  if (*byte & mask)
    return(FALSE);

  *byte |= mask;
  return(TRUE);
}

static __inline__ int clear_cid_list(unsigned char *list, uint16 cid)
{
  unsigned char *byte = &list[cid >> 3];
  unsigned char mask = (1 << (cid % 8));

  if (*byte & mask)
  {
    *byte &= ~mask;
    return(TRUE);
  }
  
  return(FALSE);
}

#define set_holder(cid, lock) \
{ \
  if (set_cid_list((lock)->holders, cid)) \
    (lock)->live_holders++; \
}

#define clear_holder(cid, lock) \
{ \
  if (clear_cid_list((lock)->holders, cid)) \
    (lock)->live_holders--; \
}

#define set_expired(cid, lock) \
{ \
  if (set_cid_list((lock)->expired, cid)) \
    (lock)->expired_holders++; \
}

#define clear_expired(cid, lock) \
{ \
  if (clear_cid_list((lock)->expired, cid)) \
    (lock)->expired_holders--; \
}

#define set_conversion(cid, lock) \
{ \
  set_cid_list((lock)->conversion, cid); \
  (lock)->conv = 1; \
}

#define clear_conversion(cid, lock) \
{ \
  clear_cid_list((lock)->conversion, cid); \
  (lock)->conv = 0; \
}

#define set_lvb(cid, lock) \
{ \
  if (set_cid_list((lock)->lvb, cid)) \
    (lock)->lvb_holders++; \
}

#define clear_lvb(cid, lock) \
{ \
  if (clear_cid_list((lock)->lvb, cid)) \
    (lock)->lvb_holders--; \
}


#endif  /*  LL_MEMEXP_DOT_H  */







