/******************************************************************************
*******************************************************************************
**
**    Copyright 2000-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: ops.c
**
*******************************************************************************
******************************************************************************/

#include "nolock.h"



#define HASH_SHIFT 8
#define HASH_SIZE (1 << HASH_SHIFT)
#define HASH_MASK (HASH_SIZE - 1)



struct nolock_space
{
  osi_spinlock_t      lock;
  osi_list_head_t     hash[HASH_SIZE];
};
typedef struct nolock_space nolock_space_t;

struct nolock_lock
{
  osi_list_head_t     list;
  unsigned int        hash;
  lm_lockname_t       name;
  osi_atomic_t        count;

  unsigned int        state;

  osi_atomic_t        lvb_count;
  lm_lvb_t            lvb;
  char                lvb_data[NOLOCK_LVB_SIZE];
  char                lvb_data_net[NOLOCK_LVB_SIZE];

  nolock_space_t      *ls;
};
typedef struct nolock_lock nolock_lock_t;





/**
 * atoi - convert a string to an integer
 * @c: the string to convert
 *
 * Returns: an integer
 */

STATIC int atoi(const char *c)
{
  int x = 0;

  while ('0' <= *c && *c <= '9')
  {
    x = x * 10 + (*c - '0');
    c++;
  }

  return x;
}


/**
 * nolock_mount - mount a nolock lockspace
 * @table_name: the name of the space to mount
 * @host_data: host specific data
 * @cb: the callback
 * @fsdata: argument to pass back with the callback
 * @jid: the journal to mount
 *
 * Returns: a lockspace on successful mount or NULL on failure
 */

STATIC lm_lockspace_t *nolock_mount(char *table_name, char *host_data,
				    lm_callback_t cb, lm_fsdata_t *fsdata,
				    unsigned int *jid, unsigned int *first)
{
  nolock_space_t *lsp;
  char *c;
  int i;


  nolock_mod_inc();


  lsp = osi_malloc(sizeof(nolock_space_t));
  if (!lsp)
  {
    nolock_mod_dec();
    return NULL;
  }

  osi_spinlock_init(&lsp->lock);
  for (i = 0; i < HASH_SIZE; i++)
    osi_list_head_init(&lsp->hash[i]);


  /*  If there is a "jid=" in the hostdata, return that jid.
      Otherwise, return zero.  */

  c = osi_strstr(host_data, "jid=");

  if (!c)
    *jid = 0;
  else
  {
    c += 4;
    *jid = atoi(c);
  }

  *first = TRUE;


  return (lm_lockspace_t *)lsp;
}


/**
 * nolock_others_may_mount - unmount a lock space
 * @lockspace: the lockspace to unmount
 *
 */

STATIC void nolock_others_may_mount(lm_lockspace_t *lockspace)
{
  return;
}


/**
 * clear_hash_bucket - clear a hash bucket
 * @head: the bucket
 *
 */

STATIC void clear_hash_bucket(osi_list_head_t *head)
{
  nolock_lock_t *lk;

  while (!osi_list_empty(head))
  {
    lk = osi_list_entry(head->next, nolock_lock_t, list);

    if (!(lk->lvb.lvb_flags & LM_LVB_PERM))
      osi_printf("nolock:  found floating lock: count=%d, ln_type=%u, ln_number=%"FMT_64"u, lvb_count=%d, lvb_flags=%d",
		 osi_atomic_read(&lk->count),
		 lk->name.ln_type, lk->name.ln_number,
		 osi_atomic_read(&lk->lvb_count), lk->lvb.lvb_flags);

    osi_list_del(&lk->list);
    osi_free(lk, sizeof(nolock_lock_t));
  }
}


/**
 * clear_lockspace - clear a lock space
 * lsp: the lock space
 *
 */

STATIC void clear_lockspace(nolock_space_t *lsp)
{
  unsigned int i;

  osi_spin_lock(&lsp->lock);

  for (i = 0; i < HASH_SIZE; i++)
    clear_hash_bucket(&lsp->hash[i]);

  osi_spin_unlock(&lsp->lock);
}


/**
 * nolock_unmount - unmount a lock space
 * @lockspace: the lockspace to unmount
 *
 */

STATIC void nolock_unmount(lm_lockspace_t *lockspace)
{
  nolock_space_t *lsp = (nolock_space_t *)lockspace;

  clear_lockspace(lsp);
  osi_free(lsp, sizeof(nolock_space_t));

  nolock_mod_dec();
}


/**
 * nolock_hash - hash a lockname
 * @ln: the lockname
 *
 */

STATIC __inline__ unsigned nolock_hash(lm_lockname_t *ln)
{
  unsigned hash = ln->ln_type;
  hash ^= (unsigned)ln->ln_number;
  hash ^= (unsigned)(ln->ln_number >> 32);
  hash ^= (unsigned)(ln->ln_number >> 16);
  return hash;
}


/**
 * __find_lock - Find a lock in a given bucket
 * @head: the head of the bucket
 * @hash: the lock name's hash
 * @name: the lock name
 *
 * Returns: the lock or NULL
 */

STATIC nolock_lock_t *__find_lock(osi_list_head_t *head, unsigned hash, lm_lockname_t *name)
{
  osi_list_head_t *tmp;
  nolock_lock_t *lk;

  tmp = head;
  for (;;) 
  {
    tmp = tmp->next;
    if (tmp == head)
      break;
    lk = osi_list_entry(tmp, nolock_lock_t, list);

    if (lk->hash != hash)
      continue;
    if (lm_name_equal(&lk->name, name))
      return lk;
  }

  return NULL;
}


/**
 * new_lock - Allocate a new lock
 * @lsp: the lock space
 * @hash: the lock name's hash
 * @name: the lock name
 *
 * Returns: the lock or NULL
 */

STATIC nolock_lock_t *new_lock(nolock_space_t *lsp, unsigned hash, lm_lockname_t *name)
{
  nolock_lock_t *lk;

  lk = osi_malloc(sizeof(nolock_lock_t));
  if (!lk)
    return NULL;

  osi_memset(lk, 0, sizeof(nolock_lock_t));

  lk->hash = hash;
  lk->name = *name;
  osi_atomic_set(&lk->count, 0);

  lk->state = LM_ST_UNLOCKED;

  osi_atomic_set(&lk->lvb_count, 0);
  lk->lvb.lvb_flags = 0;
  lk->lvb.lvb_length = NOLOCK_LVB_SIZE;
  lk->lvb.lvb_data = lk->lvb_data;

  lk->ls = lsp;

  return lk;
}


/**
 * nolock_get_lock - get a lm_lock_t given a descripton of the lock
 * @lockspace: the lockspace the lock lives in
 * @name: the name of the lock
 *
 * Returns: A lock structure or NULL on failure
 */

STATIC lm_lock_t *nolock_get_lock(lm_lockspace_t *lockspace, lm_lockname_t *name)
{
  nolock_space_t *lsp = (nolock_space_t *)lockspace;
  unsigned int hash = nolock_hash(name);
  osi_list_head_t *lh = &lsp->hash[hash & HASH_MASK];
  nolock_lock_t *lk, *nlk;


  osi_spin_lock(&lsp->lock);
  lk = __find_lock(lh, hash, name);
  if (lk) 
    osi_atomic_inc(&lk->count);
  osi_spin_unlock(&lsp->lock);

  if (lk)
    return lk;


  nlk = new_lock(lsp, hash, name);

  if (!nlk)
    return NULL;


  osi_spin_lock(&lsp->lock);
  lk = __find_lock(lh, hash, name);
  if (!lk) 
  {
    osi_list_add(&nlk->list, lh);
    lk = nlk;
    nlk = NULL;
  }
  atomic_inc(&lk->count);
  osi_spin_unlock(&lsp->lock);


  if (nlk)
    osi_free(nlk, sizeof(nolock_lock_t));

  return lk;
}


/**
 * nolock_put_lock - get rid of a lock structure
 * @lock: the lock to throw away
 *
 */

STATIC void nolock_put_lock(lm_lock_t *lock)
{
  nolock_lock_t *lk = (nolock_lock_t *)lock;
  nolock_space_t *ls = lk->ls;


  NOLOCK_ASSERT(ls, );  /*  Keep compiler from bitching on a UP system.  */
  NOLOCK_ASSERT(osi_atomic_read(&lk->count), );  
  NOLOCK_ASSERT(lk->state == LM_ST_UNLOCKED, );
  NOLOCK_ASSERT(!osi_atomic_read(&lk->lvb_count), );

  if (osi_atomic_dec_and_test(&lk->count)) 
  {
    osi_spin_lock(&ls->lock);
    if (!osi_atomic_read(&lk->count) && !(lk->lvb.lvb_flags & LM_LVB_PERM)) 
    {
      osi_list_del(&lk->list);
      osi_free(lk, sizeof(nolock_lock_t));
    }
    osi_spin_unlock(&ls->lock);
  }
}


/**
 * nolock_lock - acquire a lock
 * @lock: the lock to manipulate
 * @cur_state: the current state
 * @req_state: the requested state
 * @flags: modifier flags
 *
 * Returns: A bitmap of LM_OUT_* on success, -EXXX on failure
 */

STATIC int nolock_lock(lm_lock_t *lock, unsigned int cur_state, unsigned int req_state, unsigned int flags)
{
  nolock_lock_t *lk = (nolock_lock_t *)lock;

  NOLOCK_ASSERT(osi_atomic_read(&lk->count), );
  NOLOCK_ASSERT(lk->state == cur_state, );
  NOLOCK_ASSERT(req_state != LM_ST_UNLOCKED, );

  lk->state = req_state;

  if (osi_atomic_read(&lk->lvb_count))
  {
    if (cur_state == LM_ST_EXCLUSIVE)
      osi_memcpy(lk->lvb_data_net, lk->lvb_data, NOLOCK_LVB_SIZE);
    else
      osi_memcpy(lk->lvb_data, lk->lvb_data_net, NOLOCK_LVB_SIZE);  
  }
  
  return req_state;
}


/**
 * nolock_unlock - unlock a lock
 * @lock: the lock to manipulate
 * @cur_state: the current state
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int nolock_unlock(lm_lock_t *lock, unsigned int cur_state)
{
  nolock_lock_t *lk = (nolock_lock_t *)lock;

  NOLOCK_ASSERT(osi_atomic_read(&lk->count), );
  NOLOCK_ASSERT(lk->state == cur_state, );
  NOLOCK_ASSERT(cur_state != LM_ST_UNLOCKED, );

  lk->state = LM_ST_UNLOCKED;

  if (osi_atomic_read(&lk->lvb_count) &&
      cur_state == LM_ST_EXCLUSIVE)
    osi_memcpy(lk->lvb_data_net, lk->lvb_data, NOLOCK_LVB_SIZE);

  return 0;
}


/**
 * nolock_reset - Move the lock into an unlocked state (if it isn't already)
 * @lock: the lock to reset
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int nolock_reset(lm_lock_t *lock)
{
  NOLOCK_ASSERT(0, );
  return 0;
}


/**
 * nolock_cancel - cancel a request on a lock
 * @lock: the lock to cancel request for
 *
 */

STATIC void nolock_cancel(lm_lock_t *lock, int on)
{
  return;
}


/**
 * nolock_hold_lvb - hold on to a lock value block
 * @lock: the lock the LVB is associated with
 *
 * Returns: the LVB
 */

STATIC lm_lvb_t *nolock_hold_lvb(lm_lock_t *lock)
{
  nolock_lock_t *lk = (nolock_lock_t *)lock;

  NOLOCK_ASSERT(osi_atomic_read(&lk->count), );

  osi_atomic_inc(&lk->count);
  osi_atomic_inc(&lk->lvb_count);

  return &lk->lvb;
}


/**
 * nolock_unhold_lvb - release a LVB
 * @lock: the lock the LVB is associated with
 * @lvb: the lock value block
 *
 */

STATIC void nolock_unhold_lvb(lm_lock_t *lock, lm_lvb_t *lvb)
{
  nolock_lock_t *lk = (nolock_lock_t *)lock;

  NOLOCK_ASSERT(osi_atomic_read(&lk->count), );
  NOLOCK_ASSERT(osi_atomic_read(&lk->lvb_count), );

  if (osi_atomic_dec_and_test(&lk->lvb_count) && !(lk->lvb.lvb_flags & LM_LVB_PERM))
  {
    osi_memset(lk->lvb_data, 0, NOLOCK_LVB_SIZE);
    osi_memset(lk->lvb_data_net, 0, NOLOCK_LVB_SIZE);
  }

  osi_atomic_dec(&lk->count);
}


/**
 * nolock_sync_lvb - sync out the value of a lvb
 * @lock: the lock the LVB is associated with
 * @lvb: the lock value block
 *
 */

STATIC void nolock_sync_lvb(lm_lock_t *lock, lm_lvb_t *lvb)
{
  nolock_lock_t *lk = (nolock_lock_t *)lock;

  NOLOCK_ASSERT(osi_atomic_read(&lk->count), );
  NOLOCK_ASSERT(osi_atomic_read(&lk->lvb_count), );
  NOLOCK_ASSERT(lk->state == LM_ST_EXCLUSIVE, );

  osi_memcpy(lk->lvb_data_net, lk->lvb_data, NOLOCK_LVB_SIZE);  
}


/**
 * nolock_reset_exp - reset the expired locks for a given jid
 * @lockspace: the lockspace
 * @jid: the jid
 *
 */

STATIC void nolock_reset_exp(lm_lockspace_t *lockspace, unsigned int jid)
{
  return;
}





lm_lockops_t nolock_ops =
{
  proto_name:           "nolock",
  local_fs:             TRUE,
  mount:                nolock_mount,
  others_may_mount:     nolock_others_may_mount,
  unmount:              nolock_unmount,
  get_lock:             nolock_get_lock,
  put_lock:             nolock_put_lock,
  lock:                 nolock_lock,
  unlock:               nolock_unlock,
  reset:                nolock_reset,
  cancel:               nolock_cancel,
  hold_lvb:		nolock_hold_lvb,
  unhold_lvb:           nolock_unhold_lvb,
  sync_lvb:             nolock_sync_lvb,
  reset_exp:            nolock_reset_exp
};


