/******************************************************************************
*******************************************************************************
**
**    Copyright 2000-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: ops.c
**
*******************************************************************************
******************************************************************************/

#include "stats.h"



#define do_l(lsp) do { osi_spin_lock(&(lsp)->spin); } while (0)
#define do_u(lsp) do { osi_spin_unlock(&(lsp)->spin); } while (0)
#define do_inc(lsp, member) do { do_l((lsp)); (lsp)->member++; do_u((lsp)); } while (0)
#define do_dec(lsp, member) do { do_l((lsp)); (lsp)->member--; do_u((lsp)); } while (0)
#define do_print(lsp, member) do { osi_printf("%s = %"FMT_64"u\n", #member, (lsp)->member); } while (0)



struct stats_space
{
  lm_lockstruct_t   lower_ls;
  lm_callback_t     upper_callback;
  lm_fsdata_t       *upper_fsdata;

  osi_spinlock_t    spin;

  uint64            mount_time;
  uint64            unmount_time;

  uint64            calls_callback;
  uint64            calls_others_may_mount;
  uint64            calls_get_lock;
  uint64            calls_put_lock;
  uint64            calls_lock;
  uint64            calls_unlock;
  uint64            calls_reset;
  uint64            calls_cancel;
  uint64            calls_hold_lvb;
  uint64            calls_unhold_lvb;
  uint64            calls_sync_lvb;
  uint64            calls_reset_exp;

  uint64            current_locks;
  uint64            current_nondisk;
  uint64            current_inode;
  uint64            current_rgrp;
  uint64            current_meta;
  uint64            current_nopen;
  uint64            current_flock;
  uint64            current_plock;
  uint64            current_plock_head;

  uint64            max_locks;
  uint64            max_locks_nondisk;
  uint64            max_locks_inode;
  uint64            max_locks_rgrp;
  uint64            max_locks_meta;
  uint64            max_locks_nopen;
  uint64            max_locks_flock;
  uint64            max_locks_plock;
  uint64            max_locks_plock_head;

  uint64            current_lvbs;
  uint64            max_lvbs;

  uint64            lock_latency_count;
  uint64            lock_latency_total;
  uint64            lock_latency_min;
  uint64            lock_latency_max;
};
typedef struct stats_space stats_space_t;

struct stats_lock
{
  stats_space_t     *ls;
  lm_lockname_t     name;
  lm_lock_t         *lower_lock;
};
typedef struct stats_lock stats_lock_t;





/**
 * stats_print - print out the accumulated stats for a filesystem
 * @lsp: the lock space pointer
 *
 */

STATIC void stats_print(stats_space_t *lsp)
{
  osi_printf("\n\n\n");

  do_l(lsp);

  do_print(lsp, mount_time);
  do_print(lsp, unmount_time);

  do_print(lsp, calls_callback);
  do_print(lsp, calls_others_may_mount);
  do_print(lsp, calls_get_lock);
  do_print(lsp, calls_put_lock);
  do_print(lsp, calls_lock);
  do_print(lsp, calls_unlock);
  do_print(lsp, calls_reset);
  do_print(lsp, calls_cancel);
  do_print(lsp, calls_hold_lvb);
  do_print(lsp, calls_unhold_lvb);
  do_print(lsp, calls_sync_lvb);
  do_print(lsp, calls_reset_exp);

  do_print(lsp, current_locks);
  do_print(lsp, current_nondisk);
  do_print(lsp, current_inode);
  do_print(lsp, current_rgrp);
  do_print(lsp, current_meta);
  do_print(lsp, current_nopen);
  do_print(lsp, current_flock);
  do_print(lsp, current_plock);
  do_print(lsp, current_plock_head);

  do_print(lsp, max_locks);
  do_print(lsp, max_locks_nondisk);
  do_print(lsp, max_locks_inode);
  do_print(lsp, max_locks_rgrp);
  do_print(lsp, max_locks_meta);
  do_print(lsp, max_locks_nopen);
  do_print(lsp, max_locks_flock);
  do_print(lsp, max_locks_plock);
  do_print(lsp, max_locks_plock_head);

  do_print(lsp, max_lvbs);

  do_print(lsp, lock_latency_count);
  do_print(lsp, lock_latency_total);
  do_print(lsp, lock_latency_min);
  do_print(lsp, lock_latency_max);

  do_u(lsp);

  osi_printf("\n\n\n");
}


/**
 * stats_cb - the callback from the lock module to the lock client
 * @fsdata: opaque data from the lock client identifying the lock space
 * @type: the type of callback
 * @data: data determined by the type
 *
 */

STATIC void stats_cb(lm_fsdata_t *fsdata, unsigned int type, void *data)
{
  stats_space_t *lsp = (stats_space_t *)fsdata;

  do_inc(lsp, calls_callback);
  lsp->upper_callback(lsp->upper_fsdata, type, data);
}


/**
 * stats_mount - mount a stats lockspace
 * @table_name: the name of the space to mount
 * @host_data: host specific data
 * @cb: the callback
 * @fsdata: argument to pass back with the callback
 * @jid: the journal to mount
 *
 * Returns: a lockspace on successful mount or NULL on failure
 */

STATIC lm_lockspace_t *stats_mount(char *table_name, char *host_data,
				   lm_callback_t cb, lm_fsdata_t *fsdata,
				   unsigned int *jid, unsigned int *first)
{
  stats_space_t *lsp;
  char *lower_proto_name;
  unsigned int x;
  int error;


  stats_mod_inc();


  lsp = osi_malloc(sizeof(stats_space_t));
  if (!lsp)
  {
    stats_mod_dec();
    return NULL;
  }
  osi_memset(lsp, 0, sizeof(stats_space_t));

  
  for (x = 0; table_name[x] && table_name[x] != ':'; x++) ;
  if (!x || !table_name[x])
  {
    stats_mod_dec();
    return NULL;
  }

  lower_proto_name = osi_malloc((x + 1) * sizeof(char));
  if (!lower_proto_name)
  {
    stats_mod_dec();
    return NULL;
  }

  osi_memcpy(lower_proto_name, table_name, x);
  lower_proto_name[x] = 0;

  x++;


  lsp->upper_callback = cb;
  lsp->upper_fsdata = fsdata;

  error = lm_mount(lower_proto_name, table_name + x, host_data,
		   stats_cb, lsp,
		   &lsp->lower_ls);
  if (error)
  {
    stats_mod_dec();
    return NULL;
  }

  *jid = lsp->lower_ls.ls_jid;
  *first = lsp->lower_ls.ls_first;


  osi_free(lower_proto_name, x * sizeof(char));


  osi_spinlock_init(&lsp->spin);
  lsp->mount_time = osi_gettimeofday();
  lsp->lock_latency_min = (uint64)-1;


  return (lm_lockspace_t *)lsp;
}


/**
 * stats_others_may_mount - unmount a lock space
 * @lockspace: the lockspace to unmount
 *
 */

STATIC void stats_others_may_mount(lm_lockspace_t *lockspace)
{
  stats_space_t *lsp = (stats_space_t *)lockspace;

  do_inc(lsp, calls_others_may_mount);
  lsp->lower_ls.ls_ops->others_may_mount(lsp->lower_ls.ls_lockspace);
}


/**
 * stats_unmount - unmount a lock space
 * @lockspace: the lockspace to unmount
 *
 */

STATIC void stats_unmount(lm_lockspace_t *lockspace)
{
  stats_space_t *lsp = (stats_space_t *)lockspace;

  lsp->unmount_time = osi_gettimeofday();
  stats_print(lsp);

  lsp->lower_ls.ls_ops->unmount(lsp->lower_ls.ls_lockspace);
  osi_free(lsp, sizeof(stats_space_t));
  stats_mod_dec();
}


/**
 * stats_get_lock - get a lm_lock_t given a descripton of the lock
 * @lockspace: the lockspace the lock lives in
 * @name: the name of the lock
 *
 * Returns: A lock structure or NULL on failure
 */

STATIC lm_lock_t *stats_get_lock(lm_lockspace_t *lockspace, lm_lockname_t *name)
{
  stats_space_t *lsp = (stats_space_t *)lockspace;
  stats_lock_t *lk;

  do_inc(lsp, calls_get_lock);

  lk = osi_malloc(sizeof(stats_lock_t));
  if (!lk)
    return NULL;

  lk->ls = lsp;
  lk->name = *name;

  lk->lower_lock = lsp->lower_ls.ls_ops->get_lock(lsp->lower_ls.ls_lockspace, name);
  if (!lk->lower_lock)
  {
    osi_free(lk, sizeof(stats_lock_t));
    return NULL;
  }

  return lk;
}


/**
 * stats_put_lock - get rid of a lock structure
 * @lock: the lock to throw away
 *
 */

STATIC void stats_put_lock(lm_lock_t *lock)
{
  stats_lock_t *lk = (stats_lock_t *)lock;
  stats_space_t *lsp = lk->ls;

  do_inc(lsp, calls_put_lock);
  lsp->lower_ls.ls_ops->put_lock(lk->lower_lock);
  osi_free(lk, sizeof(stats_lock_t));
}


/**
 * do_latency - maintain lock latency stats
 * @lk: the lock
 * @latency: the latency to moosh into the stats
 *
 */

STATIC void do_latency(stats_lock_t *lk, uint64 latency)
{
  stats_space_t *lsp = lk->ls;

  lsp->lock_latency_count++;
  lsp->lock_latency_total += latency;

  if (lsp->lock_latency_min > latency)
    lsp->lock_latency_min = latency;
  if (lsp->lock_latency_max < latency)
    lsp->lock_latency_max = latency;

  if (latency > 500000)
    osi_printf("high lock latency on (%u, %"FMT_64"u):  %"FMT_64"u\n",
	       lk->name.ln_type, lk->name.ln_number,
	       latency);
}


/**
 * stats_lock - acquire a lock
 * @lock: the lock to manipulate
 * @cur_state: the current state
 * @req_state: the requested state
 * @flags: modifier flags
 *
 * Returns: A bitmap of LM_OUT_* on success, -EXXX on failure
 */

STATIC int stats_lock(lm_lock_t *lock, unsigned int cur_state, unsigned int req_state, unsigned int flags)
{
  stats_lock_t *lk = (stats_lock_t *)lock;
  stats_space_t *lsp = lk->ls;
  uint64 start, latency;
  int error;

  do_inc(lsp, calls_lock);


  start = osi_gettimeofday();
  error = lsp->lower_ls.ls_ops->lock(lk->lower_lock, cur_state, req_state, flags);
  latency = osi_gettimeofday() - start;


  if (error > 0)
  {
    do_l(lsp);
    do_latency(lk, latency);
    do_u(lsp);

    if (cur_state == LM_ST_UNLOCKED &&
	(error & LM_OUT_ST_MASK) != LM_ST_UNLOCKED)
    {
      do_l(lsp);

      lsp->current_locks++;
      switch (lk->name.ln_type)
      {
      case LM_TYPE_NONDISK:
	lsp->current_nondisk++;
	break;
      case LM_TYPE_INODE:
	lsp->current_inode++;
	break;
      case LM_TYPE_RGRP:
	lsp->current_rgrp++;
	break;
      case LM_TYPE_META:
	lsp->current_meta++;
	break;
      case LM_TYPE_NOPEN:
	lsp->current_nopen++;
	break;
      case LM_TYPE_FLOCK:
	lsp->current_flock++;
	break;
      case LM_TYPE_PLOCK:
	lsp->current_plock++;
	break;
      case LM_TYPE_PLOCK_HEAD:
	lsp->current_plock_head++;
	break;
      }

      if (lsp->max_locks < lsp->current_locks)
      {
	lsp->max_locks = lsp->current_locks;
	lsp->max_locks_nondisk = lsp->current_nondisk;
	lsp->max_locks_inode = lsp->current_inode;
	lsp->max_locks_rgrp = lsp->current_rgrp;
	lsp->max_locks_meta = lsp->current_meta;
	lsp->max_locks_nopen = lsp->current_nopen;
	lsp->max_locks_flock = lsp->current_flock;
	lsp->max_locks_plock = lsp->current_plock;
	lsp->max_locks_plock_head = lsp->current_plock_head;
      }

      do_u(lsp);
    }
  }

  
  return error;
}


/**
 * stats_unlock - unlock a lock
 * @lock: the lock to manipulate
 * @cur_state: the current state
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int stats_unlock(lm_lock_t *lock, unsigned int cur_state)
{
  stats_lock_t *lk = (stats_lock_t *)lock;
  stats_space_t *lsp = lk->ls;
  uint64 start, latency;
  int error;

  do_inc(lsp, calls_unlock);


  start = osi_gettimeofday();
  error = lsp->lower_ls.ls_ops->unlock(lk->lower_lock, cur_state);
  latency = osi_gettimeofday() - start;


  if (!error)
  {
    do_l(lsp);

    do_latency(lk, latency);

    lsp->current_locks--;
    switch (lk->name.ln_type)
    {
      case LM_TYPE_NONDISK:
	lsp->current_nondisk--;
	break;
      case LM_TYPE_INODE:
	lsp->current_inode--;
	break;
      case LM_TYPE_RGRP:
	lsp->current_rgrp--;
	break;
      case LM_TYPE_META:
	lsp->current_meta--;
	break;
      case LM_TYPE_NOPEN:
	lsp->current_nopen--;
	break;
      case LM_TYPE_FLOCK:
	lsp->current_flock--;
	break;
      case LM_TYPE_PLOCK:
	lsp->current_plock--;
	break;
      case LM_TYPE_PLOCK_HEAD:
	lsp->current_plock_head--;
	break;
    }

    do_u(lsp);
  }

  return error;
}


/**
 * stats_reset - Move the lock into an unlocked state (if it isn't already)
 * @lock: the lock to reset
 *
 * Returns: 0 on success, -EXXX on failure
 */

STATIC int stats_reset(lm_lock_t *lock)
{
  stats_lock_t *lk = (stats_lock_t *)lock;
  stats_space_t *lsp = lk->ls;
  int error;

  do_inc(lsp, calls_reset);

  error = lsp->lower_ls.ls_ops->reset(lk->lower_lock);

  return error;
}


/**
 * stats_cancel - cancel a request on a lock
 * @lock: the lock to cancel request for
 *
 */

STATIC void stats_cancel(lm_lock_t *lock, int on)
{
  stats_lock_t *lk = (stats_lock_t *)lock;
  stats_space_t *lsp = lk->ls;

  do_inc(lsp, calls_cancel);
  lsp->lower_ls.ls_ops->cancel(lk->lower_lock, on);
}


/**
 * stats_hold_lvb - hold on to a lock value block
 * @lock: the lock the LVB is associated with
 *
 * Returns: the LVB
 */

STATIC lm_lvb_t *stats_hold_lvb(lm_lock_t *lock)
{
  stats_lock_t *lk = (stats_lock_t *)lock;
  stats_space_t *lsp = lk->ls;
  lm_lvb_t *lvb;

  do_inc(lsp, calls_hold_lvb);

  lvb = lsp->lower_ls.ls_ops->hold_lvb(lk->lower_lock);

  if (lvb)
  {
    do_l(lsp);
    lsp->current_lvbs++;
    if (lsp->max_lvbs < lsp->current_lvbs)
      lsp->max_lvbs = lsp->current_lvbs;
    do_u(lsp);
  }

  return lvb;
}


/**
 * stats_unhold_lvb - release a LVB
 * @lock: the lock the LVB is associated with
 * @lvb: the lock value block
 *
 */

STATIC void stats_unhold_lvb(lm_lock_t *lock, lm_lvb_t *lvb)
{
  stats_lock_t *lk = (stats_lock_t *)lock;
  stats_space_t *lsp = lk->ls;

  do_inc(lsp, calls_unhold_lvb);
  lsp->lower_ls.ls_ops->unhold_lvb(lk->lower_lock, lvb);
  do_dec(lsp, current_lvbs);
}


/**
 * stats_sync_lvb - sync out the value of a lvb
 * @lock: the lock the LVB is associated with
 * @lvb: the lock value block
 *
 */

STATIC void stats_sync_lvb(lm_lock_t *lock, lm_lvb_t *lvb)
{
  stats_lock_t *lk = (stats_lock_t *)lock;
  stats_space_t *lsp = lk->ls;

  do_inc(lsp, calls_sync_lvb);
  lsp->lower_ls.ls_ops->sync_lvb(lk->lower_lock, lvb);
}


/**
 * stats_reset_exp - reset the expired locks for a given jid
 * @lockspace: the lockspace
 * @jid: the jid
 *
 */

STATIC void stats_reset_exp(lm_lockspace_t *lockspace, unsigned int jid)
{
  stats_space_t *lsp = (stats_space_t *)lockspace;

  do_inc(lsp, calls_reset_exp);
  lsp->lower_ls.ls_ops->reset_exp(lsp->lower_ls.ls_lockspace, jid);
}





lm_lockops_t stats_ops =
{
  proto_name:           "lockstats",
  mount:                stats_mount,
  others_may_mount:     stats_others_may_mount,
  unmount:              stats_unmount,
  get_lock:             stats_get_lock,
  put_lock:             stats_put_lock,
  lock:                 stats_lock,
  unlock:               stats_unlock,
  reset:                stats_reset,
  cancel:               stats_cancel,
  hold_lvb:		stats_hold_lvb,
  unhold_lvb:           stats_unhold_lvb,
  sync_lvb:             stats_sync_lvb,
  reset_exp:            stats_reset_exp
};


