/******************************************************************************
 *
 *    Copyright 2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file CONTRIBUTORS for a list of contributors.
 *
 *    This file is maintained by:
 *      Michael C Tilstra <conrad@sistina.com>
 *
 * This file holds all of the functions that do reads and writes to the
 * disk store.
 *
 * This was made in an attempt to seperate and hopefully modulize things a
 * bit.
 */

/* not all of these will be needed, trim the extras */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h> /* memset */
#include <errno.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mount.h> /* BLKGETSIZE */
#include <sys/uio.h> 
#include <fcntl.h>
#include <syslog.h>
#include <assert.h>
#include "memexp.h"
#include "disktypes.h"
#include "hashes.h"  /* need crc32 */

/* some prototypes that I will have to expand into some header file that
 * does not yet exit.
 */
void hash_add(MEtable_t *met, u32 elm); /* lazy programmer....*/
void clear_CHold(MEtable_t *met, int clid);
int create_Meltable( MEtable_t *met, unsigned long tsze,
      unsigned int hsze,
      unsigned int dsze, unsigned int elcnt,
      unsigned int maxcli);

#define bidcpy(a,b) memcpy((a),(b),BIDSIZE)

/**
 * load_MelDisk - load an existing disk store to use.
 * @name: < name of the disk store to use.
 * @met: <> pointer to a MEtable structure.
 *
 * This tried to load an existing disk store into memory to use.  If things
 * seem like it hasn't been used before. (ie bad magic) Then return reporting
 * that.
 *
 * returns: -1: Not yet initialized.  <-1: data messed up. 0: good to go.
 */
int load_MelDisk(char *name, MEtable_t *met)
{
   int fd;
   unsigned int i, j, k, max;
   off_t offset;
   diskheader_t dh;

   /* sanity */
   assert(HEADER_SIZE == BLK_size);

   if((fd = open(name, O_RDWR|O_SYNC)) <0) {
      log_err("Failed to open disk store: %m\n");
      return -3;
   }

   if( ioctl(fd, BLKFLSBUF, NULL) <0) {
      if(errno != 25 ) {
         /* appaerently this is the Inappropriate ioctl for device message*/
         log_err("Failed to flush all buffers on store: %m\n");
         return -3;
      }
   }

   if( lseek(fd, 0, SEEK_SET) <0 ) {
      log_err("Failed to seek to start of disk store: %m\n");
      return -1;
   }
   if (read(fd, &dh, sizeof(diskheader_t)) <0) {
      log_err("Disk head read failed: %m\n");
      return -1;
   }
   if( osi_be64_to_cpu(dh.magic) != DISKHEAD_MAGIC ) {
      log_err("Disk Header Magic failed.\n");
      return -1;
   }
   if( osi_be64_to_cpu(dh.version) != DISKHEAD_VERSION ) {
      log_err("Wrong Version number on Disk Store.\n");
      return -1;
   }

   if( create_Meltable( met, osi_be64_to_cpu( dh.total_data_field ),
            osi_be32_to_cpu( dh.num_hash_key ),
            osi_be32_to_cpu( dh.elem_data_size ),
            osi_be32_to_cpu( dh.num_max_elem ),
            osi_be32_to_cpu( dh.max_cli ) ) <0 )
      die("Failed to create memory tables.");

   met->devnode = fd;

   if(lseek(fd, HEADER_SIZE, SEEK_SET)<0) die("Failed to seek to bans: %m\n");
   max = osi_be32_to_cpu( dh.max_cli );
   /* Load the Client Table */
   for(i=0, j=0; i < CT_BLOCKS(max); i++) {
      char tmp[BLK_size];
      ct_elem_t *ct;

      if( read(fd, tmp, BLK_size) <0)
         die("Failed to read CT Block %u: %m\n", i);
      for(k=0; k < CTPB; j++, k++) {
         if( j >= max ) break;
         ct = (ct_elem_t*)(tmp + (k * sizeof(ct_elem_t)));
         if( ct->C != 'C' ) die("Client Table Magic failed on %u\n", j);
         met->IPs[j] = osi_be32_to_cpu( ct->IP );
         met->bans[j] = ct->flags;
      }
      if( j >= max ) break;
   }

   /* Copy over any Client Holds that might still exist into Element Store*/
   max = osi_be32_to_cpu( dh.max_cli );
   if( lseek(fd, HEADER_SIZE + (CT_BLOCKS(max) * BLK_size), SEEK_SET) <0)
      die("Failed to seek back to Client Hold: %m\n");

   for(i=0,k=0; i < CH_BLOCKS(max, met->datasize); i++) {
      char tmp[BLK_size];

      offset = HEADER_SIZE + /* skip header. */
               (CT_BLOCKS(max) * BLK_size) + /* skip Client Table */
               (i * BLK_size); /* skip block already read. */
      if(lseek(fd, offset, SEEK_SET) <0)
         die("Failed to seek to next (%u) Client Hold Block: %m\n", i);

      if( read(fd, tmp, BLK_size) <0)
         die("Failed to read CH group %u of %u: %m\n", i, 
               CH_BLOCKS(max, met->datasize));
      for(j=0; j< CHPB(met->datasize); k++,j++) {
         CHelem_t *ch = (CHelem_t*)(tmp + (j * CH_SIZE(met->datasize)));
         u32 pbn, crccheck;

         if( osi_be32_to_cpu(ch->magic) != CHOLDELEM_MAGIC)
            die("Magic for Client Hold %u is bad\n", k);
         crccheck = crc32(tmp+(j*CH_SIZE(met->datasize))+sizeof(CHelem_t),
                          met->datasize);
         pbn = osi_be32_to_cpu(ch->pbn);
         if( crccheck != osi_be32_to_cpu(ch->crc) ) {
            /* zero client hold and move on. */
            log_msg("CRC32 on Client Hold %u failed. Zeroing CH\n", k);
            clear_CHold(met, k);
         } else
         if( pbn != 0) {
            /* Not cleared, must copy into diskstore proper. */
            diskelem_t de;
            struct iovec iov[2];

            log_msg("Client Hold %u Needed to be copied into Element %u\n",
                  k, pbn);

            de.seq = ch->seq;
            bidcpy(de.bid, ch->bid );
            de.inUse = ch->inUse;
            de.crc = ch->crc;
            de.magic = osi_cpu_to_be32(DISKELEM_MAGIC);

            iov[0].iov_base = &de;
            iov[0].iov_len = sizeof(diskelem_t);
            iov[1].iov_base = tmp+(j*CH_SIZE(met->datasize))+sizeof(CHelem_t);
            iov[1].iov_len = met->datasize;

            offset = HEADER_SIZE + /* skip headers */
               (CT_BLOCKS(met->maxclients) * BLK_size) + /* skip Client Table */
               (CH_BLOCKS(met->maxclients, met->datasize) * BLK_size) +
                 /* skip CHolds */
               EL_OFFSET(pbn, met->datasize); /* jump to the element. */

            if(lseek(fd, offset, SEEK_SET) <0)
               die("Failed to lseek: %m");
            if( writev(fd, iov, 2) <0)
               die("Failed to write element: %m");
         }
      }
   }

   /* skip to begining of element Store. */
   offset = HEADER_SIZE +
            (CT_BLOCKS(max) * BLK_size) +
            (CH_BLOCKS(max, met->datasize) * BLK_size);
   if(lseek(fd, offset, SEEK_SET) <0)
      die("Failed to seek to start of Element Store: %m\n");

   /* load the elements. */
   max = osi_be32_to_cpu( dh.num_max_elem );
   j=0;
   for(i=0; i < DIV_RU(EL_BLOCKS(max,met->datasize),blkGroups); i++) {
      char grp[BLK_size*blkGroups];
      char *tmp = grp;
      int g;

      if( read(fd, grp, BLK_size*blkGroups) <0)
         die("Failed to read  group %u of %u: %m\n",i,
               DIV_RU(EL_BLOCKS(max,met->datasize),blkGroups));

      for(g=0; g < blkGroups; g++) {
         tmp = grp + (g *BLK_size);

         for(k=0; k < EPB(met->datasize); j++, k++) {
            u32 checkcrc;
            diskelem_t *de = (diskelem_t*)(tmp + (k * 
                  (met->datasize+sizeof(diskelem_t))));
            if( osi_be32_to_cpu(de->magic) != DISKELEM_MAGIC )
               die("Magic for disk element %d failed\n",j);
            met->elements[j].seq = osi_be64_to_cpu( de->seq );
            met->elements[j].InUse = de->inUse;
            bidcpy( met->elements[j].bid, de->bid );
            met->elements[j].Hnext = GROUNDED;
            met->elements[j].Hprev = GROUNDED;
            met->elements[j].MRUn = j+1;
            met->elements[j].MRUp = j-1;
            if( de->inUse ) met->used++;
#if 0 /* doing this is nice and cute, but causes confusing warnings */
            if( de->inUse )
#endif
            hash_add(met, j);
            memcpy( &met->datafield[ j * met->datasize ],
                    ((char *)de)+sizeof(diskelem_t), met->datasize);
            checkcrc = crc32(&met->datafield[j * met->datasize], met->datasize);
            if( checkcrc != osi_be32_to_cpu(de->crc) )
               die("CRC32 of diskelement %d failed\n", j);
         }
      }
   }

   met->MRUhead = 0;
   met->MRUtail = max -1;

   log_msg("Disk store restored to mem.\n");

   return 0;
}

/**
 * write_elem - writes an element to the diskstore
 * @met: <
 * @pbn: < pbn of element to store.
 *
 * This writes out an element and its datafield to the disk store.  The data
 * fields and elemheaders are interwoven on the disk store. (unlike memoery
 * where they are seperated.)
 *
 */
void write_elem(MEtable_t *met, u32 pbn)
{
   diskelem_t de;
   struct iovec iov[2];
   off_t offset;

   DBprintf("Entered write_elem.\n");

   de.seq = osi_cpu_to_be64( met->elements[pbn].seq );
   bidcpy(de.bid, met->elements[pbn].bid );
   de.inUse = met->elements[pbn].InUse;
   de.crc = osi_cpu_to_be32( crc32(&met->datafield[pbn*met->datasize], 
            met->datasize ));
   de.magic = osi_cpu_to_be32(DISKELEM_MAGIC);

   iov[0].iov_base = &de;
   iov[0].iov_len = sizeof(diskelem_t);
   iov[1].iov_base = &met->datafield[ pbn * met->datasize ];
   iov[1].iov_len = met->datasize;

   offset = HEADER_SIZE + /* skip headers */
      (CT_BLOCKS(met->maxclients) * BLK_size) + /* skip Client Table */
      (CH_BLOCKS(met->maxclients, met->datasize) * BLK_size) + /* skip CHolds */
      EL_OFFSET(pbn, met->datasize); /* jump to the element. */

   if(lseek(met->devnode, offset, SEEK_SET) <0)
      die("Failed to lseek: %m");
   if( writev(met->devnode, iov, 2) <0)
      die("Failed to write element: %m");

   DBprintf("Exiting write_elem cleanly.\n");
}

/**
 * write_CHold - Write the element to the Client Hold space.
 * @met: <
 * @pbn: < Element to put in CHold
 * @clid: < Client Ident
 * 
 */
void write_CHold(MEtable_t *met, u32 pbn, int clid)
{
   CHelem_t ch;
   struct iovec iov[2];
   off_t offset;

   DBprintf("Entered write_CHold\n");

   ch.pbn = osi_cpu_to_be32( pbn );
   ch.seq = osi_cpu_to_be64( met->elements[pbn].seq );
   bidcpy(ch.bid, met->elements[pbn].bid );
   ch.inUse = met->elements[pbn].InUse;
   ch.crc = osi_cpu_to_be32( crc32(&met->datafield[pbn*met->datasize], 
            met->datasize ));
   ch.magic = osi_cpu_to_be32(CHOLDELEM_MAGIC);

   iov[0].iov_base = &ch;
   iov[0].iov_len = sizeof(CHelem_t);
   iov[1].iov_base = &met->datafield[ pbn * met->datasize ];
   iov[1].iov_len = met->datasize;

   offset = HEADER_SIZE + /* skip headers */
      (CT_BLOCKS(met->maxclients) * BLK_size) + /* skip Client Table */
      CH_OFFSET(clid, met->datasize); /* jump to the element. */

   if(lseek(met->devnode, offset, SEEK_SET) <0)
      die("Failed to lseek: %m");
   if( writev(met->devnode, iov, 2) <0)
      die("Failed to write CHold: %m");

   DBprintf("Exiting write_CHold cleanly.\n");
}

/**
 * clear_CHold - Zeros out the CHold
 * @met: <
 * @clid: <
 * 
 * Since we've got to keep valid elements, this is a little more complex
 * than just writing out zeros.
 * 
 */
void clear_CHold(MEtable_t *met, int clid)
{
   u8 tmp[BLK_size]; /* disk elements are smaller than this. ALWAYS. */
   CHelem_t *de = (CHelem_t*)tmp;
   u8 *data = tmp + sizeof(CHelem_t);
   off_t offset;

   DBprintf("Entered clear_CHold\n");

   memset(tmp, 0, BLK_size);

   de->magic = osi_cpu_to_be32(CHOLDELEM_MAGIC);
   de->crc = osi_cpu_to_be32( crc32(data, met->datasize ));

   offset = HEADER_SIZE + /* skip headers */
      (CT_BLOCKS(met->maxclients) * BLK_size) + /* skip Client Table */
      CH_OFFSET(clid, met->datasize); /* jump to the element. */

   if(lseek(met->devnode, offset, SEEK_SET) <0)
      die("Failed to lseek: %m");
   if(write(met->devnode, tmp, (sizeof(CHelem_t)+met->datasize)) <0)
      die("Failed to clear CHold: %m\n");

   DBprintf("Exiting clear_CHold cleanly\n");
}

/**
 * write_ct - writes a Client Table entry to disk
 * @met: <
 * @ip: < ip that is being banned.
 * @flags: <
 * @i: < which spot to put it in.
 * 
 */
void write_ct(MEtable_t *met, u32 ip, u8 flags, int i)
{
   ct_elem_t ct;
   off_t offset;

   ct.IP = osi_cpu_to_be32( ip );
   ct.C = 'C';
   ct.flags = flags;
   offset = HEADER_SIZE +  /* get past header */
      CT_OFFSET(i);  /* skip to the correct Client Entry. */

   if(lseek(met->devnode, offset, SEEK_SET)<0)
      die("Failed to lseek ban offset: %m\n");
   if(write(met->devnode, &ct, sizeof(ct_elem_t))<0)
      die("Failed to write ban %u: %m\n", i);
}

