/******************************************************************************
 *
 *    Copyright 2000-2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file CONTRIBUTORS for a list of contributors.
 *
 *    This file is maintained by:
 *      Michael C Tilstra <conrad@sistina.com>
 */
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h> /* for memset */
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <poll.h>
#include <stropts.h>
#include <fcntl.h> /* for O_RDONLY */
#include <errno.h>
#include <limits.h> /* for OPEN_MAX */
#include <syslog.h>
#include "memexp.h"
#ifdef SOFT_RT
#include <sched.h>
#endif
#include "info_stream.h"

#ifndef MSG_WAITALL
#error "You headers are way too old.  Fix it."
/* If for some evil reason you cannot fix your header files, and you have a
 * 2.2.14 or better kernel, you can use this define below.  Other wise just
 * define it to 0.
 */
#define MSG_WAITALL       0x100   /* Wait for a full request */
#endif

static unsigned int cmdport = DEF_PORT;

/* globals. */
MEtable_t met;

/* fscking opensource. one day OPEN_MAX is defined, the next it isn't */
#ifdef OPEN_MAX
static int openmax = OPEN_MAX;
#else
static int openmax = 0;
#endif /* OPEN_MAX */

#define OM_GUESS 256

/**
 * open_max - clacs max number of open files.
 * Returns: the maximum number of file we can have open at a time.
 */
int open_max(void)
{
   if(openmax == 0) {
      errno =0;
      if((openmax = sysconf(_SC_OPEN_MAX)) < 0) {
         if( errno == 0) {
            openmax = OM_GUESS;
         }else{
            die("sysconf srror for _SC_OPEN_MAX");
         }
      }
   }
   return openmax;
}


/**
 * dump_buffer - debuggin junk.
 * @buf: < a buffer
 * @len: < size of buffer
 * 
 */
void dump_buffer(void *buf, int len)
{
   int i;
   u8 *c = buf;

   for(i=0; i<len; i++) {
      if( i % 16 == 0 ) printf("\n%p: ", &c[i]);
      printf("%2x ", c[i]);
   }
   printf("\n");
}

/*****************************************************************************/
/**
 * send_error - Send an error to the client.
 * @sk: < socket to write error packet
 * @key: < key value for packet matching
 * @ec: < Value of error packet
 * 
 * Returns: =0 Ok, <>0 error
 */
int send_error(int sk, u32 key, errorcode_t ec)
{
   char ob[sizeof(me_hdr_t)+sizeof(errorcode_t)];
   me_hdr_t *hdr = (me_hdr_t*)ob;
   errorcode_t *ecp = (errorcode_t*)(ob+sizeof(me_hdr_t));

   hdr->Opcode = Mel_Error;
   hdr->Key = key;
   hdr->length = cpu_to_be16(sizeof(errorcode_t));

   (*ecp) = ec;

   return send(sk, ob, sizeof(me_hdr_t)+sizeof(errorcode_t), 0);
}

#ifdef LPS_CODE
/*****************************************************************************/
/**
 * lps_inc - increment the lps counter
 * @lps: 
 * 
 * this should be called on every lock op.
 */
void lps_inc(lps_t *lps)
{
   lps->step ++;
}

/**
 * lsp_tick - update the lps struct
 * @lps: <>
 * 
 * This is called after every poll return.  If the alloted time has passed,
 * then figure out how many lock per second have passed in the last time
 * frame, and update the rest of the structure.
 */
void lps_tick(lps_t *lps)
{
   struct timeval tv;
   u64 curtime, rate;

   gettimeofday(&tv,NULL);
   curtime = ((uint64_t)tv.tv_sec * 1000000 + tv.tv_usec);

   if(curtime - 10000 >= lps->lasttime) {
      if(lps->step !=0) {
         rate = (curtime - lps->lasttime) / lps->step;

         lps->min = MIN(rate, lps->min);
         lps->max = MAX(rate, lps->max);
         lps->std += (rate * rate);

         lps->total += rate;

         lps->count ++;
         lps->step = 0;
      }
      lps->lasttime = curtime;
   }
}

/**
 * lps_init - initialize the lps structure.
 * @lps: <> pointer to the lps struct.
 * 
 */
void lps_init(lps_t *lps)
{
   struct timeval tv;
   gettimeofday(&tv,NULL);
   lps->step     = 0;
   lps->lasttime = ((uint64_t)tv.tv_sec * 1000000 + tv.tv_usec);
   lps->min      = ~0UL;
   lps->max      = 0;
   lps->std      = 0;
   lps->total    = 0;
   lps->count    = 0;
}
#define Declare_lps(lp) lps_t (lp)

/**
 * send_lps_stat - send the lps stats to a client
 * @sk: < socket
 * @lps: < lps stats
 * @b: <> buffer for writing.
 * @len: < size of that buffer.
 * 
 * 
 * Returns: int
 */
int send_lps_stat(int sk, lps_t *lps, char *b, size_t len)
{
   me_hdr_t *hdr = (me_hdr_t*)b;
   lps_t *lp = (lps_t*)(b + sizeof(me_hdr_t));

   if( len < sizeof(me_hdr_t)+sizeof(lps_t))
      return 0;

   lp->min = cpu_to_be64(lps->min);
   lp->max = cpu_to_be64(lps->max);
   lp->std = cpu_to_be64(lps->std);
   lp->total = cpu_to_be64(lps->total);
   lp->count = cpu_to_be64(lps->count);

   hdr->Opcode = Mel_Ok;
   hdr->length = cpu_to_be16(sizeof(lps_t));
   return send(sk, b, sizeof(me_hdr_t)+sizeof(lps_t),0);
}

#else /*LPS_CODE*/
#define lps_inc(lps)
#define lps_tick(lps)
#define lps_init(lps)
#define Declare_lps(lp)
int send_lps_stat(int sk, lps_t *lps, char *b, size_t len)
{
   me_hdr_t *hdr = (me_hdr_t*)b;
   return send_error(sk, hdr->Key, Mel_Err_NotImpl);
}
#endif /*LPS_CODE*/

/*****************************************************************************/
#ifdef LDUR_CODE
/**
 * ldur_init - initialize the ldur structure
 * @ld: > that structure
 */
void ldur_init(ldur_t *ld)
{
   ld->lasttime = 0;
   ld->min      = ~0UL;
   ld->max      = 0;
   ld->std      = 0;
   ld->total    = 0;
   ld->count    = 0;
}
#define Declare_ldur(ld) ldur_t (ld)

/**
 * ldur_start - start the timer value.
 * @ld: > 
 */
void ldur_start(ldur_t *ld)
{
   struct timeval tv;
   gettimeofday(&tv,NULL);
   ld->lasttime = ((uint64_t)tv.tv_sec * 1000000 + tv.tv_usec);
}

/**
 * ldur_stop - stop the timer
 * @ld: <>
 * 
 * Also updates all of the fields
 */
void ldur_stop(ldur_t *ld)
{
   u64 dur;
   struct timeval tv;
   gettimeofday(&tv,NULL);
   dur = ((uint64_t)tv.tv_sec * 1000000 + tv.tv_usec);
   if( dur > ld->lasttime)
      dur -= ld->lasttime;
   else
      dur = 0;

   ld->count++;
   ld->max = MAX(dur, ld->max);
   ld->min = MIN(dur, ld->min);
   ld->std += (dur * dur);
   ld->total += dur;
}

/**
 * send_ldur_stat - send the ldur stats to a client
 * @sk: < socket
 * @ldur: < ldur stats
 * @b: <> buffer for writing.
 * @len: < size of that buffer.
 * 
 * 
 * Returns: int
 */
int send_ldur_stat(int sk, ldur_t *ldur, char *b, size_t len)
{
   me_hdr_t *hdr = (me_hdr_t*)b;
   ldur_t *ld = (ldur_t*)(b + sizeof(me_hdr_t));

   if( len < sizeof(me_hdr_t)+sizeof(ldur_t))
      return 0;

   ld->min = cpu_to_be64(ldur->min);
   ld->max = cpu_to_be64(ldur->max);
   ld->std = cpu_to_be64(ldur->std);
   ld->count = cpu_to_be64(ldur->count);
   ld->total = cpu_to_be64(ldur->total);

   hdr->Opcode = Mel_Ok;
   hdr->length = cpu_to_be16(sizeof(ldur_t));
   return send(sk, b, sizeof(me_hdr_t)+sizeof(ldur_t),0);
}
#else /*LDUR_CODE*/
#define Declare_ldur(ld)
#define ldur_init(ldur)
#define ldur_start(ldur)
#define ldur_stop(ldur)
int send_ldur_stat(int sk, ldur_t *ldur, char *b, size_t len)
{
   me_hdr_t *hdr = (me_hdr_t*)b;
   return send_error(sk, hdr->Key, Mel_Err_NotImpl);
}
#endif /*LDUR_CODE*/

/**
 * send_cnt_stat - dunp counters to client
 * @sk: < socket
 * @cs: < counters.
 * @b: <> buffer for writing.
 * @len: < size of that buffer.
 * 
 * 
 * Returns: int
 */
int send_cnt_stat(int sk, counters_t *cs, char *b, size_t len)
{
   me_hdr_t *hdr = (me_hdr_t*)b;
   counters_t *cnts = (counters_t*)(b+sizeof(me_hdr_t));

   cnts->loads  = cpu_to_be32(cs->loads);
   cnts->stores = cpu_to_be32(cs->stores);
   cnts->sucstr = cpu_to_be32(cs->sucstr);
   cnts->dumps  = cpu_to_be32(cs->dumps);
   cnts->bans   = cpu_to_be32(cs->bans);
   cnts->unbans = cpu_to_be32(cs->unbans);
   cnts->errors = cpu_to_be32(cs->errors);
   cnts->verify = cpu_to_be32(cs->verify);

   hdr->Opcode = Mel_Ok;
   hdr->length = cpu_to_be16(sizeof(counters_t));
   return send(sk, b, sizeof(me_hdr_t)+sizeof(counters_t),0);
}

/*****************************************************************************/
/**
 * serv_listen - start listening to a port.
 * @port: < Port to bind to.
 * 
 * Returns: socket file descriptor
 */
int serv_listen(unsigned int port)
{
   int sk, trueint=TRUE;
   struct sockaddr_in adr;
   struct linger ling;

   if( (sk = socket(AF_INET, SOCK_STREAM, 0)) <0)
      die("Failed to create socket\n");

   if( setsockopt(sk, SOL_SOCKET, SO_REUSEADDR, &trueint, sizeof(int)) <0)
      die("Failed to set sock option SO_REUSEADDR\n");
   
   trueint=TRUE;/*jas*/
   if( setsockopt(sk, SOL_SOCKET, SO_KEEPALIVE, &trueint, sizeof(int)) <0)
      die("Failed to set sock option SO_KEEPALIVE\n");

   ling.l_onoff = TRUE;
   ling.l_linger = 4;
   if( setsockopt(sk, SOL_SOCKET, SO_LINGER, &ling, sizeof(struct linger)) <0)
      die("Failed to set sock option SO_LINGER\n");

   adr.sin_family = AF_INET;
   adr.sin_addr.s_addr = INADDR_ANY;
   adr.sin_port = htons( port );

   if( bind(sk, (struct sockaddr *)&adr, sizeof(struct sockaddr_in)) <0)
      die("Could not bind\n");

   if( listen(sk, 5) <0) die("Someone ripped off my ears\n");

   return sk;
}

/**
 * packet_size_check - just a quick bitch if things seem wrong
 * @want: < how many bytes we wanted.
 * @got: < how many we got.
 */
void __inline__ packet_size_check(size_t want, size_t got)
{
   if( want != got ) 
      log_alrt("oh cat puke. packet length should have been %d "
               "but we got %d.\n",
               want, got);
}


/**
 * loop - the server's heart.
 */
void loop(void)
{
   int i, maxi, listenSK, clisk, n;
   struct pollfd *polls;
   u32 *IP_to_FD, clIP; /* keep track of a socket's IP. */
   struct sockaddr_in adr;
   int olen;
   me_hdr_t *hdr;
   u8 *tbuf, *rest;
   unsigned int tbufsize; /*gets recalculated on Setmodepage (only grows)*/
   Declare_ldur(ldur);
   Declare_lps(lps);

   memset(&met.cnts, 0, sizeof(counters_t));

   tbufsize = sizeof(me_hdr_t) + 
      MAX( (sizeof(me_cdb_t)+sizeof(me_rdf_t)+met.datasize), SIZE_STATS);
   tbuf = malloc( tbufsize );
   if( ! tbuf ) die("failed to malloc tbuf");
   hdr = (me_hdr_t*)tbuf;
   rest = tbuf + sizeof(me_hdr_t);

   lps_init(&lps);
   ldur_init(&ldur);

   if( (polls = malloc( open_max() * sizeof(struct pollfd))) == NULL )
      die("failed to malloc pollfd array");
   memset(polls, 0, open_max() * sizeof(struct pollfd));
   for(i=0; i < open_max(); i++) polls[i].fd = -1;
   if( (IP_to_FD = malloc( open_max() * sizeof(u32))) == NULL )
      die("failed to malloc IP_to_FD array");
   memset(IP_to_FD, 0, open_max() * sizeof(u32));

   if( (listenSK = serv_listen(cmdport)) <0)
      die("Failed to listen to cmd port.");
   polls[0].fd = listenSK;
   polls[0].events = POLLIN;
   maxi=0;

   for( ;; ) {
      if( poll( polls, maxi +1, -1 ) <0)
         log_err("Poll error: %m");

      lps_tick(&lps);

      if( polls[0].revents & POLLIN ) {
         /* New client */
         /* Accept its connecction */
         n = sizeof(struct sockaddr_in);
         if( (clisk = accept(listenSK, (struct sockaddr*)&adr, &n)) <0) {
            log_err( "Accept failed: %m");
            continue;
         }
         /* TODO Authenticate it. still a nop....*/
         
         /* Add it to the poll tables. */
         for(i=1; polls[i].fd >=0 && i < open_max();i++);
         if(i >= open_max() ) {
            log_err("Maximum number of open file descriptors reached!!!!!\n");
            continue;
         }
         /* `i' should now be the next open polls entry. */
         polls[i].fd = clisk;
         polls[i].events = POLLIN;
         if(i>maxi) maxi = i;
         IP_to_FD[i] = adr.sin_addr.s_addr;

         /* TODO Need a better mapping method. */
         if( map_IP(&met, adr.sin_addr.s_addr) <0) {
            place_IP(&met, adr.sin_addr.s_addr);
         }
         /* */

         log_msg("New connection: fd %d from %x:%x\n", clisk,
               adr.sin_addr.s_addr, adr.sin_port);
      }

      for( i=1; i<= maxi; i++) {
         if( polls[i].fd < 0) continue;
         if( polls[i].revents & POLLIN ) {
            u16 packet_length=0;
            clisk = polls[i].fd;
            clIP  = IP_to_FD[i];

            n = recv(clisk, hdr, sizeof(me_hdr_t), MSG_WAITALL);
            if(n == 0) goto hangup;

            lps_inc(&lps);
            ldur_start(&ldur);

            /* get rest of packet. */
            packet_length = be16_to_cpu(hdr->length);
            n = recv(clisk, rest, packet_length, MSG_WAITALL);
            if(n == 0) goto hangup;

            /* pack Qelem */
            /* send to handler */
            
            /* react to it.*/
            if( hdr->Opcode == Mel_UnBanMe ) {
               /* mark unbanned. */
               packet_size_check(sizeof(u32), packet_length);

               drop_ban(&met, *((u32*)rest));
               met.cnts.unbans++;

               hdr->Opcode = Mel_Ok;
               hdr->length = 0;
               send(clisk, tbuf, sizeof(me_hdr_t),0);
            }else
            if( test_ban(&met, clIP) ) {
               /* is banned, return error. */
               n = send_error(clisk, hdr->Key, Mel_You_are_Banned );
            }else
            if( hdr->Opcode == Mel_BanReq ) {
               packet_size_check(sizeof(u32), packet_length);

               /* Doing test here since set always does a disk write.
                * Am trying to avoid extra disk activity.
                * */
               if( !test_ban(&met, *((u32*)rest)) ) 
                  set_ban(&met, *((u32*)rest));

               met.cnts.bans++;

               hdr->Opcode = Mel_Ok;
               hdr->length = 0;
               send(clisk, tbuf, sizeof(me_hdr_t),0);
            }else
            if( hdr->Opcode == Mel_StatReq ) {
               unsigned char Slct;
               packet_size_check(sizeof(u8), packet_length);
               Slct = *rest;
               switch(Slct) {
                  case STAT_HSHCNT:
                     {/* should pull this out.....*/
                        hshcnt_t *hshcnt = (hshcnt_t*)rest;
                        memset(hshcnt, 0, sizeof(hshcnt_t));
                        dump_hash_density(&met, hshcnt);

                        hdr->Opcode = Mel_Ok;
                        hdr->length = cpu_to_be16(sizeof(hshcnt_t));
                        n = send(clisk, tbuf,
                              sizeof(me_hdr_t)+sizeof(hshcnt_t),0);
                        if(n == 0) goto hangup;
                     }
                     break;
                  case STAT_counts:
                     n = send_cnt_stat(clisk, &met.cnts, tbuf, tbufsize);
                     if(n == 0) goto hangup;
                     break;
#ifdef LPS_CODE
                  case STAT_LPS:
                     n = send_lps_stat(clisk, &lps, tbuf, tbufsize);
                     if(n == 0) goto hangup;
                     break;
#endif
#ifdef LDUR_CODE
                  case STAT_LDUR:
                     n = send_ldur_stat(clisk, &ldur, tbuf, tbufsize);
                     if(n == 0) goto hangup;
                     break;
#endif
                  default:
                     n = send_error(clisk, hdr->Key, Mel_Err_NotImpl );
                     if(n == 0) goto hangup;
               }
            }else
            if( hdr->Opcode == Mel_Load ) {
               packet_size_check(sizeof(me_cdb_t), packet_length);

               olen = Melload( &met, tbuf, tbufsize);
               if( olen <= 0) goto hangup; /* need better response. */
               met.cnts.loads++;
               n = send(clisk, tbuf, olen, 0 );

            }else
            if( hdr->Opcode == Mel_Store ) {
               packet_size_check(sizeof(me_cdb_t)+sizeof(me_rdf_t)+met.datasize,
                     packet_length);

               olen = Melstore( &met, tbuf, tbufsize, clIP, clisk);
               met.cnts.stores++;
               if( olen < 0) goto hangup; /* need better response. */

            }else
            if( hdr->Opcode == Mel_Verify ) {
               packet_size_check(sizeof(u64), packet_length);

               olen = MelVerify(&met, tbuf, tbufsize, clIP, clisk);
               met.cnts.verify++;
               if(olen <=0) goto hangup;

            }else
            if( hdr->Opcode == Mel_Dump ) {
               packet_size_check(sizeof(dumpb_cdb_t), packet_length);

               n = Meldump( &met, hdr, (dumpb_cdb_t*)rest, clisk);
               met.cnts.dumps++;
               if( n <= 0 ) goto hangup; /* need better response. */

            }else
            if( hdr->Opcode == Mel_ModePgW ) {
               modepage_t *mp = (modepage_t *)rest;
               packet_size_check(sizeof(modepage_t), packet_length);

               /* First we need to wipe the current buffers. */
               Melreset( & met );

               if( met.devnode <0) {
                  /* don't change this if using a diskstore */
                  met.datasize = be32_to_cpu( mp->datasize );
                  met.Velemcnt = MIN( met.Relemcnt, be64_to_cpu( mp->numbufs ));
                  met.Velemcnt = MIN( met.Velemcnt,
                        (met.datatotalsize / met.datasize) );
               }
               mp->numbufs = cpu_to_be64( met.Velemcnt );
               mp->datasize = cpu_to_be32( met.datasize );

               log_msg("ModePage set: numbufs:%u datasize:%u\n",
                     met.Velemcnt,met.datasize);

               hdr->Opcode = Mel_Ok;
               n = send(clisk, tbuf, sizeof(me_hdr_t)+sizeof(modepage_t), 0);
               if(n == 0) goto hangup;

            }else
            if( hdr->Opcode == Mel_ModePgR ) {
               modepage_t *mp = (modepage_t *)rest;
               packet_size_check(sizeof(modepage_t), packet_length);

               mp->numbufs = cpu_to_be64( met.Velemcnt );
               mp->datasize = cpu_to_be32( met.datasize );

               hdr->Opcode = Mel_Ok;
               n = send(clisk, tbuf, sizeof(me_hdr_t)+sizeof(modepage_t), 0);
               if(n == 0) goto hangup;

            }else {
               log_err("Unknown header type: %d\n", hdr->Opcode);
            }

            ldur_stop(&ldur);

         }
         if( polls[i].revents & POLLHUP ) {
hangup:
            close( polls[i].fd );
            polls[i].fd = -1;
            IP_to_FD[i] = 0;
         }
      }/*for(i=1;i<=maxi;i++)*/
   }/*for(;;)*/
}

/**
 * init_systems - parse opts. build tables.
 * @argc: <
 * @argv: <
 *
 * Idea.  lets have a few master defaults, then follow other opts to
 * modify.
 */
void init_system(int argc, char *argv[] )
{
   int c;
   extern char *optarg;
   extern int optind;
   int nflg=FALSE;
   unsigned int eflg=DEF_ELSIZE;
   unsigned long Mflg=DEF_MEMSIZE;
   unsigned int cflg=DEF_CLIENTS;
   /* cute. the element header is now by default larger than the buffers
    * that they repersent.
    * oh well.
    */
   char *devname=NULL;
   char *tastr=NULL;

   while((c = getopt(argc, argv, "m:d:b:c:p:i:nh?"))!= -1){
      switch(c){
         case 'b': eflg = atoi(optarg); break;
         case 'm': Mflg = atoi(optarg); break;
         case 'c': cflg = atoi(optarg); break;
         case 'n': nflg = TRUE; break;
         case 'p': /* port change */
            cmdport = atoi(optarg);
            break;
         case 'd':
            devname = strdup(optarg);
            break;
         case 'i':
            tastr = strdup(optarg);
            break;
         case '?':
         case 'h':
         default:
            printf("Either:\n");
            printf(" memexpd -m <Memory Size> [-p port] [-b <ByteSize>]\n");
            printf("Or:\n");
            printf(" memexpd -d <datastore> [-p port] [-b <ByteSize>]\n");
            printf("\n");
            printf("  Memory Size  is how many megabytes to use.  It will\n");
            printf("    calculate the best fit for number of elements and\n");
            printf("    supporting structures.\n");
            printf("  datastore  is a device or file that was created with\n");
            printf("    initds.  DMEP state will be synced to this file to\n");
            printf("    some support of crash recovery.\n");
            printf("  port  is where this server listens for clients."
                   "Default: %u\n",cmdport);
            printf("  ByteSize  is the number of bytes each element can store."
                   " Default: %u\n",eflg);
            printf("\n");
            exit(1);
            break;
      }
   }

   if( cflg > (open_max()-2))
      printf("WARN! Operating system limits us to %d clients!\n",open_max()-2);
   cflg = MIN(cflg, (open_max()-2)); /* cannot have more than usable fds */

#ifndef DEBUG
   if(!nflg)
   {
      pid_t pid;
      int fd, i;
      if((pid=fork())<0){ die("Failed to fork off of parent: %m\n"); }
      else if(pid!=0) exit(0);

      for(i=open_max()-1; i>=0; --i) close(i); /*close everything */

      setsid();
      chdir("/");
      umask(0);
      fd = open("/dev/null", O_RDWR ); /*stdin*/
      dup(fd); /*stdout*/
      dup(fd); /*stderr*/

   }
#endif

#ifdef SOFT_RT
   /* Check to see if we're root && compiled with softrealtime. */
   if( getuid() == 0 ) {
      struct sched_param schprm;
      if( (schprm.sched_priority = sched_get_priority_max(SCHED_FIFO)) <0)
         die("Failed to find max priority for SCHED_FIFO: %m\n");
      if( sched_setscheduler(0, SCHED_FIFO, &schprm) <0)
         die("Failed to set scheduler to FIFO: %m\n");
   } else 
      die("Unable to use Soft Real Time features when not root!\n");
#endif /*SOFT_RT*/

   if(devname == NULL ) {
      /* We're not using a disk store. */
      if( create_MTab( &met, Mflg * MEGABYTE, eflg, cflg) <0 )
         die("Connot create MemExp Table");
   } else {
      if( load_MelDisk( devname, &met ) <0 ) {
         die("Failure to load disk image.");
      }
   }
   /* ok, we have a table loaded, which could have been just reloaded from
    * disk.
    */

   if(tastr) connect_info_stream(&met, tastr);
   else met.TAsock = -1;

   if( devname) free(devname);
   if( tastr) free(tastr);
}

/**
 * main - starts
 * @argc: <
 * @argv: <
 *
 * this is so complex i could gag.
 *
 * Returns: int
 */
int main(int argc, char *argv[])
{
   init_system(argc, argv);
#if defined(USE_SYSLOG) && !defined(DEBUG)
   openlog("memexpd", LOG_PID, LOG_DAEMON);
#endif
   loop();
   return 0;
}

/* vim:ai cin et sw=3 ts=3
 */
