/******************************************************************************
*******************************************************************************
**
**    Copyright 2000-2001 Sistina Software Inc
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Jonathan Brassow <brassow@sistina.com>
** 
**    Module name: lock_linux.c
**
**    Description:
**	Linux specific device lock routines.
**
*******************************************************************************
******************************************************************************/

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif  /*  MODVERSIONS  */
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/interrupt.h>

#define MAJOR_NR SCSI_DISK0_MAJOR
#include <linux/blk.h>
#include <linux/delay.h>

#include <asm/system.h>
#include <asm/io.h>

#include "scsi.h"
#include "sd.h"
#include "pool.h"
#include "pool_lock.h"

extern Scsi_Disk *rscsi_disks;
extern pool_debug_flag_t df_flags;

static void do_sd_request (void){
  osi_printf("I hate scsi compiler warnings\n");
}

int I_HATE_SCSI_COMPILER_WARINGS(void){
  do_sd();
  end_scsi_request(NULL, 0, 0);
  do_sd_request();
  return 0;
}

/**
 * pool_notify - a callback notifying pool when SCSI request is complete
 * @SCpnt
 *
 * This routine serves as an interrupt handler for the requests.
 */
static void pool_notify(Scsi_Cmnd *SCpnt)
{

  ENTER_PRINT(LOCKING_1, "pool_notify()");

  SCpnt->request.rq_status = RQ_SCSI_DONE;

  if (SCpnt->request.sem != NULL)
    up(SCpnt->request.sem);

  EXIT_PRINT(LOCKING_1, "pool_notify()");
  return;
}


/**
 * pool_to_scsi_mid - used to inject SCSI commands into SCSI midlayer
 * @dev: device to send SCSI command to
 * @cdb: Control Data Block
 * @buffer: PDF (Parameter Data Format)
 * @buffer_len: length of the PDF
 * @ec: structure that contains any SCSI error reported from device
 *
 * Returns: Linux errno
 */
int pool_to_scsi_mid(kdev_t dev, u_char *cdb, u_char *buffer,
		     int buffer_len, error_code_t *ec){
  Scsi_Device           *device;
  unsigned long 	flags=0;
  Scsi_Cmnd		*SCpnt;
  int 			wait = 1;
  struct semaphore  	sem;
  int                   error = 0;
  int                  	i;
  u_char		*tmp=NULL;
  int			tablesize;
  struct scatterlist    *sg_list=NULL;
  int			use_sg, len;

  ENTER_PRINT(LOCKING_1, "pool_to_scsi_mid()");

  /* hack - rscsi_disks exported from sd */
  device = rscsi_disks[DEVICE_NR(dev)].device;

  /* For scatter/gather we need to know how much room **
  ** the driver/HBA has to store request data         */

  /***  FIX:: some crappy drivers don't support S/G at all ! ***/

  tablesize = device->host->sg_tablesize;
  buffer_len = (buffer_len > tablesize)? tablesize: buffer_len;
  if(buffer_len/PAGE_SIZE){
    use_sg = (buffer_len%PAGE_SIZE) ? (buffer_len/PAGE_SIZE)+1:
      (buffer_len/PAGE_SIZE);

    /* does this space need to be obtained via scsi_malloc ?? */
    sg_list = (struct scatterlist *)
      osi_malloc(use_sg*sizeof(struct scatterlist));

    len = (use_sg * sizeof(struct scatterlist) + 511) & ~511;
    for(i = 0; len > PAGE_SIZE; len -= PAGE_SIZE, i++){
      if(!(sg_list[i].address = scsi_malloc(len))){
	spin_unlock_irqrestore(&io_request_lock, flags);
	for(i = 0; --i >= 0;){
	  scsi_free(sg_list[i].address, sg_list[i].length);
	}
	return(-ENOMEM);
      }
      sg_list[i].length = PAGE_SIZE;
      memcpy(sg_list[i].address, buffer+(i*PAGE_SIZE), PAGE_SIZE);
    }
    if(len){
      if(!(sg_list[i].address = scsi_malloc(len))){
	spin_unlock_irqrestore(&io_request_lock, flags);
	for(i = 0; --i >= 0;){
	  scsi_free(sg_list[i].address, sg_list[i].length);
	}
	return(-ENOMEM);
      }
	
      sg_list[i].length = len;
      memcpy(sg_list[i].address, buffer+(i*PAGE_SIZE), len);
    }
  } else {
    use_sg = 0;
    len = (buffer_len+511) & ~511;
    if (!(tmp = (u_char *)scsi_malloc(len))){
      spin_unlock_irqrestore(&io_request_lock, flags);
      return(-ENOMEM);
    }

    /* Note: we copy buffer to tmp incase of write */
    memcpy(tmp, buffer, buffer_len);

  }

  /* Now you need to make sure that the correct ptr (tmp/sg_list) **
  ** is being used when the scsi_do_cmd() is made */

  DEBUG_PRINT(LOCKING_2, "Scatter/Gather info::\n\tuse_sg: %d\n\tlen: %d\n",
	      use_sg, len);

  spin_lock_irqsave(&io_request_lock, flags);

  if (!(SCpnt = scsi_allocate_device(NULL, device, wait)))
  {
    scsi_free(tmp, len);
    spin_unlock_irqrestore(&io_request_lock, flags);
    return(-EIO);
  }

  SCpnt->tag = SC_TAG_HEAD;  /* overwritten by HBA driver? */

  sem = MUTEX_LOCKED;
  SCpnt->request.sem = &sem;
  SCpnt->request.rq_status = RQ_SCSI_BUSY;
  SCpnt->use_sg = use_sg;
  /*
#ifdef DEBUG
  {
    int i;
    osi_printf("DMEP CDB ::\n\t");
    for(i=0; i < 16; i++){
      osi_printf("%.2x ", cdb[i]);
    } osi_printf("\n\n");

    osi_printf("OUTGOING PDF ::");
    if(cdb[0] == 0x89){
    if(tmp){
      for(i=0; i < buffer_len; i++){
	if(!(i%16)){
	  osi_printf("\n\t");
	}
	osi_printf("%.2x ", tmp[i]);
      }
      osi_printf("\n");
    }}else{
      osi_printf("\n\tN/A\n");
    }
  }
#endif
  */
  scsi_do_cmd( SCpnt,           /* Scsi_Cmnd pointer */
               cdb,             /* CDB */
               /* buffer for returned lock data */
	       (use_sg)?(u_char *)sg_list:tmp,
	       buffer_len,      /* buff length (what for s/g ??) */
	       pool_notify, 	/* scsi cmd callback function */
	       DL_TIMEOUT,      /* timeout value */
	       DL_MAXRETRY);	/* number of retried attempted */

  spin_unlock_irqrestore(&io_request_lock, flags);
  down(&sem);
  spin_lock_irqsave(&io_request_lock, flags);

  SCpnt->request.sem = NULL;

  if (SCpnt->result != 0)
  {
    unsigned char *buf = SCpnt->sense_buffer;
    if(ec){
      ec->sense_key = buf[2] & 0x0F;
      ec->asc = buf[12];
      ec->ascq = buf[13];
      ec->specific[0] = buf[15];
      ec->specific[1] = 0x0;
      ec->specific[2] = 0x0;
    }
    scsi_release_command(SCpnt);
    scsi_free(tmp, len);
    spin_unlock_irqrestore(&io_request_lock, flags);

    EXIT_PRINT(LOCKING_1, "pool_to_scsi_mid()");
    return(-EIO);
  }
  /*
#ifdef DEBUG
  {
    int i;
    osi_printf("INCOMING PDF ::");
    if(cdb[0] == 0x85){
      if(tmp){
	for(i=0; i < buffer_len; i++){
	  if(!(i%16)){
	    osi_printf("\n\t");
	  }
	  osi_printf("%.2x ", tmp[i]);
	}
	osi_printf("\n");
      }
    }else {
      osi_printf("\n\tN/A\n");
    }
  }
#endif
  */
  /* Note: here we copy back in case it was a read */
  /* Hopefully, this solves both read and write issues */
  if(use_sg){
    for(i = 0; i < use_sg; i++){
      memcpy(buffer+(i*PAGE_SIZE),
	     sg_list[i].address, sg_list[i].length);
      scsi_free(sg_list[i].address, sg_list[i].length);
    }
    osi_free(sg_list, use_sg*sizeof(struct scatterlist));
  }else{
    memcpy(buffer, tmp, buffer_len);
    scsi_free(tmp, len);
  }
  wake_up(&SCpnt->device->device_wait);
  scsi_release_command(SCpnt);

  spin_unlock_irqrestore(&io_request_lock, flags);

  EXIT_PRINT(LOCKING_1, "pool_to_scsi_mid()");
  return(error);
}


