/******************************************************************************
*******************************************************************************
**
**    Copyright 2000-2001 Sistina Software Inc
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Jonathan Brassow <brassow@sistina.com>
** 
**    Module name: lock_linux.c
**
**    Description:
**	Linux specific device lock routines.
**
*******************************************************************************
******************************************************************************/

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif  /*  MODVERSIONS  */
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#define MAJOR_NR SCSI_DISK0_MAJOR
#include <linux/blk.h>
#include <linux/delay.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/semaphore.h>

#include <scsi/scsi.h>
#include <sd.h>
#include <scsi/sg.h>
#include <scsi/scsi_ioctl.h>

#include "pool.h"
#include "pool_lock.h"

extern Scsi_Disk *rscsi_disks;
extern pool_debug_flag_t df_flags;

/*
 * pool_to_scsi_mid - used to inject SCSI commands into SCSI midlayer
 * @dev: device to send SCSI command to
 * @cdb: Control Data Block
 * @buffer: PDF (Parameter Data Format)
 * @buffer_len: length of the PDF
 * @ec: structure that contains any SCSI error reported from device
 *
 * Returns: Linux errno
 */
int pool_to_scsi_mid(kdev_t dev, u_char *cdb, u_char *buffer,
		     int buffer_len, error_code_t *ec){
  Scsi_Device           *sdev;
  Scsi_Request		*scsi_req_ptr;
  unsigned char		*scsi_buffer;
  int			scsi_buf_size;
  int			direction;
  int			rtn;

  ENTER_PRINT(LOCKING_1, "pool_to_scsi_mid()");

  sdev = rscsi_disks[DEVICE_NR(dev)].device;
  scsi_buf_size = (buffer_len + 511) & ~511;
  if(scsi_buf_size > PAGE_SIZE){ scsi_buf_size = PAGE_SIZE; }
  scsi_buffer = (char *)scsi_malloc(scsi_buf_size);
  if(!scsi_buffer) { return -ENOMEM; }
  memset(scsi_buffer, 0, scsi_buf_size);

  if(cdb[0] == 0x85){
    direction = SCSI_DATA_READ;
  } else if(cdb[0] == 0x89){
    direction = SCSI_DATA_WRITE;
    memcpy(scsi_buffer, buffer, buffer_len);
  } else {
    return -EINVAL;
  }

  /* Set up correct LUN ? */

  scsi_req_ptr = scsi_allocate_request(sdev);
  if(!scsi_req_ptr) { return -EINTR; }
  scsi_req_ptr->sr_data_direction = direction;
  scsi_wait_req(scsi_req_ptr, cdb, scsi_buffer, scsi_buf_size, 
		10*HZ,	/* Normal timeout */
		5);	/* Normal retries */

  if(scsi_req_ptr->sr_result){
    unsigned char *buf = scsi_req_ptr->sr_sense_buffer;
    if(ec){
      ec->sense_key = buf[2] & 0x0F;
      ec->asc = buf[12];
      ec->ascq = buf[13];
      ec->specific[0] = buf[15];
      ec->specific[1] = 0x0;
      ec->specific[2] = 0x0;
    }
    /* Fill error structure */
    scsi_release_request(scsi_req_ptr);
    if(scsi_buffer){ scsi_free(scsi_buffer, scsi_buf_size); }

    EXIT_PRINT(LOCKING_1, "pool_to_scsi_mid()");
    return -EFAULT;
  }

  if(direction == SCSI_DATA_READ){
    memcpy(buffer, scsi_buffer, buffer_len);
  }
  rtn = scsi_req_ptr->sr_result;
  scsi_release_request(scsi_req_ptr);
  if(scsi_buffer){ scsi_free(scsi_buffer, scsi_buf_size); }

  EXIT_PRINT(LOCKING_1, "pool_to_scsi_mid()");
  return rtn;
}
