/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portions Copyright 2000-2001 Sistina Software Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Jonathan Brassow <brassow@sistina.com>
** 
**    Module name: pool_linux.c
**
**    Description:
**	Linux-specific pool driver routines.
**
*******************************************************************************
******************************************************************************/

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif  /*  MODVERSIONS  */
#define __NO_VERSION__ 
#define EXPORT_SYMTAB
#include <linux/module.h>
#include <linux/version.h>
#ifdef MODULE
char kernel_version [] = UTS_RELEASE;
#endif
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/malloc.h>
#include <linux/vmalloc.h>
#include <linux/file.h>
#include <asm/uaccess.h>
#include <linux/blk.h>
#include <linux/hdreg.h>
#include <linux/sched.h>
#include <linux/genhd.h>
#include <linux/devfs_fs_kernel.h>

#include "pool.h"

#define DEVICE_REQUEST do_pool_request

pool_t pools[POOL_MAX_DEVICES];
extern struct gendisk pool_gendisk;
/* one extra devfs handle for the directory the entries will be in. **
** The specific pool entries will be their minor+1................. */
devfs_handle_t pool_devfs_entries[POOL_MAX_DEVICES+1];
extern pool_debug_flag_t df_flags;

int pool_size[POOL_MAX_DEVICES];  	/* in KB */
static struct hd_struct pool_hd_struct[POOL_MAX_DEVICES];
static int pool_hard_blksizes[POOL_MAX_DEVICES];
static int pool_soft_blksizes[POOL_MAX_DEVICES];
static int pool_maxreadahead[POOL_MAX_DEVICES];
static char *pool_names[POOL_MAX_DEVICES];

static void pool_geninit(struct gendisk *);
static int pool_open(struct inode *inode, struct file *file);
static int pool_release(struct inode *inode, struct file *file);
int pool_ioctl(struct inode *inode, struct file *file,
	       unsigned int cmd, unsigned long arg);


static struct block_device_operations pool_blk_dops =
{
        open:           pool_open,
        release:        pool_release,
        ioctl:          pool_ioctl,
};

struct gendisk pool_gendisk=
{
  POOL_MAJOR,
  "pool",
  0,
  1,
  pool_hd_struct,
  pool_size,
  POOL_MAX_DEVICES,
  NULL,
  NULL,
  &pool_blk_dops,
  NULL,
  NULL,
  pool_names
};

/**
 * do_pool_request
 *
 * Stub routine to let Linux know we are a block device and we actually do
 * exist.
 */
static void do_pool_request(request_queue_t *q)
{
  osi_printf("pool: request function called, this should never be called\n");
}


/**
 * pool_open
 * @inode
 * @file
 *
 * This gets called for every open.  We basically just ensure the call is
 * to a valid minor number and increment the open counts accordingly.  In
 * addition, if this is the first open on the device, all the lower level
 * device comprising the pool are sent open calls.  (This ensures once a
 * user has the pool open lower level driver modules are not unloaded.)  A
 * special case is the opening of minor device 0.  This is the poolcomm
 * device and is the only pool_t that accepts assemble and remvove ioctls.
 * It is the only minor number that always exists and is not dynamically
 * assigned.  In addtion, access to it is serialized at the open level.
 * (That means we don't let more than one process open it at the same
 * time.)  This prevents multiple copies of passemble from interfering.
 * Note that I believe it is still possible for a process to open poolcomm
 * and fork.  This essentially bypasses the precaution taken here.  DO NOT
 * do this. (cjs)  Only one open (locally) is allowed on a pool to 
 * prevent unaccounted cache blocks.
 *
 * Returns: Linux errno
 */
static int pool_open(struct inode *inode, struct file *file) 
{
  int 		minor = MINOR(inode->i_rdev);
  pool_t	*pp;
  int 		error=0;

  ENTER_PRINT(POOL_1, "pool_open()");

  if (minor >= POOL_MAX_DEVICES) {
    osi_printf("POOL--poolopen: minor number out of range - %d\n", minor);
    error = -ENXIO;
    goto fail;
  }

  pp = &pools[minor];

  /* if this is an open on the comm device handle it specially. */

  if (minor == 0) { 
    pool_rw_sema_write_lock(&(pp->p_rwsema));

    /* make sure the comm device exists */
    ASSERT((pp->p_flags & PF_FREE) == 0);

    /* check to see if device is busy */
    if (!(pp->p_flags & PF_VALID)) {
      pool_rw_sema_write_unlock(&(pp->p_rwsema));
      error = -EBUSY;
      goto fail;
    }

    if (pp->p_oblk > 0 ) {
      pool_rw_sema_write_unlock(&(pp->p_rwsema));
      error = -EBUSY;
      goto fail;
    }

    if(pp->p_oblk != -1){
      pp->p_oblk++;
    } else {
      pool_rw_sema_write_unlock(&(pp->p_rwsema));
      error = -ENXIO;
      goto fail;
    }
    
    pool_rw_sema_write_unlock(&(pp->p_rwsema));
    MOD_INC_USE_COUNT;

    EXIT_PRINT(POOL_1, "pool_open()");

    return (0);
  }

  /* regular open */

  pool_rw_sema_write_lock(&(pp->p_rwsema));

  if (pp->p_flags & PF_FREE || !(pp->p_flags & PF_LOWEROPEN)) {
    pool_rw_sema_write_unlock(&(pp->p_rwsema));
    error = -ENXIO;
    goto fail;
  }

  ASSERT(pp->p_spools != NULL);

  pp->p_oblk++;

  pool_rw_sema_write_unlock(&(pp->p_rwsema));
  MOD_INC_USE_COUNT;

  EXIT_PRINT(POOL_1, "pool_open()");
  return (0);


 fail:

  EXIT_PRINT(POOL_1, "pool_open()");
  return (error);
}


/**
 * pool_release
 * @inode
 * @file
 *
 * Returns: Linux errno
 */
static int pool_release(struct inode *inode, struct file *file) 
{
  int 		minor = MINOR(inode->i_rdev);
  pool_t	*pp;

  ENTER_PRINT(POOL_1, "pool_release()");

  pp = &pools[minor];

  pool_rw_sema_write_lock(&(pp->p_rwsema));

  ASSERT((pp->p_flags & PF_VALID) != 0);
  ASSERT((pp->p_flags & PF_FREE) == 0);
  ASSERT(pp->p_oblk > 0);

  pp->p_oblk--;

  pool_rw_sema_write_unlock(&(pp->p_rwsema));
  MOD_DEC_USE_COUNT;

  EXIT_PRINT(POOL_1, "pool_release()");
  return (0);
}


/**
 * pool_map
 * @dev
 * @rdev
 * @rsector
 * @size
 * @rw
 *
 * rsector and size are in basic blocks and are relative to the
 * logical pool device.  pool_blkmap() is called to translate the 
 * logical device and logical block into the actual device and block
 * for the request.
 *
 * Returns: rdev and rsector are changed to reflect the mapping, Linux
 * errno on error, 0 if OK.
 */
static int pool_map(kdev_t dev, kdev_t *rdev, unsigned long *rsector, 
		    unsigned long size, int rw)
{
  pool_t	*pp = NULL;
  device_t	*p_dev;
  int	       	bc;
  uint64      	dev_sector;
  int		error=0;
  int		minor = MINOR(dev);

  ENTER_PRINT(MAPPING_1, "pool_map() :: minor=%d, rsector=%ld, size=%ld",
      minor, *rsector, size);
  if (minor == 0) {  	/* handle writes to poolcomm */    
    osi_printf("pool_map: can not write to poolcomm.\n");
    error = -EPERM;
    goto fail;
  }

  pp = &pools[minor];

  pool_rw_sema_read_lock(&(pp->p_rwsema));

  if (pp->p_flags & PF_FREE || !(pp->p_flags & PF_VALID) || 
      !(pp->p_flags & PF_LOWEROPEN)) {
    pool_rw_sema_read_unlock(&(pp->p_rwsema));
    osi_printf("pool_map: No such minor device %d\n", minor);
    error = -EIO;
    goto fail;
  }
  

  /* Simple bound test */

  pool_rw_sema_read_unlock(&(pp->p_rwsema));

  if ((*rsector + size ) > pp->p_capacity) {
    osi_printf("pool_map: Request exceeds device size.\n");
    osi_printf("pool_map: Attempting to reassemble pool in the hopes that"
	       " it has been grown.\n");
    error = assemble_pools(pp->p_name);
    if(error){
      osi_printf("pool_map: Attempt to grow failed.\n");
      osi_printf("\npool_map: Error with offset or size\n");
      osi_printf("pool_map: blkno=%u, bcount=%lu, capacity=%"FMT_64"u\n", 
		 (uint32)*rsector, size, pp->p_capacity);
      error = -EIO;
      goto fail;
    } else if((*rsector + size ) > pp->p_capacity){ /* recheck capacity */
      osi_printf("pool_map: Capacity still not large enough.\n");
      error = -EIO;
      goto fail;
    } 
  }

  if (pool_blkmap(pp, *rsector, size << BBSHIFT, &dev_sector, &bc, &p_dev)) {
    error = -EIO;
    goto fail;
  }

  if (sizeof(unsigned long) <= 4 && dev_sector > UINT_MAX) {
    osi_printf("poolmap: attempted access beyond 32 bit sector limit.\n");
    error = -EIO;
    goto fail;
  }

  *rdev = p_dev->d_dev;
  *rsector = dev_sector;


  /* do statistics */

#if (POOL_STATS == 1)
  {
    osi_spin_lock(&pp->p_statslock);

    if (rw == READ || rw == READA) {
      pp->p_stats.s_reads++;
      pp->p_stats.s_readblks += size;
    }
    else {
      pp->p_stats.s_writes++;
      pp->p_stats.s_writeblks += size;
    }
	
    osi_spin_unlock(&pp->p_statslock);
  }
#endif

  EXIT_PRINT(MAPPING_1, "pool_map()");
  return (0);


 fail:

  EXIT_PRINT(MAPPING_1, "pool_map()");
  return (error);

}


/**
 * pool_make_request_fn
 * @q
 * @rw: read or write (use macros)
 * @bh: buffer_head to be remapped
 *
 * This funtion simply formats the arguments for pool_map.
 *
 * Returns: 0 on failure, 1 if OK.
 */
static int pool_make_request_fn(request_queue_t *q,
                               int rw,
                               struct buffer_head *bh){

  if (pool_map(bh->b_dev, &(bh->b_rdev), &(bh->b_rsector),
	       bh->b_size >> 9, rw) < 0){
    bh->b_end_io(bh, 0);
    return 0; /* failure, buffer_IO_error has been called, don't recurse */
  } else {
    return 1; /* all ok, mapping done, call lower level driver */
  }

}


/**
 * pool_ioctl
 * @arg
 * @size
 *
 * Implements all pool ioctl's.  This includes assemble/remove,
 * statistics gathering, and other miscellaneous ioctls.  Note there should
 * be only one retrn from this function to handle the return protocol
 * correctly.  If an ioctl causes an internal routine to be called, the
 * parameters are copied to kernel space before the call is made.  NOTE:
 * structures that are in user space and are pointed to by the structure
 * passed as the ioctl arguement are not copied into kernel space before the
 * call.  Some ioctls will only work on the /dev/poolcomm device (minor=0)
 *
 * Returns: 0 if OK, Linux errno on error.
 */
int pool_ioctl(struct inode *inode, struct file *file, unsigned int cmd, 
	       unsigned long arg) 
{
  kdev_t      	dev 	= inode->i_rdev;
  int         	minor 	= MINOR(dev);
  pool_t	*pp    	= NULL;
  int           rc 	= 0;

  /*  ENTER_PRINT(POOL_1, "pool_ioctl()\n\tDevice:  %d/%d\n\tCommand: 0x%x",
	      MAJOR(dev), MINOR(dev), cmd);
  */
  if (minor < 0 || minor >= POOL_MAX_DEVICES)
    return (-ENXIO);


  pp = &pools[minor];

  pool_rw_sema_read_lock(&(pp->p_rwsema));
  if (pp->p_flags & PF_FREE) {
    pool_rw_sema_read_unlock(&(pp->p_rwsema));
    return (-ENXIO);
  }
  ASSERT(POOL_BUSY(pp));
  ASSERT(pp->p_flags & PF_VALID);
  pool_rw_sema_read_unlock(&(pp->p_rwsema));

  /* Do commands */

  switch(cmd) {

  case POOL_MEMEXP:
  {
    /* Someday I may put this struct into a .h file, but **
    ** right now I don't know where it fits best.        */
    struct {
      uint64 pk_bid;  /* pk_ stands for package, for lack of better name */
      u_char pk_cdb[16];
      uint32 pk_pdf_len;
      u_char *pk_pdf;
      error_code_t pk_ec;
      uint32 pk_major;
      uint32 pk_minor;
    } package;
    u_char *usr_pdf_ptr;
    u_char *krn_pdf_ptr;

    if (osi_copy_from_user(&package,
			   (void *)arg,
			   sizeof(package)) != 0) {
      rc = -EFAULT;
      break;
    }
    usr_pdf_ptr = package.pk_pdf;
    krn_pdf_ptr = osi_malloc(package.pk_pdf_len);
    
    if (osi_copy_from_user(krn_pdf_ptr, usr_pdf_ptr,
			   package.pk_pdf_len) != 0) {
      osi_free(krn_pdf_ptr, package.pk_pdf_len);
      rc = -EFAULT;
      break;
    }

    if(package.pk_major){
      rc = pool_memexp(MKDEV(package.pk_major, package.pk_minor),
		       REAL_DEVICE,
		       package.pk_cdb, package.pk_bid, krn_pdf_ptr,
		       package.pk_pdf_len, &(package.pk_ec));
    }else{
      rc = pool_memexp(MKDEV(POOL_MAJOR, minor),
		       LOGICAL_DEVICE,
		       package.pk_cdb, package.pk_bid, krn_pdf_ptr,
		       package.pk_pdf_len, &(package.pk_ec));
    }

    if (osi_copy_to_user(usr_pdf_ptr,
			 krn_pdf_ptr,
			 package.pk_pdf_len) != 0) {
      rc = -EFAULT;
    }
    if (osi_copy_to_user((void *)arg,
			 &package, sizeof(package)) != 0) {
      rc = -EFAULT;
    }
    osi_free(krn_pdf_ptr, package.pk_pdf_len);
    break;
  }
  case POOL_CLEAR_STATS:
  {
    /* IFDEBUG(POOL_DEBUG2,
	    osi_printf("  Clearing statistics\n");); */
    
    osi_spin_lock(&pp->p_statslock);

    osi_memset(&(pp->p_stats), 0, sizeof(pool_stat_t));

    osi_spin_unlock(&pp->p_statslock);
    break;
  }

  case POOL_DUMP_TOTALSTATS:
  {
    osi_spin_lock(&pp->p_statslock);
    
    if (osi_copy_to_user((void *)arg, &(pp->p_stats), sizeof(pool_stat_t)))
      rc = -EFAULT;
    
    osi_spin_unlock(&pp->p_statslock);
    break;
  }

  case BLKGETSIZE: 
  {
    long value = pp->p_capacity;

    if (!arg) 
      return (-EINVAL);

    if (osi_copy_to_user((void *)arg, &value, sizeof(long)))
      rc = -EFAULT;
    break;       
  }

  /* Linux specific IOCTL's */

  case HDIO_GETGEO:
    osi_printf("poolioctl: HDIO_GETGEO\n");
    return -ENXIO;

  case BLKRASET:
    if (arg > 0xff) {
      rc = -EINVAL;
      break;
    }
    
    read_ahead[MAJOR(dev)] = arg;
    break;

  case BLKRAGET:
    
    osi_copy_to_user((void *)arg, &read_ahead[MAJOR(dev)], sizeof(long));
    break;

  case BLKFLSBUF:
    fsync_dev(dev);
    invalidate_buffers(dev);
    return 0;
    /*
  case BLKRRPART:
    osi_printf("poolioctl: BLKRRPART\n");
    return -ENXIO;
   
    RO_IOCTLS(dev, arg);
    */
  case POOL_ADD: 
  {
    pool_add_t	add;

    if (minor) {
      return (-EINVAL);
    }

    if (osi_copy_from_user(&add, (void *)arg, sizeof(add)) != 0) {
      return (-EFAULT);
    }

    rc = uadd_pool(&add);

    minor = MINOR(add.pa_pool_dev);

    pool_size[minor] = (pools[minor].p_capacity >> 1);

    if (osi_copy_to_user((void *)arg, &add, sizeof(add)) != 0) {
      if (!rc)
	rc = -EFAULT;
    }
    break;
  }

  case POOL_REMOVE:
  {
    uint32 m;   /* minor number of device to remove */

    if (minor) { 
      return (-EINVAL);
    }

    if (osi_copy_from_user((void *)&m, (void *)arg,
			   sizeof(uint32)) != 0) {
      rc = -EFAULT;
      break;
    }

    rc = remove_pool(m);
    break;
  }

  case POOL_GROW:
  {
    pool_add_t	grow;

    if (minor) {
      return (-EINVAL);
    }

    if (osi_copy_from_user(&grow, (void *)arg,
			   sizeof(grow)) != 0) {
      return (-EFAULT);
    }
    
    rc = ugrow_pool(&grow);

    minor = MINOR(grow.pa_pool_dev);

    pool_size[minor] = (pools[minor].p_capacity >> 1);

    if (osi_copy_to_user((void *)arg, &grow,
			 sizeof(grow)) != 0) {
      if (!rc)
	rc = -EFAULT;
    }

    break;
  }

  case POOL_COUNT:
  {
    int count = 0;

    rc = count_pool(&count);

    if (osi_copy_to_user((void *)arg, (void *)&count,sizeof(int)) != 0) {
      if (!rc)
	rc = -EFAULT;
    }

    break;
  }
    
  case POOL_LIST:
  {
    pool_list_t	list;

    if (minor) {
      return (-EINVAL);
    }

    if (osi_copy_from_user((void *)&list, (void *)arg,
			   sizeof(pool_list_t)) != 0) {
      return (-EFAULT);
    }
    
    rc = list_pool(&list);

    if (osi_copy_to_user((void *)arg, (void *)&list,
			 sizeof(pool_list_t)) != 0) {
      if (!rc)
	rc = -EFAULT;
    }
    break;
  }

  case POOL_SPLIST:
  {
    pool_splist_t list;

    if (osi_copy_from_user(&list, (void *)arg,
			   sizeof(list)) != 0) {
      return -EFAULT;
    }

    if (list.spl_count <= 0) {
      rc = -EFAULT;
      break;
    }

    rc = list_spool(pp, &list);

    if (osi_copy_to_user((void *)arg, (void *)&list,
			 sizeof(list)) != 0) {
      if (!rc)
	rc = -EFAULT;
    }
    break;
  }
  case POOL_DEVLIST:
  {
    pool_devlist_t list;

    if (osi_copy_from_user(&list, (void *)arg,
			   sizeof(list)) != 0) {
      return -EFAULT;
    }

    if (list.pdl_count <= 0) {
      rc = -EFAULT;
      break;
    }

    rc = list_pooldevs(pp, &list);

    if (osi_copy_to_user((void *)arg, (void *)&list,
			 sizeof(list)) != 0) {
      if (!rc)
	rc = -EFAULT;
    }
    break;
  }
  case POOL_IND_CACHE:
  {
    fsync_dev(dev);
    invalidate_buffers(dev);
    break;
  }

  case POOL_DEBUG_ON:
  {
    int str_size;
    char buffer[20];

    if (osi_copy_from_user(&str_size, (void *)arg,
			   sizeof(int))) {
      return -EFAULT;
    }

    if(str_size > 19)
      return -EFAULT;

    if (osi_copy_from_user(buffer, (void *)arg+sizeof(int),
			   str_size)) {
      return -EFAULT;
    }
    buffer[str_size] = '\0';

    turn_on_pool_debug_flag(buffer);
	
    break;
  }
  case POOL_DEBUG_OFF:
  {
    int str_size;
    char buffer[20];
    if (osi_copy_from_user(&str_size, (void *)arg,
			   sizeof(int))) {
      return -EFAULT;
    }

    if(str_size > 19)
      return -EFAULT;

    if (osi_copy_from_user(buffer, (void *)arg+sizeof(int),
			   str_size)) {
      return -EFAULT;
    }
    buffer[str_size] = '\0';
    turn_off_pool_debug_flag(buffer);

    break;
  }
  default:
    rc = -EINVAL;
  }

  /*  EXIT_PRINT(POOL_1, "pool_ioctl()");*/
  return (rc);
}


static void pool_geninit (struct gendisk *gdisk)
{
  int i;
  ENTER_PRINT(POOL_1, "pool_geninit()");
  for(i=0;i<POOL_MAX_DEVICES;i++){
    pool_gendisk.part[i].start_sect=-1; /* avoid partition check */
    pool_gendisk.part[i].nr_sects=0;
    pool_gendisk.device_names[i] = NULL;
  }
  EXIT_PRINT(POOL_1, "pool_geninit()");
}




/**
 * pool_init
 *
 * This routine is called by init_module().  It does the 
 * important elements required for Linux block driver initialization
 * including registering itself as a block device in the kernel and
 * setting up the blk_dev[] function pointers.  Pool locks are also
 * initialized. 
 *
 * Returns: Linux errno
 */
static int pool_init(void)
{
  int i;

  ENTER_PRINT(POOL_1, "pool_init()");
  osi_printf("Pool %s (built %s %s) installed\n",
	 GFS_RELEASE_NAME,
	 __DATE__, __TIME__);

  if (register_blkdev(POOL_MAJOR, "pool", &pool_blk_dops)) {
    osi_printf ("Unable to get major %d for pool\n", POOL_MAJOR);
    return (-1);
  }

  read_ahead[POOL_MAJOR]= MAX_READAHEAD;

  for (i=0; i < POOL_MAX_DEVICES; i++) {
    pool_soft_blksizes[i] = 4096;		/* 512 breaks pool on md */
    pool_hard_blksizes[i] = 0;
    pool_maxreadahead[i] = MAX_READAHEAD;
    pool_size[i] = 0;
  }

  hardsect_size[POOL_MAJOR]=pool_hard_blksizes;
  /* what is blksize_size used for ?? */
  blksize_size[POOL_MAJOR]=pool_soft_blksizes;
  blk_size[POOL_MAJOR]=pool_size;
  max_readahead[POOL_MAJOR]=pool_maxreadahead;

  /* initialize pools[] array */
  osi_memset(pools, 0, POOL_MAX_DEVICES * sizeof(pool_t)); 
 
  for (i=0; i < POOL_MAX_DEVICES; i++) {
    pool_rw_sema_init(&(pools[i].p_rwsema));
    osi_spinlock_init(&pools[i].p_statslock);
    pools[i].p_flags = PF_VALID | PF_FREE;
  }
  
  pools[0].p_flags = PF_VALID;

  /* In the 2.2 kernels, the pool_geninit function was linked into **
  ** the gendisk structure and taken care of by the kernel.  Now   **
  ** we must call this function on our own - it is no longer in the**
  ** structure.................................................... */
  pool_geninit(&pool_gendisk);

  /* Maybe we don't need to do this, but I thought it would be nice **
  ** to sort by major number.					    */
  if (gendisk_head != NULL) {
    struct gendisk *gendisk_ptr = gendisk_head;

    while (gendisk_ptr->next != NULL &&
	   gendisk_ptr->major > pool_gendisk.major) {
      gendisk_ptr = gendisk_ptr->next;
    }
    pool_gendisk.next = gendisk_ptr->next;
    gendisk_ptr->next = &pool_gendisk;
  } else {
    gendisk_head = &pool_gendisk;
    pool_gendisk.next = NULL;
  }

  /* set up the devfs directory */
  pool_devfs_entries[0] = devfs_mk_dir(NULL, "pool", NULL);
  pool_devfs_entries[1] = 
    devfs_register(NULL, "poolcomm",
		 DEVFS_FL_DEFAULT, POOL_MAJOR,
		 0, S_IFBLK|S_IRWXU, &pool_blk_dops, NULL);

  blk_init_queue(BLK_DEFAULT_QUEUE(POOL_MAJOR), DEVICE_REQUEST);
  blk_queue_make_request(BLK_DEFAULT_QUEUE(POOL_MAJOR), pool_make_request_fn);
  EXIT_PRINT(POOL_1, "pool_init()");
  return (0);
}


/**
 * bdopen - open a lower level block device
 * @dev
 * @inodep
 * @filep
 *
 * Returns: Linux errno
 */
int bdopen(osi_dev_t dev, osi_vnode_t **inodep, osi_vfile_t **filep)
{
  pool_t *pp;
  int error=0;

  ENTER_PRINT(POOL_1, "bdopen()");
  if (MAJOR(dev) == POOL_MAJOR) {

    pp = &pools[MINOR(dev)];

    pool_rw_sema_write_lock(&(pp->p_rwsema));

    if (pp->p_flags & PF_FREE || !(pp->p_flags & PF_LOWEROPEN)) {
      pool_rw_sema_write_unlock(&(pp->p_rwsema));
      error = -ENXIO;
      osi_printf("bdopen of pool - error 1\n");
      goto fail;
    }

    if (!(pp->p_flags & PF_VALID)) {
      pool_rw_sema_write_unlock(&(pp->p_rwsema));
      error = -EBUSY;
      osi_printf("bdopen of pool - error 2\n");
      goto fail;
    }

    ASSERT(pp->p_spools != NULL);

    pp->p_oblk++;

    pool_rw_sema_write_unlock(&(pp->p_rwsema));
    MOD_INC_USE_COUNT;

    EXIT_PRINT(POOL_1, "Exiting bdopen\n");
    return (0);		
  }

  *inodep = get_empty_inode();
  (*inodep)->i_dev = dev;

  /* All this file stuff is a bunch of crap... No-one ever uses **
  ** the filep, but I have to keep it around just incase someone**
  ** decides to mess with it lower down........................ */
  *filep = osi_malloc(sizeof(struct file));
  memset(*filep, 0, sizeof(struct file));
  (*filep)->f_mode = 3; /* this is read/write */
  (*filep)->f_dentry = osi_malloc(sizeof(struct dentry));
  memset((*filep)->f_dentry, 0, sizeof(struct dentry));
  (*filep)->f_dentry->d_inode = (*inodep);
  (*filep)->f_op = &def_blk_fops;

  init_special_inode(*inodep, 
                    S_IFBLK | S_IRUSR | S_IWUSR, 
                    kdev_t_to_nr(dev));
  if(blkdev_open(*inodep, *filep) != 0){
    blkdev_put((*inodep)->i_bdev, BDEV_FILE);
    iput(*inodep);
    osi_free((*filep)->f_dentry, sizeof(struct dentry));
    osi_free((*filep), sizeof(struct file));
    error = -1;
  }
 
 fail:

  EXIT_PRINT(POOL_1, "bdopen()");

  return (error);
}


/**
 * bdclose - closes a lower level block device
 * @dev
 * @inodep
 * @filep
 *
 * Returns: Linux errno
 */

int bdclose(osi_dev_t dev, osi_vnode_t **inodep, osi_vfile_t **filep)
{
  pool_t *pp;
  int error=0;

  ENTER_PRINT(POOL_1, "bdclose()");

  if (MAJOR(dev) == POOL_MAJOR) {
    pp = &pools[MINOR(dev)];

    /* locking must be the responsibility of a higher layer 
       pool_rw_sema_write_lock(&(pp->p_rwsema));*/

    ASSERT((pp->p_flags & PF_FREE) == 0);
    ASSERT(pp->p_oblk > 0);

    if (pp->p_oblk > 0)
      pp->p_oblk--;

    /* locking must be the responsibility of a higher layer 
       pool_rw_sema_write_unlock(&(pp->p_rwsema));*/
    MOD_DEC_USE_COUNT;

    EXIT_PRINT(POOL_1, "bdclose()");
    return (0);
  }

  blkdev_put((*inodep)->i_bdev, BDEV_FILE);
  iput(*inodep);
  osi_free((*filep)->f_dentry, sizeof(struct dentry));
  osi_free((*filep), sizeof(struct file));

  EXIT_PRINT(POOL_1, "bdclose()");
  return (error);
}


/*
 * get_device_sector_size - an abstraction for get_hardsect_size
 * @device: the device to get the sector size for
 *
 * Returns: the sector size
 */
int get_device_sector_size(osi_dev_t device){
  return (int)get_hardsect_size(device);
}


/**
 * read_partition_list - reads devices from /proc
 * @ptr
 *
 * This routine will place all of the devices known in the gendisk
 * list into an array which will be malloced and pointed to by *ptr.
 * The number of the devices is what is returned.
 *
 * Returns: the number of devices found.
 */
int read_partition_list(osi_dev_t **ptr)
{
  int i;
  int n, num_devs = 0;
  osi_dev_t *array;
  struct gendisk *p;

  ENTER_PRINT(POOL_1, "read_partition_list()");

  for (p = gendisk_head; p; p = p->next) {
    for (n=0; n < (p->nr_real << p->minor_shift); n++) {
      if (p->part[n].nr_sects) {
	num_devs++;
      }
    }
  }

  *ptr = osi_malloc(num_devs * sizeof(osi_dev_t));
  array = *ptr;

  for (p = gendisk_head, i = 0; p; p = p->next) {
    for (n=0; n < (p->nr_real << p->minor_shift); n++) {
      if (p->part[n].nr_sects) {
	array[i++] = osi_make_dev(p->major, n);
      }
    }
  }

  EXIT_PRINT(POOL_1, "read_partition_list()");
  return num_devs;
}



/**
 * scan_dev_for_pool_label - read a pool label if on the device
 * @dev
 * @label
 *
 * This routine will scan a specific device for a pool label.  If found,
 * the label will be copied into the space pointed to by (the apply named)
 * "label".
 *
 * Returns: 1 = success, 0 = no pool label, -EXXX on error
 */
int scan_dev_for_pool_label(osi_dev_t dev, pool_label_t *label){
  int error=0;
  int rtn = 0;
  unsigned int blksize;
  struct buffer_head *bh = NULL;
  struct inode *inode = NULL;
  struct file *file = NULL;

  ENTER_PRINT(POOL_1, "scan_dev_for_pool_label()");

  error = bdopen(dev, &inode, &file);

  if (error) {
    rtn = 0;
  } else {
    bdclose(dev, &inode, &file);
    blksize = blksize_size[MAJOR(dev)][MINOR(dev)];
    bh = getblk(dev, 0, blksize);
    mark_buffer_uptodate(bh, FALSE);
    ll_rw_block(READ, 1, &bh);
    wait_on_buffer(bh);
    if (!buffer_uptodate(bh)){
      brelse(bh);
      bh =  NULL;
      rtn = -EIO;
    }
  }

  if(bh){
    pool_label_in(label, bh->b_data);

    if(label->pl_magic == POOL_MAGIC){
      brelse(bh);
      rtn = 1;
    } else {
      brelse(bh);
      rtn = 0;
    }
  }

  EXIT_PRINT(POOL_1, "scan_dev_for_pool_label()");
  return rtn;
}


/**
 * register_pool - registers pool with Linux /proc
 * @new: the new pool to be registered
 * @minor: it's corresponding minor number
 *
 * This function provides the Linux /proc system with info
 * on size of the device and it's name.
 *
 * Returns: Linux errno
 */
extern int * hardsect_size[MAX_BLKDEV];
int register_pool(pool_t *new, int minor){
  ENTER_PRINT(POOL_1, "register_pool()");
  MOD_INC_USE_COUNT;
  pool_hard_blksizes[minor] = new->p_blksize;

  /* set up gendisk structure, which is used by /proc */
  pool_gendisk.part[minor].start_sect=0;
  pool_gendisk.part[minor].nr_sects=new->p_capacity>>1;
  pool_gendisk.device_names[minor] =
     osi_malloc(strlen("pool/")+strlen(new->p_name)+1);
  osi_sprintf(pool_gendisk.device_names[minor],
  "pool/%s", new->p_name);

  /* make entry in devfs, if necessary */
  pool_devfs_entries[minor+1] = 
    devfs_register(pool_devfs_entries[0], new->p_name,
		 DEVFS_FL_DEFAULT, POOL_MAJOR,
		 minor, S_IFBLK|S_IRWXU, &pool_blk_dops, NULL);

  EXIT_PRINT(POOL_1, "register_pool()");
  return 0; /* success */
}


/**
 * unregister_pool - registers pool with Linux /proc
 * @minor: minor number of pool to be removed
 *
 * This function has the opposite effect as register_pool.
 *
 * Returns: Linux errno
 */
int unregister_pool(int minor){
  ENTER_PRINT(POOL_1, "unregister_pool()");
  if(pool_gendisk.device_names[minor] != NULL){
    MOD_DEC_USE_COUNT;
    pool_hard_blksizes[minor] = 0;
    pool_gendisk.part[minor].start_sect=0;
    pool_gendisk.part[minor].nr_sects=0;
    osi_free(pool_gendisk.device_names[minor],
	     strlen(pool_gendisk.device_names[minor]));
    pool_gendisk.device_names[minor] = NULL;
  }
  EXIT_PRINT(POOL_1, "unregister_pool()");
  return 0; /* success */
}


/**
 * init_pool
 *
 * This function is required if pool is to be compiled directly
 * into the kernel.
 *
 * Returns: 0 on success, -EXXXX on failure
 */
int __init init_pool(void)
{
  int rc = 0;
  ENTER_PRINT(POOL_1, "init_pool()");
  rc = pool_init();
#ifndef MODULE
/*
  if(!rc){
    rc = assemble_pools(NULL);
  }
*/
#endif
  EXIT_PRINT(POOL_1, "init_pool()");
  return rc;
}

/**
 * exit_pool - does all the necessary things to remove pool from the kernel
 */
void exit_pool(void)
{
  int i, j, k;
  ENTER_PRINT(POOL_1, "exit_pool()");
  for (i=0; i < POOL_MAX_DEVICES; i++) {
    pool_rw_sema_write_lock(&(pools[i].p_rwsema));
    if (pools[i].p_flags & PF_FREE) {
      pool_rw_sema_write_unlock(&(pools[i].p_rwsema));
      continue;
    }

    /* sync operations to flush any dirty buffers */

    for (j=0; j < pools[i].p_num_spools; j++) {
      for (k=0; k < pools[i].p_spools[j].sp_dmepdevs; k++) {
	fsync_dev(pools[i].p_spools[j].sp_dmepdevlist[k].dm_dev);
      }
    }

    for (j=0; j < pools[i].p_num_spools; j++) {
      for (k=0; k < pools[i].p_spools[j].sp_devs; k++)
	fsync_dev(pools[i].p_spools[j].sp_devlist[k].d_dev);
    }

    if (!(pools[i].p_flags & PF_VALID)) {
      /* attention -- return error here if no higher level check exists */
      osi_printf("ARRGG unloading module while still in use\n");
    }

    if (pools[i].p_flags & PF_LOWEROPEN) {
      close_devices(&pools[i]);
    }
    
    free_pool(&pools[i]);
    pools[i].p_flags = PF_VALID | PF_FREE;
    pool_rw_sema_write_unlock(&(pools[i].p_rwsema));
  }
  
  /* Attention -- It might be wise to lock through this section... */
  if(gendisk_head == &pool_gendisk){
    gendisk_head = pool_gendisk.next;
  } else {
    int removed = 0;
    struct gendisk *tmp_gendisk = gendisk_head;

    while(tmp_gendisk->next){
      if(tmp_gendisk->next == &pool_gendisk){
	tmp_gendisk->next = pool_gendisk.next;
	removed = 1;
	break;
      }
      tmp_gendisk = tmp_gendisk->next;
    }
    if(!removed){
      osi_printf("Pool could not be removed from the gendisk list!\n");
    }
  }


  unregister_blkdev(POOL_MAJOR, "pool");
  blk_cleanup_queue(BLK_DEFAULT_QUEUE(POOL_MAJOR));

  read_ahead[POOL_MAJOR] = 0;
  blk_size[POOL_MAJOR] = NULL;
  blksize_size[POOL_MAJOR] = NULL;
  EXIT_PRINT(POOL_1, "exit_pool()");
}

#ifdef MODULE

/**
 * init_module
 *
 * Returns: Linux errno
 */
int init_module(void)
{
  ENTER_PRINT(POOL_1, "init_module()");
  return init_pool();
  EXIT_PRINT(POOL_1, "init_module()");
  return 0;
}

/**
 * cleanup_module - calls exit_pool
 */
void cleanup_module(void)
{
  ENTER_PRINT(POOL_1, "cleanup_module()");
  exit_pool();
  EXIT_PRINT(POOL_1, "cleanup_module()");
}

#endif  /* MODULE */

EXPORT_SYMBOL(gencopy);
EXPORT_SYMBOL(pool_memexp);
EXPORT_SYMBOL(pools);


