/******************************************************************************
*******************************************************************************
**
**    Copyright 1999-2001 Sistina Software, Inc.
**    Portions Copyright 2000-2001 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Jonathan Brassow <brassow@sistina.com>
** 
**    Module name: pool_user.h
**
*******************************************************************************
******************************************************************************/

#ifndef __POOL_USER_DOT_H__
#define __POOL_USER_DOT_H__

/*  Include Files
    These should only be the ones necessary to compile the common code.  */

#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include <asm/cache.h>  /* SMP_CACHE_BYTES */

#include <global.h>

/*  Memory allocation abstraction  */

static __inline__ void *osi_malloc(int size)
{
  void *data;

  data = malloc(size);

  return data;
}

static __inline__ void osi_free(void *data, int size)
{
  free(data);
}



/*  Memory copy abstraction  */

#define osi_copy_to_user memcpy
#define osi_clear_user(x, y) memset((x), 0, (y))
#define osi_copy_from_user memcpy

#define osi_io_read(b, c, f, l) osi_copy_to_user((char *)(b) + (c), (f), (l))
#define osi_io_read_hole(b, c, l) osi_clear_user((char *)(b) + (c), (l))
#define osi_io_write(b, c, t, l) osi_copy_from_user((t), (char *)(b) + (c), (l))

#define osi_memset memset
#define osi_memcpy memcpy



/*  printf() abstraction  */

#define osi_printf printf
#define osi_sprintf sprintf
#define panic(fmt, args...) \
{ \
  fprintf(stderr, "%s: ", prog_name); \
  fprintf(stderr, fmt, ## args); \
  exit(-1); \
}
#define osi_panic panic



/*  GFS to VFS abstraction  */

typedef int osi_vfs_t;
typedef int osi_vnode_t;
typedef int osi_vfile_t;



/*  Buffer head abstraction  */

struct buffer_head
{
  unsigned long b_blocknr;
  unsigned long b_size;
  unsigned long b_state;
  char * b_data;
};
typedef struct buffer_head osi_buf_t;


/*  Device type abstraction  */

typedef unsigned short osi_dev_t;


/*  Atomic Bit Manipulation interface  */

#define osi_test_bit(nr, addr) (*(addr) &   (1 << nr))
#define osi_set_bit(nr, addr) (*(addr) |=  (1 << nr))
#define osi_clear_bit(nr, addr) (*(addr) &= ~(1 << nr))
#define osi_test_and_set_bit(nr, addr) (*(addr) |=  (1 << nr))
#define osi_test_and_clear_bit(nr, addr) (*(addr) &= ~(1 << nr))

typedef int osi_atomic_t;

#define osi_atomic_inc(x) (*(x) += 1)
#define osi_atomic_dec(x) (*(x) -= 1)
#define osi_atomic_read(x) (*(x))
#define osi_atomic_dec_and_test(x) (*(x) -= 1)
#define osi_atomic_set(x, y) (*(x) = (y))



/*  Endianness conversion abstraction  */

#include "osi_endian.h"



/*  Directory translation abstraction  */

typedef int (*osi_filldir_t)(void *, const char *, int, off_t, ino_t);



/*  struct qstr definition 
 
    "quick string" -- eases parameter passing, but more importantly
    saves "metadata" about the string (ie length and the hash).

    From /usr/include/linux/dcache.h
    */

struct qstr {
  const unsigned char * name;
  unsigned int len;
  unsigned int hash;
};
typedef struct qstr osi_qstr_t;



/*  Semaphore abstraction

    All macros take a pointer to a osi_semaphore_t structure.
    osi_sema_down_try() returns TRUE if the down() succeeds 
    */

typedef int osi_sema_t;

#define osi_sema_init(x) (*(x) = 1)
#define osi_sema_init_lkd(x) (*(x) = 0)

#define osi_sema_lock(x) (*(x) -= 1)
#define osi_sema_lock_intr(x) (*(x) -= 1)
#define osi_sema_trylock(x) (*(x) -= 1)
#define osi_sema_unlock(x) (*(x) += 1)



/*  Spinlock abstraction  */

typedef int osi_spinlock_t;
typedef int osi_rwlock_t;

#define osi_spinlock_init(x) (*(x) = 0)
#define osi_rwlock_init(x) (*(x) = 0)

#define osi_spin_lock(lock)
#define osi_spin_unlock(lock)
#define osi_spin_trylock(lock)

#define osi_read_lock(lock)
#define osi_write_lock(lock)
#define osi_read_unlock(lock)
#define osi_write_unlock(lock)
#define osi_write_trylock(lock)



/*  PID abstraction

    Returns the PID of the current process.
    */

#define osi_pid() 1

#define osi_wake_up_process(x)
typedef int osi_task_t;



/*  Time  */

#define osi_current_time() time()

static __inline__ uint64_t osi_gettimeofday(void)
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return((uint64_t)tv.tv_sec * 1000000 + tv.tv_usec);
}

typedef unsigned long osi_clock_ticks_t;
#define osi_time_stamp(x)

static __inline__ int osi_check_timeout(osi_clock_ticks_t *t, int hz)
{
  return 0;
}



/*  Schedule abstraction - Macro that makes a process temporarily
    give up control of the processor and lets other processes have
    a change to run. 
    Sleep abstraction - Sleep for x number of sections.  */

#define osi_schedule()
#define osi_sleep(x)



/*  Wait queue abstraction  */

typedef int * osi_wait_chan_t;

#define osi_wait_vars(x, y)
#define osi_init_wait_vars(x)
#define osi_init_wait_chan(x)
#define osi_asleep(x, y)
#define osi_await(x)
#define osi_unsleep(x, y)
#define osi_wakeup(x)



/*  Other Stuff  */

long long int llseek(int, long long int, int);

extern char *prog_name;


#endif  /*  __POOL_USER_DOT_H__  */




