/******************************************************************************
*******************************************************************************
**
**    Copyright 2000-2001 Sistina Software Inc
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Jonathan Brassow <brassow@sistina.com>
** 
**    Module name: common.c
**
**    Description:
**	General, non-OS specific pool driver routines.  
**
*******************************************************************************
******************************************************************************/

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif  /*  MODVERSIONS  */
#include "pool.h"
#include "pool_sptypes.h"
#include "rw_sema.h"

extern pool_t pools[POOL_MAX_DEVICES];
extern pool_debug_flag_t df_flags;
extern int pool_size[POOL_MAX_DEVICES];

/**
 * open_devices - open all pool devices block and memexp
 * @pp
 * @beginning
 *
 * (pool_t *)pp is the pool to have its devices opened.
 * (int)beginning is the subpool to begin opening devices at.
 * In the case of a new pool (or one that isn't open) this
 * beginning would be zero.  However, if the pool is being grown,
 * then beginning should be the index of the first of the added
 * subpools.
 *
 * Returns: Linux errno
 */

int open_devices(pool_t *pp, int beginning) 
{
  subpool_t           	*sp;
  device_t            	*devp;
  dmepdev_t		*dmepdevp;
  osi_dev_t 		dev;
  int			i,j;
  int			error;


  ENTER_PRINT(POOL_1,"open_devices()");

  /* clear the inode and file pointers so we know what to back out */

  for (j=beginning; j<pp->p_num_spools; j++) {

    sp = &(pp->p_spools[j]);

    for (i=0; i<sp->sp_devs; i++) {
      devp = &(sp->sp_devlist[i]);
      devp->inode = NULL;
      devp->file = NULL;
    }

    for (i=0; i<sp->sp_dmepdevs; i++) {
      dmepdevp = &(sp->sp_dmepdevlist[i]);
      dmepdevp->inode = NULL;
      dmepdevp->file = NULL;
    }
  }

  
  for (j=beginning; j<pp->p_num_spools; j++) {

    sp = &(pp->p_spools[j]);

    for (i=0; i<sp->sp_devs; i++) {
      devp = &(sp->sp_devlist[i]);
      dev = devp->d_dev;

      error = bdopen(dev, &devp->inode, &devp->file);

      if (error) {
	osi_printf("\tpool: open failed: error %d opening %d,%d\n",
	       error, MAJOR(dev), MINOR(dev));
	i--;
	goto disk_backout;
      }
    }
  }
  

  for (j=beginning; j<pp->p_num_spools; j++) {

    sp = &(pp->p_spools[j]);

    for (i=0; i<sp->sp_dmepdevs; i++) {
      dmepdevp = &(sp->sp_dmepdevlist[i]);
      dev = dmepdevp->dm_dev;
      
      error = bdopen(dev, &dmepdevp->inode, &dmepdevp->file);
      
      if (error) {
	osi_printf("\tpool: open failed: error %d opening %d,%d\n",
	       error, MAJOR(dev), MINOR(dev));
	i--;
	goto lock_backout;
      }
    }
  }


  EXIT_PRINT(POOL_1, "open_devices()");

  return (0);



 lock_backout:
  for (; j >= beginning; j--) {
    sp = &(pp->p_spools[j]);
    for (i = i > 0 ? i : sp->sp_dmepdevs - 1; i >= 0; i--) {
      dmepdevp = &(sp->sp_dmepdevlist[i]);
      dev = dmepdevp->dm_dev;
      bdclose(dev, &dmepdevp->inode, &dmepdevp->file);
    }
  }
  j = pp->p_num_spools - 1;
  i = 0;

 disk_backout:
  for (; j >= beginning; j--) {
    sp = &(pp->p_spools[j]);
    for (i = i > 0 ? i : sp->sp_devs - 1; i >= 0; i--) {
      devp = &(sp->sp_devlist[i]);
      dev = devp->d_dev;
      bdclose(dev, &devp->inode, &devp->file);
    }
  }


  EXIT_PRINT(POOL_1, "open_devices()");

  return (error);
}


/**
 * close_devices - close all device pool has open
 * @pp: pointer to the pool structure
 *
 * Returns: Linux errno
 */

int close_devices(pool_t *pp) 
{
  subpool_t           	*sp;
  dmepdev_t		*dmepdevp;
  device_t           	*devp;
  osi_dev_t 		dev;
  int                 	i,j;
  int			error = 0;


  ENTER_PRINT(POOL_1,"close_devices()");

  /* Close all storage and dmep devices */

  for (j=0; j<pp->p_num_spools; j++) {

    sp = &(pp->p_spools[j]);

    for (i=0; i<sp->sp_devs; i++) {

      devp = &(sp->sp_devlist[i]);
      dev = devp->d_dev;

      fsync_dev(dev);

      error = bdclose(dev, &devp->inode, &devp->file);

      if (error)
	osi_printf("\tclose_devices: error %d\n", error);
    }

    for (i=0; i<sp->sp_dmepdevs; i++) {
     
      dmepdevp = &(sp->sp_dmepdevlist[i]);
      dev = dmepdevp->dm_dev;

      error = bdclose(dev, &dmepdevp->inode, &dmepdevp->file);
      if (error)
	osi_printf("\tclose_devices: error %d\n", error);
    }
  }

  EXIT_PRINT(POOL_1, "close_devices()");

  return (0);
}


#ifdef DEBUG
/**
 * print_pool - print labels for debugging
 * @pp: pointer to the pool structure
 *
 * Returns: 0
 */
/* not needed right now...
static int print_pool(pool_t *pp)
{
  subpool_t           	*spoolp;
  device_t            	*devp;
  int                 	i,j;

  osi_printf("\nPool:\n");

  osi_printf("name		 = %s\n", pp->p_name);
  osi_printf("id             = %"FMT_64"x\n", pp->p_pool_id);
  osi_printf("version     	 = v%u.%u.%u\n", pp->p_version >> 16,
	 (pp->p_version >> 8) & 0xFF, pp->p_version & 0xFF);
  osi_printf("subpools   	 = %u\n", pp->p_num_spools);
  osi_printf("capacity   	 = %"FMT_64"u kB\n", pp->p_capacity/2048);
  osi_printf("oblk        	 = %u\n", pp->p_oblk);
  osi_printf("flags       	 = 0x%x\n", pp->p_flags);
  osi_printf("\n");

  for (j=0; j < pp->p_num_spools; j++) {
    spoolp = &(pp->p_spools[j]);

    osi_printf("Subpool %d:\n", j);
    osi_printf("    sp_start     = %"FMT_64"u\n", spoolp->sp_start);
    osi_printf("    sp_len       = %u\n", spoolp->sp_len);
    osi_printf("    sp_stripe    = %u\n", spoolp->sp_stripe);
    osi_printf("    sp_devs      = %u\n", spoolp->sp_devs);
    osi_printf("    sp_weight    = %u\n", spoolp->sp_tot_weight);
    osi_printf("\n");

    for (i=0; i < spoolp->sp_devs; i++) {
      devp = &(spoolp->sp_devlist[i]);
      osi_printf("    Device %d:\n", i);
      osi_printf("        d_start = %"FMT_64"u\n",devp->d_start);
      osi_printf("        d_nblks = %"FMT_64"u\n",devp->d_nblks);
    }
    osi_printf("\n");
  }

  return(0);
}
*/
#endif /* DEBUG */


/**
 * gen_pool - generate pool_t structure from an pool_add_t structure
 * @add: the structure containing info on pool to be made
 * @pp: the pool_t structure to be filled
 * @from_kern: TRUE if assemblly began in kernel, FALSE if started w/ passemble
 *
 * This function takes the information in the pool_add_t structure, which
 * is basically pool information which was read from pool labels on the
 * disk, reformatts it, and copies it into the pool_t structure.
 * The pool_add_t structure can be generated in the kernel when auto-assemble
 * is used (in which case from_kern should equal 1).  It can also be
 * generated in user-space, probably done by the passemble program (in
 * which case from_kern should equal 0).  See also: uadd_pool, kadd_pool,
 * ugrow_pool, and kgrow_pool.
 *
 * Returns: Linux errno
 */

static int gen_pool(pool_add_t *add, pool_t *pp, int from_kern)
{
  subpool_t		*spools_cpy = NULL;
  subpool_t		*sp = NULL;
  device_t    		*nd = NULL;
  dmepdev_t		*nl = NULL;
  int			i,j,k;
  int			size;
  int			error = 0;


  ENTER_PRINT(POOL_1,"gen_pool()");
  pp->p_blksize = 0; /* this is obiviously a special case */

  /* Perform a few simple tests.  Most error checking should be done by 
   * passemble.  Here we just check the pool version.
   * The on-disk pool label must match down to the subrevison.  Thus, versions
   * 0.3.0 and 0.4.0 don't match and are assumed to have different on disk
   * structures.   On the other hand opening a 0.4.2 label with a 0.4.1 or
   * 0.4.5 kernel is not an error */


  if ((add->pa_version >> 16) != (POOL_VERSION >> 16) || 
      ((add->pa_version >> 8) & 0xFF) != ((POOL_VERSION >> 8) & 0xFF)) {
    osi_printf(" ----------> %x <-----------\n", add->pa_version);
    osi_printf("Version mismatch in pool: %s:\n", add->pa_name);
    osi_printf("\ton disk: v%d.%d.%d\n", add->pa_version >> 16,
	   (add->pa_version >> 8) & 0xFF, add->pa_version & 0xFF);
    osi_printf("\tin core: v%d.%d.%d\n", POOL_VERSION >> 16,
	   (POOL_VERSION >> 8) & 0xFF, POOL_VERSION & 0xFF);
    error = -EINVAL;
    goto fail;
  }

  size = add->pa_num_spools * sizeof(subpool_t);

  spools_cpy = (subpool_t *)osi_malloc(size);

  if(from_kern){
    osi_memcpy(spools_cpy, add->pa_spools, size);
  }else{
    if (osi_copy_from_user(spools_cpy, add->pa_spools, size)) {	
      osi_free(spools_cpy, size);
      spools_cpy = NULL;
      error = -EFAULT;
      goto fail;
    }
  }



  /* fill in the devices that constitute the subpools */
  for (k=0; k < add->pa_num_spools; k++) {
    dmepdev_t	*user_copy;

    sp = &(spools_cpy[k]);

    user_copy = sp->sp_dmepdevlist;

    size = sp->sp_dmepdevs * sizeof(dmepdev_t);
    sp->sp_dmepdevlist = (dmepdev_t *)osi_malloc(size);

    if(from_kern){
      if((osi_memcpy(sp->sp_dmepdevlist,user_copy,size)) == NULL){
	goto dmep_backout;
      }
    } else {
      if (osi_copy_from_user(sp->sp_dmepdevlist, user_copy, size)) {
	goto dmep_backout;
      }
    }
    
    for (i=0; i < sp->sp_dmepdevs; i++) {
      nl = &(sp->sp_dmepdevlist[i]);
      if(from_kern){
	nl->dm_dev = nl->dm_adev;
      } else {
	nl->dm_dev = osi_u2k_dev(nl->dm_adev);
      }
    }
  }



  for (j=0; j < add->pa_num_spools; j++) {
    device_t	*user_copy;

    sp = &(spools_cpy[j]);

    user_copy = sp->sp_devlist;

    size = sp->sp_devs * sizeof(device_t);
    sp->sp_devlist = (device_t *)osi_malloc(size);
    if(from_kern){
      osi_memcpy(sp->sp_devlist, user_copy, size);
    } else {
      if (osi_copy_from_user(sp->sp_devlist, user_copy, size)) {
	goto disk_backout;
      }
    }

    for (i=0; i < sp->sp_devs; i++) {
      nd = &(sp->sp_devlist[i]);
      if(from_kern){
	nd->d_dev = nd->d_adev;
      } else {
	nd->d_dev = to_kdev_t(nd->d_adev);
      }
      if(!pp->p_blksize){ /* set the hardsector size if we don't know it */
	pp->p_blksize = get_device_sector_size(nd->d_dev);
      } else if(pp->p_blksize != get_device_sector_size(nd->d_dev)){
	osi_printf("ERROR :: All devices in a given pool"
		   " must have the same block size.\n");
	error = -EINVAL;
      }
    }

    if ((sp->sp_stripe<<BBSHIFT) % (1<<BBSHIFT) != 0) {
      osi_printf("\tpool: stripe width must be a mutliple");
      osi_printf(" of the basic block\n");
      goto disk_backout;
    }
  }


  pp->p_pool_id = add->pa_pool_id;
  pp->p_version = add->pa_version;
  osi_memset(pp->p_name, 0, POOL_NAME_SIZE);
  osi_sprintf(pp->p_name, "%s", add->pa_name);
  pp->p_capacity = add->pa_capacity;
  pp->p_num_spools = add->pa_num_spools;
  pp->p_spools = spools_cpy;
  pp->p_oblk = 0;
  
  EXIT_PRINT(POOL_1, "gen_pool()\n");


  return error;

   
  /* back out the alloc's if there was an error */

 disk_backout:

    for (; j >= 0; j--) {
      sp = &(spools_cpy[j]);
      osi_free(sp->sp_devlist, sp->sp_devs * sizeof(device_t));
      sp->sp_devlist = NULL;
    }


 dmep_backout:

    for (; k >= 0; k--) {
      sp = &(spools_cpy[k]);
      osi_free(sp->sp_dmepdevlist,
	       sp->sp_dmepdevs * sizeof(dmepdev_t));
      sp->sp_dmepdevlist = NULL;
    }

    osi_free(spools_cpy, add->pa_num_spools * sizeof(subpool_t));
    spools_cpy = NULL;

    error = -EFAULT;

 fail:

    EXIT_PRINT(POOL_1, "gen_pool()\n");

    return error;
}



/**
 * add_pool - make the kernel aware of a new pool
 * @add: ptr to struct containing new pool info
 * @from_kern: whether info was assembled in user or kernel space
 *
 * This function takes the pool configuration infomation in 'add'
 * and constructs a pool_t structure from that.  It then finds an
 * appropriate minor number and places the structure in the 
 * global pool array 'pools'.
 *
 * Returns: Linux errno
 */

static int add_pool(pool_add_t *add, int from_kern)
{
  pool_t	*new_pp = NULL;
  int  		 i;
  int		 error;


  ENTER_PRINT(POOL_1,"add_pool()");


  /* see if this pool has already been added */

  for (i=1; i < POOL_MAX_DEVICES; i++) { 

    pool_rw_sema_read_lock(&(pools[i].p_rwsema));

    if (pools[i].p_flags & PF_FREE) {
      pool_rw_sema_read_unlock(&(pools[i].p_rwsema));
      continue;
    }

    if (add->pa_pool_id == pools[i].p_pool_id) {
      if (pools[i].p_flags & PF_VALID) {
	pool_rw_sema_read_unlock(&(pools[i].p_rwsema));
	add->pa_pool_dev = osi_make_dev(POOL_MAJOR,i);

	EXIT_PRINT(POOL_1, "add_pool()\n");

	return 0;
      }
      else {
	pool_rw_sema_read_unlock(&(pools[i].p_rwsema));
	add->pa_pool_dev = osi_make_dev(POOL_MAJOR,i);
	error = -EBUSY;
	goto fail;
      }
    }
    else
      pool_rw_sema_read_unlock(&(pools[i].p_rwsema));
  }

  /* the pool is not already configured so assign a minor number to it */

  for (i=1; i < POOL_MAX_DEVICES; i++) {

    pool_rw_sema_write_lock(&(pools[i].p_rwsema));

    if (pools[i].p_flags & PF_FREE && pools[i].p_flags & PF_VALID) {

      pools[i].p_flags = PF_INIT; /* clear valid and free flags */
      ASSERT(!POOL_BUSY(&pools[i]));

      pool_rw_sema_write_unlock(&(pools[i].p_rwsema));
      break;
    } else {
      pool_rw_sema_write_unlock(&(pools[i].p_rwsema));
    }
  }
  

  if (i == POOL_MAX_DEVICES) {
    error = -ENOMEM;
    goto fail;
  }


  new_pp = &pools[i];

  /* clear out all the old stats */

  osi_memset(&new_pp->p_stats, 0, sizeof(pool_stat_t));

  add->pa_pool_dev = osi_k2u_dev(osi_make_dev(POOL_MAJOR, i));

  error = gen_pool(add, new_pp, from_kern);

  if (error != 0) {
    pool_rw_sema_write_lock(&(pools[i].p_rwsema));
    new_pp->p_flags |= PF_VALID | PF_FREE;
    pool_rw_sema_write_unlock(&(pools[i].p_rwsema));
    goto fail;
  }

  error = open_devices(new_pp, 0);

  if (error != 0) {
    pool_rw_sema_write_lock(&(pools[i].p_rwsema));
    new_pp->p_flags |= PF_VALID | PF_FREE;
    pool_rw_sema_write_unlock(&(pools[i].p_rwsema));
    goto fail;
  }

  pool_rw_sema_write_lock(&(pools[i].p_rwsema));
  new_pp->p_flags |= PF_VALID | PF_LOWEROPEN;
  register_pool(new_pp, i);
  pool_size[i] = new_pp->p_capacity;
  pool_rw_sema_write_unlock(&(pools[i].p_rwsema));

  EXIT_PRINT(POOL_1, "add_pool()");
  return (0);


 fail:
  
  EXIT_PRINT(POOL_1, "add_pool()");
  return error;
}


/**
 * kadd_pool - calls add_pool with from_kern = 1
 */
int kadd_pool(pool_add_t *add){
  return add_pool(add, TRUE);
}


/**
 * uadd_pool - calls add_pool with from_kern = 0
 */
int uadd_pool(pool_add_t *add){
  return add_pool(add, FALSE);
}


/**
 * free_pool - free memory associated with the pool_t
 * @pool: ptr to the pool_t struct
 *
 * free memory associated with the pool_t, leaving you with an
 * emptied pool_t structure.  The pool_t struct itself is not freed
 *
 * Returns: 0
 */

int free_pool(pool_t *pool)
{ 
  subpool_t           	*spoolp;
  int                 	j;


  ENTER_PRINT(POOL_1,"free_pool()");

  for (j=0; j < pool->p_num_spools; j++) {
    spoolp = &(pool->p_spools[j]);
    osi_free(spoolp->sp_dmepdevlist,
	     spoolp->sp_dmepdevs * sizeof(dmepdev_t));
    osi_free(spoolp->sp_devlist, spoolp->sp_devs * sizeof(device_t));
  }
  osi_free(pool->p_spools, pool->p_num_spools * sizeof(subpool_t));
  pool->p_num_spools = 0;
  pool->p_spools = NULL;


  EXIT_PRINT(POOL_1, "free_pool()");

  return (0);
}


/**
 * remove_pool - removes pool from list of pools in the kernel
 * @minor: minor # of the pool to be removed
 *
 * This routine checks that a pool is not in use and then removes it from
 * the list of managed pools.
 *
 * Returns: Linux errno
 */
int remove_pool(uint32_t minor)
{
  pool_t	*pp;
  int		error;

  ENTER_PRINT(POOL_1,"remove_pool()");

  pp = &pools[minor];

  pool_rw_sema_write_lock(&(pp->p_rwsema));
  if (pp->p_flags & PF_FREE) {
    error = -EINVAL;
    goto fail;
  }
  
  if (!(pp->p_flags & PF_VALID) || POOL_BUSY(pp) || (pp->p_oblk == -1)) {
    error = -EBUSY;
    goto fail;
  }
  pp->p_oblk = -1;
  pool_rw_sema_write_unlock(&(pp->p_rwsema));

  /* No need to worry about not holding the lock here because **
  ** this pool can not be opened while p_oblk is -1.......... */
  fsync_dev(osi_make_dev(POOL_MAJOR, minor));

  pool_rw_sema_write_lock(&(pp->p_rwsema));
  if (pp->p_flags & PF_LOWEROPEN){
    close_devices(pp);
  }

  free_pool(pp);

  pp->p_flags = PF_VALID | PF_FREE;

  unregister_pool(minor);
  pool_rw_sema_write_unlock(&(pp->p_rwsema));

  EXIT_PRINT(POOL_1, "remove_pool()");

  return (0);

 fail:

  pool_rw_sema_write_unlock(&(pp->p_rwsema));
  EXIT_PRINT(POOL_1, "remove_pool()");

  return (error);
}

/**
 * grow_pool - like assemble, only grows an existing pool online
 * @add: the pool to be grown
 * @from_kern: tells function if request comes from kernel or user space
 *
 * Very similar to add_pool() in most respects.  However, a new pool
 * minor number is not assigned.  If the pool label matches what is 
 * already loaded this routine returns the minor number already assigned
 * to the pool. It is returned in the
 * pa_pool_dev field of the pool_add_t structure which is copied back out
 * to user space by poolioctl upon return from this function.  If the
 * label is not already loaded, grow_pool finds the right pool_t structure
 * in the pools array and grows the pool.  It then reads in the
 * pool information and returns the minor number of the original pool (which
 * has now been grown).
 * See the pool_t definition in pool_common.h for a description of
 * the locking protocol for the pool structure.
 * 	
 * Returns: Linux errno
 */

static int grow_pool(pool_add_t *add, int from_kern){
  pool_t	*new_pp = NULL;
  pool_t	*cpy_pp = NULL;
  int  		 i,j;
  int		 error;

  ENTER_PRINT(POOL_1,"grow_pool()");
  
  /* see if this pool has already been added */

  for (i=1; i < POOL_MAX_DEVICES; i++) { 

    pool_rw_sema_read_lock(&(pools[i].p_rwsema));

    if (pools[i].p_flags & PF_FREE) {
      pool_rw_sema_read_unlock(&(pools[i].p_rwsema));
      continue;
    }

    if (add->pa_pool_id == pools[i].p_pool_id) {
      if (pools[i].p_flags & PF_VALID) {
	pool_rw_sema_read_unlock(&(pools[i].p_rwsema));
	add->pa_pool_dev = osi_make_dev(POOL_MAJOR,i);
	osi_printf("Target pool found :: attempting to grow\n");
	new_pp = &pools[i];
      }
      else {
	pool_rw_sema_read_unlock(&(pools[i].p_rwsema));
	add->pa_pool_dev = osi_make_dev(POOL_MAJOR,i);
	error = -EBUSY;
	goto fail;
      }
    }
    else
      pool_rw_sema_read_unlock(&(pools[i].p_rwsema));
  }

  if(!new_pp){
    osi_printf("Target pool not found.\n");
    error = -ENODEV;
    goto fail;
  }

  osi_printf("  Capacity before:          %"FMT_64"d\n", new_pp->p_capacity);
  cpy_pp = (pool_t *)
    osi_malloc(sizeof(pool_t));

  osi_memcpy(cpy_pp, new_pp, sizeof(pool_t));

  error = gen_pool(add, cpy_pp, from_kern);

  if (error != 0) {
    osi_free(cpy_pp, sizeof(pool_t));
    goto fail;
  }
  
  /* We're not going to open the lower level devices again, **
  ** instead; we're simply going to copy over the file and  **
  ** inode pointers					    */
  for (j=0; j < new_pp->p_num_spools; j++) {
    for (i=0; i < new_pp->p_spools[j].sp_devs; i++) {
      cpy_pp->p_spools[j].sp_devlist[i].inode =
	new_pp->p_spools[j].sp_devlist[i].inode;
      cpy_pp->p_spools[j].sp_devlist[i].file =
	new_pp->p_spools[j].sp_devlist[i].file;
    }
  }

  for (j=0; j < new_pp->p_num_spools; j++) {
    for (i=0; i < new_pp->p_spools[j].sp_dmepdevs; i++) {
      cpy_pp->p_spools[j].sp_dmepdevlist[i].inode =
	new_pp->p_spools[j].sp_dmepdevlist[i].inode;
      cpy_pp->p_spools[j].sp_dmepdevlist[i].file =
	new_pp->p_spools[j].sp_dmepdevlist[i].file;
    }
  }

  /* With this call we open the newly added devices */
  error = open_devices(cpy_pp, new_pp->p_num_spools);

  if (error != 0) {
    free_pool(cpy_pp); /* frees everything except pool_t */
    osi_free(cpy_pp, sizeof(pool_t));
    goto fail;
  }

  pool_rw_sema_write_lock(&(new_pp->p_rwsema));

  if(new_pp->p_capacity == cpy_pp->p_capacity){
    osi_printf("  Pool capacity after the grow has not changed!\n");
    free_pool(cpy_pp); /* frees everything except pool_t */
    osi_free(cpy_pp, sizeof(pool_t));
    error = -EIO;
    goto fail;
  }
    
  new_pp->p_capacity = cpy_pp->p_capacity;
  free_pool(new_pp);
  new_pp->p_num_spools = cpy_pp->p_num_spools;
  new_pp->p_spools = cpy_pp->p_spools;
  osi_free(cpy_pp, sizeof(pool_t));
  new_pp->p_flags |= PF_VALID | PF_LOWEROPEN;

  osi_printf("  Capacity after:           %"FMT_64"d\n", new_pp->p_capacity);
  pool_rw_sema_write_unlock(&(new_pp->p_rwsema));

  EXIT_PRINT(POOL_1, "grow_pool()");

  return (0);


 fail:
  
  EXIT_PRINT(POOL_1, "grow_pool()");
  
  return error;
}


/**
 * kgrow_pool - calls grow_pool with from_kern = 1
 */
int kgrow_pool(pool_add_t *add){
  return grow_pool(add, TRUE);
}


/**
 * ugrow_pool - calls grow_pool with from_kern = 0
 */
int ugrow_pool(pool_add_t *add){
  return grow_pool(add, FALSE);
}


/**
 * assemble_pools - assemble pools in the kernel
 * @poolname: NULL if all pools to be assembled; otherwise, pool to be grown
 *
 * This routine will assemble pools.  It will grow the specified pool
 * if poolname is not NULL.
 *
 * Returns: Linux errno
 */
int assemble_pools(char *poolname){
  int		error = 0;
  int		i, j, k, rtn;
  int		num_devices = 0;
  osi_dev_t	*devices = NULL;
  osi_dev_t	*pool_devices = NULL;
  int		num_labels = 0;
  pool_label_t	*labels = NULL;
  pool_add_t	*np = NULL;
  subpool_t	*ns = NULL;
  dmepdev_t	*nl = NULL;
  device_t     	*nd = NULL;
  uint64	pblk_off = 0;
  uint64	spblk_off = 0;


  ENTER_PRINT(POOL_1,"assemble_pools()");

  num_devices = read_partition_list(&devices);

  if(num_devices <= 0){
    osi_printf("Pool can not assemble without any devices!\n");
    error = -ENOTBLK;
    goto fail1;
  }

  pool_devices = osi_malloc(num_devices * sizeof(osi_dev_t));

  labels = osi_malloc(num_devices * sizeof(pool_label_t));

  for(i = 0; i < num_devices; i++){
    rtn = scan_dev_for_pool_label(devices[i], &(labels[num_labels]));
    if(rtn < 0){
      error = rtn;
      goto fail4;
    } else {
      num_labels += rtn;
      if(rtn){
	/* keep track of osi_dev_t's for later */
	pool_devices[num_labels-1] = devices[i];
      }
    }
  }

  /* Now we've got all of the pool labels.  We must construct add_pool_t's */
  j = num_labels;
  while(j != 0){
    for(i = 0; i < num_labels; i++){
      if((poolname && strcmp(poolname,labels[i].pl_pool_name)) ||
	 (labels[i].pl_magic != POOL_MAGIC)){
	/* if we are growing a pool, and this label isn't part of **
	** it, OR this label has already been looked at, continue */
	continue;
      }
      if(!np){
	np = osi_malloc(sizeof(pool_add_t));
	osi_memset(np, 0, sizeof(pool_add_t));
	np->pa_pool_id = labels[i].pl_pool_id;
	np->pa_version = labels[i].pl_version;
	np->pa_num_spools = labels[i].pl_subpools;
	osi_memcpy(np->pa_name, labels[i].pl_pool_name, POOL_NAME_SIZE);
	np->pa_spools = osi_malloc(np->pa_num_spools * sizeof(subpool_t));
	osi_memset(np->pa_spools, 0, np->pa_num_spools * sizeof(subpool_t));
      }
      if(np->pa_pool_id == labels[i].pl_pool_id){
	ns = &(np->pa_spools[labels[i].pl_sp_id]);
	ns->sp_devs = labels[i].pl_sp_devs;
	ns->sp_dmepdevs = labels[i].pl_sp_dmepdevs;
	ns->sp_type = labels[i].pl_sp_type;
	ns->sp_stripe = labels[i].pl_striping;
	if(ns->sp_devlist == NULL) {
	  ns->sp_devlist = osi_malloc(ns->sp_devs * sizeof(device_t));
	  osi_memset(ns->sp_devlist, 0, ns->sp_devs * sizeof(device_t)); 
	}
	if (ns->sp_dmepdevlist == NULL) {
	  ns->sp_dmepdevlist = 
	    osi_malloc(ns->sp_dmepdevs*sizeof(dmepdev_t));
	  osi_memset(ns->sp_dmepdevlist, 0,
		     ns->sp_dmepdevs * sizeof(dmepdev_t));
	}
	if (labels[i].pl_blocks) {
	  nd = &(ns->sp_devlist[labels[i].pl_sp_devid]);
	  nd->d_nblks = labels[i].pl_blocks;
	  nd->d_adev = pool_devices[i];
	}
	if (labels[i].pl_sp_weight){
	  nl = &(ns->sp_dmepdevlist[labels[i].pl_sp_dmepid]);
	  nl->dm_adev = pool_devices[i];
	}
	labels[i].pl_magic = 0;      /* done with this label */
	j--;
      }
    }

    pblk_off = 0;	/* logical blk offsets - scope is entire pool */
    spblk_off = 0;     	/* logical blk offsets - scope is one subpool */

    for (i=0; i < np->pa_num_spools; i++) {
      ns = &(np->pa_spools[i]);
      ns->sp_start = pblk_off;
      spblk_off = 0;
      if (ns->sp_devs == 0)		/* expected a label here */
	goto fail7;    
      for (j=0; j < ns->sp_devs; j++) {
	nd = &(ns->sp_devlist[j]);

	if (nd->d_nblks == 0)		/* expected a label here */
	  goto fail7;
	nd->d_start = spblk_off;
	spblk_off += nd->d_nblks;
      }
    
      ns->sp_len = spblk_off; /* check on this brassow */
      pblk_off += spblk_off;
    }
    np->pa_capacity = pblk_off;

    if(np && poolname){
      error = kgrow_pool(np);
      /* if this is a grow, it will happen one pool at a time **
      ** we must guard against this pool being reconsidered   */
      j = 0;
    } else {
      error = kadd_pool(np);
    }
    for(k = 0; k < np->pa_num_spools; k++){
      if(!np->pa_spools[i].sp_dmepdevlist){
	osi_free(np->pa_spools[i].sp_dmepdevlist,
		  np->pa_spools[i].sp_dmepdevs*sizeof(dmepdev_t));
      }
      if(!np->pa_spools[i].sp_devlist){
	osi_free(np->pa_spools[i].sp_devlist,
		  np->pa_spools[i].sp_devs*sizeof(device_t));
      }
    }
    osi_free(np->pa_spools,
	      np->pa_num_spools*sizeof(subpool_t));
    osi_free(np, sizeof(pool_add_t));
    if(error < 0){
      goto fail4;
    }
  } /* attention -- watch for infinite loop */

  osi_free(labels, num_devices * sizeof(pool_label_t));
  osi_free(pool_devices, num_devices*sizeof(osi_dev_t));
  osi_free(devices, num_devices*sizeof(osi_dev_t));

  EXIT_PRINT(POOL_1, "assemble_pools()");
  return error;

fail7:  /* failure while adding devices */
  for(k = 0; k < np->pa_num_spools; k++){
    if(!np->pa_spools[i].sp_dmepdevlist){
      osi_free(np->pa_spools[i].sp_dmepdevlist,
		np->pa_spools[i].sp_dmepdevs*sizeof(dmepdev_t));
    }
    if(!np->pa_spools[i].sp_devlist){
      osi_free(np->pa_spools[i].sp_devlist,
		np->pa_spools[i].sp_devs*sizeof(device_t));
    }
  }
  /* fail6: failure before devices are added */
  osi_free(np->pa_spools,
	    np->pa_num_spools*sizeof(subpool_t));
  /* fail5: failure before subpools memory is created */
  osi_free(np, sizeof(pool_add_t));
fail4: /* failure during scan_dev_for_pool_label() */
  osi_free(labels, num_devices * sizeof(pool_label_t));
  /* fail3: */
  osi_free(pool_devices, num_devices*sizeof(osi_dev_t));
  /* fail2: */
  osi_free(devices, num_devices*sizeof(osi_dev_t));
fail1: 
  EXIT_PRINT(POOL_1, "assemble_pools()");
  return error;

}


/**
 * count_pool - count the number of pools registered in the kernel
 * @count: ptr to int that will hold the number of pools in the kernel
 *
 * Returns: Linux errno
 */
int count_pool(int *count){
  int			i,j;
  pool_t		*pp;

  ENTER_PRINT(POOL_1,"count_pool()");

  for (i=1,j=0; i<POOL_MAX_DEVICES; i++) {

    pp = &pools[i];

    pool_rw_sema_read_lock(&(pp->p_rwsema));

    if (pp->p_flags & PF_FREE || !(pp->p_flags & PF_VALID)) {
      pool_rw_sema_read_unlock(&(pp->p_rwsema));
      continue;
    }
    pool_rw_sema_read_unlock(&(pp->p_rwsema));
    j++;
  }

  *count = j;

  EXIT_PRINT(POOL_1, "count_pool()");
  return(0);
}



/**
 * list_pool - gather info on add pools registered in the kernel
 * @list - ptr where the info about the pools will be placed
 *
 * routine runs through the list of in-use pools and returns
 * to the caller.  It provides a mapping between pool devices
 * label devices and provides a versioning system by returning the
 * time at which the pool was last read in.  It also provides
 * about whether the device is in use or not.
 *
 * Returns: Linux errno
 */
int list_pool(pool_list_t *list)
{
  int			 i,j,k;
  pool_status_t		*stats;
  pool_t		*pp;

  ENTER_PRINT(POOL_1,"list_pool()");

  if (list->pl_count > POOL_MAX_DEVICES)
    list->pl_count = POOL_MAX_DEVICES;

  stats = (pool_status_t *) 
    osi_malloc(list->pl_count * sizeof(pool_status_t));
  osi_memset(stats, 0, list->pl_count * sizeof(pool_status_t));

  for (i=1,j=0; i<POOL_MAX_DEVICES; i++) {

    pp = &pools[i];

    pool_rw_sema_read_lock(&(pp->p_rwsema));

    if (pp->p_flags & PF_FREE || !(pp->p_flags & PF_VALID)) {
      pool_rw_sema_read_unlock(&(pp->p_rwsema));
      continue;
    }

    stats[j].ps_pool_major = (uint32)POOL_MAJOR;
    stats[j].ps_pool_minor = (uint32)i;
    stats[j].ps_pool_id = pp->p_pool_id;
    stats[j].ps_capacity = pp->p_capacity;
    stats[j].ps_num_spools = pp->p_num_spools;
    for(k=0; k < POOL_NAME_SIZE; k++)
      stats[j].ps_name[k] = pp->p_name[k]; 
    if (POOL_BUSY(pp))
      stats[j].ps_in_use = 1;
    else 
      stats[j].ps_in_use = 0;

    pool_rw_sema_read_unlock(&(pp->p_rwsema));

    if (++j >= list->pl_count){
      break;
    }
  }

  list->pl_count = j;

  if (!(j > 0)) {
    osi_free(stats, list->pl_count * sizeof(pool_status_t));
    EXIT_PRINT(POOL_1, "list_pool()");
    return (0);
  }

  if (osi_copy_to_user(list->pl_list,stats,
		       (int)(sizeof(pool_status_t) * j))) {
    osi_free(stats, list->pl_count * sizeof(pool_status_t));
    EXIT_PRINT(POOL_1, "list_pool()");
    return -EFAULT;
  }

  osi_free(stats, list->pl_count * sizeof(pool_status_t));
  EXIT_PRINT(POOL_1, "list_pool()");
  return (0);
}


/**
 * list_spool - generate list of subpools for a given pool
 * @pp: the specific pool from which to extract subpool info
 * @param: ptr into which the subpool info will be placed.
 *
 * Generate a list of subpools for a given pool and copy their sizes out
 * to user space.  Note: the space must be available in 'param' to put
 * the information in.
 *
 * Returns: Linux errno
 */
int list_spool(pool_t *pp, pool_splist_t *param)
{
  pool_spstatus_t	*list;
  subpool_t		*spool;
  int			i;
  int			max;
  

  ENTER_PRINT(POOL_1,"list_spool()");

  max = pp->p_num_spools > param->spl_count ? param->spl_count :
    pp->p_num_spools;

  list = osi_malloc(max * sizeof(pool_spstatus_t));
  osi_memset(list, 0, max * sizeof(pool_spstatus_t));
  
  for (i=0; i < max; i++) {
    spool = &pp->p_spools[i];
      
    list[i].sps_start = spool->sp_start;
    list[i].sps_len = spool->sp_len;
    list[i].sps_stripe = spool->sp_stripe;
    list[i].sps_tot_weight = spool->sp_tot_weight;
    list[i].sps_numdevs = spool->sp_devs;
    /*    list[i].sps_dmepdevs = spool->sp_dmepdevs; */
    list[i].sps_type = spool->sp_type;
  }

  if (osi_copy_to_user(param->spl_list,list,
		       (int)(sizeof(pool_spstatus_t) * max))) {
    osi_free(list, sizeof(pool_spstatus_t) * max);
    EXIT_PRINT(POOL_1, "list_spool()");
    return -EFAULT;
  }

  param->spl_count = max;

  osi_free(list, sizeof(pool_spstatus_t) * max);
  EXIT_PRINT(POOL_1, "list_spool()");
  return (0);
}


/**
 * list_pooldevs - generate device info for all devices in a given pool
 * @pp: the pool from which to extract the device information
 * @param: ptr to where the device info will be place
 *
 * Generate a list of devices for a given pool including various information
 * specific to each device.
 *
 * Returns: Linux errno
 */
int list_pooldevs(pool_t *pp, pool_devlist_t *param)
{
  int			error = 0;
  int			i,j,k,l;
  subpool_t		*spool;
  pool_devstatus_t	*list;
  
  ENTER_PRINT(POOL_1,"list_pooldevs()");

  list = osi_malloc(param->pdl_count * sizeof(pool_devstatus_t));
  osi_memset(list, 0, param->pdl_count * sizeof(pool_devstatus_t));
  
  for (i=0,k=0; (k < param->pdl_count) && (i < pp->p_num_spools); i++) {
    spool = &pp->p_spools[i];
    for(j = 0; j < spool->sp_devs; j++, k++){
      list[k].pds_dev = osi_k2u_dev(spool->sp_devlist[j].d_adev);
      list[k].pds_blocks = spool->sp_devlist[j].d_nblks;
    }
  }

  /* Three for loops sucks.  There has to be a better way to do this */
  for (i=0; i < pp->p_num_spools; i++) {
    spool = &pp->p_spools[i];
    for(j = 0; j < spool->sp_dmepdevs; j++){
      for(l=0; l < param->pdl_count; l++){
        if(list[l].pds_dev == osi_k2u_dev(spool->sp_dmepdevlist[j].dm_adev)){
	  list[l].pds_my_weight = spool->sp_dmepdevlist[j].dm_my_weight;
        }
      }
    }
  }

  if (osi_copy_to_user(param->pdl_list,list,
	      (int)(sizeof(pool_devstatus_t) * param->pdl_count))) {
    error = -EFAULT;
  }

  osi_free(list, sizeof(pool_devstatus_t) * param->pdl_count);
  EXIT_PRINT(POOL_1, "list_pooldevs()");
  return error;
}


/**
 * pool_blkmap - maps logical storage view to a real storage device and blk
 * @pp: pool in which blk mapping is being requested
 * @psector: logical pool sector
 * @pbcount: logical pool blk count
 * @dev_sector: real device sector(s)
 * @dev_bcount: real device blk count(s)
 * @out_dev: real device(s)
 *
 * This is where the magic happens.  This function will take the logical
 * pool sector and blk count and convert it to specific devices' sectors
 * and blk counts.
 *
 * Returns: Linux errno
 */
int pool_blkmap(pool_t *pp, uint64 psector, uint32 pbcount, 
		uint64 *dev_sector, uint32 *dev_bcount, device_t **out_dev)
{
  subpool_t           	*sp = NULL;
  device_t            	*devp = NULL;
  int			i,j;
  uint32		current_spool = 0;
  uint32		current_dev;
  uint64		blk;
  uint64		sp_offset;
  uint32		devnum = 0;    /* # devices in selected spool */
  int			error;
  
  ENTER_PRINT(MAPPING_1, "pool_blkmap()");

  /* lock is to prevent pool resizing during mapping */
  pool_rw_sema_read_lock(&(pp->p_rwsema));

  /* select subpool */

  for (j=current_spool; j<pp->p_num_spools; j++) {
    sp = &(pp->p_spools[j]);

    if ((psector >= sp->sp_start) &&
	(psector < sp->sp_start+sp->sp_len)) {
      break;
    }
  }

  if (j == pp->p_num_spools) {
    osi_printf("block not in any subpools at line %d!\n", __LINE__);
    error = -EIO;
    goto fail;
  }
  
  current_spool = j;
  devnum = sp->sp_devs;

  /* Determine device within selected sub-pool */

  if (sp->sp_stripe) {

    current_dev = (uint32)(((psector - sp->sp_start) 
			    / sp->sp_stripe) % devnum);
  } 
  else {
    for (i=0; i<devnum; i++) {
      devp = &(sp->sp_devlist[i]);
      
      if (((psector - sp->sp_start) >= devp->d_start) &&
	  ((psector - sp->sp_start) < (devp->d_start + devp->d_nblks)))
	break;
    }

    if (i == devnum) {
      osi_printf("\nblock requested past end of subpool at line %d!\n",
		 __LINE__);
      error = -EIO;
      goto fail;
    }

    current_dev = i;
  }

  devp = &(sp->sp_devlist[current_dev]);


  /* Determine the block number on the device */
  if (sp->sp_stripe) {
    sp_offset = (psector - sp->sp_start);
    blk = ((sp_offset / (sp->sp_stripe*sp->sp_devs)) * 
	  sp->sp_stripe) + (sp_offset % sp->sp_stripe);
  } 
  else {
    blk = psector - sp->sp_start - devp->d_start;
  }

  /* account for partition header */

  blk += 16;	

  *dev_sector = blk;
  *out_dev = devp;

  pool_rw_sema_read_unlock(&(pp->p_rwsema));

  EXIT_PRINT(MAPPING_1, "pool_blkmap()");
  
  return (0);


 fail:

  pool_rw_sema_read_unlock(&(pp->p_rwsema));

  EXIT_PRINT(MAPPING_1, "pool_blkmap()");

  return error;
}


/**
 * gencopy - a memcpy mechanism
 * @to
 * @from
 * @size
 *
 * This is a memcpy-like function that is indifferent to whether
 * it is copying from user space to kernel space (or any other
 * combination).
 * i386   TASK_SIZE default:  0xC0000000
 * alpha  TASK_SIZE default:  0x40000000000
 *
 * Returns: Linux errno
 */
unsigned long gencopy(void *to, void *from, unsigned long size)
{

  ENTER_PRINT(POOL_1,"gencopy()");
  if ((unsigned long)from <= TASK_SIZE){    			
    if ((unsigned long)to <= TASK_SIZE)
      osi_memcpy(to, from, size);			/* user -> user */
    else{
      EXIT_PRINT(POOL_1, "gencopy()");
      return copy_from_user(to, from, size);		/* user -> kernel */
    }
  }
  else {
    if ((unsigned long)to <= TASK_SIZE){
      EXIT_PRINT(POOL_1, "gencopy()");
      return copy_to_user(to, from, size);		/* kernel -> user */
    }
    else
      osi_memcpy(to, from, size);			/* kernel -> kernel */
  }
  EXIT_PRINT(POOL_1, "gencopy()");
  return 0;
}





