/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portions Copyright 2000-2001 Sistina Software Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Jonathan Brassow <brassow@sistina.com>
**
**    Module name: ondisk.c
**
**    Description:  This module converts all pool structures that exist on
**      disk to internal processor formats.  Handles structure alignment
**      and endianess issues.  All disk structures must be multiples of
**      8 bytes.  See also GFS/src/fs/ondisk.c
**
**
*******************************************************************************
******************************************************************************/

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif  /*  MODVERSIONS  */
#include "pool.h"

#define CPIN_8(x, y, z) {osi_memcpy((x), (y), (z));}
#define CPOUT_8(x, y, z) {osi_memcpy((y), (x), (z));}
#define CPIN_16(x, y) {(x) = osi_be16_to_cpu((y));}
#define CPOUT_16(x, y) {(y) = osi_cpu_to_be16((x));}
#define CPIN_32(x, y) {(x) = osi_be32_to_cpu((y));}
#define CPOUT_32(x, y) {(y) = osi_cpu_to_be32((x));}
#define CPIN_64(x, y) {(x) = osi_be64_to_cpu((y));}
#define CPOUT_64(x, y) {(y) = osi_cpu_to_be64((x));}



/**
 * pool_label_out - copies a pool_label_t into a char buffer
 * @pl: ptr to a pool_label_t struct
 * @buf: ptr to raw space where label info will be copied
 *
 * This function is important because it takes care of all of
 * the endian issues when copying to disk.  This way, when
 * machines of different architectures are used, they will
 * be able to interpret ondisk labels correctly.  Always use
 * this function before writing to disk.
 */
void pool_label_out(pool_label_t *pl, char *buf)
{
  pool_label_t *bufpl = (pool_label_t *)buf;

  CPOUT_64(pl->pl_magic, bufpl->pl_magic);
  CPOUT_64(pl->pl_pool_id, bufpl->pl_pool_id);
  CPOUT_8(pl->pl_pool_name, bufpl->pl_pool_name, POOL_NAME_SIZE);
  CPOUT_32(pl->pl_version, bufpl->pl_version);
  CPOUT_32(pl->pl_subpools, bufpl->pl_subpools);
  CPOUT_32(pl->pl_sp_id, bufpl->pl_sp_id);
  CPOUT_32(pl->pl_sp_devs, bufpl->pl_sp_devs);
  CPOUT_32(pl->pl_sp_devid, bufpl->pl_sp_devid);
  CPOUT_32(pl->pl_sp_type, bufpl->pl_sp_type);
  CPOUT_64(pl->pl_blocks, bufpl->pl_blocks);
  CPOUT_32(pl->pl_striping, bufpl->pl_striping);  
  CPOUT_32(pl->pl_sp_dmepdevs, bufpl->pl_sp_dmepdevs);
  CPOUT_32(pl->pl_sp_dmepid, bufpl->pl_sp_dmepid);
  CPOUT_32(pl->pl_sp_weight, bufpl->pl_sp_weight);
}


/**
 * pool_label_in - copies a char buffer into a pool_label_t
 * @pl: ptr to a pool_label_t struct
 * @buf: ptr to raw space where label info is copied from
 *
 * This function is important because it takes care of all of
 * the endian issues when information from disk is about to be
 * used.  This way, when machines of different architectures
 * are used, they will be able to interpret ondisk labels
 * correctly.  Always use this function before using labels that
 * were read from disk.
 */
void pool_label_in(pool_label_t *pl, char *buf)
{
  pool_label_t *bufpl = (pool_label_t *)buf;

  CPIN_64(pl->pl_magic, bufpl->pl_magic);
  CPIN_64(pl->pl_pool_id, bufpl->pl_pool_id);
  CPIN_8(pl->pl_pool_name, bufpl->pl_pool_name, POOL_NAME_SIZE);
  CPIN_32(pl->pl_version, bufpl->pl_version);
  CPIN_32(pl->pl_subpools, bufpl->pl_subpools);
  CPIN_32(pl->pl_sp_id, bufpl->pl_sp_id);
  CPIN_32(pl->pl_sp_devs, bufpl->pl_sp_devs);
  CPIN_32(pl->pl_sp_devid, bufpl->pl_sp_devid);
  CPIN_32(pl->pl_sp_type, bufpl->pl_sp_type);  
  CPIN_64(pl->pl_blocks, bufpl->pl_blocks);
  CPIN_32(pl->pl_striping, bufpl->pl_striping);  
  CPIN_32(pl->pl_sp_dmepdevs, bufpl->pl_sp_dmepdevs);
  CPIN_32(pl->pl_sp_dmepid, bufpl->pl_sp_dmepid);
  CPIN_32(pl->pl_sp_weight, bufpl->pl_sp_weight);
}

