/******************************************************************************
*******************************************************************************
**
**    Copyright 2000-2001 Sistina Software Inc
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Jonathan Brassow <brassow@sistina.com>
** 
**    Module name: pool_incore.h
**
**    Description:
**      This header file contains everything corresponding to incore
**	memory structures.
**
*******************************************************************************
******************************************************************************/

#ifndef __POOL_INCORE_H__
#define __POOL_INCORE_H__
#include "rw_sema.h"

struct generic_device;
struct device;
struct dmepdev;
struct subpool;
struct pool;
struct pool_add;
struct pool_status;
struct pool_spstatus;
struct pool_list;
struct pool_splist;
struct pool_devstatus;
struct pool_devlist;
struct pool_stat;

typedef struct generic_device generic_device_t;
typedef struct device device_t;
typedef struct dmepdev dmepdev_t;
typedef struct subpool subpool_t;
typedef struct pool pool_t;
typedef struct pool_add pool_add_t;
typedef struct pool_status pool_status_t;
typedef struct pool_spstatus pool_spstatus_t;
typedef struct pool_list pool_list_t;
typedef struct pool_splist pool_splist_t;
typedef struct pool_devstatus pool_devstatus_t;
typedef struct pool_devlist pool_devlist_t;
typedef struct pool_stat pool_stat_t;


struct pool_status
{
  uint32		ps_pool_major;
  uint32		ps_pool_minor;
  uint64		ps_pool_id;
  uint32		ps_in_use;
  uint64		ps_capacity;
  uint32		ps_num_spools;	/* number of subpools in the pool */
  char			ps_name[POOL_NAME_SIZE];
  pool_spstatus_t	*ps_splist;
};



struct pool_list
{
  uint32		pl_count;
  pool_status_t		*pl_list;
};



struct pool_spstatus
{
  uint64		sps_start;
  uint64		sps_len;
  uint32		sps_stripe;
  uint32		sps_type;
  uint32		sps_numdevs;
  uint32		sps_tot_weight;
  pool_devstatus_t	*sps_devlist;
};



struct pool_splist
{
  uint32		spl_count;
  pool_spstatus_t	*spl_list;
};



struct pool_devstatus
{
  uint32		pds_dev;
  uint32		pds_my_weight;
  uint64		pds_blocks;
};


struct pool_devlist
{
  uint32		pdl_count;
  pool_devstatus_t	*pdl_list;
};
  

struct pool_stat
{
  uint32      	s_reads;	/* # reads */
  uint32       	s_readblks; 	/* # blocks reads */
  uint32      	s_writes;       /* # writes */
  uint32       	s_writeblks; 	/* # blocks written */
};

struct device
{
  uint64		d_start;      	/* starting logical blk in spool */
  uint64		d_nblks;	/* number of blocks in device */
  osi_dev_t             d_dev;		/* device number */
  uint32		d_adev;		/* user space device number */
  osi_vnode_t           *inode;
  osi_vfile_t           *file;
};

struct dmepdev
{
  uint32       	dm_my_weight;	/* weight of device  */
  osi_dev_t     dm_dev;		/* device number */
  uint32	dm_adev;	/* user space device number */
  osi_vnode_t 	*inode;
  osi_vfile_t  	*file;
};


struct generic_device
{
  uint64	gd_start;      	/* starting logical blk in spool */
  uint64	gd_nblks;	/* number of blocks in device */
  uint32	gd_my_weight;	/* DMEP weight of device */
  char		*gd_dev_path;	/* device path (/dev/sda) */
};

struct subpool
{
  uint64	sp_start;		/* lba of first block */
  uint32	sp_len;			/* number of blocks in subpool */
  uint32	sp_stripe;		/* stripe size for subpool */
  uint32	sp_type;      		/* type of subpool for fs */
  uint32	sp_devs;      		/* number of devices in subpool */
  device_t	*sp_devlist;		/* list of devices */
  uint32	sp_dmep_redirect; 	/* sp# that does dmep for this sp */
  uint32       	sp_tot_weight;		/* total DMEP wght of the subpool */
  uint32	sp_dmepdevs;    	/* number of dmep devices */
  dmepdev_t	*sp_dmepdevlist;	/* list of dmep devices */  
  generic_device_t	*sp_gdlist;	/* list of generic devices */
};



struct pool
{
  /* read-only values: not protected, except by POOL_VALID 
   * which prevents opening when the structure is invalid
   * (during pool adds, first open, and final close) */

  uint64		p_pool_id;	/* pool id (on-disk) */
  uint32		p_version;	/* pool version */
  uint64		p_capacity;	/* number of blocks in pool */
  uint32		p_blksize;	/* block size of devices */
  uint32		p_num_spools;	/* number of subpools */
  char			p_name[POOL_NAME_SIZE]; /* name of pool */
  subpool_t		*p_spools;	/* subpools */
  pool_rw_sema_t	p_rwsema;	/* read/write semaphore */
  int32			p_oblk;		/* # times opened as block dev */
  uint32	        p_flags;	/* status flags */

  osi_spinlock_t        p_statslock;	/* just for stats */
  pool_stat_t	        p_stats;	/* total io stats */
};


struct pool_add
{
  uint64		pa_pool_id;	/* pool id (on-disk) */
  uint32		pa_version;	/* pool version */
  uint64		pa_capacity;	/* number of blocks in pool */
  uint32		pa_num_spools;	/* number of subpools */
  char			pa_name[POOL_NAME_SIZE]; /* name of pool */
  subpool_t    		*pa_spools;	/* subpool list */
  uint32	       	pa_pool_dev;	/* pool device number (returned) */
};

#endif /* __POOL_INCORE_H__ */
