/******************************************************************************
*******************************************************************************
**
**    Copyright 2000-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland <teigland@sistina.com>
**
**    Module name:
**      main.c
**
**    Description:
**
*******************************************************************************
******************************************************************************/

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif  /*  MODVERSIONS  */
#define __NO_VERSION__
#define EXPORT_SYMTAB
#include <linux/module.h>
#include <linux/version.h>
char kernel_version [] = UTS_RELEASE;
#include <linux/proc_fs.h>
#include <linux/file.h>
#include <linux/miscdevice.h>

#include "osi.h"
#include "stomith_ops.h"
#include "release.cf"


/* Global variables */

osi_sema_t stmd_sema;
osi_sema_t stmd_write_sema;
osi_sema_t stmd_read_sema;
cluster_stomith_t stmd_victim;
cluster_stomith_t stmd_local;
int stmd_result;



/**
 * stomith_start_proc_write - for testing in-kernel stomith code
 *
 * @filp
 * @buf
 * @length
 * @data
 *
 */

int stomith_start_proc_write(struct file *filp, const char *buf, 
			     unsigned long length, void *data)
{
  cluster_stomith_t victim, local;


  if (length != 2 * sizeof(cluster_stomith_t))
  {
    printk("stomith:  stomith_start_proc_write invalid write size\n");
    return(-1);
  }

  memcpy(&victim, buf, sizeof(victim));
  memcpy(&local, buf + sizeof(victim), sizeof(local));

  do_stomith(&victim, &local);

  return(0);
}


static int stomith_open(struct inode *inode, struct file *file)
{
  MOD_INC_USE_COUNT;
  return 0;
}


static int stomith_release(struct inode *inode, struct file *file)
{
  MOD_DEC_USE_COUNT;
  return 0;
}


static ssize_t stomith_read(struct file *file, char *buffer, 
			    size_t count, loff_t *ppos)
{
  int len = sizeof(cluster_stomith_t);
  char outbuf[2*len];


  if (count != 2*len)
  {
    printk("stomith:  incorrect read count %lu, expected %u\n", 
	   (long)count, 2*len);
    return -EIO;
  }

  osi_sema_init_lkd(&stmd_read_sema);
  osi_sema_lock_intr(&stmd_read_sema);

  memcpy(outbuf, &stmd_victim, len);
  memcpy(outbuf + len, &stmd_local, len);

  if (copy_to_user(buffer, outbuf, 2*len))
    return -EFAULT;

  return(2*len);
}


static ssize_t stomith_write(struct file *file, const char *buffer,
			     size_t count, loff_t *ppos)
{
  int rv = 0, result = 0;
  unsigned char c = 0x0;


  stmd_result = -1;


  if (count == sizeof(int))
  {
    /* the return value from stomithd */

    if (copy_from_user(&result, buffer, count))
    {
      rv = -EFAULT;
      goto end;
    }

    /* printk("stomith:  result from stomithd %d\n", result); */

    if (result == 21)
      stmd_result = 0;

    rv = count;
    goto end;
  }

  else if (count == 1)
  {    
    /* stomithd writes one byte here when it starts to cause any 
       outstanding stomith request to fail */

    if (copy_from_user(&c, buffer, 1))
      rv = -EFAULT;
    else if (c == 0xFF)
    {
      /* printk("stomith:  resetting stomith\n"); */
      rv = count;
    }
    else
      rv = -EIO;
    
    goto end;
  }

  else
  {
    /* any other writes are invalid */
    
    printk("stomith:  invalid write size %lu, expected %u\n", 
	   (long)count, (int)sizeof(int));
    rv = -EIO;
  }

 end:
  *ppos += count;
  /* printk("stomith:  exit stomith_write, stmd_result %d\n", stmd_result); */
  osi_sema_unlock(&stmd_write_sema);
  return rv;
}


/* using a second proc entry is a bit of a hack for the user level 
   stomith testing tools */

static struct proc_dir_entry proc_stomith_start = {
  0,
  13, "stomith_start",
  S_IFREG | S_IRUGO,
  1, 0, 0,
  0,
  NULL,
  NULL,
  NULL,
  NULL, NULL, NULL,
  NULL,
  NULL,
  stomith_start_proc_write,
};


static struct file_operations stomith_fops = {
  NULL,           /* seek */
  stomith_read,
  stomith_write,
  NULL,           /* readdir */
  NULL,           /* poll */
  NULL,           /* ioctl */
  NULL,           /* mmap */
  stomith_open,
  NULL,           /* flush */
  stomith_release
};


static struct miscdevice stomithdev = {
  MISC_DYNAMIC_MINOR, "stomith", &stomith_fops
};


/**
 * init_module
 *
 */

int init_module(void)
{
  misc_register(&stomithdev);

  osi_sema_init(&stmd_sema);
  osi_sema_init_lkd(&stmd_read_sema);
  osi_sema_init_lkd(&stmd_write_sema);

  proc_register(&proc_root, &proc_stomith_start);

  printk("Stomith %s (built %s %s) installed\n",
	 GFS_RELEASE_NAME,
	 __DATE__, __TIME__);

  return(0);
}


/**
 * cleanup_module
 *
 */

void cleanup_module(void)
{
  proc_unregister(&proc_root, proc_stomith_start.low_ino);

  misc_deregister(&stomithdev);
}



/*  Symbols  */
EXPORT_SYMBOL(do_stomith);
EXPORT_SYMBOL(stomith_names);



