/******************************************************************************
*******************************************************************************
**
**    Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland <teigland@sistina.com>
** 
**    Module name: stomith.h
**
*******************************************************************************
******************************************************************************/

#ifndef STOMITH_DOT_H
#define STOMITH_DOT_H


#include "global.h"



/*  Magic numbers for specific STOMITH methods  */

#define STM_BRCD_ZONE          (0xDEAD0101)
#define STM_BRCD_PORT          (0xDEAD0102)
#define STM_WTI_NPS            (0xDEAD0201)
#define STM_VACM_EMP           (0xDEAD0401)
#define STM_GNBD               (0xDEAD0501)
#define STM_MEATWARE           (0xDEAD0601)
#define STM_MANUAL             (0xDEAD0602)
#define STM_VXHUB_BYPASS       (0xDEAD0701)
#define STM_APC_MS             (0xDEAD0801)



/*  Match magic numbers with tags used in gfsconf param file  */

struct stomith_names
{
  char    *name;
  uint32  number;
};
typedef struct stomith_names stomith_names_t;


#ifdef WANT_STOMITH_NAMES

const stomith_names_t stomith_names[] =
{
{"brocade_zone",           STM_BRCD_ZONE},
{"brocade_port",           STM_BRCD_PORT},
{"wti_nps",                STM_WTI_NPS},
{"vacm_emp",               STM_VACM_EMP},
{"gnbd_gclient",           STM_GNBD},
{"meatware",               STM_MEATWARE},
{"manual",                 STM_MANUAL},
{"vxhub_bypass",           STM_VXHUB_BYPASS},
{"apc_ms",                 STM_APC_MS},
{"",                       0x0}
};

#else  /*  WANT_STOMITH_NAMES  */

extern const stomith_names_t stomith_names[];

#endif /*  WANT_STOMITH_NAMES  */


/*  Number of seconds the stomith module will wait for stomith daemon  */

#define STMD_TIMEOUT 10


/*
 *  Structures for each stomith type
 */


/*  Maximum length of login names, password, etc. (multiple of 4 bytes)  */

#define STM_STRING_LEN 20


/*  Hard FC zoning via Brocade switch, model: Silkworm 2400 and 2800 with
    zoning capability  */

/*  Two approaches to zoning:

    1. All machines are in one common zone and one common configuration.
       Each machine has its own alias.  In this case, "st_command" is 'zoneremove',
       "st_zonecfg" is the name assigned to the common configuration, "st_comzone"
       is the name given to the common zone, and "st_actname" is each machine's
       alias.  The zoning proceedure goes as follows:
         > zoneremove <comzone>, <actname>
         > cfgenable <zonecfg>

    2. Each machine is placed in its own zone.  All the zones are placed into
       a common configuration.  Machine's aliases are not used.  In this case,
       "st_command" is 'cfgremove', "st_zonecfg" is the name assigned to the 
       common zone, "st_comzone" is not used, and "st_actname" is the name of
       each machine's zone.  The zoning proceedure goes as follows:
         > cfgremove <cfgname>, <actname>
	 > cfgenable <zonecfg>
*/
       
struct stm_brcd_zone_s
{
  uint32              st_ipaddr;                        /* IP address of switch */
  uint32              st_kernel_code;                   /* Use kernel stomith code */
  char                st_login[STM_STRING_LEN];         /* Switch login name */
  char                st_passwd[STM_STRING_LEN];        /* Switch login password */
  char                st_command[STM_STRING_LEN];       /* Command eg zoneremove */
  char                st_cfgname[STM_STRING_LEN];       /* Configuration name */
  char                st_comzone[STM_STRING_LEN];       /* Common zone name */

  char                st_actname[STM_STRING_LEN];       /* Alias or per-node zone */
};
typedef struct stm_brcd_zone_s stm_brcd_zone_t;



/*  Portdisable on a Brocade switch  */

struct stm_brcd_port_s
{
  uint32              st_ipaddr;                        /* IP address of switch */
  uint32              st_kernel_code;                   /* Use kernel stomith code */
  char                st_login[STM_STRING_LEN];         /* Switch login name */
  char                st_passwd[STM_STRING_LEN];        /* Switch login password */

  uint32              st_port;                          /* Port number */
};
typedef struct stm_brcd_port_s stm_brcd_port_t;



/*  Power cycle via WTI Network Power Switch (NPS), model: 115, 230  */

struct stm_wti_nps_s
{
  uint32              st_ipaddr;                        /* IP address of switch */
  uint32              st_kernel_code;                   /* Use kernel stomith code */
  char                st_passwd[STM_STRING_LEN];        /* Login password */

  uint32              st_plug;                          /* Plug number */
};
typedef struct stm_wti_nps_s stm_wti_nps_t;



/*  Power cycle with VACM EMP  */

struct stm_vacm_emp_s
{
  char                st_nexxus[STM_STRING_LEN];  
  char                st_login[STM_STRING_LEN];
  char                st_passwd[STM_STRING_LEN];

  char                st_nodeid[STM_STRING_LEN];
};
typedef struct stm_vacm_emp_s stm_vacm_emp_t;



/*  GNBD server ignores stomithed client  */

struct stm_gnbd_s
{
  uint32              st_ipaddr;
  uint32              st_pad;
};
typedef struct stm_gnbd_s stm_gnbd_t;



/*  Humanoid resets machine and then runs program  */

struct stm_meatware_s
{
  uint32              st_ipaddr;
  uint32              st_pad;
};
typedef struct stm_meatware_s stm_meatware_t;



/*  Turn off the port of the victim node on the Vixel hub  */
 
struct stm_vx_hub_s
{
  uint32              st_ipaddr;                   /* IP address of the Vixel 2100 hub */
  uint32              st_port;                     /* Port number - From 1 to 8 */
  char                st_passwd[STM_STRING_LEN];   /* Login password */
};
typedef struct stm_vx_hub_s stm_vx_hub_t;



/*  Power cycle with APC MasterSwitch  */
 
struct stm_apc_ms_s
{
  uint32              st_ipaddr;                   /* IP address of the masterswitch */
  uint32              st_outlet;                   /* Outlet number */
  char                st_login[STM_STRING_LEN];    /* Login name */
  char                st_passwd[STM_STRING_LEN];   /* Login password */
};
typedef struct stm_apc_ms_s stm_apc_ms_t;




/*  General stomith structure  */

struct cluster_stomith
{
  uint32              cs_magic;         /* Magic number */
  char                cs_name[STM_STRING_LEN];
  union 
  {    
    stm_brcd_zone_t       brcd_zone_s;
    stm_brcd_port_t       brcd_port_s;
    stm_wti_nps_t         wti_nps_s;
    stm_vacm_emp_t        vacm_emp_s;
    stm_gnbd_t            gnbd_s;
    stm_meatware_t        meatware_s;
    stm_vx_hub_t          vx_hub_s;
    stm_apc_ms_t          apc_ms_s;
  } cs_st;
};
typedef struct cluster_stomith cluster_stomith_t;



#endif  /*  STOMITH_DOT_H  */






