/******************************************************************************
*******************************************************************************
**
**    Copyright 2000-2001 Sistina Software Inc.
** 
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**	Jonathan Brassow <brassow@sistina.com>
** 
**    Module name: common.c
**
*******************************************************************************
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/sysmacros.h>
#include <sys/ioctl.h>

#include "global.h"
#include "send_pool.h"
#include "send_sg.h"
#include "pool_ioctl.h"
#include "pool_user.h"
#include "pool_ondisk.h"
#include "pool_incore.h"

#define DEFAULT_DUMP_SIZE 4088
#define SIZE_OF_SENSE_PDF 20
#define SIZE_OF_SELECT_PDF 20
#define SIZE_OF_LOAD_PDF_CONTROL_INFO 24
#define SIZE_OF_DUMP_PDF_CONTROL_INFO 8
#define SIZE_OF_STORE_PDF_CONTROL_INFO 24


int uint64_to_bit64(uint64 num, u_char *str){
  str[0] =  (num >> 56) & 0xFF;
  str[1] =  (num >> 48) & 0xFF;
  str[2] =  (num >> 40) & 0xFF;
  str[3] =  (num >> 32) & 0xFF;
  str[4] =  (num >> 24) & 0xFF;
  str[5] =  (num >> 16) & 0xFF;
  str[6] =  (num >>  8) & 0xFF;
  str[7] =  (num >>  0) & 0xFF;
  return 0;
}

int uint32_to_bit24(uint32 num, u_char *str){
  str[0] = (num >> 16) & 0xFF;
  str[1] = (num >>  8) & 0xFF;
  str[2] = (num >>  0) & 0xFF;
  return 0;
}

void *zalloc(size_t size){
  void *rtn;
  rtn = malloc(size);
  memset(rtn, 0, size);
  return rtn;
}

/*
 *
 * This function will only work on base 10 numbers!
 */
uint64 str_to_uint64(char *str){
  int i;
  uint64 rtn = 0;

  for (i = 0; i < strlen(str); i++){
    rtn *= 10;
    rtn += str[i] - 48;
  }

  return rtn;
}

/*
 *
 */
int get_pool_device_list(int fd, dev_t **devices, 	
			 int *device_count, int *dmep_count){
  int i, j;
  pool_splist_t         splist;
  pool_spstatus_t       *spstat;
  pool_devlist_t        devlist;
  int                   devstatsize;

  splist.spl_count = 10;  /* attention - this is guess */

  spstat = (pool_spstatus_t *)
    malloc(splist.spl_count * sizeof(pool_spstatus_t));
  memset(spstat, 0,  splist.spl_count* sizeof(pool_spstatus_t));
  splist.spl_list = spstat;
  
  if (ioctl(fd, POOL_SPLIST, &splist) < 0){
    fprintf(stderr, "pool communications error: %s\n", strerror(errno));
    exit(-1);
  }

  
  for(j=0, devstatsize=0; j< splist.spl_count; j++){
    devstatsize += splist.spl_list[j].sps_numdevs;
  }
  
  *device_count = devlist.pdl_count = devstatsize;
  devlist.pdl_list = malloc(devstatsize *sizeof(pool_devstatus_t));
  memset(devlist.pdl_list, 0, devlist.pdl_count * sizeof(pool_devstatus_t));
  
  if (ioctl(fd, POOL_DEVLIST, &devlist) < 0){
    fprintf(stderr, "pool communications error: %s\n", strerror(errno));
    exit(-1);
  }

  *devices = zalloc(devstatsize*sizeof(dev_t));

  /* attention - are we sure the device #'s that we have will **
  ** convert well from user spc to kernel space ??  In other  **
  ** words, we must be sure that we are getting one or the    **
  ** other back from the ioctl                                */
  j = 0;

  for(i = 0; i < devstatsize; i++){
    if(devlist.pdl_list[i].pds_my_weight){
      (*devices)[j++] = devlist.pdl_list[i].pds_dev;
    }
  }
  *dmep_count = j;
  for(i = 0; i < devstatsize; i++){
    if(!devlist.pdl_list[i].pds_my_weight){
      (*devices)[j++] = devlist.pdl_list[i].pds_dev;
    }
  }

  return 0;
}


/**
 * build_dmep_command - builds a DMEP CDB and PDF
 * @op_code: The op code (0x85 or 0x89)
 * @action: Subset command for the op code
 * @seg_num: Segment number
 * @bid: BID (# of buffers if DMEP out : Select)
 * @seq_num: Sequence Number (size of a buffer if DMEP out : Select)
 * @pbn: PBN
 * @in_use: Set or reset in use bit
 * @data: Data to be placed in the PDF
 * @data_len: The length of the data placed in the PDF (buffer size)
 * @cdb: build 16B CDB - SPACE IS CREATED
 * @pdf: this is the PDF - SPACE IS CREATED
 * @pdf_len: the length of the pdf - SPACE NOT CREATED
 *
 * It's important to note that if a particular action ignores
 * a specific field, you should indicate that by passing 0 or
 * NULL in that field of this function.
 *
 * Returns: 0 on success, -1 on failure
 */
int build_dmep_command(u_char op_code, u_char action,
		       u_char seg_num, uint64 bid,
		       uint64 seq_num, uint64 pbn,
		       int in_use,
		       u_char *data, uint data_len,
		       u_char **cdb, u_char **pdf,
		       uint *pdf_len){
  (*cdb) = (u_char *)zalloc(16);
  (*cdb)[0] = op_code;
  (*cdb)[1] = action & 0x1F;
  (*cdb)[2] = seg_num;
  (*cdb)[3] = 0x00;  /* bid type - should be an option */
  (*cdb)[15] = 0;

  if(op_code == 0x85){ /* DMEP in */
    switch(action){
    case 0:        /* Load buffer */
      uint64_to_bit64(bid, (*cdb)+4);
      *pdf_len = data_len + SIZE_OF_LOAD_PDF_CONTROL_INFO;
      *pdf = (u_char *)zalloc(*pdf_len);
      uint32_to_bit24(*pdf_len, (*cdb)+12);
      break;
    case 1: /* Dump buffer */
      uint64_to_bit64(pbn, (*cdb)+4);
      *pdf_len = DEFAULT_DUMP_SIZE + SIZE_OF_DUMP_PDF_CONTROL_INFO;
      *pdf = (u_char *)zalloc(*pdf_len);
      uint32_to_bit24(*pdf_len, (*cdb)+12);
      break;
    case 2: /* Sense config */
      *pdf_len = SIZE_OF_SENSE_PDF;
      *pdf = (u_char *)zalloc(*pdf_len);
      uint32_to_bit24(*pdf_len, (*cdb)+12);      
      break;
    default:
      break;
    }
  } else if(op_code == 0x89){  /* DMEP out */
    char *tmp_buf;
    uint32 tmp;
    switch(action){
    case 0: /* Store buffer */
      uint64_to_bit64(bid, (*cdb)+4);
      *pdf_len = data_len + SIZE_OF_STORE_PDF_CONTROL_INFO;
      tmp_buf = *pdf = (u_char *)zalloc(*pdf_len);
      uint32_to_bit24(*pdf_len, (*cdb)+12);
      uint32_to_bit24(*pdf_len, tmp_buf);
      tmp_buf[3] = 0x0;
      tmp_buf[4] = (in_use) ? 0x80 : 0x0;
      uint64_to_bit64(seq_num, tmp_buf+8);
      uint64_to_bit64(pbn, tmp_buf+16);
      strncpy(tmp_buf+24, data, data_len);
      break;
    case 2: /* Select config */
      /* attention -- is the bid ignored here? */
      *pdf_len = SIZE_OF_SELECT_PDF;
      tmp_buf = *pdf = (u_char *)zalloc(*pdf_len);
      uint32_to_bit24(*pdf_len, (*cdb)+12);
      uint32_to_bit24(*pdf_len, tmp_buf);
      tmp_buf[3] = 0x2;
      uint64_to_bit64(bid, tmp_buf+8);
      tmp = seq_num;
      uint32_to_bit24(tmp, tmp_buf+16);
      break;
    case 3: /* Enable segment */
      *pdf_len = 0;
      uint32_to_bit24(*pdf_len, (*cdb)+12);
      break;
    default:
      break;
    }
  } else {
    return -1;
  }
  return 0;
}


/**
 * send_dmep_command - Send the DMEP cmd to the right device
 * @filename: the device to send the cmd to
 * @bid: BID
 * @cdb: CDB
 * @pdf: PDF, a double ptr so data can be returned
 * @pdf_len: PDF length
 * @phys_dev: see additional comments
 *
 * The phys_dev parameter shall be zero if you are addressing a
 * SG device or a pool device.  However, it will be the device
 * number of an underlying device in a pool if you wish to talk
 * directly to that device through pool.  Note that 'filename'
 * will still be the name of the pool.
 *
 * Returns: 0 on success, -1 on failure
 */
int send_dmep_command(char *filename, uint64 bid, u_char *cdb,
		      u_char **pdf, uint pdf_len, 
		      error_code_t *ec, dev_t phys_dev){
  int fd, error;
  struct stat stat_info;
  

  if((fd = open(filename, O_RDWR)) < 0){
    fprintf(stderr, "Could not open the file, %s\n", filename);
    return -1;
  }

  if(fstat(fd, &stat_info) < 0){
    fprintf(stderr, "Stat on file, %s, failed.\n", filename);
    error = -1;
    goto fail;
  }

  if(major(stat_info.st_rdev) == 121){ /* POOL_MAJOR */
    if(!phys_dev && (cdb[1] != 0)){
      /* A command being issued to pool must be a Load or a Store **
      ** The reason for this is that the other commands are       **
      ** specific to particular devices in the pool.  One way     **
      ** around this might be to issue the other commands to all  **
      ** the DMEP devices in a pool and coalesce the information. **
      ** Even then, I'm not sure if this is what you would truely **
      ** want.                                                    */
      error = -1;
      goto fail;
    }
    error = send_to_pool(fd, bid, cdb, pdf, pdf_len, ec, phys_dev);
  } else {
    if(phys_dev){
      /* phys_dev must be zero when speaking to SG */
      error = -1;
      goto fail;
    }
    error = send_to_sg(fd, bid, cdb, pdf, pdf_len, ec);
  }

 fail:
  close(fd);
  return error;
}

