/******************************************************************************
*******************************************************************************
**
**    Copyright 2000-2001 Sistina Software Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**	Jonathan Brassow <brassow@sistina.com>
** 
**    Module name: dmep_conf.c
**
*******************************************************************************
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <getopt.h>

#include "global.h"
#include "common.h"

#define DEFAULT_NUMBER_OF_BUFFERS 150000
#define DEFAULT_SEGMENT_NUMBER 0

void print_usage(void);


int main(int argc, char *argv[]){
  int i;
  int fd;
  int destroy = 0;
  int c;
  char *filename;
  dev_t *devices;
  int device_count;
  int dmep_count;
  u_char seg_num = DEFAULT_SEGMENT_NUMBER;
  uint64 number_of_buffers = DEFAULT_NUMBER_OF_BUFFERS;
  uint64 buffer_size = DEFAULT_BUFFER_SIZE;
  uint64 memory_footprint = 0;
  uint64 created_buffers = 0;
  uint64 tmp;
  uint pdf_len;
  u_char *pdf, *cdb;
  struct stat stat_info;
  struct option long_options[] =
  {
    {"destroy", 1, NULL, 1}
  };

  if(argc < 2){
    print_usage();
    exit(EXIT_FAILURE);
  }

  filename = argv[argc-1];

  while ((c = getopt_long(argc, argv, "Sbnm",
                          long_options, NULL)) != -1) {
    switch(c){
    case 1:
      destroy = 1;
      break;
    case 'S':
      tmp = str_to_uint64(argv[optind]);
      if(tmp > 255){
	fprintf(stderr, "Segment number is out of range (max = 256)\n");
	exit(EXIT_FAILURE);
      }
      seg_num = tmp;
      break;
    case 'b':
      buffer_size = str_to_uint64(argv[optind]);
      break;
    case 'n':
      number_of_buffers = str_to_uint64(argv[optind]);
      break;
    case 'm':
      memory_footprint = str_to_uint64(argv[optind]);
      break;
    default:
      exit(EXIT_FAILURE);
      break;
    }
  }

  if(memory_footprint){
    number_of_buffers = memory_footprint / buffer_size;
  }

  if(destroy){
    buffer_size = 0;
    number_of_buffers = 0;
  } else if(buffer_size == 0 || number_of_buffers == 0){
    fprintf(stderr, "Invalid Buffer size or count.\n");
    exit(EXIT_FAILURE);
  }

  if((fd = open(filename, O_RDONLY)) < 0){
    fprintf(stderr, "Error openning file, %s\n", filename);
    exit(EXIT_FAILURE);
  }


  if(fstat(fd, &stat_info) < 0){
    fprintf(stderr, "Stat on file, %s, failed.\n", filename);
    exit(EXIT_FAILURE);
  }

  if(major(stat_info.st_rdev) == 121){ /* POOL_MAJOR */
    get_pool_device_list(fd, &devices, &device_count, &dmep_count);
    close(fd);

    if(!dmep_count){
      printf("This pool has no DMEP devices specified.\n");
      /* attention - ask if they want to proceed anyway ? */
      /* if yes, then set dmep_count to device_count */
      exit(EXIT_FAILURE);
    }

    /* attention - must keep track of how many were successfully configed */
    for(i = 0; i < dmep_count; i++){
      /* build the command for Select */
      if(0 > build_dmep_command(0x89, 0x2, seg_num, number_of_buffers,
				buffer_size, 0, 0, NULL, 0, &cdb,
				&pdf, &pdf_len)){
	fprintf(stderr, "Could not build DMEP command.\n");
	exit(EXIT_FAILURE);
      }
      /* Send the command for Select */
      if(0 > send_dmep_command(filename, 0, cdb, &pdf,
			       pdf_len, NULL, devices[i])){
	fprintf(stderr, "Configuration failure.\n");
	exit(EXIT_FAILURE);
      }
      if(!destroy){
	/* build the command for Enable */
	if(0 > build_dmep_command(0x89, 0x3, seg_num, 0,
				  0, 0, 0, NULL, 0, &cdb,
				  &pdf, &pdf_len)){
	  fprintf(stderr, "Could not build DMEP command.\n");
	  exit(EXIT_FAILURE);
	}
	/* Send the command for Enable */
	if(0 > send_dmep_command(filename, 0, cdb, &pdf,
				 pdf_len, NULL, devices[i])){
	  fprintf(stderr, "Configuration failure.\n");
	  exit(EXIT_FAILURE);
	}
      }
      created_buffers += number_of_buffers;
    }
  } else {
    close(fd);
    dmep_count = 1;
    if(0 > build_dmep_command(0x89, 0x2, seg_num, number_of_buffers,
			      buffer_size, 0, 0, NULL, 0, &cdb,
			      &pdf, &pdf_len)){
      fprintf(stderr, "Could not build DMEP command.\n");
      exit(EXIT_FAILURE);
    }
    /* Send the command for Select */
    if(0 > send_dmep_command(filename, 0, cdb, &pdf, 
			     pdf_len, NULL, 0)){
      fprintf(stderr, "Configuration failure.\n");
      exit(EXIT_FAILURE);
    }
    if(!destroy){
      /* build the command for Enable */
      if(0 > build_dmep_command(0x89, 0x3, seg_num, 0,
				0, 0, 0, NULL, 0, &cdb,
				&pdf, &pdf_len)){
	fprintf(stderr, "Could not build DMEP command.\n");
	exit(EXIT_FAILURE);
      }
      /* Send the command for Enable */
      if(0 > send_dmep_command(filename, 0, cdb, &pdf,
			       pdf_len, NULL, 0)){
	fprintf(stderr, "Configuration failure.\n");
	exit(EXIT_FAILURE);
      }
    }
    created_buffers += number_of_buffers;
  }

  printf("Configuration complete.\n");
  if(destroy){
    printf("\tSegment destroyed :: %u\n", (uint)seg_num);
    printf("\tDevices affected  :: %u\n", dmep_count);
  } else {
    printf("\tBuffers created  :: %"FMT_64"u\n", created_buffers);
    printf("\tBytes per buffer :: %"FMT_64"u\n", buffer_size);
    printf("\tSegment number   :: %u\n", (uint)seg_num);
  }
  exit(EXIT_SUCCESS);
}


void print_usage(void){
  printf("Usage :: dmep_conf [options] <device>\n"
	 "\t-S <segment_number>\n"
	 "\t-b <buffer_size>\n"
	 "\t-n <number_of_buffers>\n"
	 "\t-m memory_footprint>\n"
	 "\t--destroy\n");
}


