/******************************************************************************
*******************************************************************************
**
**    Copyright 2000-2001 Sistina Software Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**	Jonathan Brassow <brassow@sistina.com>
** 
**    Module name: send_sg.c
**
*******************************************************************************
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <scsi/scsi.h>
#include <scsi/sg.h>
#include <unistd.h>
#include <string.h>


#include "global.h"
#include "pool_lock.h"

#define MIN(x, y) (((x) < (y)) ? (x) : (y))

static int tag = 0x101674;

/**
 * send_to_pool - Sends a DMEP command through sg
 * @fd: file descriptor for an SG device
 * @bid: Buffer ID
 * @cdb: Command Descriptor Block
 * @pdf: The PDF, a double ptr so that data can be returned
 * @pdf_len: The length of the pdf
 *
 * Two methods can be used for accessing a DMEP device.  You can
 * go through pool or sg.  This is the function that take the 
 * latter approach.
 *
 * Returns: Size of pdf (>= 0) on success, -1 on failure
 */
int send_to_sg(int fd, uint64 bid, u_char *cdb,
		 u_char **pdf, uint pdf_len, error_code_t *ec){
  unsigned char *request, *reply;
  struct sg_header *sg_head;
  int value;
  int in_size, out_size;

  /*  Check sizes  */
#ifdef SG_BIG_BUFF
  if (sizeof(struct sg_header) + 16 + pdf_len > SG_BIG_BUFF)
  {
    fprintf(stderr, "scsi_cmd():  too big a transfer\n");
    return(-1);
  }
#else  /*  SG_BIG_BUFF  */
  if (sizeof(struct sg_header) + 16 + pdf_len > 4096)
  {
    fprintf(stderr, "scsi_cmd():  too big a transfer\n");
    return(-1);
  }
#endif  /*  SG_BIG_BUFF  */
 
  if(cdb[0] == 0x85){ /* MEMEXP in */
    out_size = 0;
    in_size = pdf_len;
  } else if(cdb[0] == 0x89){ /* MEMEXP out */
    out_size = pdf_len;
    in_size = 0;
  } else {
    fprintf(stderr, "send_to_sg() :: Not a DMEP command.\n");
    return -1;
  }

  /*  Get memory  */
  if (!(request = malloc(sizeof(struct sg_header) + 16 + out_size))){
    fputs("scsi_cmd():  can't get memory\n", stderr);
    return(-ENOMEM);
  }
  if (!(reply = malloc(sizeof(struct sg_header) + in_size))){
    free(request);
    fputs("scsi_cmd():  can't get memory\n", stderr);
    return(-ENOMEM);
  }
  /*  Fill in request  */

  sg_head = (struct sg_header *)request;
  memset(sg_head, 0, sizeof(struct sg_header));
  
  sg_head->pack_len = 0;
  sg_head->reply_len = sizeof(struct sg_header) + in_size;
  sg_head->pack_id = tag++;
  sg_head->twelve_byte = 0;
  
  memcpy(request + sizeof(struct sg_header), cdb, 16);
  memcpy(request + sizeof(struct sg_header) + 16, *pdf, out_size);
  /*  Write out request  */

  value = write(fd, request, sizeof(struct sg_header) + 16 + out_size);

  if (value < 0)
  {
    free(request);
    free(reply);
    fprintf(stderr, "scsi_cmd():  write returned %s (%d)\n", 
	    strerror(errno), value);
    return(value);
  }

  if (value != sizeof(struct sg_header) + 16 + out_size)
    fprintf(stderr, "scsi_cmd():  Return from write is %d (not %u)\n",
	    value,
	    (unsigned int)(sizeof(struct sg_header) + 16 + out_size));



  /*  Read in reply  */

  value = read(fd, reply, sizeof(struct sg_header) + in_size);

 
  if (value < 0)
  {
    free(request);
    free(reply);
    fprintf(stderr, "scsi_cmd():  read returned %s (%d)\n", 
	    strerror(errno), value);
    return(value);
  }
 
  sg_head = (struct sg_header *)reply;
 
  if (sg_head->result || sg_head->sense_buffer[0])
  {
    free(request);
    free(reply);
    if(ec){
      unsigned char *buf = sg_head->sense_buffer;
      ec->sense_key = buf[2] & 0x0F;
      ec->asc = buf[12];
      ec->ascq = buf[13];
      ec->specific[0] = buf[15];
      ec->specific[1] = 0x0;
      ec->specific[2] = 0x0;
    }
    return(-1);
  }

  if (in_size)
    memcpy(*pdf, reply + sizeof(struct sg_header), 
	   MIN(in_size, value - sizeof(struct sg_header)));


  free(request);
  free(reply);

  return(value - sizeof(struct sg_header));
}
