/*****************************************************************************
 *
 *    Copyright 2000-2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file CONTRIBUTORS for a list of contributors.
 *
 *    This file is maintained by:
 *      Ben Marzinski <marzinski@sistina.com>
 *
 * This is a user level interface for setting up and configuring the
 * gnbd client devices.
 *
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <netdb.h>
#include <errno.h>
#include <dirent.h>
#include <ctype.h>

#include "global.h"
#include "gnbd.h"
#define HELPER_PROGRAM
#include "debug_flags.h"
int message_flag = 1;

#define printm(fmt, args...)\
{\
  if(message_flag != 0) \
    fprintf(stderr, "gclient: " fmt, ##args); \
}



typedef struct devlist_s{
  char name[300];
  struct devlist_s *next;
} devlist_t;

int match(char *name, devlist_t *ptr){
  while (ptr != NULL){
    if (strcmp(name, ptr->name) == 0)
      return 0;
    ptr = ptr->next;
  }
  return -1;
}

devlist_t *getnext(devlist_t *ptr){
  if (ptr == NULL)
    return NULL;
  return (ptr->next);
}


devlist_t *add(char *name, devlist_t *head){
  devlist_t *ptr;
  devlist_t *tmp = (devlist_t *)malloc(sizeof(devlist_t));
  strcpy(tmp->name, name);
  tmp->next = NULL;
  if (head == NULL)
    return tmp;
  ptr = head;
  while(ptr->next != NULL)
    ptr = ptr->next;
  ptr->next = tmp;
  return head;
}

devlist_t *clear(devlist_t *ptr){
  devlist_t *tmp;
  while(ptr != NULL){
    tmp = ptr;
    ptr = ptr->next;
    free(tmp);
  }
  return NULL;
}

int do_device_remove(int comm_fd, int minor_nr){
  gnbd_device_t dev;
  errno = 0;
  dev.name = NULL;
  dev.length = 0;
  dev.action = GNBD_DEV_REMOVE;
  dev.minor_nr = minor_nr;
  if (ioctl(comm_fd, GNBD_DEVICE, &dev) < 0)
    return -1;
  return 0;
}
  

int do_disconnect(int comm_fd, char *dev_name){
  struct stat st;
  gnbd_action_t act;
  char path[300] = "/dev/gnbd/";
  errno = 0;
  memset( &act, 0, sizeof(gnbd_action_t));
  if( stat(strcat(path, dev_name), &st) < 0)
    return -1;
  act.device = minor(st.st_rdev);
  act.action = GNBD_ACT_Disconnect;
  if (ioctl(comm_fd, GNBD_ACTION, &act) < 0)
    return -1;
  return 0;
}

int do_connect(int comm_fd, u16 minor_nr){
  gnbd_action_t act;
  errno = 0;
  memset( &act, 0, sizeof(gnbd_action_t));
  act.device = minor_nr;
  act.action = GNBD_ACT_Connect;
  if (ioctl(comm_fd, GNBD_ACTION, &act) < 0)
    return -1;
  return 0;
}  
  
int do_setipport(int comm_fd, u16 minor_nr, uint16 port,
		 struct in_addr ipaddr){
  gnbd_action_t act;
  errno = 0;
  memset( &act, 0, sizeof(gnbd_action_t));
  act.device = minor_nr;
  act.port = htons(port);
  act.ip = (u32) ipaddr.s_addr;
  act.action = GNBD_ACT_SetIPPort;
  if (ioctl(comm_fd, GNBD_ACTION, &act) < 0){
    perror("ioctl failed");
    return -1;
  }
  return 0;
}

int get_proc_data(char *buf, int len){
  int n;
  int bytes = 0;
  int proc_fd;
  proc_fd = open("/proc/gnbd",O_RDONLY);
  if (proc_fd < 0){
    fprintf(stderr, "could not open /proc/gnbd: %s\n", strerror(errno));
    return -1;
  }
  while( (n = read(proc_fd, buf + bytes, len - bytes)) != 0){
    if (n < 0 && errno != EINTR){
      fprintf(stderr, "list request failed: %s\n", strerror(errno));
      close(proc_fd);
      return -1;
    }
    bytes += n;
  }
  if (bytes < len)
    buf[bytes] = 0;
  else
    buf[len-1] = 0;
  close(proc_fd);
  return bytes;
}

int proc_match(char *name){
  char buf[4096];
  char *nameptr, *bufptr;
  int bytes;
  bytes = get_proc_data(buf, 4096);
  if (bytes < 0)
    return -1;
  bufptr = buf;
  while( (nameptr = strchr(bufptr, '/')) != NULL){
    nameptr++;
    bufptr = strchr(nameptr, '/');
    if (bufptr == NULL){
      perror("/proc/gnbd formatted incorrectly\n");
      return -1;
    }
    *bufptr++ = 0;
    if (strcmp(name, nameptr) == 0)
      return 1;
  }
  return 0;
}

int do_device_create(int comm_fd, char *dev_name){
  int err;
  gnbd_device_t dev;
  errno = 0;
  err = proc_match(dev_name);
  if (err < 0){
    fprintf(stderr, "errored trying proc_match\n");
    return -1;
  }
  if (err == 1){
    errno = EEXIST;
    return -1;
  }
  dev.name = dev_name;
  dev.length = strlen(dev_name) + 1;
  dev.action = GNBD_DEV_CREATE;
  if (ioctl(comm_fd, GNBD_DEVICE, &dev) < 0){
    fprintf(stderr, "errored in ioctl\n");
    return -1;
  }
  return dev.minor_nr;
}

int connect_to_server(char *hostname, uint16 port,
		      struct in_addr *ip_addr){
  int sock_fd;
  struct sockaddr_in server;
  struct hostent *hp;
  sock_fd = socket(AF_INET, SOCK_STREAM, 0);
  if (sock_fd < 0){
    fprintf(stderr, "error creating socket: %s\n", strerror(errno));
    return -1;
  }
  server.sin_family = AF_INET;
  hp = gethostbyname(hostname);
  if (hp == NULL){
    fprintf(stderr, "cannot resolve host ip from name %s: %s", hostname,
	    strerror(errno));
    return -1;
  }
  if (memcpy(&(server.sin_addr.s_addr), hp->h_addr, hp->h_length) == NULL){
    fprintf(stderr, "cannot copy host address: %s\n", strerror(errno));
    return -1;
  }
  if (ip_addr != NULL)
    ip_addr->s_addr = server.sin_addr.s_addr;
  server.sin_port = htons(port);
  if (connect(sock_fd, (struct sockaddr *)&server, sizeof(server)) < 0){
    close(sock_fd);
    fprintf(stderr, "error connecting to server: %s\n", strerror(errno));
    return -1;
  }
  return sock_fd;
}

int stomith(char *hostname)
{
  u32 stomithip;
  int bytes;
  struct hostent *hp;
  int sock_fd;
  devlist_t *listptr, *iplist = NULL;
  char buf[4096];
  char *ipaddr, *bufptr;
  hp = gethostbyname(hostname);
  if(hp == NULL){
    fprintf(stderr, "cannot resolve stomith host ip from name %s: %s\n",
	    hostname, strerror(errno));
    return 1;
  }
  if(memcpy(&stomithip, hp->h_addr, hp->h_length) == NULL){
    fprintf(stderr, "cannot copy host address: %s\n", strerror(errno));
    return 1;
  }
  bufptr = (char *)&stomithip;
  /* fprintf(stderr, "the stomith ip is %d.%d.%d.%d\n", (int)*bufptr,
     (int)*(bufptr+1), (int)*(bufptr+2), (int)*(bufptr+3)); */
  bytes = get_proc_data(buf, 4096);
  if (bytes < 0)
    return 1;
  bufptr = buf;
  while( (ipaddr = strstr(bufptr, "IP :")) != NULL){ 
    ipaddr += 5;
    if( (bufptr = strchr(ipaddr, '\n')) == NULL){
      fprintf(stderr, "malformed /proc/gnbd entry\n");
      goto fail;
    }
    *bufptr++ = 0;
    if (match(ipaddr, iplist) < 0)
      iplist = add(ipaddr, iplist);
  }
  listptr = iplist;
  while (listptr != NULL){
    sock_fd = connect_to_server(listptr->name, (unsigned int)14243, NULL);
    if (sock_fd < 0){
      fprintf(stderr, "cannot connect to the server: %s\n", strerror(errno));
      goto fail;
    }
    write(sock_fd, "stom", 5);
    write(sock_fd, (char *)&stomithip, 4);
    read(sock_fd, buf, 5);
    write(sock_fd, "bye", 4);
    close(sock_fd);
    if (strcmp(buf, "fail") == 0){
      fprintf(stderr, "server is closing, cannot stomith\n");
      goto fail;
    }
    listptr = getnext(listptr);
  }
  fprintf(stderr, "%s stomithed\n", hostname);
  iplist = clear(iplist);
  return 0;
  
 fail:
  iplist = clear(iplist);
  return 1;
}

int getserver(char *hostname)
{
  int n;
  int total;
  int sock_fd;
  char buf[300];
  char *port;
  sock_fd = connect_to_server(hostname, (unsigned int)14243, NULL);
  if (sock_fd < 0)
    return -1;
  write(sock_fd, "port", 5);
  printf("GNBD's on host: %s\n", hostname);
  while(1){
    total = 0;
    do{
      n = read(sock_fd, buf, 4096);
      if (n < 0){
	close(sock_fd);
	fprintf(stderr, "error reading from the server: %s\n",
		strerror(errno));
	return -1;
      }
      total += n;
    }while(buf[total-1] != 0);
    if(strncmp(buf, "bye", 3) == 0)
      break;
    if ((port = strchr(buf, '/')) == NULL){
      fprintf(stderr, "malformed server message: %s\n", buf);
      close(sock_fd);
      return -1;
    }
    *port++ = 0;
    printf("Port: %5s  Name: %s\n", port, buf);
    if (write(sock_fd, "next", 5) != 5){
      fprintf(stderr, "error writing to the server: %s\n", strerror(errno));
      close(sock_fd);
      return -1;
    }
  }
  write(sock_fd, "quit", 5);
  close(sock_fd);
  return 0;
}

int validate_devs(int comm_fd){
  DIR *dp;
  int major_nr;
  u16 minor_nr;
  char buf[4096];
  int bytes;
  char *name;
  char *minor;
  char *bufptr;
  char path[300] = "/dev/gnbd/";
  devlist_t *save_list = NULL;
  devlist_t *del_list = NULL;
  devlist_t *list_ptr = NULL;
  struct dirent *entry;
  bytes = get_proc_data(buf, 4096);
  if (bytes < 0)
    return -1;
  name = strchr(buf, ':');
  name = name + 2;
  bufptr = strchr(name, '\n');
  *bufptr = 0;
  major_nr = (int)strtoul(name, &bufptr, 0);
  if (major_nr == 0 || bufptr == name){
    perror("couldn't find the major number for gnbd");
    return -1;
  }
  bufptr++;
  while( (name = strchr(bufptr, '/')) != NULL){
    name++;
    bufptr = strchr(name, '/');
    *bufptr++ = 0;
    minor = strchr(bufptr, ':');
    minor = minor + 2;
    minor_nr = (u16) strtoul(minor, &bufptr, 0);
    if (minor_nr == 0 || bufptr == minor){
      fprintf(stderr, "couldn't find the minor number for %s: %s\n",
	      name, strerror(errno));
      return -1;
    }
    bufptr++;
    if(do_connect(comm_fd, minor_nr) < 0){
      if (do_device_remove(comm_fd, (int) minor_nr) < 0){
	fprintf(stderr, "couldn't remove bad device '%s': %s\n", name,
		strerror(errno));
	return -1;
      }
      printm("invalid GNBD '%s' removed\n", name);
    }
    else{
      save_list = add(name, save_list);
    }
  }
  dp = opendir("/dev/gnbd/");
  if (dp == NULL){
    perror("error opening directory /dev/gnbd");
    clear(save_list);
    return -1;
  }
  errno = 0;
  while( (entry = readdir(dp)) != NULL){
    if (strcmp(entry->d_name, ".") == 0 ||
	strcmp(entry->d_name, "..") == 0)
      continue;
    if (match(entry->d_name, save_list) < 0)
      del_list = add(entry->d_name, del_list);
  }
  clear(save_list);
  if (closedir(dp) < 0){
    perror("error closeing directory /dev/gnbd");
    save_list = clear(save_list);
    del_list = clear(del_list);
    return -1;
  }
  list_ptr = del_list;
  while(list_ptr != NULL){
    path[10] = 0;
    if (remove(strcat(path,list_ptr->name)) < 0){
      fprintf(stderr, "error removing %s: %s\n", path, strerror(errno));
      del_list = clear(save_list);
      del_list = clear(del_list);
      return -1;
    }
    printm("removed file %s\n", path);
    list_ptr = getnext(list_ptr);
  }
  del_list = clear(del_list);
  return major_nr;
}
     
int list(void){
  int i;
  int bytes;
  char *bufptr;
  char buf[4096];
  bytes = get_proc_data(buf, 4096);
  if (bytes < 0)
    return -1;
  for(i = 0; i < bytes; i++)
    if (buf[i] == '/')
      buf[i] = ' ';
  bufptr = strchr(buf, '\n');
  if (bufptr == NULL || bufptr >= &buf[bytes]){
    perror("incorrectly formatted /proc/gnbd file?!?\n");
    return -1;
  }
  printf("GNBD's currently registered%s", bufptr);
  return 0;
}


int get_major_nr(void){
  char buf[31];
  int bytes;
  char *major;
  char *tail;
  int major_nr;
  bytes = get_proc_data(buf, 31);
  if (bytes < 0)
    return -1;
  major = strchr(buf, ':');
  major = major + 2;
  tail = strchr(major, '\n');
  *tail = 0;
  major_nr = (int)strtoul(major, &tail, 0);
  if (major_nr == 0 || tail == major)
    return -1;
  return major_nr;
}



int addclient(char *name, char *hostname, char *port_ptr){
  int fd;
  struct sockaddr_in server;
  struct hostent *hp;
  char path[300] = "/dev/gnbd/";
  char *tailptr;
  int minor_nr;
  int major_nr;
  uint16 port_nr;
  major_nr = get_major_nr();
  if (major_nr < 0)
    return 1;
  if( (fd=open("/dev/gnbdcomm", O_WRONLY)) < 0){
    if(errno == ENOENT){
      fprintf(stderr, "creating /dev/gnbdcomm\n");
      if (mknod("/dev/gnbdcomm", S_IFBLK | S_IRUSR | S_IWUSR |
                S_IRGRP | S_IROTH, makedev(major_nr, 0)) < 0){
	perror("cannot create gnbdcomm");
	return 1;
      }
    }
    else{
      perror("cannot open comm device");
      return 1;
    }
  }
  if(mkdir("/dev/gnbd", S_IRWXU | S_IRGRP | S_IXGRP |
           S_IROTH | S_IXOTH) == -1){
    if (errno == EEXIST){
      if( (major_nr = validate_devs(fd)) < 0)
        goto fail;
    }
    else{
      perror("unable to create directory /dev/gnbd/");
      goto fail;
    }
  }
  if ( (port_nr = (uint16)strtoul(port_ptr, &tailptr, 0)) == 0 &&
       tailptr == port_ptr){
    fprintf(stderr, "invalid port number %s: %s\n", port_ptr, strerror(errno));
    goto fail;
  } 
  hp = gethostbyname(hostname);
  if (hp == NULL){
    fprintf(stderr, "cannot resolve host ip from name %s: %s", hostname,
            strerror(errno));
    goto fail;
  }
  if (memcpy(&(server.sin_addr.s_addr), hp->h_addr, hp->h_length) == NULL){
    fprintf(stderr, "cannot copy host address: %s\n", strerror(errno));
    goto fail;
  }
  minor_nr = do_device_create(fd, name);
  if (minor_nr < 0){
    fprintf(stderr, "couldn't create gnbd device %s: %s\n", name,
	    strerror(errno));
    goto fail;
  }
  if (do_setipport(fd, (u16)minor_nr, port_nr, server.sin_addr) < 0){
    fprintf(stderr, "couldn't set the ip address and port for %s: %s\n",
	    name, strerror(errno));
    goto fail;
  }
  strcat(path, name);
  if (mknod(path, S_IFBLK | S_IRUSR | S_IWUSR |
	    S_IRGRP | S_IROTH, makedev(major_nr, minor_nr)) < 0){
    fprintf(stderr, "couldn't create gnbd block device file %s: %s\n",
	    path, strerror(errno));
    goto fail;
  }
  if (validate_devs(fd) < 0)
    goto fail;
  printm("imported GNBD '%s'\n", name);
  close(fd);
  return 0;

 fail:
  close(fd);
  return 1;
}

int setclients(int num, char **serv)
{
  int major_nr, minor_nr, i, n, fd, sock_fd;
  int total = 0;
  uint16 port_nr;
  struct in_addr ipaddr;
  char path[300] = "/dev/gnbd/";
  char buf[300];
  char *port;
  char *tailptr;
  errno = 0;
  major_nr = get_major_nr();
  if (major_nr < 0)
    return 1;
  if( (fd=open("/dev/gnbdcomm", O_WRONLY)) < 0){
    if(errno == ENOENT){
      fprintf(stderr, "creating /dev/gnbdcomm\n");
      if (mknod("/dev/gnbdcomm", S_IFBLK | S_IRUSR | S_IWUSR |
                S_IRGRP | S_IROTH, makedev(major_nr, 0)) < 0){
	perror("cannot create gnbdcomm");
	return 1;
      }
      if( (fd = open("/dev/gnbdcomm", O_WRONLY)) < 0){
	perror("cannot open comm device?!?!");
	return 1;
      }
    }
    else if(errno == ENODEV){
      if(unlink("/dev/gnbdcomm") < 0){
	perror("cannot remove old gnbdcomm");
	return 1;
      }
      if (mknod("/dev/gnbdcomm", S_IFBLK | S_IRUSR | S_IWUSR |
                S_IRGRP | S_IROTH, makedev(major_nr, 0)) < 0){
	perror("cannot create gnbdcomm");
	return 1;
      }
      if( (fd = open("/dev/gnbdcomm", O_WRONLY)) < 0){
	perror("cannot open comm device?!?!");
	return 1;
      }
    }
    else{
      perror("cannot open comm device");
      return 1;
    }
  }
  if(mkdir("/dev/gnbd", S_IRWXU | S_IRGRP | S_IXGRP |
	   S_IROTH | S_IXOTH) == -1){
    if (errno == EEXIST){
      if( (major_nr = validate_devs(fd)) < 0)
	goto fail;
    }
    else{
      perror("unable to create directory /dev/gnbd/");
      goto fail;
    }
  }
  for (i = 0; i < num; i++){
    sock_fd = connect_to_server(serv[i], (unsigned int)14243, &ipaddr);
    if (sock_fd < 0){
      continue;
    }
    write(sock_fd, "port", 5);
    while(1){
      path[10] = 0;
      total = 0;
      while ( (n = read(sock_fd, buf + total, 300 - total)) != 0){
	if (n < 0){
	  fprintf(stderr, "error reading from the server: %s\n",
		  strerror(errno));
	  goto sock_fail;
	}
	total += n;
	if (buf[total-1] == 0)
	  break;
      }
      if(strncmp(buf, "bye", 3) == 0)
	break;
      if ((port = strchr(buf, '/')) == NULL){
	fprintf(stderr, "malformed server message: %s\n", buf);
	goto sock_fail;
      }
      *port++ = 0;
      if ( (port_nr = (uint16)strtoul(port, &tailptr, 0)) == 0 &&
	   tailptr == port){
	fprintf(stderr, "invalid port number %s: %s\n", port, strerror(errno));
	goto sock_fail;
      } 
      errno = 0;
      if ( (minor_nr = do_device_create(fd, buf)) < 0){
	if (errno == EEXIST){
	  printm("duplicate GNBD '%s', skipping\n", buf);
	  if (write(sock_fd, "next", 5) != 5){
	    fprintf(stderr, "error writing to the server: %s\n",
		    strerror(errno));
	    goto sock_fail;
	  }
	  continue;
	}
	fprintf(stderr, "couldn't create gnbd device %s: %s\n", buf,
		strerror(errno));
	goto sock_fail;
      }
      if (do_setipport(fd, (u16)minor_nr, port_nr, ipaddr) < 0){
	fprintf(stderr, "couldn't set the ip address and port for %s: %s\n",
		buf, strerror(errno));
	goto sock_fail;
      }
      strcat(path, buf);
      if (mknod(path, S_IFBLK | S_IRUSR | S_IWUSR |
		S_IRGRP | S_IROTH, makedev(major_nr, minor_nr)) < 0){
	fprintf(stderr, "couldn't create gnbd block device file %s: %s\n",
		path, strerror(errno));
	goto sock_fail;
      }
      printm("imported GNBD '%s' from server %s\n", buf, serv[i]);
      if (write(sock_fd, "next", 5) != 5){
	fprintf(stderr, "error writing to the server: %s\n", strerror(errno));
	goto sock_fail;
      }
    }
    write(sock_fd, "quit", 5);
    close(sock_fd);
  }
  if (validate_devs(fd) < 0)
    goto fail;
  close(fd);
  return 0;
  
 sock_fail:     
  close(sock_fd);
 fail:
  close(fd);
  return 1;
}

int debug(int num, char **debug, int on){
  int i;
  int x, y;
  unsigned int flag;
  int comm_fd;
  if( (comm_fd=open("/dev/gnbdcomm", O_WRONLY)) <0) {
    perror("Cannot open comm device");
    return 1;
  }
  for (i = 0; i < num; i++){
    
    for(x = 0; gnbd_debug_flags[x].name[0]; x++){
      if (strcmp(gnbd_debug_flags[x].name, debug[i]) == 0){
	for(y = 0; y < gnbd_debug_flags[x].len; y++){
	  flag = gnbd_debug_flags[x].flag[y];
	  if (ioctl(comm_fd, (on != 0)? GNBD_DEBUG_ON : GNBD_DEBUG_OFF,
		    &flag) < 0){
	    perror("ioctl failed");
	    return 1;
	  }
	}
	break;
      }
    }
    if (gnbd_debug_flags[x].name[0] == 0){
      fprintf(stderr, "couldn't find debug flag %s\n", debug[i]);
      return 1;
    }
  }
  if (on){
    printm("successfully enabled debugging\n");
  }
  else{
    printm("successfully disabled debugging\n");
  }
  close(comm_fd);
  return 0;
}

int remove_client(char *name){
  int comm_fd;
  int found = 0;
  u16 minor = 0;
  char *name_ptr;
  char *minor_ptr;
  char *bufptr;
  char path[300] = "/dev/gnbd/";
  int bytes;
  char buf[4096];
  if (name == NULL){
    perror("gnbd client name is NULL\n");
    return -1;
  }
  if ( (comm_fd = open("/dev/gnbdcomm", O_WRONLY)) < 0){
    perror("cannot open comm device");
    return -1;
  }
  bytes = get_proc_data(buf, 4096);
  if (bytes < 0)
    return -1;
  if (remove(strcat(path,name)) < 0){
    fprintf(stderr, "blockdevice %s does not exist\n", path);
  }
  bufptr = buf;
  while( (name_ptr = strchr(bufptr, '/')) != NULL){
    name_ptr++;
    bufptr = strchr(name_ptr, '/');
    *bufptr++ = 0;
    if (strcmp(name_ptr,name) != 0)
      continue;
    minor_ptr = strchr(bufptr, ':');
    minor_ptr += 2;
    minor = (u16)strtoul(minor_ptr, &bufptr, 0);
    if (minor == 0 || bufptr == minor_ptr){
      fprintf(stderr, "couldn't find the minor number for %s: %s\n",
	      name, strerror(errno));
      close(comm_fd);
      return -1;
    }
    found = 1;
    break;
  }
  if (found == 0){
    fprintf(stderr, "couldn't find gnbd client info for %s\n", name);
    close(comm_fd);
    return 0;
  }
  if (do_device_remove(comm_fd, (int) minor) < 0){
    fprintf(stderr, "couldn't remove device %s: %s\n", name,
	    strerror(errno));
    close(comm_fd);
    return -1;
  }
  printm("removed GNBD '%s'\n", name);
  close(comm_fd);
  return 0;
}

int remove_all(void){
  DIR *dp;
  int bytes;
  char buf[4096];
  char *name_ptr;
  char *bufptr;
  devlist_t *del_list = NULL;
  devlist_t *list_ptr;
  struct dirent *entry;
  bytes = get_proc_data(buf, 4096);
  if (bytes < 0)
    return -1;
  bufptr = buf;
  while( (name_ptr = strchr(bufptr, '/')) != NULL){
    name_ptr++;
    bufptr = strchr(name_ptr, '/');
    if (bufptr == NULL){
      perror("malformed /proc/gnbd file?!?!?\n");
      return 1;
    }
    *bufptr++ = 0;
    del_list = add(name_ptr, del_list);
  }
  dp = opendir("/dev/gnbd/");
  if (dp != NULL){
    while( (entry = readdir(dp)) != NULL){
      if (strcmp(entry->d_name, ".") == 0 ||
	  strcmp(entry->d_name, "..") == 0)
	continue;
      if (match(entry->d_name, del_list) < 0)
	del_list = add(entry->d_name, del_list);
    }
    if(closedir(dp) < 0){
      perror("error closing directory /dev/gnbd");
      del_list = clear(del_list);
      return 1;
    }
  }
  list_ptr = del_list;
  while(list_ptr != NULL){
    remove_client(list_ptr->name);
    list_ptr = getnext(list_ptr);
  }
  return 0;
}

int usage(void){
  printf("Usage:\n\n");

  printf("gclient <hostname> [<other hostnames>]\n");
  printf("    Import all GNBD's exported from specified host(s).\n\n");
  
  printf("gclient -a <GNBD> <hostname> <port>\n");
  printf("    Import named GNBD from specified host and port.\n\n");

  printf("gclient -r <GNBD> [<other GNBD's>]\n");
  printf("    Remove named GNBD(s).\n\n");

  printf("gclient -l\n");
  printf("    List all imported GNBD's.\n\n");

  printf("gclient -g <hostname> [<other hostnames>]\n");
  printf("    List exported GNBD's on specified host(s).\n\n");

  printf("gclient -v\n");
  printf("    Remove all invalid GNBD's.\n\n");

  printf("gclient -d <debug flag> [<additional debug flags>]\n");
  printf("    Turn debugging on (must be compiled into gnbd.o module).\n\n");

  printf("gclient -o <debug flag> [<additional debug flags>]\n");
  printf("    Turn debugging off.\n\n");

  printf("gclient -h\n");
  printf("    Print help.\n\n");

  printf("-q toggles gclient to quiet mode, so it does not print on "
	 "success\n\n");

#if 0
  printf("Register a gnbd client for each gnbd server on the given hosts\n"
	 "   gclient <hostname> [<additional host names>]\n");
  printf("Add a gnbd client with with the given name, host, and port\n"
	 "   gclient -a <name> <hostname> <port>\n");
  printf("Remove a gnbd client\n"
	 "   gclient -r <client name> [<additional client names>]\n");
  printf("Remove all current gnbd clients\n"
	 "   gclient -r\n");
  printf("List all current gnbd clients\n"
	 "   gclient -l\n");
  printf("List all the gnbd servers on the given hosts\n"
	 "   gcleint -s <hostname> [<additional host names>]\n");
  printf("Remove all invalid gnbd clients\n"
	 "   gclient -v\n");
  printf("Turn gnbd debugging on (must be compiled into gnbd.o module)\n"
	 "   gclient -d <debug flag> [<additional debug flags>]\n");
  printf("Turn gnbd debugging off\n"
	 "   gclient -o <debug flag> [<additional debug flags>]\n");
  printf("Print this message\n"
	 "   gclient -h\n\n");
#endif
  return 1;
}


int main(int argc, char **argv)
{
  int c;
  int i;
 loop:
  c = getopt(argc, argv, "qasgrlvdoh");
  switch(c){
  case 'q':
    message_flag = 0;
    goto loop;
  case -1:
    if (argc == optind)
      return usage();
    return setclients(argc - optind, &argv[optind]);
  case 'a':
    if (argc - optind != 3)
      return usage();
    return addclient(argv[optind], argv[optind + 1], argv[optind + 2]);
  case 's':
    if (argc != optind + 1)
      return usage();
    return stomith(argv[optind]);
  case 'g':
    if (argc == optind)
      return usage();
    for (i = optind; i < argc; i++)
      getserver(argv[i]);
    return 0;
  case 'r':
    if (argc == optind)
      return remove_all();
    for (i = optind; i < argc; i++)
      remove_client(argv[i]);
    return 0;
  case 'l':
    if (optind != argc){
      fprintf(stderr, "unnecessary argument after list request: %s\n\n",
              argv[optind]);
      return usage();
    }
    return list();
  case 'v':
    if (optind != argc){
      fprintf(stderr, "unnecessary argument after validate request: %s\n\n",
	      argv[optind]);
      return usage();
    }
    if (setclients(0, NULL) == 0){
      printm("GNBD list is now uptodate\n");
      return 0;
    }
    return 1;
  case 'd':
    if (argc == optind){
      fprintf(stderr, "missing argument for debug request\n\n");
      return usage();
    }
    return debug(argc - optind, &argv[optind], 1);
  case 'o':
    if (argc == optind){
      fprintf(stderr, "missing argument for debug request\n\n");
      return usage();
    }
    return debug(argc - optind, &argv[optind], 0);
  case '?':
    if(isprint(optopt))
      fprintf(stderr, "Unknown option: -%c\n\n", optopt);
    else
      fprintf(stderr, "Unknown option\n\n");
    return usage();
  case 'h': default:
    return usage();
  }
}










