/******************************************************************************
*******************************************************************************
**
**    Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Steven Whitehouse <steve@sistina.com>
** 
*******************************************************************************
******************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <getopt.h>
#include <sys/ioctl.h>
#include <sys/mount.h>
#include <sys/stat.h>

#include "global.h"
#include "gfs_ondisk.h"
#include "gfs_ioctl.h"
#include "osi_list.h"

struct rglist_entry {
	struct osi_list_head list;
	gfs_rindex_t ri;
	gfs_rgrp_t rg;
};

struct jilist_entry {
	struct osi_list_head list;
	gfs_jindex_t ji;
};

/*
 * verbose: 0 = no messages, 1 = normal, 2 = everything
 * test: 0 = normal, 1 = don't actually write data, but do everything else
 * fspath: path to root of mounted GFS filesystem
 * device: the device upon which the GFS filesystem is mounted
 * fsoptions: the mount options used
 * devsize: the size of the device (in filesystem blocks, rounded down)
 * fssize: the size of the filesystem (in filesystem blocks, rounded down)
 * override_device_size: if non-zero, this is used for the device size
 */
static int verbose = 1;
static int test = 0;
static char fspath[4096];
static char device[1024];
static char fsoptions[4096];
static uint64 devsize;
static uint64 fssize;
static uint64 override_device_size = 0;

/*
 * fs_sb: the superblock read from the mounted filesystem
 * rglist_current: list of resource groups currently making up the filesystem
 * rglist_new: where we put the new resource groups to be written
 * jilist_current: list of current journals in the filesystem
 */
static struct gfs_sb fs_sb;
static OSI_LIST_HEAD(rglist_current);
static OSI_LIST_HEAD(rglist_new);
static OSI_LIST_HEAD(jilist_current);


#if defined(__linux__)

/**
 * device_size - Find out the size of a block device
 * @device: The name of the device
 *
 * Returns: The size of the device
 */
static uint64 device_size(char *device)
{
	int fd, error;
	long size;
	unsigned bsize = 512;

	if (override_device_size)
		return override_device_size;

	fd = open(device, O_RDONLY);
	if (fd < 0) {
		perror(device);
		exit(-1);
	}

	error = ioctl(fd, BLKGETSIZE, &size);
	if (error < 0) {
		perror("ioctl: BLKGETSIZE");
		exit(-1);
	}

	close(fd);

	if (bsize < fs_sb.sb_bsize) {
		while(bsize != fs_sb.sb_bsize) {
			bsize <<= 1;
			size >>=1;
		}
	}

	return size;
}
#else
#error No device size code available for this OS
#endif  /*  defined(__linux__)  */

/**
 * jread - Read from journaled file using ioctl()
 * @fd: The fd to read from
 * @buf: The buffer to fill
 * @size: The amount of data to read
 * @offset: The offset to read from
 *
 * Returns: Error code, or amount of data read
 */
int jread(int fd, void *buf, uint64 size, uint64 *offset)
{
  gfs_jio_t jt;
  int ret;

  jt.jio_size = size;
  jt.jio_offset = *offset;
  jt.jio_data = buf;

  ret = ioctl(fd, GFS_JREAD, &jt);
  if (ret > 0)
    (*offset) += ret;

  return ret;
}

/**
 * jwrite - Write to journaled file using ioctl()
 * @fd: The fd to write to
 * @buf: The buffer to write
 * @size: The amount of data to write
 * @offset: The offset at which to write the data
 *
 * Returns: Error code, or the amount of data written
 */
int jwrite(int fd, void *buf, uint64 size, uint64 *offset)
{
  gfs_jio_t jt;
  int ret;

  jt.jio_size = size;
  jt.jio_offset = *offset;
  jt.jio_data = buf;

  ret = ioctl(fd, GFS_JWRITE, &jt);
  if (ret > 0)
    (*offset) += ret;

  return ret;
}


/**
 * filesystem_size - Calculate the size of the filesystem
 *
 * Reads the lists of journals and resource groups in order to
 * work out where the last block of the filesystem is located.
 *
 * Returns: The calculated size
 */
static uint64 filesystem_size(void)
{
	osi_list_head_t *tmp, *head;
	struct rglist_entry *rgl;
	struct jilist_entry *jil;
	uint64 size = 0;
	uint64 extent;

	tmp = head = &rglist_current;
	for(;;) {
		tmp = tmp->next;
		if (tmp == head)
			break;
		rgl = osi_list_entry(tmp, struct rglist_entry, list);
		extent = rgl->ri.ri_addr + rgl->ri.ri_length + rgl->ri.ri_data;
		if (extent > size)
			size = extent;
	}

	tmp = head = &jilist_current;
	for(;;) {
		tmp = tmp->next;
		if (tmp == head)
			break;
		jil = osi_list_entry(tmp, struct jilist_entry, list);
		extent = jil->ji.ji_addr + jil->ji.ji_nsegment * fs_sb.sb_seg_size;
		if (extent > size)
			size = extent;
	}

	return size;
}

/**
 * gfs_jientry - Get journal index entry
 * @fd: The fd of the journal index
 * @offset: The offset at which the journal entry appears
 *
 * Reads a single entry from the journal index file and
 * adds it to the list of current journal entries.
 *
 * Returns: 1 on EOF, 0 otherwise
 */
static int get_jientry(int fd, uint64 *offset)
{
	char buffer[sizeof(gfs_jindex_t)];
	int len = jread(fd, buffer, sizeof(gfs_jindex_t), offset);
	struct jilist_entry *jil;
	if (len != sizeof(gfs_jindex_t)) {
		if (len == 0)
			return 1;

		fprintf(stderr, "Erk! Read odd size from jindex (%d)\n", len);
		exit(-1);
	}
	if ((jil = malloc(sizeof(struct jilist_entry))) == NULL)
		exit(-1);
	memset(jil, 0, sizeof(struct jilist_entry));
	gfs_jindex_in(&jil->ji, buffer);
	osi_list_add(&jil->list, &jilist_current);
	return 0;
}

/**
 * read_journals - Read the whole journal index
 * @fs_fd: An fd for some file or directory within the mounted GFS filesystem
 * @ino: The inode number of the journal index
 *
 */
static void read_journals(int fs_fd, gfs_inum_t *ino)
{
	int fd = ioctl(fs_fd, GFS_OBYNUM, ino);
	uint64 offset = 0;
	if (fd < 0) {
		perror("ioctl: GFS_OBYNUM");
		exit(-1);
	}

	while(get_jientry(fd, &offset) == 0) ;

	close(fd);
}

/**
 * get_rgrp - Read a single rindex entry
 * @fd: The fd for the rindex file
 * @offset: The offset at which to read the rindex entry
 *
 * Reads a single rindex entry and adds it to the list of current
 * resource group entries.
 *
 * Returns: 1 on EOF, 0 otherwise
 */
static int get_rgrp(int fd, uint64 *offset)
{
	char buffer[sizeof(gfs_rindex_t)];
	int len = jread(fd, buffer, sizeof(gfs_rindex_t), offset);
	struct rglist_entry *rgl;
	if (len != sizeof(gfs_rindex_t)) {
		if (len == 0)
			return 1;

		fprintf(stderr, "Erk! Read odd size from rindex (%d)\n", len);
		exit(-1);
	}
	if ((rgl = malloc(sizeof(struct rglist_entry))) == NULL)
		exit(-1);
	memset(rgl, 0, sizeof(struct rglist_entry));
	gfs_rindex_in(&rgl->ri, buffer);
	osi_list_add(&rgl->list, &rglist_current);
	return 0;
}

/**
 * read_rgrps - Reads the contents of the rindex file
 * @fs_fd: An fd for any file or directory within the mounted GFS filesytem
 * @ino: The inode number of the rindex
 *
 */
static void read_rgrps(int fs_fd, gfs_inum_t *ino)
{
	int fd = ioctl(fs_fd, GFS_OBYNUM, ino);
	uint64 offset = 0;

	if (fd < 0) {
		perror("ioctl: GFS_OBYNUM");
		exit(-1);
	}

	while(get_rgrp(fd, &offset) == 0) ;

	close(fd);
}

/**
 * write_a_block - Write a block to the current device
 * @where: The position to write the block (in filesystem blocks)
 * @rg: The (optional) resource group to write
 *
 * Writes a single disk block to the device. It has a safety check which
 * prevents it writing to the device at a position within the control of
 * the active filesystem. If @rg is NULL, it writes a single block of
 * zeros with a meta_header, otherwise the resource group is copied 
 * into the start of the block.
 */
static void write_a_block(uint64 where, gfs_rgrp_t *rg)
{
	char buffer[4096];
	uint64 fsoffset = where * (uint64)fs_sb.sb_bsize;
	int fd = open(device, O_RDWR);
	gfs_meta_header_t mh;
	mh.mh_magic = GFS_MAGIC;
	mh.mh_type = GFS_METATYPE_RB;
	mh.mh_format = GFS_FORMAT_RB;

	if (fd < 0) {
		perror(device);
		exit(-1);
	}
	if (where < fssize) {
		fprintf(stderr, "Sanity check failed: Caught trying to write to live filesystem!\n");
		exit(-1);
	}
	memset(buffer, 0, 4096);
	if (rg)
		gfs_rgrp_out(rg, buffer);
	else
		gfs_meta_header_out(&mh, buffer);
	if (lseek(fd, fsoffset, SEEK_SET) != fsoffset) {
		perror(device);
		exit(-1);
	}
	if (write(fd, buffer, fs_sb.sb_bsize) != fs_sb.sb_bsize) {
		perror("write_zero_block");
		exit(-1);
	}
	close(fd);
}

/**
 * write_whole_rgrp - Write a complete rgrp, including bitmaps
 * @rgl: The information about the resource group
 *
 * Writes a complete rgrp, including any bitmap blocks required
 * by calling write_a_block() a number of times. Calls sync() to
 * ensure data really reached disk.
 */
static void write_whole_rgrp(struct rglist_entry *rgl)
{
	uint32 l;
	uint32 nzb = rgl->ri.ri_length;
	uint64 addr = rgl->ri.ri_addr;

	write_a_block(addr++, &rgl->rg);
	for(l = 1; l < nzb; l++)
		write_a_block(addr++, NULL);
	sync();
}

/**
 * get_length - Use stat() to get the length of a file
 * @fd: The fd of the file whose length we wish to know
 *
 * Returns: The length
 */
static uint64 get_length(int fd)
{
	struct stat st;
	if (fstat(fd, &st) < 0) {
		perror("stat");
		fprintf(stderr, "Failed to get size of file. Aborting.\n");
		exit(-1);
	}
	return st.st_size;
}

/**
 * write_rindex - Writes new records to the end of the rindex file
 * @fs_fd: A fd of any file or directory withint the GFS filesystem
 * @ino: The inode number of the rindex
 *
 * This is the critical function in expanding a filesystem. It does the
 * actual write to the rindex which causes the GFS filesystem to see the
 * new resource groups which were previously added.
 */
static void write_rindex(int fs_fd, gfs_inum_t *ino)
{
	osi_list_head_t *tmp, *head;
	struct rglist_entry *rgl;
	int fd = ioctl(fs_fd, GFS_OBYNUM, ino);
	char buffer[sizeof(gfs_rindex_t)];
	uint64 orig_length;
	uint64 offset;

	if (fd < 0) {
		perror("ioctl: GFS_OBYNUM");
		exit(-1);
	}

	offset = orig_length = get_length(fd);

	/*
	 * This is the critical section.
	 * If things mess up here, it could be very difficult to put right
	 */
	tmp = head = &rglist_new;
	for(;;) {
		tmp = tmp->next;
		if (tmp == head)
			break;
		rgl = osi_list_entry(tmp, struct rglist_entry, list);
		gfs_rindex_out(&rgl->ri, buffer);
		if (jwrite(fd, buffer, sizeof(gfs_rindex_t), &offset) != sizeof(gfs_rindex_t)) {
			perror("write: rindex");
			fprintf(stderr, "Aborting...\n");
			goto abort;
		}
	}
	fsync(fd);
	/*
	 * This is the end of the critical section
	 */
	close(fd);
	return;
abort:
	/*
	 * If things did mess up in the critical section, we try to put
	 * them right here by truncating the resource index to its original
	 * length. If this fails, we have no more fall back poistions, so
	 * we let the user know.
	 */
	if (ftruncate(fd, orig_length) < 0) {
		perror("truncate");
		fprintf(stderr, "Critical: Cannot return filesystem to original state. Filesystem corruption may result\n");
		fprintf(stderr, "*** Stop cluster, unmount filsystems, correct fault manually ***\n");
		exit(-1);
	}
	close(fd);
	fprintf(stderr, "Managed to return rindex to original state\n");
	exit(-1);
}

/**
 * write_rgrps - Write the new resource groups to disk
 * @fs_fd: An fd from any file or directory on the GFS mounted filesystem
 * @ino: The inode number of the rindex
 *
 * This first writes out the new resource group information to the
 * area of the disk beyond the area the filesystem is currently
 * using and then calls write_rindex() to make the filesystem see
 * the newly written resource groups.
 */
static void write_rgrps(int fs_fd, gfs_inum_t *ino)
{
	osi_list_head_t *tmp, *head;
	struct rglist_entry *rgl;
	tmp = head = &rglist_new;
	for(;;) {
		tmp = tmp->next;
		if (tmp == head)
			break;
		rgl = osi_list_entry(tmp, struct rglist_entry, list);
		write_whole_rgrp(rgl);
	}

	sync();
	sync();
	sync();

	write_rindex(fs_fd, ino);

	sync();
	sync();
	sync();
}

/**
 * gather_info - Gathers all the information about the existing filesystem
 *
 */
static void gather_info(void)
{
	char sbuf[4096];
	int fd = open(fspath, O_RDONLY);

	if (fd < 0) {
		perror(fspath);
		exit(-1);
	}
	if (ioctl(fd, GFS_GET_SUPER, (unsigned long)sbuf) < 0) {
		perror("ioctl: GFS_GET_SUPER");
		exit(-1);
	}
	gfs_sb_in(&fs_sb, sbuf);

	read_rgrps(fd, &fs_sb.sb_rindex_di);
	read_journals(fd, &fs_sb.sb_jindex_di);
	close(fd);
	devsize = device_size(device);
	fssize = filesystem_size();
}

/**
 * print_rgrps - Print information about resource groups
 * @lh: The list of resoruce groups to print
 *
 */
static void print_rgrps(osi_list_head_t *lh)
{
	osi_list_head_t *tmp, *head;
	struct rglist_entry *rgl;
	int n = 0;

	tmp = head = lh;
	for(;;) {
		tmp = tmp->next;
		if (tmp == head)
			break;
		rgl = osi_list_entry(tmp, struct rglist_entry, list);
		n++;
		printf("RI: Addr %"FMT_64"u, RgLen %u, "
			"Start %"FMT_64"u, DataLen %u, BmapLen %u\n", 
			rgl->ri.ri_addr, rgl->ri.ri_length, 
			rgl->ri.ri_data1, rgl->ri.ri_data, rgl->ri.ri_bitbytes);
	}
	printf("RGRP: %d Resource groups in total\n", n);
}

/**
 * print_journals - Print a list of journals
 *
 */
static void print_journals(osi_list_head_t *lh)
{
	osi_list_head_t *tmp, *head;
	struct jilist_entry *jil;
	int n = 0;

	tmp = head = lh;
	for(;;) {
		tmp = tmp->next;
		if (tmp == head)
			break;
		jil = osi_list_entry(tmp, struct jilist_entry, list);
		n++;
		printf("JI: Addr %"FMT_64"u NumSeg %u SegSize %u\n", jil->ji.ji_addr, 
			jil->ji.ji_nsegment, fs_sb.sb_seg_size);
	}
	printf("JRNL: %d Journals in total\n", n);
}

/**
 * print_info - Print out various bits of (interesting?) information
 *
 */
static void print_info(void)
{
	printf("FS: Mount Point: %s\n", fspath);
	printf("FS: Device: %s\n", device);
	printf("FS: Options: %s\n", fsoptions);
	printf("FS: Size: %"FMT_64"u\n", fssize);
	if (verbose > 1) {
		printf("RGRP: Current Resource Group List:\n");
		print_rgrps(&rglist_current);
		printf("JRNL: Current Journal List:\n");
		print_journals(&jilist_current);
	}
	printf("DEV: Size: %"FMT_64"u\n", devsize);
	if (verbose > 1) {
		printf("RGRP: New Resource Group List:\n");
		print_rgrps(&rglist_new);
	}
}

#define RGRP_STUFFED_BLKS(sb) (((sb)->sb_bsize - sizeof(gfs_rgrp_t)) * GFS_NBBY)
#define RGRP_BITMAP_BLKS(sb) (((sb)->sb_bsize - sizeof(gfs_meta_header_t)) * GFS_NBBY)

/**
 * rgrp_length - Calculate the length of a resource group
 * @size: The total size of the resource group
 *
 */
uint64 rgrp_length(uint64 size)
{
	uint64 bitbytes = RGRP_BITMAP_BLKS(&fs_sb) + 1;
	uint64 stuff = RGRP_STUFFED_BLKS(&fs_sb) + 1;
	uint64 blocks = 1;

	if (size < stuff)
		goto out;
	size -= stuff;
	while(size > bitbytes) {
		blocks++;
		size -= bitbytes;
	}
	if (size)
		blocks++;
out:
	return blocks;
}

/**
 * make_rgrp - Make a new rglist_entry
 * @offset: The offset at which the new rgrp will go
 * @size: The size of the new rgrp
 *
 * Returns: The end of the new resource group
 */
uint64 make_rgrp(uint64 offset, uint64 size)
{
	struct rglist_entry *rgl = malloc(sizeof(struct rglist_entry));
	if (rgl == NULL)
		exit(-1);
	memset(rgl, 0, sizeof(struct rglist_entry));

	rgl->ri.ri_addr = offset;
	rgl->ri.ri_length = rgrp_length(size);
	rgl->ri.ri_data1 = offset + rgl->ri.ri_length;
	rgl->ri.ri_data = size - rgl->ri.ri_length;

	/* Round down to nearest multiple of GFS_NBBY */
	while(rgl->ri.ri_data & 0x03)
		rgl->ri.ri_data--;

	rgl->ri.ri_bitbytes = rgl->ri.ri_data / GFS_NBBY;

	rgl->rg.rg_header.mh_magic = GFS_MAGIC;
	rgl->rg.rg_header.mh_type = GFS_METATYPE_RG;
	rgl->rg.rg_header.mh_format = GFS_FORMAT_RG;
	rgl->rg.rg_free = rgl->ri.ri_data;

	osi_list_add(&rgl->list, &rglist_new);
	return offset + size;
}


/**
 * create_rgrps - Create a list of the new rgrps
 * 
 */
static void create_rgrps(void)
{
	uint64 space = devsize - fssize;
	uint64 optimal_rgrp_size = RGRP_STUFFED_BLKS(&fs_sb) + 
					14 * RGRP_BITMAP_BLKS(&fs_sb) + 15;
	uint64 rgrps = 1 + space / optimal_rgrp_size;
	uint64 offset = fssize;
	uint64 rgsize;
	uint64 n;

	rgsize = optimal_rgrp_size;

	for(n = 0; n < rgrps; n++)
		offset = make_rgrp(offset, (n != 0) ? rgsize :
					(space - ((rgrps-1)*rgsize)));

	if (offset > devsize) {
		fprintf(stderr, "Calculation error: Out of bounds\n");
		exit(-1);
	}
}


/**
 * update_fs - Actually perform the filesystem update
 *
 */
static void update_fs(void)
{
	int fd = open(fspath, O_RDONLY);
	if (fd < 0) {
		perror(fspath);
		exit(-1);
	}
	if (verbose)
		printf("Preparing to write new FS information...\n");
	write_rgrps(fd, &fs_sb.sb_rindex_di);
	if (verbose)
		printf("Done.\n");
	close(fd);
}

/**
 * find_fs - Find the filesystem which the user specified
 * @name: The name of a device or mount point
 *
 * Returns: 0 if the filesystem is located, 1 otherwise
 */
static int find_fs(char *name)
{
	FILE *fp = fopen("/proc/mounts", "r");
	char buffer[4096];
	char fstype[80];
	int fsdump, fspass;

	if (fp == NULL) {
		perror("open: /proc/mounts");
		exit(-1);
	}
	while((fgets(buffer, 4095, fp)) != NULL) {
		buffer[4095] = 0;
		if (strstr(buffer, name) == 0)
			continue;
		if (sscanf(buffer, "%s %s %s %s %d %d", device, fspath, fstype,
					fsoptions, &fsdump, &fspass) != 6)
			continue;
		if (strcmp(fstype, "gfs") != 0)
			continue;
		if ((strcmp(device, name) != 0) && (strcmp(fspath, name) != 0))
			continue;
		fclose(fp);
		return 0;
	}
	fprintf(stderr, "GFS Filesystem %s not found\n", name);
	fclose(fp);
	return 1;
}

/**
 * usage - Print out the usage message
 *
 * This function does not include documentation for the -D option
 * since normal users have no use for it at all. The -D option is
 * only for developers. It intended use is in combination with the
 * -t flag to find out what the result would be of trying different
 * device sizes without actually having to try them manually.
 */
static void usage(void)
{
	fprintf(stderr,
		"gfs_expand: (C) 2000-2001 Sistina Software Inc.\n"
		"Expands a GFS filesystem to fill the allocated device\n"
		"Usage:\n"
		"gfs_expand [options] /path/to/filesystem\n"
		"Options:\n"
		"     q    - Quiet, reduce verbosity\n"
		"     t    - Test, do everything except update FS\n"
		"     v    - Verbose, increase verbosity\n"
		);
}

/**
 * main - Tha main function
 * @argc: The argument count
 * @argv: The argument vector
 *
 * Runs through the filesystem expansion code for each of the specified
 * filesystems. Each filesystem specified on the command line has the
 * same options applied to it. You'll need to run the program multiple times
 * if you want to use it on several different filesystems with different
 * options for each. If you forget to specify a filesystem, then it is
 * assumed that the program has run successfully, since its done everything
 * asked of it, and it exits without printing a message.
 *
 * Returns: 0 on success, -1 otherwise
 */
int main(int argc, char *argv[])
{
	int opt;

	while ((opt = getopt(argc, argv, "D:hvqt?")) != EOF) {
		switch(opt) {
			case 'D': /* This option is for testing only */
				override_device_size = atoi(optarg);
				override_device_size <<= 11;
				break;
			case 'h':
				usage();
				exit(0);
			case 'q':
				if (verbose)
					verbose--;
				break;
			case 't':
				test = 1;
				break;
			case 'v':
				verbose++;
				break;
			case '?':
				usage();
				exit(-1);
		}
	}

	while ((argc - optind) > 0) {
		if (find_fs(argv[optind++]))
			continue;
		gather_info();
		if (fssize > devsize) {
			fprintf(stderr, "Filesystem thinks device is bigger than it really is.... skipping\n");
			continue;
		}
		if ((devsize - fssize) < 100) {
			fprintf(stderr, "Device has grown by less than 100 blocks.... skipping\n");
			continue;
		}
		create_rgrps();
		if (verbose)
			print_info();
		if (!test)
			update_fs();
	}

	return 0;
}

