/******************************************************************************
*******************************************************************************
**
**    Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Steven Whitehouse <steve@sistina.com>
** 
*******************************************************************************
******************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <getopt.h>
#include <sys/ioctl.h>
#include <sys/mount.h>
#include <sys/stat.h>

#include "global.h"
#include "gfs_ondisk.h"
#include "gfs_ioctl.h"
#include "osi_list.h"

#define BLOCK_END(buffer,type) ((buffer) + GFS_BASIC_BLOCK - sizeof(type))

struct rglist_entry {
	struct osi_list_head list;
	gfs_rindex_t ri;
	gfs_rgrp_t rg;
};

struct jilist_entry {
	struct osi_list_head list;
	gfs_jindex_t ji;
};

/*
 * verbose: 0 = no messages, 1 = normal, 2 = everything
 * test: 0 = normal, 1 = don't actually write data, but do everything else
 * fspath: path to root of mounted GFS filesystem
 * device: the device upon which the GFS filesystem is mounted
 * fsoptions: the mount options used
 * devsize: the size of the device (in filesystem blocks, rounded down)
 * fssize: the size of the filesystem (in filesystem blocks, rounded down)
 * override_device_size: if non-zero, this is used for the device size
 * number_of_journals: Number of journals to add to each fs
 * journal_size: The size of each journal, in MB
 * journal_size_blocks: The size of each journal in fs blocks
 */
static int verbose = 1;
static int test = 0;
static char fspath[4096];
static char device[1024];
static char fsoptions[4096];
static uint64 devsize;
static uint64 fssize;
static uint64 override_device_size = 0;
static unsigned int number_of_journals = 1;
static uint64 journal_size = 8;
static uint64 journal_size_blocks;

/*
 * fs_sb: the superblock read from the mounted filesystem
 * rglist_current: list of resource groups currently making up the filesystem
 * jilist_current: list of current journals in the filesystem
 * jilist_new: where we put the new resource groups to be written
 */
static struct gfs_sb fs_sb;
static OSI_LIST_HEAD(rglist_current);
static OSI_LIST_HEAD(jilist_current);
static OSI_LIST_HEAD(jilist_new);


#if defined(__linux__)

/**
 * device_size - Find out the size of a block device
 * @device: The name of the device
 *
 * Returns: The size of the device
 */
static uint64 device_size(char *device)
{
	int fd, error;
	long size;
	unsigned bsize = 512;

	if (override_device_size)
		return override_device_size;

	fd = open(device, O_RDONLY);
	if (fd < 0) {
		perror(device);
		exit(-1);
	}

	error = ioctl(fd, BLKGETSIZE, &size);
	if (error < 0) {
		perror("ioctl: BLKGETSIZE");
		exit(-1);
	}

	close(fd);

	journal_size_blocks = journal_size << (20 - 9);
	if (bsize < fs_sb.sb_bsize) {
		while(bsize != fs_sb.sb_bsize) {
			bsize <<= 1;
			size >>= 1;
			journal_size_blocks >>= 1;
		}
	}

	return size;
}
#else
#error No device size code available for this OS
#endif  /*  defined(__linux__)  */

/**
 * jread - Read from journaled file using ioctl()
 * @fd: The fd to read from
 * @buf: The buffer to fill
 * @size: The amount of data to read
 * @offset: The offset to read from
 *
 * Returns: Error code, or amount of data read
 */
int jread(int fd, void *buf, uint64 size, uint64 *offset)
{
  gfs_jio_t jt;
  int ret;

  jt.jio_size = size;
  jt.jio_offset = *offset;
  jt.jio_data = buf;

  ret = ioctl(fd, GFS_JREAD, &jt);
  if (ret > 0)
    (*offset) += ret;

  return ret;
}

/**
 * jwrite - Write to journaled file using ioctl()
 * @fd: The fd to write to
 * @buf: The buffer to write
 * @size: The amount of data to write
 * @offset: The offset at which to write the data
 *
 * Returns: Error code, or the amount of data written
 */
int jwrite(int fd, void *buf, uint64 size, uint64 *offset)
{
  gfs_jio_t jt;
  int ret;

  jt.jio_size = size;
  jt.jio_offset = *offset;
  jt.jio_data = buf;

  ret = ioctl(fd, GFS_JWRITE, &jt);
  if (ret > 0)
    (*offset) += ret;

  return ret;
}


/**
 * filesystem_size - Calculate the size of the filesystem
 *
 * Reads the lists of journals and resource groups in order to
 * work out where the last block of the filesystem is located.
 *
 * Returns: The calculated size
 */
static uint64 filesystem_size(void)
{
	osi_list_head_t *tmp, *head;
	struct rglist_entry *rgl;
	struct jilist_entry *jil;
	uint64 size = 0;
	uint64 extent;

	tmp = head = &rglist_current;
	for(;;) {
		tmp = tmp->next;
		if (tmp == head)
			break;
		rgl = osi_list_entry(tmp, struct rglist_entry, list);
		extent = rgl->ri.ri_addr + rgl->ri.ri_length + rgl->ri.ri_data;
		if (extent > size)
			size = extent;
	}

	tmp = head = &jilist_current;
	for(;;) {
		tmp = tmp->next;
		if (tmp == head)
			break;
		jil = osi_list_entry(tmp, struct jilist_entry, list);
		extent = jil->ji.ji_addr + jil->ji.ji_nsegment * fs_sb.sb_seg_size;
		if (extent > size)
			size = extent;
	}

	return size;
}

/**
 * gfs_jientry - Get journal index entry
 * @fd: The fd of the journal index
 * @offset: The offset at which the journal entry appears
 *
 * Reads a single entry from the journal index file and
 * adds it to the list of current journal entries.
 *
 * Returns: 1 on EOF, 0 otherwise
 */
static int get_jientry(int fd, uint64 *offset)
{
	char buffer[sizeof(gfs_jindex_t)];
	int len = jread(fd, buffer, sizeof(gfs_jindex_t), offset);
	struct jilist_entry *jil;
	if (len != sizeof(gfs_jindex_t)) {
		if (len == 0)
			return 1;

		fprintf(stderr, "Erk! Read odd size from jindex (%d)\n", len);
		exit(-1);
	}
	if ((jil = malloc(sizeof(struct jilist_entry))) == NULL)
		exit(-1);
	memset(jil, 0, sizeof(struct jilist_entry));
	gfs_jindex_in(&jil->ji, buffer);
	osi_list_add(&jil->list, &jilist_current);
	return 0;
}

/**
 * read_journals - Read the whole journal index
 * @fs_fd: An fd for some file or directory within the mounted GFS filesystem
 * @ino: The inode number of the journal index
 *
 */
static void read_journals(int fs_fd, gfs_inum_t *ino)
{
	int fd = ioctl(fs_fd, GFS_OBYNUM, ino);
	uint64 offset = 0;
	if (fd < 0) {
		perror("ioctl: GFS_OBYNUM");
		exit(-1);
	}

	while(get_jientry(fd, &offset) == 0) ;

	close(fd);
}

/**
 * get_rgrp - Read a single rindex entry
 * @fd: The fd for the rindex file
 * @offset: The offset at which to read the rindex entry
 *
 * Reads a single rindex entry and adds it to the list of current
 * resource group entries.
 *
 * Returns: 1 on EOF, 0 otherwise
 */
static int get_rgrp(int fd, uint64 *offset)
{
	char buffer[sizeof(gfs_rindex_t)];
	int len = jread(fd, buffer, sizeof(gfs_rindex_t), offset);
	struct rglist_entry *rgl;
	if (len != sizeof(gfs_rindex_t)) {
		if (len == 0)
			return 1;

		fprintf(stderr, "Erk! Read odd size from rindex (%d)\n", len);
		exit(-1);
	}
	if ((rgl = malloc(sizeof(struct rglist_entry))) == NULL)
		exit(-1);
	memset(rgl, 0, sizeof(struct rglist_entry));
	gfs_rindex_in(&rgl->ri, buffer);
	osi_list_add(&rgl->list, &rglist_current);
	return 0;
}

/**
 * read_rgrps - Reads the contents of the rindex file
 * @fs_fd: An fd for any file or directory within the mounted GFS filesytem
 * @ino: The inode number of the rindex
 *
 */
static void read_rgrps(int fs_fd, gfs_inum_t *ino)
{
	int fd = ioctl(fs_fd, GFS_OBYNUM, ino);
	uint64 offset = 0;

	if (fd < 0) {
		perror("ioctl: GFS_OBYNUM");
		exit(-1);
	}

	while(get_rgrp(fd, &offset) == 0) ;

	close(fd);
}

/**
 * write_a_block - Write a block to the current device
 * @where: The position to write the block (in filesystem blocks)
 *
 * Writes a single disk block to the device. It has a safety check which
 * prevents it writing to the device at a position within the control of
 * the active filesystem. 
 */
static void write_a_block(uint64 where, uint64 seq)
{
	char buffer[4096];
	uint64 fsoffset = where * (uint64)fs_sb.sb_bsize;
	int fd = open(device, O_RDWR);
	gfs_log_header_t lh;

	memset(&lh, 0, sizeof(gfs_log_header_t));

	lh.lh_header.mh_magic = GFS_MAGIC;
	lh.lh_header.mh_type = GFS_METATYPE_LH;
	lh.lh_header.mh_format = GFS_FORMAT_LH;
	lh.lh_flags = GFS_LOG_HEAD_UNMOUNT;
	lh.lh_first = where;
	lh.lh_sequence = seq;

	if (fd < 0) {
		perror(device);
		exit(-1);
	}
	if (where < fssize) {
		fprintf(stderr, "Sanity check failed: Caught trying to write to live filesystem!\n");
		exit(-1);
	}

	memset(buffer, 0, 4096);
	gfs_log_header_out(&lh, buffer);
	gfs_log_header_out(&lh, BLOCK_END(buffer, gfs_log_header_t));

	if (lseek(fd, fsoffset, SEEK_SET) != fsoffset) {
		perror(device);
		exit(-1);
	}
	if (write(fd, buffer, fs_sb.sb_bsize) != fs_sb.sb_bsize) {
		perror("write_a_block");
		exit(-1);
	}
	close(fd);
}

/**
 * write_whole_journal - Write a complete journal to the disk
 * @jil: The information about the journal
 *
 * Write a complete new journal on disk.
 */
static void write_whole_journal(struct jilist_entry *jil)
{
	uint64 seg;
	uint64 offset;

	for(seg = 0; seg < jil->ji.ji_nsegment; seg++) {
		offset = seg * fs_sb.sb_seg_size;
		write_a_block(jil->ji.ji_addr + offset, offset);
	}

	sync();
}

/**
 * get_length - Use stat() to get the length of a file
 * @fd: The fd of the file whose length we wish to know
 *
 * Returns: The length
 */
static uint64 get_length(int fd)
{
	struct stat st;
	if (fstat(fd, &st) < 0) {
		perror("stat");
		fprintf(stderr, "Failed to get size of file. Aborting.\n");
		exit(-1);
	}
	return st.st_size;
}

/**
 * write_jindex - Writes new records to the end of the jindex file
 * @fs_fd: A fd of any file or directory withint the GFS filesystem
 * @ino: The inode number of the jindex
 *
 * This is the critical function in adding journals. It does the
 * actual write to the jindex which causes the GFS filesystem to see the
 * new journals which were previously added.
 */
static void write_jindex(int fs_fd, gfs_inum_t *ino)
{
	osi_list_head_t *tmp, *head;
	struct jilist_entry *jil;
	int fd = ioctl(fs_fd, GFS_OBYNUM, ino);
	char buffer[sizeof(gfs_jindex_t)];
	uint64 orig_length;
	uint64 offset;

	if (fd < 0) {
		perror("ioctl: GFS_OBYNUM");
		exit(-1);
	}

	offset = orig_length = get_length(fd);

	/*
	 * This is the critical section.
	 * If things mess up here, it could be very difficult to put right
	 */
	tmp = head = &jilist_new;
	for(;;) {
		tmp = tmp->next;
		if (tmp == head)
			break;
		jil = osi_list_entry(tmp, struct jilist_entry, list);
		gfs_jindex_out(&jil->ji, buffer);
		if (jwrite(fd, buffer, sizeof(gfs_jindex_t), &offset) != sizeof(gfs_jindex_t)) {
			perror("write: jindex");
			fprintf(stderr, "Aborting...\n");
			goto abort;
		}
	}
	fsync(fd);
	/*
	 * This is the end of the critical section
	 */
	close(fd);
	return;
abort:
	/*
	 * If things did mess up in the critical section, we try to put
	 * them right here by truncating the resource index to its original
	 * length. If this fails, we have no more fall back poistions, so
	 * we let the user know.
	 */
	if (ftruncate(fd, orig_length) < 0) {
		perror("truncate");
		fprintf(stderr, "Critical: Cannot return filesystem to original state. Filesystem corruption may result at next mount\n");
		fprintf(stderr, "*** Correct fault manually before next mount ***\n");
		exit(-1);
	}
	close(fd);
	fprintf(stderr, "Managed to return jindex to original state\n");
	exit(-1);
}

/**
 * write_journals - Write the new journals to disk
 * @fs_fd: An fd from any file or directory on the GFS mounted filesystem
 * @ino: The inode number of the jindex
 *
 * This first writes out the new journal information to the
 * area of the disk beyond the area the filesystem is currently
 * using and then calls write_jindex() to make the filesystem see
 * the newly written journal.
 */
static void write_journals(int fs_fd, gfs_inum_t *ino)
{
	osi_list_head_t *tmp, *head;
	struct jilist_entry *jil;
	tmp = head = &jilist_new;
	for(;;) {
		tmp = tmp->next;
		if (tmp == head)
			break;
		jil = osi_list_entry(tmp, struct jilist_entry, list);
		write_whole_journal(jil);
	}

	sync();
	sync();
	sync();

	write_jindex(fs_fd, ino);

	sync();
	sync();
	sync();
}

/**
 * gather_info - Gathers all the information about the existing filesystem
 *
 */
static void gather_info(void)
{
	char sbuf[4096];
	int fd = open(fspath, O_RDONLY);

	if (fd < 0) {
		perror(fspath);
		exit(-1);
	}
	if (ioctl(fd, GFS_GET_SUPER, (unsigned long)sbuf) < 0) {
		perror("ioctl: GFS_GET_SUPER");
		exit(-1);
	}
	gfs_sb_in(&fs_sb, sbuf);

	read_rgrps(fd, &fs_sb.sb_rindex_di);
	read_journals(fd, &fs_sb.sb_jindex_di);
	close(fd);
	devsize = device_size(device);
	fssize = filesystem_size();

	/*
	 * Round size down to integer number of segments
	 */
	while(journal_size_blocks % fs_sb.sb_seg_size) {
		journal_size_blocks--;
		journal_size -= fs_sb.sb_bsize;
	}
}

/**
 * print_rgrps - Print information about resource groups
 * @lh: The list of resource groups to print
 *
 */
static void print_rgrps(osi_list_head_t *lh)
{
	osi_list_head_t *tmp, *head;
	struct rglist_entry *rgl;
	int n = 0;

	tmp = head = lh;
	for(;;) {
		tmp = tmp->next;
		if (tmp == head)
			break;
		rgl = osi_list_entry(tmp, struct rglist_entry, list);
		n++;
		printf("RI: Addr %"FMT_64"u, RgLen %u, "
			"Start %"FMT_64"u, DataLen %u, BmapLen %u\n", 
			rgl->ri.ri_addr, rgl->ri.ri_length, 
			rgl->ri.ri_data1, rgl->ri.ri_data, rgl->ri.ri_bitbytes);
	}
	printf("RGRP: %d Resource groups in total\n", n);
}

/**
 * print_journals - Print a list of journals
 *
 */
static void print_journals(osi_list_head_t *lh)
{
	osi_list_head_t *tmp, *head;
	struct jilist_entry *jil;
	int n = 0;

	tmp = head = lh;
	for(;;) {
		tmp = tmp->next;
		if (tmp == head)
			break;
		jil = osi_list_entry(tmp, struct jilist_entry, list);
		n++;
		printf("JI: Addr %"FMT_64"u NumSeg %u SegSize %u\n", jil->ji.ji_addr, 
			jil->ji.ji_nsegment, fs_sb.sb_seg_size);
	}
	printf("JRNL: %d Journals in total\n", n);
}

/**
 * print_info - Print out various bits of (interesting?) information
 *
 */
static void print_info(void)
{
	printf("FS: Mount Point: %s\n", fspath);
	printf("FS: Device: %s\n", device);
	printf("FS: Options: %s\n", fsoptions);
	printf("FS: Size: %"FMT_64"u\n", fssize);
	if (verbose > 1) {
		printf("RGRP: Current Resource Group List:\n");
		print_rgrps(&rglist_current);
		printf("JRNL: Current Journal List:\n");
		print_journals(&jilist_current);
	}
	printf("DEV: Size: %"FMT_64"u\n", devsize);
	if (verbose > 1) {
		printf("JRNL: New Journal List:\n");
		print_journals(&jilist_new);
	}
}

/**
 * make_journal - Make a new jilist_entry
 * @offset: The offset at which the new journal will go
 * @size: The size of the new journal in fs blocks
 *
 */
uint64 make_journal(uint64 offset, uint64 size)
{
	struct jilist_entry *jil = malloc(sizeof(struct jilist_entry));
	if (jil == NULL)
		exit(-1);
	memset(jil, 0, sizeof(struct jilist_entry));

	if (offset % fs_sb.sb_seg_size)
	{
	  size -= fs_sb.sb_seg_size - (offset % fs_sb.sb_seg_size);
	  offset += fs_sb.sb_seg_size - (offset % fs_sb.sb_seg_size);
	}

	jil->ji.ji_addr = offset;
	jil->ji.ji_nsegment = size / fs_sb.sb_seg_size;

	osi_list_add(&jil->list, &jilist_new);
	return offset + size;
}

/**
 * create_journals - Create a list of the new journals
 * 
 */
static int create_journals(void)
{
	uint64 offset = fssize;
	int n;

	if ((journal_size_blocks * number_of_journals) > (devsize - fssize)) {
		fprintf(stderr, "Requested size (%"FMT_64"u blocks) greater than available space (%"FMT_64"u blocks)\n", journal_size_blocks * number_of_journals, devsize - fssize);
		return -1;
	}

	for(n = 0; n < number_of_journals; n++)
		offset = make_journal(offset, journal_size_blocks);

	if (offset > devsize) {
		fprintf(stderr, "Calculation error: Out of bounds\n");
		exit(-1);
	}

	return 0;
}


/**
 * update_fs - Actually perform the filesystem update
 *
 */
static void update_fs(void)
{
	int fd = open(fspath, O_RDONLY);
	if (fd < 0) {
		perror(fspath);
		exit(-1);
	}
	if (verbose)
		printf("Preparing to write new FS information...\n");
	write_journals(fd, &fs_sb.sb_jindex_di);
	if (verbose)
		printf("Done.\n");
	close(fd);
}

/**
 * find_fs - Find the filesystem which the user specified
 * @name: The name of a device or mount point
 *
 * Returns: 0 if the filesystem is located, 1 otherwise
 */
static int find_fs(char *name)
{
	FILE *fp = fopen("/proc/mounts", "r");
	char buffer[4096];
	char fstype[80];
	int fsdump, fspass;

	if (fp == NULL) {
		perror("open: /proc/mounts");
		exit(-1);
	}
	while((fgets(buffer, 4095, fp)) != NULL) {
		buffer[4095] = 0;
		if (strstr(buffer, name) == 0)
			continue;
		if (sscanf(buffer, "%s %s %s %s %d %d", device, fspath, fstype,
					fsoptions, &fsdump, &fspass) != 6)
			continue;
		if (strcmp(fstype, "gfs") != 0)
			continue;
		if ((strcmp(device, name) != 0) && (strcmp(fspath, name) != 0))
			continue;
		fclose(fp);
		return 0;
	}
	fprintf(stderr, "GFS Filesystem %s not found\n", name);
	fclose(fp);
	return 1;
}

/**
 * usage - Print out the usage message
 *
 * This function does not include documentation for the -D option
 * since normal users have no use for it at all. The -D option is
 * only for developers. It intended use is in combination with the
 * -t flag to find out what the result would be of trying different
 * device sizes without actually having to try them manually.
 */
static void usage(void)
{
	fprintf(stderr,
		"gfs_jadd: (C) 2001 Sistina Software Inc.\n"
		"Adds journals to a GFS filesystem\n"
		"Usage:\n"
		"gfs_jadd [options] /path/to/filesystem\n"
		"Options:\n"
		"     j <num>   - Number of journals to add (default 1)\n"
		"     J <size>  - Size of journals in MB (minimum 8, default 8)\n"
		"     q         - Quiet, reduce verbosity\n"
		"     t         - Test, do everything except update FS\n"
		"     v         - Verbose, increase verbosity\n"
		);
}

/**
 * main - Tha main function
 * @argc: The argument count
 * @argv: The argument vector
 *
 * Runs through the filesystem expansion code for each of the specified
 * filesystems. Each filesystem specified on the command line has the
 * same options applied to it. You'll need to run the program multiple times
 * if you want to use it on several different filesystems with different
 * options for each. If you forget to specify a filesystem, then it is
 * assumed that the program has run successfully, since its done everything
 * asked of it, and it exits without printing a message.
 *
 * Returns: 0 on success, -1 otherwise
 */
int main(int argc, char *argv[])
{
	int opt;

	while ((opt = getopt(argc, argv, "D:hj:J:vqt?")) != EOF) {
		switch(opt) {
			case 'D': /* This option is for testing only */
				override_device_size = atoi(optarg);
				override_device_size <<= 11;
				break;
			case 'h':
				usage();
				exit(0);
			case 'j':
				number_of_journals = atoi(optarg);
				if (number_of_journals < 1) {
					usage();
					exit(-1);
				}
				break;
			case 'J':
				journal_size = atoi(optarg);
				if (journal_size < 8) {
					usage();
					exit(-1);
				}
				break;
			case 'q':
				if (verbose)
					verbose--;
				break;
			case 't':
				test = 1;
				break;
			case 'v':
				verbose++;
				break;
			case '?':
				usage();
				exit(-1);
		}
	}

	while ((argc - optind) > 0) {
		if (find_fs(argv[optind++]))
			continue;
		gather_info();
		if (fssize > devsize) {
			fprintf(stderr, "Filesystem thinks device is bigger than it really is.... skipping\n");
			continue;
		}
		if (create_journals())
			continue;
		if (verbose)
			print_info();
		if (!test)
			update_fs();
	}

	return 0;
}

