/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portions Copyright 2000-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: debug.c
**
**    Description:
**      GFS debugging code
**
*******************************************************************************
******************************************************************************/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/dir.h>
#include <ctype.h>

#include "global.h"
#include "gfs_debug.h"
#include "gfs_ioctl.h"

#include "gfs_tool.h"





/******************************************************************************
*******************************************************************************
**
** print_debug()
**
** Description:
**   Print available debug flags
**
*******************************************************************************
******************************************************************************/

void print_debug()
{
  int x;

  for (x = 0; gfs_debug_flags[x].name[0]; x++)
    printf("%-30s %u\n", gfs_debug_flags[x].name, gfs_debug_flags[x].len);
}


/******************************************************************************
*******************************************************************************
**
** void set_debug()
**
** Description:
**   This routine uses an ioctl command to change the GFS or pool driver
**   debugging level.
**
** Input(s):
**    argc - 
**  **argv - 
**
*******************************************************************************
******************************************************************************/

void set_debug(int argc, char **argv)
{
  int fd;
  int x, y, arg;
  int found;
  unsigned int flag;


  if (argc < 4)
    die("not enough arguments\n");


  fd = open(argv[2], O_RDONLY);
  if (fd < 0)
    die("can't open %s:  %s\n", argv[2], strerror(errno));


  for (arg = 3; arg < argc; arg++)
  {
    found = FALSE;

    for (x = 0; gfs_debug_flags[x].name[0]; x++)
      if (strcmp(gfs_debug_flags[x].name, argv[arg]) == 0)
      {
	found = TRUE;
	break;
      }

    if (!found)
      die("can't find debug flags %s\n", argv[arg]);

    for (y = 0; y < gfs_debug_flags[x].len; y++)
    {
      flag = gfs_debug_flags[x].flag[y];

      if (ioctl(fd, ((strcmp(argv[1], "dbon") == 0) ? GFS_DEBUG_ON : GFS_DEBUG_OFF), &flag) < 0)
	die("error doing ioctl:  %s\n", strerror(errno));
    }
  }


  close(fd);
}


/******************************************************************************
*******************************************************************************
**
** dump_debug()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void dump_debug(int argc, char **argv)
{
  int fd;
  int x, size;
  char *buf;
  int error;


  if (argc < 3)
    die("not enough arguments\n");


  fd = open(argv[2], O_RDONLY);
  if (fd < 0)
    die("can't open %s:  %s\n", argv[2], strerror(errno));


  error = ioctl(fd, GFS_DEBUG_DUMP, NULL);
  if (error < 0)
    die("error doing ioctl:  %s\n", strerror(errno));

  size = error;


  buf = (char *)malloc(size);
  if (!buf)
    die("out of memory\n");


  error = ioctl(fd, GFS_DEBUG_DUMP, buf);
  if (error < 0)
    die("error doing ioctl:  %s\n", strerror(errno));


  for (x = 0; x < size; x++)
    if (buf[x])
      putchar(buf[x]);


  free(buf);

  close(fd);
}


/******************************************************************************
*******************************************************************************
**
** dump_prof()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void dump_prof(int argc, char **argv)
{
  int fd;
  int x, size;
  gfs_profile_t *stats, *tmp, **p;
  int error;
  int changed;
  const char *name;


  if (argc < 3)
    die("not enough arguments\n");


  fd = open(argv[2], O_RDONLY);
  if (fd < 0)
    die("can't open %s:  %s\n", argv[2], strerror(errno));


  error = ioctl(fd, GFS_PROF_DUMP, NULL);
  if (error < 0)
    die("error doing ioctl:  %s\n", strerror(errno));

  size = error;


  stats = (gfs_profile_t *)malloc(size * sizeof(gfs_profile_t));
  if (!stats)
    die("out of memory\n");


  error = ioctl(fd, GFS_PROF_DUMP, stats);
  if (error < 0)
    die("error doing ioctl:  %s\n", strerror(errno));


  p = (gfs_profile_t **)malloc(size * sizeof(gfs_profile_t *));
  if (!p)
    die("out of memory");

  for (x = 0; x < size; x++)
    p[x] = &stats[x];


  do
  {
    changed = FALSE;

    for (x = 1; x < size; x++)
      if (p[x - 1]->total_micros < p[x]->total_micros)
      {
	tmp = p[x - 1];
	p[x - 1] = p[x];
	p[x] = tmp;
	changed = TRUE;
      }
  }
  while (changed);


  printf("%-20s%-10s%-10s%-10s%20s %s\n\n",
	 "called",
	 "min_micro",
	 "ave_micro",
	 "max_micro",
	 "tot_micro",
	 "function");

  for (x = 0; x < size; x++)
  {
    name = gfs_debug_flags[p[x] - stats].name;
    printf("%-20"FMT_64"u%-10"FMT_64"u%-10u%-10"FMT_64"u%20"FMT_64"u %s\n",
	   p[x]->total_calls,
	   (p[x]->min_micros > p[x]->max_micros) ? 0 : p[x]->min_micros,
	   (unsigned int)((double)p[x]->total_micros / p[x]->total_calls + 0.5),
	   p[x]->max_micros,
	   p[x]->total_micros,
	   name);
  }


  free(stats);

  close(fd);
}

