/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portions Copyright 2000-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: gfs_tool.c
**
**    Description:
**      Tool for changing debug flags in GFS, finding the subpool status,
** 	finding the file status and moving a file to a given destination
** 	subpool.
**
*******************************************************************************
******************************************************************************/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/dir.h>
#include <ctype.h>

#include "global.h"
#include "gfs_ioctl.h"

#define EXTERN
#include "gfs_tool.h"
#include "debug.h"



const char *usage[]=
{
  "Print available debug flags:\n",
  "  gfs_tool dbprint\n",
  "\n",
  "Add debug flags:\n",
  "  gfs_tool dbon <GFS_directory> <debug_flags>\n",
  "\n",
  "Remove debug flags:\n",
  "  gfs_tool dboff <GFS_directory> <debug_flags>\n",
  "\n",
  "Dump incore debug buffer:\n",
  "  gfs_tool dbdump <GFS_directory>\n",
  "\n",
  "Dump profiling stats:\n",
  "  gfs_tool prof <GFS_directory>\n",
  "\n",
  "Print file fragmentation:\n",
  "  gfs_tool frag <filename>\n",
  "\n",
  "Shrink the Linux Dcache for a file system:\n",
  "  gfs_tool shrink <filename>\n",
  "\n",
  "Have GFS dump its lock state:\n",
  "  gfs_tool lockdump <GFS_directory>\n",
  "\n",
  "Do a GFS specific \"df\"\n",
  "  gfs_tool df <GFS_directory>\n",
  "\n",
  ""
};





/******************************************************************************
*******************************************************************************
**
** print_usage()
**
** Description:
**   Print usage
**
*******************************************************************************
******************************************************************************/

static void print_usage()
{
  int x;

  for (x = 0; usage[x][0]; x++)
    printf(usage[x]);
}


/******************************************************************************
*******************************************************************************
**
** void print_frag()
**
** Description:
**   This routine uses an ioctl command to ask GFS to print fragmentation
**   information about a file to the console.
**
** Input(s):
**    argc - 
**  **argv - 
**
*******************************************************************************
******************************************************************************/

void print_frag(int argc, char **argv)
{
  int fd;


  if (argc != 3)
    die("not enough arguments\n");


  fd = open(argv[2], O_RDONLY);
  if (fd < 0)
    die("can't open %s:  %s\n", argv[2], strerror(errno));


  if (ioctl(fd, GFS_PRINT_FRAG, NULL) < 0)
    die("error doing ioctl:  %s\n", strerror(errno));


  close(fd);
}


/******************************************************************************
*******************************************************************************
**
** void shrink()
**
** Description:
**   This routine uses an ioctl command to ask GFS shrink the dcache for a
**   particular filesystem.
**
** Input(s):
**    argc - 
**  **argv - 
**
*******************************************************************************
******************************************************************************/

void shrink(int argc, char **argv)
{
  int fd;


  if (argc != 3)
    die("not enough arguments\n");


  fd = open(argv[2], O_RDONLY);
  if (fd < 0)
    die("can't open %s:  %s\n", argv[2], strerror(errno));


  if (ioctl(fd, GFS_SHRINK, NULL) < 0)
    die("error doing ioctl:  %s\n", strerror(errno));


  close(fd);
}


/******************************************************************************
*******************************************************************************
**
** void dump_lockstate()
**
** Description:
**   This routine uses an ioctl command to ask GFS dump the lockstate for a
**   particular filesystem.
**
** Input(s):
**    argc - 
**  **argv - 
**
*******************************************************************************
******************************************************************************/

void dump_lockstate(int argc, char **argv)
{
  int fd;


  if (argc != 3)
    die("not enough arguments\n");


  fd = open(argv[2], O_RDONLY);
  if (fd < 0)
    die("can't open %s:  %s\n", argv[2], strerror(errno));


  if (ioctl(fd, GFS_LOCK_DUMP, NULL) < 0)
    die("error doing ioctl:  %s\n", strerror(errno));


  close(fd);
}


/******************************************************************************
*******************************************************************************
**
** void do_df()
**
** Description:
**   Do a "df" on a filesystem
**
** Input(s):
**    argc - 
**  **argv - 
**
*******************************************************************************
******************************************************************************/

void do_df(int argc, char **argv)
{
  int fd;
  gfs_usage_t usage;
  uint64 used_data;


  if (argc != 3)
    die("not enough arguments\n");


  fd = open(argv[2], O_RDONLY);
  if (fd < 0)
    die("can't open %s:  %s\n", argv[2], strerror(errno));


  if (ioctl(fd, GFS_STATGFS, &usage) < 0)
    die("error doing ioctl:  %s\n", strerror(errno));


  used_data = usage.total_blocks -
    usage.free -
      (usage.used_dinode + usage.free_dinode) -
	(usage.used_meta + usage.free_meta);


  printf("Block size = %u\n", usage.block_size);
  printf("%-15s%-15s%-15s%-15s%-15s\n", "Type", "Total", "Used", "Free", "use%");
  printf("------------------------------------------------------------------------\n");
  printf("%-15s%-15"FMT_64"u%-15"FMT_64"u%-15"FMT_64"u%u%%\n",
	 "inodes",
	 usage.used_dinode + usage.free_dinode,
	 usage.used_dinode,
	 usage.free_dinode,
	 (unsigned int)(100.0 * usage.used_dinode / (usage.used_dinode + usage.free_dinode) + 0.5));
  printf("%-15s%-15"FMT_64"u%-15"FMT_64"u%-15"FMT_64"u%u%%\n",
	 "metadata",
	 usage.used_meta + usage.free_meta,
	 usage.used_meta,
	 usage.free_meta,
	 (unsigned int)(100.0 * usage.used_meta / (usage.used_meta + usage.free_meta) + 0.5));
  printf("%-15s%-15"FMT_64"u%-15"FMT_64"u%-15"FMT_64"u%u%%\n",
	 "data",
	 used_data + usage.free,
	 used_data,
	 usage.free,
	 (unsigned int)(100.0 * used_data / (used_data + usage.free) + 0.5));


  close(fd);
}


/******************************************************************************
*******************************************************************************
**
** main()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

int main(int argc,char *argv[])
{
  prog_name = argv[0];


  if (argc < 2)
  {
    print_usage();
    exit(EXIT_SUCCESS);
  }


  if (strcmp(argv[1], "dbprint") == 0)
  {
    print_debug();
  }
  else if (strcmp(argv[1], "dbon") == 0 || strcmp(argv[1], "dboff") == 0)
  {
    set_debug(argc, argv);
  }
  else if (strcmp(argv[1], "dbdump") == 0)
  {
    dump_debug(argc, argv);
  }
  else if (strcmp(argv[1], "prof") == 0)
  {
    dump_prof(argc, argv);
  }
  else if (strcmp(argv[1], "frag") == 0)
  {
    print_frag(argc, argv);
  }
  else if (strcmp(argv[1], "shrink") == 0)
  {
    shrink(argc, argv);
  }
  else if (strcmp(argv[1], "lockdump") == 0)
  {
    dump_lockstate(argc, argv);
  }
  else if (strcmp(argv[1], "df") == 0)
  {
    do_df(argc, argv);
  }
  else if (strcmp(argv[1], "-h") == 0)
  {
    print_usage();
  }
  else
    die("bad option %s\n", argv[1]);


  exit(EXIT_SUCCESS);
}

