/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portion Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@sistina.com>
**
**    Module name: gfsck.c
**
**    Description:
**      Filesystem checker.
**
**   This is merely an exhibition, not a competition; so please, no wagering.
**
*******************************************************************************
******************************************************************************/

/*
   Description of gfsck passes.

   Most of the passes follow a very similar proceedure of stepping through
   the file system, checking or recording something at each step.  One 
   resource group is processed at a time, so for the most part, one set of
   rgrp buffers need to be in memory at once. 
   i.e. foreach rgrp; rgrp_read(); check stuff in this rgrp; rgrp_relse();

   [This method is slowed by dinodes which are in one rgrp, but whose blocks
    are in a different rgrp.  The dinode's rgrp is in memory, but to check each
    block of the dinode, another rgrp must be read in, checked for something,
    and then released.  Caching a second MRU rgrp should help considerably.]

   For the current rgrp being checked, a loop will go through the rgrp bitmap
   stopping at each dinode or piece of metadata.  So, the bitmaps form the 
   basis for the set of dinodes which will checked, not the directory 
   structure.


   Pass 0
   ------
   * This is not really a "pass" in that it doesn't iterate over file system 
     elements.  Load the basic (mostly static) fs indeces to use in other 
     passes.

   gfsck_mount() - 
   o The superblock is read
   o The journal index inode and resource index inode are read
   o The journal indeces are read from the jiinode
   o The root dinode is read
   o The GFS incore superblock structure is set up and is used in much the
     same way it's used in GFS.

   gfsck_jindex() - 
   o Using the jiinode data, seek to the start of each journal and verify 
     headers.
   o Basic correlation of journal address's, segment sizes and length
   
   gfsck_rindex() -
   o Read each resource index structure from the riinode
   o Create rgrpd structures for each resource group using indeces
   o Read/verify/release the resource group buffers for each rgrp

   gfsck_block_ranges() -
   o set maximum valid block values for use later in gfsck, based on ri data

   Pass 1
   ------
   * Basic verification of bitmaps' picture of dinodes
   o In rgrp scan (see beginning), stop and read each piece of metadata
   o Verify the header's magic number and type
   o If it's a dinode which is still used check other dinode fields: 
     inode address, rgrp number, type, height and size, depth if a dir
   
   Pass 2
   ------
   * Verify the metadata tree (data and metadata blocks) of each dinode
   o In rgrp scan, stop and read each used dinode
   o For each dinode, call check_metatree() which does the following
   o A structure is created which describes the dinode's metadata.  The
     structure is a group of lists where each list contains the 
     metadata buffers for a subsequent level of the metadata tree.
   o Check that each buffer (representing an indirect block) is allocated
     properly (USEDMETA)
   o The buffers of the last list contain data block pointers.  Each data
     block pointer is verified in range and bitmap allocation.
   o Blocks are counted as they are checked and the dinode blocks count
     is verified.
   o Check the chain of unused dinodes in the rgrp


   Pass 3
   ------
   * Verify that directories have valid entries (and hash table, leaves, etc)
   o In rgrp scan, stop and read each directory dinode
   o If the dir is linear, call check_entries() on stuffed dir entries
   o If the dir is exhash, loop through chunks of the hash table, read each
     leaf block, verify leaf block fields and allocation, call check_entries() 
     on entries in leaf
   o For each dir entry in a block, check_entries() will verify: rec_len,
     hash value and leaf pointer index, filename
   o The dinode of each entry is read and validated
   o A count is kept of entries so the entry count in leaves and the directory
     dinode can be verified
   o A count of the subdirs is kept so that the link count can be verified
     on the directory
   o The "." entry is verified and it verifies that a ".." entry is found
   o As all the dir entries are read, an incore bitmap is marked for each entry
     indicating the dinode exists in a directory.  This bitmap is used later.
   o A list is also created of all dinodes which have more than one directory 
     entry (like hard links create).  This is detected when a dinode is 
     already set in the incore bitmap.  This list is used in a later pass.
 
   Pass 4
   ------
   * Verify that all dinodes have a home in some dir and verify 
     directory/subdir connections in both directions up/down.
   o In rgrp scan, stop and read each used dinode.
   o Check that the dinode is set in the incore bitmap created in pass 3 to
     verify the dinode is connected to a directory somewhere.
   o If the dinode is a directory, verify that the ".." entry is a valid
     parent directory and that the parent has an entry for this dir.
   o If the dinode is not a directory and has a link count other than 1,
     put a record of this dinode in a list for checking later.
     
   Pass 5
   ------
   * Detect directory cycles
   o In rgrp scan, stop and read each directory dinode
   o Loop finding parent directories using the ".." entries until the
     root directory is found.
   o A list of visited directories is kept and checked at each step
     to detect loops.

   Pass 6
   ------
   * Check that dinodes with link counts other than one are ok.  This is based
     on the lists created in passes 3 and 4.  The first list is of dinodes for
     which there were more than one dir entry.  The second list is of dinodes
     with link counts other than one.
   o First, dinodes which are in both lists are checked - the count of dir 
     entries for the dinode should equal the link count.  When this 
     correlation is verified the dinode is removed from both lists.
   o Dinodes remaining in the nlink list should have link counts set to one
   o Dinodes remaining in the dirent list should have their link counts set to
     the number of dir entries for each.

   Pass 7
   ------
   * This includes two passes over the file system.  The first checks for 
     conflicting block allocations.  A part of this is setting bits in the 
     incore bitmap for all blocks (only dinodes were set in pass 3).  The 
     second pass compares this incore bitmap with the ondisk bitmap checking 
     for any inconsistencies.
   o In rgrp scan, stop and read each dinode
   o For every block belonging to a dinode, mark the corresponding bit in the 
     incore bitmap.  If the bit is already set there is an allocation conflict.
     This duplicates most of pass 2 and the part of pass 3 which finds leaf 
     blocks.
   o For every block in each rgrp, compare the incore and ondisk bitmaps
*/



#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>

#include "gfs.h"

#define GFSCK_EXTERN
#include "gfsck.h"
#include "pass0.h"
#include "pass1.h"
#include "pass2.h"
#include "pass3.h"
#include "pass4.h"
#include "pass5.h"
#include "pass6.h"
#include "pass7.h"
#include "recover.h"



int lm_mount(char *proto_name, char *table_name, char *host_data,
	     lm_callback_t cb, lm_fsdata_t *fsdata, lm_lockstruct_t *ls)
{
  return 0;
}


/**
 * print_usage
 *
 */

void print_usage(void)
{
  printf("Usage: \n");

  printf("%s [-Vh] <BlockDevice>\n", prog_name);

  printf("  -V\tPrint program version information, then exit.\n");
  printf("  -h\tPrint this help, then exit.\n");
}



/**
 * gfsck_init
 * @argc:
 * @argv:
 *
 */

int gfsck_init(int argc, char **argv)
{
  int c;


  prog_name = argv[0];


  /* Get arguments */

  while ((c = getopt(argc, argv, "hVsndj")) != -1)
  {
    switch (c)
    {

    case 'h':
      print_usage();
      exit(EXIT_SUCCESS);
      break;


    case 'V':
      printf("gfsck -- %s Release\n", GFS_RELEASE_NAME);
      exit(EXIT_SUCCESS);
      break;

    case 's':
      print_structs = 1;
      break;
      
    case 'n':
      dry_run = 1;
      break;

    case 'd':
#if DEBUG_PRINT
      print_debug = 1;
#endif
      break;


    default:
      die("Unknown option\n");
      break;
    };
  }


  /* debugging ring buffer used in libgfs */
#ifdef DEBUG_PRINT
  memset(gfs_icdb_buf, 0, GFS_ICDB_BUF_LEN);
#endif  /*  DEBUG_PRINT  */


  /* Open device */

  if ((fsfd = open(argv[optind], O_RDWR)) < 0)
    die("Unable to open device: %s\n", argv[optind]);
  

  osi_list_head_init(&links_dirents);
  osi_list_head_init(&links_nlink);
  osi_list_head_init(&bitmaps);

  return(0);
}



/**
 * allocate_bitmaps
 * @sdp:
  *
 */

int allocate_bitmaps(gfs_sbd_t *sdp)
{
  bitmap_list_t *bl;
  gfs_rgrpd_t *rgd;
  osi_list_head_t *tmp;
  uint32 count = 0;


  for (tmp = sdp->sd_rglist.next; tmp != &sdp->sd_rglist; tmp = tmp->next)
  {
    rgd = osi_list_entry(tmp, gfs_rgrpd_t, rd_list);
    count += DIV_RU(rgd->rd_ri.ri_bitbytes, 2);
  }

  printf("Allocating %.1f MB of memory for bitmaps\n", 
	 (float)count / (1024.0 * 1024.0));

  for (tmp = sdp->sd_rglist.next; tmp != &sdp->sd_rglist; tmp = tmp->next)
  {
    rgd = osi_list_entry(tmp, gfs_rgrpd_t, rd_list);
    type_zalloc(bl, bitmap_list_t, 1);
    type_zalloc(bl->bm, char, DIV_RU(rgd->rd_ri.ri_bitbytes, 2));
    bl->rgd = rgd;
    osi_list_add(&bl->list, &bitmaps);
  }
  
  return 0;
}



/**
 * main
 * @argc:
 * @argv:
 *
 */

int main (int argc, char **argv)
{
  gfs_sbd_t *sdp;
  int error;


  gfsck_init(argc, argv);

  type_zalloc(sdp, gfs_sbd_t, 1);

  error = gfsck_mount(sdp);
  if (error)
  {
    printf("gfsck mount failed\n");
    goto out_err;
  }

  printf("Superblock done\n");

  error = gfsck_jindex(sdp);
  if (error)
  {
    printf("gfsck jindex check failed\n");
    goto out_err;
  }

  printf("Journal Index done\n");

  error = gfsck_rindex(sdp);
  if (error)
  {
    printf("gfsck rindex check failed\n");
    goto out_err;
  }

  printf("Resource Index done\n");

  error = gfsck_block_ranges(sdp);
  if (error)
  {
    printf("gfsck block range check failed\n");
    goto out_err;
  }

  allocate_bitmaps(sdp);

  printf("Pass 0:  %d resource groups\n", rg_count);

  pass_1(sdp);

  printf("Pass 1:  done         \n");

  pass_2(sdp);

  printf("Pass 2:  done         \n");

  pass_3(sdp);

  printf("Pass 3:  done         \n");

  pass_4(sdp);

  printf("Pass 4:  done         \n");

  pass_5(sdp);

  printf("Pass 5:  done         \n");

  pass_6(sdp);

  printf("Pass 6:  done         \n");

  pass_7(sdp);

  printf("Pass 7:  done         \n");

  if (print_debug)
    gfs_icdb_print();

  close(fsfd);
  printf("GFSCK Complete.\n");
  return 0;

 out_err:
  printf("GFSCK failed.\n");
  return -1;
}















