/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portion Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@sistina.com>
**
**    Module name: pass0.c  (gfsck)
**
**    Description:
**      Filesystem checker
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "gfsck.h"

#include "super.h"
#include "inode.h"
#include "file.h"
#include "blklist.h"


gfs_glock_t fakegl;


int gfs_ji_update(gfs_inode_t *ip);
gfs_rgrpd_t *gfs_rgrpd_get(gfs_sbd_t *sdp, uint64 n, int create);
int compute_bitstructs(gfs_rgrpd_t *rgd);



/**
 * gfsck_mount
 * @sdp: superblock
 *
 * analagous to gfs_read_super()
 */

int gfsck_mount(gfs_sbd_t *sdp)
{
  gfs_inode_t *ip = NULL;
  int error;


  /*  Make sure there are no invalid buffers lying around  */

  error = ioctl(fsfd, BLKFLSBUF);
  if (error)
    perror("ioctl BLKFLSBUF");

  
  /* sdp->sd_sb.sb_bsize = get_hardblocksize(sb->s_dev); */
  sdp->sd_sb.sb_bsize = 512;

  if (sdp->sd_sb.sb_bsize < GFS_BASIC_BLOCK)
    sdp->sd_sb.sb_bsize = GFS_BASIC_BLOCK;

  /* set_blocksize(sb->s_dev, sdp->sd_sb.sb_bsize); */
  GFS_ASSERT(sizeof(gfs_sb_t) <= sdp->sd_sb.sb_bsize, );


  osi_list_head_init(&sdp->sd_rglist);


  if (gfs_read_sb(sdp, NULL, 0) < 0)
  {
    return -1;
  }
  
  /* set_blocksize(sb->s_dev, sdp->sd_sb.sb_bsize); */
  

  
  /* get ji inode */

  sdp->sd_jiinode = ip = gfs_alloc_inode();
  if (!ip)
    return -1;

  memset(ip, 0, sizeof(gfs_inode_t));

  ip->i_num = sdp->sd_sb.sb_jindex_di;
  ip->i_sbd = sdp;
  ip->i_gl = &fakegl;

  error = gfs_copyin_dinode(ip);
  if (error)
    return error;

  

  /* get ri inode */

  sdp->sd_riinode = ip = gfs_alloc_inode();
  if (!ip)
    return -1;

  memset(ip, 0, sizeof(gfs_inode_t));

  ip->i_num = sdp->sd_sb.sb_rindex_di;
  ip->i_sbd = sdp;
  ip->i_gl = &fakegl;

  error = gfs_copyin_dinode(ip);
  if (error)
    return error;



  /* read in the ji data */

  error = gfs_ji_update(sdp->sd_jiinode);
  if (error)
    return -1;



  /* get root dinode */

  sdp->sd_rooti = ip = gfs_alloc_inode();
  if (!ip)
    return -1;

  memset(ip, 0, sizeof(gfs_inode_t));

  ip->i_num = sdp->sd_sb.sb_root_di;
  ip->i_sbd = sdp;
  ip->i_gl = &fakegl;
 
  error = gfs_copyin_dinode(ip);
  if (error)
    return error;
  
  return 0;
}



/**
 * gfsck_jindex - check journal index validity
 * @sdp: superblock
 *
 */

int gfsck_jindex(gfs_sbd_t *sdp)
{
  gfs_sb_t *sb = &sdp->sd_sb;
  gfs_jindex_t *ji, *prev_ji = NULL;
  gfs_log_header_t header;
  char buf[sdp->sd_sb.sb_bsize];
  uint64 off;
  int i, error;


  for (i = 0; i < sdp->sd_journals; i++)
  {
    ji = &sdp->sd_jindex[i];

    /* just verify we can read a journal header at the beginning of 
       each journal */

    off = lseek(fsfd, ji->ji_addr * sb->sb_bsize, SEEK_SET);
    if (off != ji->ji_addr * sb->sb_bsize)
    {
      warn("Bad seek to journal address: %"FMT_64"u\n", ji->ji_addr);
      return -1;
    }

    error = read(fsfd, &buf, sb->sb_bsize);
    if (error != sb->sb_bsize)
    {
      warn("Bad read at journal address: %"FMT_64"u\n", ji->ji_addr);
      return -1;
    }
    
    gfs_log_header_in(&header, buf);

    if (header.lh_header.mh_magic != GFS_MAGIC ||
	header.lh_header.mh_type != GFS_METATYPE_LH)
    {
      warn("Bad journal address or log header at block %"FMT_64"u\n", 
	   ji->ji_addr);
      return -1;
    }

    if (prev_ji)
    {
      if (prev_ji->ji_addr + (prev_ji->ji_nsegment * sdp->sd_sb.sb_seg_size) >
	  ji->ji_addr)
      {
	warn("Bad journal addr, seg count or seg size\n");
	return -1;
      }
    }
    prev_ji = ji;
  }
 
  return 0;
}



/**
 * gfsck_rindex - check resource index validity
 * @sdp: superblock
 *
 */

int gfsck_rindex(gfs_sbd_t *sdp)
{
  gfs_rgrpd_t *rgd;
  struct osi_list_head *tmp;
  gfs_rindex_t buf;
  unsigned int rg;
  int error, count1 = 0, count2 = 0;


  for (rg = 0; ; rg++)
  {
    error = gfs_internal_read(sdp->sd_riinode, (char *)&buf, 
			      rg * sizeof(gfs_rindex_t), sizeof(gfs_rindex_t));
    if (error != sizeof(gfs_rindex_t))
    {
      break;
    }

    rgd = gfs_rgrpd_get(sdp, gfs64_to_cpu(buf.ri_addr), CREATE);
    if (!rgd)
      return -1;

    gfs_rindex_in(&rgd->rd_ri, (char *)&buf);

    error = compute_bitstructs(rgd);
    if (error)
      return -1;

    count1++;
  }

  for (tmp = sdp->sd_rglist.next; tmp != &sdp->sd_rglist; tmp = tmp->next)
  {
    rgd = osi_list_entry(tmp, gfs_rgrpd_t, rd_list);

    error = gfs_rgrp_read(rgd);
    if (error)
      return -1;

    gfs_rgrp_verify(rgd);
    gfs_rgrp_relse(rgd);
    count2++;
  }

  if (count1 != count2)
    return -1;

  rg_count = count1;

  return 0;
}



/**
 * gfsck_block_ranges
 * @sdp: superblock
 *
 */

int gfsck_block_ranges(gfs_sbd_t *sdp)
{
  gfs_jindex_t *jdesc;
  gfs_rgrpd_t *rgd;
  gfs_rindex_t *ri;
  struct osi_list_head *tmp;
  char buf[sdp->sd_sb.sb_bsize];
  uint64 rmax = 0;
  uint64 jmax = 0;
  uint64 rmin = 0;
  uint64 i;
  uint64 off;
  int error;


  for (tmp = sdp->sd_rglist.next; tmp != &sdp->sd_rglist; tmp = tmp->next)
  {
    rgd = osi_list_entry(tmp, gfs_rgrpd_t, rd_list);
    ri = &rgd->rd_ri;
    if (ri->ri_data1 + ri->ri_data - 1 > rmax)
      rmax = ri->ri_data1 + ri->ri_data - 1;
    if (!rmin || ri->ri_data1 < rmin)
      rmin = ri->ri_data1;
  }


  for (i = 0; i < sdp->sd_journals; i++)
  {
    jdesc = &sdp->sd_jindex[i];

    if ((jdesc->ji_addr + jdesc->ji_nsegment * sdp->sd_sb.sb_seg_size - 1) > jmax)
      jmax = jdesc->ji_addr + jdesc->ji_nsegment * sdp->sd_sb.sb_seg_size - 1;
  }


  last_fs_block = (jmax > rmax) ? jmax : rmax;

  last_data_block = rmax;
  first_data1 = rmin;

  off = lseek(fsfd, last_fs_block * sdp->sd_sb.sb_bsize, SEEK_SET);
  if (off != last_fs_block * sdp->sd_sb.sb_bsize)
  {
    warn("Can't seek to last block in file system: %"FMT_64"u\n", last_fs_block);
    goto fail;
  }
      
  memset(buf, 0, sdp->sd_sb.sb_bsize);
  error = read(fsfd, buf, sdp->sd_sb.sb_bsize);
  if (error != sdp->sd_sb.sb_bsize)
  {
    warn("Can't read last block in file system (%u), last_fs_block: %"FMT_64"u\n",
	 error, last_fs_block);
    goto fail;
  }

  return 0;

 fail:
  return -1;
}














