/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portion Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@sistina.com>
**
**    Module name: pass2.c  (gfsck)
**
**    Description:
**      Filesystem checker
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "gfsck.h"
#include "recover.h"

#include "blklist.h"
#include "dio.h"
#include "bmap.h"
#include "inode.h"



/**
 * build_metalist
 * @ip:
 * @rgd:
 *
 */

int build_metalist(gfs_inode_t *ip, osi_list_head_t *mlp)
{
  uint32 height = ip->i_di.di_height;
  osi_buf_t *bh, *nbh;
  osi_list_head_t *prev_list, *cur_list, *tmp;
  int i, head_size;
  uint64 *ptr, block;


  bh = gfs_dgetblk_arch(ip->i_sbd, ip->i_di.di_num.no_addr, NULL);
  gfs_dreread(ip->i_sbd, bh, DR_START);
  osi_list_add(&bh->b_list, &mlp[0]);

  if (height < 2)
    return 0;

  for (i = 1; i < height; i++)
  {
    prev_list = &mlp[i - 1];
    cur_list = &mlp[i];

    for (tmp = prev_list->next; tmp != prev_list; tmp = tmp->next)
    {
      bh = osi_list_entry(tmp, osi_buf_t, b_list);

      head_size = (i > 1 ? sizeof(gfs_indirect_t) : sizeof(gfs_dinode_t));
      ptr = (uint64 *)(bh->b_data + head_size);

      for ( ; (char *)ptr < (bh->b_data + bh->b_size); ptr++)
      {
	if (!*ptr)
	  continue;

	block = gfs64_to_cpu(*ptr);

	if (check_range(ip->i_sbd, block))
	  return -1;

	nbh = gfs_dgetblk_arch(ip->i_sbd, block, NULL);
	gfs_dreread(ip->i_sbd, nbh, DR_START);
	osi_list_add(&nbh->b_list, cur_list);
      }
    }
  }
  
  return 0;
}



/**
 * check_metatree
 * @ip:
 * @rgd:
 *
 */

int check_metatree(gfs_inode_t *ip, gfs_rgrpd_t *rgd)
{
  osi_list_head_t metalist[GFS_MAX_META_HEIGHT];
  osi_list_head_t *list, *tmp;
  osi_buf_t *bh;
  uint64 block, *ptr;
  uint32 height = ip->i_di.di_height, reg_blocks = 0, indir_blocks = 0;
  int top, bottom, val, error, i, head_size;

  for (i = 0; i < GFS_MAX_META_HEIGHT; i++)
    osi_list_head_init(&metalist[i]);


  /* create metalist for each level */

  error = build_metalist(ip, &metalist[0]);
  if (error)
  {
    warn("bad metadata tree for dinode %"FMT_64"u\n", ip->i_di.di_num.no_addr);
    return -1;
  }


  /* check dinode block */

  val = get_bitmap_bit(ip->i_sbd, ip->i_di.di_num.no_addr, rgd);
  if (val != GFS_BLKST_USEDMETA)
    warn("bad bitmap value for dinode block %"FMT_64"u\n", ip->i_di.di_num.no_addr);

  if (!height)
    goto end;


  /* check indirect blocks */

  top = 1;
  bottom = height - 1;

  while (top <= bottom)
  {
    list = &metalist[top];

    for (tmp = list->next; tmp != list; tmp = tmp->next)
    {
      bh = osi_list_entry(tmp, osi_buf_t, b_list);
      val = get_bitmap_bit(ip->i_sbd, BH_BLKNO(bh), rgd);
      if (val != GFS_BLKST_USEDMETA)
	warn("bad bitmap value for indir block %"FMT_64"u\n", BH_BLKNO(bh));
      indir_blocks++;
    }
    top++;
  }


  /* check data blocks */

  list = &metalist[height - 1];

  for (tmp = list->next; tmp != list; tmp = tmp->next)
  {
    bh = osi_list_entry(tmp, osi_buf_t, b_list);
    
    head_size = (height != 1 ? sizeof(gfs_indirect_t) : sizeof(gfs_dinode_t));
    ptr = (uint64 *)(bh->b_data + head_size);

    for ( ; (char *)ptr < (bh->b_data + bh->b_size); ptr++)
    {
      if (!*ptr)
	continue;
      
      block =  gfs64_to_cpu(*ptr);
      
      if (check_range(ip->i_sbd, block))
	warn("bad range of data block pointer\n");

      val = get_bitmap_bit(ip->i_sbd, block, rgd);
      
      if (ip->i_di.di_flags & GFS_DIF_JDATA)
      {
	if (val != GFS_BLKST_USEDMETA)
	  warn("bad bitmap value\n");
      }
      else
      {
	if (val != GFS_BLKST_USED)
	  warn("bad bitmap value\n");
      }

      reg_blocks++;
    }
  }


 end:
  /* free metalists */

  for (i = 0; i < GFS_MAX_META_HEIGHT; i++)
  {
    list = &metalist[i];
    while (!osi_list_empty(list))
    {
      bh = osi_list_entry(list->next, osi_buf_t, b_list);
      osi_list_del(&bh->b_list);
      gfs_drelse_arch(bh);
    }
  }
 

  /* check block count */
   
  /* can't verify exhash dir block count until we know how many leaves */
  if (ip->i_di.di_type == FILE_DIR && (ip->i_di.di_flags & GFS_DIF_EXHASH))
    return 0;

  if (ip->i_di.di_blocks != (1 + reg_blocks + indir_blocks))
    warn("bad blocks count\n");

  return 0;
}  
  


/**
 * pass_2
 * @sdp: superblock
 *
 */

int pass_2(gfs_sbd_t *sdp)
{
  gfs_rgrpd_t *rgd;
  osi_list_head_t *tmp;
  gfs_inode_t *ip;
  uint64 block, freedi; 
  uint32 start_blk, start_buf, bsize = sdp->sd_sb.sb_bsize, freedi_count;
  int error, cnt = 0, count = 0, first, val;  
  char buf[bsize];
  

  type_zalloc(ip, gfs_inode_t, 1);
  ip->i_sbd = sdp;

  
  for (tmp = sdp->sd_rglist.next; tmp != &sdp->sd_rglist; tmp = tmp->next)
  {
    printf("Pass 2:  %.0f %% \r", 100.0 * ((float)cnt / (float)rg_count));
    fflush(0);
    cnt++;

    rgd = osi_list_entry(tmp, gfs_rgrpd_t, rd_list);

    error = gfs_rgrp_read(rgd);
    if (error)
      return -1;
    
    start_blk = 0;
    start_buf = 0;
    first = 1;


    /* check used dinodes */

    while (1)
    {
      error = next_rg_metatype(sdp, rgd, &start_blk, &start_buf, &block, buf,
			       GFS_METATYPE_DI, &first);
      if (error)
        break;

      first = 0;

      gfs_dinode_in(&ip->i_di, buf);

      if (ip->i_di.di_flags & GFS_DIF_UNUSED)
	continue;

      ip->i_num = ip->i_di.di_num;

      count++;

      error = check_metatree(ip, rgd);
      if (error < 0)
	warn("bad metatree for dinode %"FMT_64"u\n", ip->i_num.no_addr);
    }


    /* check list of unused dinodes */
    
    freedi_count = 0;
    freedi = rgd->rd_rg.rg_freedi_list.no_addr;

    while (freedi)
    {
      gfsck_read_dinode(sdp, &ip->i_di, freedi);
      if (!(ip->i_di.di_flags & GFS_DIF_UNUSED))
	warn("bad unused dinode %"FMT_64"u\n", ip->i_di.di_num.no_addr);

      val = get_bitmap_bit(ip->i_sbd, freedi, rgd);
      if (val != GFS_BLKST_USEDMETA)
	warn("bad unused dinode bitmap value%"FMT_64"u\n", 
	     ip->i_di.di_num.no_addr);

      freedi_count++;
      freedi = ip->i_di.di_next_unused.no_addr;
    }
    
    if (rgd->rd_rg.rg_freedi != freedi_count)
      warn("bad unused dinode count rg %"FMT_64"u\n", rgd->rd_ri.ri_data1);


    gfs_rgrp_relse(rgd);
  }

  free(ip);

  printf("Pass 2:  %d dinodes checked\n", count);
  
  return 0;
}




