/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portion Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@sistina.com>
**
**    Module name: pass3.c  (gfsck)
**
**    Description:
**      Filesystem checker
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "gfsck.h"

#include "blklist.h"
#include "file.h"
#include "dio.h"
#include "fs.h"

#include "pass3.h"




/**
 * record_dinode_dirent
 * @sdp:
 * @ino:
 *
 */

int record_dinode_dirent(gfs_sbd_t *sdp, uint64 ino)
{
  osi_list_head_t *tmp;
  gfs_rgrpd_t *rgd = gfs_blk2rgrpd(sdp, ino);
  bitmap_list_t *bl = NULL;
  uint64 blk;
  char *byte;
  int bit, found = 0;
  di_info_t *dinfo;

  
  for (tmp = (&bitmaps)->next; tmp != &bitmaps; tmp = tmp->next)
  {
    bl = osi_list_entry(tmp, bitmap_list_t, list);
    if (bl->rgd == rgd)
    {
      found = 1;
      break;
    }
  }

  if (!found)
    return -1;

  blk = ino - rgd->rd_ri.ri_data1;

  byte = bl->bm + (blk / 8);
  bit = blk % 8;

  if ((*byte >> bit) & 0x01)
  {
    dinfo = search_list(&links_dirents, ino);

    if (!dinfo)
    {
      type_zalloc(dinfo, di_info_t, 1);
      osi_list_add(&dinfo->din_list, &links_dirents);
      dinfo->din_addr = ino;
      dinfo->din_dirents = 2;
    }
    else
      dinfo->din_dirents++;

    return -1;
  }

  *byte |= (0x01 << bit);

  return 0;
}



/**
 * check_entries - There are two modes of operation depending on the "target"
 *   input within the status structure.  If target is zero, this routine simply
 *   cycles through all entries checking and counting them.  A non-zero
 *   target value is treated as an inode number which is searched for.  In
 *   this case, only the inode numbers of the entries are checked and if the
 *   target inode number is found we return with the status.found field
 *   set to true.
 * 
 * @ip: the inode of the dir being searched
 * @bh: buffer containing entries
 * @rgd: rgrp of the directory (unused in "target" mode)
 * @index: current offset into the hash table - used for verification
 * @type: linear or exhash
 * @status: if "target" field is set, search for the inode specified
 *
 * Returns:
 * status->found - true if target inode was found, ignore if target is zero
 * status->entries - number of entries found
 * status->dot - inode number of . entry if found
 * status->dotdot - inode number of .. entry if found
 * status->subdirs - number of entries found which are subdirs
 * status->remove - errors were found so this dir should be removed
 */

int check_entries(gfs_inode_t *ip, osi_buf_t *bh, gfs_rgrpd_t *rgd, uint32 index, 
		  int type, dir_status_t *status)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  gfs_dirent_t *dent;
  gfs_dirent_t de;
  gfs_leaf_t *leaf = NULL;
  gfs_dinode_t di;
  char *bh_end, *filename;
  uint32 count = 0, off, len;
  int error, val;

  
  if (type == DIR_LINEAR)
    dent = (gfs_dirent_t *)(BH_DATA(bh) + sizeof(gfs_dinode_t));

  else
  {
    dent = (gfs_dirent_t *)(BH_DATA(bh) + sizeof(gfs_leaf_t));
    leaf = (gfs_leaf_t *)BH_DATA(bh);
  }
  bh_end = BH_DATA(bh) + BH_SIZE(bh);
  

  while (1)                /* foreach entry */
  {
    memset(&de, 0, sizeof(gfs_dirent_t));
    gfs_dirent_in(&de, (char *)dent);
    filename = (char *)dent + sizeof(gfs_dirent_t);

    if (!de.de_inum.no_formal_ino)
      goto next;

    
    if (status->target)
    {
      if (status->target == de.de_inum.no_addr)
      {
	status->found = TRUE;
	return 0;
      }
      count++;
      goto next;
    }


    if (de.de_rec_len < GFS_DIR_REC_LEN(de.de_name_len))
    {
      warn("Dir entry with bad record or name length\n");
      goto out;
    }


    if (de.de_hash != gfs_hash(filename, de.de_name_len))
    {
      warn("Dir entry with bad hash or name length\n");
      goto out;
    }

    if (type == DIR_EXHASH)
    {
      off = de.de_hash >> (32 - ip->i_di.di_depth);
      len = 1 << (ip->i_di.di_depth - gfs16_to_cpu(leaf->lf_depth));

      if ((index > off) || (off >= (index + len)))
      {
	warn("Dir %"FMT_64"u entry with bad hash value, index: %u, off: %u\n", 
	     ip->i_num.no_addr, index, off);
	goto next;
      }
    }


    error = check_range(sdp, de.de_inum.no_addr);
    if (error)
    {
      warn("Dir entry with dinode number out of range\n");
      goto out;
    }

    
    val = get_bitmap_bit(sdp, de.de_inum.no_addr, rgd);
    if (val != GFS_BLKST_USEDMETA)
    {
      warn("Dir entry of non-existent dinode\n");
      goto out;
    }

    error = gfsck_read_dinode(sdp, &di, de.de_inum.no_addr);
    if (error)
    {
      warn("bad dinode referenced in dir entry\n");
      goto out;
    }

    if (di.di_flags & GFS_DIF_UNUSED)
    {
      warn("dir entry of unused dinode\n");
      goto out;
    }

    if (de.de_name_len == 1 && !memcmp(filename, ".", 1))
    {
      if (status->dot)
      {
	warn("Duplicate . dir entries\n");
	goto out;
      }      
      status->dot = de.de_inum.no_addr;
    }
    else if (de.de_name_len == 2 && !memcmp(filename, "..", 2))
    {
      if (status->dotdot)
      {
	warn("Duplicate .. dir entries\n");
	goto out;
      }      
      status->dotdot = de.de_inum.no_addr;
    }
    else
    {
      if (di.di_type == FILE_DIR)
	status->subdirs++;
      
      record_dinode_dirent(sdp, de.de_inum.no_addr);
    }


    count++;

  next:
    if ((char *)dent + de.de_rec_len >= bh_end)
      break;

    dent = (gfs_dirent_t *)((char *)dent + de.de_rec_len);
  }



  if ((char *)dent + de.de_rec_len != bh_end)
  {
    warn("Dir entry with bad record length\n");
    goto out;
  }

  if (type == DIR_EXHASH && count != gfs16_to_cpu(leaf->lf_entries))
  {
    warn("Bad leaf (%"FMT_64"u) entry count. found: %u, expected: %u (#%"FMT_64"u)\n", 
	 BH_BLKNO(bh), count, gfs16_to_cpu(leaf->lf_entries), ip->i_num.no_addr);
    goto out;
  }

  status->entries += count;
  return 0;


 out:

  status->remove = TRUE;
  return -1;
}




/**
 * check_leaf
 * @ip:
 *
 */

osi_buf_t *check_leaf(gfs_inode_t *ip, uint64 *lp, uint32 offset, uint32 *len)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  osi_buf_t *bh;
  gfs_leaf_t leaf;
  uint64 leaf_no = gfs64_to_cpu(lp[offset]);
  int i, error;


  error = check_range(sdp, leaf_no);
  if (error)
  {
    warn("Bad range of leaf block\n");
    return(NULL);
  }

  bh = gfs_dread(sdp, leaf_no, NULL, DR_START);

  gfs_leaf_in(&leaf, BH_DATA(bh));

  if (leaf.lf_header.mh_magic != GFS_MAGIC)
  {
    warn("Bad leaf block magic number %u\n", leaf.lf_header.mh_magic);
    gfs_drelse(bh);
    return NULL;
  }

  if (leaf.lf_header.mh_type != GFS_METATYPE_LF)
  {
    warn("Bad leaf block type\n");
    gfs_drelse(bh);
    return NULL;
  }


  *len = 1 << (ip->i_di.di_depth - leaf.lf_depth);


  for (i = 1; (i < *len) && (i < (sdp->sd_hash_ptrs - offset)); i++)
  {
    if (gfs64_to_cpu(lp[offset + i]) != leaf_no)
    {
      warn("Bad leaf pointer\n");
      gfs_drelse(bh);
      return NULL;
    }
  }


  /* Leaf pointers are read in chunks of hash_ptrs.  We may not begin
     reading the next chunk immediately after the current chunk (lp).
     This would happen if the same leaf pointers extend past the end
     of the current chunk.  In this case we verify here that if all the
     pointers to current leaf were not found, the next read will leave
     an unread gap between the end of current chunk and beginning of
     next chunk.  The unread gap is checked in check_skipped_ptrs(). */

  if ((i != *len) && ((offset + *len) <= sdp->sd_hash_ptrs))
  {
    warn("Bad number of leaf pointers\n");
    gfs_drelse(bh);
    return NULL;
  }

  return bh;
}



/**
 * dir_exhash_scan
 * @ip:
 *
 */

int dir_exhash_scan(gfs_inode_t *ip, gfs_rgrpd_t *rgd, dir_status_t *status)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  osi_buf_t *bh = NULL;
  uint32 hsize, len, ht_offset, lp_offset;
  uint32 hbsize = sdp->sd_hash_bsize;
  uint64 *lp;
  int error, val;


  hsize = 1 << ip->i_di.di_depth;

  /* even though we read in an hbsize chunk of pointers at a time which 
     corresponds to one block of pointers, the chunk of pointers actually 
     read may not correspond to one block since we read pointer chunks 
     (into the lp array) at ht_offset positions in the file.  ht_offset is 
     not always aligned to beginning of block chunks. i.e. the pointers in 
     the lp array could be from two different fs blocks */

  type_zalloc(lp, uint64, hbsize);
  
  ht_offset = 0;
  lp_offset = 0;   
  

  while (ht_offset < hsize)
  {
    if (!lp_offset)
    {
      error = gfs_internal_read(ip, (char *)lp, ht_offset * sizeof(uint64), hbsize);
      if (error != hbsize)
	goto fail;
    }


    bh = check_leaf(ip, lp, lp_offset, &len);
    if (!bh)
      goto fail;

    if (!status->target)
    {
      val = get_bitmap_bit(sdp, BH_BLKNO(bh), rgd);
      if (val != GFS_BLKST_USEDMETA)
      {
	warn("bad bitmap value for leaf block\n");
	goto fail;
      }
    }

    error = check_entries(ip, bh, rgd, ht_offset, DIR_EXHASH, status);
    if (error)
      goto fail;

    gfs_drelse(bh);


    ht_offset += len;
    lp_offset += len;


    if (lp_offset == sdp->sd_hash_ptrs)
      lp_offset = 0;

    else if (lp_offset > sdp->sd_hash_ptrs)
    {
      /* check_skipped_ptrs(ip, ht_offset - len, len); */
      lp_offset = 0;
    }
  }
  
  free(lp);

  return 0;


 fail:

  if (bh)
    gfs_drelse(bh);

  free(lp);
  status->remove = TRUE;
  return -1;
}



/**
 * dir_linear_scan
 * @ip:
 *
 */

int dir_linear_scan(gfs_inode_t *ip, gfs_rgrpd_t *rgd, dir_status_t *status)
{
  osi_buf_t *bh;

  bh = gfs_dread_arch(ip->i_sbd, ip->i_num.no_addr, NULL, DR_START);

  check_entries(ip, bh, rgd, 0, DIR_LINEAR, status);

  gfs_drelse(bh);

  return 0;
}



/**
 * pass_3
 * @sdp:
 *
 */

int pass_3(gfs_sbd_t *sdp)
{
  gfs_rgrpd_t *rgd;
  dir_status_t *status;
  osi_list_head_t *tmp;
  gfs_inode_t *ip;
  uint64 block; 
  uint32 start_blk, start_buf, bsize = sdp->sd_sb.sb_bsize;
  int error, cnt = 0, count = 0, first;  
  char buf[bsize];



  type_zalloc(ip, gfs_inode_t, 1);
  type_zalloc(status, dir_status_t, 1);

  ip->i_sbd = sdp;


  for (tmp = sdp->sd_rglist.next; tmp != &sdp->sd_rglist; tmp = tmp->next)
  {
    printf("Pass 3:  %.0f %% \r", 100.0 * ((float)cnt / (float)rg_count));
    fflush(0);
    cnt++;

    rgd = osi_list_entry(tmp, gfs_rgrpd_t, rd_list);

    error = gfs_rgrp_read(rgd);
    if (error)
      return -1;
    
    start_blk = 0;
    start_buf = 0;
    first = 1;

    while (1)
    {
      error = next_rg_metatype(sdp, rgd, &start_blk, &start_buf, &block, buf,
                               GFS_METATYPE_DI, &first);
      if (error)
        break;

      first = 0;

      gfs_dinode_in(&ip->i_di, buf);

      if (ip->i_di.di_flags & GFS_DIF_UNUSED)
        continue;

      if (ip->i_di.di_type != FILE_DIR)
	continue;
      
      ip->i_num = ip->i_di.di_num;
      count++;
      memset(status, 0, sizeof(dir_status_t));

      if (ip->i_di.di_flags & GFS_DIF_EXHASH)
	dir_exhash_scan(ip, rgd, status);
      else
	dir_linear_scan(ip, rgd, status);

      if (!status->remove)
      {	
	if (!status->dot)
	{
	  warn("No . entry found\n");
	  status->remove = TRUE;
	}

	if (status->dot != ip->i_di.di_num.no_addr)
	{
	  warn("Bad . address\n");
	  status->remove = TRUE;
	}

	if (!status->dotdot)
	{
	  warn("No .. entry found\n");
	  status->remove = TRUE;
	}

	if (status->entries != ip->i_di.di_entries)
	{
	  warn("Incorrect number of dir entries found\n");
	  status->remove = TRUE;
	}

	if (status->subdirs + 2 != ip->i_di.di_nlink)
	{
	  warn("Bad link count\n");
	  status->remove = TRUE;
	}
      }
    }

    gfs_rgrp_relse(rgd);
  }

  free(status);
  free(ip);

  printf("Pass 3:  %d directories checked\n", count);

  return 0;
}











