/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portion Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@sistina.com>
**
**    Module name: pass4.c  (gfsck)
**
**    Description:
**      Filesystem checker
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "gfsck.h"

#include "blklist.h"
#include "dir.h"
#include "fs.h"

#include "pass3.h"


/**
 * check_dotdot_entry
 * @sdp:
 * @ip:
 *
 */

int check_dotdot_entry(gfs_sbd_t *sdp, gfs_inode_t *ip)
{
  osi_qstr_t filename;
  gfs_inode_t *pip;
  dir_status_t *status;
  gfs_inum_t inum;
  int error, type = 0;


  memset(&filename, 0, sizeof(osi_qstr_t));
  filename.name = "..";
  filename.len = 2;
  
  error = gfs_dir_search(ip, &filename, &inum, &type);
  if (error < 0)
    die("check_dotdot_entry: this shouldn't happen\n");

  if (!inum.no_addr)
    die("check_dotdot_entry: this shouldn't happen\n");


  /* get the parent dinode */


  type_zalloc(pip, gfs_inode_t, 1);  
  pip->i_sbd = sdp;  
  memset(&pip->i_di, 0, sizeof(gfs_dinode_t));

  error = gfsck_read_dinode(sdp, &pip->i_di, inum.no_addr);
  if (error)
    die("error reading dinode");

  pip->i_num = pip->i_di.di_num;


  /* search parent for ip's entry */


  type_zalloc(status, dir_status_t, 1);
  status->target = inum.no_addr;
  
  if (pip->i_di.di_flags & GFS_DIF_EXHASH)
    dir_exhash_scan(pip, NULL, status);
  else
    dir_linear_scan(pip, NULL, status);
  

  if (!status->found)
  {
    warn("Bad directory linkage (#%"FMT_64"u) putting in lost+found\n", 
	 ip->i_num.no_addr);

    /* set new .. of ip, insert into l+f */

    /* if .. was bad, what will happen to real parent which has entry
       for this dinode - we need to clear that entry or fix .. if we're 
       sure parent entry is correct and .. is wrong */
  }


  free(pip);
  free(status);

  return 0;
}



/**
 * pass_4
 * @sdp:
 *
 */

int pass_4(gfs_sbd_t *sdp)
{
  gfs_rgrpd_t *rgd;
  osi_list_head_t *tmp;
  gfs_inode_t *ip;
  di_info_t *dinfo;
  uint64 block; 
  uint32 start_blk, start_buf, bsize = sdp->sd_sb.sb_bsize;
  int error, cnt = 0, count = 0, first, val;  
  char buf[bsize];
  

  type_zalloc(ip, gfs_inode_t, 1);
  ip->i_sbd = sdp;

  
  for (tmp = sdp->sd_rglist.next; tmp != &sdp->sd_rglist; tmp = tmp->next)
  {
    printf("Pass 4:  %.0f %% \r", 100.0 * ((float)cnt / (float)rg_count));
    fflush(0);
    cnt++;

    rgd = osi_list_entry(tmp, gfs_rgrpd_t, rd_list);

    error = gfs_rgrp_read(rgd);
    if (error)
      return -1;
    
    start_blk = 0;
    start_buf = 0;
    first = 1;

    while (1)
    {
      error = next_rg_metatype(sdp, rgd, &start_blk, &start_buf, &block, buf,
			       GFS_METATYPE_DI, &first);
      if (error)
        break;

      first = 0;

      gfs_dinode_in(&ip->i_di, buf);

      if (ip->i_di.di_flags & GFS_DIF_UNUSED)
	continue;
      
      ip->i_num = ip->i_di.di_num;

      count++;


      val = check_bitmap_set(sdp, ip->i_num.no_addr);
      if (!val)
      {
	if (ip->i_num.no_addr > sdp->sd_sb.sb_root_di.no_addr)
	{
	  warn("No connection to dinode #%"FMT_64"u\n", ip->i_num.no_addr);
	  /* put into l+f */
	  continue;
	}
      }


      if (ip->i_di.di_type == FILE_DIR)
	check_dotdot_entry(sdp, ip);

      else if (ip->i_di.di_nlink != 1)      /* this list used later on */
      {
	type_zalloc(dinfo, di_info_t, 1);
	osi_list_add(&dinfo->din_list, &links_nlink);
	dinfo->din_addr = ip->i_num.no_addr;
	dinfo->din_nlink = ip->i_di.di_nlink;
      }
    }
   
    gfs_rgrp_relse(rgd);
  }

  free(ip);

  printf("Pass 4:  %d dinodes checked\n", count);
  
  return 0;
}




