/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portion Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@sistina.com>
**
**    Module name: pass7.c  (gfsck)
**
**    Description:
**      Filesystem checker
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "gfsck.h"

#include "blklist.h"
#include "fs.h"
#include "dio.h"
#include "file.h"
#include "inode.h"

#include "pass2.h"



/**
 * set_bitmap
 * @sdp:
 *
 */

int set_bitmap(gfs_sbd_t *sdp, uint64 addr)
{
  osi_list_head_t *tmp;
  gfs_rgrpd_t *rgd = gfs_blk2rgrpd(sdp, addr);
  bitmap_list_t *bl = NULL;
  uint64 blk;
  char *byte;
  int bit, found = 0;

  
  for (tmp = (&bitmaps)->next; tmp != &bitmaps; tmp = tmp->next)
  {
    bl = osi_list_entry(tmp, bitmap_list_t, list);
    if (bl->rgd == rgd)
    {
      found = 1;
      break;
    }
  }

  if (!found)
    return -1;

  blk = addr - rgd->rd_ri.ri_data1;

  byte = bl->bm + (blk / 8);
  bit = blk % 8;

  if ((*byte >> bit) & 0x01)
    return -1;

  *byte |= (0x01 << bit);

  return 0;
}




/**
 * check_block_conflicts
 * @sdp:
 *
 */

int check_block_conflicts(gfs_inode_t *ip)
{
  gfs_sbd_t *sdp = ip->i_sbd;

  osi_list_head_t metalist[GFS_MAX_META_HEIGHT];
  osi_list_head_t *list, *tmp;
  osi_buf_t *bh;
  gfs_leaf_t leaf;
  uint32 hsize, len, ht_offset, lp_offset;
  uint32 hbsize = sdp->sd_hash_bsize;
  uint32 height = ip->i_di.di_height;
  uint64 *lp, leaf_no, *ptr;
  int top, bottom, head_size;
  int error, i;


  if (gfs_is_stuffed(ip))
    goto leaves;


  for (i = 0; i < GFS_MAX_META_HEIGHT; i++)
    osi_list_head_init(&metalist[i]);


  /* create metalist for each level */

  error = build_metalist(ip, &metalist[0]);
  if (error)
  {
    warn("bad metadata tree for dinode %"FMT_64"u\n", ip->i_di.di_num.no_addr);
    return -1;
  }


  /* check indirect blocks */

  top = 1;
  bottom = height - 1;

  while (top <= bottom)
  {
    list = &metalist[top];

    for (tmp = list->next; tmp != list; tmp = tmp->next)
    {
      bh = osi_list_entry(tmp, osi_buf_t, b_list);
      error = set_bitmap(sdp, BH_BLKNO(bh));
      if (error)
	return -1;

    }
    top++;
  }


  /* check data blocks */

  list = &metalist[height - 1];

  for (tmp = list->next; tmp != list; tmp = tmp->next)
  {
    bh = osi_list_entry(tmp, osi_buf_t, b_list);
    
    head_size = (height != 1 ? sizeof(gfs_indirect_t) : sizeof(gfs_dinode_t));
    ptr = (uint64 *)(bh->b_data + head_size);

    for ( ; (char *)ptr < (bh->b_data + bh->b_size); ptr++)
    {
      if (!*ptr)
        continue;
      
      error = set_bitmap(sdp, gfs64_to_cpu(*ptr));
      if (error)
	return -1;
    }
  }


  /* free metalists */

  for (i = 0; i < GFS_MAX_META_HEIGHT; i++)
  {
    list = &metalist[i];
    while (!osi_list_empty(list))
    {
      bh = osi_list_entry(list->next, osi_buf_t, b_list);
      osi_list_del(&bh->b_list);
      gfs_drelse_arch(bh);
    }
  }


 leaves:  

  /* check leaf blocks of directories; leaf block pointers are only accessible
     through the hash table.  this code is basically copied from pass 3 */

  if (ip->i_di.di_type == FILE_DIR && (ip->i_di.di_flags & GFS_DIF_EXHASH))
  {
    hsize = 1 << ip->i_di.di_depth;
    
    type_zalloc(lp, uint64, hbsize);
    
    ht_offset = 0;
    lp_offset = 0;   
    
    while (ht_offset < hsize)
    {
      if (!lp_offset)
      {
	error = gfs_internal_read(ip, (char *)lp, ht_offset * sizeof(uint64), hbsize);
	if (error != hbsize)
	  die("bad hash table read\n");
      }

      leaf_no = gfs64_to_cpu(lp[lp_offset]);

      error = set_bitmap(sdp, leaf_no);
      if (error)
	return -1;

      /* the leaf is just read to get the depth */

      bh = gfs_dread(sdp, leaf_no, NULL, DR_START);

      gfs_leaf_in(&leaf, BH_DATA(bh));

      len = 1 << (ip->i_di.di_depth - leaf.lf_depth);

      ht_offset += len;
      lp_offset += len;

      if (lp_offset == sdp->sd_hash_ptrs)
	lp_offset = 0;

      else if (lp_offset > sdp->sd_hash_ptrs)
	lp_offset = 0;
    }
  }
  
  return 0;
}




/**
 * pass_7
 * @sdp:
 *
 */

int pass_7(gfs_sbd_t *sdp)
{
  gfs_rgrpd_t *rgd;
  osi_list_head_t *tmp;
  gfs_inode_t *ip;
  uint64 block, x; 
  uint32 start_blk, start_buf, bsize = sdp->sd_sb.sb_bsize;
  int error, cnt = 0, count = 0, first, val1, val2, diffs = 0;  
  char buf[bsize];
  

  type_zalloc(ip, gfs_inode_t, 1);
  ip->i_sbd = sdp;


  /* Part a */
  
  for (tmp = sdp->sd_rglist.next; tmp != &sdp->sd_rglist; tmp = tmp->next)
  {
    printf("Pass 7a:  %.0f %% \r", 100.0 * ((float)cnt / (float)rg_count));
    fflush(0);
    cnt++;

    rgd = osi_list_entry(tmp, gfs_rgrpd_t, rd_list);

    error = gfs_rgrp_read(rgd);
    if (error)
      return -1;
    
    start_blk = 0;
    start_buf = 0;
    first = 1;

    while (1)
    {
      error = next_rg_metatype(sdp, rgd, &start_blk, &start_buf, &block, buf,
			       GFS_METATYPE_DI, &first);
      if (error)
        break;

      first = 0;

      gfs_dinode_in(&ip->i_di, buf);

      if (ip->i_di.di_flags & GFS_DIF_UNUSED)
      {
	/* used dinodes are marked as USEDMETA in ondisk bitmap */
	error = set_bitmap(sdp, ip->i_di.di_num.no_addr);
	if (error)
	  die("don't know what to do\n");
	continue;
      }

      ip->i_num = ip->i_di.di_num;

      count++;

      error = check_block_conflicts(ip);
      if (error)
	warn("Block conflict between files (#%"FMT_64"u)\n", 
	     ip->i_num.no_addr);
    }
   
    gfs_rgrp_relse(rgd);
  }

  free(ip);


  /* Part b */
  
  cnt = 0;
  for (tmp = sdp->sd_rglist.next; tmp != &sdp->sd_rglist; tmp = tmp->next)
  {
    printf("Pass 7b:  %.0f %% \r", 100.0 * ((float)cnt / (float)rg_count));
    fflush(0);
    cnt++;

    rgd = osi_list_entry(tmp, gfs_rgrpd_t, rd_list);

    error = gfs_rgrp_read(rgd);
    if (error)
      return -1;
    
    diffs = 0;
    for (x = rgd->rd_ri.ri_data1; x < rgd->rd_ri.ri_data1 + rgd->rd_ri.ri_data; x++)
    {
      val1 = get_bitmap_bit(sdp, x, rgd);
      val2 = check_bitmap_set(sdp, x);

      if (!val1 && !val2)
	continue;

      if (val2 && (val1 == GFS_BLKST_USED || val1 == GFS_BLKST_USEDMETA))
	continue;

      if (!val2 && (val1 == GFS_BLKST_FREEMETA))
	continue;

      if (rgd->rd_ri.ri_data1 == first_data1 && x <= sdp->sd_sb.sb_root_di.no_addr)
	continue;


      warn("On-disk bitmap and calculated bitmap differ. block %"FMT_64"u, %u,%u\n", x,
	   val1, val2);

      diffs++;
    }
    if (diffs)
      warn("Bitmaps diffs on %u blocks in rg %"FMT_64"u\n", diffs, rgd->rd_ri.ri_data1);
   
    gfs_rgrp_relse(rgd);
  }

  return 0;
}


















