/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portion Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@sistina.com>
**
**    Module name: util.c  (gfsck)
**
**    Description:
**      Filesystem checker
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "gfsck.h"

#include "blklist.h"
#include "bits.h"
#include "dio.h"
#include "fs.h"
#include "inode.h"


/* copied from bmap.c */

struct metapath
{
  uint64              mp_list[GFS_MAX_META_HEIGHT];
};
typedef struct metapath metapath_t;


/* prototypes not in any header files */

unsigned int gfs_calc_tree_height(gfs_inode_t *ip, uint64 size);
void find_metapath(gfs_inode_t *ip, metapath_t *mp, uint64 block);
int get_metablock(gfs_inode_t *ip,
		  osi_buf_t *bh, int level, metapath_t *mp,
		  int create, int *new, uint64 *block);
int get_datablock(gfs_inode_t *ip,
		  osi_buf_t *bh, metapath_t *mp,
		  int create, int *new, uint64 *block);




/**
 * search_list
 *
 */

di_info_t *search_list(osi_list_head_t *list, uint64 addr)
{
  osi_list_head_t *tmp;
  di_info_t *dinfo;


  for (tmp = list->next; tmp != list; tmp = tmp->next)
  {
    dinfo = osi_list_entry(tmp, di_info_t, din_list);

    if (dinfo->din_addr == addr)
      return(dinfo);
  }

  return NULL;
}



/**
 * compute_height
 * @sz:
 * @bsize:
 *
 */

int compute_height(gfs_sbd_t *sdp, uint64 sz)
{
  unsigned int height;
  uint64 space, old_space;
  unsigned int bsize = sdp->sd_sb.sb_bsize;


  if (sz <= (bsize - sizeof(gfs_dinode_t)))
    return 0;
  
  height = 1;
  space = sdp->sd_diptrs * bsize;

  while (sz > space)
  {
    old_space = space;

    height++;
    space *= sdp->sd_inptrs;

    if (space / sdp->sd_inptrs != old_space ||
        space % sdp->sd_inptrs != 0)
      break;
  }

  return height;
}



/**
 * next_rg_meta
 * @rgd:
 * @prev_bhi:
 * @prev_blk:
 * @block:
 *
 */

int next_rg_meta(gfs_rgrpd_t *rgd, uint32 *prev_bhi, uint32 *prev_blk, 
		 uint64 *block, int *first)
{
  gfs_bitmap_t *bits = NULL;
  uint32 length = rgd->rd_ri.ri_length;
  uint32 bhi = *prev_bhi;
  uint32 blk = 0;


  if (!*first)
    blk = *prev_blk + 1;

  while (bhi < length)
  {
    bits = &rgd->rd_bits[bhi];

    blk = gfs_bitfit(BH_DATA(rgd->rd_bh[bhi]) + bits->bi_offset,
		     bits->bi_len, blk, GFS_BLKST_USEDMETA);
    if (blk == BFITNOENT)
      goto next;

    *prev_bhi = bhi;
    *prev_blk = blk;

    *block = blk + (bits->bi_start * GFS_NBBY) + rgd->rd_ri.ri_data1;

    return 0;

  next:
    bhi++;
    blk = 0;
  }

  return -1;
}



/**
 * next_rg_metatype
 * @rgd:
 * @prev_bhi:
 * @prev_blk:
 * @block:
 * @outbuf: the final fs buffer
 * @type: the type of metadata we're looking for
 * @first: if set we should start at block zero
 *
 */

int next_rg_metatype(gfs_sbd_t *sdp, gfs_rgrpd_t *rgd, uint32 *prev_bhi, 
		     uint32 *prev_blk, uint64 *block, char *outbuf, uint32 type,
		     int *first)
{

  uint32 bsize = sdp->sd_sb.sb_bsize;
  char buf[sdp->sd_sb.sb_bsize];
  int error;  
  gfs_meta_header_t mheader, *mh = &mheader;


 retry:

  error = next_rg_meta(rgd, prev_blk, prev_bhi, block, first);
  if (error)
    return -1;

  memset(buf, 0, bsize);
  
  if (do_lseek(fsfd, *block * bsize))
  {
    warn("Bad lseek (#%"FMT_64"u)\n", *block);
    return -1;
  }
  
  if (do_read(fsfd, buf, bsize))
  {
    warn("Bad read (#%"FMT_64"u)\n", *block);
    return -1;
  }
  
  gfs_meta_header_in(mh, buf);
  
  if (mh->mh_magic != GFS_MAGIC)
  {
    warn("Bad meta magic (#%"FMT_64"u)\n", *block);
    return -1;
  }
  
  if (mh->mh_type != type)
  {
    *first = 0;
    goto retry;
  }

  memcpy(outbuf, buf, bsize);

  return 0;
}



/**
 * get_bitmap_bit
 * @sdp:
 * @blkno:
 * @cur_rgd:
 *
 */

int get_bitmap_bit(gfs_sbd_t *sdp, uint64 blkno, gfs_rgrpd_t *cur_rgd)
{
  gfs_rgrpd_t *rgd;
  gfs_rindex_t *ri;
  uint32 blk, count = 0, goal, i;
  gfs_bitmap_t *bits = NULL;
  unsigned char *buffer, *byte, bit_status;
  int bit, error;


  rgd = gfs_blk2rgrpd(sdp, blkno);

  ri = &rgd->rd_ri;
  
  blk = blkno - ri->ri_data1;
  
  for (i = 0; i < ri->ri_length; i++)
  {
    bits = &rgd->rd_bits[i];

    count += bits->bi_len * GFS_NBBY;

    if (blk < count)
      break;
  }

  if (i == ri->ri_length)
    die("check_bit: this shouldn't happen - blk: %u\n", blk);

  goal = blk - (count - (bits->bi_len * GFS_NBBY));

  if (cur_rgd != rgd)
  {
    error = gfs_rgrp_read(rgd);
    if (error)
      return -1;
  }

  buffer = BH_DATA(rgd->rd_bh[i]) + bits->bi_offset;
  byte = buffer + (goal / GFS_NBBY);
  bit = (goal % GFS_NBBY) * GFS_BIT_SIZE;

  bit_status = (*byte >> bit) & GFS_BIT_MASK;

  if (cur_rgd != rgd)
    gfs_rgrp_relse(rgd);

  return bit_status;
}



/**
 * check_range
 * @sdp:
 * @blk:
 *
 */

int check_range(gfs_sbd_t *sdp, uint64 blk)
{
  struct osi_list_head *tmp, *head;
  gfs_rgrpd_t *rgd = NULL;
  gfs_rindex_t *ri;


  if (blk > last_data_block)
    return -1;

  head = tmp = &sdp->sd_rglist;
  for (;;)
  {
    rgd = NULL;
    tmp = tmp->next;
    if (tmp == head)
      break;

    rgd = osi_list_entry(tmp, gfs_rgrpd_t, rd_list);
    ri = &rgd->rd_ri;

    if (blk >= (ri->ri_data1 + ri->ri_data))
      continue;
    if (blk >= ri->ri_data1)
      break;
  }
  
  if (!rgd)
    return -1;
  
  return 0;
}



/**
 * gfsck_read_dinode
 * @sdp:
 * @di:
 * @num:
 *
 */

int gfsck_read_dinode(gfs_sbd_t *sdp, gfs_dinode_t *di, uint64 num)
{
  char buf[sdp->sd_sb.sb_bsize];

  if (do_lseek(fsfd, num * sdp->sd_sb.sb_bsize))
    return -1;

  if (do_read(fsfd, buf, sizeof(gfs_dinode_t)))
    return -1;

  gfs_dinode_in(di, buf);

  if (di->di_header.mh_magic != GFS_MAGIC)
    return -1;

  if (di->di_header.mh_type != GFS_METATYPE_DI)
    return -1;

  return 0;
}



/**
 * check_bitmap_set
 * @sdp:
 * @addr:
 *
 * Returns: 1 if the blk is marked in bitmap, 0 if not
 */

int check_bitmap_set(gfs_sbd_t *sdp, uint64 addr)
{
  osi_list_head_t *tmp;
  gfs_rgrpd_t *rgd = gfs_blk2rgrpd(sdp, addr);
  bitmap_list_t *bl = NULL;
  uint64 blk;
  char *byte;
  int bit, found = 0;

  
  for (tmp = (&bitmaps)->next; tmp != &bitmaps; tmp = tmp->next)
  {
    bl = osi_list_entry(tmp, bitmap_list_t, list);
    if (bl->rgd == rgd)
    {
      found = 1;
      break;
    }
  }

  if (!found)
    return -1;

  blk = addr - rgd->rd_ri.ri_data1;

  byte = bl->bm + (blk / 8);
  bit = blk % 8;

  if ((*byte >> bit) & 0x01)
    return 1;

  return 0;
}


