/******************************************************************************
*******************************************************************************
**
**    Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@sistina.com>
**
**    Module name: main.c
**
**    Description:
**      Translate between gfscf.cf file (describes GFS cluster) and 
**      Cluster Information Device, cidev, (source of cluster info for memexp
**      locking module).
**      Nothing stomith-specific in this file!
**
*******************************************************************************
******************************************************************************/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/mount.h>

#include "cluster_config.h"

#define EXTERN
#include "gfsconf.h"
#include "interface.h"
#include "stomith_specific.h"



/**
 * add_stomith_2_node
 *
 * @params: configuration description
 * @node:
 * @methods:
 *
 */

int add_stomith_2_node(param_t *params, node_list_t *node, keypair_t *methods)
{
  stomith_list_t *new = NULL, *s, *sl = NULL;
  int error;
  keypair_t *pair;


  for (pair = methods; pair; pair = pair->next)
  {
    /* check if the name corresponds to a previously configured stomith method */

    for (sl = params->stomith_list; sl; sl = sl->next)
      if (!strcmp(pair->key, sl->name))
      {
	/* check if this new stomith method will fit into the node's block */

	if (node->length + ondisk_stomith_size(sl->magic) > 512)
	{
	  printf("%s: Limit of stomith capabilities reached for node with cid %d.\n"
		  "Skipping stomith option: %s\n", prog_name, node->cid, pair->key);
	  goto next;
	}
      
	type_zalloc(new, stomith_list_t, 1);

	memcpy(new, sl, sizeof(stomith_list_t));

	error = addparam_stomith_specific(node, new, pair->value);    
	if (error)
	{
	  free(new);
	  goto next;
	}

	node->nstomith++;
	node->length += ondisk_stomith_size(new->magic);
	break;
      }

    if (!sl)
    {
      /* check if there's a "STOMITH: NUN" defn for this name */
      
      for (sl = params->none_names; sl; sl = sl->next)
	if ((strlen(pair->key) == strlen(sl->name)) &&
	    !strcmp(pair->key, sl->name))
	{
	  type_zalloc(new, stomith_list_t, 1);	  
	  memcpy(new->name, sl->name, strlen(sl->name));
	  node->stomith = new;
	  goto next;
	}

      die("\"%s\" doesn't match any stomith configurations\n", pair->key);
    }


    /* put stomith descriptor into list for this node */
  
    if (!node->stomith)
    {
      node->stomith = new;
      new->next = NULL;
    }
    else
    {
      for (s = node->stomith; s->next; s = s->next) ;
      s->next = new;
      new->next = NULL;
    }

  next:

  }
    
  return(0);
}


/**
 * add_stomith_list
 *
 * @params: configuration description
 * @magic: 
 * @values:
 *
 */

stomith_list_t *add_stomith_list(param_t *params, uint32 magic, keypair_t *values)
{
  stomith_list_t *st;
  int error;


  type_zalloc(st, stomith_list_t, 1);

  st->magic = magic;

  error = add_stomith_specific(st, values);
  if (error)
  {
    free(st);
    return(NULL);
  }

  if (!params->stomith_list)
  {
    params->stomith_list = st;
    st->next = NULL;
  }
  else
  {
    st->next = params->stomith_list;
    params->stomith_list = st;
  }	

  return(st);
}


/**
 * add_none_name
 *
 * @params: configuration description
 * @values:
 *
 */

void add_none_name(param_t *params, keypair_t *values)
{
  stomith_list_t *st, *sl;


  if (strncmp(values->key, "name", 4))
    die("\"STOMITH: NUN\" not followed by \"name:\"\n");    


  for (sl = params->none_names; sl; sl = sl->next)
  {
    if ((strlen(values->value) == strlen(sl->name)) &&
	!strcmp(sl->name, values->value))
      return;
  }

  type_zalloc(st, stomith_list_t, 1);

  st->magic = 0;

  strcpy(st->name, values->value);

  if (!params->none_names)
  {
    params->none_names = st;
    st->next = NULL;
  }
  else
  {
    st->next = params->none_names;
    params->none_names = st;
  }	
}


/**
 * add_node_list
 *
 * @params: configuration description
 * @ipaddr: IP address of new node
 * @cid: CID of new node
 * 
 * Returns: node description
 *
 */

node_list_t *add_node_list(param_t *params, uint32 ipaddr, uint32 cid)
{
  node_list_t *node, *n;

  type_zalloc(node, node_list_t, 1);

  node->ipaddr = ipaddr;
  node->cid = cid;
  node->length = sizeof(cluster_node_t);


  /* check if this node would conflict with current node list */

  for (n = params->node_list; n; n = n->next)
  {
    if (node->ipaddr == n->ipaddr)
      die("Duplicate IP address for nodes cid=%u and cid=%u\n", node->cid, n->cid);

    if (node->cid == n->cid)
      die("Duplicate CID numbers\n");
  }


  /* add node to the list */

  if (!params->node_list)
  {
    params->node_list = node;
    node->next = NULL;
  }
  else
  {
    for (n = params->node_list; n->next; n = n->next) ;
    n->next = node;
    node->next = NULL;
  }

  return(node);
}


/**
 * create_config
 *
 * @filename: gfsconf configuration file
 * 
 * Returns: configuration description
 *
 */

param_t *create_config(char *filename)
{
  param_t *params;
  char *name;
  keypair_t *values;
  keypair_t *methods;
  node_list_t *node;
  uint32 magic, ipaddr, cid;
  int error;


  type_zalloc(params, param_t, 1);


  /* Let parser run and get initial global values */

  error = parser_init(params, filename);

  if (error < 0)
    die("bad config file \"%s\".\n", filename);


  /* Get stomith defs from parser */
  
  while (1)
  {
    error = parser_next_stomith(&name, &values);
    
    if (error < 0)
      break;

    magic = stomith_number(name);

    if (!magic)
    {
      if (!strcmp(name, "NUN"))
      { 
	add_none_name(params, values);
	continue;
      }

      die("unrecognized STOMITH type: \"%s\"\n", name);
    }

    add_stomith_list(params, magic, values);
  }
    

  /* Verify that there's something defined for STOMITH */      
  
  if (!params->none_names && !params->stomith_list)
    die("No STOMITH statement found.\n");
    

  /* Get node defs from parser */

  while (1)
  {
    error = parser_next_node(&ipaddr, &cid, &methods);

    if (error < 0)
      break;

    node = add_node_list(params, ipaddr, cid);

    add_stomith_2_node(params, node, methods);
  }

  return(params);
}


/**
 * write_config
 *
 * @params: description of configuration
 * 
 * Write the configuration to conf device.
 *
 */

int write_config(param_t *params)
{
  int fd, error;
  cluster_global_t cg;
  cluster_node_t cn;
  cluster_stomith_t cs;
  node_list_t *node;
  stomith_list_t *st;
  char buf[512];
  int block = 0, offset;


  fd = open(params->cidev, O_RDWR);
  if (fd < 0)
    die("%s: can't open device %s\n", prog_name, params->cidev);

  error = ioctl(fd, BLKFLSBUF);
  if (error)
    perror("ioctl BLKFLSBUF");
  

  /*  Write out the global struct  */
  
  memset(buf, 0, 512);
  cg.cg_magic = GFS_CLUSTER_GLOBAL;
  cg.cg_version = GFS_CLUSTER_VERSION;
  memcpy(cg.cg_lockdev, params->lockdev, GFS_LOCKNAME_LEN);
  memcpy(cg.cg_datadev, params->datadev, GFS_LOCKNAME_LEN);
  cg.cg_cbport = params->cbport;
  cg.cg_node_timeout = params->timeout;

  cluster_global_out(&cg, buf);

  do_lseek(fd, block);
  do_write(fd, buf, 512);


  /*  Write out info for each node  */
  
  for (node = params->node_list; node; node = node->next)
  {
    memset(buf, 0, 512);
    memset(&cn, 0, sizeof(cluster_node_t));
    cn.cn_magic = GFS_CLUSTER_NODE;
    cn.cn_ipaddr = node->ipaddr;
    cn.cn_cid = node->cid;
    cn.cn_num_stomith = node->nstomith;

    cluster_node_out(&cn, buf);
    offset = sizeof(cluster_node_t);

    for (st = node->stomith; st; st = st->next)
    {
      memset(&cs, 0, sizeof(cluster_stomith_t));
      cs.cs_magic = st->magic;
      memcpy(&cs.cs_name, st->name, STM_STRING_LEN);
      memcpy(&cs.cs_st, &(st->st), sizeof(st->st));

      cluster_stomith_out(&cs, buf + offset);
      offset += ondisk_stomith_size(cs.cs_magic);
    }

    block++;
    do_lseek(fd, block * CIDEV_BLOCKSIZE);
    do_write(fd, buf, 512);
  }


  /*  Write out a tail  */
  
  memset(buf, 0, 512);
  strcpy(cg.cg_lockdev, " The End....");
  strcpy(cg.cg_datadev, "Mmmm, cheese.");
  cluster_global_out(&cg, buf);

  do_lseek(fd, (block + 1) * CIDEV_BLOCKSIZE);
  do_write(fd, buf, 512);

  sync();

  close(fd);
  
  return(0);
}


/**
 * check_cidev
 *
 * @params: description of configuration
 * @active_ok: if true, proceed if cidev is active
 *
 * The function exits if version is not compatible or if cidev
 * is active and !active_ok.
 *
 */

int check_cidev(param_t *params, int active_ok)
{
  int fd, error;
  cluster_global_t cg;
  char buf[512];


  fd = open(params->cidev, O_RDONLY);
  if (fd < 0)
    die("%s: can't open device %s\n", prog_name, params->cidev);

  error = ioctl(fd, BLKFLSBUF);
  if (error)
    perror("ioctl BLKFLSBUF");
  
  
  memset(buf, 0, 512);

  do_lseek(fd, 0);
  do_read(fd, buf, sizeof(cluster_global_t));

  cluster_global_in(&cg, buf);

  if (cg.cg_magic == GFS_CLUSTER_GLOBAL)
  {
    if (!active_ok)
      die("Aborted - cidev \"%s\" may be in use.  Erase with '-e' option.\n", params->cidev);
    
    if ((((cg.cg_version >> 16) & 0xff) != ((GFS_CLUSTER_VERSION >> 16) & 0xff)) ||
	(((cg.cg_version >>  8) & 0xff) != ((GFS_CLUSTER_VERSION >>  8) & 0xff)))
    {
      printf("cg: %x\n", cg.cg_version);
      printf("ic: %x\n", GFS_CLUSTER_VERSION);
      die("Aborted - version mismatch: cidev (%u, %u), gfsconf (%u, %u)\n",
	  (cg.cg_version >> 16) & 0xff, (cg.cg_version >> 8) & 0xff,
	  (GFS_CLUSTER_VERSION >> 16) & 0xff, (GFS_CLUSTER_VERSION >> 8) & 0xff);
    }
  }


  close(fd);
  return(0);
}


/**
 * clear_cidev
 *
 * @cidev: Cluster Information Device name
 *
 */

int clear_cidev(char *cidev)
{
  int fd, error;
  cluster_global_t cg;


  fd = open(cidev, O_RDWR);
  if (fd < 0)
    die("%s: can't open device %s\n", prog_name, cidev);

  error = ioctl(fd, BLKFLSBUF);
  if (error)
    perror("ioctl BLKFLSBUF");

  
  do_lseek(fd, 0);
  do_read(fd, &cg, sizeof(cluster_global_t));

  cg.cg_magic = 0;

  do_lseek(fd, 0);
  do_write(fd, &cg, sizeof(cluster_global_t));

  sync();

  close(fd);
  return(0);
}


/**
 * read_config
 *
 * @cidev: Cluster Information Device name
 * @force: flag, if true ignore incorrect magic number
 *
 * Returns: description of configuration on cidev
 *
 */

param_t *read_config(char *cidev, int force)
{
  int fd, error;
  cluster_global_t cg;
  cluster_node_t cn;
  cluster_stomith_t cs;
  node_list_t *node, *n;
  stomith_list_t *sl, *s, *sl2, *new;
  param_t *params = NULL;
  char buf[512];
  int block = 0, offset = 0, i;


  fd = open(cidev, O_RDONLY);
  if (fd < 0)
    die("can't open device %s\n", params->cidev);

  error = ioctl(fd, BLKFLSBUF);
  if (error)
    perror("ioctl BLKFLSBUF");
  

  type_zalloc(params, param_t, 1);

  
  memset(buf, 0, 512);

  do_lseek(fd, block * CIDEV_BLOCKSIZE);
  do_read(fd, buf, sizeof(cluster_global_t));

  cluster_global_in(&cg, buf);

  if (cg.cg_magic != GFS_CLUSTER_GLOBAL)
  {
    printf("%s: Error - \"%s\" has invalid magic number\n", prog_name, cidev);
    if (!force)
      exit(-1);
  }

  
  memcpy(params->cidev, cidev, strlen(cidev));
  memcpy(params->datadev, cg.cg_datadev, GFS_LOCKNAME_LEN + 1);
  memcpy(params->lockdev, cg.cg_lockdev, GFS_LOCKNAME_LEN + 1);
  params->cbport = cg.cg_cbport;
  params->timeout = cg.cg_node_timeout;


  /* build list of everything */

  
  while (1)
  {
    block++;

    memset(buf, 0, 512);
    do_lseek(fd, block * CIDEV_BLOCKSIZE);
    do_read(fd, buf, 512);

    cluster_node_in(&cn, buf);

    if (cn.cn_magic == GFS_CLUSTER_GLOBAL)
      break;

    if (cn.cn_magic != GFS_CLUSTER_NODE)
      continue;

    
    type_zalloc(node, node_list_t, 1);
    
    node->ipaddr = cn.cn_ipaddr;
    node->cid = cn.cn_cid;
    node->nstomith = cn.cn_num_stomith;

    offset = sizeof(cluster_node_t);

    for (i = 0; i < node->nstomith; i++)
    {
      cluster_stomith_in(&cs, buf + offset);
      
      type_zalloc(sl, stomith_list_t, 1);

      sl->magic = cs.cs_magic;
      memcpy(sl->name, &cs.cs_name, STM_STRING_LEN);
      memcpy(&sl->st, &cs.cs_st, sizeof(sl->st));

      offset += ondisk_stomith_size(sl->magic);

      if (!node->stomith)
      {
	node->stomith = sl;
	sl->next = NULL;
      }
      else
      {
	for (s = node->stomith; s->next; s = s->next) ;
	s->next = sl;
	sl->next = NULL;
      }
    }

    if (!node->nstomith)
    {
      cluster_stomith_in(&cs, buf + offset);      
      type_zalloc(sl, stomith_list_t, 1);
      memcpy(sl->name, &cs.cs_name, STM_STRING_LEN);     
      node->stomith = sl;
    }
    
    if (!params->node_list)
    {
      params->node_list = node;
      node->next = NULL;
    }
    else
    {
      for (n = params->node_list; n->next; n = n->next) ;
      n->next = node;
      node->next = NULL;
    }    
  }

  
  /* build general stomith list from all node's lists */

  for (node = params->node_list; node; node = node->next)
  {
    for (sl = node->stomith; sl && node->nstomith; sl = sl->next)
    {
      for (sl2 = params->stomith_list; sl2; sl2 = sl2->next)
      {
	if ((sl->magic == sl2->magic) && !strcmp(sl->name, sl2->name))  
	  goto next;
      }

      /* add new general stomith description */
      
      type_zalloc(new, stomith_list_t, 1);

      *new = *sl;

      if (!params->stomith_list)
      {
	params->stomith_list = new;
	new->next = NULL;
      }
      else
      {
	for (s = params->stomith_list; s->next; s = s->next) ;
	s->next = new;
	new->next = NULL;
      }

    next:
    }
  }


  /* build list of none names */

  for (node = params->node_list; node; node = node->next)
  {
    if (!node->nstomith)
    {
      for (sl = params->none_names; sl; sl = sl->next)
      {
	if ((strlen(node->stomith->name) == strlen(sl->name)) &&
	     !strcmp(sl->name, node->stomith->name))
	  goto next2;
      }

      type_zalloc(new, stomith_list_t, 1);
      
      *new = *(node->stomith);
      
      if (!params->none_names)
      {
	params->none_names = new;
	new->next = NULL;
      }
      else
      {
	for (s = params->none_names; s->next; s = s->next) ;
	s->next = new;
	new->next = NULL;
      }
      
    next2:
    }
  }


  return(params);
}


/**
 * compare_nodes
 *
 * @n1: node description
 * @n2: node description
 *
 * Returns: 0 if nodes are equal, -1 if not
 *
 */

int compare_nodes(node_list_t *n1, node_list_t *n2)
{
  stomith_list_t *s1, *s2;
  int i;
  char ip[16];


  if (n1->ipaddr != n2->ipaddr)
    goto fail1;

  if (n1->cid != n2->cid)
    goto fail1;

  if (n1->nstomith != n2->nstomith)
    goto fail1;

  if (!n1->nstomith && !n2->nstomith)
    return(0);


  for (s1 = n1->stomith; s1; s1 = s1->next)
  {
    for (s2 = n2->stomith; s2; s2 = s2->next)
    {
      if ((s1->magic == s2->magic) && !strcmp(s1->name, s2->name))
      {
	for (i = 0; i < ondisk_stomith_size(s1->magic); i++)
	{
	  if ((*(char *)&s1->st + i) != (*(char *)&s2->st + i))
	    goto fail2;
	  else
	    goto next1;
	}
      }
    }

    if (!s2)
      goto fail2;

  next1:
  }


  for (s1 = n2->stomith; s1; s1 = s1->next)
  {
    for (s2 = n1->stomith; s2; s2 = s2->next)
    {
      if ((s1->magic == s2->magic) && !strcmp(s1->name, s2->name))
      {
	for (i = 0; i < ondisk_stomith_size(s1->magic); i++)
	{
	  if ((*(char *)&s1->st + i) != (*(char *)&s2->st + i))
	    goto fail2;
	  else
	    goto next2;
	}
      }
    }

    if (!s2)
      goto fail2;

  next2:
  }

  return(0);


 fail1:
  inet_ntoa(n1->ipaddr, ip);
  printf("differing node descriptions for node \"%s\"\n", ip);
  return(-1);

 fail2:
  inet_ntoa(n1->ipaddr, ip);
  printf("differing stomith descriptions for node \"%s\"\n", ip);
  return(-1);
}


/**
 * compare_config
 *
 * @p: current config
 * @q: proposed config
 *
 * Returns: 0 if p to q is a proper transition, -1 if not
 *
 */

int compare_config(param_t *p, param_t *q)
{
  node_list_t *n1, *n2;
  char ip1[16], ip2[16];

  if (strcmp(p->datadev, q->datadev))
  {
    printf("datadev mismatch: \"%s\" - \"%s\"\n", p->datadev, q->datadev);
    goto fail;
  }

  if (strcmp(p->cidev, q->cidev))
  {
    printf("cidev mismatch: \"%s\" - \"%s\"\n", p->cidev, q->cidev);
    goto fail;
  }

  if (strcmp(p->lockdev, q->lockdev))
  {
    printf("lockdev mismatch: \"%s\" - \"%s\"\n", p->lockdev, q->lockdev);
    goto fail;
  }

  if (p->cbport != q->cbport)
  {
    printf("cbport mismatch: %u - %u\n", p->cbport, q->cbport);
    goto fail;
  }

  if (p->timeout != q->timeout)
  {
    printf("timeout mismatch: %u - %u\n", p->timeout, q->timeout);
    goto fail;
  }


  /* every node in p must exist and be exactly the same in q */

  for (n1 = p->node_list; n1; n1 = n1->next)
  {
    for (n2 = q->node_list; n2; n2 = n2->next)
    {
      if (n1->ipaddr == n2->ipaddr)
      {
	if (compare_nodes(n1, n2))
	  goto fail;
	else
	  goto next;
      }

      if (n1->cid == n2->cid)
      {
	inet_ntoa(n1->ipaddr, ip1);
	inet_ntoa(n2->ipaddr, ip2);
	printf("nodes \"%s\" and \"%s\" have same cid: %u\n", ip1, ip2, n1->cid);
	goto fail;
      }
    }

    if (!n2)
    {
      inet_ntoa(n1->ipaddr, ip1);
      printf("node \"%s\" missing in new configuration\n", ip1);
      goto fail;
    }
    
  next:
  }
  
  return(0);

 fail:
  return(-1);
}


/**
 * update_config
 *
 * @old: current description of configuration
 * @new: new description
 *
 */

int update_config(param_t *old, param_t *new)
{
  int fd, error;
  cluster_global_t cg;
  cluster_node_t cn;
  cluster_stomith_t cs;
  char buf[512], outbuf[512];
  node_list_t *n1, *n2;
  stomith_list_t *st;
  unsigned int block = 0;
  unsigned int offset;


  fd = open(old->cidev, O_RDWR);
  if (fd < 0)
    die("can't open device %s\n", old->cidev);

  error = ioctl(fd, BLKFLSBUF);
  if (error)
    perror("ioctl BLKFLSBUF");

  
  memset(buf, 0, 512);
  do_lseek(fd, block * CIDEV_BLOCKSIZE);
  do_read(fd, buf, 512);
  
  cluster_global_in(&cg, buf);

  if (cg.cg_magic != GFS_CLUSTER_GLOBAL)
    die("Invalid header on cidev \"%s\"", old->cidev);

  

  /* select new nodes (those not in the old list) */

  for (n1 = new->node_list; n1; n1 = n1->next)
  {
    for (n2 = old->node_list; n2; n2 = n2->next)
      if (n1->ipaddr == n2->ipaddr)
	goto next;
    

    /* put together new buffer */

    memset(outbuf, 0, 512);
    memset(&cn, 0, sizeof(cluster_node_t));
    cn.cn_magic = GFS_CLUSTER_NODE;
    cn.cn_ipaddr = n1->ipaddr;
    cn.cn_cid = n1->cid;
    cn.cn_num_stomith = n1->nstomith;
    
    cluster_node_out(&cn, outbuf);
    offset = sizeof(cluster_node_t);
    
    for (st = n1->stomith; st; st = st->next)
    {
      memset(&cs, 0, sizeof(cluster_stomith_t));
      cs.cs_magic = st->magic;
      memcpy(&cs.cs_name, st->name, STM_STRING_LEN);
      memcpy(&cs.cs_st, &(st->st), sizeof(st->st));

      cluster_stomith_out(&cs, outbuf + offset);
      offset += ondisk_stomith_size(cs.cs_magic);
    }


    /* find first empty block */

    block = 1;

    while (1)
    {      
      memset(buf, 0, 512);
      do_lseek(fd, block * CIDEV_BLOCKSIZE);
      do_read(fd, buf, 512);
      
      cluster_node_in(&cn, buf);

      if (cn.cn_magic == GFS_CLUSTER_GLOBAL)
	break;
      
      if (cn.cn_magic == 0)
      {
	/* empty block, put new config here */
	
	do_lseek(fd, block * CIDEV_BLOCKSIZE);
	do_write(fd, outbuf, 512);

	goto next;
      }
      
      block++;
    }


    /* no empty blocks, add one to end */
     
    memset(buf, 0, 512);
    strcpy(cg.cg_lockdev, " The End....");
    strcpy(cg.cg_datadev, "Mmmm, biscuits.");
    
    cluster_global_out(&cg, buf);
    
    do_lseek(fd, (block + 1) * CIDEV_BLOCKSIZE);
    do_write(fd, buf, 512);
    
    do_lseek(fd, block * CIDEV_BLOCKSIZE);
    do_write(fd, outbuf, 512);
    
  next:
  }

  sync();
 
  return(0);
}


/**
 * remove_node
 *
 * @rmnode: IP address of node to remove
 * @cidev: the name of the cidev
 *
 */

int remove_node(char *rmnode, char *cidev)
{
  uint32 ip;
  int fd, error;
  cluster_global_t cg;
  cluster_node_t cn;
  char buf[512];
  int block = 0;


  inet_aton(rmnode, &ip);

  fd = open(cidev, O_RDWR);
  if (fd < 0)
    die("can't open device %s\n", cidev);

  error = ioctl(fd, BLKFLSBUF);
  if (error)
    perror("ioctl BLKFLSBUF");


  memset(buf, 0, 512);

  do_lseek(fd, block * CIDEV_BLOCKSIZE);
  do_read(fd, buf, sizeof(cluster_global_t));

  cluster_global_in(&cg, buf);

  if (cg.cg_magic != GFS_CLUSTER_GLOBAL)
    die("\"%s\" has invalid magic number\n", cidev);

  
  while (1)
  {
    block++;

    memset(buf, 0, 512);
    do_lseek(fd, block * CIDEV_BLOCKSIZE);
    do_read(fd, buf, 512);

    cluster_node_in(&cn, buf);

    if (cn.cn_magic == GFS_CLUSTER_GLOBAL)
      die("Node with IP address \"%s\" not found.\n", rmnode);

    if (cn.cn_magic == GFS_CLUSTER_NODE && cn.cn_ipaddr == ip)
    {
      cn.cn_magic = 0;
      
      cluster_node_out(&cn, buf);

      do_lseek(fd, block * CIDEV_BLOCKSIZE);
      do_write(fd, buf, 512);
      break;
    }
  }

  sync();

  return(0);
}


/**
 * print_config
 *
 * @params: describes configuration
 *
 * Output is in config file format.
 *
 */

int print_config(param_t *params)
{
  char ip[16];
  node_list_t *node;
  stomith_list_t *sl;

  printf("# Parameter file for gfsconf\n\n");
  
  printf("datadev:  %s\n", params->datadev);
  printf("cidev:    %s\n", params->cidev);
  printf("lockdev:  %s\n", params->lockdev);
  printf("cbport:  %u\n", params->cbport);
  printf("timeout: %u\n", params->timeout);
  printf("\n");


  for (sl = params->stomith_list; sl; sl = sl->next)
  {
    printf("STOMITH:\t%s\n", stomith_name(sl->magic));
    printf("name:\t\t%s\n", sl->name);
    print_stomith_specific(sl, PRINT_STOMITH_GENERAL);
    printf("\n");
  }

  for (sl = params->none_names; sl; sl = sl->next)
  {
    printf("STOMITH:\tNUN\n");
    printf("name:\t\t%s\n", sl->name);
    printf("\n");
  }

  for (node = params->node_list; node; node = node->next)
  {
    inet_ntoa(node->ipaddr, ip);

    printf("node:  %s  %u\n", ip, node->cid);

    for (sl = node->stomith; sl; sl = sl->next)
    {
      printf("SM: %s ", sl->name);
      print_stomith_specific(sl, PRINT_STOMITH_SPECIFIC);
    }
    printf("\n");
  }

  return(0);
}


/**
 * stat_device
 *
 * @cidev: name of cidev
 *
 */

void stat_device(char *cidev)
{
  struct stat devstat;
  int error;

  memset(&devstat, 0, sizeof(struct stat));

  error = stat(cidev, &devstat);
  if (error < 0)
    die("bad stat on \"%s\"", cidev);

  if (!S_ISBLK(devstat.st_mode))
    die("Error - \"%s\" is not a block device\n", cidev);
}


/**
 * stat_file
 *
 * @file: name of file
 *
 */

void stat_file(char *file)
{
  struct stat devstat;
  int error;

  memset(&devstat, 0, sizeof(struct stat));

  error = stat(file, &devstat);
  if (error < 0)
    die("bad stat on \"%s\"", file);

  if (!S_ISREG(devstat.st_mode))
    die("Error - \"%s\" is not a regular file\n", file);
}


/**
 * print_params
 *
 * @params: describes configuration
 *
 */

void print_params(param_t *params)
{
  node_list_t *node;
  stomith_list_t *st;
  char buf[16];

  printf("Data Device:               %s\n", params->datadev);
  printf("Cluster Info Device:       %s\n", params->cidev);
  printf("Lock Device:               %s\n", params->lockdev);
  printf("Call-back Port:            %u\n", params->cbport);
  printf("Timeout:                   %u\n", params->timeout);
  
  for (node = params->node_list; node; node = node->next)
  {
    printf("\n");

    inet_ntoa(node->ipaddr, buf);        
    printf("Node IP address: %s\tCID: %u\n", buf, node->cid);
    
    for (st = node->stomith; st; st = st->next)
    {
      printf("  STOMITH: %s (%s)\n", stomith_name(st->magic), st->name);
      print_stomith_specific(st, PRINT_STOMITH_ALL);
    }
  }
}


/**
 * print_usage
 *
 *
 */

void print_usage(void)
{
  printf("Usage: \n\n");

  printf("%s -c <parameter file>\n", prog_name);
  printf("\tCreate a new GFS cluster configuration based on given file.\n\n");

  printf("%s -e <cidev>\n", prog_name);
  printf("\tErase the cluster configuration on specified cidev.\n\n");

  printf("%s -r <IP address> <cidev>\n", prog_name);
  printf("\tRemove the node with given IP address from cidev.\n\n");

  printf("%s -p <cidev>\n", prog_name);
  printf("\tPrint cluster configuration in parameter file format.\n\n");

  printf("%s -v <cidev>\n", prog_name);
  printf("\tPrint verbose, descriptive listing for each node.\n\n");

  printf("%s -a <paramater file>\n", prog_name);
  printf("\tAppend new node description to cidev per param file.\n\n");

  printf("%s -d <paramater file>\n", prog_name);
  printf("\tJust parse the paramater file and print it back out.\n\n");

  printf("Other options:\n");
  printf("\t-n dry-run: don't actually make changes to cidev.\n");
  printf("\t-q quiet: no output\n");
  printf("\t-f force: no confirmation or ignore basic errors.\n");
}


/**
 * main 
 *
 * @argc:
 * @argv:
 *
 */

int main(int argc, char **argv)
{
  param_t *params, *params1, *params2;
  int create = 0, append = 0, erase = 0, remove = 0, parsedump = 0;
  int quiet = 0, verbose = 0, print = 0, force = 0, dryrun = 0;
  char *infile = NULL, *cidev = NULL, *rmnode = NULL;
  int c;
  int error;

  prog_name = argv[0];



  while ((c = getopt(argc, argv, "d:c:e:r:v:p:a:nqfh")) != -1)
  {
    switch(c)
    {
    case 'h':
      print_usage();
      exit(0);

    case 'q':
      quiet = TRUE;
      break;

    case 'f':
      force = TRUE;
      break;

    case 'c':
      create = TRUE;
      infile = optarg;
      break;
      
    case 'a':
      append = TRUE;
      infile = optarg;
      break;

    case 'e':
      erase = TRUE;
      cidev = optarg;
      break;

    case 'd':
      parsedump = TRUE;
      infile = optarg;
      break;

    case 'v':
      verbose = TRUE;
      cidev = optarg;
      break;

    case 'p':
      print = TRUE;
      cidev = optarg;
      break;

    case 'r':
      remove = TRUE;
      rmnode = optarg;
      cidev = argv[argc - 1];
      break;

    case 'n':
      dryrun = TRUE;
      break;

    default:
      print_usage();
      exit(EXIT_FAILURE);
    }
  }

  
  if (parsedump)
  {
    stat_file(infile);
    params = create_config(infile);
    print_config(params);
    exit(0);
  }


  if (create)
  {
    stat_file(infile);
    params = create_config(infile);

    stat_device(params->cidev);

    check_cidev(params, (force ? TRUE : FALSE));
    
    if (!dryrun)
      write_config(params);
    
    if (!quiet)
      print_config(params);

    exit(0);
  }

  
  if (append)
  {
    stat_file(infile);
    params1 = create_config(infile);		/* proposed (new) */

    stat_device(params1->cidev);

    check_cidev(params1, TRUE);

    params2 = read_config(params1->cidev, 0);	/* current (old) */

    error = compare_config(params2, params1);
    if (error)
      die("Aborted - inconsistency between current config and \"%s\".\n", infile);

    if (!dryrun)
      update_config(params2, params1);

    if (!quiet)
      print_config(params1);

    exit(0);
  }


  if (remove)
  {
    stat_device(cidev);

    params = read_config(cidev, 0);

    check_cidev(params, TRUE);

    if (!dryrun)
      remove_node(rmnode, cidev);

    if (!quiet)
    {
      params2 = read_config(cidev, 0);
      print_config(params2);
    }

    exit(0);
  }

  
  if (erase)
  {
    stat_device(cidev);

    if (!dryrun)
      clear_cidev(cidev);
    
    if (!quiet)
      printf("Erased configuration on \"%s\".\n", cidev);

    exit(0);
  }


  if (print)
  {
    stat_device(cidev);

    params = read_config(cidev, force);

    check_cidev(params, TRUE);

    print_config(params);
    exit(0);
  }


  if (verbose)
  {
    stat_device(cidev);

    params = read_config(cidev, force);
    
    check_cidev(params, TRUE);

    print_params(params);
    exit(0);
  }


  printf("%s: error - unknown option.\n\n", prog_name);
  print_usage();

  return(0);
}

















