/******************************************************************************
*******************************************************************************
**
**    Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@sistina.com>
**
**    Module name: stomith_specific.c
**
**    Description:
**      Routines for handling config file options for each specific STOMITH
**      method.
**
*******************************************************************************
******************************************************************************/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>


#define WANT_STOMITH_NAMES
#include "gfsconf.h"
#include "interface.h"
#include "stomith_specific.h"



static int add_brcd_zone(stomith_list_t *st, keypair_t *list)
{
  int found = 0;
  stm_brcd_zone_t *silk;
  keypair_t *pair;
  char *key, *value;

  silk = &st->st.brcd_zone_s;


  for (pair = list; pair; pair = pair->next)
  {
    key = pair->key;
    value = pair->value;
    
    if (!strncmp(key, "name", 4))
    {
      if (strlen(value) > STM_STRING_LEN)
      {
	fprintf(stderr, "%s: name \"%s\" is too long\n", prog_name, value);
	exit(EXIT_FAILURE);
      }

      strcpy(st->name, value);
      found++;
    }
	
    else if (!strncmp(key, "ipaddr", 6))
    {
      inet_aton(value, &silk->st_ipaddr);
      found++;
    }

    else if (!strncmp(key, "login", 5))
    {
      if (strlen(value) > STM_STRING_LEN)
      {
	fprintf(stderr, "%s: login name \"%s\" is too long\n", prog_name, value);
	exit(EXIT_FAILURE);
      }

      strcpy(silk->st_login, value);
      found++;
    }

    else if (!strncmp(key, "passwd", 6))
    {
      if (strlen(value) > STM_STRING_LEN)
      {
	fprintf(stderr, "%s: password \"%s\" is too long\n", prog_name, value);
	exit(EXIT_FAILURE);
      }

      strcpy(silk->st_passwd, value);
      found++;
    }

    else if (!strncmp(key, "cfgname", 7))
    {
      if (strlen(value) > STM_STRING_LEN)
      {
	fprintf(stderr, "%s: cfgname \"%s\" is too long\n", prog_name, value);
	exit(EXIT_FAILURE);
      }
      
      strcpy(silk->st_cfgname, value);
      found++;           
    }

    else if (!strncmp(key, "comzone", 7))
    {
      if (strlen(value) > STM_STRING_LEN)
      {
	fprintf(stderr, "%s: common zone name \"%s\" is too long\n", prog_name, value);
	exit(EXIT_FAILURE);
      }
      
      strcpy(silk->st_comzone, value);
      found++;           
    }

    else if (!strncmp(key, "command", 7))
    {
      if (strlen(value) > STM_STRING_LEN)
      {
	fprintf(stderr, "%s: zone name \"%s\" is too long\n", prog_name, value);
	exit(EXIT_FAILURE);
      }

      strcpy(silk->st_command, value);
      found++;
    }

    else
    {
      fprintf(stderr, "%s: unrecognized STOMITH-specific field: \"%s\"\n", prog_name, key);
      exit(EXIT_FAILURE);
    }
  }

  /* 7 specific config lines for brocade_zone:
     1 name, 2 ipaddr, 3 login, 4 passwd, 5 cfgname, 6 comzone, 7 command */

  if (found != 7)
  {
    fprintf(stderr, "%s: incomplete STOMITH definition for \"%s\".\n", prog_name,
	    stomith_name(st->magic));
    exit(EXIT_FAILURE);
  }

  return(0);
}



static int add_brcd_port(stomith_list_t *st, keypair_t *list)
{
  int found = 0;
  stm_brcd_port_t *bp;
  keypair_t *pair;
  char *key, *value;

  bp = &st->st.brcd_port_s;


  for (pair = list; pair; pair = pair->next)
  {
    key = pair->key;
    value = pair->value;
    
    if (!strncmp(key, "name", 4))
    {
      if (strlen(value) > STM_STRING_LEN)
      {
	fprintf(stderr, "%s: name \"%s\" is too long\n", prog_name, value);
	exit(EXIT_FAILURE);
      }

      strcpy(st->name, value);
      found++;
    }
	
    else if (!strncmp(key, "ipaddr", 6))
    {
      inet_aton(value, &bp->st_ipaddr);
      found++;
    }

    else if (!strncmp(key, "login", 5))
    {
      if (strlen(value) > STM_STRING_LEN)
      {
	fprintf(stderr, "%s: login name \"%s\" is too long\n", prog_name, value);
	exit(EXIT_FAILURE);
      }

      strcpy(bp->st_login, value);
      found++;
    }

    else if (!strncmp(key, "passwd", 6))
    {
      if (strlen(value) > STM_STRING_LEN)
      {
	fprintf(stderr, "%s: password \"%s\" is too long\n", prog_name, value);
	exit(EXIT_FAILURE);
      }

      strcpy(bp->st_passwd, value);
      found++;
    }

    else
    {
      fprintf(stderr, "%s: unrecognized STOMITH-specific field: \"%s\"\n", prog_name, key);
      exit(EXIT_FAILURE);
    }
  }

  /* 4 specific config lines for brocade_port:
     1 name, 2 ipaddr, 3 login, 4 passwd */

  if (found != 4)
  {
    fprintf(stderr, "%s: incomplete STOMITH definition for \"%s\".\n", prog_name,
	    stomith_name(st->magic));
    exit(EXIT_FAILURE);
  }

  return(0);
}



static int add_wti_nps(stomith_list_t *st, keypair_t *list)
{
  int found = 0;
  stm_wti_nps_t *nps;
  keypair_t *pair;
  char *key, *value;

  nps = &st->st.wti_nps_s;


  for (pair = list; pair; pair = pair->next)
  {
    key = pair->key;
    value = pair->value;

    if (!strncmp(key, "name", 4))
    {
      if (strlen(value) > STM_STRING_LEN)
      {
	fprintf(stderr, "%s: name \"%s\" is too long\n", prog_name, value);
	exit(EXIT_FAILURE);
      }

      strcpy(st->name, value);
      found++;
    }

    else if (!strncmp(key, "ipaddr", 6))
    {
      inet_aton(value, &nps->st_ipaddr);
      found++;
    }

    else if (!strncmp(key, "passwd", 6))
    {
      if (strlen(value) > STM_STRING_LEN)
      {
	fprintf(stderr, "%s: password \"%s\" is too long\n", prog_name, value);
	exit(EXIT_FAILURE);
      }

      strcpy(nps->st_passwd, value);
      found++;
    }

    else
    {
      fprintf(stderr, "%s: unrecognized STOMITH-specific field: \"%s\"\n", prog_name, key);
      exit(EXIT_FAILURE);
    }
  }

  /* 3 specific config lines for wti_nps */  

  if (found != 3)
  {
    fprintf(stderr, "%s: incomplete STOMITH definition for \"%s\".\n", prog_name,
	    stomith_name(st->magic));
    exit(EXIT_FAILURE);
  }

  return(0);
}



static int add_vacm_emp(stomith_list_t *st, keypair_t *list)
{
  int found = 0;
  stm_vacm_emp_t *va;
  keypair_t *pair;
  char *key, *value;

  va = &st->st.vacm_emp_s;


  for (pair = list; pair; pair = pair->next)
  {
    key = pair->key;
    value = pair->value;

    if (!strncmp(key, "name", 4))
    {
      if (strlen(value) > STM_STRING_LEN)
      {
	fprintf(stderr, "%s: name \"%s\" is too long\n", prog_name, value);
	exit(EXIT_FAILURE);
      }

      strcpy(st->name, value);
      found++;
    }

    else if (!strncmp(key, "nexxus", 6))
    {
      if (strlen(value) > STM_STRING_LEN)
      {
	fprintf(stderr, "%s: nexxus name \"%s\" is too long\n", prog_name, value);
	exit(EXIT_FAILURE);
      }

      strcpy(va->st_nexxus, value);
      found++;
    }

    else if (!strncmp(key, "login", 5))
    {
      if (strlen(value) > STM_STRING_LEN)
      {
	fprintf(stderr, "%s: login name \"%s\" is too long\n", prog_name, value);
	exit(EXIT_FAILURE);
      }

      strcpy(va->st_login, value);
      found++;
    }

    else if (!strncmp(key, "passwd", 6))
    {
      if (strlen(value) > STM_STRING_LEN)
      {
	fprintf(stderr, "%s: passwd \"%s\" is too long\n", prog_name, value);
	exit(EXIT_FAILURE);
      }

      strcpy(va->st_passwd, value);
      found++;
    }

    else
    {
      fprintf(stderr, "%s: unrecognized STOMITH-specific field: \"%s\"\n", prog_name, key);
      exit(EXIT_FAILURE);
    }
  }

  /* 4 specific config lines for wti_nps */  

  if (found != 4)
  {
    fprintf(stderr, "%s: incomplete STOMITH definition for \"%s\".\n", prog_name,
	    stomith_name(st->magic));
    exit(EXIT_FAILURE);
  }

  return(0);
}



static int add_gnbd(stomith_list_t *st, keypair_t *list)
{
  int found = 0;
  stm_gnbd_t *sg;
  keypair_t *pair;
  char *key, *value;

  sg = &st->st.gnbd_s;


  for (pair = list; pair; pair = pair->next)
  {
    key = pair->key;
    value = pair->value;

    if (!strncmp(key, "name", 4))
    {
      if (strlen(value) > STM_STRING_LEN)
      {
	fprintf(stderr, "%s: name \"%s\" is too long\n", prog_name, value);
	exit(EXIT_FAILURE);
      }

      strcpy(st->name, value);
      found++;
    }

    else
    {
      fprintf(stderr, "%s: unrecognized STOMITH-specific field: \"%s\"\n", prog_name, key);
      exit(EXIT_FAILURE);
    }
  }

  /* 1 specific config line for gnbd */  

  if (found != 1)
  {
    fprintf(stderr, "%s: incomplete STOMITH definition for \"%s\".\n", prog_name,
	    stomith_name(st->magic));
    exit(EXIT_FAILURE);
  }

  return(0);
}



static int add_meatware(stomith_list_t *st, keypair_t *list)
{
  int found = 0;
  stm_meatware_t *meat;
  keypair_t *pair;
  char *key, *value;

  meat = &st->st.meatware_s;


  for (pair = list; pair; pair = pair->next)
  {
    key = pair->key;
    value = pair->value;

    if (!strncmp(key, "name", 4))
    {
      if (strlen(value) > STM_STRING_LEN)
      {
	fprintf(stderr, "%s: name \"%s\" is too long\n", prog_name, value);
	exit(EXIT_FAILURE);
      }

      strcpy(st->name, value);
      found++;
    }

    else
    {
      fprintf(stderr, "%s: unrecognized STOMITH-specific field: \"%s\"\n", prog_name, key);
      exit(EXIT_FAILURE);
    }
  }

  /* 1 specific config line for gnbd */  

  if (found != 1)
  {
    fprintf(stderr, "%s: incomplete STOMITH definition for \"%s\".\n", prog_name,
	    stomith_name(st->magic));
    exit(EXIT_FAILURE);
  }

  return(0);
}



static int add_vx_hub(stomith_list_t *st, keypair_t *list)
{
  int found = 0;
  stm_vx_hub_t *hub;
  keypair_t *pair;
  char *key, *value;

  hub = &st->st.vx_hub_s;


  for (pair = list; pair; pair = pair->next)
  {
    key = pair->key;
    value = pair->value;

    if (!strncmp(key, "name", 4))
    {
      if (strlen(value) > STM_STRING_LEN)
      {
        fprintf(stderr, "%s: name \"%s\" is too long\n", prog_name, value);
        exit(EXIT_FAILURE);
      }

      strcpy(st->name, value);
      found++;
    }

    else if (!strncmp(key, "ipaddr", 6))
    {
      inet_aton(value, &hub->st_ipaddr);
      found++;
    }

    else if (!strncmp(key, "passwd", 6))
    {
      if (strlen(value) > STM_STRING_LEN)
      {
        fprintf(stderr, "%s: password \"%s\" is too long\n", prog_name, value);
        exit(EXIT_FAILURE);
      }

      strcpy(hub->st_passwd, value);
      found++;
    }

    else
    {
      fprintf(stderr, "%s: unrecognized STOMITH-specific field: \"%s\"\n", prog_name, key);
      exit(EXIT_FAILURE);
    }
  }

  /* 3 specific config lines for vx_hub - same as those for wti_nps */

  if (found != 3)
  {
    fprintf(stderr, "%s: incomplete STOMITH definition for \"%s\".\n", prog_name,
            stomith_name(st->magic));
    exit(EXIT_FAILURE);
  }

  return(0);
}



static int add_apc_ms(stomith_list_t *st, keypair_t *list)
{
  int found = 0;
  stm_apc_ms_t *apc;
  keypair_t *pair;
  char *key, *value;

  apc = &st->st.apc_ms_s;


  for (pair = list; pair; pair = pair->next)
  {
    key = pair->key;
    value = pair->value;

    if (!strncmp(key, "name", 4))
    {
      if (strlen(value) > STM_STRING_LEN)
      {
        fprintf(stderr, "%s: name \"%s\" is too long\n", prog_name, value);
        exit(EXIT_FAILURE);
      }

      strcpy(st->name, value);
      found++;
    }

    else if (!strncmp(key, "ipaddr", 6))
    {
      inet_aton(value, &apc->st_ipaddr);
      found++;
    }

    else if (!strncmp(key, "login", 5))
    {
      if (strlen(value) > STM_STRING_LEN)
      {
        fprintf(stderr, "%s: login \"%s\" is too long\n", prog_name, value);
        exit(EXIT_FAILURE);
      }

      strcpy(apc->st_login, value);
      found++;
    }

    else if (!strncmp(key, "passwd", 6))
    {
      if (strlen(value) > STM_STRING_LEN)
      {
        fprintf(stderr, "%s: password \"%s\" is too long\n", prog_name, value);
        exit(EXIT_FAILURE);
      }

      strcpy(apc->st_passwd, value);
      found++;
    }

    else
    {
      fprintf(stderr, "%s: unrecognized STOMITH-specific field: \"%s\"\n", prog_name, key);
      exit(EXIT_FAILURE);
    }
  }

  /* 4 specific lines for apc_ms */

  if (found != 4)
  {
    fprintf(stderr, "%s: incomplete STOMITH definition for \"%s\".\n", prog_name,
            stomith_name(st->magic));
    exit(EXIT_FAILURE);
  }

  return(0);
}



static int addparam_brcd_zone(stomith_list_t *st, char *value)
{
  stm_brcd_zone_t *silk = &st->st.brcd_zone_s;


  if (!value)
  {
    fprintf(stderr, "%s: node line missing brocade_zone node id.\n", prog_name);    
    exit(EXIT_FAILURE);
  }

  if (!strcmp(value, ""))
  {
    fprintf(stderr, "%s: node line brocade_zone bad name of \"%s\".\n", prog_name, value);
    exit(EXIT_FAILURE);
  }

  if (strlen(value) > STM_STRING_LEN)
  {
    fprintf(stderr, "%s: actname \"%s\" is too long\n", prog_name, value);
    exit(EXIT_FAILURE);
  }

  strcpy(silk->st_actname, value);

  return(0);
}



static int addparam_brcd_port(stomith_list_t *st, char *value)
{
  stm_brcd_port_t *bp = &st->st.brcd_port_s;


  if (!value)
  {
    fprintf(stderr, "%s: node line missing brocade_port port number.\n", prog_name);    
    exit(EXIT_FAILURE);
  }

  if (!strcmp(value, ""))
  {
    fprintf(stderr, "%s: node line brocade_zone bad port number.\n", prog_name);
    exit(EXIT_FAILURE);
  }

  bp->st_port = atoi(value);

  return(0);
}



static int addparam_wti_nps(stomith_list_t *st, char *value)
{
  stm_wti_nps_t *nps = &st->st.wti_nps_s;

  if (!value)
  {
    fprintf(stderr, "%s: node line missing wti_nps plug number.\n", prog_name);    
    exit(EXIT_FAILURE);
  }

  if (!strcmp(value, ""))
  {
    fprintf(stderr, "%s: wti_nps bad plug number of \"%s\".\n", prog_name, value);
    exit(EXIT_FAILURE);
  }

  nps->st_plug = atoi(value);

  return(0);
}



static int addparam_vacm_emp(stomith_list_t *st, char *value)
{
  stm_vacm_emp_t *va = &st->st.vacm_emp_s;

  if (!value)
  {
    fprintf(stderr, "%s: node line missing vacm_emp node id.\n", prog_name);    
    exit(EXIT_FAILURE);
  }

  if (!strcmp(value, ""))
  {
    fprintf(stderr, "%s: vacm_emp bad node id of \"%s\".\n", prog_name, value);
    exit(EXIT_FAILURE);
  }

  if (strlen(value) > STM_STRING_LEN)
  {
    fprintf(stderr, "%s: node id \"%s\" is too long\n", prog_name, value);
    exit(EXIT_FAILURE);
  }

  strcpy(va->st_nodeid, value);

  return(0);
}



static int addparam_gnbd(stomith_list_t *st, node_list_t *node)
{
  stm_gnbd_t *sg = &st->st.gnbd_s;

  sg->st_ipaddr = node->ipaddr;

  return(0);
}



static int addparam_meatware(stomith_list_t *st, node_list_t *node)
{
  stm_meatware_t *meat = &st->st.meatware_s;

  meat->st_ipaddr = node->ipaddr;

  return(0);
}



static int addparam_vx_hub(stomith_list_t *st, char *value)
{
  stm_vx_hub_t *hub = &st->st.vx_hub_s;


  if (!value)
  {
    fprintf(stderr, "%s: node line missing vx_hub plug number.\n", prog_name);    
    exit(EXIT_FAILURE);
  }

  if (!strcmp(value, ""))
  {
    fprintf(stderr, "%s: vxhub bad plug number of \"%s\".\n", prog_name, value);
    exit(EXIT_FAILURE);
  }

  hub->st_port = atoi(value);

  return(0);
}



static int addparam_apc_ms(stomith_list_t *st, char *value)
{
  stm_apc_ms_t *apc = &st->st.apc_ms_s;

  if (!value)
  {
    fprintf(stderr, "%s: node line missing apc_ws outlet number.\n", prog_name);    
    exit(EXIT_FAILURE);
  }

  if (!strcmp(value, ""))
  {
    fprintf(stderr, "%s: apc_ms bad outlet number of \"%s\".\n", prog_name, value);
    exit(EXIT_FAILURE);
  }

  apc->st_outlet = atoi(value);

  return(0);
}



static void print_brcd_zone(stomith_list_t *st, int flags)
{
  stm_brcd_zone_t *silk = &st->st.brcd_zone_s;
  char ip[20];

  inet_ntoa(silk->st_ipaddr, ip);

  switch(flags)
  {
  case PRINT_STOMITH_GENERAL:
    printf("ipaddr:\t\t%s\n", ip);
    printf("login:\t\t%s\n", silk->st_login);
    printf("passwd:\t\t%s\n", silk->st_passwd);
    printf("command:\t%s\n", silk->st_command);
    printf("cfgname:\t%s\n", silk->st_cfgname);
    printf("comzone:\t%s\n", silk->st_comzone);
    break;

  case PRINT_STOMITH_ALL:
    printf("    %s, %s, %s, %s, %s, %s, %s\n", ip, silk->st_login,
	   silk->st_passwd, silk->st_command, silk->st_cfgname, silk->st_comzone,
	   silk->st_actname);
    break;

  case PRINT_STOMITH_SPECIFIC:
    printf("%s\n", silk->st_actname);
    break;
  }    
}



static void print_brcd_port(stomith_list_t *st, int flags)
{
  stm_brcd_port_t *bp = &st->st.brcd_port_s;
  char ip[20];

  inet_ntoa(bp->st_ipaddr, ip);

  switch(flags)
  {
  case PRINT_STOMITH_GENERAL:
    printf("ipaddr:\t\t%s\n", ip);
    printf("login:\t\t%s\n", bp->st_login);
    printf("passwd:\t\t%s\n", bp->st_passwd);
    break;

  case PRINT_STOMITH_ALL:
    printf("    %s, %s, %s\n", ip, bp->st_login,
	   bp->st_passwd);
    break;

  case PRINT_STOMITH_SPECIFIC:
    printf("%d\n", bp->st_port);
    break;
  }    
}



static void print_wti_nps(stomith_list_t *st, int flags)
{
  stm_wti_nps_t *nps = &st->st.wti_nps_s;
  char ip[20];

  inet_ntoa(nps->st_ipaddr, ip);

  switch(flags)
  {
  case PRINT_STOMITH_GENERAL:
    printf("ipaddr:\t\t%s\n", ip);
    printf("passwd:\t\t%s\n", nps->st_passwd);
    break;

  case PRINT_STOMITH_ALL:
    printf("    %s, %s, %u\n", ip, 
	   nps->st_passwd, nps->st_plug);
    break;

  case PRINT_STOMITH_SPECIFIC:
    printf("%u\n", nps->st_plug);
    break;
  }
}



static void print_vacm_emp(stomith_list_t *st, int flags)
{
  stm_vacm_emp_t *va = &st->st.vacm_emp_s;

  switch(flags)
  {
  case PRINT_STOMITH_GENERAL:
    printf("nexxus:\t\t%s\n", va->st_nexxus);
    printf("login:\t\t%s\n", va->st_login);
    printf("passwd:\t\t%s\n", va->st_passwd);
    break;

  case PRINT_STOMITH_ALL:
    printf("    %s, %s, %s, %s\n", va->st_nexxus, va->st_login, va->st_passwd, 
	   va->st_nodeid);
    break;

  case PRINT_STOMITH_SPECIFIC:
    printf("%s\n", va->st_nodeid);
    break;
  }
}



static void print_gnbd(stomith_list_t *st, int flags)
{
  stm_gnbd_t *sg = &st->st.gnbd_s;
  char ip[20];

  inet_ntoa(sg->st_ipaddr, ip);

  switch(flags)
  {
  case PRINT_STOMITH_GENERAL:
    break;

  case PRINT_STOMITH_ALL:
    printf("    %s\n", ip);
    break;

  case PRINT_STOMITH_SPECIFIC:
    break;
  }
}



static void print_meatware(stomith_list_t *st, int flags)
{
  stm_meatware_t *meat = &st->st.meatware_s;
  char ip[20];

  inet_ntoa(meat->st_ipaddr, ip);

  switch(flags)
  {
  case PRINT_STOMITH_GENERAL:
    break;

  case PRINT_STOMITH_ALL:
    printf("    %s\n", ip);
    break;

  case PRINT_STOMITH_SPECIFIC:
    break;
  }
}



static void print_vx_hub(stomith_list_t *st, int flags)
{
  stm_vx_hub_t *hub = &st->st.vx_hub_s;
  char ip[20];

  inet_ntoa(hub->st_ipaddr, ip);

  switch(flags)
  {
  case PRINT_STOMITH_GENERAL:
    printf("ipaddr:\t\t%s\n", ip);
    printf("passwd:\t\t%s\n", hub->st_passwd);
    break;

  case PRINT_STOMITH_ALL:
    printf("    %s, %s, %u\n", ip, 
           hub->st_passwd, hub->st_port);
    break;

  case PRINT_STOMITH_SPECIFIC:
    printf("%u\n", hub->st_port);
    break;
  }
}



static void print_apc_ms(stomith_list_t *st, int flags)
{
  stm_apc_ms_t *apc = &st->st.apc_ms_s;
  char ip[20];

  inet_ntoa(apc->st_ipaddr, ip);

  switch(flags)
  {
  case PRINT_STOMITH_GENERAL:
    printf("ipaddr:\t\t%s\n", ip);
    printf("login:\t\t%s\n", apc->st_login);
    printf("passwd:\t\t%s\n", apc->st_passwd);
    break;

  case PRINT_STOMITH_ALL:
    printf("    %s, %s, %s, %u\n", ip, 
	   apc->st_login, apc->st_passwd, apc->st_outlet);
    break;

  case PRINT_STOMITH_SPECIFIC:
    printf("%u\n", apc->st_outlet);
    break;
  }
}



/* 
 * addparam_make_model() routines scan the machine-specific parameters following
 * a stomith method listed for a node.  
 */


int addparam_stomith_specific(node_list_t *node, stomith_list_t *st, char *value)
{
  int error = 0;

  switch(st->magic)
  {
  case STM_BRCD_ZONE:
    error = addparam_brcd_zone(st, value);
    break;

  case STM_BRCD_PORT:
    error = addparam_brcd_port(st, value);
    break;

  case STM_WTI_NPS:
    error = addparam_wti_nps(st, value);
    break;

  case STM_VACM_EMP:
    error = addparam_vacm_emp(st, value);
    break;

  case STM_GNBD:
    error = addparam_gnbd(st, node);
    break;

  case STM_MEATWARE:
  case STM_MANUAL:
    error = addparam_meatware(st, node);
    break;

  case STM_VXHUB_BYPASS:
    error = addparam_vx_hub(st, value);
    break;

  case STM_APC_MS:
    error = addparam_apc_ms(st, value);
    break;

  default:
    error = -1;
  }

  return(error);
}



/* 
 * add_make_model() routines scan for the unique parameters required 
 * for a stomith method.  i.e. those following a "STOMITH: make_model"
 */


int add_stomith_specific(stomith_list_t *st, keypair_t *list)
{
  int error = 0;

  switch(st->magic)
  {
  case STM_BRCD_ZONE:
    error = add_brcd_zone(st, list);
    break;

  case STM_BRCD_PORT:
    error = add_brcd_port(st, list);
    break;

  case STM_WTI_NPS:
    error = add_wti_nps(st, list);
    break;
    
  case STM_VACM_EMP:
    error = add_vacm_emp(st, list);
    break;

  case STM_GNBD:
    error = add_gnbd(st, list);
    break;

  case STM_MEATWARE:
  case STM_MANUAL:
    error = add_meatware(st, list);
    break;

  case STM_VXHUB_BYPASS:
    error = add_vx_hub(st, list);
    break;

  case STM_APC_MS:
    error = add_apc_ms(st, list);
    break;

  default:
    error = -1;
  }

  return(error);
}



void print_stomith_specific(stomith_list_t *st, int flags)
{
  switch(st->magic)
  {
  case STM_BRCD_ZONE:
    print_brcd_zone(st, flags);
    break;

  case STM_BRCD_PORT:
    print_brcd_port(st, flags);
    break;

  case STM_WTI_NPS:
    print_wti_nps(st, flags);
    break;
    
  case STM_VACM_EMP:
    print_vacm_emp(st, flags);
    break;

  case STM_GNBD:
    print_gnbd(st, flags);
    break;

  case STM_MEATWARE:
  case STM_MANUAL:
    print_meatware(st, flags);
    break;

  case STM_VXHUB_BYPASS:
    print_vx_hub(st, flags);
    break;

  case STM_APC_MS:
    print_apc_ms(st, flags);
    break;

  default:
  }
}    
  

