/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portions Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: gfshex.c
**
**    Description:
**	GFS-specific routines supportive of hexedit.
**
*******************************************************************************
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include "gfs_ondisk.h"
#include "osi_endian.h"

#include "hexedit.h"
#include "gfshex.h"





/******************************************************************************
*******************************************************************************
**
** do_dinode_extended()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void do_dinode_extended(gfs_dinode_t *di, char *buf)
{
  unsigned int x, y, count;
  gfs_dirent_t de;
  uint64 p, last;


  if (di->di_height > 0)
  {
    printf("\nIndirect Pointers\n\n");

    for (x = sizeof(gfs_dinode_t), y = 0;
	 x < bsize;
	 x += sizeof(uint64), y++)
    {
      p = gfs64_to_cpu(*(uint64 *)(buf + x));

      if (p)
	printf("  %u -> %"FMT_64"u\n", y, p);
    }
  }


  else if (di->di_type == FILE_DIR &&
	   !(di->di_flags & GFS_DIF_EXHASH))
  {
    printf("\nDirectory Entries:\n");

    for (x = sizeof(gfs_dinode_t); x < bsize; x += de.de_rec_len)
    {
      printf("\n");
      gfs_dirent_in(&de, buf + x);
      if (de.de_inum.no_formal_ino)
	gfs_dirent_print(&de, buf + x + sizeof(gfs_dirent_t), TRUE);
    }
  }


  else if (di->di_type == FILE_DIR &&
	   (di->di_flags & GFS_DIF_EXHASH) &&
	   di->di_height == 0)
  {
    printf("\nLeaf Pointers:\n\n");

    last = gfs64_to_cpu(*(uint64 *)(buf + sizeof(gfs_dinode_t)));
    count = 0;
    
    for (x = sizeof(gfs_dinode_t), y = 0;
	 y < (1 << di->di_depth);
	 x += sizeof(uint64), y++)
    {
      p = gfs64_to_cpu(*(uint64 *)(buf + x));

      if (p != last)
      {
	printf("  %u:  %"FMT_64"u\n", count, last);
	last = p;
	count = 1;
      }
      else
	count++;

      if ((y + 1) * sizeof(uint64) == di->di_size)
	printf("  %u:  %"FMT_64"u\n", count, last);
    }
  }
}


/******************************************************************************
*******************************************************************************
**
** do_leaf_extended()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void do_leaf_extended(char *buf)
{
  int x;
  gfs_dirent_t de;


  printf("\nDirectory Entries:\n");

  for (x = sizeof(gfs_leaf_t); x < bsize; x += de.de_rec_len)
  {
    printf("\n");
    gfs_dirent_in(&de, buf + x);
    if (de.de_inum.no_formal_ino)
	gfs_dirent_print(&de, buf + x + sizeof(gfs_dirent_t), TRUE);
  }
}


/******************************************************************************
*******************************************************************************
**
** int display_gfs()
**
** Description:
**   This routine...
**
** Input(s):
**  *buffer   - 
**   extended - 
**
** Returns:
**   0 if OK, 1 on error.
**
*******************************************************************************
******************************************************************************/

int display_gfs(int extended)
{
  gfs_meta_header_t mh;
  gfs_sb_t sb;
  gfs_rgrp_t rg;
  gfs_dinode_t di;
  gfs_leaf_t lf;
  gfs_log_header_t lh;
  gfs_log_descriptor_t ld;

  char *buf;
  uint32 magic;


  type_alloc(buf, char, bsize);

  do_lseek(fd, block * bsize);
  do_read(fd, buf, bsize);

  
  magic = gfs32_to_cpu(*(uint32 *)buf);


  switch (magic)
  {
  case GFS_MAGIC:
    gfs_meta_header_in(&mh, buf);

    switch (mh.mh_type)
    {
    case GFS_METATYPE_SB:
      printf("Superblock:\n\n");
      gfs_sb_in(&sb, buf);
      gfs_sb_print(&sb, 0);

      if (extended)
	printf("\nNo Extended data\n");

      break;


    case GFS_METATYPE_RG:
      printf("Resource Group Header:\n\n");
      gfs_rgrp_in(&rg, buf);
      gfs_rgrp_print(&rg, 0);

      if (extended)
	printf("\nNo Extended data\n");

      break;


    case GFS_METATYPE_RB:
      printf("Resource Group Bitmap:\n\n");
      gfs_meta_header_print(&mh, 0);

      if (extended)
	printf("\nNo Extended data\n");

      break;


    case GFS_METATYPE_DI:
      printf("Dinode:\n\n");
      gfs_dinode_in(&di, buf);
      gfs_dinode_print(&di, 0);

      if (extended)
	do_dinode_extended(&di, buf);

      break;


    case GFS_METATYPE_LF:
      printf("Leaf:\n\n");
      gfs_leaf_in(&lf, buf);
      gfs_leaf_print(&lf, 0);

      if (extended)
	do_leaf_extended(buf);

      break;


    case GFS_METATYPE_IN:
      printf("Indirect Block:\n\n");
      gfs_meta_header_print(&mh, 0);

      if (extended)
	printf("\nNo Extended data\n");

      break;


    case GFS_METATYPE_JD:
      printf("Journaled File Block:\n\n");
      gfs_meta_header_print(&mh, 0);

      if (extended)
	printf("\nNo Extended data\n");

      break;


    case GFS_METATYPE_LH:
      printf("Log Header:\n\n");
      gfs_log_header_in(&lh, buf);
      gfs_log_header_print(&lh, 0);

      if (extended)
	printf("\nNo Extended data\n");

      break;


    case GFS_METATYPE_LD:
      printf("Log Descriptor:\n\n");
      gfs_desc_in(&ld, buf);
      gfs_desc_print(&ld, 0);

      if (extended)
	printf("\nNo Extended data\n");

      break;


    default:
      printf("Unknown metadata type\n");
      break;
    }

    break;


  default:
    printf("Unknown block type\n");
    break;
  };


  free(buf);


  return(0);
}


/******************************************************************************
*******************************************************************************
**
** int edit_gfs()
**
** Description:
**   This routine...
**
** Input(s):
**  *buffer   - 
**   extended - 
**
** Returns:
**   0 if OK, 1 on error.
**
*******************************************************************************
******************************************************************************/

int edit_gfs(char *arg1, char *arg2, char *arg3)
{
  int row, col, byte;

  if (!strncmp(arg3, "", 1))
  {
    fprintf(stderr, "%s:  invalid number of arguments\n", prog_name);
    return(-EINVAL);
  }


  row = atol(arg1);
  col = atol(arg2);
  sscanf(arg3, "%x", &byte);


  if (row >= SCREEN_HEIGHT)
  {
    fprintf(stderr, "%s:  row is out of range for set\n", prog_name);
    return(-EINVAL);
  }
  
  if (col >= SCREEN_WIDTH)
  {
    fprintf(stderr, "%s:  column is out of range for set\n", prog_name);
    return(-EINVAL);
  }

  if (byte > 255)
  {
    fprintf(stderr, "%s:  byte value is out of range for set\n", prog_name);
    return(-EINVAL);
  }


  do_lseek(fd, block * bsize + start * SCREEN_WIDTH + 
	   row * SCREEN_WIDTH + col);
  
  do_write(fd, &byte, 1);

  fsync(fd);

  return(0);
}






