/******************************************************************************
*******************************************************************************
**
**    Copyright 2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Seth D. Van Oort <oort@sistina.com>
** 
*******************************************************************************
******************************************************************************/

#include <stdlib.h>
#include <stdio.h>

#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netinet/in.h>

/* Here's how this is supposed to work. Each participating machine has their own
 * circular buffer to which they successively write incrementing integers. All
 * of the machines' buffers are in the same fs block. Each time a machine writes
 * to its buffer, only one integer is written. In the examples below, when I say
 * that a machine writes, it implies an arbitrary number of writes that are not
 * broken up by a write from a different machine.
 *
 *  Suppose we have machine A and B with a sequence of writes looking like
 *
 * A writes ...
 *
 * | Buffer A                | Buffer B                  |
 * | 1,2,3,4,0,0,0,0,0,0,0,0 | 0,0,0,0,0,0,0,0,0,0,0,0 0 |
 *
 * B writes ...
 *
 * | 1,2,3,4,0,0,0,0,0,0,0,0 | 1,2,0,0,0,0,0,0,0,0,0,0,0 |
 *
 * At this point, both A and B have the block buffered. If A now writes ...
 *
 * | 1,2,3,4,5,6,0,0,0,0,0,0 | 1,2,0,0,0,0,0,0,0,0,0,0,0 | 
 *
 * ... and B does not invalidate the buffer correctly, when B writes we will
 * get something like ...
 *
 * | 1,2,3,4,0,0,0,0,0,0,0,0 | 1,2,3,4,0,0,0,0,0,0,0,0,0 |
 *
 * Now the next time A writes, it will look like 
 *
 * | 1,2,3,4,0,0,7,0,0,0,0,0 | 1,2,3,4,0,0,0,0,0,0,0,0,0 |
 *
 * Because B did not invalidate the buffer, it erased some of A's work which
 * is evident from the gap between the 4 and 7. Note that A could cycle through
 * its buffer and write the gap over with correct data before any other machine
 * checks and realizes the inconsistency. The variable read_pct controls how
 * often the buffer is checked. If read_pct = 100, on average SOME participating
 * machine will check the block once in the time it takes the average machine to
 * cycle through its buffer. read_pct = 200 doubles the frequency, etc.
 */

int block_size = 4096;

/* number of concurrent clients */

int nr_clients = 1;

/* the higher, the better chance of catching a coherency slip
 * that has occurred before it gets written over
 */

int read_pct = 100; 
int disk_busy_pct = 100; /* how busy to keep the disk */
int disk_latency = 3000; /* micros */
int rollover; /* how often you want each client counter to roll over */

int client_nr = 0, offset, buf_size;
char *filename = "caching";

void die (char *s) {
  fprintf(stderr, s);
  exit(-1);
}

#define OPTION_STRING "b:c:n:"

void parse_args(int argc, char **argv)
{
  int optchar;

  while ((optchar = getopt(argc, argv, OPTION_STRING)) > 0)
    {

      switch (optchar)
	{
	case 'c':
	  client_nr = atoi(optarg);
	  break;

	case 'n':
	  nr_clients = atoi(optarg);
	  break;

	case 'b':
	  block_size = atoi(optarg);
	  break;

	}
    }
}

void check_coherency();

int fd, int_size;
int offset_char, offset_int;
int buf_size_char, buf_size_int;

int main (int argc, char **argv) 
{
  int current = 0, ret, total_writes = 0;
  int read_mod;

  parse_args(argc, argv);

  int_size = block_size / sizeof(int);
  offset_char = block_size / nr_clients * client_nr; /* in chars */
  offset_int = offset_char / 4;
  buf_size_char = block_size / nr_clients; /* in chars */
  buf_size_int = buf_size_char / sizeof(int); /* in ints */
  read_mod = buf_size * 100 / read_pct;
  rollover = buf_size * 40;
  

  fd = open (filename, O_RDWR);

  if (fd < 0)
    die ("can't open file\n");

  while (1)
    {
      int cycle_time, pause, tmp;
      cycle_time = disk_latency * nr_clients;
      pause = cycle_time * 100 / disk_busy_pct;

      pause = random() % pause;

      usleep(pause);

      tmp = ntohl(current);

      lseek(fd, offset + (current % buf_size_int) * sizeof(int), SEEK_SET);
      ret = write(fd, &tmp, sizeof(int));

      if (ret < 0)
	die ("error on write\n");

      if ((total_writes % read_mod) == read_mod - 1)
	check_coherency();

      current = (current + 1) % rollover;
      total_writes++;
    }

  return 0;
}


int is_a_drop(int a, int b, int rollover)
{
  if ((a == rollover -1) && b == 0)
    return 0;

  if (b < a)
    return 1;
  else
    return 0;
}

void do_autopsy(int *buf, int size, int client_nr)
{
  int x;

  printf("autopsy for client %d\n", client_nr);

  for (x=0; x < size; x++) {
    if (is_a_drop(buf[x], buf[(x+1) % size], rollover))
      printf("\ndrop\n");

    printf("%d ", buf[x]);

    if (x % 8 == 7)
      printf("\n");
  }
}

void check_client(int client_nr, int *buf, int size)
{
  int nr_drops = 0, x;

  for (x=0; x < size; x++) {
    if (is_a_drop(buf[x], buf[(x+1) % size], rollover))
      nr_drops++;
  }

  if (nr_drops > 1)
    do_autopsy(buf, size, client_nr);
}

void check_coherency()
{
  int ret, i;

  char buf[10240];

  lseek(fd, 0, SEEK_SET);
  ret = read(fd, buf, block_size);

  if (ret < 0)
    die ("error on read\n");

  for (i = 0; i < nr_clients; i++)
    {
      check_client(i, (int *)(buf + block_size / nr_clients * i), block_size / nr_clients);
    }
}


  
