/******************************************************************************
*******************************************************************************
**
**    Copyright 2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Description:
**      Implement a simple dd command that can do I/O using mmap().
**
*******************************************************************************
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/mman.h>



#define die(fmt, args...) \
do \
{ \
  fprintf(stderr, "%s: ", prog_name); \
  fprintf(stderr, fmt, ##args); \
  exit(EXIT_FAILURE); \
} \
while (0);

#define TRUE (1)
#define FALSE (0)



char *prog_name;
unsigned long long read_offset = 0;
unsigned long long write_offset = 0;


void do_bothmapped(int ifd, int ofd, unsigned bs)
{
  char *in;
  char *out;

  in = mmap(NULL, bs, PROT_READ, MAP_SHARED, ifd, read_offset);
  if (in == MAP_FAILED)
    die("can't mmap file: %s\n", strerror(errno));

  out = mmap(NULL, bs, PROT_WRITE, MAP_SHARED, ofd, write_offset);
  if (out == MAP_FAILED)
    die("can't mmap file: %s\n", strerror(errno));
  
  memcpy(out, in, bs);

  munmap(in, bs);
  munmap(out, bs);

  read_offset += bs;
  write_offset += bs;
}

void do_omapped(int ifd, int ofd, unsigned bs)
{
  char *out;
  unsigned int err;

  out = mmap(NULL, bs, PROT_WRITE, MAP_SHARED, ofd, write_offset);
  if (out == MAP_FAILED)
    die("can't mmap file: %s\n", strerror(errno));

  err = read(ifd, out, bs);
  if (err != bs)
    die("bad read return %u: %s\n", err, strerror(errno));
  
  munmap(out, bs);

  read_offset += bs;
  write_offset += bs;
}

void do_imapped(int ifd, int ofd, unsigned bs)
{
  char *in;
  unsigned int err;

  in = mmap(NULL, bs, PROT_READ, MAP_SHARED, ifd, read_offset);
  if (in == MAP_FAILED)
    die("can't mmap file: %s\n", strerror(errno));

  err = write(ofd, in, bs);
  if (err != bs)
    die("bad write return %u: %s\n", err, strerror(errno));

  munmap(in, bs);

  read_offset += bs;
  write_offset += bs;
}


void do_mapped_read(int fd, char *buf, unsigned int bs)
{
  char *data;

  data = mmap(NULL, bs, PROT_READ, MAP_SHARED, fd, read_offset);
  if (data == MAP_FAILED)
    die("can't mmap file: %s\n", strerror(errno));

  memcpy(buf, data, bs);

  munmap(data, bs);

  read_offset += bs;
}


void do_raw_read(int fd, char *buf, unsigned int bs)
{
  unsigned int out;

  out = read(fd, buf, bs);
  if (out != bs)
    die("bad read return %u: %s\n", out, strerror(errno));
}


void do_mapped_write(int fd, char *buf, unsigned int bs)
{
  char *data;

  data = mmap(NULL, bs, PROT_WRITE, MAP_SHARED, fd, write_offset);
  if (data == MAP_FAILED)
    die("can't mmap file: %s\n", strerror(errno));

  memcpy(data, buf, bs);

  munmap(data, bs);

  write_offset += bs;
}


void do_raw_write(int fd, char *buf, unsigned int bs)
{
  unsigned int out;

  out = write(fd, buf, bs);
  if (out != bs)
    die("bad write return %u: %s\n", out, strerror(errno));
}


int main(int argc, char *argv[])
{
  int arg;

  char *ifname = NULL;
  char *ofname = NULL;
  int imapped = FALSE, omapped = FALSE;
  unsigned int bs = 4096;
  unsigned long long move = 0;
  unsigned long long count = 0;
  int do_fsync = FALSE;
  int verbose = FALSE;
  int notrunc = FALSE;
  int directcopy = FALSE;
  
  int done_bs = FALSE;
  int done_amount = FALSE;
  int done_fsync = FALSE;
  int done_notrunc = FALSE;
  int done_directcopy = FALSE;

  int ifd = -1, ofd = -1;

  char c;
  char *data;
  unsigned long long x;

  struct timeval tv1, tv2;
  unsigned long long t1, t2;
  double seconds, mbps;

  struct stat st;
  int error;



  prog_name = argv[0];



  if (argc == 1)
    die("options:  if, imf, omf, of, bs, move, count, fsync, notrunc, -v\n");



  for (arg = 1; arg < argc; arg++)
  {
    if (strncmp(argv[arg], "if=", 3) == 0)
    {
      if (ifname)
	die("you're only allowed one \"i?f=\"\n");

      ifname = argv[arg] + 3;
      imapped = FALSE;
    }
    else if (strncmp(argv[arg], "imf=", 4) == 0)
    {
      if (ifname)
	die("you're only allowed one \"i?f=\"\n");

      ifname = argv[arg] + 4;
      imapped = TRUE;
    }
    else if (strncmp(argv[arg], "of=", 3) == 0)
    {
      if (ofname)
	die("you're only allowed one \"o?f=\"\n");

      ofname = argv[arg] + 3;
      omapped = FALSE;
    }
    else if (strncmp(argv[arg], "omf=", 4) == 0)
    {
      if (ofname)
	die("you're only allowed one \"o?f=\"\n");

      ofname = argv[arg] + 4;
      omapped = TRUE;
    }
    else if (strncmp(argv[arg], "bs=", 3) == 0)
    {
      if (done_bs)
	die("you're only allowed one \"bs=\"\n");
      done_bs = TRUE;

      bs = atoi(argv[arg] + 3);
      c = argv[arg][strlen(argv[arg]) - 1];
      if (c == 'k')
	bs <<= 10;
      else if (c == 'm')
	bs <<= 20;
      else if (c == 'g')
	bs <<= 30;

      if (!bs || bs > 4194304)
	die("invalid blocksize: %u\n", bs);
    }
    else if (strncmp(argv[arg], "move=", 5) == 0)
    {
      if (done_amount)
	die("you're only allowed one amount\n");
      done_amount = TRUE;

      move = atoll(argv[arg] + 5);
      c = argv[arg][strlen(argv[arg]) - 1];
      if (c == 'k')
	move <<= 10;
      else if (c == 'm')
	move <<= 20;
      else if (c == 'g')
	move <<= 30;

      if (!move)
	die("zero move\n");
    }
    else if (strncmp(argv[arg], "count=", 6) == 0)
    {
      if (done_amount)
	die("you're only allowed one amount\n");
      done_amount = TRUE;

      count = atoll(argv[arg] + 6);

      if (!count)
	die("zero count\n");
    }
    else if (strncmp(argv[arg], "fsync=", 6) == 0)
    {
      if (done_fsync)
	die("you're only allowed one \"fsync=\"\n");
      done_fsync = TRUE;

      do_fsync = atoi(argv[arg] + 6);
    }
    else if (strncmp(argv[arg], "notrunc=", 8) == 0)
    {
      if (done_notrunc)
	die("you're only allowed one \"notrunc=\"\n");
      done_notrunc = TRUE;

      notrunc = atoi(argv[arg] + 8);
    }
    else if (strncmp(argv[arg], "direct=", 8) == 0)
    {
      if (done_directcopy)
	die("you're only allowed one \"direct=\"\n");
      done_directcopy = TRUE;

      directcopy = atoi(argv[arg] + 8);
    }
    else if (strcmp(argv[arg], "-v") == 0)
    {
      verbose = TRUE;
    }
    else
    {
      die("unknown/unimplemented option: %s\n", argv[arg]);
    }
  }



  if (!done_amount)
    die("need an amount to move\n");

  if (verbose)
  {
    fprintf(stderr, "ifname = %s (%s)\n", ifname, (imapped) ? "mapped" : "raw");
    fprintf(stderr, "ofname = %s (%s)\n", ofname, (omapped) ? "mapped" : "raw");
    fprintf(stderr, "bs = %u\n", bs);
    fprintf(stderr, "move = %qu\n", move);
    fprintf(stderr, "count = %qu\n", count);
    fprintf(stderr, "fsync = %d\n", do_fsync);
    fprintf(stderr, "notrunc = %d\n", notrunc);
  }

  if (!count)
  {
    if (move % bs != 0)
      die("move size (%qu) not divisible by block size (%u)\n", move, bs);
    count = move / bs;
  }



  data = (char *)malloc(bs);
  if (!data)
    die("out of memory\n");
  memset(data, 0, bs);



  gettimeofday(&tv1, NULL);



  if (ifname)
  {
    if (strcmp(ifname, "-") == 0)
      ifd = STDIN_FILENO;
    else
    {
      ifd = open(ifname, O_RDONLY);
      if (ifd < 0)
	die("can't open file %s: %s\n", ifname, strerror(errno));
      if (verbose)
      {
	error = fstat(ifd, &st);
	if (error)
	  die("can't stat file %s: %s\n", ifname, strerror(errno));
	fprintf(stderr, "input inode = %llu\n", (unsigned long long)st.st_ino);
      }
    }
  }

  if (ofname)
  {
    if (strcmp(ofname, "-") == 0)
      ofd = STDOUT_FILENO;
    else
    {
      ofd = open(ofname, O_RDWR | O_CREAT | ((notrunc) ? 0 : O_TRUNC), 0644);
      if (ofd < 0)
	die("can't open file %s: %s\n", ofname, strerror(errno));
      if (verbose)
      {
	error = fstat(ofd, &st);
	if (error)
	  die("can't stat file %s: %s\n", ofname, strerror(errno));
	fprintf(stderr, "output inode = %llu\n", (unsigned long long)st.st_ino);
      }
    }
  }



  if (ofd >= 0 && omapped)
  {
    if (ftruncate(ofd, bs * count) < 0)
      die("can't truncate bigger: %s\n", strerror(errno));
  }



  for (x = 0; x < count; x++)
  {
    if (directcopy && ifd >= 0 && ofd >= 0 && (imapped || omapped))
    {
	if (imapped && omapped)
	    do_bothmapped(ifd, ofd, bs);
	else if (imapped)
	    do_imapped(ifd, ofd, bs);
	else
	    do_omapped(ifd, ofd, bs);
    }  
    else
    {
	if (ifd >= 0)
	{
	  if (imapped)
	    do_mapped_read(ifd, data, bs);
	  else
	    do_raw_read(ifd, data, bs);
	}

	if (ofd >= 0)
	{
	  if (omapped)
	    do_mapped_write(ofd, data, bs);
	  else
	    do_raw_write(ofd, data, bs);
	}
    }
    if (verbose)
      fprintf(stderr, "%qu\n", x);
  }

 

  if (ifd >= 0)
    close(ifd);

  if (ofd >= 0)
  {
    if (do_fsync)
      fsync(ofd);
    close(ofd);
  }



  gettimeofday(&tv2, NULL);

  t1 = (unsigned long long)tv1.tv_sec * 1000000 + tv1.tv_usec;
  t2 = (unsigned long long)tv2.tv_sec * 1000000 + tv2.tv_usec;

  seconds = (t2 - t1) / 1000000.0;
  mbps = bs * count / 1048576.0 / seconds;

  fprintf(stderr, "seconds = %.4f, MB/s = %.4f\n", seconds, mbps);



  exit(EXIT_SUCCESS);
}



