/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portions Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: fs_geometry.c
**
**    Description:
**	Functions that compute the rgrp/journal geometry
**
*******************************************************************************
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <assert.h>
#include <time.h>

#include "global.h"
#include "pool_sptypes.h"
#include "pool.h"
#include "gfs_ondisk.h"

#include "mkfs_gfs.h"



#define MIN(x, y) (((x) < (y)) ? (x) : (y))
#define MAX(x, y) (((x) > (y)) ? (x) : (y))

#define HEADER_SIZE (14 * 1024)





/******************************************************************************
*******************************************************************************
**
** how_many_rgrps()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

static uint64 how_many_rgrps(commandline_t *comline, pool_spstatus_t *sps)
{
  uint32 optimal_rgrp_size;
  uint64 nrgrp;


  /*  Number of resource groups based on a rgrp header size of HEADER_SIZE.  */

  optimal_rgrp_size = HEADER_SIZE * GFS_NBBY;

  nrgrp = sps->sps_len / optimal_rgrp_size;


  /*  If the fs size is less than optimal_rgrp_size, default to 10 rgrps.
      We're guaranteed by the pool_geometry code that there are at least
      100 blocks in the subpool, so we know there's enough room for the
      header and at least nine data block.  */

  if (nrgrp < 10)
    nrgrp = 10;


  if (comline->debug)
    printf("  nrgrp = %"FMT_64"u\n", nrgrp);


  return nrgrp;
}


/******************************************************************************
*******************************************************************************
**
** compute_rgrp_layout()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

rgrp_list_t *compute_rgrp_layout(commandline_t *comline, pool_splist_t *splist)
{
  pool_spstatus_t *sps;
  rgrp_list_t *rlist = NULL, *rtmp, *rlast = NULL;
  uint64 rgrp, nrgrp;
  uint32 sp;
  int first_sp = TRUE;


  for (sp = 0; sp < splist->spl_count; sp++)
  {
    sps = &splist->spl_list[sp];

    if (sps->sps_type == SPTYPE_GFS_DATA)
    {
      /*  If this is the first subpool reserve space for the superblock  */

      if (first_sp)
	sps->sps_len -= comline->sb_addr + 1;


      if (comline->debug)
	printf("\nData Subpool %d\n", sp);

      nrgrp = how_many_rgrps(comline, sps);


      for (rgrp = 0; rgrp < nrgrp; rgrp++)
      {
	type_zalloc(rtmp, rgrp_list_t, 1);

	rtmp->subpool = sp;

	if (rgrp)
	{
	  rtmp->rg_offset = rlast->rg_offset + rlast->rg_length;
	  rtmp->rg_length = sps->sps_len / nrgrp;
	}
	else
	{
	  rtmp->rg_offset = sps->sps_start;
	  rtmp->rg_length = sps->sps_len - (nrgrp - 1) * (sps->sps_len / nrgrp);

	  if (first_sp)
	    rtmp->rg_offset += comline->sb_addr + 1;
	}

	if (rlast)
	  rlast->next = rtmp;
	else
	  rlist = rtmp;

	rlast = rtmp;
      }

      first_sp = FALSE;

      comline->rgrps += nrgrp;
    }
  }


  if (comline->debug)
  {
    printf("\n");

    for (rtmp = rlist; rtmp; rtmp = rtmp->next)
      printf("subpool %u:  rg_o = %"FMT_64"u, rg_l = %"FMT_64"u\n",
	     rtmp->subpool,
	     rtmp->rg_offset, rtmp->rg_length);
  }


  return rlist;
}


/******************************************************************************
*******************************************************************************
**
** compute_journal_layout()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

journal_list_t *compute_journal_layout(commandline_t *comline, pool_splist_t *splist)
{
  pool_spstatus_t *sps;
  journal_list_t *jlist = NULL, *jtmp, *jlast = NULL;
  int sp, j = 0;
  uint64 boffset, bcount;


  for (sp = 0; sp < splist->spl_count; sp++)
  {
    sps = &splist->spl_list[sp];

    if (sps->sps_type == SPTYPE_GFS_JOURNAL)
    {
      type_zalloc(jtmp, journal_list_t, 1);

      /*  Align the journals on seg_size boundries  */

      boffset = sps->sps_start;
      bcount = sps->sps_len;

      if (bcount * comline->bsize < 32 << 20)
	die("journal %d is too small (minimum size is 32 MB)\n", j);

      if (boffset % comline->seg_size)
      {
	bcount -= comline->seg_size - (boffset % comline->seg_size);
	boffset += comline->seg_size - (boffset % comline->seg_size);
      }

      jtmp->start = boffset;
      jtmp->segments = bcount / comline->seg_size;

      if (jlist)
	jlast->next = jtmp;
      else
	jlist = jtmp;

      jlast = jtmp;

      j++;
    }
  }


  if (comline->debug)
  {
    printf("\n");

    for (jtmp = jlist, j = 0; jtmp; jtmp = jtmp->next, j++)
      printf("journal %u:  start = %"FMT_64"u, segments = %u\n",
	     j, jtmp->start, jtmp->segments);
  }


  return jlist;
}

