/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portions Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: pool_geometry.c
**
**    Description:
**	Functions that find out about the underlying device
**
*******************************************************************************
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <assert.h>
#include <time.h>

#include "global.h"
#define WANT_SPTYPE_NAMES
#include "pool_sptypes.h"
#include "pool.h"
#include "gfs_ondisk.h"

#include "mkfs_gfs.h"



/*  Needed for BLKGETSIZE ioctl().
    Replace it with whatever the appropriate thing is for your OS.  */
#if defined(__linux__)

#include <sys/mount.h>

#elif defined(__FreeBSD__)

#include <pool_fbsd.h>
#include <sys/disklabel.h>
#include <ctype.h>
#define BLKGETSIZE DIOCGDINFO
#define atoll(x) strtoq(x,(char **)NULL,10)

#endif  /*  defined(__FreeBSD__)  */

#ifndef BTOBB
#define BTOBB(x) (((x) + GFS_BASIC_BLOCK - 1) / GFS_BASIC_BLOCK)
#endif  /*  BTOBB  */





/******************************************************************************
*******************************************************************************
**
** get_pool_geometry()
**
** Description:
**   Talk to the pool and find out the number of subpools, their size, and
**   the number of Glocks.
**
** Input(s):
**   comline - the command line arguments (Specifies the pool name)
**
** Output(s):
**   splist - the structures containing the subpool geometry
**
*******************************************************************************
******************************************************************************/

int get_pool_geometry(commandline_t *comline, pool_splist_t *splist)
{
  int fd;
  int i, data_subpools = 0;


  /*  Open pool device  */

  fd = open(comline->device, O_RDONLY);
  if (fd < 0)
    die("Error opening \"%s\": %s\n", comline->device, strerror(errno));


  /*  Get subpool information  */

  for (i = 128; i < 65536; i += 128) 
  {
    splist->spl_count = i;
    type_zalloc(splist->spl_list, pool_spstatus_t, i);

    if (ioctl(fd, POOL_SPLIST, splist) != 0) 
      return -1;
    
    if (splist->spl_count <= i)
      break;
    
    free(splist->spl_list);
  }
  
  if (i >= 65536)
    die("unable to get pool subpool list: to many subpools.\n");

  assert(splist->spl_count > 0);
  

  close(fd);


  if (comline->journals || comline->jsize != MKFS_DEFAULT_JSIZE)
  {
    printf("Warning:  ignoring -j and -J command line options!\n");
    printf("  (Use -i if you want to ignore the Pool layout)\n\n");
  }

  comline->journals = 0;


  /*  Make sure all the subpools are aligned  */

  for (i = 0; i < splist->spl_count; i++)
  {
    switch (splist->spl_list[i].sps_type)
    {
    case SPTYPE_GFS_DATA:
      data_subpools++;
      break;
    case SPTYPE_GFS_JOURNAL:
      comline->journals++;
      break;
    default:
      die("subpool %u is of unknown type (%u)\n", i, splist->spl_list[i].sps_type);
      break;
    };
  }


  if (!comline->journals)
    die("No journals\n");

  if (!data_subpools)
    die("No data subpools\n");


  return 0;
}


/******************************************************************************
*******************************************************************************
**
** device_size()
**
** Description:
**   Linux
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

#if defined(__linux__)

static uint64 device_size(char *device)
{
  int fd, error;
  long size;


  fd = open(device, O_RDONLY);
  if (fd < 0)
    die("can't open %s\n", device);

  error = ioctl(fd, BLKGETSIZE, &size);
  if (error < 0)
    die("can't determine device size: %s\n", strerror(errno));

  close(fd);


  return size;
}

#endif  /*  defined(__linux__)  */


/******************************************************************************
*******************************************************************************
**
** device_size()
**
** Description:
**   FreeBSD
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

#if defined(__FreeBSD__)

static uint64 device_size(char *device)
{
  int fd, error;
  long size;
  static struct disklabel lab;	
  register struct partition *pp;
  char *cp;  
  char vflag = 0; /* need to add this later */


  fd = open(device, O_RDONLY);
  if (fd < 0)
    die("can't open %s\n", device);

  error = ioctl(fd, BLKGETSIZE, (char *)&lab);
  if (error < 0)
    die("can't determine device size: %s\n", strerror(errno));

  close(fd);


  cp = strchr(comline->device, '\0');
  if (cp == comline->device){
	fprintf(stderr,"null special file name");
	exit(-1);
  }
  cp--;
  if (!vflag && (*cp < 'a' || *cp > 'h') && !isdigit(*cp)){
	fprintf(stderr,"%s: can't figure out file system partition",
		  comline->device);
	exit(-1);
  }
  if (vflag)
	pp = &lab.d_partitions[0];
  else
	pp = &lab.d_partitions[*cp - 'a'];
  if (pp->p_size == 0)
	fprintf(stderr,"%s: `%c' partition is unavailable", comline->device, *cp);
  size = pp->p_size;


  return size;
}

#endif  /*  defined(__FreeBSD__)  */


/******************************************************************************
*******************************************************************************
**
** fake_pool_geometry()
**
** Description:
**   Talk to the underlying device and figure out how many blocks it has.
**   Stuff the value into a pool_splist_t structure with only on subpool.
**
** Input(s):
**   comline - the command line arguments (Specifies the pool name)
**
** Output(s):
**   splist - the structures containing the subpool geometry
**
*******************************************************************************
******************************************************************************/

void fake_pool_geometry(commandline_t *comline, pool_splist_t *splist)
{
  struct stat st;
  uint64 size=0, jsize=0;
  int i;


  if (stat(comline->device, &st) < 0)
    die("can't stat file %s\n", comline->device);

  
  if (S_ISREG(st.st_mode))
    size = st.st_size / GFS_BASIC_BLOCK;
  else if (S_ISCHR(st.st_mode) || S_ISBLK(st.st_mode))
    size = device_size(comline->device);
  else
    die("%s has strange file type\n", comline->device);


  if (comline->debug)
    printf("\nPartition size = %"FMT_64"u\n", size);


  if (!comline->journals)
    die("No journals specified (use -j)\n");

  if (!comline->jsize)
    die("journal size is zero (use -J)\n");

  jsize = comline->jsize * (1 << 20) / GFS_BASIC_BLOCK;

  if (comline->journals * jsize > size)
    die("Partition too small for number/size of journals\n");


  splist->spl_count = comline->journals + 1;

  type_zalloc(splist->spl_list, pool_spstatus_t, splist->spl_count);

  splist->spl_list[0].sps_start = 0;
  splist->spl_list[0].sps_len = size - comline->journals * jsize;
  splist->spl_list[0].sps_type = SPTYPE_GFS_DATA;

  for (i = 1; i <= comline->journals; i++)
  {
    splist->spl_list[i].sps_start = splist->spl_list[i - 1].sps_start + splist->spl_list[i - 1].sps_len;
    splist->spl_list[i].sps_len = jsize;
    splist->spl_list[i].sps_type = SPTYPE_GFS_JOURNAL;
  }
}


/******************************************************************************
*******************************************************************************
**
** fix_pool_geometry()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void fix_pool_geometry(commandline_t *comline, pool_splist_t *splist)
{
  int i, x;
  uint64 offset, len;
  uint32 bbsize = BTOBB(comline->bsize);


  if (comline->debug)
  {
    printf("\nSubPool Geometry:  (in basic blocks)\n");
    for (i = 0; i < splist->spl_count; i++)
    {
      printf("  Subpool #%d:\n", i);

      printf("    BStart = %"FMT_64"u, BLength = %"FMT_64"u\n",
	     splist->spl_list[i].sps_start, splist->spl_list[i].sps_len);

      printf("    type = ");
      for (x = 0; sptype_names[x].name[0]; x++)
	if (splist->spl_list[i].sps_type == sptype_names[x].label)
	  printf("%s\n", sptype_names[x].name);
    }
  }


  /*  Make sure all the subpools are aligned  */

  for (i = 0; i < splist->spl_count; i++)
  {
    offset = splist->spl_list[i].sps_start;
    len = splist->spl_list[i].sps_len;


    if (len < 100 * bbsize)
      die("subpool %d is way too small (%"FMT_64"u bytes)\n", i, len * GFS_BASIC_BLOCK); 


    if (offset % bbsize)
    {
      len -= bbsize - (offset % bbsize);
      offset += bbsize - (offset % bbsize);
    }


    splist->spl_list[i].sps_start = offset / bbsize;
    splist->spl_list[i].sps_len = len / bbsize;
  }


  if (comline->debug)
  {
    printf("\nSubPool Geometry:  (in FS blocks)\n");
    for (i = 0; i < splist->spl_count; i++)
    {
      printf("  Subpool #%d:\n", i);

      printf("    BStart = %"FMT_64"u, BLength = %"FMT_64"u\n",
	     splist->spl_list[i].sps_start, splist->spl_list[i].sps_len);

      printf("    type = ");
      for (x = 0; sptype_names[x].name[0]; x++)
	if (splist->spl_list[i].sps_type == sptype_names[x].label)
	  printf("%s\n", sptype_names[x].name);
    }

    printf("journals = %u\n", comline->journals);
  }
}


