/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-2000 Regents of the University of Minnesota
**    Portions Copyright 1999-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: structures.c
**
**    Description:
**	Functions that write out data to the disk
**
*******************************************************************************
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <assert.h>
#include <time.h>

#include "global.h"
#include "pool.h"
#include "gfs_ondisk.h"
#include "osi_endian.h"

#include "mkfs_gfs.h"


#define MKFS_ROOT_MODE	            (0755)


#define do_lseek(fd, off) \
{ \
  if (lseek((fd), (off), SEEK_SET) != (off)) \
    die("bad seek: %s on line %d of file %s\n", \
	strerror(errno),__LINE__, __FILE__); \
}

#define do_read(fd, buff, len) \
{ \
  if (read((fd), (buff), (len)) != (len)) \
    die("bad read: %s on line %d of file %s\n", \
	strerror(errno), __LINE__, __FILE__); \
}

#define do_write(fd, buff, len) \
{ \
  if (write((fd), (buff), (len)) != (len)) \
    die("bad write: %s on line %d of file %s\n", \
	strerror(errno), __LINE__, __FILE__); \
}


#define DIV_RU(x, y) (((x) + (y) - 1) / (y))





/******************************************************************************
*******************************************************************************
**
** rgrplength2bitblocks()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

unsigned int rgrplength2bitblocks(commandline_t *comline, unsigned int length)
{
  unsigned int bitbytes;
  unsigned int old_blocks = 0, blocks;
  int tries = 0;

  for (;;)
  {
    bitbytes = (length - old_blocks) / GFS_NBBY;
    blocks = 1;

    if (bitbytes > comline->bsize - sizeof(gfs_rgrp_t))
    {
      bitbytes -= comline->bsize - sizeof(gfs_rgrp_t);
      blocks += DIV_RU(bitbytes, (comline->bsize - sizeof(gfs_rgrpbits_t)));
    }

    if (blocks == old_blocks)
      break;

    old_blocks = blocks;

    if (tries++ > 10)
      die("confused computing bitblock length\n");
  }

  return blocks;
}


/******************************************************************************
*******************************************************************************
**
** write_sb()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void write_sb(commandline_t *comline, pool_splist_t *splist, rgrp_list_t *rlist)
{
  gfs_sb_t *sb;
  uint64 jindex_dinode;
  char buf[comline->bsize];
  int x;


  memset(buf, 0, comline->bsize);

  for (x = 0; x < comline->sb_addr; x++)
  {
    do_lseek(comline->fd, x * comline->bsize);
    do_write(comline->fd, buf, comline->bsize);
  }


  /*  Figure out the location of the journal index inode  */

  jindex_dinode = rlist->rg_offset + rgrplength2bitblocks(comline, rlist->rg_length);


  /*  Now, fill in the superblock  */

  type_zalloc(sb, gfs_sb_t, 1);
  comline->sb = sb;


  sb->sb_header.mh_magic = GFS_MAGIC;
  sb->sb_header.mh_type = GFS_METATYPE_SB;
  sb->sb_header.mh_format = GFS_FORMAT_SB;

  sb->sb_fs_format = GFS_FORMAT_FS;
  sb->sb_multihost_format = GFS_FORMAT_MULTI;

  sb->sb_bsize = comline->bsize;
  sb->sb_bsize_shift = ffs(comline->bsize) - 1;
  sb->sb_seg_size = comline->seg_size;

  sb->sb_jindex_di.no_formal_ino = jindex_dinode;
  sb->sb_jindex_di.no_addr = jindex_dinode;
  sb->sb_rindex_di.no_formal_ino = jindex_dinode + 1;
  sb->sb_rindex_di.no_addr = jindex_dinode + 1;
  sb->sb_root_di.no_formal_ino = jindex_dinode + 2;
  sb->sb_root_di.no_addr = jindex_dinode + 2;

  strcpy(sb->sb_lockproto, comline->lockproto);
  strcpy(sb->sb_locktable, comline->locktable);


  gfs_sb_out(sb, buf);


  do_lseek(comline->fd, comline->sb_addr * comline->bsize);
  do_write(comline->fd, buf, comline->bsize);


  if (comline->debug)
  {
    printf("\nSuperblock:\n");
    gfs_sb_print(sb, 0);
  }
}


/******************************************************************************
*******************************************************************************
**
** compute_height()
**
** Description:
**   Code ripped right from GFS' build_height()
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

static int compute_height(uint64 sz, uint32 bsize)
{
  uint32 diptrs = (bsize - sizeof(gfs_dinode_t)) / sizeof(uint64);
  uint32 inptrs = (bsize - sizeof(gfs_indirect_t)) / sizeof(uint64);

  int height;
  uint64 space, old_space;


  bsize -= sizeof(gfs_jdata_t);

  height = 1;
  space = diptrs * bsize;


  for (;;)
  {
    if (sz <= space)
      break;

    old_space = space;

    height++;
    space *= inptrs;

    if (space / inptrs != old_space ||
	space % inptrs != 0)
      break;
  }


  return height;
}


/******************************************************************************
*******************************************************************************
**
** build_tree()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

static void build_tree(commandline_t *comline, gfs_dinode_t *di, uint64 addr, uint32 blocks)
{
  gfs_indirect_t ind;
  char *buf;
  uint32 x, offset;
  uint32 height;

  uint32 indblocks;
  uint64 tmp_addr;
  uint32 tmp_blocks;

  
  di->di_height = compute_height(di->di_size, comline->bsize);


  if (di->di_height == 1)
  {
    type_zalloc(buf, char, comline->bsize);

    for (x = 0; x < blocks; x++)
      ((uint64 *)(buf + sizeof(gfs_dinode_t)))[x] = cpu_to_gfs64(addr + x);

    gfs_dinode_out(di, buf);

    do_lseek(comline->fd, di->di_num.no_addr * comline->bsize);
    do_write(comline->fd, buf, comline->bsize);

    free(buf);
  }
  else
  {
    tmp_addr = addr;
    tmp_blocks = blocks;


    for (height = di->di_height; height > 1; height--)
    {
      memset(&ind, 0, sizeof(gfs_indirect_t));
      ind.in_header.mh_magic = GFS_MAGIC;
      ind.in_header.mh_type = GFS_METATYPE_IN;
      ind.in_header.mh_format = GFS_FORMAT_IN;


      indblocks = DIV_RU((tmp_blocks * sizeof(uint64)),
			 (comline->bsize - sizeof(gfs_indirect_t)));

      type_zalloc(buf, char, indblocks * comline->bsize);


      offset = 0;
      for (x = 0; x < tmp_blocks; x++)
      {
	if (!(offset % comline->bsize))
	{
	  gfs_indirect_out(&ind, buf + offset);
	  offset += sizeof(gfs_indirect_t);
	}

	*((uint64 *)(buf + offset)) = cpu_to_gfs64(tmp_addr + x);
	offset += sizeof(uint64);
      }


      do_lseek(comline->fd, comline->rgrp0_next * comline->bsize);
      do_write(comline->fd, buf, indblocks * comline->bsize);

      free(buf);

      tmp_addr = comline->rgrp0_next;
      tmp_blocks = indblocks;

      di->di_blocks += indblocks;

      comline->rgrp0_next += indblocks;
    }


    type_zalloc(buf, char, comline->bsize);

    for (x = 0; x < tmp_blocks; x++)
      ((uint64 *)(buf + sizeof(gfs_dinode_t)))[x] = cpu_to_gfs64(tmp_addr + x);

    gfs_dinode_out(di, buf);

    do_lseek(comline->fd, di->di_num.no_addr * comline->bsize);
    do_write(comline->fd, buf, comline->bsize);    

    free(buf);
  }
}


/******************************************************************************
*******************************************************************************
**
** fill_jindex()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

static char *fill_jindex(commandline_t *comline, journal_list_t *jlist)
{
  journal_list_t *jtmp;
  gfs_jindex_t ji;
  char *buf;
  int j = 0;


  type_alloc(buf, char, comline->journals * sizeof(gfs_jindex_t));


  for (jtmp = jlist; jtmp; jtmp = jtmp->next)
  {
    memset(&ji, 0, sizeof(gfs_jindex_t));

    ji.ji_addr = jtmp->start;
    ji.ji_nsegment = jtmp->segments;

    gfs_jindex_out(&ji, buf + j * sizeof(gfs_jindex_t));

    j++;
  }


  if (comline->debug)
  {
    printf("\nJournal Index data:\n");

    for (j = 0; j < comline->journals; j++)
    {
      gfs_jindex_in(&ji, buf + j * sizeof(gfs_jindex_t));

      printf("\n  Journal %d\n", j);
      gfs_jindex_print(&ji, 0);
    }
  }


  return buf;
}


/******************************************************************************
*******************************************************************************
**
** write_jindex()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void write_jindex(commandline_t *comline, journal_list_t *jlist)
{
  gfs_dinode_t di;
  gfs_jdata_t jd;
  char *buf, *data;
  uint64 addr;
  uint32 blocks;
  unsigned int x, left, jbsize = comline->bsize - sizeof(gfs_jdata_t);


  memset(&di, 0, sizeof(gfs_dinode_t));
 
  di.di_header.mh_magic = GFS_MAGIC;
  di.di_header.mh_type = GFS_METATYPE_DI;
  di.di_header.mh_format = GFS_FORMAT_DI;

  di.di_num = comline->sb->sb_jindex_di;

  di.di_mode = MKFS_ROOT_MODE;
  di.di_nlink = 1;
  di.di_size = comline->journals * sizeof(gfs_jindex_t);
  di.di_blocks = 1;
  di.di_atime = di.di_mtime = di.di_ctime = time(NULL);

  di.di_flags = GFS_DIF_JDATA;
  di.di_payload_format = GFS_FORMAT_JI;
  di.di_type = FILE_REG;


  data = fill_jindex(comline, jlist);


  if (di.di_size < comline->bsize - sizeof(gfs_dinode_t))
  {
    type_zalloc(buf, char, comline->bsize);

    gfs_dinode_out(&di, buf);

    memcpy(buf + sizeof(gfs_dinode_t), data, comline->journals * sizeof(gfs_jindex_t));

    do_lseek(comline->fd, di.di_num.no_addr * comline->bsize);
    do_write(comline->fd, buf, comline->bsize);

    free(buf);
  }
  else
  {
    blocks = DIV_RU(di.di_size, (comline->bsize - sizeof(gfs_jdata_t)));
    di.di_blocks += blocks;

    addr = comline->rgrp0_next;


    memset(&jd, 0, sizeof(gfs_jdata_t));
    jd.mh_magic = GFS_MAGIC;
    jd.mh_type = GFS_METATYPE_JD;
    jd.mh_format = GFS_FORMAT_JD;


    type_zalloc(buf, char, blocks * comline->bsize);

    left = comline->journals * sizeof(gfs_jindex_t);
    for (x = 0; x < blocks; x++)
    {
      gfs_meta_header_out(&jd, buf + x * comline->bsize);
      memcpy(buf + x * comline->bsize + sizeof(gfs_jdata_t),
	     data + x * jbsize,
	     (left > jbsize) ? jbsize : left);
      left -= jbsize;
    }


    do_lseek(comline->fd, addr * comline->bsize);
    do_write(comline->fd, buf, blocks * comline->bsize);

    free(buf);


    comline->rgrp0_next += blocks;

    build_tree(comline, &di, addr, blocks);
  }


  free(data);


  if (comline->debug)
  {
    printf("\nJournal index dinode:\n");
    gfs_dinode_print(&di, 0);
  }
}


/******************************************************************************
*******************************************************************************
**
** fill_rindex()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

static char *fill_rindex(commandline_t *comline, rgrp_list_t *rlist)
{
  gfs_rindex_t *ri, rindex;
  rgrp_list_t *rtmp;
  char *buf;
  int r = 0;
  uint32 length, blocks;


  type_alloc(buf, char, comline->rgrps * sizeof(gfs_rindex_t));


  for (rtmp = rlist; rtmp; rtmp = rtmp->next)
  {
    length = rgrplength2bitblocks(comline, rtmp->rg_length);

    blocks = rtmp->rg_length - length;
    blocks -= blocks % GFS_NBBY;
    comline->fssize += blocks;

    type_zalloc(ri, gfs_rindex_t, 1);
    rtmp->ri = ri;

    ri->ri_addr = rtmp->rg_offset;
    ri->ri_length = length;

    ri->ri_data1 = rtmp->rg_offset + length;
    ri->ri_data = blocks;

    ri->ri_bitbytes = blocks / GFS_NBBY;

    gfs_rindex_out(ri, buf + r * sizeof(gfs_rindex_t));

    r++;
  }


  if (comline->debug)
  {
    printf("\nResource Index data:\n");

    for (r = 0; r < comline->rgrps; r++)
    {
      gfs_rindex_in(&rindex, buf + r * sizeof(gfs_rindex_t));

      printf("\n  Resource index %d\n", r);
      gfs_rindex_print(&rindex, 0);
    }
  }


  return buf;
}


/******************************************************************************
*******************************************************************************
**
** write_rindex()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void write_rindex(commandline_t *comline, rgrp_list_t *rlist)
{
  gfs_dinode_t di;
  gfs_jdata_t jd;
  char *buf, *data;
  uint64 addr;
  uint32 blocks;
  unsigned int x, left, jbsize = comline->bsize - sizeof(gfs_jdata_t);


  memset(&di, 0, sizeof(gfs_dinode_t));

  di.di_header.mh_magic = GFS_MAGIC;
  di.di_header.mh_type = GFS_METATYPE_DI;
  di.di_header.mh_format = GFS_FORMAT_DI;

  di.di_num = comline->sb->sb_rindex_di;

  di.di_mode = MKFS_ROOT_MODE;
  di.di_nlink = 1;
  di.di_size = comline->rgrps * sizeof(gfs_rindex_t);
  di.di_blocks = 1;
  di.di_atime = di.di_mtime = di.di_ctime = time(NULL);

  di.di_flags = GFS_DIF_JDATA;
  di.di_payload_format = GFS_FORMAT_RI;
  di.di_type = FILE_REG;


  data = fill_rindex(comline, rlist);


  if (di.di_size < comline->bsize - sizeof(gfs_dinode_t))
  {
    type_zalloc(buf, char, comline->bsize);

    gfs_dinode_out(&di, buf);
    
    memcpy(buf + sizeof(gfs_dinode_t), data, comline->rgrps * sizeof(gfs_rindex_t));

    do_lseek(comline->fd, di.di_num.no_addr * comline->bsize);
    do_write(comline->fd, buf, comline->bsize);

    free(buf);
  }
  else
  {
    blocks = DIV_RU(di.di_size, (comline->bsize - sizeof(gfs_jdata_t)));
    di.di_blocks += blocks;

    addr = comline->rgrp0_next;


    memset(&jd, 0, sizeof(gfs_jdata_t));
    jd.mh_magic = GFS_MAGIC;
    jd.mh_type = GFS_METATYPE_JD;
    jd.mh_format = GFS_FORMAT_JD;


    type_zalloc(buf, char, blocks * comline->bsize);

    left = comline->rgrps * sizeof(gfs_rindex_t);
    for (x = 0; x < blocks; x++)
    {
      gfs_meta_header_out(&jd, buf + x * comline->bsize);
      memcpy(buf + x * comline->bsize + sizeof(gfs_jdata_t),
	     data + x * jbsize,
	     (left > jbsize) ? jbsize : left);
      left -= jbsize;
    }


    do_lseek(comline->fd, addr * comline->bsize);
    do_write(comline->fd, buf, blocks * comline->bsize);

    free(buf);


    comline->rgrp0_next += blocks;

    build_tree(comline, &di, addr, blocks);
  }


  free(data);


  if (comline->debug)
  {
    printf("\nResource index dinode:\n");
    gfs_dinode_print(&di, 0);
  }
}


/******************************************************************************
*******************************************************************************
**
** write_root()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void write_root(commandline_t *comline)
{
  gfs_dinode_t di;
  gfs_dirent_t de;

  char buf[comline->bsize];


  memset(&di, 0, sizeof(gfs_dinode_t));
  memset(buf, 0, comline->bsize);


  di.di_header.mh_magic = GFS_MAGIC;
  di.di_header.mh_type = GFS_METATYPE_DI;
  di.di_header.mh_format = GFS_FORMAT_DI;

  di.di_num = comline->sb->sb_root_di;

  di.di_mode = MKFS_ROOT_MODE;
  di.di_nlink = 2;
  di.di_size = comline->bsize - sizeof(gfs_dinode_t);
  di.di_blocks = 1;
  di.di_atime = di.di_mtime = di.di_ctime = time(NULL);

  di.di_flags = GFS_DIF_JDATA;
  di.di_payload_format = GFS_FORMAT_DE;
  di.di_type = FILE_DIR;

  di.di_entries = 2;

  gfs_dinode_out(&di, buf);


  /*  Fill in .  */

  memset(&de, 0, sizeof(gfs_dirent_t));

  de.de_inum = comline->sb->sb_root_di;
  de.de_hash = 0x0ED4E242;
  de.de_rec_len = GFS_DIR_REC_LEN(1);
  de.de_name_len = 1;
  de.de_type = FILE_DIR;

  gfs_dirent_out(&de, buf + sizeof(gfs_dinode_t));
  memcpy(buf + sizeof(gfs_dinode_t) + sizeof(gfs_dirent_t), ".", 1);

  if (comline->debug)
  {
    printf("\nRoot Dinode dirent:\n");
    gfs_dirent_print(&de, ".", TRUE);
  }


  /*  Fill in ..  */

  memset(&de, 0, sizeof(gfs_dirent_t));

  de.de_inum = comline->sb->sb_root_di;
  de.de_hash = 0x9608161C;
  de.de_rec_len = comline->bsize - GFS_DIR_REC_LEN(1) - sizeof(gfs_dinode_t);
  de.de_name_len = 2;
  de.de_type = FILE_DIR;

  gfs_dirent_out(&de, buf + sizeof(gfs_dinode_t) + GFS_DIR_REC_LEN(1));
  memcpy(buf + sizeof(gfs_dinode_t) + GFS_DIR_REC_LEN(1) + sizeof(gfs_dirent_t), "..", 2);

  if (comline->debug)
  {
    printf("\nRoot Dinode dirent:\n");
    gfs_dirent_print(&de, "..", TRUE);
  }


  do_lseek(comline->fd, di.di_num.no_addr * comline->bsize);
  do_write(comline->fd, buf, comline->bsize);


  if (comline->debug)
  {
    printf("\nRoot dinode:\n");
    gfs_dinode_print(&di, 0);
  }
}


/******************************************************************************
*******************************************************************************
**
** write_rgrps()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void write_rgrps(commandline_t *comline, rgrp_list_t *rlist)
{
  gfs_rgrp_t rg;
  gfs_rgrpbits_t rb;
  rgrp_list_t *rtmp;

  char *buf, *data;
  unsigned int x, offset;
  unsigned int byte, bit;
  unsigned int blk_count;
  unsigned int r = 0;


  memset(&rb, 0, sizeof(gfs_rgrpbits_t));
  rb.mh_magic = GFS_MAGIC;
  rb.mh_type = GFS_METATYPE_RB;
  rb.mh_format = GFS_FORMAT_RB;


  for (rtmp = rlist; rtmp; rtmp = rtmp->next)
  {
    memset(&rg, 0, sizeof(gfs_rgrp_t));
    rg.rg_header.mh_magic = GFS_MAGIC;
    rg.rg_header.mh_type = GFS_METATYPE_RG;
    rg.rg_header.mh_format = GFS_FORMAT_RG;
    rg.rg_free = rtmp->ri->ri_data;

    type_zalloc(data, char, rtmp->ri->ri_bitbytes);
    type_zalloc(buf, char, rtmp->ri->ri_length * comline->bsize);


    /*  Deal with the special case of the first dinode  */

    if (!r)
    {
      if (comline->rgrp0_next > rtmp->ri->ri_data1 + rtmp->ri->ri_data)
	die("Send email to kpreslan@sistina.com and tell him he's an idiot\n");

      /*  Compensate for the hidden dinodes  */

      *data = (GFS_BLKST_USEDMETA << 2 * GFS_BIT_SIZE) |
	(GFS_BLKST_USEDMETA << GFS_BIT_SIZE) |
	  GFS_BLKST_USEDMETA;
      rg.rg_free -= 3;
      rg.rg_useddi = 3;

      /*  Compensate for the data the hidden dinodes point to  */

      for (x = rtmp->ri->ri_data1 + 3; x < comline->rgrp0_next; x++)
      {
	byte = (x - rtmp->ri->ri_data1) / GFS_NBBY;
	bit = (x - rtmp->ri->ri_data1) % GFS_NBBY;

	data[byte] |= GFS_BLKST_USEDMETA << (bit * GFS_BIT_SIZE);

	rg.rg_free--;
	rg.rg_usedmeta++;
      }
    }


    gfs_rgrp_out(&rg, buf);
    offset = sizeof(gfs_rgrp_t);
    blk_count = 1;

    for (x = 0; x < rtmp->ri->ri_bitbytes; x++)
    {
      if (!(offset % comline->bsize))
      {
	gfs_meta_header_out(&rb, buf + offset);
	offset += sizeof(gfs_rgrpbits_t);
	blk_count++;
      }

      buf[offset] = data[x];
      offset++;
    }

    if (blk_count != rtmp->ri->ri_length)
      die("RG underflow (rg = %u, blk_count = %u, rtmp->ri->ri_length = %u)\n",
	  r, blk_count, rtmp->ri->ri_length);


    do_lseek(comline->fd, rtmp->ri->ri_addr * comline->bsize);
    do_write(comline->fd, buf, rtmp->ri->ri_length * comline->bsize);


    free(buf);
    free(data);


    if (comline->debug)
    {
      printf("\nResource group header %d\n", r);
      gfs_rgrp_print(&rg, 0);
    }

    r++;
  }
}


/******************************************************************************
*******************************************************************************
**
** write_journals()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void write_journals(commandline_t *comline, journal_list_t *jlist)
{
  gfs_log_header_t lh;
  journal_list_t *jtmp;

  char buf[comline->bsize];
  uint32 seg, sequence;
  int x = 0;


  srandom(time(NULL));


  for (jtmp = jlist; jtmp; jtmp = jtmp->next)
  {
    if (comline->debug)
      printf("Starting journal %d\n", x++);

    sequence = jtmp->segments / (RAND_MAX + 1.0) * random();

    for (seg = 0; seg < jtmp->segments; seg++)
    {
      memset(buf, 0, comline->bsize);
      memset(&lh, 0, sizeof(gfs_log_header_t));

      lh.lh_header.mh_magic = GFS_MAGIC;
      lh.lh_header.mh_type = GFS_METATYPE_LH;
      lh.lh_header.mh_format = GFS_FORMAT_LH;
      lh.lh_flags = GFS_LOG_HEAD_UNMOUNT;
      lh.lh_first = jtmp->start + seg * comline->seg_size;
      lh.lh_sequence = sequence;
      /*  Don't care about tail  */
      /*  Don't care about nopen dump  */

      gfs_log_header_out(&lh, buf);
      gfs_log_header_out(&lh, buf + GFS_BASIC_BLOCK - sizeof(gfs_log_header_t));

      do_lseek(comline->fd, lh.lh_first * comline->bsize);
      do_write(comline->fd, buf, comline->bsize);

      if (++sequence == jtmp->segments)
	sequence = 0;

      if (!(seg % 100))
	if (comline->debug)
	  printf("seg #%u\n", seg);
	
    }
  }
}


