/******************************************************************************
*******************************************************************************
**
**    Copyright 2001 Sistina Software Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Jonathan Brassow <brassow@sistina.com>
** 
**    Module name: common.c
**
**    Description:
**	This file, along with common.h, contain all of the functions and
**	structures that can be shared between the Pool related tools.  This
**	is done in an effort to reduce the amount of code and make it easier
**	to read.
**
*******************************************************************************
******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mount.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>

#include "common.h"
#include "pool_ondisk.h"  /* needed for pool label structures */
#define WANT_SPTYPE_NAMES
#include "pool_sptypes.h"
#include "pool_ioctl.h"

/**************************************************
 **  Structures only needed by this file.
 *************************************************/
typedef struct label_list_s{
  pool_label_t		ll_label;
  dev_t			ll_dev;
  struct label_list_s 	*next;
} label_list_t;


/**************************************************
 **  Functions only needed by this file.
 *************************************************/

static int pool_loaded(char *poolname, int commfd);
static int pool_in_use(char *poolname, int commfd);
static int is_blkdev(const char *path);
static int ext2sb_check(char *path);
static int clean_pool_dir(int commfd);
static int labels_exist_with_same_name(pool_t *pool);
static int labels_exist_on_devices(pool_t *pool);
static int seek_and_destroy_labels(const char *pool_name);
static int erase_conflicting_labels(pool_t *pool);
static int stripe_devs_same_size(pool_t *pool);
static label_list_t *sort_label_list(label_list_t *head);
static label_list_t *find_and_sort_labels(void);


/*
 * pool_loaded
 * @poolname
 * @commfd - the pool comm device file descriptor
 *
 * This function depends on the comm device being open.
 * It returns whether or not a pool with the given name
 * exists in the kernel.
 *
 * Returns: TRUE or FLASE
 */
static int pool_loaded(char *poolname, int commfd){
  int		i;
  int		size = 0;
  pool_list_t	list;

  if (ioctl(commfd, POOL_COUNT, &size) < 0) {
    fprintf(stderr, "pool communications error.\n");
    return TRUE;
  }

  list.pl_count = size;
  list.pl_list = (pool_status_t *)malloc(sizeof(pool_status_t) * size);

  if (ioctl(commfd, POOL_LIST, &list) < 0) {
    fprintf(stderr, "pool communications error.\n");
    return TRUE;
  }

  for(i = 0; i < list.pl_count; i++){
    if(!strcmp(list.pl_list[i].ps_name, poolname) ||
       !strcmp(poolname, "all")){
      return TRUE;  /* Pool is loaded */
    }
  }

  return FALSE;  /* Not found, therefore, not loaded */
}


/*
 * pool_in_use
 * @poolname
 * @commfd - the pool comm device file descriptor
 *
 * This function depends on the comm device being open.
 * It returns whether or not the pool is in use.
 *
 * Returns: TRUE or FLASE
 */
static int pool_in_use(char *poolname, int commfd){
  int		i;
  int		size = 0;
  pool_list_t	list;

  if (ioctl(commfd, POOL_COUNT, &size) < 0) {
    fprintf(stderr, "pool communications error.\n");
    return TRUE;
  }

  list.pl_count = size;
  list.pl_list = (pool_status_t *)malloc(sizeof(pool_status_t) * size);

  if (ioctl(commfd, POOL_LIST, &list) < 0) {
    fprintf(stderr, "pool communications error.\n");
    return TRUE;
  }

  for(i = 0; i < list.pl_count; i++){
    if(!strcmp(list.pl_list[i].ps_name, poolname) ||
       !strcmp(poolname, "all")){
      return list.pl_list[i].ps_in_use;
    }
  }
  return FALSE;  /* Not found */
}


/*
 * is_blkdev - is this a blk device
 * @path - the path (e.g. /dev/sda)
 *
 * Returns: 1 if TRUE, 0 if FALSE
 */
static int is_blkdev(const char *path){
  struct stat dev_stats;

  if ((stat(path, &dev_stats)) < 0) {
    return(0); /* Can't tell if it is blk dev - assume no */
  }
  else if (!S_ISBLK(dev_stats.st_mode)) {
    return(0); /* No, it isn't a blk device */
  }

  return 1;  /* Yes, it is a blk device */
}


/* ext2sb_check - check for an ext2 FS super block
 * @path: the path to the device to run the check on
 *
 * Returns: 1 sb found, 0 no sb found, -1 on error
 */
#define EXT2_SUPER_MAGIC 0x53EF
#define EXT2_FS_BLKSIZE 1024 /* attention -- this is not always true ! */
static int ext2sb_check(char *path){
  uint16	ext2_magic = 0;
  uint16	*ext2_magic_ptr = NULL;
  int		fd;
  char		buffer[EXT2_FS_BLKSIZE];

  if((fd = open(path, O_RDONLY)) < 0){
    fprintf(stderr, "Unable to open device \"%s\".\n", path);
    return -1;
  }

  if(lseek(fd, EXT2_FS_BLKSIZE, SEEK_SET) < 0){
    fprintf(stderr, "Unable to perform lseek on %s.\n", path);
    return -1;
  }

  switch(read(fd, buffer, EXT2_FS_BLKSIZE)){
  case -1:
    fprintf(stderr, "Unable to read %s.\n", path);
    return -1;
    break;
  case EXT2_FS_BLKSIZE:
    break;
  default: /* device is probably smaller than 2kB */
    return 0;
    break;
  }

  ext2_magic_ptr = (uint16 *)&(buffer[56]);

  ext2_magic = (((*ext2_magic_ptr & 0x00ff) << 8) |
		((*ext2_magic_ptr & 0xff00) >> 8));

  if(EXT2_SUPER_MAGIC == ext2_magic){
    return 1;
  }
  return 0;
}


/*
 * clean_pool_dir - get rid of unused pool device nodes
 * @commfd - pool comm file descriptor
 *
 * Returns: 0 on success, -1 on error
 */
static int clean_pool_dir(int commfd){
  int			i;
  int			first = 1;
  char			path[LINE_MAX];
  DIR			*dir;
  struct dirent		*file;
  struct stat		statbuf;
  pool_list_t		*pl = NULL;

  if((dir = opendir("/dev/pool/")) == NULL){
    return -1;
  }

  rewinddir(dir);
  while ((file = readdir(dir)) != NULL){
    if(!strcmp(".", file->d_name) || !strcmp("..", file->d_name)){
      continue;
    }
    strcpy(path, "/dev/pool/");
    strcat(path, file->d_name);
    if(first){
      pl = get_pool_info(commfd);
      first = 0;
    }

    if(pl && (stat(path, &statbuf) >= 0)){
      for(i = 0; i < pl->pl_count; i++){
	if(pl->pl_list[i].ps_pool_minor == (uint32)MINOR(statbuf.st_rdev)){
	  break;
	}
      }
      if(i != pl->pl_count){ /* pool loaded, skip unlink */
	continue;
      }
    }
    unlink(path);
  }
  closedir(dir);

  return 0;
}


/*
 * labels_exist_with_same_name - check if name is already used
 * @pool: the pool whose name to check for
 *
 * This function will scan all block devices for labels that have
 * a name that is the same as the name of the pool passed as a
 * parameter
 *
 * Returns: TRUE if others are found, FALSE if no, and -1 on error
 */
static int labels_exist_with_same_name(pool_t *pool){
  int			rtn = 0;
  label_list_t		*ll, *tmp;

  ll=find_and_sort_labels();

  for(tmp = ll; tmp; tmp = tmp->next){
    if(!strcmp(pool->p_name, tmp->ll_label.pl_pool_name)){
      rtn = 1;
      break;
    }
  }
  for(; ll;){
    tmp = ll;
    ll = ll->next;
    free(tmp);
  }
  return rtn;
}


/*
 * labels_exist_on_devices - check devices for labels
 * @pool - the pool whose devices will be checked
 *
 * This function takes a pool_t structure as an argument
 * and tells whether the devices specified in the struct
 * have labels written on them.
 *
 * Returns 1 if 1+ labels found, 0 if no labels, -1 if error
 */
static int labels_exist_on_devices(pool_t *pool){
  int			i, j, devfd;
  subpool_t		*sp;
  generic_device_t	*gd;
  char			buf[sizeof(pool_label_t)];
  pool_label_t		label;

  memset(buf, 0, sizeof(pool_label_t));

  for(i=0; i < pool->p_num_spools; i++){
    sp = &(pool->p_spools[i]);
    for(j=0; j < sp->sp_devs; j++){
      gd = &(sp->sp_gdlist[j]);
      if((devfd = open(gd->gd_dev_path, O_RDONLY)) < 0){
	fprintf(stderr, "Failed to open %s.\n", gd->gd_dev_path);
	return -1;
      }
      if(read(devfd, buf, sizeof(pool_label_t)) < sizeof(pool_label_t)){
	fprintf(stderr, "Could not read label from %s\n", gd->gd_dev_path);
	close(devfd);
	return -1;
      }

      pool_label_in(&label, buf);
      if(label.pl_magic == POOL_MAGIC){
	return 1;
      }
    }
  }
  return 0;
}


/*
 * seek_and_destroy_labels - destroy all pool labels with given name
 * @pool_name - name of pool to destroy
 *
 * Returns: 0 on Ok, -1 on error
 */
static int seek_and_destroy_labels(const char *pool_name){
  FILE		*fp;
  int		line_num = 0;
  char		a[50],b[50],c[50],name[LINE_MAX];
  char		tmp_str[LINE_MAX];
  char		proc_line[LINE_MAX];

  if((fp = fopen("/proc/partitions", "r")) == NULL){
    fprintf(stderr, "Could not open /proc/partitions\n");
    return -1;
  }
      
  while(fgets(proc_line, LINE_MAX, fp) != NULL){
    int			fd;
    char		buf[sizeof(pool_label_t)];
    pool_label_t	label;

    if(line_num++ <= 2) /* skip first two lines */
      continue;

    sscanf(proc_line,"%s %s %s %s", a, b, c, name);
    
    sprintf(tmp_str,"/dev/%s",name);
    if((fd = open(tmp_str, O_RDWR)) < 0){
      fprintf(stderr, "Could not open %s\n", tmp_str);
      return -1;
    }
    if (read(fd, buf, sizeof(pool_label_t)) < sizeof(pool_label_t)){
      fprintf(stderr, "Could not read %s\n", tmp_str);
      close(fd);
      return -1;
    }
    pool_label_in(&label, buf);
    if(!strcmp(pool_name, label.pl_pool_name)){
      memset(buf, 0, sizeof(pool_label_t));
      lseek(fd,0,SEEK_SET);
      if(write(fd, buf, sizeof(pool_label_t)) < sizeof(pool_label_t)){
	fprintf(stderr, "Could not write to %s\n", tmp_str);
	close(fd);
	return -1;
      }
    }
    fsync(fd);
    close(fd);
  }
  fclose(fp);
  return 0;
}


/*
 * erase_conflicting_labels
 * @pool - ptr to the pool_t structure
 *
 * This function checks for labels on the devices mentioned in the
 * pool_t structure.  If labels are found, it will erase them AND
 * any other labels for that same pool whether they are on the
 * devices mentioned in the pool_t or not.  (It will seek out and
 * erase all the labels for a pool if any member is found to
 * conflict.
 *
 * Returns: 0 on Ok, -1 on error
 */
static int erase_conflicting_labels(pool_t *pool){
  int			i, j, devfd;
  subpool_t		*sp;
  generic_device_t	*gd;
  char			buf[sizeof(pool_label_t)];
  pool_label_t		label;

  memset(buf, 0, sizeof(pool_label_t));

  for(i=0; i < pool->p_num_spools; i++){
    sp = &(pool->p_spools[i]);
    for(j=0; j < sp->sp_devs; j++){
      gd = &(sp->sp_gdlist[j]);
      if((devfd = open(gd->gd_dev_path, O_RDONLY)) < 0){
	fprintf(stderr, "Failed to open %s.\n", gd->gd_dev_path);
	return -1;
      }
      if(read(devfd, buf, sizeof(pool_label_t)) < sizeof(pool_label_t)){
	fprintf(stderr, "Could not read label from %s\n", gd->gd_dev_path);
	close(devfd);
	return -1;
      }

      pool_label_in(&label, buf);
      if(label.pl_magic == POOL_MAGIC){
	if(seek_and_destroy_labels(label.pl_pool_name) < 0){
	  return -1;
	}
      }
    }
  }
  return 0;
}

/*
 * stripe_devs_same_size - checks if striped devs are the same size
 * @pool - pool_t structure
 *
 * This function looks in the pool_t structure for striped groups of
 * block devices.  It checks to make sure those devices are the same
 * size.  If not, it changes them to all be the same size (the smallest)
 * and returns 0.  Otherwise, 1 (TRUE) is returned.  The structure
 * is guaranteed to leave this function with equal size stripe devs.
 *
 * Returns: TRUE if devs are same, FALSE if altered
 */
static int stripe_devs_same_size(pool_t *pool){
  int			i,j;
  int			flag=0;
  int			min_dev_size = -1;
  subpool_t		*sp;
  generic_device_t	*gd;

  for (i=0; i < pool->p_num_spools; i++) {
    sp = &(pool->p_spools[i]);
    if(sp->sp_stripe){
      if(sp->sp_devs <= 0){
	continue;
      }
      for(j = 0; j < sp->sp_devs; j++){
	gd = &(sp->sp_gdlist[j]);
	if(min_dev_size == -1){
	  min_dev_size = gd->gd_nblks;
	}
	if((gd->gd_nblks != 0) && (min_dev_size != gd->gd_nblks)){
	  min_dev_size = (min_dev_size < gd->gd_nblks) ?
	    min_dev_size : gd->gd_nblks;
	  flag = 1;
	}
      }
      for(j = 0; j < sp->sp_devs; j++){
	gd = &(sp->sp_gdlist[j]);
	gd->gd_nblks = min_dev_size;
      }
    }
  }
  if(flag){
    return 0;
  } else {
    return 1;
  }
}


/*
 * sort_label_list - sort according to label pool name
 * @head: the beginning of the label list
 *
 * This function sorts the label list according to the
 * pool name found in the pool label.  It's not the fastest
 * sort, but it's plenty fast enough for what it does.
 *
 * Returns: head of the sorted list
 */
static label_list_t *sort_label_list(label_list_t *head){
  label_list_t		*sorted = NULL;
  label_list_t		*smallest;
  label_list_t		*tmp;
  
  while(head){
    /* find smallest */
    for(tmp = head, smallest = head; tmp; tmp = tmp->next){
      if(strcmp(smallest->ll_label.pl_pool_name,
		tmp->ll_label.pl_pool_name) < 0){
	smallest = tmp;
      }
    }
    /* remove smallest */
    tmp = head;
    if(tmp == smallest){
      head = tmp->next;
    } else {
      for(; tmp->next != smallest; tmp = tmp->next);
      tmp->next = smallest->next;
    }

    /* insert smallest into sorted list */
    smallest->next = sorted;
    sorted = smallest;
  }
  return sorted; 
}


/*
 * find_and_sort_labels - like the function implies
 *
 * This function uses /proc/partitions to go through all
 * the known block devices and search for labels.  Once
 * a list of labels is compiled, the list is sorted.
 * This groups all labels for the same pool together,
 * making it easier to assemble them.
 *
 * Returns: ptr to head of sorted list, else NULL
 */
static label_list_t *find_and_sort_labels(void){
  FILE		*fp;
  char		proc_line[LINE_MAX];
  int		line_num = 0;
  char		maj[20], min[20], blocks[20], name[LINE_MAX];
  char		tmp_str[LINE_MAX];
  char		buffer[sizeof(pool_label_t)];
  pool_label_t	label;
  label_list_t	*head = NULL;
  label_list_t	*tmp = NULL;

  if((fp = fopen("/proc/partitions", "r")) == NULL){
    fprintf(stderr, "Could not open /proc/partitions\n");
    return NULL;
  }
      
  while(fgets(proc_line, LINE_MAX, fp) != NULL){
    int		fd;
    if(line_num++ < 2) /* skip first two lines */
      continue;
    
    sscanf(proc_line,"%s %s %s %s", maj, min, blocks,name);
    sprintf(tmp_str,"/dev/%s",name);

    if((fd = open(tmp_str, O_RDONLY)) < 0){
      fprintf(stderr, "Could not open %s\n", tmp_str);
      continue;
    }

    if(read(fd, buffer, sizeof(pool_label_t)) < sizeof(pool_label_t)){
      fprintf(stderr, "Could not read %s\n", tmp_str);
      close (fd);
      fclose(fp);
      return NULL;
    }

    pool_label_in(&label, buffer);
    close(fd);
    if(label.pl_magic == POOL_MAGIC){
      if((tmp = (label_list_t *)malloc(sizeof(label_list_t))) == NULL){
	fprintf(stderr, "Not enough memory.\n");
	while(head){
	  tmp = head;
	  head = head->next;
	  free(tmp);
	}
	return NULL;
      }
      memcpy(&(tmp->ll_label), &label, sizeof(pool_label_t));
      {
	struct stat statbuf;
	if((stat(tmp_str, &statbuf)) < 0){
	  return NULL;
	}
	tmp->ll_dev = statbuf.st_rdev;
      }
      tmp->next = head;
      head = tmp;
    }
  }
  fclose(fp);

  /* Now we've got our list, so let's sort it */

  head = sort_label_list(head);

  return head;
}


/**************************************************
 **  Functions needed globally (and found in .h).
 *************************************************/


/*
 * devpath_to_devt - convert a device path to it's dev_t
 * @path: the path to convert
 *
 * Returns: dev_t on success, 0 otherwise
 */
dev_t devpath_to_devt(char *path){
  struct stat		stats;

  if(stat(path, &stats)){
    return 0;
  }

  return stats.st_rdev;
}


/*
 * devt_to_devpath - convert dev_t to it's device path
 * @device: the dev_t to convert
 *
 * Returns: path as a string, NULL if not found
 */
char *devt_to_devpath(dev_t device){
  DIR			*dir;
  struct dirent 	*file;
  struct stat		stats;
  char			*path = NULL;
  int			d_namlen;
  unsigned int		major = major(device);
  unsigned int		minor = minor(device);

  dir = opendir("/dev/");
  rewinddir(dir);

  /* check the dev directory */
  while ((file = readdir(dir)) != NULL) {

    d_namlen = strlen(file->d_name);

    path = (char *)malloc(strlen("/dev/") + d_namlen + 1);

    if (path == NULL) {
      fprintf(stderr, "unable to allocate memory\n");
      goto fail;
    }

    strcpy(path, "/dev/");
    strcat(path, file->d_name);
      
    stat(path, &stats);
    if (S_ISDIR(stats.st_mode) || S_ISCHR(stats.st_mode)) {
      free(path);
      continue;
    }
    if (major(stats.st_rdev) == major && minor(stats.st_rdev) == minor) {
      closedir(dir);
      return path;
    }
    free(path);
    continue;
  }

  /* check the gnbd directory */
  closedir(dir);
  dir = opendir("/dev/");
  rewinddir(dir);
  while ((file = readdir(dir)) != NULL) {
    d_namlen = strlen(file->d_name);

    path = (char *)malloc(strlen("/dev/gnbd/") + d_namlen + 1);

    if (path == NULL) {
      fprintf(stderr, "unable to allocate memory\n");
      goto fail;
    }

    strcpy(path, "/dev/gnbd/");
    strcat(path, file->d_name);
      
    stat(path, &stats);
    if (S_ISDIR(stats.st_mode) || S_ISCHR(stats.st_mode)) {
      free(path);
      continue;
    }
    if (major(stats.st_rdev) == major && minor(stats.st_rdev) == minor) {
      closedir(dir);
      return path;
    }
    free(path);
    continue;
  }
  
 fail:
  closedir(dir);
  return NULL;
}  

/*
 * get_sptype_name - convert type number to string equivalent
 * @type - type number (defined in pool_sptypes.h)
 *
 * Returns: type as a string, NULL if not found
 */
char *get_sptype_name(uint32 type){
  int 	x;
  char	buffer[LINE_MAX];
  char	*type_name = NULL;

  for (x = 0; sptype_names[x].name[0]; x++) {
    if (type == sptype_names[x].label){
      sprintf(buffer, "%s", sptype_names[x].name);
      type_name = (char *)malloc(strlen(buffer));
      strcpy(type_name, buffer);
    }
  }
  return type_name;
}


/*
 * print_pool_t - print pool information from pool_t
 * @pool - ptr to pool we want info on
 *
 * Returns:
 */
void print_pool_t(pool_t *pool){
  uint			i, j;
  subpool_t		*sp;
  generic_device_t	*gd;
  
  printf("\n"
	 "**********************************************************\n");
  printf("* Pool information:                                      *\n");
  printf("* Pool name           :      %20s        *\n", pool->p_name);
  printf("* Number of subpools  :      %20u        *\n", pool->p_num_spools);

  for (i=0; i < pool->p_num_spools; i++) {
    printf("*========================================================*\n");
    sp = &(pool->p_spools[i]);
    printf("* Subpool %-20d                           *\n", i);
    printf("*   Number of devices :      %20u        *\n", sp->sp_devs);
    printf("*   Stripe size       :      %20u        *\n", sp->sp_stripe);
    printf("*   Type              :      ");

    printf("%20s        *\n", get_sptype_name(sp->sp_type));

    for (j=0; j < sp->sp_devs; j++) {
      printf("* - - - - - - - - - - - - - - - - - - - - - - - - - - - -*\n");
      gd = &(sp->sp_gdlist[j]);
      printf("*\tDevice %3d    :      %20s        *\n", j, gd->gd_dev_path);
      printf("*\t  Weight      :      %20u        *\n", gd->gd_my_weight);
      printf("*\t  Blocks      :      %20"FMT_64"u        *\n", gd->gd_nblks);
    }
  }
  printf("**********************************************************\n\n");
}


/*
 * print_pool_list_t - print pool info from pool_list_t
 * @pl: ptr to pool_list_t structure
 * @poolname: the pool to print (accepts keyword 'all')
 *
 * A couple things to note about this function...  The pools are
 * probably already loaded, as you get the pool_list_t ptr
 * basically from ioctl's.
 *
 * Returns:
 */
void print_pool_list_t(pool_list_t *pl, char *poolname){
  int			i,j,k;
  float			capacity;
  char			capacity_str[20];
  pool_status_t		*ps;
  pool_spstatus_t	*sps;
  pool_devstatus_t	*pds;

  for(i=0; i < pl->pl_count; i++){
    if(!strcmp(poolname, "all") || 
       !strcmp(pl->pl_list[i].ps_name, poolname)){
      ps = &(pl->pl_list[i]);
      capacity = ps->ps_capacity/2;/* attention -- assumes 512 byte blks */
      if(capacity > 1024){
	capacity /= 1024;
	sprintf(capacity_str, "%6.1f MB", capacity);
      } else {
	sprintf(capacity_str, "%6.1f kB", capacity);
      }
      if(capacity > 1024){
	capacity /= 1024;
	sprintf(capacity_str, "%6.1f GB", capacity);
      }
      if(capacity > 1024){
	capacity /= 1024;
	sprintf(capacity_str, "%6.1f TB", capacity);
      }
      if(capacity > 1024){
	capacity /= 1024;
	sprintf(capacity_str, "%6.1f PB", capacity);
      }

      printf("\n"
	     "**********************************************************\n");
      printf("* Pool information:                                      *\n");
      printf("* Pool name           :      %20s        *\n",
	     ps->ps_name);
      printf("* Number of subpools  :      %20u        *\n",
	     ps->ps_num_spools);
      printf("* Total Capacity      :      %20s        *\n",
	     capacity_str);
      printf("* In use              :      %20s        *\n",
	     (ps->ps_in_use) ? "YES" : "NO");
      for (j=0; j < ps->ps_num_spools; j++) {
	printf("*========================================================*\n");
	sps = &(ps->ps_splist[j]);
	printf("* Subpool %-20d                           *\n", j);
	printf("*   Number of devices :      %20u        *\n",
	       sps->sps_numdevs);
	printf("*   Stripe size       :      %20u        *\n",
	       sps->sps_stripe);
	printf("*   Total weight      :      %20u        *\n",
	       sps->sps_tot_weight);
	printf("*   Type              :      %20s        *\n",
	       get_sptype_name(sps->sps_type));

	for (k=0; k < sps->sps_numdevs; k++) {
	  pds = &(sps->sps_devlist[k]);
	  printf("* - - - - - - - - - - - - - - "
		 "- - - - - - - - - - - - - -*\n");
	  pds = &(sps->sps_devlist[k]);
	  printf("*\tDevice %20s                      *\n",
		 devt_to_devpath(pds->pds_dev));
	  printf("*\t  Weight      :      %20u        *\n",
		 pds->pds_my_weight);
	  printf("*\t  Blocks      :      %20"FMT_64"u        *\n", pds->pds_blocks);
	}
      }
      printf("**********************************************************\n\n");
    }
  }

}

/*
 * get_pool_info - gets information on all loaded pools
 * @commfd - the file descriptor for the poolcomm (or 0)
 *
 * The function will obtain information on all of the pools
 * and put it in a pool_list_t.  It takes the fd of the pool
 * comm device, but if 0 is given, it will open the comm itself.
 *
 * Returns: ptr to pool_list_t on success, NULL on failure
 */
pool_list_t *get_pool_info(int commfd){
  int			i, j;
  int			poolfd;
  int			I_opened_the_comm = 0;
  uint32		pool_count;
  pool_list_t		*plist;
  pool_status_t		*pstat;
  pool_splist_t		splist;
  pool_devstatus_t	*devstat;
  pool_devlist_t	total_devlist;

  if(!commfd){
    if ((commfd = open("/dev/poolcomm", O_RDONLY)) < 0)  {
      fprintf(stderr, "Couldn't open comm device: %s.\n", strerror(errno));
      return NULL;
    }
    I_opened_the_comm = 1;
  }

  if (ioctl(commfd, POOL_COUNT, &pool_count) < 0){
    fprintf(stderr, "pool communications error: %s\n", strerror(errno));
    close(commfd);
    return NULL;
  }

  if(!pool_count) return NULL;

  plist = (pool_list_t *)malloc(sizeof(pool_list_t));
  memset(plist, 0, sizeof(pool_list_t));
  plist->pl_count = pool_count;

  plist->pl_list = 
    (pool_status_t *)malloc(plist->pl_count * sizeof(pool_status_t));

  if (ioctl(commfd, POOL_LIST, plist) < 0){
    fprintf(stderr, "pool communications error: %s\n", strerror(errno));
    close(commfd);
    return NULL;
  }

  for(i=0; i < plist->pl_count; i++){
    int		index = 0;
    int		dev_count = 0;
    char	poolpath[LINE_MAX];

    pstat = &(plist->pl_list[i]);

    strcpy(poolpath, "/dev/pool/");
    strcat(poolpath, pstat->ps_name);
    if ((poolfd = open(poolpath, O_RDONLY)) < 0)  {
      fprintf(stderr, "Couldn't open pool device: %s.\n", poolpath);
      return NULL;
    }
    
    splist.spl_count = pstat->ps_num_spools;
    splist.spl_list = (pool_spstatus_t *)
      malloc(pstat->ps_num_spools * sizeof(pool_spstatus_t));
    memset(splist.spl_list, 0, splist.spl_count * sizeof(pool_spstatus_t));
    
    if (ioctl(poolfd, POOL_SPLIST, &splist) < 0){
      fprintf(stderr, "Could not communicate with %s\n", poolpath);
      return NULL;
    }

    pstat->ps_splist = splist.spl_list;

    for(j=0, total_devlist.pdl_count = 0; j < splist.spl_count; j++){
      total_devlist.pdl_count += splist.spl_list[j].sps_numdevs;
    }

    total_devlist.pdl_list = (pool_devstatus_t *)
      malloc(total_devlist.pdl_count * sizeof(pool_devstatus_t));
    memset(total_devlist.pdl_list, 0, 
	   (total_devlist.pdl_count * sizeof(pool_devstatus_t)));

    if (ioctl(poolfd, POOL_DEVLIST, &total_devlist) < 0){
      fprintf(stderr, "Could not communicate with %s\n", poolpath);
      return NULL;
    }

    for(j=0; j < splist.spl_count; j++){
      dev_count = splist.spl_list[j].sps_numdevs;
      devstat = (pool_devstatus_t *)
	malloc(dev_count * sizeof(pool_devstatus_t));
      splist.spl_list[j].sps_devlist = devstat;
      memset(devstat, 0, 
	     (dev_count * sizeof(pool_devstatus_t)));
      memcpy(devstat, 
	     &(total_devlist.pdl_list[index]),
	     (dev_count * sizeof(pool_devstatus_t)));
      index += dev_count;
    }
    free(total_devlist.pdl_list);
    close(poolfd);
  }
  if(I_opened_the_comm){
    close(commfd);
  }

  return plist;
}  

/*
 * remove_pool - remove a pool from the kernel
 * @poolname - name of the pool to be removed (keyword "all" accepted)
 *
 * This function will remove one (or all) pool(s) from the kernel.
 * The opposite of assembling a pool.
 *
 * Returns: 0 if ok, -1 on error
 */
int remove_pool(char *poolname){
  int		i;
  int		commfd;
  int		size = 0;
  pool_list_t	list;

  if((commfd = open("/dev/poolcomm", O_RDONLY)) < 0){
    fprintf(stderr, "Unable to open /dev/poolcomm.\n");
    return -1;
  }

  if(pool_in_use(poolname, commfd)){
    fprintf(stderr, "Pool is in use.  Can not remove.\n");
    return -1;
  }

  if (ioctl(commfd, POOL_COUNT, &size) < 0) {
    fprintf(stderr, "pool communications error.\n");
    return -1;
  }

  list.pl_count = size;
  
  list.pl_list = (pool_status_t *)malloc(sizeof(pool_status_t) * size);

  if (ioctl(commfd, POOL_LIST, &list) < 0) {
    fprintf(stderr, "pool communications error.\n");
    return -1;
  }

  for(i = 0; i < list.pl_count; i++){
    if(!strcmp(list.pl_list[i].ps_name, poolname) ||
       !strcmp(poolname, "all")){
      if (ioctl(commfd, POOL_REMOVE, &(list.pl_list[i].ps_pool_minor)) < 0) {
	fprintf(stderr, "Unable to remove pool %s.\n", 
		list.pl_list[i].ps_name);
	return -1;
      } else {
	printf("Unloaded %s.\n", list.pl_list[i].ps_name);
      }
    }
  }
  clean_pool_dir(commfd);
  close(commfd);
  return 0;
}


/*
 * assemble_pool - assemble the pool(s)
 * @poolname: name of the pool to be assembled (keyword "all" accepted)
 * @flags: PTOOLS_<flags> found in common.h
 *
 * This function makes the kernel aware of the pools that have been
 * constructed.  It reads the labels off of the devices, figures out
 * which devices belong to a pool, and makes the kernel aware of them.
 *
 * Returns: 0 if ok, -1 if error
 */
int assemble_pool(char *poolname, int flags){
  int		i,j;
  int		grow = flags & PTOOLS_GROW;
  int		verbose = flags & PTOOLS_VERBOSE;
  int		commfd;
  int		popper = 0;
  label_list_t	*head = NULL;
  label_list_t	*tmp = NULL;
  pool_label_t	*label;
  pool_add_t	*np;
  subpool_t	*ns;
  dmepdev_t	*nl;
  device_t	*nd;
  uint64	spblk_off, pblk_off;
  uint64	pid;

  mkdir("/dev/pool", 0755);
  mknod("/dev/poolcomm", 0600 | S_IFBLK, makedevice(POOL_MAJOR, 0));

  if ((commfd = open("/dev/poolcomm", O_RDONLY)) < 0) {
    fprintf(stderr, "Couldn't open comm device.\n");
    fprintf(stderr, "Be sure pool module is installed "
	    "and you are root.\n");
    return -1;
  }

  clean_pool_dir(commfd);

  if((head = find_and_sort_labels()) == NULL){
    fprintf(stderr, "No Pool labels found.\n");
    return 0;
  }

  while(head){
    if(!strcmp(head->ll_label.pl_pool_name, poolname) ||
       !strcmp(poolname, "all")){
      if(!grow & pool_loaded(head->ll_label.pl_pool_name, commfd)){
	tmp = head; head = head->next; free(tmp);
	continue;
      }
      label = &(head->ll_label);
      np = (pool_add_t *)malloc(sizeof(pool_add_t));
      memset(np, 0, sizeof(pool_add_t)); 
      np->pa_pool_id = label->pl_pool_id;
      np->pa_version = label->pl_version;
      np->pa_num_spools = label->pl_subpools;
      memcpy(np->pa_name, label->pl_pool_name, POOL_NAME_SIZE);

      np->pa_spools = 
	(subpool_t *)malloc(np->pa_num_spools * sizeof(subpool_t));
      memset(np->pa_spools, 0, np->pa_num_spools * sizeof(subpool_t)); 
      pid = np->pa_pool_id;
      
      for(;(head) && !strcmp(head->ll_label.pl_pool_name,
			     label->pl_pool_name);){
	label = &(head->ll_label);
	if(label->pl_pool_id != np->pa_pool_id){ 
	  fprintf(stderr, "Multiple pools with the name \"%s\" found.\n",
		     label->pl_pool_name);
	  tmp = head;  head = head->next;  free(tmp);
	  continue;
	}
	ns = &(np->pa_spools[label->pl_sp_id]);
	ns->sp_devs = label->pl_sp_devs;
	ns->sp_dmepdevs = label->pl_sp_dmepdevs;
	ns->sp_type = label->pl_sp_type;
	ns->sp_stripe = label->pl_striping;
	if (ns->sp_devlist == NULL) {
	  ns->sp_devlist = (device_t *)malloc(ns->sp_devs * sizeof(device_t));
	  memset(ns->sp_devlist, 0, ns->sp_devs * sizeof(device_t)); 
	}
	if (ns->sp_dmepdevlist == NULL) {
	  ns->sp_dmepdevlist = (dmepdev_t *)malloc(ns->sp_dmepdevs * 
						 sizeof(dmepdev_t));
	  memset(ns->sp_dmepdevlist, 0, ns->sp_dmepdevs * sizeof(dmepdev_t));
	}
	if (label->pl_blocks) {
	  nd = &(ns->sp_devlist[label->pl_sp_devid]);
	  nd->d_nblks = label->pl_blocks;
	  nd->d_adev = head->ll_dev;
	}
	if (label->pl_sp_weight) {
	  nl = &(ns->sp_dmepdevlist[label->pl_sp_dmepid]);
	  nl->dm_my_weight = label->pl_sp_weight;
	  nl->dm_adev = head->ll_dev;
	}
	tmp = head;  head = head->next;  free(tmp);
      }

      pblk_off = 0;	/* logical blk offsets - scope is entire pool */
      spblk_off = 0;    /* logical blk offsets - scope is one subpool */

      for (i=0; i < np->pa_num_spools; i++) {
	uint32 sp_weight;
	ns = &(np->pa_spools[i]);

	ns->sp_start = pblk_off;
	spblk_off = 0;
	sp_weight = 0;

	if (ns->sp_devs == 0){
	  fprintf(stderr, "Pool %s, subpool %d has no devices.\n",
		  np->pa_name, i);
	  goto exit_err;
	}
    
	for (j=0; j < ns->sp_devs; j++) {
	  nd = &(ns->sp_devlist[j]);

	  if (nd->d_nblks == 0){
	    fprintf(stderr, "Pool %s, subpool %d, device %d has no blocks.\n",
		    np->pa_name, i,j);
	    goto exit_err;
	  }

	  nd->d_start = spblk_off;
	  spblk_off += nd->d_nblks;
	}
    
	for (j=0; j < ns->sp_dmepdevs; j++) {
	  nl = &(ns->sp_dmepdevlist[j]);

	  if (nl->dm_my_weight == 0){
	    fprintf(stderr, "Subpool %d, dmep %d has no weight.\n", i,j);
	    goto exit_err;
	  }
	  sp_weight += nl->dm_my_weight;
	}
    
	ns->sp_len = spblk_off;
	ns->sp_tot_weight = sp_weight;
	pblk_off += spblk_off;    
      }
      np->pa_capacity = pblk_off;

      if(!grow){  /* Assemble */
	if (ioctl(commfd, POOL_ADD, (void *)np) < 0) {
	  fprintf(stderr, "Adding %s (ioctl) failed: %s.\n", 
		  np->pa_name, strerror(errno));
	} else {
	  char buf_bdsk[LINE_MAX];
	  struct stat statbuf;
	
	  popper = 1;
	  strcpy(buf_bdsk, "/dev/pool/");
	  strncat(buf_bdsk, np->pa_name, POOL_NAME_SIZE);

	  if (stat(buf_bdsk, &statbuf) != 0){
	    if (mknod(buf_bdsk, 0600 | S_IFBLK, (dev_t)np->pa_pool_dev) < 0)
	      fprintf(stderr, "Unable to make %s: %s.\n",
		      buf_bdsk, strerror(errno));
	  }

	  if(verbose){
	    pool_list_t *pl;
	    pl = get_pool_info(commfd);
	    print_pool_list_t(pl, np->pa_name);
	  }
	  else {
	    printf("Added %s.\n", np->pa_name);
	  }
	}
      } else {
	popper = 1;
	if(verbose){
	  pool_list_t *pl;
	  pl = get_pool_info(commfd);
	  printf("********************  POOL BEFORE GROW  ******************\n");
	  print_pool_list_t(pl, np->pa_name);
	}
	if (ioctl(commfd, POOL_GROW, (void *)np) < 0) {
	  fprintf(stderr, "Growing %s (ioctl) failed: %s.\n", 
		  np->pa_name, strerror(errno));
	  goto exit_err;
	} else {
	  if(verbose){
	    pool_list_t *pl;
	    pl = get_pool_info(commfd);
	    printf("********************  POOL AFTER GROW  "
		   "*******************\n");
	    print_pool_list_t(pl, np->pa_name);
	  } else {
	    printf("Pool \"%s\" has been grown.\n", np->pa_name);
	  }
	}
      }
      for(i = 0; i < np->pa_num_spools; i++){
	ns = &(np->pa_spools[i]);
	if(ns->sp_devlist){ free(ns->sp_devlist); }
	if(ns->sp_dmepdevlist){ free(ns->sp_dmepdevlist); }
      }
      free(np->pa_spools);
      free(np);
    }else{
      tmp = head;  head = head->next;  free(tmp);
      continue;
    }	
  }

  close(commfd);
  if(!popper)
    printf("No pools added.\n");
  return 0;

 exit_err:
  /* free memory */
  for(i = 0; i < np->pa_num_spools; i++){
    ns = &(np->pa_spools[i]);
    if(ns->sp_devlist){ free(ns->sp_devlist); }
    if(ns->sp_dmepdevlist){ free(ns->sp_dmepdevlist); }
  }
  free(np->pa_spools);
  free(np);

  close(commfd);
  return -1;
}


/*
 * compute_and_print_lost_blks
 * @pool
 *
 * This fuction will compute and the print the blocks lost from
 * forming the pool.  Attention -- assumes 512 blocks.
 * 
 * Returns:
 */
void compute_and_print_lost_blks(pool_t *pool){
  int 			i,j;
  uint64		subpool_size;
  uint64		label_blks = 0;
  uint64		boundary_blks = 0;
  uint64		uneven_stripe_dev = 0;
  uint64		stripe_blks = 0;
  uint64		min_dev_size;
  subpool_t		*sp;
  generic_device_t	*gd;
  
  for(i = 0; i < pool->p_num_spools; i++){
    sp = &(pool->p_spools[i]);

    subpool_size = 0;

    for(j = 0; j < sp->sp_devs; j++){
      subpool_size += sp->sp_gdlist[j].gd_nblks;
    }
    
    if(subpool_size % 16){
      for(j = 0; j < sp->sp_devs; j++){
	label_blks += 16;				/* used by label */
	if(sp->sp_gdlist[j].gd_nblks){
	  boundary_blks += sp->sp_gdlist[j].gd_nblks % 16; /* 8k boundary */
	}
      }
    }
    
    if(sp->sp_stripe){
      min_dev_size = -1;
      if(sp->sp_devs <= 0){
	continue;
      }
      for(j = 0; j < sp->sp_devs; j++){
	gd = &(sp->sp_gdlist[j]);
	if(min_dev_size == -1){
	  min_dev_size = gd->gd_nblks;
	}
	if((gd->gd_nblks != 0) && (min_dev_size != gd->gd_nblks)){
	  min_dev_size = (min_dev_size < gd->gd_nblks) ?
	    min_dev_size : gd->gd_nblks;
	}
      }
      for(j = 0; j < sp->sp_devs; j++){
	uint64		tmp = 0;
	gd = &(sp->sp_gdlist[j]);
	uneven_stripe_dev += gd->gd_nblks - min_dev_size; /* same size devs */
	tmp = min_dev_size-16;  /* label (accounted for) */
	/* stripe boundaries */
	stripe_blks += tmp - (tmp/sp->sp_stripe)*sp->sp_stripe;
      }
    }
  }
  printf("\n");
  printf("Blocks used for pool labels               : %14"FMT_64"u\n",
	 label_blks);
  printf("Blocks lost to boundaries constraints     : %14"FMT_64"u\n",
	 boundary_blks);
  printf("Blocks lost to due to stripe size         : %14"FMT_64"u\n", 
	 stripe_blks);
  printf("Blocks lost to non-uniform stripe devices : %14"FMT_64"u\n",
	 uneven_stripe_dev);
}


/*
 * scan_and_print_labels
 * 
 * This function looks at all partitions.  If they have a label, it will
 * report the name of the pool the device belongs to.
 *
 * Returns: 0 if ok, -1 on error
 */
int scan_and_print_labels(void){
  FILE		*fp;
  char		proc_line[LINE_MAX];
  int		line_num = 0;
  char		maj[20], min[20], blocks[20], name[LINE_MAX];
  char		tmp_str[LINE_MAX];
  char		buffer[sizeof(pool_label_t)];
  pool_label_t	label;

  if((fp = fopen("/proc/partitions", "r")) == NULL){
    fprintf(stderr, "Could not open /proc/partitions\n");
    return -1;
  }
      

  printf("  %-25s%25s\n", "Device", "Pool Label");
  printf("  ======                                  ==========\n");
  while(fgets(proc_line, LINE_MAX, fp) != NULL){
    int		fd;
    if(line_num++ < 2) /* skip first two lines */
      continue;
    
    sscanf(proc_line,"%s %s %s %s", maj, min, blocks,name);
    sprintf(tmp_str,"/dev/%s",name);
    if((fd = open(tmp_str, O_RDONLY)) < 0){
      fprintf(stderr, "Could not open %s\n", tmp_str);
      fclose(fp);
      return -1;
    }
    if(read(fd, buffer, sizeof(pool_label_t)) < sizeof(pool_label_t)){
      fprintf(stderr, "Could not read %s\n", tmp_str);
      close (fd);
      fclose(fp);
      return -1;
    }

    pool_label_in(&label, buffer);
    close(fd);
    printf("  %-25s%25s\n", tmp_str,
	   (label.pl_magic == POOL_MAGIC) ?
	   label.pl_pool_name : 
	   (ext2sb_check(tmp_str)) ?
	   "<- ext2fs ->" :
	   "< none >");
  }
  fclose(fp);

  return 0;
}


/*
 * erase_pool_labels - erases pool labels from disk
 * @poolname - name of the pool whose labels are to be erased
 *
 * If the keyword of all is given, then erase all labels.
 *
 * Returns: 0 if ok, -1 on error
 */
int erase_pool_labels(char *poolname, int flags){
  int		force = PTOOLS_FORCE & flags;
  FILE		*fp;
  char		proc_line[LINE_MAX];
  int		line_num = 0;
  char		maj[20], min[20], blocks[20], name[LINE_MAX];
  char		tmp_str[LINE_MAX];
  char		buffer[sizeof(pool_label_t)];
  pool_label_t	label;
  int		commfd;

  if((commfd = open("/dev/poolcomm", O_RDONLY)) < 0){
    fprintf(stderr, "Unable to open /dev/poolcomm.\n");
    return -1;
  }

  if(pool_loaded(poolname, commfd)){  /* note special case for all */
    fprintf(stderr, "%s is loaded.  Can not erase labels.\n",
	   (!strcmp(poolname, "all"))? "A pool": poolname);
    return -1;
  }

  close(commfd);

  if(!force){
    char	response[LINE_MAX];

    fprintf(stderr, "Warning!  About to destroy %s on disk.\n"
	    "Are you certain you want to continue? [yN] ",
	    (!strcmp(poolname, "all")) ? "all labels" : poolname);
    scanf("%s", response);

    if (tolower(response[0]) != 'y') {
      fprintf(stderr, "Destroy %s canceled.\n", poolname);
      return(-1);
    }
  }

  if((fp = fopen("/proc/partitions", "r")) == NULL){
    fprintf(stderr, "Could not open /proc/partitions\n");
    return -1;
  }
      
  while(fgets(proc_line, LINE_MAX, fp) != NULL){
    int		fd;
    if(line_num++ <= 2) /* skip first two lines */
      continue;
    
    sscanf(proc_line,"%s %s %s %s", maj, min, blocks,name);
    sprintf(tmp_str,"/dev/%s",name);
    if((fd = open(tmp_str, O_RDWR)) < 0){
      fprintf(stderr, "Could not open %s\n", tmp_str);
      fclose(fp);
      return -1;
    }
    if(read(fd, buffer, sizeof(pool_label_t)) < sizeof(pool_label_t)){
      fprintf(stderr, "Could not read %s\n", tmp_str);
      close (fd);
      fclose(fp);
      return -1;
    }

    pool_label_in(&label, buffer);
    if((label.pl_magic == POOL_MAGIC) && 
       (!strcmp(label.pl_pool_name, poolname) ||
	!strcmp(poolname, "all"))){
      printf("Erasing label on %s\n", tmp_str);
      memset(buffer, 0, sizeof(pool_label_t));
      if(lseek(fd, 0, SEEK_SET) < 0){
	fprintf(stderr, "Failed to erase label on %s\n", tmp_str);
	return -1;
      }	
      if(write(fd, buffer, sizeof(pool_label_t)) < 0){
	fprintf(stderr, "Failed to erase label on %s\n", tmp_str);
	return -1;
      }
    }
    close(fd);
  }
  fclose(fp);

  return 0;

}


/*
 * get_pool_params_from_file - translate config file to incore structures
 * @file: the pool configuration file
 * @pool: ptr that will store config information
 *
 * This fuction will read a pool configuration file and store the
 * information in the standard incore structures.
 *
 * Returns: 0 if ok, -1 on error
 */
int get_pool_params_from_file(pool_t **pool, const char *file){
  FILE		*fp = NULL;
  char	        buffer[LINE_MAX];
  char	        arg1[24], arg2[24], arg3[24], arg4[24];
  int		line_cnt=0;
  int		size = 0;
  int		i;
  pool_t	*new_pool = NULL;

  if ((fp = fopen(file, "r")) == NULL) {
    fprintf(stderr, "Error opening the parameter file \"%s\"\n", file);
    goto fail;
  }

  new_pool = (pool_t *)malloc(sizeof(pool_t));
  memset(new_pool, 0, sizeof(pool_t));

  bzero(buffer, LINE_MAX);
  while(fgets(buffer, LINE_MAX, fp) != NULL){
    char	keyword[24];
    line_cnt++;

    bzero(keyword, 24);
    bzero(arg1, 24);
    bzero(arg2, 24);
    bzero(arg3, 24);
    bzero(arg4, 24);

    if (buffer[0] == '#')
      continue;
      
    if (buffer[0] == '\n')
      continue;
      
    /* Scan for a keyword and up to five arguments */
      
    sscanf(buffer, "%s %s %s %s %s", keyword, arg1,arg2,arg3,arg4);

    /* Check for various recognized keywords */

    /*************
     **  poolname
     ************/
    if(!strcmp(keyword, "poolname")){
      if (strlen(arg1) <= 0) {
	fprintf(stderr, "\"%s\", line %d: error: poolname error\n%s\n",
		file, line_cnt, buffer);
	goto fail;
      }
      if(!strcmp(arg1, "all")){  /* "all" is a keyword */
	fprintf(stderr, "\"%s\", line %d: error: poolname error\n%s\n"
		"\"%s\" is a keyword.\n",
		file, line_cnt, buffer, arg1);
	goto fail;
      }
      strncpy(new_pool->p_name, arg1, POOL_NAME_SIZE);
    }

    /*************
     **  subpools
     ************/
    else if(!strcmp(keyword, "subpools")){
      if(new_pool->p_num_spools){
	/* multiple specifications for number of subpools */
	fprintf(stderr, "\"%s\", line %d: error: subpools error\n%s\n",
		file, line_cnt, buffer);
	goto fail;
      }
      if ((new_pool->p_num_spools = atoi(arg1)) < 1) {
	/* invalid argument */
	fprintf(stderr, "\"%s\", line %d: error: subpools error\n%s\n",
		file, line_cnt, buffer);
	goto fail;
      }

      size = new_pool->p_num_spools * sizeof(subpool_t);
      new_pool->p_spools = (subpool_t *)malloc(size);
      memset(new_pool->p_spools, 0, size);
    }

    /************
     **  subpool
     ***********/
    else if(!strcmp(keyword, "subpool")){
      int 	done = FALSE;
      int	spid = atoi(arg1);
      int	stripe = atoi(arg2);
      int	spdevs = atoi(arg3);
      subpool_t	*sp = NULL;

      if (new_pool->p_num_spools == 0) {
	/* number of subpools not specified */
	fprintf(stderr, "\"%s\", line %d: error: "
		"\"subpools\" keyword must be assigned"
		" before \"subpool\"\n%s\n", file, line_cnt, buffer);
	goto fail;
      }
      if (spid >= new_pool->p_num_spools) {
	fprintf(stderr, "\"%s\", line %d: error: "
		"bad subpool id"
		"\n%s\n",
		file, line_cnt, buffer);
	goto fail;
      }
      if (spdevs < 1) {
	fprintf(stderr, "\"%s\", line %d: error: "
		"subpool must have at least 1 device"
		"\n%s\n",
		file, line_cnt, buffer);
	goto fail;
      }
      sp = &(new_pool->p_spools[spid]);
      if(sp->sp_devs){
	fprintf(stderr, "\"%s\", line %d: error: "
		"duplicate subpool id's"
		"\n%s\n",
		file, line_cnt, buffer);
	goto fail;
      }
      sp->sp_devs = spdevs;
      if(stripe%8){
	fprintf(stderr, "Stripe sizes must be a multiple of 8 (4kB).\n");
	goto fail;
      }
      sp->sp_stripe = stripe;

      for (i = 0; sptype_names[i].name[0]; i++) {
        if (strcmp(sptype_names[i].name, arg4) == 0) {
	  sp->sp_type = sptype_names[i].label;
	  done = TRUE;
	}
      }

      if (*arg4 && !done) {
	fprintf(stderr, "\"%s\", line %d: error: "
		"bad subpool type"
		"\n%s\n",
		file, line_cnt, buffer);
	goto fail;
      }
      
      size = sp->sp_devs * sizeof(generic_device_t);
      sp->sp_gdlist = (generic_device_t *)malloc(size);
      memset(sp->sp_gdlist, 0, size);
    }

    /***************
     **  pooldevice
     **************/
    else if(!strcmp(keyword, "pooldevice")){
      int	spid = atoi(arg1);
      int	pdid = atoi(arg2);
      int	blks = 0;
      int	line_num = 0;
      int	dmep_weight =0;
      char	proc_line[LINE_MAX];
      FILE	*tmp_fp;
      int	tmp_fd;
      char	maj[24], min[24], blocks[24], name[24], tmp_str[24];
      subpool_t	*sp;

      if(!is_blkdev(arg3)){
	fprintf(stderr, "\"%s\", line %d: error: "
		"Bad device or path"
		"\n%s\n",
		file, line_cnt, buffer);
	goto fail;
      }

      if((tmp_fp = fopen("/proc/partitions", "r")) == NULL){
	fprintf(stderr, "Could not open /proc/partitions\n");
	goto fail;
      }
      
      while(fgets(proc_line, LINE_MAX, tmp_fp) != NULL){
	if(line_num++ <= 2) /* skip first two lines */
	  continue;

	sscanf(proc_line,"%s %s %s %s", maj, min, blocks,name);

	sprintf(tmp_str,"/dev/%s",name);
	if(!strcmp(tmp_str,arg3)){
	  /* convert to 512 byte blocks */
	  blks = (atoll(blocks) << 1);
	  break;
	}
      }
      fclose(tmp_fp);

      if(!blks){
	if ((tmp_fd = open(arg3, O_RDONLY)) < 0) {
	  fprintf(stderr, "\"%s\", line %d: error: "
		  "Could not open device"
		  "\n%s\n",
		  file, line_cnt, buffer);
	  goto fail;
	}
            
	if(ioctl(tmp_fd,BLKGETSIZE,&blks)<0) {
	  fprintf(stderr, "\"%s\", line %d: error: "
		  "Could not obtain size of device"
		  "\n%s\n",
		  file, line_cnt, buffer);
	  close(tmp_fd);
	  goto fail;
	}
	close(tmp_fd);

	if(blks <= 0) {
	  fprintf(stderr,
		  "Warning:  Ptool detected 0 blocks for %s\n",
		  arg3);
	}
      }
      dmep_weight = atoi(arg4);

      if(spid >= new_pool->p_num_spools){
	fprintf(stderr, "\"%s\", line %d: error: "
		"bad subpool id"
		"\n%s\n",
		file, line_cnt, buffer);
	goto fail;
      }

      sp = &(new_pool->p_spools[spid]);

      /* I used to think this check was necessary, now I do not.
      if (sp->sp_type == SPTYPE_GFS_JOURNAL && dmep_weight != 0) {
	fprintf(stderr, "\"%s\", line %d: error: "
		"Journal device should have DMEP weight of zero"
		"\n%s\n",
		file, line_cnt, buffer);
	goto fail;
	}*/

      if(pdid >= sp->sp_devs){
	fprintf(stderr, "\"%s\", line %d: error: "
		"bad device id"
		"\n%s\n",
		file, line_cnt, buffer);
	goto fail;
      }

      sp->sp_gdlist[pdid].gd_dev_path = (char *)malloc(strlen(arg3));
      strcpy(sp->sp_gdlist[pdid].gd_dev_path, arg3);
      sp->sp_gdlist[pdid].gd_my_weight = dmep_weight;

      if(!blks){
	if(!dmep_weight){
	  fprintf(stderr, "\"%s\", line %d: error: "
		"No blocks and no DMEP - device is useless"
		"\n%s\n",
		file, line_cnt, buffer);
	  goto fail;
	}
      } else {
	uint64 tmp;
	generic_device_t *gd = &(sp->sp_gdlist[pdid]);
	/* Attention -- this is based on 512 blocks and should be fixed */

	tmp = gd->gd_nblks = blks-16;/* 16 blocks offset for label */
	if(sp->sp_stripe){
	  gd->gd_nblks = (gd->gd_nblks/sp->sp_stripe)*sp->sp_stripe;
	}
      }
    } else {
      fprintf(stderr, "\"%s\", line %d: error: "
	      "Invalid keyword"
	      "\n%s\n",
	      file, line_cnt, buffer);
      goto fail;
    }
    bzero(buffer, LINE_MAX);
  }

  fclose(fp);

  /* verify that a pool name was specified */
  if(strlen(new_pool->p_name) <= 0){
    fprintf(stderr, "Error.  No pool name was specified.\n");
    goto fail;
  }

  /* verify all subpools and pooldevices are defined */

  for (i=0; i < new_pool->p_num_spools; i++) {
    int		j;
    subpool_t	*sp = NULL;

    sp = &(new_pool->p_spools[i]);
    if (sp->sp_devs == 0) {
      fprintf(stderr, "Error.  subpool %d has no devices.\n", i);
      goto fail;
    }

    for (j=0; j < sp->sp_devs; j++) {
      generic_device_t *gd = NULL;
      gd = &(sp->sp_gdlist[j]);
      if ((gd->gd_nblks == 0) && (gd->gd_my_weight == 0)) {
	fprintf(stderr, "Error. subpool %d, device %d not defined.\n", i, j);
	goto fail;
      }
    }
  }



  *pool = new_pool;
  return 0;

 fail:
  if(new_pool){
    if(new_pool->p_spools){
      for(i=0; i < new_pool->p_num_spools; i++){
	if(new_pool->p_spools[i].sp_devs){
	  free(new_pool->p_spools[i].sp_gdlist);
	}
      }
      free(new_pool->p_spools);
    }
    free(new_pool);
  }
  return -1;
}


/*
 * write_pool_labels - writes the labels out on the disks
 * @pool - the pool_t ptr whose info will be written
 *
 * This function takes a pool_t structure and translates
 * that into labels which can be written to the disks. This
 * function may change the information in the pool_t, so be
 * careful.
 *
 * Returns 0 if Ok, -1 on error
 */

int write_pool_labels(pool_t *pool, int flags){
  int			i,j;
  int			grow  = flags & PTOOLS_GROW;
  int			force = flags & PTOOLS_FORCE;
  char			buffer[sizeof(pool_label_t)];
  subpool_t		*sp;
  pool_label_t		label;

  memset(&label, 0, sizeof(pool_label_t));
  memset(&buffer, 0, sizeof(pool_label_t));
  srand(time(NULL));

  label.pl_magic = POOL_MAGIC;
  label.pl_version = POOL_VERSION;
  if(!grow){
    label.pl_pool_id=
      (pool->p_spools[0].sp_gdlist[0].gd_nblks << 32) + rand();
  } else {
    label.pl_pool_id = pool->p_pool_id;
  }
  strncpy(label.pl_pool_name, pool->p_name, POOL_NAME_SIZE);
  label.pl_subpools = pool->p_num_spools;
  
  /* We can not allow labels to be written that have a name in common **
  ** with another pool's labels which have already been written...... */
  switch(labels_exist_with_same_name(pool)){
  case 1:
    if(!grow && !force){
      fprintf(stderr, "The pool name \"%s\" is already used.\n",
	      pool->p_name);
      return -1;
    }
    break;
  case 0:
    break;
  default:
    fprintf(stderr, "Unable to check devices for labels.\n");
    return -1;
  }

  /* If force is true, the devices in the pool_t, which is about **
  ** to be written, will *not* be checked for other labels.  So, **
  ** it would be possible to overwrite other pool labels.        */
  switch(labels_exist_on_devices(pool)){
  case 1:  /* labels were found */
    if(!force){
      char	response[LINE_MAX];
      printf("\nHey Ho!  You are attempting to write");
      printf(" over previously written labels.\n");
      printf(" Before writing these labels, you must");
      printf(" erase any conflicting, unwanted pools.\n\n");
      printf("Would you like the conflicting labels erased?\n"
	     "\t(E)rase, (Q)uit : ");
      scanf("%s", response);
      printf("\n");
      if(tolower(response[0]) != 'e'){
	return -1;
      }
    }
    if(erase_conflicting_labels(pool) < 0){
      fprintf(stderr, "Failed to erase conflicting labels.\n"
	      "This is potentially hazardous, as some labels may have\n"
	      " been erase and others have not.  Try running ptool\n"
	      " again, if this does not fix the problem, you should\n"
	      " contact the GFS team.\n\n");
      return -1;
    }
    break;
  case 0:  /* no labels were found */
    break;
  default: /* an error occurred while checking for labels */
    return -1;
    break;
  }


  /* check for ext2 FS on partitions.......................... **
  ** We may be best served to make this a separate function... */
  {
    int		rtn, result = 0;
    for(i=0; i < pool->p_num_spools; i++){
      sp = &(pool->p_spools[i]);
      for(j=0; j < sp->sp_devs; j++)
	if((rtn = ext2sb_check(sp->sp_gdlist[j].gd_dev_path))){
	  if(result != -1) result = rtn;
	}
    }
    switch(result){
    case 0: /* ext2 fs not found on any of the devices */
      break;
    case 1: /* ext2 fs was found on one or more devices */
      if(!force){
	char	response[LINE_MAX];
	printf("\nHey Ho!  You are attempting to write labels");
	printf(" onto a device with an EXT2 FS.\n");
	printf("This may be a partition you are using, are you"
	       " sure you want to do this?\n"
	       "\t(Y)es, (Q)uit : ");
	scanf("%s", response);
	printf("\n");
	if(tolower(response[0]) != 'y'){
	  return -1;
	}
      }
      break;
    default: /* an error has occurred */
      return -1;
      break;
    }
  }

  /* We check here if the striped devices in the pool are the  **
  ** same size.  If force is true, the devices will be altered **
  ** so that they are the same size. It goes something like:   **
  ** force    stripe check                                     **
  **   0	   0	   dev sizes differ, query user        **
  **   0	   1	   dev sizes equal, ok                 **
  **   1	   0	   force makes devs equal to smallest  **
  **   1	   1	   dev sizes equal, ok                 */
  if(!stripe_devs_same_size(pool) && !force){
    
    char	response[LINE_MAX];
    fprintf(stderr, "\n\nDevices used in striped subpool are"
	    " not of the same size!\n"
	    "If you continue, the devices you are striping will "
	    "be normalized to the smallest device.\n"
	    "\n\t(C)ontinue, (Q)uit : ");
    scanf("%s", response);
    if(tolower(response[0]) != 'c'){
      return -1;
    }
  }

  /* (8kB) boundary check */
  /* Attention -- Assuming 512 block size */ 
  for (i=0; i < pool->p_num_spools; i++) {
    int		subpool_size;

    sp = &(pool->p_spools[i]);
    subpool_size = 0;

    for(j = 0; j < sp->sp_devs; j++){
      subpool_size += sp->sp_gdlist[j].gd_nblks;
    }
    
    if(subpool_size % 16){
      for(j = 0; j < sp->sp_devs; j++){
	if(sp->sp_gdlist[j].gd_nblks){
	  sp->sp_gdlist[j].gd_nblks =
	    (sp->sp_gdlist[j].gd_nblks/16)*16;
	}	
      }      
    }
  }

  for(i=0; i < pool->p_num_spools; i++){
    int devid_cnt, dmepid_cnt;

    sp = &(pool->p_spools[i]);
    devid_cnt = dmepid_cnt = 0;
    label.pl_sp_devs = label.pl_sp_dmepdevs = 0;

    label.pl_sp_id = i;
    for(j=0; j < sp->sp_devs; j++){
      if (sp->sp_gdlist[j].gd_my_weight){
	label.pl_sp_dmepdevs++;
      }
      if(sp->sp_gdlist[j].gd_nblks){
	label.pl_sp_devs++;
      }
    }

    label.pl_sp_type = sp->sp_type;
    label.pl_striping = sp->sp_stripe;

    for (j=0; j < sp->sp_devs; j++) {
      int		fd;
      generic_device_t	*gd;

      gd = &(sp->sp_gdlist[j]);

      label.pl_blocks = gd->gd_nblks;
      label.pl_sp_weight = gd->gd_my_weight;

      if (label.pl_blocks)
	label.pl_sp_devid = devid_cnt++;
      else
	label.pl_sp_devid = (uint32)(-1);

      if (label.pl_sp_weight)
	label.pl_sp_dmepid = dmepid_cnt++;
      else
	label.pl_sp_dmepid = (uint32)(-1);
      
      memset(buffer, 0, sizeof(pool_label_t));
      pool_label_out(&label, buffer);

      if ((fd = open(gd->gd_dev_path, O_WRONLY)) < 0) {
	fprintf(stderr, "Failed to open %s.\n", gd->gd_dev_path);
	return -1;
      }

      if (write(fd, buffer, sizeof(pool_label_t)) < sizeof(pool_label_t)) {
	fprintf(stderr, "Failed to write to %s\n", gd->gd_dev_path);
	return -1;
      }

      fsync(fd);
      close(fd);
    }
  }
  return 0;
}






/*
 * grow_pool - grow an in-use pool
 * @pool: description of the pool to be grown
 * @flags: PTOOLS_VERBOSE, PTOOLS_FORCE, etc
 *
 * This function will do some sanity checks, write new labels to
 * disk, and reassemble the new pool.
 *
 * Returns: 0 if ok, -1 on error
 */
int grow_pool(pool_t *pool, int flags){
  int			force = flags & PTOOLS_FORCE;
  /*int			verbose = flags & PTOOLS_VERBOSE;*/
  int			error = 0;
  int			i,j;
  pool_list_t		*pl;
  pool_status_t		*ps = NULL;
  pool_spstatus_t	*sps;

  flags |= PTOOLS_GROW;  /* add grow to the flags */

  /* Growing will only work in 2.2 kernel right now */

  if((pl = get_pool_info(0)) == NULL){
    fprintf(stderr, "Unable to query for loaded pools.\n");
    error = -1;
    goto fail;
  }

  for(i = 0; i < pl->pl_count; i++){
    ps = &(pl->pl_list[i]);
    if(!strcmp(ps->ps_name, pool->p_name)){
      break;
    }
  }

  if(i == pl->pl_count){
    fprintf(stderr, "Pool, %s, is not loaded.\n", pool->p_name);
    error = -1;
    goto fail;
  }

  /* new number of subpools should be greater than before */
  if(ps->ps_num_spools >= pool->p_num_spools){
    fprintf(stderr, "The number of subpools in the new pool should be\n"
	    "\tgreater than the number in the old.\n");
    error = -1;
    goto fail;
  }

  /* All devices in old pool should be in the new */
  for(i = 0; i < ps->ps_num_spools; i++){
    dev_t		device;
    pool_devstatus_t	*pds;

    subpool_t		*sp;
    generic_device_t	*gd;

    sp  = &(pool->p_spools[i]); /* the new */
    sps = &(ps->ps_splist[i]);  /* the old */

    for(j = 0; j < sps->sps_numdevs; j++){
      if(sp->sp_devs != sps->sps_numdevs){
	fprintf(stderr, "When growing, you can add subpools, but "
		"not change them.\n");
	error = -1;
	goto fail;
      }
      
      gd  = &(sp->sp_gdlist[j]);
      pds = &(sps->sps_devlist[j]);
      
      device = devpath_to_devt(gd->gd_dev_path);
      if(minor(pds->pds_dev) != minor(device) ||
	 major(pds->pds_dev) != major(device)){
	fprintf(stderr, "Devices in subpools don't match up.\n");
	error = -1;
	goto fail;
      }
    }
  }

  pool->p_pool_id = ps->ps_pool_id;

  if(!force){
    char	response[LINE_MAX];

    fprintf(stderr, "Warning!  You are about to write new labels for %s\n"
	    "Are you certain you want to continue? [yN] ",
	    ps->ps_name);
    scanf("%s", response);

    if (tolower(response[0]) != 'y') {
      fprintf(stderr, "Grow %s canceled.\n", pool->p_name);
      return(-1);
    }
  }

  /* should check here if the new portion will overwrite other labels */
  /* and if any of those conflicting labels belong to pools that are  **
  ** loaded.......................................................... */

  /* must add force option if not there... assume all interactive checks **
  ** have been made .................................................... */
  if(write_pool_labels(pool, PTOOLS_FORCE | flags)<0){
    error = -1;
    goto fail;
  }

  assemble_pool(pool->p_name, flags);

  flags &= ~PTOOLS_GROW;  /* remove grow flag */

 fail:

  if(pl){
    for(i=0; i < pl->pl_count; i++){
      ps = &(pl->pl_list[i]);
      if(ps && ps->ps_splist){
	for(j = 0; j < ps->ps_num_spools; j++){
	  sps = &(ps->ps_splist[j]);
	  if(sps->sps_devlist){ free(sps->sps_devlist); }
	}
	free(ps->ps_splist);
      }
    }
    free(pl->pl_list);
    free(pl);
  }
  return error;
}



