/******************************************************************************
*******************************************************************************
**
**    Copyright 2001 Sistina Software Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Jonathan Brassow <brassow@sistina.com>
** 
**    Module name: pinfo.c
**
**    Description:
**	Pinfo is used to gain information on the pools that are loaded in
**	the kernel (live pools).
**
*******************************************************************************
******************************************************************************/
#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#include "common.h"
#include "pool_ioctl.h"

static void print_usage(void);
static void print_pool_config_file(char *poolname);
static void print_pool_stats(char *poolname, int interval);
static void clear_pool_stats(void);
static void print_pool_info(char *poolname, int verbose);


int main(int argc, char *argv[]){
  int c;
  int verbose = 0;
  int file_output = 0;
  int stats = 0;
  int interval = 0;
  int clearstats = 0;
  char *poolname = NULL;

  while ((c = getopt(argc, argv, "hf:vs:t:c")) != -1) {
    switch (c) {
    case 'h':
      print_usage();
      exit(EXIT_SUCCESS);
      break;
    case 'f':
      file_output = 1;
      poolname = optarg;
      break;
    case 'v':
      verbose = 1;
      break;
    case 's':
      stats = 1;
      poolname = optarg;
      break;
    case 't':
      interval = atoi(optarg);
      break;
    case 'c':
      clearstats = 1;
      break;
    default:
      print_usage();
      exit(EXIT_FAILURE);
    }
  }
  if(file_output){
    if(!poolname){
      exit(EXIT_FAILURE);
    }
    print_pool_config_file(poolname);
  }
  else if(stats | interval){
    poolname = argv[optind];
    if (!poolname){
      poolname = "all";
    } 
    if(!stats){
      fprintf(stderr, "The '-t' option requires '-s' option.\n");
      exit(EXIT_FAILURE);
    }
    print_pool_stats(poolname, interval);
  }
  else if(clearstats){
    clear_pool_stats();
  }
  else {
    poolname = argv[optind];
    if (!poolname){
      poolname = "all";
    }
    print_pool_info(poolname, verbose);
  }
  
  exit(EXIT_SUCCESS);
}

static void print_usage(void){
  printf("Usage: pinfo [-hsfcdt]\n");
  printf("  -h         Print this help, then exit.\n");
  printf("  -f <pool>  Print the configuration parameter file"
	 " for the pool specified.\n");
  printf("  -v         Verbose.  More complete information.\n");
  printf("  -c         Clear all pool statistics in kernel.\n");
  printf("  -s <pool>  Print pool device IO statistics.\n");
  printf("  -t sec     Use with -s for continuous updates with sec delay.\n");
}

static void print_pool_config_file(char *poolname){
  int			i, j, k;
  pool_list_t		*pl;
  pool_status_t		*ps;
  pool_spstatus_t	*sps;
  pool_devstatus_t	*pds;

  if((pl = get_pool_info(0)) == NULL){
    return;
  }
  for(i=0; i < pl->pl_count; i++){
    if(!strcmp(pl->pl_list[i].ps_name, poolname)){
      ps = &(pl->pl_list[i]);
      printf("poolname\t%s\n\n", ps->ps_name);
      printf("subpools\t%u\n\n", ps->ps_num_spools);
      for(j=0; j < ps->ps_num_spools; j++){
	sps = &(ps->ps_splist[j]);
	printf("subpool\t%d\t%u\t%u\t%s\n",
	       j, sps->sps_stripe, sps->sps_numdevs,
	       get_sptype_name(sps->sps_type));
      } printf("\n");
      for(j=0; j < ps->ps_num_spools; j++){
	sps = &(ps->ps_splist[j]);
	for(k=0; k < sps->sps_numdevs; k++){
	  pds = &(sps->sps_devlist[k]);
	  printf("pooldevice\t%d\t%d\t%s\t%u\n",
		 j, k, devt_to_devpath(pds->pds_dev), pds->pds_my_weight);
	}
      }
    }
  } 

}

static void print_pool_stats(char *poolname, int interval){
  int			i, fd;
  char			node[256];
  pool_list_t		*pl;
  pool_stat_t		iostat;

  if((pl = get_pool_info(0)) == NULL){
    fprintf(stderr, "Unable to obtain pool list.\n");
    return;
  }

  while(1){
    for(i=0; i < pl->pl_count; i++){
      if(!strcmp(poolname, "all") ||
	 !strcmp(poolname, pl->pl_list[i].ps_name)){
	memset(node, 0, sizeof(node));
	memset(&iostat, 0, sizeof(pool_stat_t));
	strcpy(node, "/dev/pool/");
	strcat(node, pl->pl_list[i].ps_name);
	if ((fd = open(node, O_RDONLY)) < 0) {
	  fprintf(stderr, "Couldn't open pool device %s\n", node);
	  continue;
	}      
	if (ioctl(fd, POOL_DUMP_TOTALSTATS, &iostat) < 0){
	  fprintf(stderr, "Unable to obtain stats on %s.\n", node);
	}
      
	printf("[%10s]", pl->pl_list[i].ps_name);
	printf("  rd: %d  rdblk: %d  wr: %d  wrblk: %d\n", iostat.s_reads, 
	       iostat.s_readblks, iostat.s_writes, iostat.s_writeblks);

	fflush(stdout);

	close(fd);
      }
    }
    if(interval < 1){
      break;
    } else {
      sleep(interval);
    }
  }
}

static void clear_pool_stats(void){
  int			i, fd;
  char			node[256];
  pool_list_t		*pl;
  

  if((pl = get_pool_info(0)) == NULL){
    fprintf(stderr, "Unable to obtain pool list.\n");
    return;
  }

  for(i=0; i < pl->pl_count; i++){
    memset(node, 0, sizeof(node));
    strcpy(node, "/dev/pool/");
    strcat(node, pl->pl_list[i].ps_name);
    if ((fd = open(node, O_RDONLY)) < 0) {
      fprintf(stderr, "Couldn't open pool device %s\n", node);
      continue;
    }      
            
    if (ioctl(fd, POOL_CLEAR_STATS, NULL) < 0){
      fprintf(stderr, "Unable to clear pool stats for %s\n",
	      node);
    }
    close(fd);
  }    
}

static void print_pool_info(char *poolname, int verbose){
  int			i, j, k;
  pool_list_t		*pl;
  pool_status_t		*ps;
  pool_spstatus_t	*sps;
  pool_devstatus_t	*pds;
  float			capacity;
  char			capacity_str[20];

  if((pl = get_pool_info(0)) == NULL){
    printf("No pool information available.\n");
    return;
  }

  if(!verbose){
    printf("Pool Number/Name               Capacity       In use\n"
	   "================               ========       ======\n");
  }
  for(i=0; i < pl->pl_count; i++){
    if(!strcmp(poolname, "all") || 
       !strcmp(pl->pl_list[i].ps_name, poolname)){
      ps = &(pl->pl_list[i]);
      capacity = ps->ps_capacity/2;/* attention -- assumes 512 byte blks */
      if(capacity > 1024){
	capacity /= 1024;
	sprintf(capacity_str, "%6.1f MB", capacity);
      } else {
	sprintf(capacity_str, "%6.1f kB", capacity);
      }
      if(capacity > 1024){
	capacity /= 1024;
	sprintf(capacity_str, "%6.1f GB", capacity);
      }
      if(capacity > 1024){
	capacity /= 1024;
	sprintf(capacity_str, "%6.1f TB", capacity);
      }
      if(capacity > 1024){
	capacity /= 1024;
	sprintf(capacity_str, "%6.1f PB", capacity);
      }

      if(verbose){
	printf("\n"
	       "**********************************************************\n");
	printf("* Pool information:                                      *\n");
	printf("* Pool name           :      %20s        *\n", ps->ps_name);
	printf("* Number of subpools  :      %20u        *\n", ps->ps_num_spools);
	printf("* Total Capacity      :      %20s        *\n", capacity_str);
	printf("* In use              :      %20s        *\n",
	       (ps->ps_in_use) ? "YES" : "NO");
	for (j=0; j < ps->ps_num_spools; j++) {
	  printf("*========================================================*\n");
	  sps = &(ps->ps_splist[j]);
	  printf("* Subpool %-20d                           *\n", j);
	  printf("*   Number of devices :      %20u        *\n", sps->sps_numdevs);
	  printf("*   Stripe size       :      %20u        *\n", sps->sps_stripe);
	  printf("*   Total weight      :      %20u        *\n", sps->sps_tot_weight);
	  printf("*   Type              :      %20s        *\n", get_sptype_name(sps->sps_type));

	  for (k=0; k < sps->sps_numdevs; k++) {
	    pds = &(sps->sps_devlist[k]);
	    printf("* - - - - - - - - - - - - - - - - - - - - - - - - - - - -*\n");
	    pds = &(sps->sps_devlist[k]);
	    printf("*\tDevice %20s                      *\n",
		   devt_to_devpath(pds->pds_dev));
	    printf("*\t  Weight      :      %20u        *\n", pds->pds_my_weight);
	    printf("*\t  Blocks      :      %20"FMT_64"u        *\n", pds->pds_blocks);
	  }
	}
	printf("**********************************************************\n\n");
      } else {
	printf("Pool %d: %-22s%s   %10s\n",
	       i+1,
	       ps->ps_name,
	       capacity_str,
	       (ps->ps_in_use) ? "YES" : "NO");
      }
    }
  }
}

