/******************************************************************************
*******************************************************************************
**
**    Copyright 2001 Sistina Software Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Jonathan Brassow <brassow@sistina.com>
** 
**    Module name: ptool.c
**
**    Description:
**	ptool is a command that is used to write Pool labels to disk based
**	upon a configuration file.
**
**	ptool can also be used for other things relating to pool, like:
**		1. turning on debugging in the pool module in the kernel
**		2. turning off debugging in the pool module in the kernel
**	These uses for ptool will *not* be documented in the man pages
**	because they are not needed for proper use of pool.  So, the only
**	way you will come across their existence is if you are reading
**	this :)
**
*******************************************************************************
******************************************************************************/
#include <fcntl.h>
#include <sys/ioctl.h>

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>

#include "common.h"
#include "pool_ioctl.h"

static void print_usage(void);
static int toggle_debug_flags(char *, char *, int);

int main(int argc, char *argv[])
{
  pool_t	*pool;
  int		c;
  char		*debug_flags = NULL;
  int		debug_switch = 0;
  int		flags = 0;
  int		erase = 0;
  struct option	long_options[] = 
  {
    {"on", 1, NULL, 1},
    {"off", 1, NULL,0}
  };

  if(argc < 2){
    print_usage();
    exit(EXIT_FAILURE);
  }
  while ((c = getopt_long(argc, argv, "efghsv",
			  long_options, NULL)) != -1) {
    switch (c) {
    case 'e':
      erase = 1;
      break;
    case 'f':
      flags |= PTOOLS_FORCE;
      break;
    case 'g':
      printf("'ptool -g' is depricated.  Use 'pgrow' instead.\n");
      exit(EXIT_SUCCESS);
      break;
    case 'h':
      print_usage();
      exit(EXIT_SUCCESS);
      break;
    case 's':
      if(scan_and_print_labels() < 0){
	exit(EXIT_FAILURE);
      } else {
	exit(EXIT_SUCCESS);
      }
      break;
    case 'v':
      flags |= PTOOLS_VERBOSE;
      break;
    case 0:
    case 1:
      debug_flags = argv[optind-1];
      debug_switch = c;
      break;
    default:
      print_usage();
      exit(EXIT_FAILURE);
    }	
  }

  if(!argv[optind]){
    print_usage();
    exit(EXIT_FAILURE);
  }

  if(erase){
    if(erase_pool_labels(argv[optind], flags) < 0){
      exit(EXIT_FAILURE);
    } else {
      exit(EXIT_SUCCESS);
    }
  }

  if(debug_flags){
    if(toggle_debug_flags(argv[optind], debug_flags, debug_switch) < 0){
      fprintf(stderr, "Invalid debug flag.\n");
      exit(EXIT_FAILURE);
    }
    exit(EXIT_SUCCESS);
  }

  if (get_pool_params_from_file(&pool, argv[optind]) < 0) {
    fprintf(stderr, "Error occurred while reading parameter file.\n");
    exit(EXIT_FAILURE);
  }
  
  if(flags & PTOOLS_VERBOSE){
    compute_and_print_lost_blks(pool);
  }

  if(write_pool_labels(pool, flags) < 0){
    exit(EXIT_FAILURE);
  }

  if(flags & PTOOLS_VERBOSE){
    print_pool_t(pool);
  }

  printf("Pool labels successfully written.\n");
  exit(EXIT_SUCCESS);
}


/*
 * print_usage - Print ptool usage
 *
 */
static void print_usage(void){
  printf("Usage: ptool [-dfhqg] <file>\n");
  printf("  <file>     Build storage pool labels according to file.\n");
  printf("  -e         Erase pool labels (use <poolname> in place of <file>.\n");
  printf("  -f         Force label write even if label exists on device.\n");
  printf("  -h         Print this help, then exit.\n");
  printf("  -s         Scan for and show all pool labels.\n");
  printf("  -v         Verbose.\n");
}


/*
 * toggle_debug_flags - turn on/off debugging flags
 * @poolname: name of a pool
 * @mnemonic: which debugging type to turn on (mapping, locking, pool)
 * @switch: 1/0 (on/off)
 *
 * The mnemonics might be somewhat confusing here.  Mapping has to do
 * specifically with those portions of pool that do the mapping.  The
 * reason this is separate is because it produces *alot* of print
 * messages and will slow the kernel greatly.  Otherwise, the mnemonics
 * are intuitive - pool for pool functions, locking for locking functions.
 *
 * Returns: 0 on success, -1 on failure.
 */
static int toggle_debug_flags(char *poolname, char *mnemonic, int swtch){
  int fd;

  struct tmp_struct {
    int str_len;
    char str[20];
  } arg;

  arg.str_len = strlen(mnemonic);
  strncpy(arg.str, mnemonic, 20);

  if((fd = open(poolname, O_RDONLY)) < 0){
    fprintf(stderr, "Failed to open the pool: %s\n", poolname);
    return -1;
  }

  if(swtch){
    printf("Turning on \"%s\" debugging.\n", arg.str);
    if(ioctl(fd,POOL_DEBUG_ON, &arg) < 0){
      fprintf(stderr, "IOCTL failed: could be due to invalid flag.\n");
      close(fd);
      return -1;
    }
  } else {
    printf("Turning off \"%s\" debugging.\n", arg.str);
    if(ioctl(fd,POOL_DEBUG_OFF, &arg) < 0){
      fprintf(stderr, "IOCTL failed: could be due to invalid flag.\n");
      close(fd);
      return -1;
    }
  }

  close(fd);
  return 0;
}
