/******************************************************************************
*******************************************************************************
**
**    Copyright 2000-2001 Sistina Software, Inc.  <nygaard@sistina.com>
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Erling Nygaard <nygaard@sistina.com>
** 
**    Module name: scsi_lib.c
**
*******************************************************************************
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <scsi/scsi.h>
#include <scsi/sg.h>

#include "scsi_lib.h"

#define MIN(x, y) (((x) < (y)) ? (x) : (y))





int scsi_open(char *device)
{
  return(open(device, O_RDWR));
}





int scsi_close(int fd)
{
  return(close(fd));
}


/******************************************************************************
*******************************************************************************
**
** scsi_cmd()
**
** Description:
**   Do a SCSI command
**
** Input(s):
**   fd - File descriptor for the device
**   cmd_size - the size of the CDB
**   cmd_buff - the CDB
**   out_size - the size of the data to be written to the device
**              in the data out phase
**   out_buff - the data to be written in the data out phase
**   in_size - the size of the data to be read from the device
**             in the data in phase
**   in_buff - the data read from the device is stored here
**   sense_size	- the size of sense data
**   sense_buff - the sense to be sent back
**   tag - a unique integer that the command is tagged with
**   quiet - suppress error messages
**
** Returns:
**   The number of bytes read on success, -Exxx on error
**
*******************************************************************************
******************************************************************************/

int scsi_cmd(int fd,
	     unsigned int cmd_size, unsigned char *cmd_buff,
	     unsigned int out_size, unsigned char *out_buff,
	     unsigned int in_size, unsigned char *in_buff,
	     unsigned int sense_size, unsigned char *sense_buff,
	     unsigned int tag, int quiet)
{
  unsigned char *request, *reply;
  struct sg_header *sg_head;
  int value, x;
  


  /*  Check sizes  */

#ifdef SG_BIG_BUFF
  if (sizeof(struct sg_header) + cmd_size + out_size > SG_BIG_BUFF ||
      sizeof(struct sg_header) + in_size > SG_BIG_BUFF)
  {
    if (!quiet)
      fprintf(stderr, "scsi_cmd():  too big a transfer\n");
    return(-1);
  }
#else  /*  SG_BIG_BUFF  */
  if (sizeof(struct sg_header) + cmd_size + out_size > 4096 ||
      sizeof(struct sg_header) + in_size > 4096)
  {
    if (!quiet)
      fprintf(stderr, "scsi_cmd():  too big a transfer\n");
    return(-1);
  }
#endif  /*  SG_BIG_BUFF  */
 


  /*  Get memory  */
  
  if (!(request = malloc(sizeof(struct sg_header) + cmd_size + out_size)))
  {
    if (!quiet)
      fputs("scsi_cmd():  can't get memory\n", stderr);
    return(-ENOMEM);
  }

  if (!(reply = malloc(sizeof(struct sg_header) + in_size)))
  {
    free(request);
    if (!quiet)
      fputs("scsi_cmd():  can't get memory\n", stderr);
    return(-ENOMEM);
  }



  /*  Fill in request  */

  sg_head = (struct sg_header *)request;

  memset(sg_head, 0, sizeof(struct sg_header));
  sg_head->pack_len = sizeof(struct sg_header) + cmd_size + out_size;
  sg_head->reply_len = sizeof(struct sg_header) + in_size;
  sg_head->pack_id = tag;
  sg_head->twelve_byte = (cmd_size == 12);

  memcpy(request + sizeof(struct sg_header), cmd_buff, cmd_size);

  if (out_buff)
    memcpy(request + sizeof(struct sg_header) + cmd_size, out_buff, out_size);



  /*  Write out request  */

  value = write(fd, request, sizeof(struct sg_header) + cmd_size + out_size);



  if (value < 0)
  {
    free(request);
    free(reply);
    if (!quiet)
      fprintf(stderr, "scsi_cmd():  write returned %s (%d)\n", 
	      strerror(errno), value);
    return(value);
  }

  if (!quiet && value != sizeof(struct sg_header) + cmd_size + out_size)
    fprintf(stderr, "scsi_cmd():  Return from write is %d (not %u)\n",
	    value,
	    (unsigned int)(sizeof(struct sg_header) + cmd_size + out_size));



  /*  Read in reply  */

  value = read(fd, reply, sizeof(struct sg_header) + in_size);

 
  if (value < 0)
  {
    free(request);
    free(reply);
    if (!quiet)
      fprintf(stderr, "scsi_cmd():  read returned %s (%d)\n", 
	      strerror(errno), value);
    return(value);
  }
 
  sg_head = (struct sg_header *)reply;
 
  if (sg_head->result || sg_head->sense_buffer[0])
  {
    free(request);
    free(reply);
    if (!quiet)
    {
      /*fprintf(stderr, "scsi_cmd():  result = %d\n", sg_head->result);
       *fprintf(stderr, "scsi_cmd():  sense = ");
       *for (x = 0; x < sizeof(sg_head->sense_buffer); x++){
       *fprintf(stderr, "%.2X ", sg_head->sense_buffer[x]);
       *sense_buff[x] = sg_head->sense_buffer[x];
       *}
       *fprintf(stderr, "\n");
       */
      if(sense_size)
	for(x=0;x<sense_size;x++)
	 sense_buff[x] = sg_head->sense_buffer[x];	
    }
    return(-1);
  }

  if (in_buff)
    memcpy(in_buff, reply + sizeof(struct sg_header), MIN(in_size, value - sizeof(struct sg_header)));


  free(request);
  free(reply);

  return(value - sizeof(struct sg_header));
}


