#define _POSIX_SOURCE
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <getopt.h>
#include <sysexits.h>
#include <errno.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>

int verbose = 0;

char *get_real_dst(char *src, char *dst)
{
	char *fn = strrchr(src, '/');
	char *rd;

	if (fn == NULL)
		fn = src;

	rd = malloc(strlen(dst) + strlen(fn) + 2);
	if (rd == NULL) {
		fprintf(stderr, "OOM: get_real_dst\n");
		exit(EX_TEMPFAIL);
	}

	strcpy(rd, dst);
	if (*(dst + strlen(dst) - 1) != '/')
		strcat(rd, "/");
	strcat(rd, fn);

	return rd;
}

void put_real_dst(char *dst)
{
	free(dst);
}

void do_copy(char *src, char *dst, int isdir)
{
	char *realdst = dst;
	int s_fd, d_fd;
	struct stat sbuf;
	unsigned char zero = 0;
	char *in, *out;

	if (isdir)
		realdst = get_real_dst(src, dst);

	if (verbose)
		printf("%s -> %s\n", src, realdst);

	if ((s_fd = open(src, O_RDONLY | O_NOCTTY)) < 0) {
		perror(src);
		goto out;
	}

	if (stat(src, &sbuf) < 0) {
		perror(src);
		close(s_fd);
		goto out;
	}

	if ((d_fd = open(realdst, O_RDWR | O_CREAT | O_NOCTTY | O_TRUNC, S_IRWXU | S_IRWXG)) < 0) {
		perror(dst);
		close(s_fd);
		goto out;
	}

	if (sbuf.st_size  == 0) {
		close(s_fd);
		close(d_fd);
		goto out;
	}

	if (lseek(d_fd, sbuf.st_size - 1, SEEK_SET) < 0) {
		perror("lseek");
		close(s_fd);
		close(d_fd);
		goto out;
	}

	if (write(d_fd, &zero, 1) < 0) {
		perror("write");
		close(s_fd);
		close(d_fd);
		goto out;
	}

	if ((in = mmap(NULL, sbuf.st_size, PROT_READ, MAP_SHARED, s_fd, 0)) == MAP_FAILED) {
		perror("mmap source");
		close(s_fd);
		close(d_fd);
		goto out;
	}

	close(s_fd);

	if ((out = mmap(NULL, sbuf.st_size, PROT_READ|PROT_WRITE, MAP_SHARED, d_fd, 0)) == MAP_FAILED) {
		perror("mmap destination");
		munmap(in, sbuf.st_size);
		close(d_fd);
		goto out;
	}

	close(d_fd);

	memcpy(out, in, sbuf.st_size);

	munmap(in, sbuf.st_size);
	munmap(out, sbuf.st_size);

out:
	if (isdir)
		put_real_dst(realdst);
}

void usage(void)
{
	fprintf(stderr, "Usage:\n"
			"mmc [opts] <srcfile> [srcfile..] <dstdir>\n"
			"mmc [opts] <srcfile> <dstfile>\n");
}


int main(int argc, char *argv[])
{
	int opt;
	char *dst;
	struct stat sbuf;

	while ((opt = getopt(argc, argv, "h?v")) != EOF) {
		switch(opt) {
			case 'v':
				verbose++;
				break;
			case 'h':
			case '?':
				usage();
				return(EX_USAGE);
		}
	}

	if ((argc - optind) < 2) {
		usage();
		return(EX_USAGE);
	}

	dst = argv[optind + (argc - optind) - 1];
	argc--;

	if (stat(dst, &sbuf) < 0) {
		if (errno != ENOENT) {
			perror("stat");
			return(EX_UNAVAILABLE);
		}
	} else {
		if (!S_ISDIR(sbuf.st_mode)) {
			if ((argc - optind) != 1) {
				fprintf(stderr, "Refusing to overwrite \"%s\"\n", dst);
				usage();
				exit(EX_USAGE);
			}
			if (!S_ISREG(sbuf.st_mode)) {
				fprintf(stderr, "Destination \"%s\" is not a regular file\n", dst);
				usage();
				exit(EX_USAGE);
			}
		}
	}

	while(optind < argc) {
		do_copy(argv[optind++], dst, S_ISDIR(sbuf.st_mode));
	}

	return(EX_OK);
}


