/******************************************************************************
 *    Copyright 2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file CONTRIBUTORS for a list of contributors.
 *
 *    This file is maintained by:
 *      Michael C Tilstra <conrad@sistina.com>
 */
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <readline/readline.h>
#include <readline/history.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include "global.h"
#include "memexp.h"

#include "lsm_info.h"

#undef die
#define die(msg, args...) {fprintf(stderr, msg, ## args);exit(1);}

/**
 * show_error - printout the errorcode
 * @rbuf: 
 * 
 * 
 * Returns: void
 */
void show_error(u8 *rbuf)

{
   printf(" Error received: ");
#define CaseError(x) case (x): printf("%s\n",#x); break
   switch ( *rbuf ) {
      CaseError(Mel_Err_BadSize);
      CaseError(Mel_Err_BidZero);
      CaseError(Mel_Err_NoBid);
      CaseError(Mel_Err_BadSeq);
      CaseError(Mel_Err_BadPbn);
      CaseError(Mel_Pbn_Space_Full);
      CaseError(Mel_You_are_Banned);
      CaseError(Mel_Verify_Failed);
      CaseError(Mel_Err_NotImpl);
      default: printf("Unknown errorcode %d\n", *rbuf);
   }
#undef CaseError
}

typedef enum {
   apcalc,
   gnuplot
} stat_ouptu_style;

#ifdef LPS_CODE
/**
 * lps_print_vals - print out info based on the lps data
 * @lps: 
 * 
 * 
 * Returns: void
 */
void lps_print_vals(lps_t *lps, stat_ouptu_style sop)
{
   u64 devi = 0;
   u64 part1;/* These parts may have to be 64bit to avoid overflow.*/
   u64 part2;
   u64 part3;
   u64 mean=0;

   part1 = (u64)lps->count * (u64)lps->std;
   part2 = (u64)lps->total * (u64)lps->total;
   part3 = (u64)lps->count * ((u64)lps->count -(u64)1);
   if(part3 != 0)
      devi  = (part1 - part2) / part3;
   if(lps->count !=0)
         mean = lps->total / lps->count;

   switch(sop){
   case apcalc:
      printf("usec per lock op:\n");
      printf(" Min: %"FMT_64"u  Max: %"FMT_64"u  Mean: %"FMT_64"u\n",
            lps->min, lps->max, mean);
      printf(" Called %"FMT_64"u times, with total rate %"FMT_64"u\n",
            lps->count, lps->total);
      printf(" Squared Std Deviation: %"FMT_64"u\n", devi);
      printf("  (%"FMT_64"u - %"FMT_64"u) / %"FMT_64"u\n",
            part1, part2, part3);
      printf(" 99%%conf = (2.576 * sqrt(%"FMT_64"u))/sqrt(%"FMT_64"u)\n",
            devi, lps->count);
      break;
   case gnuplot:
      printf("set data style lines\n");
      printf("ssd = %"FMT_64"u\n", devi);
      printf("mu = %"FMT_64"u\n", mean);
      printf("nd(x) = (1/sqrt(2*pi*ssd))*exp((-(x-mu)**2)/(2*ssd))\n");
      printf("plot nd(x) \n");
      break;
   default:
      printf("bad code\n");
   }

}
#endif /*LPS_CODE*/

#ifdef LDUR_CODE
/**
 * ldur_print_vals - Prints lock op durration statitics.
 * @ldur: < stuct we got from server.
 */
void ldur_print_vals(ldur_t *ldur, stat_ouptu_style sop)
{
   u64 devi = 0;
   u64 part1;/* These parts may have to be 64bit to avoid overflow.*/
   u64 part2;
   u64 part3;
   u64 mean=0;

   part1 = (u64)ldur->count * (u64)ldur->std;
   part2 = (u64)ldur->total * (u64)ldur->total;
   part3 = (u64)ldur->count * ((u64)ldur->count -(u64)1);
   if(part3 != 0)
      devi  = (part1 - part2) / part3;
   if(ldur->count !=0)
         mean = ldur->total / ldur->count;

   switch(sop) {
   case apcalc:
      printf("Lock op durration values:\n");
      printf(" Min: %"FMT_64"u  Max: %"FMT_64"u  Mean: %"FMT_64"u\n",
            ldur->min, ldur->max, mean);
      printf(" Called %"FMT_64"u times, with total rate %"FMT_64"u\n",
            ldur->count, ldur->total);
      printf(" Squared Std Deviation: %"FMT_64"u\n", devi);
      printf("  (%"FMT_64"u - %"FMT_64"u) / %"FMT_64"u\n",
            part1, part2, part3);
      printf(" 99%%conf = (2.576 * sqrt(%"FMT_64"u))/sqrt(%"FMT_64"u)\n",
            devi, ldur->count);
      break;
   case gnuplot:
      printf("set data style lines\n");
      printf("ssd = %"FMT_64"u\n", devi);
      printf("mu = %"FMT_64"u\n", mean);
      printf("nd(x) = (1/sqrt(2*pi*ssd))*exp((-(x-mu)**2)/(2*ssd))\n");
      printf("plot nd(x) \n");
      break;
   default:
      printf("bad code\n");
   }

}
#endif /*LDUR_CODE*/

/**
 * requestStat - send request to server
 * @sk: < socket file descriptor
 * @sel: < selector code
 * 
 * Also grabs the return header, and handles errors server may send.
 * 
 * Returns: int
 */
int requestStat(int sk, char sel)
{
   char buf[sizeof(me_hdr_t)+sizeof(u8)];
   me_hdr_t *hdr = (me_hdr_t*)buf;
   char *sl = (char*)(buf+sizeof(me_hdr_t));

   hdr->Opcode = Mel_StatReq;
   hdr->Key = 4242;
   hdr->length = cpu_to_be16(sizeof(u8));
   *sl = sel;

   if(write(sk, buf, (sizeof(me_hdr_t)+sizeof(u8))) <0) return -1;
   if(read(sk, buf, sizeof(me_hdr_t)) <0) return -1;

   if( hdr->Opcode == Mel_Error ) {
      read(sk, buf, sizeof(errorcode_t));
      show_error(buf);
      return -1;
   }

   return 0;
}

/******************************************************************************
 * main
 *   prgm [] [-p port] ip
 */
int main(int argc, char **argv)
{
   int sock, n;
   struct sockaddr_in adr;
   me_hdr_t hdr;
   int c;
   extern char *optarg;
   extern int optind;
   char gethashstats = FALSE;
   char getcounts = FALSE;
   char getlps = FALSE;
   char getldur = FALSE;

   memset(&adr, 0, sizeof(struct sockaddr_in));
   adr.sin_family = AF_INET;
   adr.sin_port = htons( 40046 );

   while((c = getopt(argc, argv, "scldp:h?")) != -1) {
      switch(c) {
         case 's': gethashstats = TRUE; break;
         case 'c': getcounts = TRUE; break;
         case 'l': getlps = TRUE; break;
         case 'd': getldur = TRUE; break;
         case 'p':
            adr.sin_port = htons(atoi(optarg));
            break;
         case 'h':
         case '?':
         default:
            printf("Usage: %s [] [-p port] ip\n", argv[0]);
            exit(1);
      }
   }
   
   if( optind >= argc ) {
      printf("Need to specify an ip address.... (%d)(%d)\n",optind, argc);
      exit(1);
   }

   /* probably do gethostbyname for this. */
   adr.sin_addr.s_addr = inet_addr( argv[optind] );


   if( (sock = socket(AF_INET, SOCK_STREAM, 0)) <0)
      die("Failed to create socket.\n");

   if( connect(sock, (struct sockaddr*)&adr, sizeof(struct sockaddr_in))<0)
      die("Could not connect.\n");


   /* ok, always grab modepage, then grab what ever others they ask for. */
   {
      modepage_t mp;
      memset(&mp, 0, sizeof(modepage_t));

      hdr.Opcode = Mel_ModePgR;
      hdr.Key = 0x6869;
      hdr.length = cpu_to_be16(sizeof(modepage_t));

      n = write(sock, &hdr, sizeof(me_hdr_t));
      if(n<0) die("write error: %m\n");
      n = write(sock, &mp, sizeof(modepage_t));
      if(n<0) die("write error: %m\n");

      n = read(sock, &hdr, sizeof(me_hdr_t));
      if(n<0) die("read error: %m\n");
      n = read(sock, &mp, sizeof(modepage_t));
      if(n<0) die("read error: %m\n");

      printf("ModePage returns:\n  Datasize=%d\n  Number of buffers=%"FMT_64"d\n",
            be32_to_cpu(mp.datasize), be64_to_cpu(mp.numbufs));
   }

   if(gethashstats) {
      hshcnt_t hshcnt;
      int i;

      if( requestStat(sock, STAT_HSHCNT) <0) die("error. %m\n");

      n = read(sock, &hshcnt, sizeof(hshcnt_t));
      if(n<0) printf("read got n=%d\n",n);
      printf("hash size : number of buckets at this size\n");
      for(i=0;i<MAXdens;i++) {
         printf("  %.2u: %7u   ", i, be16_to_cpu(hshcnt.bcnt[i]));
         if( i % 4 ==3 ) printf("\n");
      }
      printf(" There are %u memory buffers in use\n",
            be32_to_cpu(hshcnt.inuse));
      printf(" There are %u keys in the hashtable\n",
            be32_to_cpu(hshcnt.numkeys));
      printf(" There is a maximum of %u elements.\n",
            be32_to_cpu(hshcnt.maxelems));
   }
   if(getcounts) {
      counters_t cnts;

      if( requestStat(sock, STAT_counts) <0) die("error. %m\n");

      if( read(sock, &cnts, sizeof(counters_t)) <0) die("Read error:%m\n");
      printf("There have been:\n");
      printf("   loads : %u\n", be32_to_cpu(cnts.loads));
      printf("   stores: %u\n", be32_to_cpu(cnts.stores));
      printf("   sucstr: %u\n", be32_to_cpu(cnts.sucstr));
      printf("   verify: %u\n", be32_to_cpu(cnts.verify));
      printf("   dumps : %u\n", be32_to_cpu(cnts.dumps));
      printf("   bans  : %u\n", be32_to_cpu(cnts.bans));
      printf("   unbans: %u\n", be32_to_cpu(cnts.unbans));
      printf("   errors: %u\n", be32_to_cpu(cnts.errors));
   }
   if(getlps) {
#ifdef LPS_CODE
      lps_t lps;

      if( requestStat(sock, STAT_LPS) <0) die("error. %m\n");

      if( read(sock, &lps, sizeof(lps_t)) <0) die("Read error:%m\n");

      lps.min = be64_to_cpu(lps.min);
      lps.max = be64_to_cpu(lps.max);
      lps.std = be64_to_cpu(lps.std);
      lps.total = be64_to_cpu(lps.total);
      lps.count = be64_to_cpu(lps.count);

      lps_print_vals(&lps, apcalc);
#else
      printf("LPS support not compiled in.\n");
#endif /*LPS_CODE*/
   }
   if(getldur) {
#ifdef LDUR_CODE
      ldur_t ld;

      if( requestStat(sock, STAT_LDUR) <0) die("error. %m\n");

      if( read(sock, &ld, sizeof(ldur_t)) <0) die("Read error:%m\n");

      ld.min = be64_to_cpu(ld.min);
      ld.max = be64_to_cpu(ld.max);
      ld.std = be64_to_cpu(ld.std);
      ld.total = be64_to_cpu(ld.total);
      ld.count = be64_to_cpu(ld.count);

      ldur_print_vals(&ld, apcalc);
#else
      printf("LDUR support not compiled in.\n");
#endif /*LDUR_CODE*/
   }

   return 0;
}
